
#
# This work is licensed under THE ANTI-CAPITALIST SOFTWARE LICENSE (v 1.4) 
# To view a copy of this license, visit https://directory.fsf.org/wiki/License:ANTI-1.4

from numpy import dot, outer, argsort, array, sin, cos, hstack, sum as npsum, square, diagonal, subtract, cross, argmin, round as npround, atleast_2d, vstack, reshape, shape, unique, where, floor
from numpy.linalg import norm
from scipy.optimize import linear_sum_assignment
from plotly import graph_objects as go
from plotly.io import to_html

from .scsd_symmetry import *

#copy from scsd
s2c = lambda th1, ph1: array((sin(ph1) * cos(th1), sin(ph1) * sin(th1), cos(ph1)))
c2q = lambda om, vec: hstack([array([cos(om/2)]), (vec * sin(om/2))])
s2q = lambda th, ph, om: array([sin(om/2)*sin(ph)*cos(th), sin(om/2)*sin(ph)*sin(th), sin(om/2)*cos(ph), cos(om/2)])
uq2r = lambda x,y,z,w:((1-2*(y**2 + z**2), 2*(x*y - w*z), 2*(x*z + w*y)),(2*(x*y + w*z), 1-2*(x**2 + z**2), 2*(y*z - w*x)),(2*(x*z - w*y), 2*(y*z + w*x), 1-2*(x**2 + y**2)))
q2uq = lambda x,y,z,w: [x,y,z,w]/norm([x,y,z,w])
q2r  = lambda x,y,z,w: uq2r(*q2uq(x,y,z,w))
costmat_gen = lambda m1,m2: npsum(square(diagonal(subtract.outer(m2,m1),0,1,3)),axis = -1)

def assign_mat_for_symm(model, typo, theta, phi, rotation):
    omega     = (2*pi)/rotation
    refvec    = s2c(theta, phi)
    rot_mat   = q2r(*s2q(theta, phi, omega))
    trans_atoms = model.copy()[:,:3]

    if typo in ['improperrotation','mirror']:
        trans_atoms = trans_atoms-outer(2*dot(trans_atoms, refvec),refvec)
    if typo in ['rotation','improperrotation']:
        trans_atoms = dot(rot_mat,trans_atoms.T).T
    if typo == 'inversion':
        trans_atoms = -model[:,:3]

    costmat  = costmat_gen(model[:,:3],trans_atoms)
    row_ind, col_ind = linear_sum_assignment(costmat)
    return col_ind.argsort().tolist()

def return_assign_mat(model, point_group_name):
    ops, pgt, ops_order = [x.get(point_group_name.capitalize()) for x in [point_group_dict,pgt_dict,ordered_ops_dict]]
    ops_vals = [operations_dict.get(x) for x in ops_order]
    return {z:[assign_mat_for_symm(model,*x) for x in y] for z,y in zip(ops_order,ops_vals)}

def generate_superlist(pairs,strop):
    for i,pair in enumerate(pairs):
        for j,at in enumerate(strop):
            if at == pair[0]:
                strop.insert(j,pair[0])
                strop.insert(j+1,pair[1])
                pairs.pop(i)
                return pairs,strop
    return pairs,strop

def generate_line_thru_points(ats,maxdist = 1.75):
    pairs = [[[i,j] for i,x in enumerate(ats) if ((norm(x-y)<maxdist) and (i>j))] for j,y in enumerate(ats)]
    pairs =  [y for x in pairs for y in x]
    strop = pairs[0]
    for x in range(100):
        pairs,strop = generate_superlist(pairs,strop)
        pairs,strop = generate_superlist([x[::-1] for x in pairs],strop)
    return strop

class scsd_model:
    def __init__(self, name, atom_positions, point_group_name, assign_mat=None, pca=None,  database_path=None, maxdist = 1.75,
                    mondrian_limits = None):
        self.name = name
        self.ats = atom_positions
        self.ats_3 = self.ats[:,:3].astype(float)
        self.ats_3_round_4 = npround(self.ats[:,:3].astype(float),4)
        if len(self.ats[0]) == 4:
            self.ats_round_4 = hstack((self.ats_3_round_4,atleast_2d(self.ats[:,3]).T))
        else:
            self.ats_round_4 = self.ats_3_round_4
        self.ptgr = point_group_name.capitalize()
        self.symm = scsd_symmetry(self.ptgr)
        self.pca = pca
        self.assign_mat = assign_mat
        self.database_path = database_path
        self.maxdist = maxdist
        self.mondrian_limits = mondrian_limits
        self.ptgr_html = symm_typog_html.get(self.ptgr,self.ptgr)

    def gen_assign_mat(self):
        self.assign_mat = return_assign_mat(self.ats_3, self.ptgr)

    def __repr__(self):
        not_Nones = [x for x,y in zip('name,ats,ptgr,pca,assign_mat,database_path,maxdist,mondrian_limits'.split(','),[self.name ,self.ats,
        self.ptgr,self.pca,self.assign_mat,self.database_path, self.maxdist, self.mondrian_limits]) if not isinstance(y,type(None))]
        return "scsd_model: " + self.name + "; attributes: " + ', '.join(not_Nones)

    def importable(self):
        l1 = [f'''model_objs_dict['{self.name}'] = scsd_model('{self.name}', array({str(self.ats_round_4.tolist())}), '{self.ptgr}' ''']
        if type(self.pca) is not type(None):
            l1.append(', pca = {'+', '.join([f'"{k}":array({str(v.round(5).tolist())})' for k,v in self.pca.items()])+'}')
        if type(self.database_path) is not type(None):
            l1.append(f", database_path = '{self.database_path}'")
        if type(self.maxdist) is not type(None):
            l1.append(f", maxdist = {self.maxdist}")
        if type(self.mondrian_limits) is not type(None):
            l1.append(f", mondrian_limits = {self.mondrian_limits}")
        l1.append(')')
        return ''.join(l1)

        #if type(self.pca) is type(None):
        #    return "model_objs_dict['{}'] = scsd_model('{}', array({}), '{}')\n".format(self.name, self.name, str(self.ats_round_4.tolist()),self.ptgr)
        #else:
        #    return f'''model_objs_dict['{self.name}'] = scsd_model('{self.name}', array({str(self.ats_round_4.tolist())}), '{self.ptgr}', pca = {'{'+', '.join([f'"{k}":array({str(v.round(5).tolist())})' for k,v in self.pca.items()])+'}' if type(self.pca) is not None else 'None'}, database_path = {"'"+self.database_path+"'" if type(self.database_path) == str else 'None'}, maxdist = {self.maxdist if type(self.maxdist) in [int,float] else 'None'}, mondrian_limits = {self.mondrian_limits if type(self.mondrian_limits) is not None else 'None'} )'''
#
        #  maxdist = {self.maxdist if type(self.maxdist) in [int,float] else 'None'}, mondrian_limits = {self.mondrian_limits if type(self.mondrian_limits) is not None else 'None'} )'''

    def html_table(self):
        return "<table border='1' id = 'html_table'>"+'\n'.join(['<tr>{}</tr>'.format(''.join(['<td>{}</td>'.format(value) for value in row])) for row in vstack((atleast_2d(['x','y','z','at'][:len(self.ats_round_4[0])]),self.ats_round_4.tolist()))]) + '</table>'

    def plotly_plot_model(self):
        line_thru = generate_line_thru_points(self.ats_3, self.maxdist)

        if len(self.ats.T) == 3:
            a = self.ats_3[line_thru].T
            b = self.ats_3.T
            fig = go.Figure()
            fig.add_trace(go.Scatter3d(x=b[0], y=b[1],z = b[2], mode = 'markers',name = 'Model',legendgroup = 'Model'))
            fig.add_trace(go.Scatter3d(x=a[0], y=a[1],z = a[2], mode = 'lines',showlegend = False,legendgroup = 'Model', line = dict(color = 'black')))
        if len(self.ats.T) == 4:
            at_colors = {'C':'gray', 'N':'cornflowerblue', 'S': 'goldenrod', 'O':'red', 'H':'floralwhite','P':'brown'}
            a,b = self.ats_3[line_thru].T, self.ats.T
            fig = go.Figure()
            for index, atype in enumerate(unique(b[3])):
                d = b[:3,where(b[3]==atype)[0]].astype(float)
                if atype in at_colors.keys():
                    fig.add_trace(go.Scatter3d(x=d[0], y=d[1],z = d[2], mode = 'markers', marker = {'color':at_colors.get(atype)},  name = atype,legendgroup = 'Model'))
                else:
                    fig.add_trace(go.Scatter3d(x=d[0], y=d[1],z = d[2], mode = 'markers', name = atype, legendgroup = 'Model'))
            fig.add_trace(go.Scatter3d(x=a[0], y=a[1],z = a[2], mode = 'lines',showlegend = False,legendgroup = 'Model', line = dict(color = 'black')))
        fig.layout.scene.camera.projection = dict(type = 'orthographic')
        fig.layout.scene.aspectmode = 'data'

        return fig

    def visualize_symm_ops(self, as_type = 'html', symm_ops = None): #
        self.symm = scsd_symmetry(self.ptgr)
        ops_verb = [y for x in self.symm.ops_verb_order for y in x]
        if type(symm_ops) == str:
            ops_verb = [y for x in [z for z,a in zip(self.symm.ops_verb_order, self.symm.pgt.get(symm_ops)) if (a == 1)] for y in x]
        if type(symm_ops) == list:
            if symm_ops[0] == 1:
                ops_temp = [z for i, z in enumerate(self.symm.ops_verb_order) if symm_ops[i]==1]
            else:
                ops_temp = [z for i, z in enumerate(self.symm.ops_verb_order) if all([self.symm.pgt.get(b)[i]==1 for b in symm_ops])]
            ops_verb = [y for x in ops_temp for y in x]

        fig = self.plotly_plot_model()
        scale = max(self.ats_3.flatten())

        for op in ops_verb[1:]:
            typo, th, ph, n_rot = op
            if typo.lower() == 'rotation':
                x,y,z = s2c(th,ph) * scale*1.2
                fig.add_trace(go.Scatter3d(x=[-x,x],y=[-y,y],z=[-z,z], marker = dict(symbol = 'circle-open'), name = str(n_rot)+'-fold rotation'))
            if typo.lower() == 'improperrotation':
                x,y,z = s2c(th,ph) * scale*1.1
                fig.add_trace(go.Scatter3d(x=[-x,x],y=[-y,y],z=[-z,z], marker = dict(symbol = 'circle-open'), name = str(n_rot)+'-fold improper rotation'))
            if typo.lower() == 'mirror':
                x1,y1,z1 = s2c(th,ph+(pi/2)) * scale
                x2,y2,z2 = cross(s2c(th,ph+pi/2),s2c(th,ph)) * scale
                x3,y3,z3 = x1+x2,y1+y2,z1+z2
                x4,y4,z4 = x1-x2,y1-y2,z1-z2
                x,y,z = [x3,x4,-x3,-x4,x3],[y3,y4,-y3,-y4,y3],[z3,z4,-z3,-z4,z3]
                sax = argmin([norm(v) for v in [x,y,z]])
                fig.add_trace(go.Scatter3d(x=x,y=y,z=z, name = 'Mirror', mode = 'lines', surfaceaxis = sax, surfacecolor = "rgba(150,50,50,0.2)"))
            if typo.lower() == 'inversion':
                x,y,z = [0],[0],[0]
                fig.add_trace(go.Scatter3d(x=x,y=y,z=z, name = 'Inversion'))
        fig.layout.scene.camera.projection = dict(type = 'orthographic')
        fig.layout.scene.aspectmode = 'data'
        if as_type.lower() == 'html':
            return to_html(fig)
        elif as_type.lower() == 'fig':
            return fig

    def visualize_principal_component(self, irrep, mode_number, magnitude = 1.0, as_type = 'fig', color = 'black'): #, symm_ops = None

        traces = []

        a0 = self.ats_3
        a1 = (self.ats_3 + (array(self.pca.get(irrep)[mode_number - 1]).astype(float).reshape(shape(self.ats_3)) * magnitude))
        for xyz0,xyz1 in zip(a0,a1):
            x0,y0,z0 = xyz0
            x1,y1,z1 = xyz1
            traces.append(go.Scatter3d(x=[x0,x1],y=[y0,y1],z=[z0,z1], mode = 'lines', line = {'color':color},legendgroup = irrep +'('+str(mode_number)+')', name = irrep +'('+str(mode_number)+')'))
        x,y,z = a1.T
        u,v,w = (a1-a0).T/5
        traces.append(go.Cone(x=x,y=y,z=z, u=u,v=v,w=w, colorscale = [color,color], showscale=False, legendgroup = irrep +'('+str(mode_number)+')'))
        if as_type == 'traces':
            return traces
        fig = self.plotly_plot_model()
        [fig.add_traces(x) for x in traces]
        fig.layout.scene.camera.projection = dict(type = 'orthographic')
        fig.layout.scene.aspectmode = 'data'
        fig.update_layout(showlegend=False)

        if as_type.lower() == 'fig':
            return fig
        elif as_type.lower() == 'html':
            return to_html(fig)

    def visualize_symm_and_pc(self, as_type = 'html', symm_ops = None): #
        fig = self.visualize_symm_ops(as_type = 'fig')

        if type(self.pca) is not type(None):
            scale_pca = floor((max(self.ats_3.flatten())**2)/4)
            line_thru = generate_line_thru_points(self.ats_3, self.maxdist)

            for k,v in self.pca.items():
                if len(v)>0:
                    ats = self.ats_3 + array(v[0]).reshape(shape(self.ats_3))*scale_pca
                    x,y,z = ats[line_thru].T
                    tracename = irrep_typog_html.get(k,k) + ' (1)'
                    fig.add_trace(go.Scatter3d(x=x,y=y,z=z, name = tracename, visible = 'legendonly'))
                if len(v)>1:
                    ats = self.ats_3 + array(v[1]).reshape(shape(self.ats_3))*scale_pca
                    x,y,z = ats[line_thru].T
                    tracename = irrep_typog_html.get(k,k) + ' (2)'
                    fig.add_trace(go.Scatter3d(x=x,y=y,z=z, name = tracename, visible = 'legendonly'))

        fig.layout.scene.camera.projection = dict(type = 'orthographic')
        fig.layout.scene.aspectmode = 'data'
        if as_type.lower() == 'html':
            return to_html(fig)
        elif as_type.lower() == 'fig':
            return fig

    def headbox(self, filename):
        return "SCSD model generated from {} using {} symmetry, available as <i>{}</i>".format(filename,symm_typog_html.get(self.ptgr,self.ptgr), self.name)

    def chem_formula(self):
        if len(self.ats[0]) == 4:
            uq = unique(self.ats[:,3], return_counts = True)
            return ''.join([f'{a}<sub>{str(b)}</sub>' for a,b in zip(*uq)])
        else: return ''

model_ptpop = array([['-2.998393989496222', '-1.2048999389471107', '-2.0868654036021708', 'O'], ['-2.998393989496222', '1.204899938947111', '-2.0868654036021708', 'O'], ['-1.2048999389471107', '2.998393989496222', '-2.0868654036021708', 'O'], ['1.2048999389471111', '2.998393989496222', '-2.0868654036021708', 'O'], ['2.998393989496222', '1.204899938947111', '-2.0868654036021708', 'O'], ['2.998393989496222', '-1.2048999389471107', '-2.0868654036021708', 'O'], ['1.2048999389471107', '-2.998393989496222', '-2.0868654036021708', 'O'], ['-1.204899938947111', '-2.998393989496222', '-2.0868654036021708', 'O'], ['-2.998393989496222', '-1.204899938947111', '2.0868654036021708', 'O'], ['-2.998393989496222', '1.2048999389471107', '2.0868654036021708', 'O'], ['-1.2048999389471107', '2.998393989496222', '2.0868654036021708', 'O'], ['1.204899938947111', '2.998393989496222', '2.0868654036021708', 'O'], ['2.998393989496222', '1.2048999389471107', '2.0868654036021708', 'O'], ['2.998393989496222', '-1.204899938947111', '2.0868654036021708', 'O'], ['1.2048999389471107', '-2.998393989496222', '2.0868654036021708', 'O'], ['-1.2048999389471111', '-2.998393989496222', '2.0868654036021708', 'O'], ['-2.8586621306199405', '-2.3501016759829083e-17', '-1.8488927466117464e-32', 'O'], ['1.9854358816520224e-16', '2.8586621306199405', '2.4754757847413967e-16', 'O'], ['2.8586621306199405', '2.3501016759829083e-17', '1.8488927466117464e-32', 'O'], ['-1.9854358816520224e-16', '-2.8586621306199405', '-2.4754757847413967e-16', 'O'], ['9.360081346632462e-17', '2.3201706607295787', '1.4741278320302014', 'P'], ['2.3201706607295787', '-2.0462311592884027e-16', '1.474127832030201', 'P'], ['-2.705132701156106e-16', '-2.3201706607295787', '1.474127832030201', 'P'], ['-2.3201706607295787', '4.84686651905793e-17', '1.4741278320302014', 'P'], ['-2.3201706607295787', '2.0462311592884027e-16', '-1.474127832030201', 'P'], ['2.705132701156106e-16', '2.3201706607295787', '-1.474127832030201', 'P'], ['2.3201706607295787', '-4.84686651905793e-17', '-1.4741278320302014', 'P'], ['-9.360081346632462e-17', '-2.3201706607295787', '-1.4741278320302014', 'P'], ['3.0814879110195774e-33', '1.268463646805813e-16', '-1.4648129517334991', 'PT'], ['-3.0814879110195774e-33', '-1.268463646805813e-16', '1.4648129517334991', 'PT']])
model_tmbodipy = array([[0.0, 0.0, -1.3471, 'B'], [-0.0, 2.5613, 1.4257, 'C'], [-0.0, 3.3453, 0.3011, 'C'], [0.0, 2.5272, -0.8254, 'C'], [0.0, -2.5272, -0.8254, 'C'], [-0.0, -3.3453, 0.3011, 'C'], [-0.0, -2.5613, 1.4257, 'C'], [-0.0, -0.0, 1.6366, 'C'], [-0.0, 1.215, 0.9613, 'C'], [-0.0, -1.215, 0.9613, 'C'], [-0.0, 3.0614, 2.8336, 'C'], [0.0, 2.9387, -2.2587, 'C'], [0.0, -2.9387, -2.2587, 'C'], [-0.0, -3.0614, 2.8336, 'C'], [-1.1306, 0.0, -2.1482, 'F'], [1.1306, 0.0, -2.1482, 'F'], [0.0, 1.2392, -0.4341, 'N'], [0.0, -1.2392, -0.4341, 'N']])
model_coron = array([[0.0, 2.8518, 0.0], [0.0, 1.4217, 0.0], [-1.23123, 0.71085, 0.0], [1.23123, 0.71085, -0.0], [-1.25098, 3.50954, 0.0], [-2.41386, 2.83815, 0.0], [-2.46973, 1.4259, 0.0], [-3.66484, 0.67139, 0.0], [-3.66484, -0.67139, 0.0], [2.46973, 1.4259, -0.0], [2.41386, 2.83815, -0.0], [1.25098, 3.50954, 0.0], [-1.23123, -0.71085, 0.0], [1.23123, -0.71085, -0.0], [-2.46973, -1.4259, 0.0], [3.66484, 0.67139, -0.0], [-0.0, -1.4217, -0.0], [2.46973, -1.4259, -0.0], [-2.41386, -2.83815, 0.0], [3.66484, -0.67139, -0.0], [-0.0, -2.8518, -0.0], [2.41386, -2.83815, -0.0], [-1.25098, -3.50954, -0.0], [1.25098, -3.50954, -0.0]])
model_bcp = array([[ 9.10284417e-06,-1.57665886e-05, 9.43138464e-01], [ 9.10284417e-06,-1.57665886e-05,-9.43138464e-01], [ 1.23268196e+00, 1.07036079e-05,-1.19096467e-16], [-6.16349916e-01, 1.06754937e+00, 1.07501951e-16], [-6.16350250e-01,-1.06752854e+00, 1.16393846e-17]])
model_dpm = array([[5.566576374209448e-17, 1.3442915151292316, -0.9090909114505692], [-2.9814583249802527e-17, 1.22136535903189, 0.4869090952682944], [-6.306374341225482e-17, 2.511971386171238, 1.0299090881740312], [2.504960914053955e-19, 3.394640345536385, -0.004090911625781618], [7.201479850754641e-17, 2.646036534220928, -1.1760909113988771], [-7.010546335797796e-17, -7.010546335797796e-17, 1.1449091020658058], [-2.981458324980251e-17, -1.22136535903189, 0.4869090952682942], [5.566576374209448e-17, -1.3442915151292316, -0.9090909114505692], [7.20147985075464e-17, -2.646036534220928, -1.1760909113988771], [2.504960914054448e-19, -3.394640345536385, -0.004090911625782033], [-6.306374341225482e-17, -2.511971386171238, 1.0299090881740312]])
model_macac = array([[-2.56201,  1.231  , -0.     ], [-2.56201, -1.231  ,  0.     ], [-3.21582, -0.     ,  0.     ], [-3.38219,  2.48555, -0.     ], [-3.38219, -2.48555,  0.     ], [ 2.56201, -1.231  ,  0.     ], [ 2.56201,  1.231  , -0.     ], [ 3.21582,  0.     ,  0.     ], [ 3.38219, -2.48555,  0.     ], [ 3.38219,  2.48555, -0.     ], [ 0.     ,  0.     ,  0.     ], [-1.30833, -1.39517,  0.     ], [-1.30833,  1.39517, -0.     ], [ 1.30833,  1.39517, -0.     ], [ 1.30833, -1.39517,  0.     ]])
model_tbcor = array([['-4.86565', '-3.70926', '-0.0', 'C'], ['-5.52827', '-2.55893', '-0.0', 'C'], ['-4.8777', '-1.36082', '-0.0', 'C'], ['-3.5431', '-1.2645', '-0.0', 'C'], ['-2.84995', '-2.47099', '-0.0', 'C'], ['-3.55746', '-3.69375', '-0.0', 'C'], ['-1.41245', '-2.44547', '-0.0', 'C'], ['-0.67922', '-3.61244', '-0.0', 'C'], ['-0.71573', '-1.23272', '-0.0', 'C'], ['-1.43828', '0.0', '-0.0', 'C'], ['0.71573', '-1.23272', '-0.0', 'C'], ['1.41245', '-2.44547', '-0.0', 'C'], ['0.67922', '-3.61244', '-0.0', 'C'], ['2.84995', '-2.47099', '0.0', 'C'], ['3.5431', '-1.2645', '0.0', 'C'], ['-2.83989', '0.0', '-0.0', 'C'], ['4.8777', '-1.36082', '0.0', 'C'], ['5.52827', '-2.55893', '0.0', 'C'], ['4.86565', '-3.70926', '0.0', 'C'], ['3.55746', '-3.69375', '-0.0', 'C'], ['4.86565', '3.70926', '0.0', 'C'], ['5.52827', '2.55893', '0.0', 'C'], ['4.8777', '1.36082', '0.0', 'C'], ['3.5431', '1.2645', '0.0', 'C'], ['2.84995', '2.47099', '0.0', 'C'], ['3.55746', '3.69375', '0.0', 'C'], ['1.41245', '2.44547', '0.0', 'C'], ['0.67922', '3.61244', '0.0', 'C'], ['0.71573', '1.23272', '0.0', 'C'], ['1.43828', '0.0', '0.0', 'C'], ['-0.71573', '1.23272', '0.0', 'C'], ['-1.41245', '2.44547', '0.0', 'C'], ['-0.67922', '3.61244', '0.0', 'C'], ['-2.84995', '2.47099', '-0.0', 'C'], ['-3.5431', '1.2645', '-0.0', 'C'], ['2.83989', '0.0', '0.0', 'C'], ['-4.8777', '1.36082', '-0.0', 'C'], ['-5.52827', '2.55893', '-0.0', 'C'], ['-4.86565', '3.70926', '-0.0', 'C'], ['-3.55746', '3.69375', '0.0', 'C']])
model_bcr  = array([[   11.73209013 ,   1.991244472 ,   0   ],[ 11.69359284 ,   0.664132642 ,   0   ],[ 10.44319555 ,   -0.149631208    ,   0   ],[ 9.195836389 ,   0.343273498 ,   0   ],[ 7.987281273 ,   -0.531020632    ,   0   ],[ 6.76858644  ,   0.055442964 ,   0   ],[ 5.498066394 ,   -0.685596023    ,   0   ],[ 8.115275538 ,   -2.029693474    ,   0   ],[ 4.334326857 ,   -0.022018592    ,   0   ],[ 3.007442233 ,   -0.691696464    ,   0   ],[ 2.912655849 ,   -2.191453424    ,   0   ],[ 1.899781398 ,   0.086477802 ,   0   ],[ 0.52062724  ,   -0.422888115    ,   0   ],[ -0.52062724 ,   0.422888115 ,   0   ],[ -1.899781398    ,   -0.086477802    ,   0   ],[ -2.912655849    ,   2.191453424 ,   0   ],[ -3.007442233    ,   0.691696464 ,   0   ],[ -4.334326857    ,   0.022018592 ,   0   ],[ -5.498066394    ,   0.685596023 ,   0   ],[ -6.76858644 ,   -0.055442964    ,   0   ],[ -7.987281273    ,   0.531020632 ,   0   ],[ -8.115275538    ,   2.029693474 ,   0   ],[ -9.195836389    ,   -0.343273498    ,   0   ],[ -10.44319555    ,   0.149631208 ,   0   ],[ -11.69359284    ,   -0.664132642    ,   0   ],[ -11.73209013    ,   -1.991244472    ,   0   ]])
pca_betacarotene = {'Ag': array([[-1.27249950e-01,  1.80990123e-02, -5.06434550e-22, -9.90530348e-02, -2.36423941e-02, -6.99097338e-24, -5.41015486e-02, -1.56662631e-02,  1.10702315e-25, -4.22076815e-02,  3.87077143e-03, -0.00000000e+00, -2.23383276e-02, -9.82415918e-03, -0.00000000e+00,  2.34459790e-03,  4.71779865e-04, -0.00000000e+00,  1.09657960e-02, -1.51401557e-03, -0.00000000e+00, -8.50415257e-03, -2.05563732e-02, -0.00000000e+00,  1.18750352e-02, -7.69071304e-03, -0.00000000e+00,  1.12237900e-02,  1.84636775e-02, -0.00000000e+00,  3.77772656e-02,  1.53623707e-02, -0.00000000e+00, -2.20216217e-03, -1.08424848e-02, -0.00000000e+00,  1.82360071e-03,  9.28892593e-03, -0.00000000e+00, -1.82360071e-03, -9.28892593e-03, -0.00000000e+00,  2.20216217e-03,  1.08424848e-02, -0.00000000e+00, -3.77772656e-02, -1.53623707e-02, -0.00000000e+00, -1.12237900e-02, -1.84636775e-02, -0.00000000e+00, -1.18750352e-02,  7.69071304e-03, -0.00000000e+00, -1.09657960e-02,  1.51401557e-03, -0.00000000e+00, -2.34459790e-03, -4.71779865e-04, -0.00000000e+00,  2.23383276e-02,  9.82415918e-03, -0.00000000e+00,  8.50415257e-03,  2.05563732e-02, -0.00000000e+00,  4.22076815e-02, -3.87077143e-03, -0.00000000e+00,  5.41015486e-02,  1.56662631e-02, -0.00000000e+00,  9.90530348e-02,  2.36423941e-02, -0.00000000e+00,  1.27249950e-01, -1.80990123e-02, -0.00000000e+00],[-9.33708457e-03,  4.65184803e-02,  4.55738255e-20, -9.15724680e-03,  6.83568691e-02, -2.35750063e-22,  3.29474121e-03,  2.20887049e-02,  7.02011292e-24,  5.25111817e-03, -1.19969391e-02, -0.00000000e+00,  2.13641595e-02, -2.65889076e-02, -0.00000000e+00,  4.10898895e-03, -6.04858070e-02, -0.00000000e+00,  7.80460463e-03, -5.98085476e-02, -0.00000000e+00, -6.32199496e-03,  1.97172735e-03, -0.00000000e+00, -3.70952419e-03, -6.10742976e-02, -0.00000000e+00, -3.75502533e-03, -5.04368315e-02, -0.00000000e+00,  6.70828536e-03, -3.05818229e-02, -0.00000000e+00, -3.03206496e-03, -3.19705539e-02, -0.00000000e+00, -1.81906360e-03, -7.42294798e-03, -0.00000000e+00,  1.81906360e-03,  7.42294798e-03, -0.00000000e+00,  3.03206496e-03,  3.19705539e-02, -0.00000000e+00, -6.70828536e-03,  3.05818229e-02, -0.00000000e+00,  3.75502533e-03,  5.04368315e-02, -0.00000000e+00,  3.70952419e-03,  6.10742976e-02, -0.00000000e+00, -7.80460463e-03,  5.98085476e-02, -0.00000000e+00, -4.10898895e-03,  6.04858070e-02, -0.00000000e+00, -2.13641595e-02,  2.65889076e-02, -0.00000000e+00,  6.32199496e-03, -1.97172735e-03, -0.00000000e+00, -5.25111817e-03,  1.19969391e-02, -0.00000000e+00, -3.29474121e-03, -2.20887049e-02, -0.00000000e+00,  9.15724680e-03, -6.83568691e-02, -0.00000000e+00,  9.33708457e-03, -4.65184803e-02, -0.00000000e+00]]), 'Bg': array([[ 2.40163998e-18, -1.11455511e-18, -1.23571526e-01,  1.97378249e-22,  1.02663805e-18,  7.22153969e-02,  7.46192165e-28, -9.16217025e-21,  4.48279627e-02, -8.75781283e-32,  7.92178033e-20,  4.48118819e-02,  0.00000000e+00,  2.64946349e-19,  1.68300518e-02,  0.00000000e+00, -2.31956877e-19,  7.17562646e-04,  0.00000000e+00, -1.15360954e-19, -2.36680099e-02,  0.00000000e+00, -3.76382571e-19,  1.77707731e-02,  0.00000000e+00, -1.80593869e-19, -1.71606979e-02, -2.73524711e-19,  6.09551083e-20, -3.40591657e-02,  2.48285722e-19, -3.31104350e-18, -8.45828219e-02, -7.23186269e-20,  1.16135982e-19, -3.25388827e-03, -3.02766205e-19, -6.32810541e-20, -1.65302619e-02,  3.02766205e-19,  6.32810541e-20,  1.65302619e-02,  7.23186269e-20, -1.16135982e-19,  3.25388827e-03, -2.48285722e-19,  3.31104350e-18,  8.45828219e-02,  2.73524711e-19, -6.09551083e-20,  3.40591657e-02,  0.00000000e+00,  1.80593869e-19,  1.71606979e-02,  0.00000000e+00,  1.15360954e-19,  2.36680099e-02,  0.00000000e+00,  2.31956877e-19, -7.17562646e-04,  0.00000000e+00, -2.64946349e-19, -1.68300518e-02,  0.00000000e+00,  3.76382571e-19, -1.77707731e-02,  0.00000000e+00, -7.92178033e-20, -4.48118819e-02,  0.00000000e+00,  9.16217040e-21, -4.48279627e-02,  0.00000000e+00, -1.02664607e-18, -7.22153969e-02,  0.00000000e+00,  6.15859325e-19,  1.23571526e-01],[-2.14826199e-18, -1.02955685e-18,  4.02377159e-02,  4.01975419e-23,  2.31946727e-19,  4.14871877e-02, -6.98706826e-29, -1.42208720e-19,  3.01306614e-02,  6.78917421e-32, -2.25177914e-19, -4.39319842e-02,  0.00000000e+00, -6.72298348e-20, -2.57430151e-02,  0.00000000e+00, -1.89030972e-19, -7.14430032e-02,  0.00000000e+00, -5.81500734e-20, -4.87369020e-02,  0.00000000e+00,  7.88416926e-19,  5.75678033e-02,  0.00000000e+00, -3.06883799e-19, -6.09298581e-02,  5.81486285e-20,  9.83410764e-20, -2.91393215e-02,  5.75986951e-19, -9.68292342e-20,  1.50753976e-02,  4.45650394e-19,  6.00665648e-20, -3.38079982e-02,  3.24057160e-19,  2.13936831e-19, -1.76915191e-03, -3.24057160e-19, -2.13936831e-19,  1.76915191e-03, -4.45650394e-19, -6.00665648e-20,  3.38079982e-02, -5.75986951e-19,  9.68292342e-20, -1.50753976e-02, -5.81486285e-20, -9.83410764e-20,  2.91393215e-02,  0.00000000e+00,  3.06883799e-19,  6.09298581e-02,  0.00000000e+00,  5.81500734e-20,  4.87369020e-02,  0.00000000e+00,  1.89030972e-19,  7.14430032e-02,  0.00000000e+00,  6.72298348e-20,  2.57430151e-02,  0.00000000e+00, -7.88416926e-19, -5.75678033e-02,  0.00000000e+00,  2.25177914e-19,  4.39319842e-02,  0.00000000e+00,  1.42208720e-19, -3.01306614e-02,  0.00000000e+00, -2.31948041e-19, -4.14871877e-02,  0.00000000e+00,  1.57101608e-18, -4.02377159e-02]]), 'Au': array([[ 2.41938515e-23, -4.28392954e-19, -2.67002744e-02, -1.55079728e-27,  2.91598492e-19, -6.20525096e-02,  8.61068321e-32,  3.32380160e-20, -6.88231308e-02, -1.48248332e-35, -2.66206699e-20, -4.62749768e-02, -0.00000000e+00, -6.58577504e-20, -2.15181206e-02, -0.00000000e+00,  5.37244798e-20,  1.78424408e-04, -0.00000000e+00, -4.55308131e-21,  1.92587830e-02, -0.00000000e+00,  3.54615670e-19, -2.45495685e-02, -0.00000000e+00,  9.52296546e-20,  3.44525259e-02,  1.06928190e-20,  7.80930198e-20,  4.40561933e-02,  1.02076855e-19, -2.88743907e-19,  5.00689042e-02,  4.00569582e-20,  5.86254259e-20,  4.86880613e-02, -1.98332234e-19, -2.39103252e-20,  5.33785269e-02, -1.78859680e-19,  3.77125075e-20,  5.33785269e-02,  1.31295538e-20,  6.97140938e-20,  4.86880613e-02, -3.83548650e-19,  2.58836036e-19,  5.00689042e-02, -2.72278305e-19,  1.66131446e-20,  4.40561933e-02, -0.00000000e+00, -2.85972561e-20,  3.44525259e-02, -0.00000000e+00, -3.49961852e-20,  1.92587830e-02, -0.00000000e+00, -3.09120273e-21,  1.78424408e-04, -0.00000000e+00,  1.72426024e-19, -2.15181206e-02, -0.00000000e+00, -2.58858264e-19, -2.45495685e-02, -0.00000000e+00, -1.00711887e-19, -4.62749768e-02, -0.00000000e+00, -8.61646356e-20, -6.88231308e-02, -0.00000000e+00, -3.03024328e-19, -6.20525096e-02, -0.00000000e+00,  1.28512519e-19, -2.67002744e-02],[ 1.04649682e-17, -9.59328923e-19,  1.26966937e-01, -4.90096514e-22,  1.10451084e-18,  6.32540248e-02,  7.67249903e-26, -4.03649983e-19, -3.32386669e-02,  1.26850139e-31,  1.03002232e-19, -1.01772174e-02, -0.00000000e+00, -3.07305553e-19, -4.38070617e-02, -0.00000000e+00, -4.63127779e-20, -1.92193975e-03, -0.00000000e+00,  3.12362987e-20, -1.28433516e-02, -0.00000000e+00, -3.50534889e-19, -1.13362340e-01, -0.00000000e+00, -7.32384590e-22,  8.16191805e-03,  3.71863004e-19,  2.55232042e-20,  4.27771916e-03, -1.15471419e-18,  6.44817542e-19, -3.53759732e-02,  3.38819363e-20, -2.54272081e-19,  2.30075941e-02,  4.34351714e-19, -3.40960027e-19,  2.36052565e-02,  4.36795722e-19,  2.34309764e-20,  2.36052565e-02,  4.49774287e-19, -1.37464157e-19,  2.30075941e-02,  8.49920149e-19, -5.88231009e-19, -3.53759732e-02,  5.40376178e-19, -1.08975332e-20,  4.27771916e-03, -0.00000000e+00,  7.96424077e-20,  8.16191805e-03, -0.00000000e+00,  2.66851638e-19, -1.28433516e-02, -0.00000000e+00, -3.01021692e-19, -1.92193975e-03, -0.00000000e+00, -1.94186504e-19, -4.38070617e-02, -0.00000000e+00, -5.03506016e-19, -1.13362340e-01, -0.00000000e+00,  1.71771888e-19, -1.01772174e-02, -0.00000000e+00,  2.95829022e-19, -3.32386669e-02, -0.00000000e+00, -4.41758710e-19,  6.32540248e-02, -0.00000000e+00,  1.22411178e-18,  1.26966937e-01]]), 'Bu': array([[ 2.16742635e-02,  4.22821792e-02,  3.49402311e-25,  1.08672233e-02,  6.45074320e-02, -8.30196475e-27, -2.91138018e-03,  5.92186349e-02,  1.21208337e-28,  1.42003310e-03,  3.07439675e-02,  0.00000000e+00, -1.24455601e-03,  1.83741689e-02,  0.00000000e+00,  2.25347171e-03,  5.17638584e-03,  0.00000000e+00,  2.62499746e-03, -1.31194793e-02,  0.00000000e+00, -8.17358356e-03,  1.79949657e-02,  0.00000000e+00,  3.81316745e-03, -3.13181397e-02,  0.00000000e+00,  4.38709748e-03, -4.62435326e-02,  0.00000000e+00, -2.59132359e-02, -3.50661333e-02,  0.00000000e+00, -5.19218469e-03, -5.51653391e-02,  0.00000000e+00, -3.67909723e-03, -5.85113562e-02,  0.00000000e+00, -3.67909723e-03, -5.85113562e-02,  0.00000000e+00, -5.19218469e-03, -5.51653391e-02,  0.00000000e+00, -2.59132359e-02, -3.50661333e-02,  0.00000000e+00,  4.38709748e-03, -4.62435326e-02,  0.00000000e+00,  3.81316745e-03, -3.13181397e-02,  0.00000000e+00,  2.62499746e-03, -1.31194793e-02,  0.00000000e+00,  2.25347171e-03,  5.17638584e-03,  0.00000000e+00, -1.24455601e-03,  1.83741689e-02,  0.00000000e+00, -8.17358356e-03,  1.79949657e-02,  0.00000000e+00,  1.42003310e-03,  3.07439675e-02,  0.00000000e+00, -2.91138018e-03,  5.92186349e-02,  0.00000000e+00,  1.08672233e-02,  6.45074320e-02,  0.00000000e+00,  2.16742635e-02,  4.22821792e-02,  0.00000000e+00],[ 6.41256114e-02, -8.41723769e-04,  4.24760421e-20,  3.10548831e-02,  3.00413624e-02,  2.26609435e-21,  2.84903649e-02, -7.39444856e-03, -4.84961411e-23,  1.91118434e-02, -3.09008331e-02, -0.00000000e+00,  1.29358560e-02, -2.09810790e-02, -0.00000000e+00,  7.94524208e-03, -3.80637948e-02, -0.00000000e+00,  1.60155923e-04, -1.86012325e-02, -0.00000000e+00, -7.42495208e-02, -1.12936919e-02, -0.00000000e+00,  6.80639754e-03, -6.40960728e-03, -0.00000000e+00, -2.68509029e-03,  1.70919801e-02, -0.00000000e+00, -1.01212725e-01,  3.25624282e-02, -0.00000000e+00,  8.66538708e-03,  2.90191741e-02, -0.00000000e+00, -1.74245710e-03,  2.59661925e-02, -0.00000000e+00, -1.74245710e-03,  2.59661925e-02, -0.00000000e+00,  8.66538708e-03,  2.90191741e-02, -0.00000000e+00, -1.01212725e-01,  3.25624282e-02, -0.00000000e+00, -2.68509029e-03,  1.70919801e-02, -0.00000000e+00,  6.80639754e-03, -6.40960728e-03, -0.00000000e+00,  1.60155923e-04, -1.86012325e-02, -0.00000000e+00,  7.94524208e-03, -3.80637948e-02, -0.00000000e+00,  1.29358560e-02, -2.09810790e-02, -0.00000000e+00, -7.42495208e-02, -1.12936919e-02, -0.00000000e+00,  1.91118434e-02, -3.09008331e-02, -0.00000000e+00,  2.84903649e-02, -7.39444856e-03, -0.00000000e+00,  3.10548831e-02,  3.00413624e-02, -0.00000000e+00,  6.41256114e-02, -8.41723769e-04, -0.00000000e+00]])}
model_azatrip = array([[ 1.4138e+00,  0.0000e+00, -0.0000e+00], [ 2.1165e+00,  1.1993e+00, -0.0000e+00], [ 3.4691e+00,  1.1653e+00, -0.0000e+00], [ 4.1468e+00, -1.0000e-04, -0.0000e+00], [ 3.4691e+00, -1.1654e+00, -0.0000e+00], [ 2.1165e+00, -1.1993e+00, -0.0000e+00], [ 1.4426e+00, -2.4986e+00, -0.0000e+00], [-1.9600e-02, -2.4326e+00, -0.0000e+00], [-7.2530e-01, -3.5870e+00, -0.0000e+00], [-2.0733e+00, -3.5913e+00,  0.0000e+00], [-2.7437e+00, -2.4217e+00,  0.0000e+00], [-2.0968e+00, -1.2333e+00,  0.0000e+00], [-7.0690e-01, -1.2244e+00,  0.0000e+00], [-2.8853e+00,  0.0000e+00,  0.0000e+00], [-2.0969e+00,  1.2333e+00,  0.0000e+00], [-2.7438e+00,  2.4216e+00,  0.0000e+00], [-2.0734e+00,  3.5913e+00,  0.0000e+00], [-7.2530e-01,  3.5870e+00,  0.0000e+00], [-1.9600e-02,  2.4326e+00,  0.0000e+00], [-7.0690e-01,  1.2244e+00,  0.0000e+00], [ 1.4427e+00,  2.4988e+00, -0.0000e+00], [-0.0000e+00,  0.0000e+00,  0.0000e+00]])
model_texaphyrin = array([[1.0894, 0.0, -3.4725], [0.6766, 0.0, -4.8639], [-0.6766, 0.0, -4.8639], [-1.0894, 0.0, -3.4725], [-2.4369, 0.0, -3.0802], [-2.9988, 0.0, -1.813], [-3.1224, -0.0, 0.3568], [-4.4755, 0.0, -0.1176], [-4.3989, 0.0, -1.4812], [-2.6207, -0.0, 1.6919], [-0.7056, -0.0, 3.0971], [-1.4023, -0.0, 4.3212], [-0.7099, -0.0, 5.497], [0.7099, -0.0, 5.497], [1.4023, -0.0, 4.3212], [0.7056, -0.0, 3.0971], [2.6207, -0.0, 1.6919], [3.1224, -0.0, 0.3568], [2.9988, 0.0, -1.813], [4.3989, 0.0, -1.4812], [4.4755, 0.0, -0.1176], [2.4369, 0.0, -3.0802], [0.0, 0.0, -2.6403], [-2.2533, 0.0, -0.6724], [-1.3321, -0.0, 1.8571], [1.3321, -0.0, 1.8571], [2.2533, 0.0, -0.6724]])
model_corenothiophene  = array( [[-3.53, -3.71683, 0.0], [-2.43784, -2.83087, 0.0], [-2.47069, -1.42012, 0.0], [-3.70157, -0.70515, 0.0], [-4.92522, -1.36356, 0.0], [-6.02515, -0.68539, 0.0], [-6.02515, 0.68539, 0.0], [-4.92522, 1.36356, 0.0], [-3.70157, 0.70515, 0.0], [-2.47069, 1.42012, 0.0], [-2.43784, 2.83087, -0.0], [-3.53, 3.71683, -0.0], [-3.08888, 5.00654, -0.0], [-1.22692, 3.51971, -0.0], [-0.0, 2.85216, -0.0], [-1.22692, -3.51971, 0.0], [-3.08888, -5.00654, 0.0], [-0.0, 1.43408, -0.0], [-1.24826, 0.72157, 0.0], [-1.24826, -0.72157, 0.0], [3.53, 3.71683, -0.0], [2.43784, 2.83087, -0.0], [2.47069, 1.42012, -0.0], [3.70157, 0.70515, -0.0], [4.92522, 1.36356, -0.0], [6.02515, 0.68539, -0.0], [6.02515, -0.68539, -0.0], [4.92522, -1.36356, -0.0], [3.70157, -0.70515, -0.0], [2.47069, -1.42012, -0.0], [2.43784, -2.83087, 0.0], [3.53, -3.71683, 0.0], [3.08888, -5.00654, 0.0], [1.22692, -3.51971, 0.0], [0.0, -2.85216, 0.0], [1.22692, 3.51971, -0.0], [3.08888, 5.00654, -0.0], [0.0, -1.43408, 0.0], [1.24826, -0.72157, -0.0], [1.24826, 0.72157, -0.0], [-1.42358, 5.21284, -0.0], [-1.42358, -5.21284, 0.0], [1.42358, -5.21284, 0.0], [1.42358, 5.21284, -0.0]])

model_objs_dict = {}

model_objs_dict['porphyrin'] = scsd_model('porphyrin', array([[-1.09, 2.827, 0.0], [-0.6767, 4.203, 0.0], [0.6767, 4.203, 0.0], [1.09, 2.827, 0.0], [2.413, 2.413, 0.0], [2.827, 1.09, 0.0], [4.203, 0.6767, 0.0], [4.203, -0.6767, 0.0], [2.827, -1.09, 0.0], [2.413, -2.413, 0.0], [1.09, -2.827, 0.0], [0.6767, -4.203, 0.0], [-0.6767, -4.203, 0.0], [-1.09, -2.827, 0.0], [-2.413, -2.413, 0.0], [-2.827, -1.09, 0.0], [-4.203, -0.6767, 0.0], [-4.203, 0.6767, 0.0], [-2.827, 1.09, 0.0], [-2.413, 2.413, 0.0], [0.0, 1.998, 0.0], [1.998, 0.0, 0.0], [0.0, -1.998, 0.0], [-1.998, 0.0, 0.0]]), 'D4h' , pca = {"A1g":array([[-0.00064, -0.01613, -0.0, -0.00196, -0.0932, -0.0, 0.00196, -0.0932, -0.0, 0.00064, -0.01613, -0.0, 0.00327, 0.00327, 0.0, -0.01613, 0.00064, 0.0, -0.0932, 0.00196, 0.0, -0.0932, -0.00196, -0.0, -0.01613, -0.00064, -0.0, 0.00327, -0.00327, -0.0, 0.00064, 0.01613, 0.0, 0.00196, 0.0932, 0.0, -0.00196, 0.0932, 0.0, -0.00064, 0.01613, 0.0, -0.00327, -0.00327, -0.0, 0.01613, -0.00064, -0.0, 0.0932, -0.00196, -0.0, 0.0932, 0.00196, 0.0, 0.01613, 0.00064, 0.0, -0.00327, 0.00327, 0.0, 0.0, 0.02665, 0.0, 0.02665, -0.0, -0.0, -0.0, -0.02665, -0.0, -0.02665, 0.0, 0.0], [0.02045, -0.04109, -0.0, 0.01351, -0.00484, -0.0, -0.01351, -0.00484, -0.0, -0.02045, -0.04109, -0.0, -0.04181, -0.04181, -0.0, -0.04109, -0.02045, -0.0, -0.00484, -0.01351, -0.0, -0.00484, 0.01351, 0.0, -0.04109, 0.02045, 0.0, -0.04181, 0.04181, 0.0, -0.02045, 0.04109, 0.0, -0.01351, 0.00484, 0.0, 0.01351, 0.00484, 0.0, 0.02045, 0.04109, 0.0, 0.04181, 0.04181, 0.0, 0.04109, 0.02045, 0.0, 0.00484, 0.01351, 0.0, 0.00484, -0.01351, -0.0, 0.04109, -0.02045, -0.0, 0.04181, -0.04181, -0.0, -0.0, -0.07037, -0.0, -0.07037, 0.0, 0.0, 0.0, 0.07037, 0.0, 0.07037, -0.0, -0.0]]), "A2g":array([[-0.00511, 0.01758, -0.0, -0.07253, 0.01222, 0.0, -0.07253, -0.01222, 0.0, -0.00511, -0.01758, -0.0, 0.02628, -0.02628, -0.0, 0.01758, 0.00511, 0.0, 0.01222, 0.07253, 0.0, -0.01222, 0.07253, -0.0, -0.01758, 0.00511, -0.0, -0.02628, -0.02628, 0.0, 0.00511, -0.01758, 0.0, 0.07253, -0.01222, -0.0, 0.07253, 0.01222, -0.0, 0.00511, 0.01758, 0.0, -0.02628, 0.02628, 0.0, -0.01758, -0.00511, -0.0, -0.01222, -0.07253, -0.0, 0.01222, -0.07253, 0.0, 0.01758, -0.00511, 0.0, 0.02628, 0.02628, -0.0, 0.02913, -0.0, -0.0, -0.0, -0.02913, 0.0, -0.02913, 0.0, 0.0, 0.0, 0.02913, -0.0], [0.02627, -0.04002, 0.0, -0.01468, -0.03652, 0.0, -0.01468, 0.03652, 0.0, 0.02627, 0.04002, 0.0, -0.01879, 0.01879, 0.0, -0.04002, -0.02627, 0.0, -0.03652, 0.01468, 0.0, 0.03652, 0.01468, -0.0, 0.04002, -0.02627, -0.0, 0.01879, 0.01879, -0.0, -0.02627, 0.04002, -0.0, 0.01468, 0.03652, -0.0, 0.01468, -0.03652, -0.0, -0.02627, -0.04002, -0.0, 0.01879, -0.01879, -0.0, 0.04002, 0.02627, -0.0, 0.03652, -0.01468, -0.0, -0.03652, -0.01468, 0.0, -0.04002, 0.02627, 0.0, -0.01879, -0.01879, 0.0, 0.04897, 0.0, 0.0, 0.0, -0.04897, -0.0, -0.04897, -0.0, -0.0, -0.0, 0.04897, 0.0]]), "B1g":array([[1e-05, -0.03075, -0.0, 0.00022, -0.09022, -0.0, -0.00022, -0.09022, -0.0, -1e-05, -0.03075, -0.0, 0.0031, -0.0031, -0.0, 0.03075, 1e-05, -0.0, 0.09022, 0.00022, -0.0, 0.09022, -0.00022, -0.0, 0.03075, -1e-05, -0.0, 0.0031, 0.0031, -0.0, -1e-05, 0.03075, 0.0, -0.00022, 0.09022, 0.0, 0.00022, 0.09022, 0.0, 1e-05, 0.03075, 0.0, -0.0031, 0.0031, 0.0, -0.03075, -1e-05, 0.0, -0.09022, -0.00022, 0.0, -0.09022, 0.00022, 0.0, -0.03075, 1e-05, 0.0, -0.0031, -0.0031, 0.0, 0.0, 0.00367, -0.0, -0.00367, -0.0, -0.0, -0.0, -0.00367, 0.0, 0.00367, 0.0, 0.0], [0.01245, -0.05228, -0.0, 0.00399, 0.01611, 0.0, -0.00399, 0.01611, 0.0, -0.01245, -0.05228, -0.0, 0.00807, -0.00807, 0.0, 0.05228, 0.01245, -0.0, -0.01611, 0.00399, 0.0, -0.01611, -0.00399, 0.0, 0.05228, -0.01245, -0.0, 0.00807, 0.00807, 0.0, -0.01245, 0.05228, 0.0, -0.00399, -0.01611, -0.0, 0.00399, -0.01611, -0.0, 0.01245, 0.05228, 0.0, -0.00807, 0.00807, -0.0, -0.05228, -0.01245, 0.0, 0.01611, -0.00399, -0.0, 0.01611, 0.00399, -0.0, -0.05228, 0.01245, 0.0, -0.00807, -0.00807, -0.0, -0.0, -0.0979, -0.0, 0.0979, -0.0, -0.0, 0.0, 0.0979, 0.0, -0.0979, 0.0, 0.0]]), "B2g":array([[0.03084, -0.02424, -0.0, -0.01799, -0.01611, -0.0, -0.01799, 0.01611, 0.0, 0.03084, 0.02424, 0.0, 0.04737, 0.04737, 0.0, 0.02424, 0.03084, 0.0, 0.01611, -0.01799, -0.0, -0.01611, -0.01799, -0.0, -0.02424, 0.03084, 0.0, -0.04737, 0.04737, 0.0, -0.03084, 0.02424, 0.0, 0.01799, 0.01611, 0.0, 0.01799, -0.01611, -0.0, -0.03084, -0.02424, -0.0, -0.04737, -0.04737, -0.0, -0.02424, -0.03084, -0.0, -0.01611, 0.01799, 0.0, 0.01611, 0.01799, 0.0, 0.02424, -0.03084, -0.0, 0.04737, -0.04737, -0.0, 0.05623, 0.0, 0.0, 0.0, 0.05623, 0.0, -0.05623, -0.0, -0.0, -0.0, -0.05623, -0.0], [0.03038, -0.00761, -0.0, 0.08573, -0.00427, -0.0, 0.08573, 0.00427, 0.0, 0.03038, 0.00761, 0.0, 0.0091, 0.0091, 0.0, 0.00761, 0.03038, 0.0, 0.00427, 0.08573, 0.0, -0.00427, 0.08573, 0.0, -0.00761, 0.03038, 0.0, -0.0091, 0.0091, 0.0, -0.03038, 0.00761, 0.0, -0.08573, 0.00427, 0.0, -0.08573, -0.00427, -0.0, -0.03038, -0.00761, -0.0, -0.0091, -0.0091, -0.0, -0.00761, -0.03038, -0.0, -0.00427, -0.08573, -0.0, 0.00427, -0.08573, -0.0, 0.00761, -0.03038, -0.0, 0.0091, -0.0091, -0.0, -0.00281, -0.0, -0.0, -0.0, -0.00281, -0.0, 0.00281, 0.0, 0.0, 0.0, 0.00281, 0.0]]), "A1u":array([[0.0, -0.0, 0.0449, -0.0, -0.0, 0.0801, 0.0, -0.0, -0.0801, -0.0, -0.0, -0.0449, 0.0, -0.0, -0.0, -0.0, 0.0, 0.0449, -0.0, 0.0, 0.0801, -0.0, 0.0, -0.0801, -0.0, -0.0, -0.0449, 0.0, 0.0, -0.0, -0.0, 0.0, 0.0449, -0.0, 0.0, 0.0801, -0.0, 0.0, -0.0801, 0.0, 0.0, -0.0449, -0.0, 0.0, 0.0, 0.0, -0.0, 0.0449, 0.0, -0.0, 0.0801, 0.0, -0.0, -0.0801, 0.0, 0.0, -0.0449, -0.0, -0.0, 0.0, 0.0, 0.0, -0.0, 0.0, -0.0, -0.0, -0.0, -0.0, 0.0, -0.0, 0.0, 0.0], [-0.0, 0.0, -0.0801, 0.0, -0.0, 0.0449, -0.0, 0.0, -0.0449, -0.0, 0.0, 0.0801, 0.0, -0.0, 0.0, 0.0, -0.0, -0.0801, 0.0, -0.0, 0.0449, -0.0, -0.0, -0.0449, -0.0, -0.0, 0.0801, 0.0, -0.0, 0.0, 0.0, -0.0, -0.0801, -0.0, 0.0, 0.0449, 0.0, -0.0, -0.0449, 0.0, -0.0, 0.0801, -0.0, 0.0, -0.0, -0.0, 0.0, -0.0801, -0.0, 0.0, 0.0449, 0.0, 0.0, -0.0449, 0.0, 0.0, 0.0801, -0.0, 0.0, -0.0, -0.0, -0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -0.0, -0.0, -0.0, -0.0]]), "A2u":array([[0.0, 0.0, -0.00166, -0.0, -0.0, 0.07952, -0.0, -0.0, 0.07952, 0.0, 0.0, -0.00166, 0.0, 0.0, -0.03686, 0.0, -0.0, -0.00166, 0.0, -0.0, 0.07952, 0.0, -0.0, 0.07952, 0.0, -0.0, -0.00166, -0.0, 0.0, -0.03686, 0.0, -0.0, -0.00166, -0.0, -0.0, 0.07952, -0.0, -0.0, 0.07952, -0.0, -0.0, -0.00166, -0.0, -0.0, -0.03686, -0.0, 0.0, -0.00166, 0.0, -0.0, 0.07952, 0.0, -0.0, 0.07952, -0.0, 0.0, -0.00166, 0.0, -0.0, -0.03686, 0.0, -0.0, -0.05077, -0.0, 0.0, -0.05077, 0.0, -0.0, -0.05077, -0.0, 0.0, -0.05077], [0.0, -0.0, 0.03985, -0.0, -0.0, 0.03023, 0.0, -0.0, 0.03023, -0.0, -0.0, 0.03985, -0.0, -0.0, 0.06487, -0.0, -0.0, 0.03985, 0.0, 0.0, 0.03023, -0.0, -0.0, 0.03023, 0.0, 0.0, 0.03985, -0.0, 0.0, 0.06487, -0.0, 0.0, 0.03985, 0.0, 0.0, 0.03023, -0.0, 0.0, 0.03023, 0.0, 0.0, 0.03985, 0.0, 0.0, 0.06487, 0.0, 0.0, 0.03985, 0.0, -0.0, 0.03023, 0.0, 0.0, 0.03023, -0.0, 0.0, 0.03985, 0.0, -0.0, 0.06487, -0.0, 0.0, 0.04498, 0.0, -0.0, 0.04498, -0.0, 0.0, 0.04498, 0.0, -0.0, 0.04498]]), "B1u":array([[0.0, -0.0, 0.04752, -0.0, -0.0, 0.03177, 0.0, -0.0, -0.03177, 0.0, 0.0, -0.04752, 0.0, -0.0, -0.09142, 0.0, 0.0, -0.04752, -0.0, 0.0, -0.03177, -0.0, -0.0, 0.03177, 0.0, -0.0, 0.04752, 0.0, 0.0, 0.09142, 0.0, 0.0, 0.04752, -0.0, 0.0, 0.03177, -0.0, 0.0, -0.03177, 0.0, -0.0, -0.04752, 0.0, -0.0, -0.09142, -0.0, -0.0, -0.04752, 0.0, 0.0, -0.03177, 0.0, -0.0, 0.03177, -0.0, 0.0, 0.04752, 0.0, 0.0, 0.09142, -0.0, -0.0, 0.0, 0.0, -0.0, 0.0, 0.0, 0.0, -0.0, -0.0, 0.0, -0.0], [-0.0, -0.0, 0.03489, 0.0, 0.0, 0.05342, -0.0, -0.0, -0.05342, -0.0, 0.0, -0.03489, -0.0, 0.0, 0.07339, 0.0, -0.0, -0.03489, -0.0, -0.0, -0.05342, 0.0, 0.0, 0.05342, 0.0, 0.0, 0.03489, -0.0, -0.0, -0.07339, -0.0, 0.0, 0.03489, -0.0, -0.0, 0.05342, 0.0, 0.0, -0.05342, 0.0, -0.0, -0.03489, -0.0, 0.0, 0.07339, -0.0, 0.0, -0.03489, 0.0, 0.0, -0.05342, -0.0, -0.0, 0.05342, -0.0, -0.0, 0.03489, -0.0, -0.0, -0.07339, -0.0, -0.0, 0.0, 0.0, -0.0, -0.0, 0.0, 0.0, -0.0, 0.0, 0.0, -0.0]]), "B2u":array([[0.0, 0.0, 0.03254, 0.0, 0.0, 0.09246, 0.0, 0.0, 0.09246, -0.0, 0.0, 0.03254, -0.0, -0.0, 0.0, -0.0, -0.0, -0.03254, 0.0, 0.0, -0.09246, 0.0, 0.0, -0.09246, 0.0, -0.0, -0.03254, -0.0, 0.0, 0.0, -0.0, -0.0, 0.03254, 0.0, -0.0, 0.09246, 0.0, -0.0, 0.09246, 0.0, -0.0, 0.03254, 0.0, 0.0, -0.0, 0.0, 0.0, -0.03254, -0.0, 0.0, -0.09246, -0.0, 0.0, -0.09246, -0.0, 0.0, -0.03254, 0.0, -0.0, -0.0, -0.0, 0.0, 0.0, 0.0, -0.0, -0.0, 0.0, -0.0, 0.0, -0.0, 0.0, -0.0], [0.0, 0.0, 0.04735, -0.0, -0.0, -0.01667, -0.0, -0.0, -0.01667, -0.0, -0.0, 0.04735, -0.0, -0.0, -0.0, -0.0, 0.0, -0.04735, -0.0, -0.0, 0.01667, -0.0, 0.0, 0.01667, 0.0, 0.0, -0.04735, -0.0, -0.0, -0.0, 0.0, 0.0, 0.04735, -0.0, 0.0, -0.01667, -0.0, 0.0, -0.01667, 0.0, 0.0, 0.04735, 0.0, 0.0, 0.0, 0.0, -0.0, -0.04735, 0.0, 0.0, 0.01667, 0.0, -0.0, 0.01667, -0.0, -0.0, -0.04735, 0.0, 0.0, 0.0, 0.0, 0.0, 0.12197, 0.0, 0.0, -0.12197, 0.0, -0.0, 0.12197, -0.0, 0.0, -0.12197]]), "Egx":array([[-0.0, -0.0, -0.01201, 0.0, 0.0, 0.09248, 0.0, 0.0, 0.09248, 0.0, 0.0, -0.01201, -0.0, 0.0, -0.05872, 0.0, -0.0, -0.03004, 0.0, -0.0, -0.02025, 0.0, 0.0, 0.02025, 0.0, 0.0, 0.03004, -0.0, 0.0, 0.05872, 0.0, 0.0, 0.01201, -0.0, -0.0, -0.09248, -0.0, -0.0, -0.09248, -0.0, -0.0, 0.01201, 0.0, -0.0, 0.05872, -0.0, 0.0, 0.03004, -0.0, 0.0, 0.02025, -0.0, -0.0, -0.02025, -0.0, -0.0, -0.03004, 0.0, -0.0, -0.05872, -0.0, -0.0, -0.07302, -0.0, -0.0, -0.0, 0.0, -0.0, 0.07302, -0.0, -0.0, 0.0], [0.0, 0.0, -0.05041, 0.0, -0.0, -0.07318, -0.0, -0.0, -0.07318, -0.0, 0.0, -0.05041, 0.0, -0.0, -0.06371, -0.0, -0.0, -0.02818, 0.0, -0.0, -0.01792, 0.0, 0.0, 0.01792, -0.0, 0.0, 0.02818, 0.0, 0.0, 0.06371, -0.0, 0.0, 0.05041, -0.0, 0.0, 0.07318, 0.0, 0.0, 0.07318, 0.0, -0.0, 0.05041, -0.0, 0.0, 0.06371, 0.0, 0.0, 0.02818, -0.0, 0.0, 0.01792, -0.0, -0.0, -0.01792, 0.0, -0.0, -0.02818, -0.0, -0.0, -0.06371, -0.0, 0.0, -0.0332, 0.0, 0.0, -0.0, 0.0, 0.0, 0.0332, 0.0, 0.0, 0.0]]), "Egy":array([[-0.0, -0.0, 0.03063, -0.0, 0.0, 0.02025, 0.0, 0.0, -0.02025, 0.0, -0.0, -0.03063, 0.0, 0.0, -0.0592, 0.0, -0.0, -0.01191, 0.0, -0.0, 0.09212, 0.0, -0.0, 0.09212, 0.0, 0.0, -0.01191, 0.0, -0.0, -0.0592, 0.0, 0.0, -0.03063, 0.0, -0.0, -0.02025, -0.0, -0.0, 0.02025, -0.0, 0.0, 0.03063, -0.0, -0.0, 0.0592, -0.0, 0.0, 0.01191, -0.0, 0.0, -0.09212, -0.0, 0.0, -0.09212, -0.0, -0.0, 0.01191, -0.0, 0.0, 0.0592, -0.0, -0.0, -0.0, -0.0, 0.0, -0.07178, 0.0, -0.0, 0.0, -0.0, -0.0, 0.07178], [-0.0, -0.0, -0.02679, -0.0, 0.0, -0.01677, -0.0, 0.0, 0.01677, -0.0, -0.0, 0.02679, -0.0, 0.0, 0.05998, 0.0, -0.0, 0.0531, 0.0, -0.0, 0.07348, 0.0, -0.0, 0.07348, 0.0, -0.0, 0.0531, 0.0, 0.0, 0.05998, -0.0, -0.0, 0.02679, 0.0, -0.0, 0.01677, 0.0, -0.0, -0.01677, 0.0, 0.0, -0.02679, 0.0, -0.0, -0.05998, -0.0, 0.0, -0.0531, -0.0, 0.0, -0.07348, -0.0, 0.0, -0.07348, -0.0, 0.0, -0.0531, -0.0, -0.0, -0.05998, 0.0, -0.0, 0.0, -0.0, -0.0, 0.03973, -0.0, -0.0, -0.0, -0.0, 0.0, -0.03973]]), "Egx+y":array([[-0.0, -0.0, 0.02804, 0.0, -0.0, -0.04776, 0.0, -0.0, -0.07512, -0.0, 0.0, -0.01241, 0.0, 0.0, 0.0, -0.0, -0.0, 0.01241, -0.0, -0.0, 0.07512, 0.0, 0.0, 0.04776, -0.0, -0.0, -0.02804, 0.0, -0.0, -0.07774, 0.0, 0.0, -0.02804, -0.0, 0.0, 0.04776, -0.0, 0.0, 0.07512, 0.0, -0.0, 0.01241, -0.0, -0.0, -0.0, 0.0, 0.0, -0.01241, 0.0, 0.0, -0.07512, -0.0, -0.0, -0.04776, 0.0, 0.0, 0.02804, -0.0, 0.0, 0.07774, -0.0, 0.0, 0.04779, 0.0, -0.0, -0.04779, 0.0, -0.0, -0.04779, -0.0, 0.0, 0.04779], [0.0, 0.0, 0.05683, -0.0, -0.0, 0.06545, 0.0, 0.0, 0.03969, 0.0, -0.0, 0.01702, -0.0, -0.0, 0.0, -0.0, 0.0, -0.01702, 0.0, -0.0, -0.03969, -0.0, 0.0, -0.06545, 0.0, -0.0, -0.05683, 0.0, -0.0, -0.0889, -0.0, -0.0, -0.05683, 0.0, 0.0, -0.06545, -0.0, -0.0, -0.03969, -0.0, 0.0, -0.01702, 0.0, 0.0, -0.0, 0.0, -0.0, 0.01702, -0.0, 0.0, 0.03969, 0.0, -0.0, 0.06545, -0.0, 0.0, 0.05683, -0.0, 0.0, 0.0889, -0.0, 0.0, 0.02656, 0.0, -0.0, -0.02656, 0.0, -0.0, -0.02656, -0.0, 0.0, 0.02656]]), "Egx-y":array([[0.0, -0.0, -0.01197, 0.0, -0.0, -0.0743, 0.0, -0.0, -0.0479, 0.0, -0.0, 0.02809, -0.0, -0.0, 0.07876, 0.0, -0.0, 0.02809, 0.0, 0.0, -0.0479, -0.0, 0.0, -0.0743, -0.0, 0.0, -0.01197, 0.0, -0.0, -0.0, -0.0, 0.0, 0.01197, -0.0, 0.0, 0.0743, -0.0, 0.0, 0.0479, -0.0, 0.0, -0.02809, 0.0, 0.0, -0.07876, -0.0, 0.0, -0.02809, -0.0, -0.0, 0.0479, 0.0, -0.0, 0.0743, 0.0, -0.0, 0.01197, -0.0, 0.0, 0.0, -0.0, -0.0, 0.04835, 0.0, -0.0, 0.04835, 0.0, 0.0, -0.04835, -0.0, 0.0, -0.04835], [0.0, -0.0, -0.01775, 0.0, -0.0, -0.04064, -0.0, -0.0, -0.0647, 0.0, -0.0, -0.05685, 0.0, 0.0, -0.08859, -0.0, 0.0, -0.05685, -0.0, 0.0, -0.0647, -0.0, -0.0, -0.04064, -0.0, 0.0, -0.01775, 0.0, -0.0, -0.0, 0.0, -0.0, 0.01775, -0.0, 0.0, 0.04064, 0.0, 0.0, 0.0647, -0.0, 0.0, 0.05685, -0.0, -0.0, 0.08859, 0.0, -0.0, 0.05685, 0.0, -0.0, 0.0647, 0.0, 0.0, 0.04064, 0.0, -0.0, 0.01775, -0.0, 0.0, 0.0, 0.0, 0.0, -0.02576, -0.0, 0.0, -0.02576, -0.0, -0.0, 0.02576, 0.0, -0.0, 0.02576]]), "Eux":array([[-0.01154, 0.00374, 0.0, -0.00278, 0.00203, 0.0, -0.00278, -0.00203, -0.0, -0.01154, -0.00374, -0.0, -0.01066, 6e-05, -0.0, 0.0349, 0.0021, -0.0, 0.16325, 0.00225, -0.0, 0.16325, -0.00225, -0.0, 0.0349, -0.0021, -0.0, -0.01066, -6e-05, -0.0, -0.01154, 0.00374, 0.0, -0.00278, 0.00203, 0.0, -0.00278, -0.00203, -0.0, -0.01154, -0.00374, -0.0, -0.01066, 6e-05, 0.0, 0.0349, 0.0021, 0.0, 0.16325, 0.00225, 0.0, 0.16325, -0.00225, 0.0, 0.0349, -0.0021, 0.0, -0.01066, -6e-05, 0.0, -0.01701, 0.0, -0.0, -0.03404, 0.0, -0.0, -0.01701, 0.0, 0.0, -0.03404, 0.0, 0.0], [0.00873, 0.02164, 0.0, 0.0962, 0.01494, 0.0, 0.0962, -0.01494, -0.0, 0.00873, -0.02164, -0.0, -0.03242, -0.04996, -0.0, -0.02453, -0.01594, -0.0, 0.00059, -0.00946, 0.0, 0.00059, 0.00946, 0.0, -0.02453, 0.01594, 0.0, -0.03242, 0.04996, 0.0, 0.00873, 0.02164, 0.0, 0.0962, 0.01494, 0.0, 0.0962, -0.01494, -0.0, 0.00873, -0.02164, -0.0, -0.03242, -0.04996, -0.0, -0.02453, -0.01594, -0.0, 0.00059, -0.00946, -0.0, 0.00059, 0.00946, -0.0, -0.02453, 0.01594, 0.0, -0.03242, 0.04996, 0.0, -0.03845, 0.0, -0.0, -0.02359, 0.0, -0.0, -0.03845, 0.0, 0.0, -0.02359, 0.0, 0.0]]), "Euy":array([[-0.00064, 0.03401, -0.0, -0.00111, 0.16528, -0.0, 0.00111, 0.16528, -0.0, 0.00064, 0.03401, -0.0, 0.00299, -0.00974, -0.0, -0.00067, -0.01023, 0.0, 0.002, -0.00381, -0.0, -0.002, -0.00381, 0.0, 0.00067, -0.01023, -0.0, -0.00299, -0.00974, 0.0, -0.00064, 0.03401, 0.0, -0.00111, 0.16528, 0.0, 0.00111, 0.16528, 0.0, 0.00064, 0.03401, 0.0, 0.00299, -0.00974, -0.0, -0.00067, -0.01023, 0.0, 0.002, -0.00381, -0.0, -0.002, -0.00381, 0.0, 0.00067, -0.01023, -0.0, -0.00299, -0.00974, 0.0, -0.0, -0.03976, 0.0, 0.0, -0.01217, 0.0, -0.0, -0.03976, -0.0, 0.0, -0.01217, 0.0], [0.017, -0.02278, -0.0, 0.01104, 0.00125, -0.0, -0.01104, 0.00125, -0.0, -0.017, -0.02278, 0.0, -0.03867, -0.03119, 0.0, -0.01575, 0.01151, 0.0, -0.00974, 0.10425, 0.0, 0.00974, 0.10425, -0.0, 0.01575, 0.01151, -0.0, 0.03867, -0.03119, -0.0, 0.017, -0.02278, -0.0, 0.01104, 0.00125, 0.0, -0.01104, 0.00125, 0.0, -0.017, -0.02278, 0.0, -0.03867, -0.03119, 0.0, -0.01575, 0.01151, 0.0, -0.00974, 0.10425, 0.0, 0.00974, 0.10425, -0.0, 0.01575, 0.01151, -0.0, 0.03867, -0.03119, -0.0, -0.0, -0.0349, 0.0, 0.0, -0.03824, 0.0, -0.0, -0.0349, -0.0, 0.0, -0.03824, -0.0]]), "Eux+y":array([[0.00869, -0.01986, -0.0, 0.00588, -0.08967, -0.0, 0.00283, -0.08997, -0.0, 0.00641, -0.01968, -0.0, 0.00102, 0.00102, 0.0, -0.01968, 0.00641, 0.0, -0.08997, 0.00283, 0.0, -0.08967, 0.00588, 0.0, -0.01986, 0.00869, 0.0, 0.00817, 0.00817, -0.0, 0.00869, -0.01986, -0.0, 0.00588, -0.08967, -0.0, 0.00283, -0.08997, -0.0, 0.00641, -0.01968, -0.0, 0.00102, 0.00102, -0.0, -0.01968, 0.00641, 0.0, -0.08997, 0.00283, 0.0, -0.08967, 0.00588, 0.0, -0.01986, 0.00869, 0.0, 0.00817, 0.00817, -0.0, 0.00919, 0.01917, 0.0, 0.01917, 0.00919, -0.0, 0.00919, 0.01917, 0.0, 0.01917, 0.00919, -0.0], [0.01818, -0.0031, -0.0, 0.0721, 0.01303, -0.0, 0.05921, -0.00464, -0.0, -0.00092, -0.02852, -0.0, -0.04571, -0.04571, -0.0, -0.02852, -0.00092, 0.0, -0.00464, 0.05921, 0.0, 0.01303, 0.0721, 0.0, -0.0031, 0.01818, 0.0, 0.00827, 0.00827, 0.0, 0.01818, -0.0031, -0.0, 0.0721, 0.01303, -0.0, 0.05921, -0.00464, -0.0, -0.00092, -0.02852, -0.0, -0.04571, -0.04571, 0.0, -0.02852, -0.00092, 0.0, -0.00464, 0.05921, 0.0, 0.01303, 0.0721, 0.0, -0.0031, 0.01818, 0.0, 0.00827, 0.00827, -0.0, -0.02365, -0.02184, 0.0, -0.02184, -0.02365, 0.0, -0.02365, -0.02184, -0.0, -0.02184, -0.02365, -0.0]]), "Eux-y":array([[-0.0041, -0.01562, -0.0, 0.00109, -0.09013, -0.0, 0.00042, -0.09081, -0.0, -0.00488, -0.02062, -0.0, -0.00455, 0.00455, -0.0, 0.02062, 0.00488, 0.0, 0.09081, -0.00042, 0.0, 0.09013, -0.00109, 0.0, 0.01562, 0.0041, 0.0, -0.00887, 0.00887, 0.0, -0.0041, -0.01562, -0.0, 0.00109, -0.09013, -0.0, 0.00042, -0.09081, -0.0, -0.00488, -0.02062, -0.0, -0.00455, 0.00455, 0.0, 0.02062, 0.00488, 0.0, 0.09081, -0.00042, 0.0, 0.09013, -0.00109, 0.0, 0.01562, 0.0041, 0.0, -0.00887, 0.00887, -0.0, -0.007, 0.02109, 0.0, -0.02109, 0.007, -0.0, -0.007, 0.02109, 0.0, -0.02109, 0.007, -0.0], [0.00833, -0.02631, -0.0, -0.0581, -0.01035, -0.0, -0.07202, 0.00426, -0.0, -0.01594, -0.00341, -0.0, -0.00744, 0.00744, 0.0, 0.00341, 0.01594, 0.0, -0.00426, 0.07202, 0.0, 0.01035, 0.0581, 0.0, 0.02631, -0.00833, 0.0, 0.05364, -0.05364, -0.0, 0.00833, -0.02631, -0.0, -0.0581, -0.01035, -0.0, -0.07202, 0.00426, -0.0, -0.01594, -0.00341, -0.0, -0.00744, 0.00744, -0.0, 0.00341, 0.01594, 0.0, -0.00426, 0.07202, 0.0, 0.01035, 0.0581, 0.0, 0.02631, -0.00833, 0.0, 0.05364, -0.05364, 0.0, 0.02722, -0.01632, 0.0, 0.01632, -0.02722, -0.0, 0.02722, -0.01632, 0.0, 0.01632, -0.02722, -0.0]])}, database_path = 'porphyrin_scsd_20221012.pkl', maxdist = 1.75)
#model_objs_dict['porphyrin'] = scsd_model('porphyrin', array([[-1.09, 2.827, 0.0], [-0.6767, 4.203, 0.0], [0.6767, 4.203, 0.0], [1.09, 2.827, 0.0], [2.413, 2.413, 0.0], [2.827, 1.09, 0.0], [4.203, 0.6767, 0.0], [4.203, -0.6767, 0.0], [2.827, -1.09, 0.0], [2.413, -2.413, 0.0], [1.09, -2.827, 0.0], [0.6767, -4.203, 0.0], [-0.6767, -4.203, 0.0], [-1.09, -2.827, 0.0], [-2.413, -2.413, 0.0], [-2.827, -1.09, 0.0], [-4.203, -0.6767, 0.0], [-4.203, 0.6767, 0.0], [-2.827, 1.09, 0.0], [-2.413, 2.413, 0.0], [0.0, 1.998, 0.0], [1.998, 0.0, 0.0], [0.0, -1.998, 0.0], [-1.998, 0.0, 0.0]]), 'D4h' , pca = {"A1g":array([[-0.00055, -0.01625, -0.0, -0.00189, -0.09331, -0.0, 0.00189, -0.09331, -0.0, 0.00055, -0.01625, -0.0, 0.0031, 0.0031, 0.0, -0.01625, 0.00055, 0.0, -0.09331, 0.00189, 0.0, -0.09331, -0.00189, -0.0, -0.01625, -0.00055, -0.0, 0.0031, -0.0031, -0.0, 0.00055, 0.01625, 0.0, 0.00189, 0.09331, 0.0, -0.00189, 0.09331, 0.0, -0.00055, 0.01625, 0.0, -0.0031, -0.0031, -0.0, 0.01625, -0.00055, -0.0, 0.09331, -0.00189, -0.0, 0.09331, 0.00189, 0.0, 0.01625, 0.00055, 0.0, -0.0031, 0.0031, 0.0, 0.0, 0.02645, 0.0, 0.02645, -0.0, -0.0, -0.0, -0.02645, -0.0, -0.02645, 0.0, 0.0], [0.02048, -0.04106, -0.0, 0.01354, -0.00462, -0.0, -0.01354, -0.00462, -0.0, -0.02048, -0.04106, -0.0, -0.04185, -0.04185, -0.0, -0.04106, -0.02048, -0.0, -0.00462, -0.01354, -0.0, -0.00462, 0.01354, 0.0, -0.04106, 0.02048, 0.0, -0.04185, 0.04185, 0.0, -0.02048, 0.04106, 0.0, -0.01354, 0.00462, 0.0, 0.01354, 0.00462, 0.0, 0.02048, 0.04106, 0.0, 0.04185, 0.04185, 0.0, 0.04106, 0.02048, 0.0, 0.00462, 0.01354, 0.0, 0.00462, -0.01354, -0.0, 0.04106, -0.02048, -0.0, 0.04185, -0.04185, -0.0, -0.0, -0.07043, -0.0, -0.07043, 0.0, 0.0, 0.0, 0.07043, 0.0, 0.07043, -0.0, -0.0]]), "A2g":array([[0.01182, 0.02324, 0.0, -0.04753, 0.01569, 0.0, -0.04753, -0.01569, 0.0, 0.01182, -0.02324, 0.0, 0.03969, -0.03969, 0.0, 0.02324, -0.01182, 0.0, 0.01569, 0.04753, 0.0, -0.01569, 0.04753, -0.0, -0.02324, -0.01182, -0.0, -0.03969, -0.03969, -0.0, -0.01182, -0.02324, -0.0, 0.04753, -0.01569, -0.0, 0.04753, 0.01569, -0.0, -0.01182, 0.02324, -0.0, -0.03969, 0.03969, -0.0, -0.02324, 0.01182, -0.0, -0.01569, -0.04753, -0.0, 0.01569, -0.04753, 0.0, 0.02324, 0.01182, 0.0, 0.03969, 0.03969, 0.0, 0.04161, -0.0, 0.0, -0.0, -0.04161, 0.0, -0.04161, 0.0, -0.0, 0.0, 0.04161, -0.0], [0.03628, 0.01169, 0.0, 0.05274, 0.00682, -0.0, 0.05274, -0.00682, -0.0, 0.03628, -0.01169, 0.0, 0.0289, -0.0289, 0.0, 0.01169, -0.03628, 0.0, 0.00682, -0.05274, 0.0, -0.00682, -0.05274, -0.0, -0.01169, -0.03628, -0.0, -0.0289, -0.0289, -0.0, -0.03628, -0.01169, -0.0, -0.05274, -0.00682, 0.0, -0.05274, 0.00682, 0.0, -0.03628, 0.01169, -0.0, -0.0289, 0.0289, -0.0, -0.01169, 0.03628, -0.0, -0.00682, 0.05274, -0.0, 0.00682, 0.05274, 0.0, 0.01169, 0.03628, 0.0, 0.0289, 0.0289, 0.0, 0.02654, 0.0, 0.0, 0.0, -0.02654, 0.0, -0.02654, -0.0, -0.0, -0.0, 0.02654, -0.0]]), "B1g":array([[-0.00013, -0.03121, -0.0, 0.00025, -0.09, -0.0, -0.00025, -0.09, -0.0, 0.00013, -0.03121, -0.0, 0.00355, -0.00355, -0.0, 0.03121, -0.00013, -0.0, 0.09, 0.00025, -0.0, 0.09, -0.00025, -0.0, 0.03121, 0.00013, -0.0, 0.00355, 0.00355, 0.0, 0.00013, 0.03121, 0.0, -0.00025, 0.09, 0.0, 0.00025, 0.09, 0.0, -0.00013, 0.03121, 0.0, -0.00355, 0.00355, 0.0, -0.03121, 0.00013, 0.0, -0.09, -0.00025, 0.0, -0.09, 0.00025, 0.0, -0.03121, -0.00013, 0.0, -0.00355, -0.00355, -0.0, 0.0, 0.00255, 0.0, -0.00255, 0.0, 0.0, -0.0, -0.00255, -0.0, 0.00255, -0.0, -0.0], [-0.01194, 0.05158, 0.0, -0.00388, -0.01684, -0.0, 0.00388, -0.01684, -0.0, 0.01194, 0.05158, 0.0, -0.00855, 0.00855, 0.0, -0.05158, -0.01194, 0.0, 0.01684, -0.00388, -0.0, 0.01684, 0.00388, -0.0, -0.05158, 0.01194, 0.0, -0.00855, -0.00855, 0.0, 0.01194, -0.05158, -0.0, 0.00388, 0.01684, 0.0, -0.00388, 0.01684, 0.0, -0.01194, -0.05158, -0.0, 0.00855, -0.00855, -0.0, 0.05158, 0.01194, -0.0, -0.01684, 0.00388, 0.0, -0.01684, -0.00388, 0.0, 0.05158, -0.01194, -0.0, 0.00855, 0.00855, -0.0, 0.0, 0.09746, 0.0, -0.09746, 0.0, 0.0, -0.0, -0.09746, -0.0, 0.09746, -0.0, -0.0]]), "B2g":array([[0.03076, -0.02424, -0.0, -0.01811, -0.01616, -0.0, -0.01811, 0.01616, 0.0, 0.03076, 0.02424, 0.0, 0.04729, 0.04729, 0.0, 0.02424, 0.03076, 0.0, 0.01616, -0.01811, -0.0, -0.01616, -0.01811, -0.0, -0.02424, 0.03076, 0.0, -0.04729, 0.04729, 0.0, -0.03076, 0.02424, 0.0, 0.01811, 0.01616, 0.0, 0.01811, -0.01616, -0.0, -0.03076, -0.02424, -0.0, -0.04729, -0.04729, -0.0, -0.02424, -0.03076, -0.0, -0.01616, 0.01811, 0.0, 0.01616, 0.01811, 0.0, 0.02424, -0.03076, -0.0, 0.04729, -0.04729, -0.0, 0.05625, -0.0, -0.0, -0.0, 0.05625, 0.0, -0.05625, 0.0, 0.0, 0.0, -0.05625, -0.0], [0.03044, -0.00773, -0.0, 0.08556, -0.00419, -0.0, 0.08556, 0.00419, 0.0, 0.03044, 0.00773, 0.0, 0.00924, 0.00924, 0.0, 0.00773, 0.03044, 0.0, 0.00419, 0.08556, 0.0, -0.00419, 0.08556, 0.0, -0.00773, 0.03044, 0.0, -0.00924, 0.00924, 0.0, -0.03044, 0.00773, 0.0, -0.08556, 0.00419, 0.0, -0.08556, -0.00419, -0.0, -0.03044, -0.00773, -0.0, -0.00924, -0.00924, -0.0, -0.00773, -0.03044, -0.0, -0.00419, -0.08556, -0.0, 0.00419, -0.08556, -0.0, 0.00773, -0.03044, -0.0, 0.00924, -0.00924, -0.0, -0.0028, -0.0, -0.0, -0.0, -0.0028, -0.0, 0.0028, 0.0, 0.0, 0.0, 0.0028, 0.0]]), "A1u":array([[0.0, 0.0, 0.04419, -0.0, -0.0, 0.08081, 0.0, -0.0, -0.08081, 0.0, -0.0, -0.04419, 0.0, 0.0, -0.0, -0.0, -0.0, 0.04419, -0.0, 0.0, 0.08081, -0.0, -0.0, -0.08081, 0.0, -0.0, -0.04419, -0.0, -0.0, 0.0, -0.0, -0.0, 0.04419, 0.0, 0.0, 0.08081, -0.0, 0.0, -0.08081, -0.0, 0.0, -0.04419, -0.0, -0.0, 0.0, 0.0, 0.0, 0.04419, 0.0, -0.0, 0.08081, 0.0, 0.0, -0.08081, -0.0, 0.0, -0.04419, 0.0, 0.0, -0.0, -0.0, 0.0, -0.0, 0.0, 0.0, -0.0, 0.0, -0.0, 0.0, -0.0, -0.0, 0.0], [0.0, 0.0, -0.08081, -0.0, 0.0, 0.04419, 0.0, 0.0, -0.04419, 0.0, -0.0, 0.08081, -0.0, -0.0, 0.0, 0.0, -0.0, -0.08081, -0.0, 0.0, 0.04419, 0.0, -0.0, -0.04419, 0.0, -0.0, 0.08081, -0.0, -0.0, -0.0, -0.0, 0.0, -0.08081, 0.0, -0.0, 0.04419, -0.0, -0.0, -0.04419, -0.0, 0.0, 0.08081, 0.0, 0.0, -0.0, -0.0, 0.0, -0.08081, 0.0, -0.0, 0.04419, -0.0, 0.0, -0.04419, -0.0, 0.0, 0.08081, 0.0, 0.0, 0.0, -0.0, -0.0, 0.0, -0.0, 0.0, 0.0, 0.0, 0.0, -0.0, 0.0, -0.0, -0.0]]), "A2u":array([[-0.0, -0.0, 0.01192, 0.0, -0.0, -0.0625, 0.0, 0.0, -0.0625, -0.0, -0.0, 0.01192, 0.0, -0.0, 0.04429, -0.0, -0.0, 0.01192, 0.0, 0.0, -0.0625, -0.0, 0.0, -0.0625, -0.0, -0.0, 0.01192, -0.0, 0.0, 0.04429, 0.0, 0.0, 0.01192, 0.0, 0.0, -0.0625, 0.0, 0.0, -0.0625, 0.0, 0.0, 0.01192, -0.0, 0.0, 0.04429, 0.0, 0.0, 0.01192, 0.0, 0.0, -0.0625, -0.0, 0.0, -0.0625, 0.0, 0.0, 0.01192, 0.0, -0.0, 0.04429, -0.0, 0.0, 0.05687, 0.0, -0.0, 0.05687, -0.0, 0.0, 0.05687, 0.0, -0.0, 0.05687], [0.0, -0.0, -0.03119, -0.0, 0.0, 0.00578, -0.0, -0.0, 0.00578, 0.0, 0.0, -0.03119, -0.0, -0.0, 0.11347, 0.0, 0.0, -0.03119, -0.0, 0.0, 0.00578, -0.0, 0.0, 0.00578, 0.0, -0.0, -0.03119, 0.0, 0.0, 0.11347, 0.0, -0.0, -0.03119, 0.0, -0.0, 0.00578, 0.0, 0.0, 0.00578, -0.0, -0.0, -0.03119, 0.0, 0.0, 0.11347, -0.0, -0.0, -0.03119, 0.0, 0.0, 0.00578, -0.0, -0.0, 0.00578, -0.0, 0.0, -0.03119, -0.0, -0.0, 0.11347, -0.0, 0.0, -0.06258, 0.0, -0.0, -0.06258, -0.0, 0.0, -0.06258, 0.0, -0.0, -0.06258]]), "B1u":array([[-0.0, -0.0, 0.04753, -0.0, -0.0, 0.03175, 0.0, -0.0, -0.03175, -0.0, 0.0, -0.04753, -0.0, -0.0, -0.09145, 0.0, -0.0, -0.04753, -0.0, 0.0, -0.03175, -0.0, -0.0, 0.03175, -0.0, -0.0, 0.04753, 0.0, 0.0, 0.09145, -0.0, 0.0, 0.04753, 0.0, 0.0, 0.03175, -0.0, 0.0, -0.03175, 0.0, -0.0, -0.04753, 0.0, 0.0, -0.09145, -0.0, 0.0, -0.04753, 0.0, -0.0, -0.03175, 0.0, 0.0, 0.03175, 0.0, 0.0, 0.04753, -0.0, 0.0, 0.09145, -0.0, -0.0, 0.0, 0.0, -0.0, -0.0, 0.0, 0.0, -0.0, -0.0, 0.0, 0.0], [0.0, -0.0, 0.03153, 0.0, 0.0, 0.05721, -0.0, 0.0, -0.05721, -0.0, -0.0, -0.03153, -0.0, 0.0, 0.07251, -0.0, -0.0, -0.03153, 0.0, 0.0, -0.05721, 0.0, 0.0, 0.05721, -0.0, -0.0, 0.03153, -0.0, -0.0, -0.07251, -0.0, 0.0, 0.03153, -0.0, -0.0, 0.05721, 0.0, -0.0, -0.05721, 0.0, 0.0, -0.03153, -0.0, 0.0, 0.07251, 0.0, 0.0, -0.03153, -0.0, 0.0, -0.05721, -0.0, -0.0, 0.05721, 0.0, 0.0, 0.03153, -0.0, -0.0, -0.07251, -0.0, -0.0, 0.0, 0.0, -0.0, 0.0, 0.0, 0.0, -0.0, -0.0, 0.0, -0.0]]), "B2u":array([[0.0, -0.0, -0.03253, -0.0, 0.0, -0.09246, -0.0, 0.0, -0.09246, 0.0, -0.0, -0.03253, -0.0, -0.0, 0.0, -0.0, -0.0, 0.03253, 0.0, -0.0, 0.09246, 0.0, -0.0, 0.09246, -0.0, -0.0, 0.03253, -0.0, 0.0, 0.0, -0.0, 0.0, -0.03253, -0.0, -0.0, -0.09246, -0.0, -0.0, -0.09246, -0.0, 0.0, -0.03253, 0.0, 0.0, -0.0, 0.0, 0.0, 0.03253, -0.0, -0.0, 0.09246, -0.0, -0.0, 0.09246, 0.0, 0.0, 0.03253, 0.0, -0.0, -0.0, 0.0, -0.0, 1e-05, -0.0, -0.0, -1e-05, -0.0, 0.0, 1e-05, 0.0, 0.0, -1e-05], [-0.0, 0.0, 0.04713, -0.0, -0.0, -0.01658, -0.0, -0.0, -0.01658, 0.0, 0.0, 0.04713, 0.0, 0.0, 0.0, 0.0, -0.0, -0.04713, -0.0, 0.0, 0.01658, -0.0, -0.0, 0.01658, 0.0, -0.0, -0.04713, 0.0, -0.0, 0.0, 0.0, -0.0, 0.04713, -0.0, 0.0, -0.01658, -0.0, 0.0, -0.01658, -0.0, -0.0, 0.04713, -0.0, -0.0, -0.0, -0.0, 0.0, -0.04713, 0.0, -0.0, 0.01658, 0.0, -0.0, 0.01658, -0.0, 0.0, -0.04713, -0.0, 0.0, -0.0, 0.0, 0.0, 0.12258, 0.0, 0.0, -0.12258, 0.0, -0.0, 0.12258, -0.0, 0.0, -0.12258]]), "Egx":array([[0.0, -0.0, -0.01863, 0.0, 0.0, 0.07721, 0.0, 0.0, 0.07721, 0.0, -0.0, -0.01863, 0.0, 0.0, -0.06365, 0.0, 0.0, -0.03214, -0.0, 0.0, -0.02119, -0.0, 0.0, 0.02119, 0.0, 0.0, 0.03214, 0.0, -0.0, 0.06365, 0.0, 0.0, 0.01863, -0.0, -0.0, -0.07721, -0.0, -0.0, -0.07721, -0.0, 0.0, 0.01863, -0.0, -0.0, 0.06365, -0.0, -0.0, 0.03214, 0.0, -0.0, 0.02119, 0.0, -0.0, -0.02119, -0.0, -0.0, -0.03214, -0.0, 0.0, -0.06365, -0.0, 0.0, -0.07437, 0.0, -0.0, 0.0, 0.0, 0.0, 0.07437, 0.0, 0.0, -0.0], [0.0, 0.0, -0.04242, 0.0, 0.0, 0.00917, -0.0, 0.0, 0.00917, 0.0, 0.0, -0.04242, -0.0, 0.0, 0.04843, 0.0, -0.0, 0.04572, 0.0, 0.0, 0.04919, 0.0, 0.0, -0.04919, 0.0, -0.0, -0.04572, -0.0, 0.0, -0.04843, 0.0, -0.0, 0.04242, -0.0, -0.0, -0.00917, 0.0, -0.0, -0.00917, -0.0, -0.0, 0.04242, 0.0, -0.0, -0.04843, -0.0, 0.0, -0.04572, -0.0, -0.0, -0.04919, -0.0, -0.0, 0.04919, -0.0, 0.0, 0.04572, 0.0, -0.0, 0.04843, -0.0, 0.0, -0.11014, 0.0, -0.0, 0.0, 0.0, 0.0, 0.11014, 0.0, 0.0, -0.0]]), "Egy":array([[0.0, -0.0, 0.03184, -0.0, 0.0, 0.02109, 0.0, 0.0, -0.02109, -0.0, 0.0, -0.03184, 0.0, -0.0, -0.06231, -0.0, -0.0, -0.01941, -0.0, -0.0, 0.07739, -0.0, -0.0, 0.07739, -0.0, 0.0, -0.01941, -0.0, -0.0, -0.06231, -0.0, 0.0, -0.03184, 0.0, -0.0, -0.02109, -0.0, -0.0, 0.02109, 0.0, -0.0, 0.03184, -0.0, 0.0, 0.06231, 0.0, 0.0, 0.01941, 0.0, 0.0, -0.07739, 0.0, 0.0, -0.07739, 0.0, -0.0, 0.01941, 0.0, 0.0, 0.06231, -0.0, 0.0, -0.0, 0.0, 0.0, -0.0759, 0.0, 0.0, 0.0, 0.0, -0.0, 0.0759], [0.0, -0.0, -0.04222, 0.0, 0.0, -0.04455, -0.0, 0.0, 0.04455, 0.0, -0.0, 0.04222, -0.0, 0.0, 0.05222, -0.0, 0.0, -0.05117, 0.0, -0.0, 0.00179, 0.0, -0.0, 0.00179, -0.0, 0.0, -0.05117, 0.0, 0.0, 0.05222, 0.0, -0.0, 0.04222, -0.0, -0.0, 0.04455, 0.0, -0.0, -0.04455, -0.0, 0.0, -0.04222, 0.0, -0.0, -0.05222, 0.0, -0.0, 0.05117, -0.0, 0.0, -0.00179, -0.0, 0.0, -0.00179, 0.0, -0.0, 0.05117, -0.0, -0.0, -0.05222, 0.0, -0.0, 0.0, -0.0, 0.0, -0.1161, -0.0, -0.0, -0.0, -0.0, -0.0, 0.1161]]), "Egx+y":array([[-0.0, 0.0, 0.03537, 0.0, -0.0, -0.0399, 0.0, -0.0, -0.06893, -0.0, -0.0, -0.00901, -0.0, -0.0, -0.0, -0.0, -0.0, 0.00901, -0.0, 0.0, 0.06893, -0.0, 0.0, 0.0399, 0.0, 0.0, -0.03537, -0.0, 0.0, -0.08773, 0.0, -0.0, -0.03537, -0.0, 0.0, 0.0399, -0.0, 0.0, 0.06893, 0.0, 0.0, 0.00901, 0.0, 0.0, 0.0, 0.0, 0.0, -0.00901, 0.0, -0.0, -0.06893, 0.0, -0.0, -0.0399, -0.0, -0.0, 0.03537, 0.0, -0.0, 0.08773, -0.0, -0.0, 0.05293, -0.0, -0.0, -0.05293, 0.0, 0.0, -0.05293, 0.0, 0.0, 0.05293], [0.0, -0.0, -0.00546, -0.0, -0.0, 0.0308, -0.0, -0.0, -0.03563, -0.0, 0.0, -0.06479, 0.0, -0.0, -0.0, 0.0, 0.0, 0.06479, -0.0, 0.0, 0.03563, 0.0, 0.0, -0.0308, 0.0, 0.0, 0.00546, -0.0, -0.0, -0.06229, -0.0, 0.0, 0.00546, 0.0, 0.0, -0.0308, 0.0, 0.0, 0.03563, 0.0, -0.0, 0.06479, -0.0, 0.0, 0.0, -0.0, -0.0, -0.06479, 0.0, -0.0, -0.03563, -0.0, -0.0, 0.0308, -0.0, -0.0, -0.00546, 0.0, 0.0, 0.06229, 0.0, 0.0, -0.08217, -0.0, 0.0, 0.08217, -0.0, -0.0, 0.08217, 0.0, -0.0, -0.08217]]), "Egx-y":array([[0.0, 0.0, 0.00924, -0.0, 0.0, 0.06906, -0.0, 0.0, 0.03875, -0.0, -0.0, -0.03615, 0.0, 0.0, -0.08902, 0.0, -0.0, -0.03615, 0.0, -0.0, 0.03875, 0.0, -0.0, 0.06906, 0.0, -0.0, 0.00924, 0.0, -0.0, -0.0, -0.0, -0.0, -0.00924, 0.0, -0.0, -0.06906, 0.0, -0.0, -0.03875, 0.0, 0.0, 0.03615, -0.0, -0.0, 0.08902, -0.0, 0.0, 0.03615, -0.0, 0.0, -0.03875, -0.0, 0.0, -0.06906, -0.0, 0.0, -0.00924, -0.0, 0.0, 0.0, 0.0, -0.0, -0.05229, -0.0, 0.0, -0.05229, -0.0, 0.0, 0.05229, 0.0, -0.0, 0.05229], [0.0, -0.0, -0.06536, -0.0, -0.0, -0.0264, -0.0, -0.0, 0.03857, -0.0, -0.0, -0.00236, -0.0, -0.0, 0.07483, -0.0, 0.0, -0.00236, -0.0, -0.0, 0.03857, -0.0, 0.0, -0.0264, -0.0, 0.0, -0.06536, 0.0, -0.0, -0.0, -0.0, 0.0, 0.06536, 0.0, 0.0, 0.0264, 0.0, 0.0, -0.03857, 0.0, 0.0, 0.00236, 0.0, 0.0, -0.07483, 0.0, -0.0, 0.00236, 0.0, 0.0, -0.03857, 0.0, -0.0, 0.0264, 0.0, -0.0, 0.06536, -0.0, 0.0, 0.0, 0.0, -0.0, -0.0799, -0.0, 0.0, -0.0799, -0.0, 0.0, 0.0799, 0.0, -0.0, 0.0799]]), "Eux":array([[-0.02584, 0.00612, 0.0, 0.006, 0.00107, 0.0, 0.006, -0.00107, -0.0, -0.02584, -0.00612, -0.0, -0.0386, -0.00557, -0.0, -0.00221, -0.00292, -0.0, 0.11776, -0.00079, -0.0, 0.11776, 0.00079, -0.0, -0.00221, 0.00292, -0.0, -0.0386, 0.00557, 0.0, -0.02584, 0.00612, 0.0, 0.006, 0.00107, 0.0, 0.006, -0.00107, -0.0, -0.02584, -0.00612, -0.0, -0.0386, -0.00557, -0.0, -0.00221, -0.00292, 0.0, 0.11776, -0.00079, 0.0, 0.11776, 0.00079, 0.0, -0.00221, 0.00292, 0.0, -0.0386, 0.00557, 0.0, -0.04299, 0.0, -0.0, -0.07084, 0.0, 0.0, -0.04299, 0.0, 0.0, -0.07084, 0.0, -0.0], [0.01473, 0.01626, 0.0, 0.09259, 0.00963, 0.0, 0.09259, -0.00963, -0.0, 0.01473, -0.01626, -0.0, -0.0259, -0.03893, -0.0, -0.02743, -0.01604, -0.0, -0.02673, -0.01251, 0.0, -0.02673, 0.01251, 0.0, -0.02743, 0.01604, 0.0, -0.0259, 0.03893, 0.0, 0.01473, 0.01626, 0.0, 0.09259, 0.00963, 0.0, 0.09259, -0.00963, -0.0, 0.01473, -0.01626, -0.0, -0.0259, -0.03893, -0.0, -0.02743, -0.01604, -0.0, -0.02673, -0.01251, -0.0, -0.02673, 0.01251, -0.0, -0.02743, 0.01604, 0.0, -0.0259, 0.03893, 0.0, -0.02712, 0.0, 0.0, -0.02671, 0.0, -0.0, -0.02712, 0.0, -0.0, -0.02671, 0.0, 0.0]]), "Euy":array([[0.00222, -0.00343, -0.0, -0.00041, 0.10343, -0.0, 0.00041, 0.10343, -0.0, -0.00222, -0.00343, -0.0, -0.01041, -0.04086, 0.0, -0.00611, -0.02641, 0.0, -0.00516, 0.01956, 0.0, 0.00516, 0.01956, -0.0, 0.00611, -0.02641, -0.0, 0.01041, -0.04086, -0.0, 0.00222, -0.00343, 0.0, -0.00041, 0.10343, 0.0, 0.00041, 0.10343, 0.0, -0.00222, -0.00343, 0.0, -0.01041, -0.04086, 0.0, -0.00611, -0.02641, 0.0, -0.00516, 0.01956, 0.0, 0.00516, 0.01956, -0.0, 0.00611, -0.02641, -0.0, 0.01041, -0.04086, -0.0, -0.0, -0.05383, 0.0, -0.0, -0.05216, -0.0, -0.0, -0.05383, -0.0, -0.0, -0.05216, -0.0], [0.01517, -0.02419, -0.0, 0.01, -0.03957, -0.0, -0.01, -0.03957, 0.0, -0.01517, -0.02419, 0.0, -0.04088, -0.0223, 0.0, -0.01592, 0.01378, 0.0, -0.01183, 0.09247, 0.0, 0.01183, 0.09247, -0.0, 0.01592, 0.01378, -0.0, 0.04088, -0.0223, -0.0, 0.01517, -0.02419, -0.0, 0.01, -0.03957, -0.0, -0.01, -0.03957, 0.0, -0.01517, -0.02419, 0.0, -0.04088, -0.0223, 0.0, -0.01592, 0.01378, 0.0, -0.01183, 0.09247, 0.0, 0.01183, 0.09247, -0.0, 0.01592, 0.01378, -0.0, 0.04088, -0.0223, -0.0, -0.0, -0.00958, 0.0, 0.0, -0.02999, 0.0, -0.0, -0.00958, -0.0, 0.0, -0.02999, 0.0]]), "Eux+y":array([[0.0123, -0.00206, -0.0, -0.01071, -0.06776, -0.0, -0.00912, -0.0655, -0.0, 0.01651, 0.00716, -0.0, 0.03506, 0.03506, -0.0, 0.00716, 0.01651, 0.0, -0.0655, -0.00912, 0.0, -0.06776, -0.01071, 0.0, -0.00206, 0.0123, 0.0, 0.01841, 0.01841, -0.0, 0.0123, -0.00206, -0.0, -0.01071, -0.06776, -0.0, -0.00912, -0.0655, -0.0, 0.01651, 0.00716, -0.0, 0.03506, 0.03506, 0.0, 0.00716, 0.01651, 0.0, -0.0655, -0.00912, 0.0, -0.06776, -0.01071, 0.0, -0.00206, 0.0123, 0.0, 0.01841, 0.01841, -0.0, 0.02697, 0.03848, 0.0, 0.03848, 0.02697, -0.0, 0.02697, 0.03848, 0.0, 0.03848, 0.02697, -0.0], [0.02064, -0.00644, -0.0, 0.0707, -0.0182, -0.0, 0.05673, -0.03316, -0.0, -0.00136, -0.02933, -0.0, -0.04276, -0.04276, 0.0, -0.02933, -0.00136, 0.0, -0.03316, 0.05673, 0.0, -0.0182, 0.0707, 0.0, -0.00644, 0.02064, 0.0, 0.01192, 0.01192, -0.0, 0.02064, -0.00644, -0.0, 0.0707, -0.0182, -0.0, 0.05673, -0.03316, -0.0, -0.00136, -0.02933, -0.0, -0.04276, -0.04276, -0.0, -0.02933, -0.00136, 0.0, -0.03316, 0.05673, 0.0, -0.0182, 0.0707, 0.0, -0.00644, 0.02064, 0.0, 0.01192, 0.01192, 0.0, -0.01952, -0.00932, 0.0, -0.00932, -0.01952, -0.0, -0.01952, -0.00932, 0.0, -0.00932, -0.01952, -0.0]]), "Eux-y":array([[-0.01827, 0.00296, -0.0, 0.00432, -0.06617, -0.0, 0.00273, -0.07082, -0.0, -0.01681, -0.00213, -0.0, -0.0205, 0.0205, 0.0, 0.00213, 0.01681, 0.0, 0.07082, -0.00273, 0.0, 0.06617, -0.00432, 0.0, -0.00296, 0.01827, 0.0, -0.0215, 0.0215, -0.0, -0.01827, 0.00296, -0.0, 0.00432, -0.06617, -0.0, 0.00273, -0.07082, -0.0, -0.01681, -0.00213, -0.0, -0.0205, 0.0205, 0.0, 0.00213, 0.01681, 0.0, 0.07082, -0.00273, 0.0, 0.06617, -0.00432, 0.0, -0.00296, 0.01827, 0.0, -0.0215, 0.0215, 0.0, -0.0299, 0.03712, -0.0, -0.03712, 0.0299, 0.0, -0.0299, 0.03712, 0.0, -0.03712, 0.0299, -0.0], [0.00194, -0.02819, -0.0, -0.05667, -0.02523, -0.0, -0.07264, -0.00915, -0.0, -0.01874, -0.00664, -0.0, -0.0087, 0.0087, -0.0, 0.00664, 0.01874, 0.0, 0.00915, 0.07264, 0.0, 0.02523, 0.05667, 0.0, 0.02819, -0.00194, 0.0, 0.04596, -0.04596, -0.0, 0.00194, -0.02819, -0.0, -0.05667, -0.02523, -0.0, -0.07264, -0.00915, -0.0, -0.01874, -0.00664, -0.0, -0.0087, 0.0087, 0.0, 0.00664, 0.01874, 0.0, 0.00915, 0.07264, 0.0, 0.02523, 0.05667, 0.0, 0.02819, -0.00194, 0.0, 0.04596, -0.04596, 0.0, 0.02308, -0.01576, 0.0, 0.01576, -0.02308, -0.0, 0.02308, -0.01576, 0.0, 0.01576, -0.02308, -0.0]])}, database_path = 'porphyrin_scsd_20220808.pkl', maxdist = 1.75)
#model_objs_dict['porphyrin'] = scsd_model('porphyrin', array([[-1.09, 2.827, 0.0], [-0.6767, 4.203, 0.0], [0.6767, 4.203, 0.0], [1.09, 2.827, 0.0], [2.413, 2.413, 0.0], [2.827, 1.09, 0.0], [4.203, 0.6767, 0.0], [4.203, -0.6767, 0.0], [2.827, -1.09, 0.0], [2.413, -2.413, 0.0], [1.09, -2.827, 0.0], [0.6767, -4.203, 0.0], [-0.6767, -4.203, 0.0], [-1.09, -2.827, 0.0], [-2.413, -2.413, 0.0], [-2.827, -1.09, 0.0], [-4.203, -0.6767, 0.0], [-4.203, 0.6767, 0.0], [-2.827, 1.09, 0.0], [-2.413, 2.413, 0.0], [0.0, 1.998, 0.0], [1.998, 0.0, 0.0], [0.0, -1.998, 0.0], [-1.998, 0.0, 0.0]]), 'D4h', database_path = 'porphyrin_scsd_20220713.pkl', maxdist = 1.75) # pca = {"A1g":array([[-0.00063, -0.01855, -0.0, -0.00216, -0.10655, -0.0, 0.00216, -0.10655, -0.0, 0.00063, -0.01855, -0.0, 0.00354, 0.00354, 0.0, -0.01855, 0.00063, 0.0, -0.10655, 0.00216, 0.0, -0.10655, -0.00216, -0.0, -0.01855, -0.00063, -0.0, 0.00354, -0.00354, -0.0, 0.00063, 0.01855, 0.0, 0.00216, 0.10655, 0.0, -0.00216, 0.10655, 0.0, -0.00063, 0.01855, 0.0, -0.00354, -0.00354, -0.0, 0.01855, -0.00063, -0.0, 0.10655, -0.00216, -0.0, 0.10655, 0.00216, 0.0, 0.01855, 0.00063, 0.0, -0.00354, 0.00354, 0.0, 0.0, 0.03021, 0.0, 0.03021, -0.0, -0.0, -0.0, -0.03021, -0.0, -0.03021, 0.0, 0.0], [0.02351, -0.04714, -0.0, 0.01555, -0.0053, -0.0, -0.01555, -0.0053, -0.0, -0.02351, -0.04714, -0.0, -0.04804, -0.04804, -0.0, -0.04714, -0.02351, -0.0, -0.0053, -0.01555, -0.0, -0.0053, 0.01555, 0.0, -0.04714, 0.02351, 0.0, -0.04804, 0.04804, 0.0, -0.02351, 0.04714, 0.0, -0.01555, 0.0053, 0.0, 0.01555, 0.0053, 0.0, 0.02351, 0.04714, 0.0, 0.04804, 0.04804, 0.0, 0.04714, 0.02351, 0.0, 0.0053, 0.01555, 0.0, 0.0053, -0.01555, -0.0, 0.04714, -0.02351, -0.0, 0.04804, -0.04804, -0.0, -0.0, -0.08086, -0.0, -0.08086, 0.0, 0.0, 0.0, 0.08086, 0.0, 0.08086, -0.0, -0.0]]), "A2g":array([[0.0145, 0.02851, 0.0, -0.05832, 0.01925, 0.0, -0.05832, -0.01925, 0.0, 0.0145, -0.02851, 0.0, 0.0487, -0.0487, 0.0, 0.02851, -0.0145, 0.0, 0.01925, 0.05832, 0.0, -0.01925, 0.05832, -0.0, -0.02851, -0.0145, -0.0, -0.0487, -0.0487, -0.0, -0.0145, -0.02851, -0.0, 0.05832, -0.01925, -0.0, 0.05832, 0.01925, -0.0, -0.0145, 0.02851, -0.0, -0.0487, 0.0487, -0.0, -0.02851, 0.0145, -0.0, -0.01925, -0.05832, -0.0, 0.01925, -0.05832, 0.0, 0.02851, 0.0145, 0.0, 0.0487, 0.0487, 0.0, 0.05106, -0.0, 0.0, -0.0, -0.05106, 0.0, -0.05106, 0.0, -0.0, 0.0, 0.05106, -0.0], [0.04571, 0.01474, 0.0, 0.06646, 0.00859, -0.0, 0.06646, -0.00859, -0.0, 0.04571, -0.01474, 0.0, 0.03642, -0.03642, 0.0, 0.01474, -0.04571, 0.0, 0.00859, -0.06646, 0.0, -0.00859, -0.06646, -0.0, -0.01474, -0.04571, -0.0, -0.03642, -0.03642, -0.0, -0.04571, -0.01474, -0.0, -0.06646, -0.00859, 0.0, -0.06646, 0.00859, 0.0, -0.04571, 0.01474, -0.0, -0.03642, 0.03642, -0.0, -0.01474, 0.04571, -0.0, -0.00859, 0.06646, -0.0, 0.00859, 0.06646, 0.0, 0.01474, 0.04571, 0.0, 0.03642, 0.03642, 0.0, 0.03344, 0.0, 0.0, 0.0, -0.03344, 0.0, -0.03344, -0.0, -0.0, -0.0, 0.03344, -0.0]]), "B1g":array([[-0.00015, -0.03554, -0.0, 0.00029, -0.10249, -0.0, -0.00029, -0.10249, -0.0, 0.00015, -0.03554, -0.0, 0.00404, -0.00404, -0.0, 0.03554, -0.00015, -0.0, 0.10249, 0.00029, -0.0, 0.10249, -0.00029, -0.0, 0.03554, 0.00015, -0.0, 0.00404, 0.00404, 0.0, 0.00015, 0.03554, 0.0, -0.00029, 0.10249, 0.0, 0.00029, 0.10249, 0.0, -0.00015, 0.03554, 0.0, -0.00404, 0.00404, 0.0, -0.03554, 0.00015, 0.0, -0.10249, -0.00029, 0.0, -0.10249, 0.00029, 0.0, -0.03554, -0.00015, 0.0, -0.00404, -0.00404, -0.0, 0.0, 0.0029, 0.0, -0.0029, 0.0, 0.0, -0.0, -0.0029, -0.0, 0.0029, -0.0, -0.0], [-0.01423, 0.06145, 0.0, -0.00462, -0.02006, -0.0, 0.00462, -0.02006, -0.0, 0.01423, 0.06145, 0.0, -0.01018, 0.01018, 0.0, -0.06145, -0.01423, 0.0, 0.02006, -0.00462, -0.0, 0.02006, 0.00462, -0.0, -0.06145, 0.01423, 0.0, -0.01018, -0.01018, 0.0, 0.01423, -0.06145, -0.0, 0.00462, 0.02006, 0.0, -0.00462, 0.02006, 0.0, -0.01423, -0.06145, -0.0, 0.01018, -0.01018, -0.0, 0.06145, 0.01423, -0.0, -0.02006, 0.00462, 0.0, -0.02006, -0.00462, 0.0, 0.06145, -0.01423, -0.0, 0.01018, 0.01018, -0.0, 0.0, 0.11611, 0.0, -0.11611, 0.0, 0.0, -0.0, -0.11611, -0.0, 0.11611, -0.0, -0.0]]), "B2g":array([[0.03631, -0.02862, -0.0, -0.02138, -0.01908, -0.0, -0.02138, 0.01908, 0.0, 0.03631, 0.02862, 0.0, 0.05582, 0.05582, 0.0, 0.02862, 0.03631, 0.0, 0.01908, -0.02138, -0.0, -0.01908, -0.02138, -0.0, -0.02862, 0.03631, 0.0, -0.05582, 0.05582, 0.0, -0.03631, 0.02862, 0.0, 0.02138, 0.01908, 0.0, 0.02138, -0.01908, -0.0, -0.03631, -0.02862, -0.0, -0.05582, -0.05582, -0.0, -0.02862, -0.03631, -0.0, -0.01908, 0.02138, 0.0, 0.01908, 0.02138, 0.0, 0.02862, -0.03631, -0.0, 0.05582, -0.05582, -0.0, 0.0664, -0.0, -0.0, -0.0, 0.0664, 0.0, -0.0664, 0.0, 0.0, 0.0, -0.0664, -0.0], [0.03469, -0.00881, -0.0, 0.09751, -0.00478, -0.0, 0.09751, 0.00478, 0.0, 0.03469, 0.00881, 0.0, 0.01053, 0.01053, 0.0, 0.00881, 0.03469, 0.0, 0.00478, 0.09751, 0.0, -0.00478, 0.09751, 0.0, -0.00881, 0.03469, 0.0, -0.01053, 0.01053, 0.0, -0.03469, 0.00881, 0.0, -0.09751, 0.00478, 0.0, -0.09751, -0.00478, -0.0, -0.03469, -0.00881, -0.0, -0.01053, -0.01053, -0.0, -0.00881, -0.03469, -0.0, -0.00478, -0.09751, -0.0, 0.00478, -0.09751, -0.0, 0.00881, -0.03469, -0.0, 0.01053, -0.01053, -0.0, -0.00319, -0.0, -0.0, -0.0, -0.00319, -0.0, 0.00319, 0.0, 0.0, 0.0, 0.00319, 0.0]]), "A1u":array([[0.0, 0.0, 0.0608, -0.0, -0.0, 0.11117, 0.0, -0.0, -0.11117, 0.0, -0.0, -0.0608, 0.0, 0.0, -0.0, -0.0, -0.0, 0.0608, -0.0, 0.0, 0.11117, -0.0, -0.0, -0.11117, 0.0, -0.0, -0.0608, -0.0, -0.0, 0.0, -0.0, -0.0, 0.0608, 0.0, 0.0, 0.11117, -0.0, 0.0, -0.11117, -0.0, 0.0, -0.0608, -0.0, -0.0, 0.0, 0.0, 0.0, 0.0608, 0.0, -0.0, 0.11117, 0.0, 0.0, -0.11117, -0.0, 0.0, -0.0608, 0.0, 0.0, -0.0, -0.0, 0.0, -0.0, 0.0, 0.0, -0.0, 0.0, -0.0, 0.0, -0.0, -0.0, 0.0], [0.0, 0.0, -0.11637, -0.0, 0.0, 0.06364, 0.0, 0.0, -0.06364, 0.0, -0.0, 0.11637, -0.0, -0.0, 0.0, 0.0, -0.0, -0.11637, -0.0, 0.0, 0.06364, 0.0, -0.0, -0.06364, 0.0, -0.0, 0.11637, -0.0, -0.0, -0.0, -0.0, 0.0, -0.11637, 0.0, -0.0, 0.06364, -0.0, -0.0, -0.06364, -0.0, 0.0, 0.11637, 0.0, 0.0, -0.0, -0.0, 0.0, -0.11637, 0.0, -0.0, 0.06364, -0.0, 0.0, -0.06364, -0.0, 0.0, 0.11637, 0.0, 0.0, 0.0, -0.0, -0.0, 0.0, -0.0, 0.0, 0.0, 0.0, 0.0, -0.0, 0.0, -0.0, -0.0]]), "A2u":array([[-0.0, -0.0, 0.01859, 0.0, -0.0, -0.09746, 0.0, 0.0, -0.09746, -0.0, -0.0, 0.01859, 0.0, -0.0, 0.06905, -0.0, -0.0, 0.01859, 0.0, 0.0, -0.09746, -0.0, 0.0, -0.09746, -0.0, -0.0, 0.01859, -0.0, 0.0, 0.06905, 0.0, 0.0, 0.01859, 0.0, 0.0, -0.09746, 0.0, 0.0, -0.09746, 0.0, 0.0, 0.01859, -0.0, 0.0, 0.06905, 0.0, 0.0, 0.01859, 0.0, 0.0, -0.09746, -0.0, 0.0, -0.09746, 0.0, 0.0, 0.01859, 0.0, -0.0, 0.06905, -0.0, 0.0, 0.08868, 0.0, -0.0, 0.08868, -0.0, 0.0, 0.08868, 0.0, -0.0, 0.08868], [0.0, -0.0, -0.04292, -0.0, 0.0, 0.00796, -0.0, -0.0, 0.00796, 0.0, 0.0, -0.04292, -0.0, -0.0, 0.15614, 0.0, 0.0, -0.04292, -0.0, 0.0, 0.00796, -0.0, 0.0, 0.00796, 0.0, -0.0, -0.04292, 0.0, 0.0, 0.15614, 0.0, -0.0, -0.04292, 0.0, -0.0, 0.00796, 0.0, 0.0, 0.00796, -0.0, -0.0, -0.04292, 0.0, 0.0, 0.15614, -0.0, -0.0, -0.04292, 0.0, 0.0, 0.00796, -0.0, -0.0, 0.00796, -0.0, 0.0, -0.04292, -0.0, -0.0, 0.15614, -0.0, 0.0, -0.08611, 0.0, -0.0, -0.08611, -0.0, 0.0, -0.08611, 0.0, -0.0, -0.08611]]), "B1u":array([[-0.0, -0.0, 0.07027, -0.0, -0.0, 0.04695, 0.0, -0.0, -0.04695, -0.0, 0.0, -0.07027, -0.0, -0.0, -0.1352, 0.0, -0.0, -0.07027, -0.0, 0.0, -0.04695, -0.0, -0.0, 0.04695, -0.0, -0.0, 0.07027, 0.0, 0.0, 0.1352, -0.0, 0.0, 0.07027, 0.0, 0.0, 0.04695, -0.0, 0.0, -0.04695, 0.0, -0.0, -0.07027, 0.0, 0.0, -0.1352, -0.0, 0.0, -0.07027, 0.0, -0.0, -0.04695, 0.0, 0.0, 0.04695, 0.0, 0.0, 0.07027, -0.0, 0.0, 0.1352, -0.0, -0.0, 0.0, 0.0, -0.0, -0.0, 0.0, 0.0, -0.0, -0.0, 0.0, 0.0], [0.0, -0.0, 0.04881, 0.0, 0.0, 0.08857, -0.0, 0.0, -0.08857, -0.0, -0.0, -0.04881, -0.0, 0.0, 0.11225, -0.0, -0.0, -0.04881, 0.0, 0.0, -0.08857, 0.0, 0.0, 0.08857, -0.0, -0.0, 0.04881, -0.0, -0.0, -0.11225, -0.0, 0.0, 0.04881, -0.0, -0.0, 0.08857, 0.0, -0.0, -0.08857, 0.0, 0.0, -0.04881, -0.0, 0.0, 0.11225, 0.0, 0.0, -0.04881, -0.0, 0.0, -0.08857, -0.0, -0.0, 0.08857, 0.0, 0.0, 0.04881, -0.0, -0.0, -0.11225, -0.0, -0.0, 0.0, 0.0, -0.0, 0.0, 0.0, 0.0, -0.0, -0.0, 0.0, -0.0]]), "B2u":array([[0.0, -0.0, -0.04257, -0.0, 0.0, -0.12098, -0.0, 0.0, -0.12098, 0.0, -0.0, -0.04257, -0.0, -0.0, 0.0, -0.0, -0.0, 0.04257, 0.0, -0.0, 0.12098, 0.0, -0.0, 0.12098, -0.0, -0.0, 0.04257, -0.0, 0.0, 0.0, -0.0, 0.0, -0.04257, -0.0, -0.0, -0.12098, -0.0, -0.0, -0.12098, -0.0, 0.0, -0.04257, 0.0, 0.0, -0.0, 0.0, 0.0, 0.04257, -0.0, -0.0, 0.12098, -0.0, -0.0, 0.12098, 0.0, 0.0, 0.04257, 0.0, -0.0, -0.0, 0.0, -0.0, 2e-05, -0.0, -0.0, -2e-05, -0.0, 0.0, 2e-05, 0.0, 0.0, -2e-05], [-0.0, 0.0, 0.06456, -0.0, -0.0, -0.02271, -0.0, -0.0, -0.02271, 0.0, 0.0, 0.06456, 0.0, 0.0, 0.0, 0.0, -0.0, -0.06456, -0.0, 0.0, 0.02271, -0.0, -0.0, 0.02271, 0.0, -0.0, -0.06456, 0.0, -0.0, 0.0, 0.0, -0.0, 0.06456, -0.0, 0.0, -0.02271, -0.0, 0.0, -0.02271, -0.0, -0.0, 0.06456, -0.0, -0.0, -0.0, -0.0, 0.0, -0.06456, 0.0, -0.0, 0.02271, 0.0, -0.0, 0.02271, -0.0, 0.0, -0.06456, -0.0, 0.0, -0.0, 0.0, 0.0, 0.1679, 0.0, 0.0, -0.1679, 0.0, -0.0, 0.1679, -0.0, 0.0, -0.1679]]), "Egx":array([[0.0, -0.0, -0.02943, 0.0, 0.0, 0.12197, 0.0, 0.0, 0.12197, 0.0, -0.0, -0.02943, 0.0, 0.0, -0.10056, 0.0, 0.0, -0.05077, -0.0, 0.0, -0.03347, -0.0, 0.0, 0.03347, 0.0, 0.0, 0.05077, 0.0, -0.0, 0.10056, 0.0, 0.0, 0.02943, -0.0, -0.0, -0.12197, -0.0, -0.0, -0.12197, -0.0, 0.0, 0.02943, -0.0, -0.0, 0.10056, -0.0, -0.0, 0.05077, 0.0, -0.0, 0.03347, 0.0, -0.0, -0.03347, -0.0, -0.0, -0.05077, -0.0, 0.0, -0.10056, -0.0, 0.0, -0.1175, 0.0, -0.0, 0.0, 0.0, 0.0, 0.1175, 0.0, 0.0, -0.0], [0.0, 0.0, -0.06454, 0.0, 0.0, 0.01395, -0.0, 0.0, 0.01395, 0.0, 0.0, -0.06454, -0.0, 0.0, 0.07368, 0.0, -0.0, 0.06955, 0.0, 0.0, 0.07484, 0.0, 0.0, -0.07484, 0.0, -0.0, -0.06955, -0.0, 0.0, -0.07368, 0.0, -0.0, 0.06454, -0.0, -0.0, -0.01395, 0.0, -0.0, -0.01395, -0.0, -0.0, 0.06454, 0.0, -0.0, -0.07368, -0.0, 0.0, -0.06955, -0.0, -0.0, -0.07484, -0.0, -0.0, 0.07484, -0.0, 0.0, 0.06955, 0.0, -0.0, 0.07368, -0.0, 0.0, -0.16757, 0.0, -0.0, 0.0, 0.0, 0.0, 0.16757, 0.0, 0.0, -0.0]]), "Egy":array([[0.0, -0.0, 0.04801, -0.0, 0.0, 0.0318, 0.0, 0.0, -0.0318, -0.0, 0.0, -0.04801, 0.0, -0.0, -0.09395, -0.0, -0.0, -0.02927, -0.0, -0.0, 0.11669, -0.0, -0.0, 0.11669, -0.0, 0.0, -0.02927, -0.0, -0.0, -0.09395, -0.0, 0.0, -0.04801, 0.0, -0.0, -0.0318, -0.0, -0.0, 0.0318, 0.0, -0.0, 0.04801, -0.0, 0.0, 0.09395, 0.0, 0.0, 0.02927, 0.0, 0.0, -0.11669, 0.0, 0.0, -0.11669, 0.0, -0.0, 0.02927, 0.0, 0.0, 0.09395, -0.0, 0.0, -0.0, 0.0, 0.0, -0.11443, 0.0, 0.0, 0.0, 0.0, -0.0, 0.11443], [0.0, -0.0, -0.06235, 0.0, 0.0, -0.06581, -0.0, 0.0, 0.06581, 0.0, -0.0, 0.06235, -0.0, 0.0, 0.07713, -0.0, 0.0, -0.07558, 0.0, -0.0, 0.00265, 0.0, -0.0, 0.00265, -0.0, 0.0, -0.07558, 0.0, 0.0, 0.07713, 0.0, -0.0, 0.06235, -0.0, -0.0, 0.06581, 0.0, -0.0, -0.06581, -0.0, 0.0, -0.06235, 0.0, -0.0, -0.07713, 0.0, -0.0, 0.07558, -0.0, 0.0, -0.00265, -0.0, 0.0, -0.00265, 0.0, -0.0, 0.07558, -0.0, -0.0, -0.07713, 0.0, -0.0, 0.0, -0.0, 0.0, -0.17149, -0.0, -0.0, -0.0, -0.0, -0.0, 0.17149]]), "Egx+y":array([[-0.0, 0.0, 0.05306, 0.0, -0.0, -0.05986, 0.0, -0.0, -0.1034, -0.0, -0.0, -0.01351, -0.0, -0.0, -0.0, -0.0, -0.0, 0.01351, -0.0, 0.0, 0.1034, -0.0, 0.0, 0.05986, 0.0, 0.0, -0.05306, -0.0, 0.0, -0.13161, 0.0, -0.0, -0.05306, -0.0, 0.0, 0.05986, -0.0, 0.0, 0.1034, 0.0, 0.0, 0.01351, 0.0, 0.0, 0.0, 0.0, 0.0, -0.01351, 0.0, -0.0, -0.1034, 0.0, -0.0, -0.05986, -0.0, -0.0, 0.05306, 0.0, -0.0, 0.13161, -0.0, -0.0, 0.07941, -0.0, -0.0, -0.07941, 0.0, 0.0, -0.07941, 0.0, 0.0, 0.07941], [0.0, -0.0, -0.00834, -0.0, -0.0, 0.04701, -0.0, -0.0, -0.05439, -0.0, 0.0, -0.09888, 0.0, -0.0, -0.0, 0.0, 0.0, 0.09888, -0.0, 0.0, 0.05439, 0.0, 0.0, -0.04701, 0.0, 0.0, 0.00834, -0.0, -0.0, -0.09506, -0.0, 0.0, 0.00834, 0.0, 0.0, -0.04701, 0.0, 0.0, 0.05439, 0.0, -0.0, 0.09888, -0.0, 0.0, 0.0, -0.0, -0.0, -0.09888, 0.0, -0.0, -0.05439, -0.0, -0.0, 0.04701, -0.0, -0.0, -0.00834, 0.0, 0.0, 0.09506, 0.0, 0.0, -0.12541, -0.0, 0.0, 0.12541, -0.0, -0.0, 0.12541, 0.0, -0.0, -0.12541]]), "Egx-y":array([[0.0, 0.0, 0.01432, -0.0, 0.0, 0.10708, -0.0, 0.0, 0.06008, -0.0, -0.0, -0.05606, 0.0, 0.0, -0.13802, 0.0, -0.0, -0.05606, 0.0, -0.0, 0.06008, 0.0, -0.0, 0.10708, 0.0, -0.0, 0.01432, 0.0, -0.0, -0.0, -0.0, -0.0, -0.01432, 0.0, -0.0, -0.10708, 0.0, -0.0, -0.06008, 0.0, 0.0, 0.05606, -0.0, -0.0, 0.13802, -0.0, 0.0, 0.05606, -0.0, 0.0, -0.06008, -0.0, 0.0, -0.10708, -0.0, 0.0, -0.01432, -0.0, 0.0, 0.0, 0.0, -0.0, -0.08107, -0.0, 0.0, -0.08107, -0.0, 0.0, 0.08107, 0.0, -0.0, 0.08107], [0.0, -0.0, -0.09897, -0.0, -0.0, -0.03997, -0.0, -0.0, 0.0584, -0.0, -0.0, -0.00357, -0.0, -0.0, 0.11331, -0.0, 0.0, -0.00357, -0.0, -0.0, 0.0584, -0.0, 0.0, -0.03997, -0.0, 0.0, -0.09897, 0.0, -0.0, -0.0, -0.0, 0.0, 0.09897, 0.0, 0.0, 0.03997, 0.0, 0.0, -0.0584, 0.0, 0.0, 0.00357, 0.0, 0.0, -0.11331, 0.0, -0.0, 0.00357, 0.0, 0.0, -0.0584, 0.0, -0.0, 0.03997, 0.0, -0.0, 0.09897, -0.0, 0.0, 0.0, 0.0, -0.0, -0.12098, -0.0, 0.0, -0.12098, -0.0, 0.0, 0.12098, 0.0, -0.0, 0.12098]]), "Eux":array([[-0.0336, 0.00795, 0.0, 0.00781, 0.0014, 0.0, 0.00781, -0.0014, -0.0, -0.0336, -0.00795, -0.0, -0.05019, -0.00724, -0.0, -0.00287, -0.0038, -0.0, 0.15311, -0.00102, -0.0, 0.15311, 0.00102, -0.0, -0.00287, 0.0038, -0.0, -0.05019, 0.00724, 0.0, -0.0336, 0.00795, 0.0, 0.00781, 0.0014, 0.0, 0.00781, -0.0014, -0.0, -0.0336, -0.00795, -0.0, -0.05019, -0.00724, -0.0, -0.00287, -0.0038, 0.0, 0.15311, -0.00102, 0.0, 0.15311, 0.00102, 0.0, -0.00287, 0.0038, 0.0, -0.05019, 0.00724, 0.0, -0.0559, 0.0, -0.0, -0.0921, 0.0, 0.0, -0.0559, 0.0, 0.0, -0.0921, 0.0, -0.0], [0.01776, 0.0196, 0.0, 0.11159, 0.0116, 0.0, 0.11159, -0.0116, -0.0, 0.01776, -0.0196, -0.0, -0.03121, -0.04692, -0.0, -0.03306, -0.01933, -0.0, -0.03222, -0.01508, 0.0, -0.03222, 0.01508, 0.0, -0.03306, 0.01933, 0.0, -0.03121, 0.04692, 0.0, 0.01776, 0.0196, 0.0, 0.11159, 0.0116, 0.0, 0.11159, -0.0116, -0.0, 0.01776, -0.0196, -0.0, -0.03121, -0.04692, -0.0, -0.03306, -0.01933, -0.0, -0.03222, -0.01508, -0.0, -0.03222, 0.01508, -0.0, -0.03306, 0.01933, 0.0, -0.03121, 0.04692, 0.0, -0.03269, 0.0, 0.0, -0.03219, 0.0, -0.0, -0.03269, 0.0, -0.0, -0.03219, 0.0, 0.0]]), "Euy":array([[0.00295, -0.00456, -0.0, -0.00055, 0.13758, -0.0, 0.00055, 0.13758, -0.0, -0.00295, -0.00456, -0.0, -0.01385, -0.05435, 0.0, -0.00813, -0.03513, 0.0, -0.00687, 0.02602, 0.0, 0.00687, 0.02602, -0.0, 0.00813, -0.03513, -0.0, 0.01385, -0.05435, -0.0, 0.00295, -0.00456, 0.0, -0.00055, 0.13758, 0.0, 0.00055, 0.13758, 0.0, -0.00295, -0.00456, 0.0, -0.01385, -0.05435, 0.0, -0.00813, -0.03513, 0.0, -0.00687, 0.02602, 0.0, 0.00687, 0.02602, -0.0, 0.00813, -0.03513, -0.0, 0.01385, -0.05435, -0.0, -0.0, -0.0716, 0.0, -0.0, -0.06938, -0.0, -0.0, -0.0716, -0.0, -0.0, -0.06938, -0.0], [0.01796, -0.02865, -0.0, 0.01184, -0.04685, -0.0, -0.01184, -0.04685, 0.0, -0.01796, -0.02865, 0.0, -0.0484, -0.02641, 0.0, -0.01885, 0.01631, 0.0, -0.014, 0.10949, 0.0, 0.014, 0.10949, -0.0, 0.01885, 0.01631, -0.0, 0.0484, -0.02641, -0.0, 0.01796, -0.02865, -0.0, 0.01184, -0.04685, -0.0, -0.01184, -0.04685, 0.0, -0.01796, -0.02865, 0.0, -0.0484, -0.02641, 0.0, -0.01885, 0.01631, 0.0, -0.014, 0.10949, 0.0, 0.014, 0.10949, -0.0, 0.01885, 0.01631, -0.0, 0.0484, -0.02641, -0.0, -0.0, -0.01135, 0.0, 0.0, -0.0355, 0.0, -0.0, -0.01135, -0.0, 0.0, -0.0355, 0.0]]), "Eux+y":array([[0.01445, -0.00242, -0.0, -0.01258, -0.07962, -0.0, -0.01072, -0.07696, -0.0, 0.0194, 0.00841, -0.0, 0.0412, 0.0412, -0.0, 0.00841, 0.0194, 0.0, -0.07696, -0.01072, 0.0, -0.07962, -0.01258, 0.0, -0.00242, 0.01445, 0.0, 0.02164, 0.02164, -0.0, 0.01445, -0.00242, -0.0, -0.01258, -0.07962, -0.0, -0.01072, -0.07696, -0.0, 0.0194, 0.00841, -0.0, 0.0412, 0.0412, 0.0, 0.00841, 0.0194, 0.0, -0.07696, -0.01072, 0.0, -0.07962, -0.01258, 0.0, -0.00242, 0.01445, 0.0, 0.02164, 0.02164, -0.0, 0.03169, 0.04521, 0.0, 0.04521, 0.03169, -0.0, 0.03169, 0.04521, 0.0, 0.04521, 0.03169, -0.0], [0.0232, -0.00724, -0.0, 0.07946, -0.02045, -0.0, 0.06377, -0.03727, -0.0, -0.00153, -0.03297, -0.0, -0.04806, -0.04806, 0.0, -0.03297, -0.00153, 0.0, -0.03727, 0.06377, 0.0, -0.02045, 0.07946, 0.0, -0.00724, 0.0232, 0.0, 0.0134, 0.0134, -0.0, 0.0232, -0.00724, -0.0, 0.07946, -0.02045, -0.0, 0.06377, -0.03727, -0.0, -0.00153, -0.03297, -0.0, -0.04806, -0.04806, -0.0, -0.03297, -0.00153, 0.0, -0.03727, 0.06377, 0.0, -0.02045, 0.07946, 0.0, -0.00724, 0.0232, 0.0, 0.0134, 0.0134, 0.0, -0.02194, -0.01048, 0.0, -0.01048, -0.02194, -0.0, -0.02194, -0.01048, 0.0, -0.01048, -0.02194, -0.0]]), "Eux-y":array([[-0.02157, 0.00349, -0.0, 0.0051, -0.07811, -0.0, 0.00322, -0.0836, -0.0, -0.01984, -0.00251, -0.0, -0.02421, 0.02421, 0.0, 0.00251, 0.01984, 0.0, 0.0836, -0.00322, 0.0, 0.07811, -0.0051, 0.0, -0.00349, 0.02157, 0.0, -0.02538, 0.02538, -0.0, -0.02157, 0.00349, -0.0, 0.0051, -0.07811, -0.0, 0.00322, -0.0836, -0.0, -0.01984, -0.00251, -0.0, -0.02421, 0.02421, 0.0, 0.00251, 0.01984, 0.0, 0.0836, -0.00322, 0.0, 0.07811, -0.0051, 0.0, -0.00349, 0.02157, 0.0, -0.02538, 0.02538, 0.0, -0.0353, 0.04383, -0.0, -0.04383, 0.0353, 0.0, -0.0353, 0.04383, 0.0, -0.04383, 0.0353, -0.0], [0.00229, -0.03331, -0.0, -0.06697, -0.02982, -0.0, -0.08584, -0.01081, -0.0, -0.02215, -0.00784, -0.0, -0.01029, 0.01029, -0.0, 0.00784, 0.02215, 0.0, 0.01081, 0.08584, 0.0, 0.02982, 0.06697, 0.0, 0.03331, -0.00229, 0.0, 0.05431, -0.05431, -0.0, 0.00229, -0.03331, -0.0, -0.06697, -0.02982, -0.0, -0.08584, -0.01081, -0.0, -0.02215, -0.00784, -0.0, -0.01029, 0.01029, 0.0, 0.00784, 0.02215, 0.0, 0.01081, 0.08584, 0.0, 0.02982, 0.06697, 0.0, 0.03331, -0.00229, 0.0, 0.05431, -0.05431, 0.0, 0.02727, -0.01863, 0.0, 0.01863, -0.02727, -0.0, 0.02727, -0.01863, 0.0, 0.01863, -0.02727, -0.0]])}
model_objs_dict['chlorin'] = scsd_model('chlorin', array([['1.0752', '-0.0', '-3.1518', 'C'], ['2.3722', '-0.0', '-2.7781', 'C'], ['2.8244', '-0.0', '-1.4898', 'C'], ['4.1639', '-0.0', '-1.0416', 'C'], ['4.2083', '0.0', '0.2755', 'C'], ['2.8066', '0.0', '0.6901', 'C'], ['2.3902', '0.0', '1.9932', 'C'], ['1.076', '0.0', '2.4407', 'C'], ['0.6744', '0.0', '3.8129', 'C'], ['-0.6744', '0.0', '3.8129', 'C'], ['-1.076', '0.0', '2.4407', 'C'], ['-2.3902', '0.0', '1.9932', 'C'], ['-2.8066', '0.0', '0.6901', 'C'], ['-4.2083', '0.0', '0.2755', 'C'], ['-4.1639', '-0.0', '-1.0416', 'C'], ['-2.8244', '-0.0', '-1.4898', 'C'], ['-2.3722', '-0.0', '-2.7781', 'C'], ['-1.0752', '-0.0', '-3.1518', 'C'], ['-1.9865', '-0.0', '-0.3806', 'N'], ['-0.0', '-0.0', '-2.3759', 'N'], ['1.9865', '-0.0', '-0.3806', 'N'], ['0.0', '0.0', '1.635', 'N']]), 'C2v' , pca = {"A1":array([[-0.04331, -0.0, 0.07057, -0.04533, -0.0, 0.03961, -0.03536, 0.0, 0.01896, -0.03037, -0.0, -0.003, -0.02918, -0.0, -0.03125, -0.03125, -0.0, -0.02786, -0.01994, 0.0, -0.03291, -0.01125, 0.0, -0.02707, -0.00866, 0.0, -0.0322, 0.00866, 0.0, -0.0322, 0.01125, 0.0, -0.02707, 0.01994, 0.0, -0.03291, 0.03125, -0.0, -0.02786, 0.02918, -0.0, -0.03125, 0.03037, -0.0, -0.003, 0.03536, 0.0, 0.01896, 0.04533, -0.0, 0.03961, 0.04331, -0.0, 0.07057, 0.04504, -0.0, -0.00188, -0.0, -0.0, 0.08384, -0.04504, -0.0, -0.00188, 0.0, 0.0, -0.02978], [-0.01949, 0.0, 0.05577, 0.02617, 0.0, -0.08304, 0.02807, -0.0, -0.04253, 0.03132, 0.0, -0.03071, 0.03499, 0.0, -0.00069, 0.02937, 0.0, -0.00175, 0.03789, -0.0, 0.0109, 0.01924, -0.0, 0.01191, 0.01347, 0.0, 0.00913, -0.01347, 0.0, 0.00913, -0.01924, -0.0, 0.01191, -0.03789, -0.0, 0.0109, -0.02937, 0.0, -0.00175, -0.03499, 0.0, -0.00069, -0.03132, 0.0, -0.03071, -0.02807, -0.0, -0.04253, -0.02617, 0.0, -0.08304, 0.01949, 0.0, 0.05577, -0.02323, 0.0, -0.02245, -0.0, -0.0, 0.17732, 0.02323, 0.0, -0.02245, -0.0, -0.0, 0.00962]]), "A2":array([[0.0, -0.08134, -0.0, -0.0, -0.10766, 0.0, -0.0, -0.04809, 0.0, -0.0, -0.01978, -0.0, -0.0, 0.0448, -0.0, -0.0, 0.0486, 0.0, -0.0, 0.07528, 0.0, -0.0, 0.04155, -0.0, 0.0, 0.02885, -0.0, -0.0, -0.02885, -0.0, -0.0, -0.04155, -0.0, -0.0, -0.07528, -0.0, -0.0, -0.0486, -0.0, -0.0, -0.0448, 0.0, -0.0, 0.01978, -0.0, -0.0, 0.04809, -0.0, -0.0, 0.10766, -0.0, 0.0, 0.08134, -0.0, -0.0, 0.00405, -0.0, 0.0, 0.0, -0.0, -0.0, -0.00405, 0.0, -0.0, -0.0, -0.0], [0.0, 0.09984, -0.0, -0.0, 0.03518, 0.0, -0.0, -0.02868, 0.0, -0.0, -0.08025, -0.0, -0.0, -0.03801, 0.0, -0.0, 0.02817, -0.0, -0.0, 0.07779, 0.0, -0.0, 0.0563, -0.0, 0.0, 0.0397, -0.0, -0.0, -0.0397, -0.0, -0.0, -0.0563, -0.0, -0.0, -0.07779, -0.0, -0.0, -0.02817, 0.0, -0.0, 0.03801, -0.0, -0.0, 0.08025, -0.0, -0.0, 0.02868, -0.0, -0.0, -0.03518, -0.0, 0.0, -0.09984, -0.0, -0.0, -0.01608, 0.0, 0.0, 0.0, -0.0, -0.0, 0.01608, -0.0, 0.0, -0.0, -0.0]]), "B1":array([[0.00444, 0.0, 0.10425, 0.03286, -0.0, 0.0992, 0.03935, -0.0, 0.0066, 0.00457, -0.0, -0.0403, -0.03516, 0.0, -0.05584, -0.03718, -0.0, -0.01826, -0.03802, 0.0, -0.0077, 0.00014, -0.0, 0.00433, 0.04285, 0.0, 0.00041, 0.04285, -0.0, -0.00041, 0.00014, 0.0, -0.00433, -0.03802, -0.0, 0.0077, -0.03718, 0.0, 0.01826, -0.03516, 0.0, 0.05584, 0.00457, 0.0, 0.0403, 0.03935, -0.0, -0.0066, 0.03286, -0.0, -0.0992, 0.00444, 0.0, -0.10425, 0.00937, -0.0, -0.01024, -0.03813, 0.0, -0.0, 0.00937, -0.0, 0.01024, -0.00831, 0.0, -0.0], [-0.05904, 0.0, 0.05148, -0.04427, 0.0, 0.02607, -0.03099, -0.0, 0.04255, 0.00636, -0.0, 0.0245, 0.03272, -0.0, 0.0281, 0.03607, -0.0, 0.03436, 0.04542, -0.0, 0.03134, 0.03365, 0.0, -2e-05, 0.01311, -0.0, 0.0126, 0.01311, -0.0, -0.0126, 0.03365, -0.0, 2e-05, 0.04542, 0.0, -0.03134, 0.03607, 0.0, -0.03436, 0.03272, 0.0, -0.0281, 0.00636, -0.0, -0.0245, -0.03099, 0.0, -0.04255, -0.04427, 0.0, -0.02607, -0.05904, 0.0, -0.05148, -0.00317, 0.0, -0.0373, -0.07678, 0.0, -0.0, -0.00317, -0.0, 0.0373, 0.01707, 0.0, -0.0]]), "B2":array([[-0.0, 0.0823, -0.0, 0.0, 0.03187, -0.0, -0.0, -0.01943, 0.0, -0.0, -0.08717, 0.0, -0.0, -0.0985, -0.0, -0.0, -0.03626, 0.0, -0.0, -0.00864, 0.0, -0.0, 0.02895, -0.0, -0.0, 0.07365, -0.0, -0.0, 0.07365, -0.0, 0.0, 0.02895, -0.0, -0.0, -0.00864, -0.0, -0.0, -0.03626, 0.0, -0.0, -0.0985, -0.0, -0.0, -0.08717, 0.0, -0.0, -0.01943, 0.0, -0.0, 0.03187, -0.0, -0.0, 0.0823, -0.0, -0.0, 0.00258, 0.0, -0.0, 0.05139, -0.0, -0.0, 0.00258, 0.0, -0.0, 0.00993, -0.0], [-0.0, 0.00431, -0.0, -0.0, 0.04716, 0.0, 0.0, 0.01468, 0.0, 0.0, 0.06552, 0.0, 0.0, 0.02187, -0.0, 0.0, -0.0467, 0.0, 0.0, -0.04383, -0.0, -0.0, -0.02165, 0.0, 0.0, 0.09644, 0.0, 0.0, 0.09644, 0.0, 0.0, -0.02165, 0.0, 0.0, -0.04383, -0.0, 0.0, -0.0467, 0.0, 0.0, 0.02187, -0.0, 0.0, 0.06552, 0.0, 0.0, 0.01468, 0.0, 0.0, 0.04716, -0.0, 0.0, 0.00431, 0.0, 0.0, -0.06348, 0.0, -0.0, -0.06523, 0.0, 0.0, -0.06348, -0.0, 0.0, -0.08346, 0.0]])}, database_path = 'chlorin_scsd_20220713.pkl', maxdist = 1.75)
model_objs_dict['porphycene'] = scsd_model('porphycene', array([['0.7052', '2.415', '-0.0', 'C'], ['1.5918', '3.5337', '-0.0', 'C'], ['2.8513', '3.0389', '-0.0', 'C'], ['2.7636', '1.6084', '-0.0', 'C'], ['3.8568', '0.7029', '-0.0', 'C'], ['3.8568', '-0.7029', '0.0', 'C'], ['2.7636', '-1.6084', '0.0', 'C'], ['2.8513', '-3.0389', '0.0', 'C'], ['1.5918', '-3.5337', '0.0', 'C'], ['0.7052', '-2.415', '0.0', 'C'], ['-0.7052', '-2.415', '0.0', 'C'], ['-1.5918', '-3.5337', '0.0', 'C'], ['-2.8513', '-3.0389', '0.0', 'C'], ['-2.7636', '-1.6084', '0.0', 'C'], ['-3.8568', '-0.7029', '0.0', 'C'], ['-3.8568', '0.7029', '-0.0', 'C'], ['-2.7636', '1.6084', '-0.0', 'C'], ['-2.8513', '3.0389', '-0.0', 'C'], ['-1.5918', '3.5337', '-0.0', 'C'], ['-0.7052', '2.415', '-0.0', 'C'], ['1.4336', '1.273', '-0.0', 'N'], ['1.4336', '-1.273', '0.0', 'N'], ['-1.4336', '-1.273', '0.0', 'N'], ['-1.4336', '1.273', '-0.0', 'N']]), 'D2h' , pca = {"Ag":array([[0.01767, -0.0537, 0.0, -0.05957, -0.03172, -0.0, -0.01712, -0.00833, -0.0, 0.02082, -0.042, -0.0, 0.03925, 0.02683, 0.0, 0.03925, -0.02683, 0.0, 0.02082, 0.042, 0.0, -0.01712, 0.00833, -0.0, -0.05957, 0.03172, -0.0, 0.01767, 0.0537, 0.0, -0.01767, 0.0537, 0.0, 0.05957, 0.03172, 0.0, 0.01712, 0.00833, 0.0, -0.02082, 0.042, 0.0, -0.03925, -0.02683, -0.0, -0.03925, 0.02683, -0.0, -0.02082, -0.042, -0.0, 0.01712, -0.00833, 0.0, 0.05957, -0.03172, 0.0, -0.01767, -0.0537, -0.0, 0.03792, -0.04239, -0.0, 0.03792, 0.04239, 0.0, -0.03792, 0.04239, 0.0, -0.03792, -0.04239, -0.0], [0.06596, 0.04769, -0.0, -0.03266, 0.02447, -0.0, 0.03075, -0.00344, 0.0, -0.01017, -0.04457, -0.0, -0.0268, 0.06249, 0.0, -0.0268, -0.06249, -0.0, -0.01017, 0.04457, 0.0, 0.03075, 0.00344, 0.0, -0.03266, -0.02447, -0.0, 0.06596, -0.04769, 0.0, -0.06596, -0.04769, -0.0, 0.03266, -0.02447, 0.0, -0.03075, 0.00344, -0.0, 0.01017, 0.04457, 0.0, 0.0268, -0.06249, -0.0, 0.0268, 0.06249, 0.0, 0.01017, -0.04457, -0.0, -0.03075, -0.00344, -0.0, 0.03266, 0.02447, 0.0, -0.06596, 0.04769, -0.0, -0.01269, 0.02546, 0.0, -0.01269, -0.02546, -0.0, 0.01269, -0.02546, -0.0, 0.01269, 0.02546, 0.0]]), "B1g":array([[0.04047, -0.07455, 0.0, -0.01335, 0.03267, 0.0, -0.01402, 0.0315, 0.0, -0.00365, -0.03877, 0.0, -0.00378, -0.03053, -0.0, 0.00378, -0.03053, -0.0, 0.00365, -0.03877, -0.0, 0.01402, 0.0315, 0.0, 0.01335, 0.03267, 0.0, -0.04047, -0.07455, 0.0, -0.04047, 0.07455, -0.0, 0.01335, -0.03267, -0.0, 0.01402, -0.0315, -0.0, 0.00365, 0.03877, -0.0, 0.00378, 0.03053, 0.0, -0.00378, 0.03053, 0.0, -0.00365, 0.03877, 0.0, -0.01402, -0.0315, -0.0, -0.01335, -0.03267, -0.0, 0.04047, 0.07455, -0.0, 0.00257, 0.09648, -0.0, -0.00257, 0.09648, -0.0, -0.00257, -0.09648, 0.0, 0.00257, -0.09648, 0.0], [-0.02193, -0.03479, -0.0, -0.01807, -0.0262, 0.0, -0.0331, -0.03741, 0.0, -0.05863, -0.04376, 0.0, -0.03245, -0.01915, 0.0, 0.03245, -0.01915, 0.0, 0.05863, -0.04376, -0.0, 0.0331, -0.03741, -0.0, 0.01807, -0.0262, -0.0, 0.02193, -0.03479, 0.0, 0.02193, 0.03479, 0.0, 0.01807, 0.0262, -0.0, 0.0331, 0.03741, -0.0, 0.05863, 0.04376, -0.0, 0.03245, 0.01915, -0.0, -0.03245, 0.01915, -0.0, -0.05863, 0.04376, 0.0, -0.0331, 0.03741, 0.0, -0.01807, 0.0262, 0.0, -0.02193, 0.03479, -0.0, -0.05922, -0.02946, -0.0, 0.05922, -0.02946, 0.0, 0.05922, 0.02946, 0.0, -0.05922, 0.02946, -0.0]]), "B2g":array([[-0.0, 0.0, 0.09881, -0.0, 0.0, -0.01114, 0.0, -0.0, -0.09856, 0.0, -0.0, -0.01116, -0.0, 0.0, -0.02683, -0.0, -0.0, 0.02683, 0.0, 0.0, 0.01116, 0.0, 0.0, 0.09856, -0.0, -0.0, 0.01114, 0.0, 0.0, -0.09881, 0.0, 0.0, -0.09881, 0.0, -0.0, 0.01114, -0.0, 0.0, 0.09856, -0.0, 0.0, 0.01116, 0.0, -0.0, 0.02683, 0.0, 0.0, -0.02683, -0.0, -0.0, -0.01116, -0.0, -0.0, -0.09856, 0.0, 0.0, -0.01114, -0.0, -0.0, 0.09881, -0.0, -0.0, 0.10767, 0.0, 0.0, -0.10767, 0.0, 0.0, -0.10767, -0.0, -0.0, 0.10767], [0.0, -0.0, 0.03135, 0.0, 0.0, 0.08803, 0.0, 0.0, -0.02415, 0.0, 0.0, -0.09809, -0.0, -0.0, -0.08771, -0.0, 0.0, 0.08771, 0.0, -0.0, 0.09809, -0.0, -0.0, 0.02415, 0.0, -0.0, -0.08803, 0.0, 0.0, -0.03135, -0.0, 0.0, -0.03135, -0.0, -0.0, -0.08803, -0.0, -0.0, 0.02415, -0.0, -0.0, 0.09809, 0.0, 0.0, 0.08771, 0.0, -0.0, -0.08771, -0.0, 0.0, -0.09809, 0.0, 0.0, -0.02415, -0.0, 0.0, 0.08803, -0.0, -0.0, 0.03135, 0.0, -0.0, -0.0738, -0.0, 0.0, 0.0738, -0.0, 0.0, 0.0738, 0.0, -0.0, -0.0738]]), "B3g":array([[-0.0, 0.0, -0.01559, 0.0, -0.0, -0.11875, 0.0, 0.0, -0.08731, -0.0, 0.0, 0.02086, 0.0, -0.0, 0.08201, 0.0, 0.0, 0.08201, -0.0, -0.0, 0.02086, 0.0, -0.0, -0.08731, -0.0, 0.0, -0.11875, 0.0, 0.0, -0.01559, 0.0, 0.0, 0.01559, 0.0, 0.0, 0.11875, -0.0, -0.0, 0.08731, 0.0, -0.0, -0.02086, -0.0, 0.0, -0.08201, -0.0, -0.0, -0.08201, 0.0, 0.0, -0.02086, -0.0, 0.0, 0.08731, 0.0, -0.0, 0.11875, -0.0, -0.0, 0.01559, 0.0, -0.0, 0.05233, -0.0, 0.0, 0.05233, -0.0, 0.0, -0.05233, 0.0, -0.0, -0.05233], [-0.0, 0.0, 0.04435, -0.0, -0.0, -0.00577, -0.0, 0.0, 0.00551, 0.0, 0.0, 0.05509, -0.0, -0.0, -0.10679, -0.0, 0.0, -0.10679, 0.0, -0.0, 0.05509, -0.0, -0.0, 0.00551, 0.0, -0.0, -0.00577, -0.0, -0.0, 0.04435, 0.0, -0.0, -0.04435, -0.0, -0.0, 0.00577, 0.0, -0.0, -0.00551, -0.0, -0.0, -0.05509, 0.0, 0.0, 0.10679, 0.0, -0.0, 0.10679, -0.0, 0.0, -0.05509, 0.0, 0.0, -0.00551, -0.0, 0.0, 0.00577, 0.0, 0.0, -0.04435, -0.0, -0.0, 0.15472, -0.0, 0.0, 0.15472, 0.0, 0.0, -0.15472, 0.0, -0.0, -0.15472]]), "Au":array([[0.0, -0.0, 0.01856, -0.0, -0.0, 0.10929, -0.0, -0.0, 0.13317, 0.0, 0.0, 0.05604, -0.0, 0.0, 0.033, -0.0, 0.0, -0.033, -0.0, 0.0, -0.05604, -0.0, -0.0, -0.13317, -0.0, -0.0, -0.10929, -0.0, -0.0, -0.01856, -0.0, -0.0, 0.01856, -0.0, -0.0, 0.10929, -0.0, 0.0, 0.13317, -0.0, -0.0, 0.05604, -0.0, -0.0, 0.033, -0.0, -0.0, -0.033, -0.0, -0.0, -0.05604, -0.0, 0.0, -0.13317, -0.0, -0.0, -0.10929, -0.0, -0.0, -0.01856, -0.0, -0.0, 0.00495, -0.0, -0.0, -0.00495, -0.0, 0.0, 0.00495, -0.0, 0.0, -0.00495], [-0.0, -0.0, 0.03138, 0.0, -0.0, 0.06039, -0.0, -0.0, -0.00685, -0.0, -0.0, -0.06873, -0.0, -0.0, -0.05045, -0.0, -0.0, 0.05045, -0.0, -0.0, 0.06873, -0.0, -0.0, 0.00685, -0.0, -0.0, -0.06039, -0.0, -0.0, -0.03138, -0.0, -0.0, 0.03138, -0.0, -0.0, 0.06039, -0.0, 0.0, -0.00685, -0.0, 0.0, -0.06873, -0.0, 0.0, -0.05045, -0.0, 0.0, 0.05045, -0.0, 0.0, 0.06873, -0.0, 0.0, 0.00685, -0.0, -0.0, -0.06039, -0.0, -0.0, -0.03138, -0.0, 0.0, -0.15217, -0.0, 0.0, 0.15217, -0.0, -0.0, -0.15217, -0.0, -0.0, 0.15217]]), "B1u":array([[-0.0, 0.0, 0.05076, 0.0, 0.0, 0.10443, 0.0, 0.0, 0.0376, 0.0, 0.0, -0.04815, 0.0, 0.0, -0.11465, 0.0, 0.0, -0.11465, 0.0, 0.0, -0.04815, 0.0, 0.0, 0.0376, 0.0, 0.0, 0.10443, 0.0, 0.0, 0.05076, 0.0, 0.0, 0.05076, 0.0, 0.0, 0.10443, 0.0, 0.0, 0.0376, 0.0, 0.0, -0.04815, 0.0, 0.0, -0.11465, 0.0, 0.0, -0.11465, 0.0, 0.0, -0.04815, 0.0, 0.0, 0.0376, 0.0, 0.0, 0.10443, 0.0, 0.0, 0.05076, 0.0, 0.0, -0.02999, 0.0, 0.0, -0.02999, 0.0, 0.0, -0.02999, 0.0, 0.0, -0.02999], [0.0, -0.0, -0.07786, 0.0, 0.0, 0.0361, 0.0, 0.0, 0.09454, 0.0, 0.0, 0.00307, 0.0, 0.0, 0.05787, 0.0, 0.0, 0.05787, 0.0, 0.0, 0.00307, 0.0, 0.0, 0.09454, 0.0, 0.0, 0.0361, 0.0, 0.0, -0.07786, 0.0, 0.0, -0.07786, 0.0, 0.0, 0.0361, 0.0, 0.0, 0.09454, 0.0, 0.0, 0.00307, 0.0, 0.0, 0.05787, 0.0, 0.0, 0.05787, 0.0, 0.0, 0.00307, 0.0, 0.0, 0.09454, 0.0, 0.0, 0.0361, 0.0, 0.0, -0.07786, 0.0, 0.0, -0.11372, 0.0, 0.0, -0.11372, 0.0, 0.0, -0.11372, 0.0, 0.0, -0.11372]]), "B2u":array([[0.00676, 0.02346, -0.0, -0.04099, 0.08251, 0.0, -0.01487, 0.09615, 0.0, 0.02672, 0.01224, -0.0, 0.00821, -0.02988, -0.0, 0.00821, 0.02988, -0.0, 0.02672, -0.01224, -0.0, -0.01487, -0.09615, -0.0, -0.04099, -0.08251, -0.0, 0.00676, -0.02346, -0.0, 0.00676, 0.02346, 0.0, -0.04099, 0.08251, 0.0, -0.01487, 0.09615, 0.0, 0.02672, 0.01224, 0.0, 0.00821, -0.02988, 0.0, 0.00821, 0.02988, 0.0, 0.02672, -0.01224, 0.0, -0.01487, -0.09615, -0.0, -0.04099, -0.08251, -0.0, 0.00676, -0.02346, -0.0, 0.01416, -0.01522, -0.0, 0.01416, 0.01522, -0.0, 0.01416, -0.01522, 0.0, 0.01416, 0.01522, 0.0], [0.00346, -0.00713, -0.0, 0.0294, 0.01241, -0.0, 0.06515, 0.04764, -0.0, -0.02667, 0.03403, 0.0, -0.04371, 0.06713, 0.0, -0.04371, -0.06713, 0.0, -0.02667, -0.03403, 0.0, 0.06515, -0.04764, -0.0, 0.0294, -0.01241, -0.0, 0.00346, 0.00713, -0.0, 0.00346, -0.00713, 0.0, 0.0294, 0.01241, 0.0, 0.06515, 0.04764, 0.0, -0.02667, 0.03403, -0.0, -0.04371, 0.06713, -0.0, -0.04371, -0.06713, -0.0, -0.02667, -0.03403, -0.0, 0.06515, -0.04764, 0.0, 0.0294, -0.01241, 0.0, 0.00346, 0.00713, 0.0, -0.02764, 0.01543, 0.0, -0.02764, -0.01543, 0.0, -0.02764, 0.01543, -0.0, -0.02764, -0.01543, -0.0]]), "B3u":array([[0.01508, 0.05317, 0.0, 0.0692, 0.02308, 0.0, 0.03743, -0.03292, 0.0, -0.03941, -0.02595, -0.0, -0.01099, -0.04561, 0.0, 0.01099, -0.04561, -0.0, 0.03941, -0.02595, 0.0, -0.03743, -0.03292, -0.0, -0.0692, 0.02308, -0.0, -0.01508, 0.05317, 0.0, 0.01508, 0.05317, 0.0, 0.0692, 0.02308, 0.0, 0.03743, -0.03292, -0.0, -0.03941, -0.02595, -0.0, -0.01099, -0.04561, -0.0, 0.01099, -0.04561, 0.0, 0.03941, -0.02595, 0.0, -0.03743, -0.03292, 0.0, -0.0692, 0.02308, -0.0, -0.01508, 0.05317, -0.0, -0.01609, 0.02822, -0.0, 0.01609, 0.02822, 0.0, -0.01609, 0.02822, 0.0, 0.01609, 0.02822, -0.0], [0.01984, 0.02279, -0.0, -0.01063, -0.05056, -0.0, -0.00328, -0.0665, -0.0, 0.03322, 0.03808, -0.0, 0.02464, -0.0021, 0.0, -0.02464, -0.0021, -0.0, -0.03322, 0.03808, 0.0, 0.00328, -0.0665, 0.0, 0.01063, -0.05056, 0.0, -0.01984, 0.02279, 0.0, 0.01984, 0.02279, 0.0, -0.01063, -0.05056, 0.0, -0.00328, -0.0665, 0.0, 0.03322, 0.03808, 0.0, 0.02464, -0.0021, -0.0, -0.02464, -0.0021, 0.0, -0.03322, 0.03808, -0.0, 0.00328, -0.0665, -0.0, 0.01063, -0.05056, -0.0, -0.01984, 0.02279, -0.0, 0.05263, 0.05829, -0.0, -0.05263, 0.05829, 0.0, 0.05263, 0.05829, 0.0, -0.05263, 0.05829, -0.0]])}, database_path = 'porphycene_scsd_20220713.pkl', maxdist = 1.75)
model_objs_dict['corrphycene'] = scsd_model('corrphycene', array([['-3.4178', '0.0', '-0.5268', 'C'], ['-2.9469', '-0.0', '0.8031', 'C'], ['-3.8348', '-0.0', '1.9183', 'C'], ['-3.0997', '-0.0', '3.0368', 'C'], ['-1.7046', '-0.0', '2.6424', 'C'], ['-0.7034', '-0.0', '3.6312', 'C'], ['0.7034', '-0.0', '3.6312', 'C'], ['1.7046', '-0.0', '2.6424', 'C'], ['3.0997', '-0.0', '3.0368', 'C'], ['3.8348', '-0.0', '1.9183', 'C'], ['2.9469', '-0.0', '0.8031', 'C'], ['3.4178', '0.0', '-0.5268', 'C'], ['2.6016', '0.0', '-1.6529', 'C'], ['2.9365', '0.0', '-3.0533', 'C'], ['1.7466', '0.0', '-3.7517', 'C'], ['0.7108', '0.0', '-2.7825', 'C'], ['-0.7108', '0.0', '-2.7825', 'C'], ['-1.7466', '0.0', '-3.7517', 'C'], ['-2.9365', '0.0', '-3.0533', 'C'], ['-2.6016', '0.0', '-1.6529', 'C'], ['-1.6213', '-0.0', '1.2687', 'N'], ['1.6213', '-0.0', '1.2687', 'N'], ['1.2479', '0.0', '-1.5333', 'N'], ['-1.2479', '0.0', '-1.5333', 'N']]), 'C2v' , pca = {"A1":array([[0.06933, -0.0, 0.00856, 0.09734, -0.0, 0.00281, 0.05982, -0.0, 0.00648, 0.0278, -0.0, -0.00224, 0.05064, 0.0, -0.02249, 0.00134, 0.0, 0.00044, -0.00134, 0.0, 0.00044, -0.05064, 0.0, -0.02249, -0.0278, -0.0, -0.00224, -0.05982, -0.0, 0.00648, -0.09734, -0.0, 0.00281, -0.06933, 0.0, 0.00856, -0.04861, 0.0, 0.01731, 0.00091, 0.0, -0.00531, 0.03344, 0.0, -0.01687, -0.00759, 0.0, 0.01483, 0.00759, 0.0, 0.01483, -0.03344, 0.0, -0.01687, -0.00091, 0.0, -0.00531, 0.04861, 0.0, 0.01731, 0.10769, 0.0, -0.03599, -0.10769, 0.0, -0.03599, -0.05619, -0.0, 0.03247, 0.05619, -0.0, 0.03247], [0.01354, 0.0, -0.01497, 0.01324, 0.0, 0.01666, -0.04326, -0.0, 0.02608, -0.07223, -0.0, 0.05708, -0.02886, 0.0, 0.03939, -0.02182, -0.0, 0.05911, 0.02182, -0.0, 0.05911, 0.02886, 0.0, 0.03939, 0.07223, -0.0, 0.05708, 0.04326, -0.0, 0.02608, -0.01324, -0.0, 0.01666, -0.01354, 0.0, -0.01497, -0.00157, 0.0, -0.03725, 0.03532, 0.0, -0.06416, 0.04097, 0.0, -0.0702, 0.00576, 0.0, -0.01543, -0.00576, 0.0, -0.01543, -0.04097, 0.0, -0.0702, -0.03532, 0.0, -0.06416, 0.00157, 0.0, -0.03725, 0.02836, -0.0, 0.01839, -0.02836, -0.0, 0.01839, -0.01348, -0.0, -0.01472, 0.01348, -0.0, -0.01472]]), "A2":array([[-0.0, 0.04267, -0.0, -0.0, -0.00867, 0.0, -0.0, -0.0973, -0.0, -0.0, -0.13279, 0.0, -0.0, -0.05401, -0.0, 0.0, -0.02874, -0.0, -0.0, 0.02874, -0.0, -0.0, 0.05401, -0.0, -0.0, 0.13279, -0.0, -0.0, 0.0973, -0.0, -0.0, 0.00867, -0.0, -0.0, -0.04267, -0.0, -0.0, -0.06029, -0.0, -0.0, -0.1341, -0.0, -0.0, -0.11096, -0.0, -0.0, -0.02215, -0.0, -0.0, 0.02215, -0.0, -0.0, 0.11096, -0.0, -0.0, 0.1341, 0.0, -0.0, 0.06029, 0.0, -0.0, 0.0135, -0.0, -0.0, -0.0135, -0.0, -0.0, 0.00426, -0.0, -0.0, -0.00426, 0.0], [0.0, 0.08576, 0.0, -0.0, 0.03702, -0.0, -0.0, 0.03547, -0.0, -0.0, -0.06958, 0.0, -0.0, -0.097, -0.0, 0.0, -0.0686, -0.0, -0.0, 0.0686, -0.0, -0.0, 0.097, -0.0, -0.0, 0.06958, -0.0, -0.0, -0.03547, -0.0, -0.0, -0.03702, 0.0, -0.0, -0.08576, -0.0, -0.0, -0.07025, -0.0, -0.0, 0.05979, -0.0, -0.0, 0.10526, -0.0, -0.0, -0.01638, -0.0, -0.0, 0.01638, -0.0, -0.0, -0.10526, -0.0, -0.0, -0.05979, 0.0, -0.0, 0.07025, 0.0, -0.0, -0.05494, -0.0, -0.0, 0.05494, -0.0, -0.0, -0.11633, -0.0, -0.0, 0.11633, 0.0]]), "B1":array([[0.00542, 0.0, -0.04744, 0.01003, -0.0, -0.04072, 0.00266, 0.0, -0.02296, 0.01042, 0.0, -1e-05, 0.035, -0.0, -0.0274, 0.08831, 0.0, -0.02204, 0.08831, -0.0, 0.02204, 0.035, 0.0, 0.0274, 0.01042, -0.0, 1e-05, 0.00266, -0.0, 0.02296, 0.01003, 0.0, 0.04072, 0.00542, 0.0, 0.04744, -0.03359, -0.0, 0.04841, -0.04471, -0.0, 0.07009, -0.03331, 0.0, 0.05124, -0.01176, -0.0, 0.03139, -0.01176, 0.0, -0.03139, -0.03331, -0.0, -0.05124, -0.04471, 0.0, -0.07009, -0.03359, 0.0, -0.04841, 0.03157, -0.0, -0.01746, 0.03157, 0.0, 0.01746, -0.06004, 0.0, 0.02615, -0.06004, -0.0, -0.02615], [0.0611, 0.0, 0.05366, 0.02992, -0.0, 0.03617, -0.02485, -0.0, 0.02678, -0.09456, 0.0, -0.04696, -0.03936, 0.0, -0.06461, 0.04115, -0.0, -0.05306, 0.04115, -0.0, 0.05306, -0.03936, 0.0, 0.06461, -0.09456, -0.0, 0.04696, -0.02485, 0.0, -0.02678, 0.02992, 0.0, -0.03617, 0.0611, 0.0, -0.05366, 0.02044, 0.0, -0.02372, 0.0111, -0.0, -0.00351, -0.02215, -0.0, -0.00441, -0.0053, 0.0, 0.00509, -0.0053, 0.0, -0.00509, -0.02215, 0.0, 0.00441, 0.0111, -0.0, 0.00351, 0.02044, 0.0, 0.02372, 0.03774, -0.0, 0.00114, 0.03774, 0.0, -0.00114, -0.01528, 0.0, -0.01485, -0.01528, -0.0, 0.01485]]), "B2":array([[0.0, -0.10234, 0.0, -0.0, -0.08761, 0.0, -0.0, -0.1101, -0.0, 0.0, -0.01857, 0.0, -0.0, 0.04595, 0.0, -0.0, 0.15104, 0.0, 0.0, 0.15104, 0.0, 0.0, 0.04595, 0.0, 0.0, -0.01857, 0.0, 0.0, -0.1101, -0.0, 0.0, -0.08761, -0.0, 0.0, -0.10234, 0.0, 0.0, -0.02698, -0.0, 0.0, -0.01761, -0.0, -0.0, 0.0523, 0.0, 0.0, 0.07818, 0.0, 0.0, 0.07818, 0.0, 0.0, 0.0523, 0.0, 0.0, -0.01761, -0.0, 0.0, -0.02698, -0.0, 0.0, -0.00558, 0.0, 0.0, -0.00558, -0.0, 0.0, 0.04133, -0.0, 0.0, 0.04133, -0.0], [-0.0, 0.04187, 0.0, 0.0, 0.02818, -0.0, 0.0, -0.05667, 0.0, 0.0, -0.07724, 0.0, -0.0, -0.00168, 0.0, 0.0, -0.01802, 0.0, -0.0, -0.01802, 0.0, 0.0, -0.00168, 0.0, 0.0, -0.07724, 0.0, 0.0, -0.05667, 0.0, 0.0, 0.02818, -0.0, 0.0, 0.04187, 0.0, 0.0, 0.05977, -0.0, 0.0, -0.09502, -0.0, 0.0, -0.12231, 0.0, 0.0, 0.02902, 0.0, 0.0, 0.02902, 0.0, -0.0, -0.12231, 0.0, 0.0, -0.09502, 0.0, 0.0, 0.05977, 0.0, 0.0, 0.06008, -0.0, 0.0, 0.06008, -0.0, 0.0, 0.15202, -0.0, 0.0, 0.15202, -0.0]])}, database_path = 'corrphycene_scsd_20220713.pkl', maxdist = 1.75)
model_objs_dict['corrole'] = scsd_model('corrole', array([['-2.7327', '-0.0', '0.9917', 'C'], ['-3.4068', '-0.0', '2.2424', 'C'], ['-2.4831', '-0.0', '3.2197', 'C'], ['-1.2063', '-0.0', '2.625', 'C'], ['-0.0', '-0.0', '3.2894', 'C'], ['1.2063', '-0.0', '2.625', 'C'], ['2.4831', '-0.0', '3.2197', 'C'], ['3.4068', '-0.0', '2.2424', 'C'], ['2.7327', '-0.0', '0.9917', 'C'], ['3.3421', '0.0', '-0.2598', 'C'], ['2.5645', '0.0', '-1.4148', 'C'], ['2.9128', '0.0', '-2.7915', 'C'], ['1.7576', '0.0', '-3.5284', 'C'], ['0.6952', '0.0', '-2.627', 'C'], ['-0.6952', '0.0', '-2.627', 'C'], ['-1.7576', '0.0', '-3.5284', 'C'], ['-2.9128', '0.0', '-2.7915', 'C'], ['-2.5645', '0.0', '-1.4148', 'C'], ['-3.3421', '0.0', '-0.2598', 'C'], ['-1.3604', '-0.0', '1.2442', 'N'], ['1.3604', '-0.0', '1.2442', 'N'], ['1.2074', '0.0', '-1.3462', 'N'], ['-1.2074', '0.0', '-1.3462', 'N']]), 'C2v' , pca = {"A1":array([[-0.04692, 0.0, -0.01278, -0.00288, 0.0, -0.00905, 0.01931, 0.0, -0.03078, -0.00493, 0.0, -0.04405, 0.0, 0.0, -0.05467, 0.00493, 0.0, -0.04405, -0.01931, 0.0, -0.03078, 0.00288, 0.0, -0.00905, 0.04692, 0.0, -0.01278, 0.04953, 0.0, -0.01009, 0.02185, -0.0, 0.00623, -0.03112, -0.0, 0.00914, -0.04657, -0.0, 0.04505, -0.00377, -0.0, 0.06579, 0.00377, -0.0, 0.06579, 0.04657, -0.0, 0.04505, 0.03112, -0.0, 0.00914, -0.02185, -0.0, 0.00623, -0.04953, 0.0, -0.01009, -0.04544, 0.0, -0.03424, 0.04544, 0.0, -0.03424, 0.03467, -0.0, 0.04208, -0.03467, -0.0, 0.04208], [0.05726, -0.0, -0.0163, 0.04336, 0.0, -0.04115, 0.02989, 0.0, -0.03586, 0.02285, 0.0, -0.01918, 0.0, 0.0, -0.00086, -0.02285, 0.0, -0.01918, -0.02989, 0.0, -0.03586, -0.04336, 0.0, -0.04115, -0.05726, 0.0, -0.0163, -0.03705, -0.0, 0.00177, -0.03458, -0.0, 0.02148, -0.02014, -0.0, 0.0341, -0.01269, -0.0, 0.02911, -0.0059, -0.0, 0.01696, 0.0059, -0.0, 0.01696, 0.01269, -0.0, 0.02911, 0.02014, -0.0, 0.0341, 0.03458, -0.0, 0.02148, 0.03705, -0.0, 0.00177, 0.06288, 0.0, -0.02742, -0.06288, 0.0, -0.02742, -0.03993, -0.0, 0.03785, 0.03993, -0.0, 0.03785]]), "A2":array([[0.0, 0.00777, 0.0, 0.0, 0.08565, 0.0, 0.0, 0.09702, -0.0, 0.0, 0.02281, 0.0, 0.0, -0.0, 0.0, 0.0, -0.02281, 0.0, 0.0, -0.09702, 0.0, 0.0, -0.08565, -0.0, 0.0, -0.00777, -0.0, 0.0, 0.02328, -0.0, 0.0, 0.0344, -0.0, 0.0, 0.09266, 0.0, -0.0, 0.07764, 0.0, 0.0, 0.0125, 0.0, 0.0, -0.0125, 0.0, 0.0, -0.07764, -0.0, 0.0, -0.09266, -0.0, 0.0, -0.0344, 0.0, 0.0, -0.02328, 0.0, 0.0, -0.02907, -0.0, 0.0, 0.02907, 0.0, 0.0, -0.01719, -0.0, 0.0, 0.01719, 0.0], [-0.0, -0.03569, -0.0, -0.0, 0.03164, -0.0, -0.0, 0.05834, -0.0, 0.0, -0.00493, 0.0, -0.0, -0.0, 0.0, 0.0, 0.00493, 0.0, 0.0, -0.05834, 0.0, 0.0, -0.03164, 0.0, 0.0, 0.03569, -0.0, 0.0, 0.04444, -0.0, 0.0, 0.04154, -0.0, 0.0, -0.04977, 0.0, -0.0, -0.07959, 0.0, 0.0, 0.00847, 0.0, 0.0, -0.00847, 0.0, 0.0, 0.07959, -0.0, 0.0, 0.04977, -0.0, 0.0, -0.04154, 0.0, 0.0, -0.04444, 0.0, 0.0, -0.0593, -0.0, 0.0, 0.0593, 0.0, 0.0, 0.08629, -0.0, 0.0, -0.08629, 0.0]]), "B1":array([[-0.01576, -0.0, -0.0342, -0.03198, 0.0, -0.0427, -0.04795, 0.0, -0.03242, -0.04017, 0.0, -0.01655, -0.05072, 0.0, -0.0, -0.04017, -0.0, 0.01655, -0.04795, -0.0, 0.03242, -0.03198, -0.0, 0.0427, -0.01576, 0.0, 0.0342, 0.00393, -0.0, 0.04043, 0.0225, -0.0, 0.03148, 0.04511, 0.0, 0.03217, 0.05614, 0.0, 0.01987, 0.03904, -0.0, 0.00751, 0.03904, 0.0, -0.00751, 0.05614, 0.0, -0.01987, 0.04511, 0.0, -0.03217, 0.0225, -0.0, -0.03148, 0.00393, 0.0, -0.04043, -0.02092, 0.0, -0.01852, -0.02092, -0.0, 0.01852, 0.01937, -0.0, 0.01496, 0.01937, -0.0, -0.01496], [1e-05, -0.0, 0.01917, -0.05052, -0.0, 0.05116, -0.05947, -0.0, 0.0527, -0.01962, 0.0, 0.00831, -0.01353, 0.0, -0.0, -0.01962, 0.0, -0.00831, -0.05947, 0.0, -0.0527, -0.05052, 0.0, -0.05116, 1e-05, 0.0, -0.01917, 0.03219, 0.0, -0.00381, 0.05525, 0.0, -0.02254, -0.01701, 0.0, 0.00482, -0.02739, -0.0, -0.0001, 0.00893, 0.0, -0.03614, 0.00893, -0.0, 0.03614, -0.02739, -0.0, 0.0001, -0.01701, 0.0, -0.00482, 0.05525, -0.0, 0.02254, 0.03219, -0.0, 0.00381, -0.00222, 0.0, 0.00893, -0.00222, -0.0, -0.00893, 0.08649, 0.0, -0.06009, 0.08649, -0.0, 0.06009]]), "B2":array([[-0.0, 0.00248, -0.0, -0.0, -0.02842, 0.0, -0.0, -0.02767, -0.0, -0.0, 0.00396, -0.0, 0.0, -0.00028, -0.0, 0.0, 0.00396, -0.0, -0.0, -0.02767, 0.0, -0.0, -0.02842, 0.0, -0.0, 0.00248, 0.0, -0.0, -0.00455, 0.0, -0.0, 0.02797, 0.0, -0.0, -0.10198, 0.0, 0.0, -0.08719, -0.0, 0.0, 0.05464, -0.0, -0.0, 0.05464, -0.0, -0.0, -0.08719, -0.0, -0.0, -0.10198, 0.0, -0.0, 0.02797, 0.0, -0.0, -0.00455, 0.0, -0.0, 0.02646, 0.0, -0.0, 0.02646, 0.0, -0.0, 0.13455, 0.0, -0.0, 0.13455, 0.0], [-0.0, -0.01526, -0.0, -0.0, 0.0085, -0.0, 0.0, 0.0721, 0.0, -0.0, 0.08214, -0.0, 0.0, 0.12174, -0.0, 0.0, 0.08214, -0.0, -0.0, 0.0721, 0.0, -0.0, 0.0085, 0.0, -0.0, -0.01526, 0.0, -0.0, -0.05277, -0.0, -0.0, -0.047, -0.0, -0.0, -0.04724, 0.0, 0.0, -0.03241, 0.0, 0.0, -0.02454, -0.0, -0.0, -0.02454, -0.0, -0.0, -0.03241, -0.0, -0.0, -0.04724, -0.0, -0.0, -0.047, -0.0, -0.0, -0.05277, 0.0, -0.0, 0.02753, -0.0, -0.0, 0.02753, 0.0, -0.0, -0.02965, -0.0, -0.0, -0.02965, -0.0]])}, database_path = 'corrole_scsd_20220713.pkl', maxdist = 1.75)
model_objs_dict['norcorrole'] = scsd_model('norcorrole', array([['2.5656', '-0.7374', '-0.0', 'C'], ['-1.2572', '-2.5724', '0.0', 'C'], ['-0.0', '-3.2567', '0.0', 'C'], ['1.2572', '-2.5724', '-0.0', 'C'], ['3.4342', '-1.8641', '-0.0', 'C'], ['2.6364', '-2.9932', '-0.0', 'C'], ['-2.6364', '-2.9932', '0.0', 'C'], ['-2.5656', '-0.7374', '0.0', 'C'], ['-3.4342', '-1.8641', '0.0', 'C'], ['-2.5656', '0.7374', '0.0', 'C'], ['1.2572', '2.5724', '-0.0', 'C'], ['0.0', '3.2567', '0.0', 'C'], ['-1.2572', '2.5724', '0.0', 'C'], ['-3.4342', '1.8641', '0.0', 'C'], ['-2.6364', '2.9932', '0.0', 'C'], ['2.6364', '2.9932', '-0.0', 'C'], ['2.5656', '0.7374', '-0.0', 'C'], ['3.4342', '1.8641', '-0.0', 'C'], ['1.3016', '-1.2098', '-0.0', 'N'], ['-1.3016', '-1.2098', '0.0', 'N'], ['-1.3016', '1.2098', '0.0', 'N'], ['1.3016', '1.2098', '-0.0', 'N']]), 'D2h' , pca = {"Ag":array([[0.05867, 0.01725, 0.0, 0.00388, -0.0654, -0.0, 0.0, 0.0714, 0.0, -0.00388, -0.0654, -0.0, -0.07514, 0.00166, -0.0, -0.02577, 0.00968, -0.0, 0.02577, 0.00968, 0.0, -0.05867, 0.01725, -0.0, 0.07514, 0.00166, 0.0, -0.05867, -0.01725, -0.0, -0.00388, 0.0654, 0.0, -0.0, -0.0714, -0.0, 0.00388, 0.0654, 0.0, 0.07514, -0.00166, 0.0, 0.02577, -0.00968, 0.0, -0.02577, -0.00968, -0.0, 0.05867, -0.01725, 0.0, -0.07514, -0.00166, -0.0, -0.0118, -0.00827, -0.0, 0.0118, -0.00827, 0.0, 0.0118, 0.00827, 0.0, -0.0118, 0.00827, -0.0], [0.02377, 0.03605, 0.0, -0.01504, -0.02035, -0.0, -0.0, -0.05585, -0.0, 0.01504, -0.02035, -0.0, 0.02702, 0.01845, 0.0, -0.00717, 0.09463, 0.0, 0.00717, 0.09463, 0.0, -0.02377, 0.03605, 0.0, -0.02702, 0.01845, -0.0, -0.02377, -0.03605, -0.0, 0.01504, 0.02035, 0.0, 0.0, 0.05585, 0.0, -0.01504, 0.02035, 0.0, -0.02702, -0.01845, -0.0, 0.00717, -0.09463, -0.0, -0.00717, -0.09463, -0.0, 0.02377, -0.03605, -0.0, 0.02702, -0.01845, 0.0, 0.0118, 0.03117, 0.0, -0.0118, 0.03117, 0.0, -0.0118, -0.03117, -0.0, 0.0118, -0.03117, -0.0]]), "B1g":array([[-0.01259, -0.02249, 0.0, -0.01151, -0.07015, -0.0, -0.04105, -0.0, 0.0, -0.01151, 0.07015, -0.0, 0.06541, -0.02723, -0.0, -0.01447, 0.00667, 0.0, -0.01447, -0.00667, 0.0, -0.01259, 0.02249, -0.0, 0.06541, 0.02723, 0.0, 0.01259, 0.02249, -0.0, 0.01151, 0.07015, 0.0, 0.04105, -0.0, -0.0, 0.01151, -0.07015, 0.0, -0.06541, 0.02723, 0.0, 0.01447, -0.00667, -0.0, 0.01447, 0.00667, -0.0, 0.01259, -0.02249, 0.0, -0.06541, -0.02723, -0.0, 0.0292, 0.02862, -0.0, 0.0292, -0.02862, -0.0, -0.0292, -0.02862, 0.0, -0.0292, 0.02862, 0.0], [-0.09285, 0.01773, -0.0, 0.02288, -0.03948, 0.0, 0.00976, -0.0, -0.0, 0.02288, 0.03948, 0.0, -0.01331, 0.00333, 0.0, 0.03153, -0.0469, -0.0, 0.03153, 0.0469, -0.0, -0.09285, -0.01773, 0.0, -0.01331, -0.00333, -0.0, 0.09285, -0.01773, 0.0, -0.02288, 0.03948, -0.0, -0.00976, -0.0, 0.0, -0.02288, -0.03948, -0.0, 0.01331, -0.00333, -0.0, -0.03153, 0.0469, 0.0, -0.03153, -0.0469, 0.0, 0.09285, 0.01773, -0.0, 0.01331, 0.00333, 0.0, -0.02078, -0.02581, -0.0, -0.02078, 0.02581, -0.0, 0.02078, 0.02581, 0.0, 0.02078, -0.02581, 0.0]]), "B2g":array([[-0.0, 0.0, -0.01201, 0.0, -0.0, -0.03806, 0.0, -0.0, -0.02511, -0.0, -0.0, -0.03806, -0.0, 0.0, 0.06229, 0.0, -0.0, 0.0435, -0.0, -0.0, 0.0435, 0.0, 0.0, -0.01201, -0.0, 0.0, 0.06229, 0.0, -0.0, 0.01201, -0.0, -0.0, 0.03806, -0.0, -0.0, 0.02511, 0.0, -0.0, 0.03806, 0.0, -0.0, -0.06229, -0.0, 0.0, -0.0435, 0.0, 0.0, -0.0435, -0.0, -0.0, 0.01201, 0.0, -0.0, -0.06229, -0.0, 0.0, -0.08159, 0.0, 0.0, -0.08159, 0.0, -0.0, 0.08159, -0.0, -0.0, 0.08159], [-0.0, 0.0, 0.06429, -0.0, -0.0, -0.02649, 0.0, -0.0, -0.06637, -0.0, -0.0, -0.02649, -0.0, -0.0, 0.06007, 0.0, 0.0, -0.01498, 0.0, 0.0, -0.01498, 0.0, -0.0, 0.06429, 0.0, -0.0, 0.06007, 0.0, -0.0, -0.06429, 0.0, -0.0, 0.02649, -0.0, -0.0, 0.06637, 0.0, -0.0, 0.02649, 0.0, 0.0, -0.06007, -0.0, -0.0, 0.01498, -0.0, -0.0, 0.01498, -0.0, 0.0, -0.06429, -0.0, 0.0, -0.06007, 0.0, -0.0, 0.05098, 0.0, -0.0, 0.05098, -0.0, 0.0, -0.05098, -0.0, 0.0, -0.05098]]), "B3g":array([[-0.0, -0.0, -0.05836, -0.0, 0.0, 0.00858, 0.0, 0.0, 0.0, -0.0, 0.0, -0.00858, -0.0, -0.0, 0.03011, 0.0, -0.0, 0.06506, 0.0, -0.0, -0.06506, 0.0, -0.0, 0.05836, 0.0, -0.0, -0.03011, 0.0, 0.0, 0.05836, 0.0, 0.0, -0.00858, 0.0, 0.0, 0.0, 0.0, 0.0, 0.00858, 0.0, 0.0, -0.03011, -0.0, 0.0, -0.06506, -0.0, 0.0, 0.06506, -0.0, 0.0, -0.05836, -0.0, 0.0, 0.03011, 0.0, -0.0, -0.08789, -0.0, -0.0, 0.08789, -0.0, 0.0, 0.08789, 0.0, 0.0, -0.08789], [-0.0, 0.0, 0.04532, -0.0, -0.0, 0.07766, -0.0, -0.0, -0.0, -0.0, -0.0, -0.07766, 0.0, -0.0, 0.04329, 0.0, 0.0, -0.0437, -0.0, 0.0, 0.0437, 0.0, 0.0, -0.04532, 0.0, -0.0, -0.04329, 0.0, -0.0, -0.04532, 0.0, -0.0, -0.07766, -0.0, -0.0, -0.0, 0.0, -0.0, 0.07766, -0.0, 0.0, -0.04329, -0.0, -0.0, 0.0437, 0.0, -0.0, -0.0437, -0.0, -0.0, 0.04532, -0.0, 0.0, 0.04329, 0.0, -0.0, -0.04003, -0.0, -0.0, 0.04003, -0.0, 0.0, 0.04003, 0.0, 0.0, -0.04003]]), "Au":array([[-0.0, 0.0, 0.02711, 0.0, 0.0, -0.10759, -0.0, 0.0, -0.0, 0.0, 0.0, 0.10759, 0.0, 0.0, -0.1082, 0.0, 0.0, -0.00319, 0.0, -0.0, 0.00319, 0.0, -0.0, -0.02711, 0.0, 0.0, 0.1082, 0.0, -0.0, 0.02711, 0.0, 0.0, -0.10759, 0.0, 0.0, 0.0, 0.0, 0.0, 0.10759, 0.0, 0.0, -0.1082, 0.0, -0.0, -0.00319, 0.0, 0.0, 0.00319, 0.0, 0.0, -0.02711, 0.0, 0.0, 0.1082, 0.0, 0.0, 0.00392, 0.0, -0.0, -0.00392, 0.0, -0.0, 0.00392, 0.0, 0.0, -0.00392], [0.0, 0.0, -0.0231, 0.0, 0.0, 0.0587, 0.0, 0.0, 0.0, 0.0, 0.0, -0.0587, 0.0, 0.0, -0.06096, 0.0, -0.0, -0.10207, 0.0, 0.0, 0.10207, 0.0, -0.0, 0.0231, 0.0, 0.0, 0.06096, 0.0, -0.0, -0.0231, 0.0, 0.0, 0.0587, -0.0, 0.0, -0.0, 0.0, 0.0, -0.0587, 0.0, 0.0, -0.06096, 0.0, 0.0, -0.10207, 0.0, -0.0, 0.10207, 0.0, 0.0, 0.0231, 0.0, 0.0, 0.06096, 0.0, 0.0, 0.00516, 0.0, -0.0, -0.00516, 0.0, -0.0, 0.00516, 0.0, 0.0, -0.00516]]), "B1u":array([[-0.0, 0.0, -0.01921, 0.0, 0.0, 0.00849, -0.0, 0.0, -0.01314, 0.0, 0.0, 0.00849, 0.0, 0.0, 0.04069, 0.0, 0.0, 0.07582, 0.0, 0.0, 0.07582, 0.0, 0.0, -0.01921, 0.0, 0.0, 0.04069, 0.0, 0.0, -0.01921, 0.0, 0.0, 0.00849, -0.0, 0.0, -0.01314, 0.0, 0.0, 0.00849, 0.0, 0.0, 0.04069, 0.0, 0.0, 0.07582, 0.0, 0.0, 0.07582, 0.0, 0.0, -0.01921, 0.0, 0.0, 0.04069, 0.0, 0.0, -0.09922, 0.0, 0.0, -0.09922, 0.0, 0.0, -0.09922, 0.0, 0.0, -0.09922], [0.0, -0.0, 0.03461, 0.0, 0.0, 0.0782, 0.0, 0.0, -0.08371, 0.0, 0.0, 0.0782, 0.0, 0.0, -0.0695, 0.0, 0.0, 0.01219, 0.0, 0.0, 0.01219, 0.0, 0.0, 0.03461, 0.0, 0.0, -0.0695, 0.0, 0.0, 0.03461, 0.0, 0.0, 0.0782, 0.0, 0.0, -0.08371, 0.0, 0.0, 0.0782, 0.0, 0.0, -0.0695, 0.0, 0.0, 0.01219, 0.0, 0.0, 0.01219, 0.0, 0.0, 0.03461, 0.0, 0.0, -0.0695, 0.0, 0.0, -0.01365, 0.0, 0.0, -0.01365, 0.0, 0.0, -0.01365, 0.0, 0.0, -0.01365]]), "B2u":array([[-0.04131, 0.01274, 0.0, 0.03001, 0.01404, 0.0, 0.02538, -0.0, -0.0, 0.03001, -0.01404, -0.0, -0.021, 0.05764, -0.0, 0.03811, 0.04816, -0.0, 0.03811, -0.04816, 0.0, -0.04131, -0.01274, -0.0, -0.021, -0.05764, 0.0, -0.04131, 0.01274, -0.0, 0.03001, 0.01404, -0.0, 0.02538, -0.0, -0.0, 0.03001, -0.01404, 0.0, -0.021, 0.05764, 0.0, 0.03811, 0.04816, 0.0, 0.03811, -0.04816, -0.0, -0.04131, -0.01274, 0.0, -0.021, -0.05764, -0.0, -0.01848, -0.03556, -0.0, -0.01848, 0.03556, 0.0, -0.01848, -0.03556, 0.0, -0.01848, 0.03556, -0.0], [-0.01269, -0.01805, -0.0, -0.03285, 0.0696, 0.0, -0.04378, 0.0, 0.0, -0.03285, -0.0696, -0.0, 0.03862, 0.01418, 0.0, 0.03734, -0.04299, 0.0, 0.03734, 0.04299, -0.0, -0.01269, 0.01805, 0.0, 0.03862, -0.01418, -0.0, -0.01269, -0.01805, 0.0, -0.03285, 0.0696, 0.0, -0.04378, 0.0, 0.0, -0.03285, -0.0696, -0.0, 0.03862, 0.01418, -0.0, 0.03734, -0.04299, -0.0, 0.03734, 0.04299, 0.0, -0.01269, 0.01805, -0.0, 0.03862, -0.01418, 0.0, -0.00852, -0.02119, 0.0, -0.00852, 0.02119, -0.0, -0.00852, -0.02119, -0.0, -0.00852, 0.02119, 0.0]]), "B3u":array([[-0.00176, -0.01164, -0.0, 0.02841, 0.03133, -0.0, 0.0, 0.08559, 0.0, -0.02841, 0.03133, -0.0, -0.01309, -0.05256, -0.0, -0.05246, -0.03959, -0.0, 0.05246, -0.03959, 0.0, 0.00176, -0.01164, 0.0, 0.01309, -0.05256, 0.0, -0.00176, -0.01164, -0.0, 0.02841, 0.03133, 0.0, -0.0, 0.08559, -0.0, -0.02841, 0.03133, 0.0, -0.01309, -0.05256, -0.0, -0.05246, -0.03959, -0.0, 0.05246, -0.03959, 0.0, 0.00176, -0.01164, -0.0, 0.01309, -0.05256, 0.0, 0.00549, 0.02967, -0.0, -0.00549, 0.02967, -0.0, 0.00549, 0.02967, 0.0, -0.00549, 0.02967, 0.0], [0.02449, -0.01192, 0.0, 0.00268, 0.01757, -0.0, 0.0, 0.09458, 0.0, -0.00268, 0.01757, -0.0, 0.04238, -0.01415, 0.0, 0.10112, -0.02214, -0.0, -0.10112, -0.02214, -0.0, -0.02449, -0.01192, 0.0, -0.04238, -0.01415, -0.0, 0.02449, -0.01192, -0.0, 0.00268, 0.01757, 0.0, -0.0, 0.09458, -0.0, -0.00268, 0.01757, 0.0, 0.04238, -0.01415, 0.0, 0.10112, -0.02214, 0.0, -0.10112, -0.02214, 0.0, -0.02449, -0.01192, -0.0, -0.04238, -0.01415, -0.0, -0.00665, -0.01667, -0.0, 0.00665, -0.01667, -0.0, -0.00665, -0.01667, 0.0, 0.00665, -0.01667, 0.0]])}, database_path = 'norcorrole_scsd_20220713.pkl', maxdist = 1.75)
model_objs_dict['phthalocyanine'] = scsd_model('phthalocyanine', array([['-2.7567', '-1.1138', '-0.0', 'C'], ['-4.1495', '-0.6983', '-0.0', 'C'], ['-5.3338', '-1.4185', '-0.0', 'C'], ['-6.5161', '-0.6981', '-0.0', 'C'], ['-6.5161', '0.6981', '0.0', 'C'], ['-5.3338', '1.4185', '0.0', 'C'], ['-4.1495', '0.6983', '0.0', 'C'], ['-2.7567', '1.1138', '0.0', 'C'], ['-1.1138', '2.7567', '0.0', 'C'], ['-0.6983', '4.1495', '0.0', 'C'], ['-1.4185', '5.3338', '0.0', 'C'], ['-0.6981', '6.5161', '0.0', 'C'], ['0.6981', '6.5161', '0.0', 'C'], ['1.4185', '5.3338', '0.0', 'C'], ['0.6983', '4.1495', '0.0', 'C'], ['1.1138', '2.7567', '0.0', 'C'], ['2.7567', '1.1138', '0.0', 'C'], ['4.1495', '0.6983', '0.0', 'C'], ['5.3338', '1.4185', '0.0', 'C'], ['6.5161', '0.6981', '0.0', 'C'], ['6.5161', '-0.6981', '-0.0', 'C'], ['5.3338', '-1.4185', '-0.0', 'C'], ['4.1495', '-0.6983', '-0.0', 'C'], ['2.7567', '-1.1138', '-0.0', 'C'], ['1.1138', '-2.7567', '-0.0', 'C'], ['0.6983', '-4.1495', '-0.0', 'C'], ['1.4185', '-5.3338', '-0.0', 'C'], ['0.6981', '-6.5161', '-0.0', 'C'], ['-0.6981', '-6.5161', '-0.0', 'C'], ['-1.4185', '-5.3338', '-0.0', 'C'], ['-0.6983', '-4.1495', '-0.0', 'C'], ['-1.1138', '-2.7567', '-0.0', 'C'], ['-2.3833', '-2.3833', '-0.0', 'N'], ['-1.9637', '0.0', '-0.0', 'N'], ['-2.3833', '2.3833', '0.0', 'N'], ['0.0', '1.9637', '0.0', 'N'], ['2.3833', '2.3833', '0.0', 'N'], ['1.9637', '-0.0', '0.0', 'N'], ['2.3833', '-2.3833', '-0.0', 'N'], ['-0.0', '-1.9637', '-0.0', 'N']]), 'D4h' , pca = {"A1g":array([[0.01252, 0.00197, 0.0, 0.02482, 0.00029, 0.0, 0.0384, -0.00221, -0.0, 0.05071, 0.00078, 0.0, 0.05071, -0.00078, -0.0, 0.0384, 0.00221, 0.0, 0.02482, -0.00029, -0.0, 0.01252, -0.00197, -0.0, 0.00197, -0.01252, -0.0, 0.00029, -0.02482, -0.0, -0.00221, -0.0384, -0.0, 0.00078, -0.05071, -0.0, -0.00078, -0.05071, -0.0, 0.00221, -0.0384, -0.0, -0.00029, -0.02482, -0.0, -0.00197, -0.01252, -0.0, -0.01252, -0.00197, -0.0, -0.02482, -0.00029, -0.0, -0.0384, 0.00221, 0.0, -0.05071, -0.00078, -0.0, -0.05071, 0.00078, 0.0, -0.0384, -0.00221, -0.0, -0.02482, 0.00029, 0.0, -0.01252, 0.00197, 0.0, -0.00197, 0.01252, 0.0, -0.00029, 0.02482, 0.0, 0.00221, 0.0384, 0.0, -0.00078, 0.05071, 0.0, 0.00078, 0.05071, 0.0, -0.00221, 0.0384, 0.0, 0.00029, 0.02482, 0.0, 0.00197, 0.01252, 0.0, 0.00524, 0.00524, 0.0, 0.00979, -0.0, -0.0, 0.00524, -0.00524, -0.0, -0.0, -0.00979, -0.0, -0.00524, -0.00524, -0.0, -0.00979, 0.0, 0.0, -0.00524, 0.00524, 0.0, 0.0, 0.00979, 0.0], [-0.03463, -0.00975, -0.0, -0.01774, -0.0053, -0.0, 0.00403, -0.00615, -0.0, 0.02218, 0.00028, 0.0, 0.02218, -0.00028, -0.0, 0.00403, 0.00615, 0.0, -0.01774, 0.0053, 0.0, -0.03463, 0.00975, 0.0, -0.00975, 0.03463, 0.0, -0.0053, 0.01774, 0.0, -0.00615, -0.00403, -0.0, 0.00028, -0.02218, -0.0, -0.00028, -0.02218, -0.0, 0.00615, -0.00403, -0.0, 0.0053, 0.01774, 0.0, 0.00975, 0.03463, 0.0, 0.03463, 0.00975, 0.0, 0.01774, 0.0053, 0.0, -0.00403, 0.00615, 0.0, -0.02218, -0.00028, -0.0, -0.02218, 0.00028, 0.0, -0.00403, -0.00615, -0.0, 0.01774, -0.0053, -0.0, 0.03463, -0.00975, -0.0, 0.00975, -0.03463, -0.0, 0.0053, -0.01774, -0.0, 0.00615, 0.00403, 0.0, -0.00028, 0.02218, 0.0, 0.00028, 0.02218, 0.0, -0.00615, 0.00403, 0.0, -0.0053, -0.01774, -0.0, -0.00975, -0.03463, -0.0, -0.00923, -0.00923, -0.0, -0.0717, 0.0, 0.0, -0.00923, 0.00923, 0.0, 0.0, 0.0717, 0.0, 0.00923, 0.00923, 0.0, 0.0717, -0.0, -0.0, 0.00923, -0.00923, -0.0, -0.0, -0.0717, -0.0]]), "A2g":array([[0.00397, 0.02828, -0.0, 0.00517, 0.02566, -0.0, 0.02684, 0.00175, 0.0, 0.01316, -0.02902, -0.0, -0.01316, -0.02902, 0.0, -0.02684, 0.00175, -0.0, -0.00517, 0.02566, 0.0, -0.00397, 0.02828, 0.0, 0.02828, -0.00397, -0.0, 0.02566, -0.00517, -0.0, 0.00175, -0.02684, -0.0, -0.02902, -0.01316, -0.0, -0.02902, 0.01316, -0.0, 0.00175, 0.02684, -0.0, 0.02566, 0.00517, -0.0, 0.02828, 0.00397, -0.0, -0.00397, -0.02828, 0.0, -0.00517, -0.02566, 0.0, -0.02684, -0.00175, -0.0, -0.01316, 0.02902, 0.0, 0.01316, 0.02902, -0.0, 0.02684, -0.00175, 0.0, 0.00517, -0.02566, -0.0, 0.00397, -0.02828, -0.0, -0.02828, 0.00397, 0.0, -0.02566, 0.00517, 0.0, -0.00175, 0.02684, 0.0, 0.02902, 0.01316, 0.0, 0.02902, -0.01316, 0.0, -0.00175, -0.02684, 0.0, -0.02566, -0.00517, 0.0, -0.02828, -0.00397, 0.0, -0.01225, 0.01225, 0.0, -0.0, 0.02183, 0.0, 0.01225, 0.01225, -0.0, 0.02183, 0.0, -0.0, 0.01225, -0.01225, -0.0, 0.0, -0.02183, -0.0, -0.01225, -0.01225, 0.0, -0.02183, -0.0, 0.0], [-0.02052, 0.03164, -0.0, -0.0091, -0.00723, -0.0, -0.03725, -0.01342, 0.0, -0.01606, -0.01624, 0.0, 0.01606, -0.01624, -0.0, 0.03725, -0.01342, -0.0, 0.0091, -0.00723, 0.0, 0.02052, 0.03164, -0.0, 0.03164, 0.02052, -0.0, -0.00723, 0.0091, -0.0, -0.01342, 0.03725, -0.0, -0.01624, 0.01606, -0.0, -0.01624, -0.01606, -0.0, -0.01342, -0.03725, -0.0, -0.00723, -0.0091, -0.0, 0.03164, -0.02052, -0.0, 0.02052, -0.03164, -0.0, 0.0091, 0.00723, 0.0, 0.03725, 0.01342, -0.0, 0.01606, 0.01624, -0.0, -0.01606, 0.01624, 0.0, -0.03725, 0.01342, 0.0, -0.0091, 0.00723, -0.0, -0.02052, -0.03164, 0.0, -0.03164, -0.02052, 0.0, 0.00723, -0.0091, 0.0, 0.01342, -0.03725, 0.0, 0.01624, -0.01606, 0.0, 0.01624, 0.01606, 0.0, 0.01342, 0.03725, 0.0, 0.00723, 0.0091, 0.0, -0.03164, 0.02052, 0.0, -0.00814, 0.00814, 0.0, -0.0, 0.00761, 0.0, 0.00814, 0.00814, -0.0, 0.00761, 0.0, -0.0, 0.00814, -0.00814, -0.0, 0.0, -0.00761, -0.0, -0.00814, -0.00814, 0.0, -0.00761, -0.0, 0.0]]), "B1g":array([[-0.00997, -0.00074, 0.0, -0.02561, 0.00112, 0.0, -0.03965, 0.00048, 0.0, -0.0537, 0.00047, 0.0, -0.0537, -0.00047, 0.0, -0.03965, -0.00048, 0.0, -0.02561, -0.00112, 0.0, -0.00997, 0.00074, 0.0, 0.00074, -0.00997, -0.0, -0.00112, -0.02561, -0.0, -0.00048, -0.03965, -0.0, -0.00047, -0.0537, -0.0, 0.00047, -0.0537, -0.0, 0.00048, -0.03965, -0.0, 0.00112, -0.02561, -0.0, -0.00074, -0.00997, -0.0, 0.00997, 0.00074, -0.0, 0.02561, -0.00112, -0.0, 0.03965, -0.00048, -0.0, 0.0537, -0.00047, -0.0, 0.0537, 0.00047, -0.0, 0.03965, 0.00048, -0.0, 0.02561, 0.00112, -0.0, 0.00997, -0.00074, -0.0, -0.00074, 0.00997, 0.0, 0.00112, 0.02561, 0.0, 0.00048, 0.03965, 0.0, 0.00047, 0.0537, 0.0, -0.00047, 0.0537, 0.0, -0.00048, 0.03965, 0.0, -0.00112, 0.02561, 0.0, 0.00074, 0.00997, 0.0, -0.00435, 0.00435, 0.0, -0.00428, -0.0, 0.0, -0.00435, -0.00435, 0.0, 0.0, -0.00428, -0.0, 0.00435, -0.00435, -0.0, 0.00428, 0.0, -0.0, 0.00435, 0.00435, -0.0, -0.0, 0.00428, 0.0], [0.02356, 0.03697, 0.0, 0.01309, 0.00131, 0.0, 0.00188, 0.01621, 0.0, -0.01454, 0.00123, 0.0, -0.01454, -0.00123, 0.0, 0.00188, -0.01621, 0.0, 0.01309, -0.00131, 0.0, 0.02356, -0.03697, -0.0, -0.03697, 0.02356, 0.0, -0.00131, 0.01309, 0.0, -0.01621, 0.00188, 0.0, -0.00123, -0.01454, -0.0, 0.00123, -0.01454, -0.0, 0.01621, 0.00188, -0.0, 0.00131, 0.01309, 0.0, 0.03697, 0.02356, -0.0, -0.02356, -0.03697, -0.0, -0.01309, -0.00131, -0.0, -0.00188, -0.01621, -0.0, 0.01454, -0.00123, -0.0, 0.01454, 0.00123, -0.0, -0.00188, 0.01621, -0.0, -0.01309, 0.00131, -0.0, -0.02356, 0.03697, 0.0, 0.03697, -0.02356, -0.0, 0.00131, -0.01309, -0.0, 0.01621, -0.00188, -0.0, 0.00123, 0.01454, 0.0, -0.00123, 0.01454, 0.0, -0.01621, -0.00188, 0.0, -0.00131, -0.01309, -0.0, -0.03697, -0.02356, 0.0, -0.00341, 0.00341, 0.0, 0.06227, -0.0, -0.0, -0.00341, -0.00341, 0.0, -0.0, 0.06227, 0.0, 0.00341, -0.00341, -0.0, -0.06227, 0.0, 0.0, 0.00341, 0.00341, -0.0, 0.0, -0.06227, -0.0]]), "B2g":array([[0.00758, 0.00619, 0.0, 0.00763, -0.00635, -0.0, 0.02075, -0.02907, -0.0, 0.00973, -0.05147, -0.0, -0.00973, -0.05147, -0.0, -0.02075, -0.02907, -0.0, -0.00763, -0.00635, -0.0, -0.00758, 0.00619, 0.0, -0.00619, 0.00758, 0.0, 0.00635, 0.00763, 0.0, 0.02907, 0.02075, 0.0, 0.05147, 0.00973, 0.0, 0.05147, -0.00973, -0.0, 0.02907, -0.02075, -0.0, 0.00635, -0.00763, -0.0, -0.00619, -0.00758, -0.0, -0.00758, -0.00619, -0.0, -0.00763, 0.00635, 0.0, -0.02075, 0.02907, 0.0, -0.00973, 0.05147, 0.0, 0.00973, 0.05147, 0.0, 0.02075, 0.02907, 0.0, 0.00763, 0.00635, 0.0, 0.00758, -0.00619, -0.0, 0.00619, -0.00758, -0.0, -0.00635, -0.00763, -0.0, -0.02907, -0.02075, -0.0, -0.05147, -0.00973, -0.0, -0.05147, 0.00973, 0.0, -0.02907, 0.02075, 0.0, -0.00635, 0.00763, 0.0, 0.00619, 0.00758, 0.0, 0.00929, 0.00929, 0.0, 0.0, 0.01149, 0.0, -0.00929, 0.00929, 0.0, -0.01149, -0.0, -0.0, -0.00929, -0.00929, -0.0, -0.0, -0.01149, -0.0, 0.00929, -0.00929, -0.0, 0.01149, 0.0, 0.0], [-0.01359, -0.0228, -0.0, -0.01088, -0.02118, -0.0, -0.02758, -0.01742, -0.0, -0.01331, -0.01324, -0.0, 0.01331, -0.01324, -0.0, 0.02758, -0.01742, -0.0, 0.01088, -0.02118, -0.0, 0.01359, -0.0228, -0.0, 0.0228, -0.01359, -0.0, 0.02118, -0.01088, -0.0, 0.01742, -0.02758, -0.0, 0.01324, -0.01331, -0.0, 0.01324, 0.01331, 0.0, 0.01742, 0.02758, 0.0, 0.02118, 0.01088, 0.0, 0.0228, 0.01359, 0.0, 0.01359, 0.0228, 0.0, 0.01088, 0.02118, 0.0, 0.02758, 0.01742, 0.0, 0.01331, 0.01324, 0.0, -0.01331, 0.01324, 0.0, -0.02758, 0.01742, 0.0, -0.01088, 0.02118, 0.0, -0.01359, 0.0228, 0.0, -0.0228, 0.01359, 0.0, -0.02118, 0.01088, 0.0, -0.01742, 0.02758, 0.0, -0.01324, 0.01331, 0.0, -0.01324, -0.01331, -0.0, -0.01742, -0.02758, -0.0, -0.02118, -0.01088, -0.0, -0.0228, -0.01359, -0.0, -0.02053, -0.02053, -0.0, 0.0, -0.01785, -0.0, 0.02053, -0.02053, -0.0, 0.01785, -0.0, -0.0, 0.02053, 0.02053, 0.0, -0.0, 0.01785, 0.0, -0.02053, 0.02053, 0.0, -0.01785, 0.0, 0.0]]), "A1u":array([[0.0, -0.0, 0.01155, 0.0, -0.0, 0.01877, 0.0, 0.0, 0.05966, 0.0, 0.0, 0.03502, 0.0, 0.0, -0.03502, 0.0, -0.0, -0.05966, 0.0, -0.0, -0.01877, 0.0, -0.0, -0.01155, -0.0, 0.0, 0.01155, -0.0, -0.0, 0.01877, 0.0, -0.0, 0.05966, 0.0, 0.0, 0.03502, 0.0, -0.0, -0.03502, -0.0, -0.0, -0.05966, -0.0, 0.0, -0.01877, 0.0, -0.0, -0.01155, -0.0, 0.0, 0.01155, -0.0, 0.0, 0.01877, -0.0, -0.0, 0.05966, -0.0, -0.0, 0.03502, -0.0, -0.0, -0.03502, -0.0, 0.0, -0.05966, -0.0, 0.0, -0.01877, -0.0, 0.0, -0.01155, 0.0, -0.0, 0.01155, 0.0, 0.0, 0.01877, -0.0, 0.0, 0.05966, -0.0, -0.0, 0.03502, -0.0, 0.0, -0.03502, 0.0, 0.0, -0.05966, 0.0, -0.0, -0.01877, -0.0, 0.0, -0.01155, -0.0, -0.0, 0.0, -0.0, -0.0, -0.0, -0.0, 0.0, 0.0, 0.0, -0.0, 0.0, 0.0, 0.0, -0.0, 0.0, 0.0, 0.0, 0.0, -0.0, -0.0, -0.0, 0.0, -0.0], [0.0, -0.0, 0.0021, 0.0, -0.0, 0.02577, -0.0, -0.0, 0.03057, -0.0, -0.0, -0.06657, -0.0, -0.0, 0.06657, -0.0, 0.0, -0.03057, -0.0, -0.0, -0.02577, -0.0, 0.0, -0.0021, -0.0, 0.0, 0.0021, -0.0, 0.0, 0.02577, -0.0, -0.0, 0.03057, 0.0, 0.0, -0.06657, -0.0, 0.0, 0.06657, 0.0, -0.0, -0.03057, 0.0, 0.0, -0.02577, -0.0, -0.0, -0.0021, -0.0, 0.0, 0.0021, -0.0, 0.0, 0.02577, 0.0, 0.0, 0.03057, 0.0, 0.0, -0.06657, 0.0, 0.0, 0.06657, 0.0, -0.0, -0.03057, 0.0, 0.0, -0.02577, 0.0, -0.0, -0.0021, 0.0, -0.0, 0.0021, 0.0, -0.0, 0.02577, 0.0, 0.0, 0.03057, -0.0, -0.0, -0.06657, 0.0, -0.0, 0.06657, -0.0, 0.0, -0.03057, -0.0, -0.0, -0.02577, 0.0, 0.0, -0.0021, -0.0, -0.0, 0.0, -0.0, -0.0, -0.0, -0.0, 0.0, 0.0, 0.0, -0.0, 0.0, 0.0, 0.0, -0.0, 0.0, 0.0, 0.0, 0.0, -0.0, -0.0, -0.0, 0.0, -0.0]]), "A2u":array([[0.0, 0.0, 0.0263, 0.0, -0.0, 0.00364, -0.0, 0.0, -0.01972, 0.0, -0.0, -0.04278, -0.0, -0.0, -0.04278, -0.0, 0.0, -0.01972, 0.0, 0.0, 0.00364, -0.0, 0.0, 0.0263, 0.0, -0.0, 0.0263, 0.0, 0.0, 0.00364, -0.0, 0.0, -0.01972, 0.0, 0.0, -0.04278, -0.0, 0.0, -0.04278, 0.0, 0.0, -0.01972, -0.0, -0.0, 0.00364, -0.0, -0.0, 0.0263, -0.0, -0.0, 0.0263, -0.0, 0.0, 0.00364, 0.0, -0.0, -0.01972, 0.0, 0.0, -0.04278, 0.0, 0.0, -0.04278, 0.0, -0.0, -0.01972, -0.0, -0.0, 0.00364, 0.0, -0.0, 0.0263, -0.0, 0.0, 0.0263, -0.0, -0.0, 0.00364, 0.0, -0.0, -0.01972, -0.0, -0.0, -0.04278, -0.0, -0.0, -0.04278, -0.0, -0.0, -0.01972, 0.0, 0.0, 0.00364, 0.0, 0.0, 0.0263, -0.0, -0.0, 0.02275, 0.0, -0.0, 0.04237, -0.0, 0.0, 0.02275, -0.0, -0.0, 0.04237, 0.0, 0.0, 0.02275, 0.0, -0.0, 0.04237, 0.0, -0.0, 0.02275, 0.0, -0.0, 0.04237], [0.0, -0.0, -0.00237, -0.0, 0.0, -0.03695, -0.0, 0.0, -0.01779, -0.0, -0.0, 0.03147, -0.0, -0.0, 0.03147, -0.0, 0.0, -0.01779, -0.0, 0.0, -0.03695, 0.0, 0.0, -0.00237, -0.0, -0.0, -0.00237, -0.0, -0.0, -0.03695, 0.0, -0.0, -0.01779, -0.0, 0.0, 0.03147, 0.0, 0.0, 0.03147, 0.0, -0.0, -0.01779, -0.0, 0.0, -0.03695, 0.0, 0.0, -0.00237, 0.0, 0.0, -0.00237, 0.0, -0.0, -0.03695, -0.0, 0.0, -0.01779, 0.0, 0.0, 0.03147, 0.0, 0.0, 0.03147, 0.0, -0.0, -0.01779, 0.0, -0.0, -0.03695, -0.0, -0.0, -0.00237, 0.0, 0.0, -0.00237, 0.0, 0.0, -0.03695, -0.0, 0.0, -0.01779, -0.0, -0.0, 0.03147, 0.0, -0.0, 0.03147, -0.0, -0.0, -0.01779, 0.0, -0.0, -0.03695, -0.0, -0.0, -0.00237, 0.0, 0.0, -0.01078, 0.0, -0.0, 0.06206, 0.0, -0.0, -0.01078, 0.0, -0.0, 0.06206, -0.0, -0.0, -0.01078, 0.0, 0.0, 0.06206, -0.0, 0.0, -0.01078, -0.0, 0.0, 0.06206]]), "B1u":array([[0.0, -0.0, 0.02356, -0.0, 0.0, 0.01855, 0.0, -0.0, 0.04134, 0.0, 0.0, 0.02154, 0.0, -0.0, -0.02154, 0.0, 0.0, -0.04134, -0.0, 0.0, -0.01855, 0.0, -0.0, -0.02356, -0.0, 0.0, -0.02356, 0.0, -0.0, -0.01855, -0.0, -0.0, -0.04134, -0.0, 0.0, -0.02154, -0.0, 0.0, 0.02154, -0.0, -0.0, 0.04134, -0.0, 0.0, 0.01855, 0.0, -0.0, 0.02356, 0.0, 0.0, 0.02356, 0.0, -0.0, 0.01855, -0.0, 0.0, 0.04134, -0.0, -0.0, 0.02154, -0.0, 0.0, -0.02154, -0.0, -0.0, -0.04134, 0.0, -0.0, -0.01855, -0.0, 0.0, -0.02356, 0.0, -0.0, -0.02356, -0.0, 0.0, -0.01855, 0.0, 0.0, -0.04134, 0.0, -0.0, -0.02154, -0.0, -0.0, 0.02154, 0.0, 0.0, 0.04134, 0.0, -0.0, 0.01855, -0.0, 0.0, 0.02356, 0.0, 0.0, 0.04003, 0.0, -0.0, 0.0, 0.0, -0.0, -0.04003, -0.0, -0.0, 0.0, -0.0, -0.0, 0.04003, -0.0, 0.0, -0.0, -0.0, 0.0, -0.04003, 0.0, 0.0, -0.0], [0.0, 0.0, 0.03922, -0.0, -0.0, 0.028, -0.0, 0.0, -0.03176, 0.0, 0.0, -0.01903, -0.0, -0.0, 0.01903, 0.0, 0.0, 0.03176, -0.0, 0.0, -0.028, -0.0, -0.0, -0.03922, -0.0, 0.0, -0.03922, 0.0, 0.0, -0.028, 0.0, -0.0, 0.03176, 0.0, -0.0, 0.01903, -0.0, 0.0, -0.01903, -0.0, 0.0, -0.03176, 0.0, 0.0, 0.028, -0.0, -0.0, 0.03922, -0.0, 0.0, 0.03922, 0.0, 0.0, 0.028, 0.0, -0.0, -0.03176, -0.0, -0.0, -0.01903, 0.0, 0.0, 0.01903, -0.0, 0.0, 0.03176, 0.0, -0.0, -0.028, 0.0, 0.0, -0.03922, 0.0, -0.0, -0.03922, -0.0, -0.0, -0.028, -0.0, 0.0, 0.03176, -0.0, 0.0, 0.01903, 0.0, -0.0, -0.01903, 0.0, -0.0, -0.03176, -0.0, -0.0, 0.028, 0.0, 0.0, 0.03922, -0.0, 0.0, 0.01396, 0.0, 0.0, -0.0, 0.0, -0.0, -0.01396, 0.0, 0.0, -0.0, 0.0, -0.0, 0.01396, -0.0, -0.0, 0.0, -0.0, 0.0, -0.01396, -0.0, -0.0, 0.0]]), "B2u":array([[-0.0, 0.0, 0.00884, 0.0, -0.0, 0.02469, -0.0, -0.0, 0.03898, 0.0, -0.0, 0.05164, 0.0, -0.0, 0.05164, -0.0, -0.0, 0.03898, 0.0, -0.0, 0.02469, -0.0, -0.0, 0.00884, 0.0, 0.0, -0.00884, -0.0, -0.0, -0.02469, -0.0, 0.0, -0.03898, -0.0, -0.0, -0.05164, -0.0, -0.0, -0.05164, 0.0, 0.0, -0.03898, -0.0, -0.0, -0.02469, -0.0, 0.0, -0.00884, 0.0, -0.0, 0.00884, -0.0, -0.0, 0.02469, 0.0, -0.0, 0.03898, -0.0, -0.0, 0.05164, -0.0, -0.0, 0.05164, 0.0, -0.0, 0.03898, -0.0, -0.0, 0.02469, 0.0, 0.0, 0.00884, -0.0, -0.0, -0.00884, 0.0, 0.0, -0.02469, 0.0, -0.0, -0.03898, -0.0, 0.0, -0.05164, 0.0, 0.0, -0.05164, -0.0, -0.0, -0.03898, 0.0, 0.0, -0.02469, 0.0, -0.0, -0.00884, -0.0, -0.0, -0.0, 0.0, 0.0, 0.0017, -0.0, 0.0, -0.0, -0.0, -0.0, -0.0017, 0.0, 0.0, 0.0, -0.0, -0.0, 0.0017, 0.0, -0.0, 0.0, -0.0, 0.0, -0.0017], [0.0, -0.0, -0.02967, -0.0, 0.0, -0.03449, -0.0, -0.0, -0.01081, -0.0, -0.0, 0.03038, -0.0, -0.0, 0.03038, -0.0, -0.0, -0.01081, 0.0, -0.0, -0.03449, -0.0, 0.0, -0.02967, -0.0, -0.0, 0.02967, -0.0, 0.0, 0.03449, 0.0, 0.0, 0.01081, 0.0, -0.0, -0.03038, -0.0, -0.0, -0.03038, 0.0, 0.0, 0.01081, 0.0, 0.0, 0.03449, 0.0, -0.0, 0.02967, -0.0, 0.0, -0.02967, 0.0, -0.0, -0.03449, 0.0, 0.0, -0.01081, 0.0, 0.0, 0.03038, 0.0, 0.0, 0.03038, 0.0, -0.0, -0.01081, -0.0, 0.0, -0.03449, 0.0, -0.0, -0.02967, 0.0, 0.0, 0.02967, 0.0, -0.0, 0.03449, -0.0, -0.0, 0.01081, -0.0, 0.0, -0.03038, 0.0, 0.0, -0.03038, -0.0, -0.0, 0.01081, -0.0, -0.0, 0.03449, -0.0, 0.0, 0.02967, -0.0, -0.0, -0.0, 0.0, -0.0, -0.0393, -0.0, 0.0, -0.0, 0.0, -0.0, 0.0393, 0.0, 0.0, 0.0, -0.0, 0.0, -0.0393, 0.0, -0.0, 0.0, -0.0, 0.0, 0.0393]]), "Egx":array([[-0.0, -0.0, -0.02145, -0.0, 0.0, -0.01813, 0.0, -0.0, -0.04251, -0.0, 0.0, -0.02219, -0.0, -0.0, 0.02219, 0.0, 0.0, 0.04251, -0.0, 0.0, 0.01813, -0.0, 0.0, 0.02145, -0.0, -0.0, 0.02823, 0.0, 0.0, 0.01238, -0.0, -0.0, -0.0146, -0.0, -0.0, -0.04185, 0.0, -0.0, -0.04185, 0.0, -0.0, -0.0146, -0.0, 0.0, 0.01238, 0.0, -0.0, 0.02823, -0.0, 0.0, 0.02145, 0.0, -0.0, 0.01813, -0.0, 0.0, 0.04251, 0.0, -0.0, 0.02219, 0.0, 0.0, -0.02219, -0.0, -0.0, -0.04251, 0.0, -0.0, -0.01813, -0.0, -0.0, -0.02145, 0.0, -0.0, -0.02823, -0.0, -0.0, -0.01238, 0.0, 0.0, 0.0146, 0.0, 0.0, 0.04185, -0.0, 0.0, 0.04185, -0.0, 0.0, 0.0146, 0.0, -0.0, -0.01238, -0.0, -0.0, -0.02823, 0.0, -0.0, -0.03212, -0.0, 0.0, -0.0, 0.0, -0.0, 0.03212, 0.0, -0.0, 0.0331, -0.0, 0.0, 0.03212, -0.0, -0.0, -0.0, -0.0, 0.0, -0.03212, -0.0, -0.0, -0.0331], [0.0, -0.0, 0.00613, 0.0, 0.0, 0.01904, -0.0, 0.0, 0.05644, 0.0, 0.0, 0.03367, 0.0, -0.0, -0.03367, -0.0, -0.0, -0.05644, 0.0, -0.0, -0.01904, -0.0, -0.0, -0.00613, -0.0, -0.0, 0.0344, 0.0, -0.0, 0.03413, -0.0, -0.0, 0.00468, 0.0, -0.0, -0.03096, -0.0, -0.0, -0.03096, 0.0, -0.0, 0.00468, -0.0, -0.0, 0.03413, 0.0, -0.0, 0.0344, -0.0, -0.0, -0.00613, -0.0, -0.0, -0.01904, 0.0, -0.0, -0.05644, -0.0, -0.0, -0.03367, -0.0, 0.0, 0.03367, 0.0, 0.0, 0.05644, -0.0, 0.0, 0.01904, -0.0, 0.0, 0.00613, 0.0, -0.0, -0.0344, -0.0, 0.0, -0.03413, 0.0, 0.0, -0.00468, -0.0, 0.0, 0.03096, 0.0, 0.0, 0.03096, -0.0, 0.0, -0.00468, 0.0, 0.0, -0.03413, -0.0, -0.0, -0.0344, 0.0, -0.0, -0.00958, -0.0, 0.0, -0.0, 0.0, -0.0, 0.00958, 0.0, -0.0, 0.04192, -0.0, 0.0, 0.00958, -0.0, -0.0, -0.0, -0.0, 0.0, -0.00958, -0.0, -0.0, -0.04192]]), "Egy":array([[0.0, 0.0, 0.031, -0.0, 0.0, 0.01456, 0.0, 0.0, -0.01372, 0.0, 0.0, -0.04453, 0.0, 0.0, -0.04453, 0.0, 0.0, -0.01372, -0.0, -0.0, 0.01456, 0.0, 0.0, 0.031, -0.0, 0.0, 0.02082, -0.0, 0.0, 0.01642, 0.0, -0.0, 0.03781, 0.0, -0.0, 0.01936, 0.0, -0.0, -0.01936, 0.0, -0.0, -0.03781, 0.0, 0.0, -0.01642, 0.0, -0.0, -0.02082, 0.0, -0.0, -0.031, 0.0, -0.0, -0.01456, -0.0, -0.0, 0.01372, -0.0, -0.0, 0.04453, -0.0, -0.0, 0.04453, -0.0, -0.0, 0.01372, 0.0, 0.0, -0.01456, 0.0, -0.0, -0.031, 0.0, -0.0, -0.02082, 0.0, -0.0, -0.01642, -0.0, 0.0, -0.03781, -0.0, 0.0, -0.01936, -0.0, 0.0, 0.01936, -0.0, 0.0, 0.03781, -0.0, -0.0, 0.01642, -0.0, 0.0, 0.02082, -0.0, -0.0, 0.03327, 0.0, -0.0, 0.03703, -0.0, 0.0, 0.03327, -0.0, 0.0, 0.0, 0.0, 0.0, -0.03327, 0.0, 0.0, -0.03703, 0.0, -0.0, -0.03327, 0.0, 0.0, -0.0], [-0.0, 0.0, -0.03606, -0.0, 0.0, -0.02822, 0.0, 0.0, -0.00042, -0.0, 0.0, 0.02239, -0.0, -0.0, 0.02239, 0.0, -0.0, -0.00042, -0.0, 0.0, -0.02822, 0.0, -0.0, -0.03606, 0.0, -0.0, 0.00882, -0.0, 0.0, 0.01943, -0.0, 0.0, 0.06333, -0.0, 0.0, 0.03935, 0.0, 0.0, -0.03935, -0.0, 0.0, -0.06333, -0.0, 0.0, -0.01943, 0.0, 0.0, -0.00882, 0.0, 0.0, 0.03606, 0.0, -0.0, 0.02822, -0.0, -0.0, 0.00042, 0.0, -0.0, -0.02239, 0.0, 0.0, -0.02239, -0.0, 0.0, 0.00042, 0.0, -0.0, 0.02822, 0.0, 0.0, 0.03606, -0.0, 0.0, -0.00882, 0.0, -0.0, -0.01943, 0.0, -0.0, -0.06333, 0.0, -0.0, -0.03935, -0.0, -0.0, 0.03935, 0.0, -0.0, 0.06333, 0.0, -0.0, 0.01943, -0.0, -0.0, 0.00882, 0.0, 0.0, -0.01196, 0.0, 0.0, -0.04006, 0.0, -0.0, -0.01196, 0.0, 0.0, 0.0, -0.0, -0.0, 0.01196, 0.0, -0.0, 0.04006, -0.0, 0.0, 0.01196, -0.0, 0.0, -0.0]]), "Egx+y":array([[-0.0, 0.0, 0.00753, 0.0, -0.0, -0.00189, -0.0, -0.0, -0.03981, 0.0, 0.0, -0.04848, 0.0, -0.0, -0.01832, 0.0, 0.0, 0.01839, -0.0, -0.0, 0.0237, 0.0, -0.0, 0.03914, -0.0, -0.0, 0.03914, -0.0, 0.0, 0.0237, -0.0, -0.0, 0.01839, 0.0, -0.0, -0.01832, 0.0, -0.0, -0.04848, -0.0, -0.0, -0.03981, -0.0, -0.0, -0.00189, -0.0, -0.0, 0.00753, -0.0, -0.0, -0.00753, -0.0, 0.0, 0.00189, 0.0, 0.0, 0.03981, -0.0, -0.0, 0.04848, -0.0, 0.0, 0.01832, -0.0, -0.0, -0.01839, 0.0, 0.0, -0.0237, -0.0, 0.0, -0.03914, 0.0, 0.0, -0.03914, 0.0, -0.0, -0.0237, 0.0, 0.0, -0.01839, -0.0, 0.0, 0.01832, -0.0, 0.0, 0.04848, 0.0, 0.0, 0.03981, 0.0, 0.0, 0.00189, 0.0, 0.0, -0.00753, 0.0, 0.0, 0.0, 0.0, 0.0, 0.02775, -0.0, 0.0, 0.05002, -0.0, -0.0, 0.02775, -0.0, -0.0, -0.0, -0.0, -0.0, -0.02775, 0.0, -0.0, -0.05002, 0.0, 0.0, -0.02775], [0.0, -0.0, -0.031, -0.0, 0.0, -0.03714, 0.0, -0.0, -0.0422, -0.0, -0.0, -0.00592, -0.0, -0.0, 0.04187, -0.0, -0.0, 0.03775, 0.0, 0.0, -0.0106, -0.0, 0.0, -0.01491, 0.0, 0.0, -0.01491, -0.0, -0.0, -0.0106, 0.0, 0.0, 0.03775, -0.0, -0.0, 0.04187, 0.0, 0.0, -0.00592, 0.0, -0.0, -0.0422, -0.0, 0.0, -0.03714, -0.0, 0.0, -0.031, 0.0, 0.0, 0.031, 0.0, -0.0, 0.03714, -0.0, 0.0, 0.0422, 0.0, 0.0, 0.00592, 0.0, 0.0, -0.04187, 0.0, 0.0, -0.03775, -0.0, -0.0, 0.0106, 0.0, -0.0, 0.01491, -0.0, -0.0, 0.01491, 0.0, 0.0, 0.0106, -0.0, -0.0, -0.03775, 0.0, 0.0, -0.04187, -0.0, -0.0, 0.00592, -0.0, 0.0, 0.0422, 0.0, -0.0, 0.03714, 0.0, -0.0, 0.031, -0.0, 0.0, 0.0, 0.0, -0.0, -0.02403, 0.0, 0.0, -0.00919, 0.0, -0.0, -0.02403, 0.0, -0.0, -0.0, -0.0, 0.0, 0.02403, -0.0, -0.0, 0.00919, -0.0, 0.0, 0.02403]]), "Egx-y":array([[-0.0, -0.0, -0.03821, -0.0, 0.0, -0.02318, 0.0, 0.0, -0.02109, 0.0, 0.0, 0.01618, 0.0, 0.0, 0.049, -0.0, 0.0, 0.04254, -0.0, -0.0, 0.00358, -0.0, 0.0, -0.00556, 0.0, 0.0, 0.00556, 0.0, 0.0, -0.00358, -0.0, -0.0, -0.04254, -0.0, -0.0, -0.049, -0.0, 0.0, -0.01618, -0.0, -0.0, 0.02109, 0.0, 0.0, 0.02318, 0.0, -0.0, 0.03821, -0.0, 0.0, 0.03821, 0.0, -0.0, 0.02318, -0.0, -0.0, 0.02109, -0.0, -0.0, -0.01618, -0.0, -0.0, -0.049, 0.0, -0.0, -0.04254, 0.0, 0.0, -0.00358, 0.0, -0.0, 0.00556, -0.0, -0.0, -0.00556, -0.0, -0.0, 0.00358, 0.0, 0.0, 0.04254, 0.0, 0.0, 0.049, 0.0, -0.0, 0.01618, 0.0, 0.0, -0.02109, -0.0, -0.0, -0.02318, -0.0, 0.0, -0.03821, 0.0, 0.0, -0.04959, -0.0, 0.0, -0.02585, 0.0, 0.0, 0.0, -0.0, 0.0, 0.02585, -0.0, -0.0, 0.04959, 0.0, -0.0, 0.02585, -0.0, -0.0, -0.0, 0.0, -0.0, -0.02585], [0.0, -0.0, -0.02173, 0.0, 0.0, -0.00486, -0.0, -0.0, 0.03868, -0.0, -0.0, 0.04153, -0.0, 0.0, -0.00789, 0.0, -0.0, -0.04034, 0.0, 0.0, -0.02922, 0.0, 0.0, -0.02586, 0.0, 0.0, 0.02586, -0.0, -0.0, 0.02922, -0.0, 0.0, 0.04034, -0.0, 0.0, 0.00789, 0.0, 0.0, -0.04153, -0.0, -0.0, -0.03868, 0.0, 0.0, 0.00486, 0.0, -0.0, 0.02173, -0.0, 0.0, 0.02173, -0.0, -0.0, 0.00486, 0.0, 0.0, -0.03868, 0.0, 0.0, -0.04153, 0.0, -0.0, 0.00789, -0.0, 0.0, 0.04034, -0.0, -0.0, 0.02922, -0.0, -0.0, 0.02586, -0.0, -0.0, -0.02586, 0.0, 0.0, -0.02922, 0.0, -0.0, -0.04034, 0.0, -0.0, -0.00789, -0.0, -0.0, 0.04153, 0.0, 0.0, 0.03868, -0.0, -0.0, -0.00486, -0.0, 0.0, -0.02173, -0.0, 0.0, -0.01971, 0.0, 0.0, -0.03003, -0.0, -0.0, -0.0, -0.0, -0.0, 0.03003, 0.0, -0.0, 0.01971, -0.0, -0.0, 0.03003, 0.0, 0.0, 0.0, 0.0, 0.0, -0.03003]]), "Eux":array([[-0.01243, -0.00387, 0.0, -0.01533, -0.00317, 0.0, -0.02083, -0.00151, 0.0, -0.02779, -0.00415, 0.0, -0.02779, 0.00415, 0.0, -0.02083, 0.00151, 0.0, -0.01533, 0.00317, 0.0, -0.01243, 0.00387, 0.0, -0.0115, 0.00723, 0.0, 0.00574, 0.00845, 0.0, 0.03642, 0.02299, 0.0, 0.07106, 0.01205, 0.0, 0.07106, -0.01205, -0.0, 0.03642, -0.02299, -0.0, 0.00574, -0.00845, -0.0, -0.0115, -0.00723, -0.0, -0.01243, -0.00387, -0.0, -0.01533, -0.00317, -0.0, -0.02083, -0.00151, -0.0, -0.02779, -0.00415, -0.0, -0.02779, 0.00415, -0.0, -0.02083, 0.00151, -0.0, -0.01533, 0.00317, -0.0, -0.01243, 0.00387, -0.0, -0.0115, 0.00723, 0.0, 0.00574, 0.00845, 0.0, 0.03642, 0.02299, 0.0, 0.07106, 0.01205, 0.0, 0.07106, -0.01205, -0.0, 0.03642, -0.02299, -0.0, 0.00574, -0.00845, -0.0, -0.0115, -0.00723, -0.0, -0.0105, -0.00605, -0.0, -0.01093, 0.0, 0.0, -0.0105, 0.00605, 0.0, -0.01875, -0.0, -0.0, -0.0105, -0.00605, -0.0, -0.01093, 0.0, -0.0, -0.0105, 0.00605, 0.0, -0.01875, -0.0, 0.0], [-0.02094, -0.00451, 0.0, -0.00118, 0.00092, 0.0, 0.03296, -0.00607, 0.0, 0.06671, -0.00113, 0.0, 0.06671, 0.00113, 0.0, 0.03296, 0.00607, 0.0, -0.00118, -0.00092, 0.0, -0.02094, 0.00451, 0.0, -0.02199, 0.00687, 0.0, -0.01198, 0.00959, 0.0, -0.00812, 0.02459, 0.0, 0.01225, 0.01283, 0.0, 0.01225, -0.01283, -0.0, -0.00812, -0.02459, -0.0, -0.01198, -0.00959, -0.0, -0.02199, -0.00687, -0.0, -0.02094, -0.00451, -0.0, -0.00118, 0.00092, -0.0, 0.03296, -0.00607, -0.0, 0.06671, -0.00113, -0.0, 0.06671, 0.00113, -0.0, 0.03296, 0.00607, -0.0, -0.00118, -0.00092, -0.0, -0.02094, 0.00451, -0.0, -0.02199, 0.00687, 0.0, -0.01198, 0.00959, 0.0, -0.00812, 0.02459, 0.0, 0.01225, 0.01283, 0.0, 0.01225, -0.01283, -0.0, -0.00812, -0.02459, -0.0, -0.01198, -0.00959, -0.0, -0.02199, -0.00687, -0.0, -0.02356, -0.00773, 0.0, -0.03146, 0.0, 0.0, -0.02356, 0.00773, 0.0, -0.01683, 0.0, -0.0, -0.02356, -0.00773, -0.0, -0.03146, 0.0, -0.0, -0.02356, 0.00773, -0.0, -0.01683, 0.0, 0.0]]), "Euy":array([[-0.00578, -0.00675, 0.0, -0.00975, 0.00659, 0.0, -0.02908, 0.03671, 0.0, -0.01376, 0.06661, 0.0, 0.01376, 0.06661, -0.0, 0.02908, 0.03671, -0.0, 0.00975, 0.00659, -0.0, 0.00578, -0.00675, -0.0, 0.0036, -0.01073, -0.0, 0.00176, -0.01535, -0.0, -0.00057, -0.02301, 0.0, 0.00165, -0.02827, 0.0, -0.00165, -0.02827, 0.0, 0.00057, -0.02301, 0.0, -0.00176, -0.01535, -0.0, -0.0036, -0.01073, 0.0, -0.00578, -0.00675, 0.0, -0.00975, 0.00659, 0.0, -0.02908, 0.03671, 0.0, -0.01376, 0.06661, 0.0, 0.01376, 0.06661, -0.0, 0.02908, 0.03671, -0.0, 0.00975, 0.00659, -0.0, 0.00578, -0.00675, -0.0, 0.0036, -0.01073, -0.0, 0.00176, -0.01535, 0.0, -0.00057, -0.02301, -0.0, 0.00165, -0.02827, -0.0, -0.00165, -0.02827, -0.0, 0.00057, -0.02301, -0.0, -0.00176, -0.01535, 0.0, -0.0036, -0.01073, 0.0, -0.00718, -0.01069, 0.0, -0.0, -0.01808, -0.0, 0.00718, -0.01069, -0.0, -0.0, -0.01211, -0.0, -0.00718, -0.01069, 0.0, -0.0, -0.01808, -0.0, 0.00718, -0.01069, -0.0, -0.0, -0.01211, 0.0], [0.00936, 0.02687, -0.0, 0.01219, 0.01719, -0.0, 0.03401, 0.00747, -0.0, 0.02092, -0.00882, -0.0, -0.02092, -0.00882, 0.0, -0.03401, 0.00747, 0.0, -0.01219, 0.01719, 0.0, -0.00936, 0.02687, 0.0, 0.00197, 0.01573, -0.0, 0.00084, -0.00312, 0.0, 0.00756, -0.03146, 0.0, -0.00108, -0.0595, 0.0, 0.00108, -0.0595, 0.0, -0.00756, -0.03146, 0.0, -0.00084, -0.00312, 0.0, -0.00197, 0.01573, -0.0, 0.00936, 0.02687, -0.0, 0.01219, 0.01719, -0.0, 0.03401, 0.00747, -0.0, 0.02092, -0.00882, -0.0, -0.02092, -0.00882, 0.0, -0.03401, 0.00747, 0.0, -0.01219, 0.01719, 0.0, -0.00936, 0.02687, 0.0, 0.00197, 0.01573, 0.0, 0.00084, -0.00312, -0.0, 0.00756, -0.03146, -0.0, -0.00108, -0.0595, -0.0, 0.00108, -0.0595, -0.0, -0.00756, -0.03146, -0.0, -0.00084, -0.00312, -0.0, -0.00197, 0.01573, 0.0, 0.00358, 0.02073, -0.0, 0.0, 0.00949, 0.0, -0.00358, 0.02073, 0.0, 0.0, 0.02032, -0.0, 0.00358, 0.02073, -0.0, 0.0, 0.00949, 0.0, -0.00358, 0.02073, 0.0, 0.0, 0.02032, 0.0]]), "Eux+y":array([[-0.01146, -0.01018, 0.0, -0.01544, 0.00255, 0.0, -0.02911, 0.02268, 0.0, -0.02579, 0.04334, 0.0, -0.01285, 0.04902, 0.0, -0.00106, 0.02648, 0.0, -0.00576, 0.00784, 0.0, -0.00447, -0.00311, 0.0, -0.00311, -0.00447, -0.0, 0.00784, -0.00576, -0.0, 0.02648, -0.00106, -0.0, 0.04902, -0.01285, -0.0, 0.04334, -0.02579, -0.0, 0.02268, -0.02911, -0.0, 0.00255, -0.01544, -0.0, -0.01018, -0.01146, -0.0, -0.01146, -0.01018, 0.0, -0.01544, 0.00255, 0.0, -0.02911, 0.02268, 0.0, -0.02579, 0.04334, 0.0, -0.01285, 0.04902, 0.0, -0.00106, 0.02648, 0.0, -0.00576, 0.00784, 0.0, -0.00447, -0.00311, 0.0, -0.00311, -0.00447, -0.0, 0.00784, -0.00576, -0.0, 0.02648, -0.00106, -0.0, 0.04902, -0.01285, -0.0, 0.04334, -0.02579, -0.0, 0.02268, -0.02911, -0.0, 0.00255, -0.01544, -0.0, -0.01018, -0.01146, -0.0, -0.01108, -0.01108, 0.0, -0.0074, -0.01168, -0.0, -0.00254, -0.00254, 0.0, -0.01168, -0.0074, 0.0, -0.01108, -0.01108, -0.0, -0.0074, -0.01168, -0.0, -0.00254, -0.00254, 0.0, -0.01168, -0.0074, 0.0], [0.02015, 0.01755, 0.0, 0.0076, 0.0065, 0.0, 0.00703, 0.00194, 0.0, -0.02646, -0.00886, 0.0, -0.05763, -0.00443, 0.0, -0.04702, 0.00822, 0.0, -0.01245, 0.00982, 0.0, 0.00444, 0.01551, 0.0, 0.01551, 0.00444, -0.0, 0.00982, -0.01245, -0.0, 0.00822, -0.04702, -0.0, -0.00443, -0.05763, -0.0, -0.00886, -0.02646, -0.0, 0.00194, 0.00703, -0.0, 0.0065, 0.0076, -0.0, 0.01755, 0.02015, -0.0, 0.02015, 0.01755, -0.0, 0.0076, 0.0065, -0.0, 0.00703, 0.00194, -0.0, -0.02646, -0.00886, -0.0, -0.05763, -0.00443, 0.0, -0.04702, 0.00822, 0.0, -0.01245, 0.00982, 0.0, 0.00444, 0.01551, -0.0, 0.01551, 0.00444, 0.0, 0.00982, -0.01245, -0.0, 0.00822, -0.04702, -0.0, -0.00443, -0.05763, -0.0, -0.00886, -0.02646, 0.0, 0.00194, 0.00703, 0.0, 0.0065, 0.0076, 0.0, 0.01755, 0.02015, 0.0, 0.01961, 0.01961, 0.0, 0.0182, 0.01047, 0.0, 0.00981, 0.00981, -0.0, 0.01047, 0.0182, -0.0, 0.01961, 0.01961, -0.0, 0.0182, 0.01047, -0.0, 0.00981, 0.00981, 0.0, 0.01047, 0.0182, 0.0]]), "Eux-y":array([[-0.00212, 0.00364, 0.0, -0.00247, -0.00353, -0.0, 0.00691, -0.02324, -0.0, -0.00772, -0.04698, -0.0, -0.02967, -0.0452, -0.0, -0.03618, -0.0258, -0.0, -0.01758, -0.00245, -0.0, -0.0125, 0.00661, -0.0, -0.00661, 0.0125, 0.0, 0.00245, 0.01758, 0.0, 0.0258, 0.03618, 0.0, 0.0452, 0.02967, 0.0, 0.04698, 0.00772, 0.0, 0.02324, -0.00691, 0.0, 0.00353, 0.00247, 0.0, -0.00364, 0.00212, 0.0, -0.00212, 0.00364, -0.0, -0.00247, -0.00353, -0.0, 0.00691, -0.02324, -0.0, -0.00772, -0.04698, -0.0, -0.02967, -0.0452, -0.0, -0.03618, -0.0258, -0.0, -0.01758, -0.00245, -0.0, -0.0125, 0.00661, -0.0, -0.00661, 0.0125, 0.0, 0.00245, 0.01758, 0.0, 0.0258, 0.03618, 0.0, 0.0452, 0.02967, 0.0, 0.04698, 0.00772, 0.0, 0.02324, -0.00691, 0.0, 0.00353, 0.00247, 0.0, -0.00364, 0.00212, -0.0, -0.00258, 0.00258, 0.0, -0.00801, 0.01294, 0.0, -0.0121, 0.0121, 0.0, -0.01294, 0.00801, -0.0, -0.00258, 0.00258, 0.0, -0.00801, 0.01294, 0.0, -0.0121, 0.0121, -0.0, -0.01294, 0.00801, -0.0], [-0.01257, 0.01666, 0.0, 0.00072, 0.00955, 0.0, 0.03217, 0.00074, 0.0, 0.05167, -0.015, 0.0, 0.03935, -0.00698, 0.0, 0.01462, 0.00574, 0.0, -0.00572, 0.01047, 0.0, -0.01708, 0.01845, 0.0, -0.01845, 0.01708, 0.0, -0.01047, 0.00572, 0.0, -0.00574, -0.01462, 0.0, 0.00698, -0.03935, 0.0, 0.015, -0.05167, 0.0, -0.00074, -0.03217, 0.0, -0.00955, -0.00072, 0.0, -0.01666, 0.01257, 0.0, -0.01257, 0.01666, -0.0, 0.00072, 0.00955, -0.0, 0.03217, 0.00074, -0.0, 0.05167, -0.015, -0.0, 0.03935, -0.00698, -0.0, 0.01462, 0.00574, -0.0, -0.00572, 0.01047, -0.0, -0.01708, 0.01845, -0.0, -0.01845, 0.01708, -0.0, -0.01047, 0.00572, -0.0, -0.00574, -0.01462, -0.0, 0.00698, -0.03935, -0.0, 0.015, -0.05167, -0.0, -0.00074, -0.03217, -0.0, -0.00955, -0.00072, -0.0, -0.01666, 0.01257, -0.0, -0.01481, 0.01481, 0.0, -0.01957, 0.01005, -0.0, -0.01912, 0.01912, 0.0, -0.01005, 0.01957, 0.0, -0.01481, 0.01481, -0.0, -0.01957, 0.01005, -0.0, -0.01912, 0.01912, -0.0, -0.01005, 0.01957, 0.0]])}, database_path = 'phthalocyanine_scsd_20220713.pkl', maxdist = 1.75)
model_objs_dict['subporphyrin'] = scsd_model('subporphyrin', array([[2.102, 1.1605, -0.0], [3.4535, 0.6902, -0.0], [3.4535, -0.6902, -0.0], [2.102, -1.1605, -0.0], [1.3862, -2.4009, -0.0], [-0.046, -2.4007, -0.0], [-1.1291, -3.3359, -0.0], [-2.3245, -2.6458, 0.0], [-2.056, -1.2402, 0.0], [-2.7723, 0.0, 0.0], [-2.056, 1.2402, 0.0], [-2.3245, 2.6458, 0.0], [-1.1291, 3.3359, 0.0], [-0.046, 2.4007, 0.0], [1.3862, 2.4009, -0.0], [1.3697, 0.0, -0.0], [-0.6848, -1.1862, 0.0], [-0.6848, 1.1862, 0.0]]), 'D3h' , pca = {"A'1":array([[-0.0588, -0.01183, 0.0, -0.10317, 0.00155, 0.0, -0.10317, -0.00155, 0.0, -0.0588, 0.01183, 0.0, -0.02962, 0.0513, 0.0, 0.01916, 0.05684, 0.0, 0.05293, 0.08857, -0.0, 0.05024, 0.09012, -0.0, 0.03964, 0.04501, -0.0, 0.05924, -0.0, -0.0, 0.03964, -0.04501, -0.0, 0.05024, -0.09012, -0.0, 0.05293, -0.08857, -0.0, 0.01916, -0.05684, -0.0, -0.02962, -0.0513, 0.0, -0.002, 0.0, 0.0, 0.001, 0.00173, -0.0, 0.001, -0.00173, -0.0], [-0.07558, 0.02663, -0.0, 0.06443, 0.02608, -0.0, 0.06443, -0.02608, -0.0, -0.07558, -0.02663, 0.0, -0.04123, 0.07142, 0.0, 0.06086, 0.05214, -0.0, -0.00963, -0.06883, -0.0, -0.0548, -0.04276, 0.0, 0.01473, 0.07877, 0.0, 0.08247, 0.0, -0.0, 0.01473, -0.07877, -0.0, -0.0548, 0.04276, 0.0, -0.00963, 0.06883, 0.0, 0.06086, -0.05214, -0.0, -0.04123, -0.07142, 0.0, -0.03429, 0.0, 0.0, 0.01715, 0.0297, -0.0, 0.01715, -0.0297, -0.0]]), "A'2":array([[-0.05623, -0.05558, -0.0, -0.08041, 0.02258, -0.0, 0.08041, 0.02258, -0.0, 0.05623, -0.05558, -0.0, -0.02571, -0.01484, -0.0, -0.02002, 0.07649, -0.0, 0.05976, 0.05835, -0.0, -0.02065, -0.08093, -0.0, -0.07625, -0.02091, -0.0, -0.0, 0.02969, 0.0, 0.07625, -0.02091, 0.0, 0.02065, -0.08093, 0.0, -0.05976, 0.05835, 0.0, 0.02002, 0.07649, 0.0, 0.02571, -0.01484, 0.0, -0.0, -0.05951, 0.0, -0.05154, 0.02976, -0.0, 0.05154, 0.02976, 0.0], [0.03973, -0.02227, 0.0, 0.05001, 0.07431, 0.0, -0.05001, 0.07431, 0.0, -0.03973, -0.02227, 0.0, -0.04341, -0.02506, 0.0, -0.03915, -0.02327, 0.0, 0.03935, -0.08047, 0.0, 0.08936, 0.00615, 0.0, 0.00058, 0.04555, -0.0, 0.0, 0.05012, -0.0, -0.00058, 0.04555, -0.0, -0.08936, 0.00615, -0.0, -0.03935, -0.08047, -0.0, 0.03915, -0.02327, -0.0, 0.04341, -0.02506, -0.0, 0.0, -0.08724, 0.0, -0.07555, 0.04362, -0.0, 0.07555, 0.04362, 0.0]]), "A''1":array([[0.0, 0.0, 0.12365, 0.0, 0.0, 0.11676, 0.0, 0.0, -0.11676, -0.0, -0.0, -0.12365, -0.0, 0.0, -0.0, 0.0, 0.0, 0.12365, -0.0, -0.0, 0.11676, -0.0, 0.0, -0.11676, -0.0, 0.0, -0.12365, -0.0, -0.0, -0.0, 0.0, 0.0, 0.12365, 0.0, -0.0, 0.11676, 0.0, -0.0, -0.11676, -0.0, 0.0, -0.12365, 0.0, -0.0, 0.0, -0.0, 0.0, 0.0, 0.0, 0.0, -0.0, -0.0, 0.0, -0.0], [-0.0, 0.0, 0.11744, -0.0, 0.0, -0.12438, 0.0, 0.0, 0.12438, -0.0, 0.0, -0.11744, -0.0, 0.0, -0.0, 0.0, 0.0, 0.11744, -0.0, -0.0, -0.12438, 0.0, 0.0, 0.12438, 0.0, 0.0, -0.11744, -0.0, -0.0, -0.0, -0.0, 0.0, 0.11744, -0.0, -0.0, -0.12438, 0.0, -0.0, 0.12438, -0.0, 0.0, -0.11744, 0.0, -0.0, 0.0, 0.0, 0.0, 0.0, -0.0, -0.0, -0.0, 0.0, -0.0, -0.0]]), "A''2":array([[0.0, -0.0, 0.0376, -0.0, -0.0, -0.12083, -0.0, -0.0, -0.12083, 0.0, -0.0, 0.0376, 0.0, -0.0, 0.01021, 0.0, -0.0, 0.0376, 0.0, 0.0, -0.12083, 0.0, 0.0, -0.12083, 0.0, -0.0, 0.0376, 0.0, 0.0, 0.01021, 0.0, -0.0, 0.0376, 0.0, 0.0, -0.12083, -0.0, 0.0, -0.12083, 0.0, -0.0, 0.0376, -0.0, 0.0, 0.01021, -0.0, 0.0, 0.15624, -0.0, 0.0, 0.15624, -0.0, 0.0, 0.15624], [0.0, -0.0, -0.00566, 0.0, 0.0, -0.05681, -0.0, 0.0, -0.05681, 0.0, -0.0, -0.00566, -0.0, 0.0, 0.22477, -0.0, -0.0, -0.00566, -0.0, 0.0, -0.05681, -0.0, -0.0, -0.05681, 0.0, -0.0, -0.00566, -0.0, 0.0, 0.22477, 0.0, -0.0, -0.00566, 0.0, -0.0, -0.05681, -0.0, -0.0, -0.05681, 0.0, -0.0, -0.00566, -0.0, 0.0, 0.22477, -0.0, 0.0, -0.09983, -0.0, 0.0, -0.09983, -0.0, 0.0, -0.09983]]), "E'1":array([[0.00831, 0.01729, -0.0, 0.1749, -0.007, -0.0, 0.1749, 0.007, -0.0, 0.00831, -0.01729, -0.0, -0.01252, 0.00805, -0.0, -0.02679, 0.03125, 0.0, -0.01795, 0.11832, 0.0, -0.03005, 0.11135, 0.0, -0.02133, 0.00613, 0.0, -0.02647, -0.0, 0.0, -0.02133, -0.00613, 0.0, -0.03005, -0.11135, -0.0, -0.01795, -0.11832, -0.0, -0.02679, -0.03125, -0.0, -0.01252, -0.00805, -0.0, -0.06876, -0.0, 0.0, -0.02695, -0.02414, -0.0, -0.02695, 0.02414, 0.0], [-0.0101, 0.02415, -0.0, -0.0243, -0.01695, 0.0, -0.0243, 0.01695, 0.0, -0.0101, -0.02415, 0.0, 0.02963, -0.05195, -0.0, -0.01732, -0.06472, -0.0, -0.10818, -0.02666, -0.0, -0.05783, 0.09444, 0.0, 0.06325, 0.02653, -0.0, 0.11961, 0.0, -0.0, 0.06325, -0.02653, -0.0, -0.05783, -0.09444, -0.0, -0.10818, 0.02666, 0.0, -0.01732, 0.06472, 0.0, 0.02963, 0.05195, 0.0, 0.03057, 0.0, 0.0, 0.04976, -0.01108, -0.0, 0.04976, 0.01108, 0.0]]), "E'2":array([[0.02092, 0.02884, -0.0, -0.06336, 0.05958, 0.0, -0.08939, 0.05509, 0.0, -0.028, 0.03925, 0.0, -0.01401, 0.01608, 0.0, 0.00548, -0.01134, -0.0, -0.08027, -0.13004, -0.0, -0.07248, -0.13453, -0.0, -0.01256, -0.00092, 0.0, 0.02093, -0.00409, -0.0, 0.048, -0.00462, -0.0, 0.0924, -0.04987, -0.0, 0.08328, -0.02508, -0.0, 0.01451, 0.03254, -0.0, 0.02093, 0.03625, 0.0, 0.02439, 0.01967, -0.0, 0.02439, 0.04224, -0.0, 0.00484, 0.03096, 0.0], [-0.01297, 0.02226, -0.0, 0.02397, 0.07861, 0.0, 0.01217, 0.07925, 0.0, -0.04825, 0.01027, 0.0, -0.06716, 0.00602, 0.0, -0.04099, -0.04703, 0.0, 0.01863, 0.03097, -0.0, 0.01751, 0.03162, -0.0, -0.02023, -0.05901, -0.0, 0.03879, -0.05515, -0.0, 0.03302, -0.03665, -0.0, 0.06255, 0.05017, -0.0, 0.05609, 0.06006, -0.0, 0.02576, -0.0001, -0.0, 0.03879, 0.06719, 0.0, -0.07386, -0.03105, 0.0, -0.07386, -0.12793, 0.0, 0.01004, -0.07949, -0.0]]), "E'3":array([[-0.02092, -0.03664, 0.0, -0.08771, -0.06883, 0.0, -0.07049, -0.07639, 0.0, 0.01011, -0.02142, 0.0, 0.01345, -0.0233, 0.0, 0.01349, -0.01947, 0.0, 0.1014, 0.02285, -0.0, 0.10346, 0.04155, -0.0, 0.04219, -0.00021, -0.0, 0.01345, -0.00548, -0.0, -0.01859, 0.00175, -0.0, -0.07255, 0.14079, -0.0, -0.08565, 0.13323, -0.0, 0.00778, 0.01697, -0.0, -0.00198, -0.01439, -0.0, 0.02724, -0.0135, -0.0, -0.00193, -0.03034, 0.0, 0.02724, -0.04718, 0.0], [0.03988, 0.01016, -0.0, 0.05529, -0.10836, -0.0, -0.02335, -0.10926, 0.0, -0.05306, 0.01983, 0.0, -0.04094, 0.07092, -0.0, 0.00935, 0.05587, -0.0, 0.10629, -0.03441, -0.0, 0.0662, -0.10206, -0.0, -0.02874, -0.02945, 0.0, -0.04094, 0.01259, 0.0, -0.01496, 0.00658, 0.0, 0.01675, -0.02722, 0.0, 0.0152, -0.02811, 0.0, 0.00178, 0.01625, 0.0, 0.00957, 0.04176, -0.0, -0.03614, 0.07402, 0.0, -0.04603, 0.06831, 0.0, -0.03614, 0.0626, -0.0]]), "E''1":array([[0.0, 0.0, 0.06521, 0.0, -0.0, 0.04736, 0.0, 0.0, -0.04736, -0.0, -0.0, -0.06521, 0.0, -0.0, -0.12962, 0.0, 0.0, -0.08639, -0.0, -0.0, 0.11643, -0.0, -0.0, 0.1638, -0.0, 0.0, -0.02118, 0.0, 0.0, 0.0, 0.0, 0.0, 0.02118, 0.0, -0.0, -0.1638, 0.0, -0.0, -0.11643, -0.0, 0.0, 0.08639, -0.0, 0.0, 0.12962, -0.0, 0.0, -0.0, 0.0, 0.0, -0.1421, -0.0, 0.0, 0.1421], [0.0, 0.0, 0.10218, 0.0, 0.0, 0.06725, 0.0, 0.0, -0.06725, -0.0, 0.0, -0.10218, -0.0, 0.0, -0.15219, 0.0, 0.0, 0.0427, -0.0, -0.0, -0.02549, -0.0, -0.0, 0.04175, -0.0, -0.0, 0.14488, -0.0, 0.0, 0.0, 0.0, 0.0, -0.14488, 0.0, -0.0, -0.04175, 0.0, -0.0, 0.02549, -0.0, -0.0, -0.0427, 0.0, 0.0, 0.15219, -0.0, 0.0, 0.0, -0.0, 0.0, 0.18782, -0.0, 0.0, -0.18782]]), "E''2":array([[0.0, 0.0, -0.01658, 0.0, 0.0, 0.15828, 0.0, 0.0, 0.11028, -0.0, 0.0, -0.08421, -0.0, 0.0, -0.11963, -0.0, 0.0, -0.06763, -0.0, -0.0, -0.048, -0.0, -0.0, 0.048, -0.0, -0.0, 0.06763, -0.0, 0.0, 0.11963, 0.0, -0.0, 0.08421, 0.0, -0.0, -0.11028, 0.0, -0.0, -0.15828, -0.0, -0.0, 0.01658, -0.0, 0.0, -0.0, -0.0, 0.0, -0.13914, 0.0, 0.0, 0.0, -0.0, 0.0, 0.13914], [-0.0, -0.0, -0.13725, -0.0, -0.0, -0.04029, -0.0, -0.0, 0.02575, 0.0, -0.0, -0.04377, 0.0, -0.0, 0.14641, 0.0, -0.0, 0.09348, 0.0, -0.0, 0.06604, 0.0, 0.0, -0.06604, 0.0, 0.0, -0.09348, 0.0, 0.0, -0.14641, -0.0, 0.0, 0.04377, -0.0, 0.0, -0.02575, -0.0, 0.0, 0.04029, -0.0, 0.0, 0.13725, -0.0, -0.0, 0.0, 0.0, -0.0, -0.17667, 0.0, -0.0, 0.0, 0.0, -0.0, 0.17667]]), "E''3":array([[0.0, 0.0, -0.08475, 0.0, 0.0, 0.11299, 0.0, 0.0, 0.16765, -0.0, 0.0, -0.01217, -0.0, 0.0, -0.0, -0.0, 0.0, 0.01217, -0.0, -0.0, -0.16765, -0.0, -0.0, -0.11299, -0.0, -0.0, 0.08475, -0.0, 0.0, 0.13354, 0.0, -0.0, 0.07257, 0.0, -0.0, 0.05466, 0.0, -0.0, -0.05466, 0.0, -0.0, -0.07257, -0.0, 0.0, -0.13354, -0.0, 0.0, -0.13427, -0.0, 0.0, 0.13427, -0.0, 0.0, -0.0], [0.0, 0.0, 0.04857, -0.0, -0.0, -0.03384, -0.0, -0.0, 0.03379, 0.0, 0.0, 0.14436, -0.0, 0.0, -0.0, -0.0, 0.0, -0.14436, 0.0, 0.0, -0.03379, 0.0, 0.0, 0.03384, -0.0, -0.0, -0.04857, -0.0, -0.0, 0.14151, 0.0, -0.0, 0.09579, -0.0, 0.0, 0.06763, -0.0, 0.0, -0.06763, 0.0, -0.0, -0.09579, 0.0, 0.0, -0.14151, 0.0, 0.0, 0.19003, -0.0, -0.0, -0.19003, 0.0, -0.0, 0.0]])}, database_path = 'subporphyrin_scsd_20220713.pkl', maxdist = 1.75)
model_objs_dict['bodipy'] = scsd_model('bodipy', array([['0.0', '0.0', '-1.2741', 'B'], ['2.554', '-0.0', '1.497', 'C'], ['3.3646', '-0.0', '0.3858', 'C'], ['2.5162', '0.0', '-0.7301', 'C'], ['-2.5162', '0.0', '-0.7301', 'C'], ['-3.3646', '-0.0', '0.3858', 'C'], ['-2.554', '-0.0', '1.497', 'C'], ['-0.0', '-0.0', '1.7428', 'C'], ['1.2126', '-0.0', '1.04', 'C'], ['-1.2126', '-0.0', '1.04', 'C'], ['0.0', '1.1244', '-2.0829', 'F'], ['0.0', '-1.1244', '-2.0829', 'F'], ['1.2315', '0.0', '-0.3442', 'N'], ['-1.2315', '0.0', '-0.3442', 'N']]), 'C2v' , pca = {"A1":array([[0.0, 0.0, -0.05815, -0.07521, -0.0, 0.03119, -0.0062, -0.0, 0.13454, 0.05905, -0.0, 0.06296, -0.05905, -0.0, 0.06296, 0.0062, -0.0, 0.13454, 0.07521, -0.0, 0.03119, 0.0, 0.0, -0.0961, -0.0175, 0.0, -0.06592, 0.0175, 0.0, -0.06592, 0.0, -0.00955, -0.01359, 0.0, 0.00955, -0.01359, 0.02922, 0.0, -0.04441, -0.02922, 0.0, -0.04441], [0.0, -0.0, 0.02844, -0.04865, 0.0, -0.0706, -0.11549, 0.0, -0.00622, -0.10765, -0.0, 0.06315, 0.10765, -0.0, 0.06315, 0.11549, 0.0, -0.00622, 0.04865, 0.0, -0.0706, 0.0, 0.0, -0.03333, -0.00843, 0.0, -0.03472, 0.00843, 0.0, -0.03472, -0.0, -0.05189, 0.04674, -0.0, 0.05189, 0.04674, -0.0485, -0.0, 0.01377, 0.0485, -0.0, 0.01377], [0.0, -0.0, 0.05561, 0.0638, 0.0, -0.03089, 0.06502, -0.0, -0.00378, 0.03524, 0.0, -0.01239, -0.03524, 0.0, -0.01239, -0.06502, -0.0, -0.00378, -0.0638, 0.0, -0.03089, 0.0, 0.0, -0.10132, 0.02813, 0.0, -0.06033, -0.02813, 0.0, -0.06033, -0.0, -0.03253, 0.17826, -0.0, 0.03253, 0.17826, 0.02888, -0.0, 0.00213, -0.02888, -0.0, 0.00213], [-0.0, -0.0, 0.08507, -0.07591, 0.0, -0.1034, 0.0795, 0.0, -0.01624, 0.07066, -0.0, 0.04069, -0.07066, -0.0, 0.04069, -0.0795, 0.0, -0.01624, 0.07591, 0.0, -0.1034, -0.0, -0.0, 0.1066, -0.0159, 0.0, -0.01717, 0.0159, 0.0, -0.01717, 0.0, -0.07095, -0.02986, 0.0, 0.07095, -0.02986, -0.00578, -0.0, 0.02863, 0.00578, -0.0, 0.02863], [0.0, 0.0, 0.08772, -0.08067, -0.0, 0.02874, -0.09057, 0.0, -0.00464, 0.06138, 0.0, -0.10014, -0.06138, 0.0, -0.10014, 0.09057, 0.0, -0.00464, 0.08067, -0.0, 0.02874, -0.0, -0.0, 0.0592, -0.03573, -0.0, 0.00128, 0.03573, -0.0, 0.00128, -0.0, 0.01952, 0.05619, -0.0, -0.01952, 0.05619, 0.02771, 0.0, -0.01745, -0.02771, 0.0, -0.01745], [-0.0, -0.0, 0.13803, 0.05771, -0.0, 0.0519, -0.03446, 0.0, 0.02537, 0.01517, 0.0, -0.04348, -0.01517, 0.0, -0.04348, 0.03446, -0.0, 0.02537, -0.05771, -0.0, 0.0519, 0.0, 0.0, -0.09163, 0.02312, 0.0, 0.00683, -0.02312, 0.0, 0.00683, 0.0, -0.14175, -0.07956, 0.0, 0.14175, -0.07956, -0.00036, -0.0, 0.02303, 0.00036, -0.0, 0.02303], [0.0, -0.0, -0.10433, -0.05658, 0.0, 0.07886, 0.06555, -0.0, -0.01937, -0.05796, 0.0, -0.04056, 0.05796, 0.0, -0.04056, -0.06555, -0.0, -0.01937, 0.05658, -0.0, 0.07886, -0.0, -0.0, 0.0285, -0.06904, -0.0, 0.00762, 0.06904, -0.0, 0.00762, -0.0, -0.10175, 0.04112, -0.0, 0.10175, 0.04112, -0.02201, -0.0, -0.00302, 0.02201, -0.0, -0.00302], [0.0, -0.0, 0.16394, -0.09508, -0.0, 0.01795, 0.08873, 0.0, -0.01334, -0.07436, 0.0, -0.00939, 0.07436, 0.0, -0.00939, -0.08873, -0.0, -0.01334, 0.09508, -0.0, 0.01795, 0.0, 0.0, -0.163, -0.02473, -0.0, 0.04226, 0.02473, -0.0, 0.04226, 0.0, 0.08061, -0.01855, 0.0, -0.08061, -0.01855, -0.01203, 0.0, 0.00513, 0.01203, 0.0, 0.00513], [-0.0, -0.0, 0.0792, 0.01936, -0.0, 0.04827, 0.03415, 0.0, 0.04884, -0.0554, 0.0, -0.03335, 0.0554, 0.0, -0.03335, -0.03415, -0.0, 0.04884, -0.01936, -0.0, 0.04827, -0.0, -0.0, 0.11764, 0.03078, 0.0, -0.14563, -0.03078, 0.0, -0.14563, 0.0, 0.03753, -0.01586, 0.0, -0.03753, -0.01586, -0.07211, 0.0, -0.00239, 0.07211, 0.0, -0.00239], [0.0, -0.0, 0.02403, -0.0582, -0.0, 0.10223, -0.00676, 0.0, -0.04994, -0.00259, -0.0, 0.08529, 0.00259, -0.0, 0.08529, 0.00676, 0.0, -0.04994, 0.0582, -0.0, 0.10223, -0.0, -0.0, 0.04969, 0.12516, 0.0, 0.00453, -0.12516, 0.0, 0.00453, -0.0, -0.01051, 0.02575, -0.0, 0.01051, 0.02575, 0.06361, -0.0, 0.04822, -0.06361, -0.0, 0.04822], [-0.0, -0.0, 0.03441, 0.01139, -0.0, 0.00043, 0.01188, -0.0, 0.15021, -0.01855, 0.0, 0.0051, 0.01855, 0.0, 0.0051, -0.01188, -0.0, 0.15021, -0.01139, -0.0, 0.00043, -0.0, -0.0, 0.08135, 0.02314, -0.0, 0.12303, -0.02314, -0.0, 0.12303, -0.0, 0.00163, 0.06152, -0.0, -0.00163, 0.06152, -0.04559, -0.0, 0.06555, 0.04559, -0.0, 0.06555], [-0.0, 0.0, 0.09249, 0.04957, -0.0, 0.06636, -0.00837, -0.0, -0.02702, 0.03314, 0.0, 0.10776, -0.03314, -0.0, 0.10776, 0.00837, 0.0, -0.02702, -0.04957, -0.0, 0.06636, -0.0, -0.0, 0.04241, -0.08516, -0.0, 0.02756, 0.08516, -0.0, 0.02756, -0.0, 0.00366, 0.02692, -0.0, -0.00366, 0.02692, -0.04798, 0.0, -0.07559, 0.04798, 0.0, -0.07559], [-0.0, 0.0, 0.04408, 0.00467, -0.0, -0.03915, 0.02179, -0.0, 0.03396, -0.08247, 0.0, -0.01442, 0.08247, 0.0, -0.01442, -0.02179, -0.0, 0.03396, -0.00467, 0.0, -0.03915, -0.0, -0.0, 0.07115, 0.04658, -0.0, 0.03333, -0.04658, -0.0, 0.03333, 0.0, -0.03315, -0.00376, 0.0, 0.03315, -0.00376, 0.10221, 0.0, -0.16379, -0.10221, 0.0, -0.16379]]), "A2":array([[-0.0, -0.0, -0.0, -0.0, 0.14594, 0.0, -0.0, 0.04222, -0.0, -0.0, -0.11327, -0.0, -0.0, 0.11327, 0.0, -0.0, -0.04222, 0.0, -0.0, -0.14594, -0.0, 0.0, 0.0, -0.0, -0.0, 0.02723, -0.0, -0.0, -0.02723, 0.0, 0.12148, 0.0, 0.0, -0.12148, 0.0, -0.0, -0.0, -0.10618, 0.0, -0.0, 0.10618, -0.0], [-0.0, -0.0, 0.0, 0.0, 0.07394, 0.0, 0.0, -0.0908, 0.0, 0.0, -0.08495, -0.0, 0.0, 0.08495, 0.0, 0.0, 0.0908, -0.0, 0.0, -0.07394, -0.0, -0.0, 0.0, 0.0, 0.0, 0.1298, 0.0, 0.0, -0.1298, -0.0, -0.10515, -0.0, -0.0, 0.10515, -0.0, 0.0, 0.0, 0.06914, 0.0, 0.0, -0.06914, -0.0], [0.0, -0.0, -0.0, -0.0, -0.01848, 0.0, -0.0, -0.04651, -0.0, -0.0, 0.08125, 0.0, -0.0, -0.08125, -0.0, -0.0, 0.04651, 0.0, -0.0, 0.01848, -0.0, -0.0, -0.0, -0.0, -0.0, 0.11862, -0.0, -0.0, -0.11862, 0.0, 0.16985, -0.0, 0.0, -0.16985, -0.0, -0.0, -0.0, 0.09416, -0.0, -0.0, -0.09416, 0.0], [-0.0, 0.0, 0.0, 0.0, 0.17136, -0.0, -0.0, 0.10122, -0.0, -0.0, 0.16635, -0.0, -0.0, -0.16635, 0.0, -0.0, -0.10122, 0.0, -0.0, -0.17136, 0.0, -0.0, 0.0, -0.0, -0.0, 0.01527, 0.0, -0.0, -0.01527, -0.0, -0.06152, -0.0, 0.0, 0.06152, -0.0, -0.0, -0.0, 0.03183, 0.0, -0.0, -0.03183, -0.0], [0.0, -0.0, 0.0, 0.0, -0.07811, 0.0, -0.0, 0.20216, -0.0, -0.0, -0.06809, 0.0, -0.0, 0.06809, -0.0, -0.0, -0.20216, 0.0, -0.0, 0.07811, -0.0, -0.0, -0.0, -0.0, -0.0, 0.09976, 0.0, -0.0, -0.09976, -0.0, -0.01819, -0.0, 0.0, 0.01819, -0.0, -0.0, -0.0, 0.05042, -0.0, -0.0, -0.05042, 0.0]]), "B1":array([[-0.00898, 0.0, 0.0, -0.04206, 0.0, 0.02475, 0.03039, -0.0, 0.07561, 0.08602, -0.0, 0.0091, 0.08602, -0.0, -0.0091, 0.03039, 0.0, -0.07561, -0.04206, 0.0, -0.02475, 0.04109, -0.0, 0.0, -0.00835, 0.0, -0.05382, -0.00835, -0.0, 0.05382, -0.12807, 0.0, 0.0, -0.12807, -0.0, 0.0, 0.04842, 0.0, -0.06522, 0.04842, -0.0, 0.06522], [0.09087, -0.0, 0.0, -0.11673, 0.0, -0.03953, -0.03153, 0.0, -0.066, 0.07218, 0.0, -0.02635, 0.07218, -0.0, 0.02635, -0.03153, -0.0, 0.066, -0.11673, 0.0, 0.03953, -0.08045, 0.0, 0.0, -0.00645, 0.0, -0.0654, -0.00645, -0.0, 0.0654, 0.06132, -0.0, -0.0, 0.06132, 0.0, -0.0, 0.02064, -0.0, -0.02128, 0.02064, -0.0, 0.02128], [0.05943, 0.0, -0.0, 0.00145, -0.0, 0.00644, 0.04443, 0.0, -0.06951, 0.02956, -0.0, 0.06574, 0.02956, 0.0, -0.06574, 0.04443, -0.0, 0.06951, 0.00145, -0.0, -0.00644, -0.14213, 0.0, -0.0, 0.05806, -0.0, 0.06708, 0.05806, -0.0, -0.06708, -0.09079, 0.0, 0.0, -0.09079, -0.0, -0.0, -0.0107, 0.0, 0.04848, -0.0107, 0.0, -0.04848], [-0.07999, 0.0, 0.0, -0.00709, -0.0, 0.11363, -0.02049, 0.0, -0.02232, 0.03154, 0.0, -0.16868, 0.03154, 0.0, 0.16868, -0.02049, 0.0, 0.02232, -0.00709, -0.0, -0.11363, -0.07225, -0.0, 0.0, 0.00442, -0.0, 0.07742, 0.00442, -0.0, -0.07742, -0.00749, 0.0, 0.0, -0.00749, -0.0, 0.0, 0.06655, 0.0, 0.01274, 0.06655, -0.0, -0.01274], [0.07848, 0.0, 0.0, 0.02404, 0.0, -0.06656, 0.12571, -0.0, 0.04747, 0.01228, -0.0, -0.07389, 0.01228, -0.0, 0.07389, 0.12571, 0.0, -0.04747, 0.02404, -0.0, 0.06656, -0.06466, 0.0, -0.0, -0.10873, -0.0, 0.03012, -0.10873, 0.0, -0.03012, 0.00599, 0.0, -0.0, 0.00599, -0.0, -0.0, -0.03777, 0.0, 0.00221, -0.03777, -0.0, -0.00221], [0.29403, 0.0, -0.0, 0.00411, 0.0, 0.0718, -0.06123, 0.0, -0.00803, -0.05866, 0.0, -0.05106, -0.05866, 0.0, 0.05106, -0.06123, -0.0, 0.00803, 0.00411, 0.0, -0.0718, 0.10972, -0.0, 0.0, 0.00066, 0.0, 0.02389, 0.00066, -0.0, -0.02389, -0.05473, 0.0, 0.0, -0.05473, -0.0, 0.0, -0.0384, 0.0, -0.02426, -0.0384, -0.0, 0.02426], [0.02332, -0.0, 0.0, 0.06314, -0.0, -0.10717, -0.13594, -0.0, 0.04721, 0.08935, -0.0, -0.01017, 0.08935, -0.0, 0.01017, -0.13594, 0.0, -0.04721, 0.06314, -0.0, 0.10717, -0.00336, 0.0, 0.0, -0.02177, -0.0, 0.064, -0.02177, 0.0, -0.064, -0.025, -0.0, 0.0, -0.025, -0.0, -0.0, 0.02411, -0.0, 0.04074, 0.02411, 0.0, -0.04074], [0.01242, -0.0, 0.0, -0.0798, 0.0, -0.07822, 0.01309, 0.0, 0.08312, -0.08548, -0.0, -0.04261, -0.08548, 0.0, 0.04261, 0.01309, -0.0, -0.08312, -0.0798, 0.0, 0.07822, -0.01897, 0.0, -0.0, 0.10434, -0.0, 0.029, 0.10434, 0.0, -0.029, -0.01205, 0.0, 0.0, -0.01205, 0.0, 0.0, 0.04157, -0.0, 0.01596, 0.04157, 0.0, -0.01596], [0.03127, -0.0, -0.0, 0.09573, -0.0, 0.02629, -0.03598, 0.0, 0.06449, -0.01642, -0.0, -0.01891, -0.01642, 0.0, 0.01891, -0.03598, 0.0, -0.06449, 0.09573, -0.0, -0.02629, -0.18722, 0.0, 0.0, 0.05459, 0.0, -0.09157, 0.05459, -0.0, 0.09157, 0.01318, -0.0, -0.0, 0.01318, 0.0, 0.0, -0.03129, 0.0, -0.05975, -0.03129, 0.0, 0.05975], [-0.01995, -0.0, 0.0, -0.02625, -0.0, 0.02524, 0.01327, -0.0, 0.04304, 0.08268, -0.0, -0.04451, 0.08268, -0.0, 0.04451, 0.01327, -0.0, -0.04304, -0.02625, 0.0, -0.02524, 0.05679, -0.0, 0.0, 0.07135, -0.0, -0.03403, 0.07135, -0.0, 0.03403, 0.00904, 0.0, -0.0, 0.00904, -0.0, -0.0, -0.14365, 0.0, 0.09924, -0.14365, 0.0, -0.09924], [0.03352, 0.0, 0.0, 0.02678, -0.0, 0.01053, 0.00269, -0.0, -0.00461, -0.04179, 0.0, -0.01349, -0.04179, -0.0, 0.01349, 0.00269, -0.0, 0.00461, 0.02678, 0.0, -0.01053, -0.00863, 0.0, -0.0, -0.04361, 0.0, -0.11927, -0.04361, -0.0, 0.11927, -0.03767, -0.0, -0.0, -0.03767, 0.0, -0.0, 0.09085, -0.0, 0.16604, 0.09085, 0.0, -0.16604]]), "B2":array([[0.0, -0.02156, 0.0, 0.0, -0.02436, 0.0, 0.0, 0.08383, 0.0, 0.0, 0.11427, 0.0, 0.0, 0.11427, 0.0, 0.0, 0.08383, 0.0, 0.0, -0.02436, -0.0, 0.0, -0.1171, 0.0, 0.0, -0.05309, 0.0, 0.0, -0.05309, -0.0, 0.0, -0.07925, -0.07581, -0.0, -0.07925, 0.07581, 0.0, 0.03364, 0.0, 0.0, 0.03364, 0.0], [0.0, -0.04302, -0.0, -0.0, 0.08851, -0.0, -0.0, 0.11816, 0.0, -0.0, -0.01028, -0.0, -0.0, -0.01028, 0.0, -0.0, 0.11816, 0.0, -0.0, 0.08851, 0.0, -0.0, -0.07806, -0.0, -0.0, -0.04405, 0.0, -0.0, -0.04405, -0.0, -0.0, 0.03778, 0.10024, 0.0, 0.03778, -0.10024, -0.0, -0.09968, -0.0, -0.0, -0.09968, 0.0], [0.0, 0.0892, -0.0, 0.0, -0.0783, 0.0, -0.0, -0.01694, 0.0, -0.0, 0.06395, 0.0, -0.0, 0.06395, 0.0, -0.0, -0.01694, -0.0, -0.0, -0.0783, -0.0, 0.0, -0.07963, -0.0, -0.0, -0.04062, -0.0, -0.0, -0.04062, 0.0, -0.0, 0.13093, 0.06463, 0.0, 0.13093, -0.06463, -0.0, 0.0483, 0.0, -0.0, 0.0483, -0.0], [-0.0, 0.03418, 0.0, 0.0, 0.08085, -0.0, 0.0, -0.03511, 0.0, 0.0, -0.08804, -0.0, 0.0, -0.08804, -0.0, 0.0, -0.03511, -0.0, 0.0, 0.08085, -0.0, 0.0, -0.24751, 0.0, 0.0, 0.06221, 0.0, 0.0, 0.06221, 0.0, 0.0, 0.01231, -0.03725, -0.0, 0.01231, 0.03725, 0.0, 0.06851, 0.0, 0.0, 0.06851, 0.0], [-0.0, -0.16952, 0.0, -0.0, -0.05969, 0.0, 0.0, -0.1051, -0.0, -0.0, -0.02104, 0.0, -0.0, -0.02104, 0.0, -0.0, -0.1051, 0.0, -0.0, -0.05969, 0.0, 0.0, -0.09316, -0.0, -0.0, -0.07261, -0.0, -0.0, -0.07261, 0.0, -0.0, -0.06163, 0.07407, 0.0, -0.06163, -0.07407, -0.0, -0.01914, 0.0, -0.0, -0.01914, 0.0], [0.0, 0.13391, -0.0, 0.0, -0.07351, -0.0, -0.0, 0.00248, -0.0, 0.0, -0.08949, -0.0, 0.0, -0.08949, 0.0, 0.0, 0.00248, 0.0, 0.0, -0.07351, 0.0, -0.0, -0.04476, 0.0, 0.0, -0.08119, -0.0, 0.0, -0.08119, -0.0, 0.0, -0.01057, -0.07143, -0.0, -0.01057, 0.07143, 0.0, -0.10444, 0.0, 0.0, -0.10444, -0.0], [0.0, -0.11099, 0.0, -0.0, -0.09494, 0.0, -0.0, 0.11409, -0.0, 0.0, -0.10696, 0.0, 0.0, -0.10696, 0.0, 0.0, 0.11409, -0.0, 0.0, -0.09494, 0.0, -0.0, 0.01967, 0.0, 0.0, 0.00852, 0.0, 0.0, 0.00852, -0.0, 0.0, 0.02994, -0.00571, -0.0, 0.02994, 0.00571, 0.0, 0.08006, 0.0, 0.0, 0.08006, -0.0]])}, database_path = 'bodipy_scsd_20220805.pkl', maxdist = 1.75)
#model_objs_dict['bodipy'] = scsd_model('bodipy', array([['0.0', '0.0', '-1.2741', 'B'], ['2.554', '-0.0', '1.497', 'C'], ['3.3646', '-0.0', '0.3858', 'C'], ['2.5162', '0.0', '-0.7301', 'C'], ['-2.5162', '0.0', '-0.7301', 'C'], ['-3.3646', '-0.0', '0.3858', 'C'], ['-2.554', '-0.0', '1.497', 'C'], ['-0.0', '-0.0', '1.7428', 'C'], ['1.2126', '-0.0', '1.04', 'C'], ['-1.2126', '-0.0', '1.04', 'C'], ['0.0', '1.1244', '-2.0829', 'F'], ['0.0', '-1.1244', '-2.0829', 'F'], ['1.2315', '0.0', '-0.3442', 'N'], ['-1.2315', '0.0', '-0.3442', 'N']]), 'C2v' , pca = {"A1":array([[-0.0, 0.0, -0.06189, -0.07264, -0.0, 0.02784, -0.00183, -0.0, 0.13011, 0.06253, -0.0, 0.05809, -0.06253, -0.0, 0.05809, 0.00183, -0.0, 0.13011, 0.07264, -0.0, 0.02784, 0.0, 0.0, -0.09927, -0.01686, 0.0, -0.06911, 0.01686, 0.0, -0.06911, 0.0, -0.00774, -0.01835, 0.0, 0.00774, -0.01835, 0.03061, 0.0, -0.048, -0.03061, 0.0, -0.048], [-0.0, 0.0, 0.02438, -0.0601, 0.0, -0.06529, -0.12085, 0.0, -0.00401, -0.10712, -0.0, 0.05713, 0.10712, -0.0, 0.05713, 0.12085, 0.0, -0.00401, 0.0601, 0.0, -0.06529, 0.0, 0.0, -0.03344, -0.01565, 0.0, -0.03431, 0.01565, 0.0, -0.03431, -0.0, -0.05143, 0.03986, -0.0, 0.05143, 0.03986, -0.05031, 0.0, 0.01116, 0.05031, 0.0, 0.01116]]), "A2":array([[0.0, 0.0, 0.0, 0.0, -0.10078, 0.0, 0.0, -0.14668, 0.0, 0.0, -0.12409, 0.0, 0.0, 0.12409, -0.0, 0.0, 0.14668, -0.0, 0.0, 0.10078, -0.0, -0.0, 0.0, 0.0, 0.0, -0.05098, -0.0, 0.0, 0.05098, 0.0, 0.05785, -0.0, 0.0, -0.05785, 0.0, -0.0, 0.0, -0.06426, -0.0, 0.0, 0.06426, 0.0], [0.0, 0.0, 0.0, 0.0, 0.15551, -0.0, 0.0, 0.05661, -0.0, 0.0, -0.10075, -0.0, 0.0, 0.10075, 0.0, 0.0, -0.05661, 0.0, 0.0, -0.15551, 0.0, 0.0, -0.0, 0.0, 0.0, 0.03205, -0.0, 0.0, -0.03205, 0.0, 0.11574, -0.0, 0.0, -0.11574, 0.0, -0.0, 0.0, -0.0998, 0.0, 0.0, 0.0998, -0.0]]), "B1":array([[-0.04997, 0.0, 0.0, 0.06066, -0.0, -0.08032, 0.01524, -0.0, -0.10254, -0.03324, -0.0, -0.07149, -0.03324, -0.0, 0.07149, 0.01524, -0.0, 0.10254, 0.06066, -0.0, 0.08032, 0.06635, -0.0, 0.0, 0.04087, -0.0, -0.03478, 0.04087, -0.0, 0.03478, -0.07463, 0.0, -0.0, -0.07463, 0.0, 0.0, -0.01709, 0.0, -0.03203, -0.01709, -0.0, 0.03203], [-0.01117, 0.0, 0.0, -0.04016, 0.0, 0.02196, 0.03069, -0.0, 0.0728, 0.08443, -0.0, 0.00675, 0.08443, -0.0, -0.00675, 0.03069, 0.0, -0.0728, -0.04016, 0.0, -0.02196, 0.04367, -0.0, 0.0, -0.00724, 0.0, -0.05515, -0.00724, -0.0, 0.05515, -0.13148, 0.0, 0.0, -0.13148, -0.0, 0.0, 0.0475, 0.0, -0.06669, 0.0475, -0.0, 0.06669]]), "B2":array([[-0.0, 0.02142, 0.0, 0.0, 0.02622, -0.0, -0.0, -0.08276, 0.0, -0.0, -0.11382, -0.0, -0.0, -0.11382, -0.0, -0.0, -0.08276, -0.0, -0.0, 0.02622, -0.0, -0.0, 0.11926, -0.0, -0.0, 0.05467, -0.0, -0.0, 0.05467, 0.0, -0.0, 0.07836, 0.07468, 0.0, 0.07836, -0.07468, -0.0, -0.03301, 0.0, -0.0, -0.03301, 0.0], [-0.0, -0.05654, 0.0, -0.0, 0.09528, 0.0, -0.0, 0.11743, 0.0, -0.0, -0.01998, -0.0, -0.0, -0.01998, 0.0, -0.0, 0.11743, 0.0, -0.0, 0.09528, 0.0, -0.0, -0.07077, -0.0, -0.0, -0.0415, 0.0, -0.0, -0.0415, -0.0, -0.0, 0.01967, 0.09319, 0.0, 0.01967, -0.09319, -0.0, -0.10725, -0.0, -0.0, -0.10725, 0.0]])}, database_path = 'bodipy_scsd_20220713.pkl', maxdist = 1.75)
model_objs_dict['subphthalocyanine'] = scsd_model('subphthalocyanine', array([['-1.978', '1.1419', '0.0', 'C'], ['-3.2574', '0.7147', '0.0', 'C'], ['-4.333', '1.4246', '0.0', 'C'], ['-4.333', '-1.4246', '0.0', 'C'], ['-3.2574', '-0.7147', '0.0', 'C'], ['-1.978', '-1.1418', '0.0', 'C'], ['0.0001', '-2.2839', '-0.0', 'C'], ['1.0098', '-3.1783', '-0.0', 'C'], ['0.9327', '-4.4648', '-0.0', 'C'], ['2.0869', '-5.0212', '-0.0', 'C'], ['3.305', '-4.3179', '-0.0', 'C'], ['3.4003', '-3.0402', '-0.0', 'C'], ['2.2476', '-2.4636', '-0.0', 'C'], ['1.9779', '-1.142', '-0.0', 'C'], ['1.9779', '1.1421', '-0.0', 'C'], ['2.2477', '2.4637', '-0.0', 'C'], ['3.4003', '3.0402', '-0.0', 'C'], ['3.305', '4.3178', '-0.0', 'C'], ['2.0869', '5.0211', '0.0', 'C'], ['0.9328', '4.4648', '0.0', 'C'], ['1.0098', '3.1784', '0.0', 'C'], ['0.0001', '2.2839', '0.0', 'C'], ['-5.3918', '0.7032', '0.0', 'C'], ['-5.3918', '-0.7033', '0.0', 'C'], ['-1.3344', '2.3112', '0.0', 'N'], ['-1.3615', '0.0', '0.0', 'N'], ['-1.3344', '-2.3112', '0.0', 'N'], ['0.6807', '-1.1791', '-0.0', 'N'], ['2.6687', '0.0', '-0.0', 'N'], ['0.6807', '1.1791', '-0.0', 'N']]), 'D3h' , pca = {"A'1":array([[-0.00317, 0.00245, -0.0, -0.02195, 0.00185, 0.0, -0.05496, 0.00089, 0.0, -0.05496, -0.00089, 0.0, -0.02195, -0.00185, 0.0, -0.00317, -0.00245, 0.0, -0.00054, -0.00397, -0.0, 0.00938, -0.01994, -0.0, 0.02671, -0.04804, -0.0, 0.04528, -0.07862, -0.0, 0.04544, -0.07852, -0.0, 0.02825, -0.04716, -0.0, 0.01258, -0.01809, -0.0, 0.0037, -0.00152, -0.0, 0.0037, 0.00152, -0.0, 0.01258, 0.01809, -0.0, 0.02825, 0.04716, -0.0, 0.04544, 0.07852, -0.0, 0.04528, 0.07862, -0.0, 0.02671, 0.04804, -0.0, 0.00938, 0.01994, -0.0, -0.00054, 0.00397, 0.0, -0.09072, 9e-05, 0.0, -0.09072, -9e-05, 0.0, -0.00384, 0.00666, 0.0, 0.00613, 0.0, -0.0, -0.00384, -0.00666, 0.0, -0.00306, 0.00531, 0.0, 0.00769, 0.0, -0.0, -0.00306, -0.00531, 0.0], [0.01387, -0.01796, 0.0, 0.0388, -0.01381, -0.0, 0.01441, -0.07989, -0.0, 0.01441, 0.07989, 0.0, 0.0388, 0.01381, -0.0, 0.01387, 0.01796, -0.0, 0.00862, 0.02099, 0.0, -0.00744, 0.04051, 0.0, 0.06198, 0.05243, -0.0, 0.02363, -0.01052, -0.0, -0.0027, -0.02572, -0.0, -0.07639, -0.02746, 0.0, -0.03136, 0.0267, 0.0, -0.02249, 0.00303, 0.0, -0.02249, -0.00303, 0.0, -0.03136, -0.0267, 0.0, -0.07639, 0.02746, 0.0, -0.0027, 0.02572, 0.0, 0.02363, 0.01052, -0.0, 0.06198, -0.05243, -0.0, -0.00744, -0.04051, -0.0, 0.00862, -0.02099, -0.0, -0.02093, -0.0152, 0.0, -0.02093, 0.0152, 0.0, 0.01369, -0.02371, -0.0, 0.01185, 0.0, -0.0, 0.01369, 0.02371, -0.0, -0.00592, 0.01026, 0.0, -0.02737, 0.0, 0.0, -0.00592, -0.01026, 0.0]]), "A'2":array([[0.01062, 0.00672, 0.0, 0.0212, -0.00751, 0.0, 0.04617, -0.0206, 0.0, -0.04617, -0.0206, 0.0, -0.0212, -0.00751, -0.0, -0.01062, 0.00672, 0.0, -0.01113, 0.00584, -0.0, -0.0041, 0.02212, -0.0, -0.00525, 0.05028, -0.0, 0.05742, 0.06033, -0.0, 0.08095, 0.01956, -0.0, 0.04092, -0.02968, 0.0, 0.01711, -0.01461, 0.0, -0.00051, -0.01255, -0.0, 0.00051, -0.01255, 0.0, -0.01711, -0.01461, -0.0, -0.04092, -0.02968, 0.0, -0.08095, 0.01956, -0.0, -0.05742, 0.06033, -0.0, 0.00525, 0.05028, -0.0, 0.0041, 0.02212, 0.0, 0.01113, 0.00584, 0.0, 0.02353, -0.07989, 0.0, -0.02353, -0.07989, 0.0, 0.00275, 0.00159, -0.0, -0.0, 0.01318, -0.0, -0.00275, 0.00159, -0.0, -0.01141, -0.00659, 0.0, 0.0, -0.00317, 0.0, 0.01141, -0.00659, 0.0], [-0.01348, 0.01113, 0.0, -0.0157, 0.01271, 0.0, -0.06198, 0.003, -0.0, 0.06198, 0.003, 0.0, 0.0157, 0.01271, -0.0, 0.01348, 0.01113, 0.0, -0.0029, -0.01724, 0.0, -0.00316, -0.01995, 0.0, 0.02839, -0.05517, -0.0, 0.0717, -0.01251, -0.0, 0.02502, 0.06835, -0.0, -0.03359, 0.05218, -0.0, -0.01886, 0.00724, -0.0, -0.01638, 0.00611, -0.0, 0.01638, 0.00611, 0.0, 0.01886, 0.00724, -0.0, 0.03359, 0.05218, 0.0, -0.02502, 0.06835, -0.0, -0.0717, -0.01251, -0.0, -0.02839, -0.05517, 0.0, 0.00316, -0.01995, 0.0, 0.0029, -0.01724, 0.0, -0.04668, -0.05584, 0.0, 0.04668, -0.05584, 0.0, -0.00594, -0.00343, 0.0, 0.0, 0.01034, -0.0, 0.00594, -0.00343, 0.0, -0.00896, -0.00517, 0.0, 0.0, 0.00686, -0.0, 0.00896, -0.00517, 0.0]]), "A''1":array([[0.0, -0.0, -0.02399, 0.0, 0.0, -0.03826, 0.0, 0.0, -0.11292, 0.0, 0.0, 0.11292, -0.0, 0.0, 0.03826, -0.0, -0.0, 0.02399, 0.0, -0.0, -0.02399, 0.0, 0.0, -0.03826, -0.0, -0.0, -0.11292, 0.0, -0.0, -0.062, 0.0, -0.0, 0.062, 0.0, -0.0, 0.11292, -0.0, -0.0, 0.03826, 0.0, -0.0, 0.02399, 0.0, -0.0, -0.02399, -0.0, 0.0, -0.03826, -0.0, 0.0, -0.11292, -0.0, -0.0, -0.062, -0.0, -0.0, 0.062, -0.0, -0.0, 0.11292, -0.0, 0.0, 0.03826, 0.0, -0.0, 0.02399, 0.0, 0.0, -0.062, 0.0, 0.0, 0.062, 0.0, -0.0, -0.0, 0.0, -0.0, 0.0, 0.0, -0.0, -0.0, 0.0, -0.0, 0.0, 0.0, -0.0, 0.0, 0.0, -0.0, -0.0], [0.0, -0.0, -0.12906, -0.0, -0.0, -0.03231, 0.0, -0.0, 0.04527, 0.0, 0.0, -0.04527, -0.0, -0.0, 0.03231, 0.0, -0.0, 0.12906, 0.0, -0.0, -0.12906, 0.0, 0.0, -0.03231, -0.0, -0.0, 0.04527, -0.0, 0.0, -0.01256, 0.0, 0.0, 0.01256, 0.0, 0.0, -0.04527, -0.0, 0.0, 0.03231, 0.0, -0.0, 0.12906, 0.0, -0.0, -0.12906, 0.0, 0.0, -0.03231, -0.0, 0.0, 0.04527, 0.0, -0.0, -0.01256, 0.0, -0.0, 0.01256, 0.0, -0.0, -0.04527, 0.0, -0.0, 0.03231, 0.0, -0.0, 0.12906, 0.0, -0.0, -0.01256, -0.0, -0.0, 0.01256, 0.0, -0.0, 0.0, -0.0, 0.0, 0.0, 0.0, -0.0, 0.0, -0.0, 0.0, 0.0, 0.0, -0.0, -0.0, 0.0, 0.0, -0.0]]), "A''2":array([[-0.0, 0.0, -0.04729, -0.0, -0.0, 5e-05, -0.0, -0.0, 0.0484, -0.0, -0.0, 0.0484, -0.0, 0.0, 5e-05, -0.0, 0.0, -0.04729, -0.0, 0.0, -0.04729, -0.0, 0.0, 5e-05, -0.0, 0.0, 0.0484, -0.0, 0.0, 0.09391, -0.0, 0.0, 0.09391, -0.0, 0.0, 0.0484, -0.0, 0.0, 5e-05, -0.0, 0.0, -0.04729, -0.0, 0.0, -0.04729, 0.0, 0.0, 5e-05, 0.0, 0.0, 0.0484, 0.0, 0.0, 0.09391, 0.0, 0.0, 0.09391, 0.0, -0.0, 0.0484, 0.0, 0.0, 5e-05, -0.0, 0.0, -0.04729, -0.0, -0.0, 0.09391, -0.0, -0.0, 0.09391, -0.0, 0.0, -0.03534, 0.0, 0.0, -0.08478, -0.0, 0.0, -0.03534, 0.0, -0.0, -0.08478, -0.0, 0.0, -0.03534, -0.0, 0.0, -0.08478], [-0.0, 0.0, -0.05844, -0.0, -0.0, -0.00417, -0.0, -0.0, -0.02288, -0.0, -0.0, -0.02288, -0.0, -0.0, -0.00417, -0.0, 0.0, -0.05844, -0.0, 0.0, -0.05844, -0.0, 0.0, -0.00417, -0.0, 0.0, -0.02288, -0.0, 0.0, -0.07882, -0.0, 0.0, -0.07882, -0.0, 0.0, -0.02288, -0.0, 0.0, -0.00417, -0.0, 0.0, -0.05844, 0.0, 0.0, -0.05844, 0.0, 0.0, -0.00417, 0.0, 0.0, -0.02288, 0.0, 0.0, -0.07882, 0.0, 0.0, -0.07882, 0.0, -0.0, -0.02288, 0.0, 0.0, -0.00417, -0.0, 0.0, -0.05844, -0.0, -0.0, -0.07882, -0.0, -0.0, -0.07882, -0.0, 0.0, -0.03921, -0.0, 0.0, -0.1192, -0.0, 0.0, -0.03921, -0.0, 0.0, -0.1192, -0.0, 0.0, -0.03921, -0.0, 0.0, -0.1192]]), "E'1":array([[-0.00102, 0.00192, -0.0, -0.02474, 0.00164, 0.0, -0.07795, -0.00181, 0.0, -0.07795, 0.00181, 0.0, -0.02474, -0.00164, 0.0, -0.00102, -0.00192, -0.0, 0.00389, 0.00593, 0.0, 0.01018, 0.02475, 0.0, 0.01092, 0.05525, 0.0, 0.01338, 0.08773, 0.0, 0.01491, 0.08518, 0.0, 0.01591, 0.05024, 0.0, 0.01274, 0.01705, 0.0, 0.00491, 0.00032, 0.0, 0.00491, -0.00032, -0.0, 0.01274, -0.01705, -0.0, 0.01591, -0.05024, -0.0, 0.01491, -0.08518, -0.0, 0.01338, -0.08773, -0.0, 0.01092, -0.05525, -0.0, 0.01018, -0.02475, -0.0, 0.00389, -0.00593, -0.0, -0.1356, -5e-05, 0.0, -0.1356, 5e-05, 0.0, 0.00304, -0.00428, -0.0, 0.00139, -0.0, -0.0, 0.00304, 0.00428, 0.0, -0.00491, -0.00364, -0.0, 0.01045, -0.0, -0.0, -0.00491, 0.00364, 0.0], [-0.01239, -0.00357, -0.0, -0.01254, 0.00235, 0.0, 0.01946, -0.00369, 0.0, 0.01946, 0.00369, 0.0, -0.01254, -0.00235, 0.0, -0.01239, 0.00357, 0.0, -0.00573, 0.01526, 0.0, 0.01167, 0.02906, 0.0, 0.03187, 0.05017, 0.0, 0.0807, 0.03653, 0.0, 0.10961, -0.00624, -0.0, 0.07231, -0.0125, -0.0, 0.02269, 0.00526, -0.0, 0.00725, -8e-05, 0.0, 0.00725, 8e-05, -0.0, 0.02269, -0.00526, -0.0, 0.07231, 0.0125, -0.0, 0.10961, 0.00624, -0.0, 0.0807, -0.03653, -0.0, 0.03187, -0.05017, -0.0, 0.01167, -0.02906, -0.0, -0.00573, -0.01526, -0.0, 0.06891, -0.00365, 0.0, 0.06891, 0.00365, 0.0, -0.00137, -0.01276, -0.0, -0.02173, 0.0, -0.0, -0.00137, 0.01276, 0.0, -0.01223, 0.00548, 0.0, 0.02074, 0.0, -0.0, -0.01223, -0.00548, -0.0]]), "E'2":array([[-0.00418, 0.00739, -0.0, -0.01313, 0.01765, 0.0, -0.03674, 0.03127, 0.0, -0.02432, 0.03542, 0.0, -0.00668, 0.01681, 0.0, 0.00191, 0.00448, 0.0, 0.00293, 0.00389, -0.0, 0.0179, 0.00262, -0.0, 0.04283, -0.00335, -0.0, 0.06529, -0.02186, -0.0, 0.06797, -0.02775, -0.0, 0.04545, -0.01618, -0.0, 0.02185, -0.00255, -0.0, 0.00849, 7e-05, -0.0, -0.00365, -0.00052, 0.0, -0.01063, -0.01674, 0.0, -0.02694, -0.05495, -0.0, -0.05425, -0.09523, -0.0, -0.05535, -0.0946, -0.0, -0.03412, -0.0508, 0.0, -0.00918, -0.01758, -0.0, 0.00138, -0.00342, -0.0, -0.05802, 0.04498, 0.0, -0.05158, 0.04562, 0.0, -0.00362, 0.00215, 0.0, 8e-05, -0.00609, -0.0, 0.00367, 0.00635, 0.0, -0.00532, -0.00297, 0.0, 0.00367, -0.00206, -0.0, 8e-05, 0.00015, 0.0], [-0.01976, -0.00381, -0.0, -0.01936, 0.01807, 0.0, -0.02047, 0.04069, 0.0, 0.03909, 0.05614, -0.0, 0.00155, 0.02226, 0.0, -0.00423, -0.00222, 0.0, 0.0002, -0.00477, -0.0, 0.0185, 0.01247, -0.0, 0.02908, 0.06192, 0.0, 0.04832, 0.09112, 0.0, 0.06144, 0.05088, -0.0, 0.04547, 0.00262, -0.0, 0.02532, -0.00773, -0.0, 0.00658, -0.01902, -0.0, -0.01061, -0.02157, -0.0, -0.00528, -0.01752, -0.0, 0.0227, 0.0084, -0.0, 0.04163, 0.0546, -0.0, 0.02647, 0.06335, 0.0, -0.00407, 0.02385, 0.0, -0.01254, -0.01333, 0.0, -0.01338, -0.01998, 0.0, 0.01334, 0.07865, 0.0, 0.05476, 0.0874, 0.0, -0.01472, -0.002, 0.0, -0.01384, -0.01363, 0.0, 0.00563, 0.00975, -0.0, -0.00488, -0.0188, -0.0, 0.00563, -0.01374, -0.0, -0.01384, -0.02397, 0.0]]), "E'3":array([[0.00323, -0.00616, 0.0, -0.0081, -0.01791, 0.0, -0.02764, -0.0351, 0.0, -0.03264, -0.03452, 0.0, -0.0133, -0.01837, 0.0, -0.00355, -0.0064, 0.0, 5e-05, 0.0004, 0.0, -0.01031, 0.01877, 0.0, -0.03064, 0.05192, 0.0, -0.05379, 0.0912, -0.0, -0.05209, 0.09218, -0.0, -0.02964, 0.0525, 0.0, -0.0111, 0.01831, 0.0, -0.00037, 0.00016, 0.0, 0.00732, -0.00013, -0.0, 0.02256, 0.00234, -0.0, 0.04622, 0.01101, -0.0, 0.06996, 0.02172, -0.0, 0.06996, 0.01975, -0.0, 0.04422, 0.00639, -0.0, 0.01956, -0.00194, -0.0, 0.00372, -0.00588, -0.0, -0.05209, -0.05071, 0.0, -0.05379, -0.04972, 0.0, 0.00434, -0.00752, 0.0, 0.00161, 0.00424, -0.0, -0.00182, -0.00396, 0.0, 0.00161, -0.0028, 0.0, 0.00434, -0.0004, -0.0, -0.00448, 0.00072, 0.0], [0.00675, -0.0046, 0.0, 0.00537, 0.01148, -0.0, -0.01655, 0.04879, -0.0, -0.00906, 0.0528, -0.0, 0.01083, 0.01284, -0.0, 0.00933, -0.00518, -0.0, 0.00854, -0.01363, 0.0, 0.00692, -0.01471, 0.0, -0.01627, 0.02017, 0.0, -0.03755, 0.06188, 0.0, -0.03482, 0.06346, 0.0, -0.00933, 0.02418, 0.0, 0.00928, -0.01335, 0.0, 0.00754, -0.01421, 0.0, -0.00017, -0.01067, 0.0, -0.01654, -0.00296, 0.0, -0.0412, 0.03425, 0.0, -0.06563, 0.07302, 0.0, -0.0585, 0.08221, -0.0, -0.03398, 0.03873, 0.0, -0.01262, 0.00109, 0.0, 0.00061, -0.00815, 0.0, -0.04194, 0.09177, -0.0, -0.03042, 0.09335, -0.0, -0.00088, 0.00152, -0.0, 0.01274, -0.01795, -0.0, 0.01071, -0.00517, -0.0, 0.01274, -0.02207, 0.0, -0.00088, -0.01186, 0.0, 0.00917, -0.02001, 0.0]]), "E''1":array([[0.0, -0.0, 0.03096, 0.0, 0.0, 0.02283, -0.0, 0.0, 0.05008, 0.0, 0.0, -0.05008, 0.0, 0.0, -0.02283, 0.0, 0.0, -0.03096, 0.0, -0.0, -0.03839, 0.0, -0.0, -0.00462, 0.0, -0.0, 0.04336, 0.0, -0.0, 0.12638, 0.0, -0.0, 0.1511, 0.0, -0.0, 0.09344, 0.0, -0.0, 0.01821, 0.0, -0.0, -0.00743, -0.0, -0.0, 0.00743, -0.0, -0.0, -0.01821, -0.0, -0.0, -0.09344, -0.0, 0.0, -0.1511, -0.0, 0.0, -0.12638, -0.0, 0.0, -0.04336, -0.0, 0.0, 0.00462, 0.0, -0.0, 0.03839, -0.0, 0.0, 0.02472, 0.0, 0.0, -0.02472, -0.0, 0.0, 0.04767, 0.0, -0.0, 0.0, 0.0, -0.0, -0.04767, 0.0, -0.0, -0.0332, -0.0, -0.0, -0.0, 0.0, -0.0, 0.0332], [0.0, -0.0, 0.05142, -0.0, 0.0, 0.0509, -0.0, -0.0, 0.14024, -0.0, -0.0, -0.14024, -0.0, -0.0, -0.0509, 0.0, -0.0, -0.05142, 0.0, -0.0, -0.01956, 0.0, 0.0, -0.0154, 0.0, 0.0, -0.0939, -0.0, 0.0, -0.09997, -0.0, 0.0, -0.01922, -0.0, 0.0, 0.04634, -0.0, -0.0, 0.03549, 0.0, -0.0, 0.03186, 0.0, -0.0, -0.03186, 0.0, -0.0, -0.03549, 0.0, -0.0, -0.04634, 0.0, 0.0, 0.01922, 0.0, 0.0, 0.09997, 0.0, 0.0, 0.0939, -0.0, 0.0, 0.0154, 0.0, -0.0, 0.01956, -0.0, -0.0, 0.08076, -0.0, -0.0, -0.08076, 0.0, -0.0, 0.06359, 0.0, -0.0, -0.0, 0.0, -0.0, -0.06359, 0.0, -0.0, 0.02285, 0.0, -0.0, -0.0, 0.0, -0.0, -0.02285]]), "E''2":array([[0.0, -0.0, 0.04121, -0.0, -0.0, 0.00508, -0.0, -0.0, -0.0531, -0.0, -0.0, -0.0893, -0.0, -0.0, -0.01349, -0.0, -0.0, 0.01243, 0.0, -0.0, -0.01243, -0.0, 0.0, 0.01349, -0.0, 0.0, 0.0893, -0.0, 0.0, 0.15875, -0.0, 0.0, 0.14199, -0.0, 0.0, 0.0531, -0.0, 0.0, -0.00508, 0.0, -0.0, -0.04121, 0.0, -0.0, -0.02878, 0.0, 0.0, -0.01857, 0.0, 0.0, -0.0362, 0.0, 0.0, -0.01676, 0.0, 0.0, 0.01676, 0.0, 0.0, 0.0362, 0.0, -0.0, 0.01857, 0.0, -0.0, 0.02878, -0.0, -0.0, -0.14199, -0.0, -0.0, -0.15875, 0.0, -0.0, 0.04695, 0.0, 0.0, 0.04037, 0.0, -0.0, 0.0, 0.0, -0.0, -0.04037, 0.0, -0.0, -0.04695, -0.0, -0.0, -0.0], [-0.0, 0.0, -0.01928, -0.0, 0.0, -0.01495, -0.0, 0.0, -0.08002, 0.0, 0.0, 0.05629, 0.0, 0.0, 0.03474, 0.0, 0.0, 0.03108, 0.0, -0.0, -0.03108, 0.0, -0.0, -0.03474, -0.0, -0.0, -0.05629, -0.0, 0.0, 0.00396, -0.0, 0.0, 0.08405, 0.0, -0.0, 0.08002, 0.0, -0.0, 0.01495, 0.0, -0.0, 0.01928, -0.0, -0.0, 0.05036, -0.0, -0.0, 0.04969, -0.0, -0.0, 0.1363, -0.0, 0.0, 0.08008, -0.0, 0.0, -0.08008, -0.0, 0.0, -0.1363, -0.0, 0.0, -0.04969, -0.0, 0.0, -0.05036, -0.0, 0.0, -0.08405, 0.0, 0.0, -0.00396, -0.0, 0.0, -0.06119, 0.0, -0.0, 0.01708, 0.0, 0.0, -0.0, 0.0, -0.0, -0.01708, -0.0, -0.0, 0.06119, -0.0, -0.0, -0.0]]), "E''3":array([[-0.0, 0.0, -0.00842, 0.0, 0.0, 0.01778, 0.0, 0.0, 0.09441, 0.0, 0.0, 0.04682, 0.0, 0.0, -0.00524, 0.0, 0.0, -0.03951, -0.0, -0.0, -0.0311, 0.0, -0.0, -0.02301, 0.0, -0.0, -0.04759, 0.0, -0.0, -0.02368, 0.0, -0.0, 0.02368, 0.0, -0.0, 0.04759, 0.0, -0.0, 0.02301, 0.0, 0.0, 0.0311, -0.0, 0.0, 0.03951, -0.0, -0.0, 0.00524, -0.0, -0.0, -0.04682, -0.0, -0.0, -0.13662, -0.0, -0.0, -0.1603, -0.0, -0.0, -0.09441, -0.0, 0.0, -0.01778, -0.0, 0.0, 0.00842, 0.0, 0.0, 0.1603, 0.0, 0.0, 0.13662, -0.0, 0.0, -0.0, -0.0, -0.0, -0.03461, 0.0, 0.0, -0.04817, -0.0, 0.0, -0.0, -0.0, -0.0, 0.04817, 0.0, -0.0, 0.03461], [0.0, 0.0, 0.0343, -0.0, 0.0, 0.03472, -0.0, 0.0, 0.05413, 0.0, 0.0, -0.08516, 0.0, 0.0, -0.01722, 0.0, 0.0, -0.01871, 0.0, -0.0, -0.05302, 0.0, -0.0, -0.05194, -0.0, 0.0, -0.13929, -0.0, 0.0, -0.08059, -0.0, 0.0, 0.08059, -0.0, 0.0, 0.13929, 0.0, -0.0, 0.05194, 0.0, -0.0, 0.05302, -0.0, -0.0, 0.01871, -0.0, -0.0, 0.01722, -0.0, -0.0, 0.08516, -0.0, 0.0, 0.09677, -0.0, 0.0, 0.01618, 0.0, 0.0, -0.05413, -0.0, 0.0, -0.03472, -0.0, 0.0, -0.0343, -0.0, 0.0, -0.01618, -0.0, 0.0, -0.09677, -0.0, 0.0, 0.0, 0.0, -0.0, 0.0244, 0.0, 0.0, -0.06219, 0.0, -0.0, 0.0, -0.0, -0.0, 0.06219, -0.0, -0.0, -0.0244]])}, database_path = 'subphthalocyanine_scsd_20220727.pkl', maxdist = 1.75, mondrian_limits = [-1.0, -1.0])

#model_objs_dict['blue_box'] = scsd_model('blue_box', array([['0.7383', '3.4705', '-0.0', 'C'], ['-5.1135', '1.3886', '0.0', 'C'], ['-4.9421', '2.8846', '0.0', 'C'], ['-2.8229', '3.3026', '1.1591', 'C'], ['-1.4762', '3.4344', '1.1823', 'C'], ['-0.7383', '3.4705', '0.0', 'C'], ['-1.4762', '3.4344', '-1.1823', 'C'], ['-2.8229', '3.3026', '-1.1591', 'C'], ['1.4762', '3.4344', '1.1823', 'C'], ['2.8229', '3.3026', '1.1591', 'C'], ['2.8229', '3.3026', '-1.1591', 'C'], ['1.4762', '3.4344', '-1.1823', 'C'], ['4.9421', '2.8846', '-0.0', 'C'], ['5.1135', '1.3886', '-0.0', 'C'], ['5.1503', '0.6819', '-1.1896', 'C'], ['5.1503', '-0.6819', '-1.1896', 'C'], ['5.1503', '-0.6819', '1.1896', 'C'], ['5.1503', '0.6819', '1.1896', 'C'], ['-1.4762', '-3.4344', '-1.1823', 'C'], ['-2.8229', '-3.3026', '-1.1591', 'C'], ['-2.8229', '-3.3026', '1.1591', 'C'], ['-1.4762', '-3.4344', '1.1823', 'C'], ['-4.9421', '-2.8846', '0.0', 'C'], ['-5.1135', '-1.3886', '0.0', 'C'], ['-5.1503', '-0.6819', '1.1896', 'C'], ['-5.1503', '0.6819', '1.1896', 'C'], ['-5.1503', '0.6819', '-1.1896', 'C'], ['-5.1503', '-0.6819', '-1.1896', 'C'], ['-0.7383', '-3.4705', '0.0', 'C'], ['5.1135', '-1.3886', '-0.0', 'C'], ['4.9421', '-2.8846', '-0.0', 'C'], ['2.8229', '-3.3026', '-1.1591', 'C'], ['1.4762', '-3.4344', '-1.1823', 'C'], ['0.7383', '-3.4705', '-0.0', 'C'], ['1.4762', '-3.4344', '1.1823', 'C'], ['2.8229', '-3.3026', '1.1591', 'C'], ['-3.4819', '3.1975', '0.0', 'N'], ['3.4819', '3.1975', '-0.0', 'N'], ['-3.4819', '-3.1975', '0.0', 'N'], ['3.4819', '-3.1975', '-0.0', 'N']]), 'D2h' , pca = {"Ag":array([[0.00059, -0.00333, -0.0, -0.00198, -5e-05, -0.0, -0.00196, 0.00066, -0.0, -0.0011, -0.00242, 0.03933, 0.00018, -0.00357, 0.04028, -0.00059, -0.00333, -0.0, 0.00018, -0.00357, -0.04028, -0.0011, -0.00242, -0.03933, -0.00018, -0.00357, 0.04028, 0.0011, -0.00242, 0.03933, 0.0011, -0.00242, -0.03933, -0.00018, -0.00357, -0.04028, 0.00196, 0.00066, 0.0, 0.00198, -5e-05, 0.0, 0.00205, 0.00056, -0.04063, 0.00205, -0.00056, -0.04063, 0.00205, -0.00056, 0.04063, 0.00205, 0.00056, 0.04063, 0.00018, 0.00357, -0.04028, -0.0011, 0.00242, -0.03933, -0.0011, 0.00242, 0.03933, 0.00018, 0.00357, 0.04028, -0.00196, -0.00066, -0.0, -0.00198, 5e-05, -0.0, -0.00205, -0.00056, 0.04063, -0.00205, 0.00056, 0.04063, -0.00205, 0.00056, -0.04063, -0.00205, -0.00056, -0.04063, -0.00059, 0.00333, 0.0, 0.00198, 5e-05, 0.0, 0.00196, -0.00066, 0.0, 0.0011, 0.00242, -0.03933, -0.00018, 0.00357, -0.04028, 0.00059, 0.00333, 0.0, -0.00018, 0.00357, 0.04028, 0.0011, 0.00242, 0.03933, 6e-05, -0.00201, -0.0, -6e-05, -0.00201, -0.0, 6e-05, 0.00201, 0.0, -6e-05, 0.00201, 0.0], [-1e-05, -0.04191, 0.0, -0.02223, 0.00051, -0.0, -0.00812, -0.00086, -0.0, -0.00254, -0.0256, -0.00295, -0.00095, -0.03806, -0.00319, 1e-05, -0.04191, -0.0, -0.00095, -0.03806, 0.00319, -0.00254, -0.0256, 0.00295, 0.00095, -0.03806, -0.00319, 0.00254, -0.0256, -0.00295, 0.00254, -0.0256, 0.00295, 0.00095, -0.03806, 0.00319, 0.00812, -0.00086, 0.0, 0.02223, 0.00051, 0.0, 0.0255, 0.00017, 0.00307, 0.0255, -0.00017, 0.00307, 0.0255, -0.00017, -0.00307, 0.0255, 0.00017, -0.00307, -0.00095, 0.03806, 0.00319, -0.00254, 0.0256, 0.00295, -0.00254, 0.0256, -0.00295, -0.00095, 0.03806, -0.00319, -0.00812, 0.00086, -0.0, -0.02223, -0.00051, -0.0, -0.0255, -0.00017, -0.00307, -0.0255, 0.00017, -0.00307, -0.0255, 0.00017, 0.00307, -0.0255, -0.00017, 0.00307, 1e-05, 0.04191, 0.0, 0.02223, -0.00051, 0.0, 0.00812, 0.00086, 0.0, 0.00254, 0.0256, 0.00295, 0.00095, 0.03806, 0.00319, -1e-05, 0.04191, 0.0, 0.00095, 0.03806, -0.00319, 0.00254, 0.0256, -0.00295, -0.00427, -0.0169, -0.0, 0.00427, -0.0169, -0.0, -0.00427, 0.0169, 0.0, 0.00427, 0.0169, 0.0]]), "B1g":array([[-0.02252, -0.00709, -0.0, -0.01307, 0.01568, -0.0, -0.02196, 0.01612, -0.0, -0.02386, 0.01932, -0.00083, -0.02321, 0.01284, -0.00083, -0.02252, 0.00709, -0.0, -0.02321, 0.01284, 0.00083, -0.02386, 0.01932, 0.00083, -0.02321, -0.01284, 0.00083, -0.02386, -0.01932, 0.00083, -0.02386, -0.01932, -0.00083, -0.02321, -0.01284, -0.00083, -0.02196, -0.01612, 0.0, -0.01307, -0.01568, 0.0, -0.00668, -0.01485, 1e-05, 0.00668, -0.01485, -1e-05, 0.00668, -0.01485, 1e-05, -0.00668, -0.01485, -1e-05, 0.02321, 0.01284, -0.00083, 0.02386, 0.01932, -0.00083, 0.02386, 0.01932, 0.00083, 0.02321, 0.01284, 0.00083, 0.02196, 0.01612, -0.0, 0.01307, 0.01568, -0.0, 0.00668, 0.01485, -1e-05, -0.00668, 0.01485, 1e-05, -0.00668, 0.01485, -1e-05, 0.00668, 0.01485, 1e-05, 0.02252, 0.00709, 0.0, 0.01307, -0.01568, 0.0, 0.02196, -0.01612, 0.0, 0.02386, -0.01932, 0.00083, 0.02321, -0.01284, 0.00083, 0.02252, -0.00709, 0.0, 0.02321, -0.01284, -0.00083, 0.02386, -0.01932, -0.00083, -0.02372, 0.02043, -0.0, -0.02372, -0.02043, -0.0, 0.02372, 0.02043, 0.0, 0.02372, -0.02043, 0.0], [0.01177, -0.01516, 0.0, 0.0174, -0.00144, -0.0, 0.01166, 0.00565, -0.0, 0.01274, 0.03328, 0.01434, 0.01372, 0.02797, 0.01083, 0.01177, 0.01516, -0.0, 0.01372, 0.02797, -0.01083, 0.01274, 0.03328, -0.01434, 0.01372, -0.02797, -0.01083, 0.01274, -0.03328, -0.01434, 0.01274, -0.03328, 0.01434, 0.01372, -0.02797, 0.01083, 0.01166, -0.00565, 0.0, 0.0174, 0.00144, 0.0, 0.01104, 0.0043, -0.00074, -0.01104, 0.0043, 0.00074, -0.01104, 0.0043, -0.00074, 0.01104, 0.0043, 0.00074, -0.01372, 0.02797, 0.01083, -0.01274, 0.03328, 0.01434, -0.01274, 0.03328, -0.01434, -0.01372, 0.02797, -0.01083, -0.01166, 0.00565, -0.0, -0.0174, -0.00144, -0.0, -0.01104, -0.0043, 0.00074, 0.01104, -0.0043, -0.00074, 0.01104, -0.0043, 0.00074, -0.01104, -0.0043, -0.00074, -0.01177, 0.01516, 0.0, -0.0174, 0.00144, 0.0, -0.01166, -0.00565, 0.0, -0.01274, -0.03328, -0.01434, -0.01372, -0.02797, -0.01083, -0.01177, -0.01516, 0.0, -0.01372, -0.02797, 0.01083, -0.01274, -0.03328, 0.01434, 0.01645, 0.02544, -0.0, 0.01645, -0.02544, 0.0, -0.01645, 0.02544, -0.0, -0.01645, -0.02544, 0.0]]), "B2g":array([[-0.0, 0.0, 0.01437, 0.0, -0.0, 0.00582, -0.0, 0.0, 0.01257, -0.01687, 0.03251, 0.01402, -0.01708, 0.03315, 0.01433, 0.0, -0.0, 0.01437, 0.01708, -0.03315, 0.01433, 0.01687, -0.03251, 0.01402, 0.01708, 0.03315, 0.01433, 0.01687, 0.03251, 0.01402, -0.01687, -0.03251, 0.01402, -0.01708, -0.03315, 0.01433, -0.0, 0.0, 0.01257, -0.0, -0.0, 0.00582, 0.00021, 0.02133, 0.00286, -0.00021, 0.02133, -0.00286, 0.00021, -0.02133, -0.00286, -0.00021, -0.02133, 0.00286, -0.01708, -0.03315, -0.01433, -0.01687, -0.03251, -0.01402, 0.01687, 0.03251, -0.01402, 0.01708, 0.03315, -0.01433, -0.0, -0.0, -0.01257, 0.0, 0.0, -0.00582, -0.00021, -0.02133, -0.00286, 0.00021, -0.02133, 0.00286, -0.00021, 0.02133, 0.00286, 0.00021, 0.02133, -0.00286, 0.0, 0.0, -0.01437, -0.0, 0.0, -0.00582, -0.0, -0.0, -0.01257, 0.01687, -0.03251, -0.01402, 0.01708, -0.03315, -0.01433, -0.0, 0.0, -0.01437, -0.01708, 0.03315, -0.01433, -0.01687, 0.03251, -0.01402, -0.0, -0.0, 0.01347, -0.0, -0.0, 0.01347, 0.0, 0.0, -0.01347, 0.0, 0.0, -0.01347], [-0.0, -0.0, 0.01722, -0.0, 0.0, 0.00773, 0.0, -0.0, 0.01283, -0.01347, -0.02652, 0.01452, -0.01396, -0.03051, 0.01713, 0.0, 0.0, 0.01722, 0.01396, 0.03051, 0.01713, 0.01347, 0.02652, 0.01452, 0.01396, -0.03051, 0.01713, 0.01347, -0.02652, 0.01452, -0.01347, 0.02652, 0.01452, -0.01396, 0.03051, 0.01713, 0.0, -0.0, 0.01283, -0.0, 0.0, 0.00773, 0.00133, 0.02965, 0.00361, -0.00133, 0.02965, -0.00361, 0.00133, -0.02965, -0.00361, -0.00133, -0.02965, 0.00361, -0.01396, 0.03051, -0.01713, -0.01347, 0.02652, -0.01452, 0.01347, -0.02652, -0.01452, 0.01396, -0.03051, -0.01713, 0.0, 0.0, -0.01283, 0.0, -0.0, -0.00773, -0.00133, -0.02965, -0.00361, 0.00133, -0.02965, 0.00361, -0.00133, 0.02965, 0.00361, 0.00133, 0.02965, -0.00361, 0.0, -0.0, -0.01722, -0.0, -0.0, -0.00773, 0.0, 0.0, -0.01283, 0.01347, 0.02652, -0.01452, 0.01396, 0.03051, -0.01713, -0.0, -0.0, -0.01722, -0.01396, -0.03051, -0.01713, -0.01347, -0.02652, -0.01452, 0.0, 0.0, 0.01322, -0.0, 0.0, 0.01322, 0.0, -0.0, -0.01322, -0.0, -0.0, -0.01322]]), "B3g":array([[0.0, -0.0, 0.00125, 0.0, -0.0, -0.00634, 0.0, 0.0, -0.00674, 0.0, -0.03588, -0.00459, 0.00022, -0.03717, -0.00242, -0.0, 0.0, -0.00125, -0.00022, 0.03717, -0.00242, -0.0, 0.03588, -0.00459, 0.00022, 0.03717, 0.00242, 0.0, 0.03588, 0.00459, -0.0, -0.03588, 0.00459, -0.00022, -0.03717, 0.00242, 0.0, 0.0, 0.00674, -0.0, -0.0, 0.00634, -0.04076, -0.00021, 0.00632, -0.04076, 0.00021, 0.00632, 0.04076, -0.00021, 0.00632, 0.04076, 0.00021, 0.00632, -0.00022, -0.03717, -0.00242, -0.0, -0.03588, -0.00459, 0.0, 0.03588, -0.00459, 0.00022, 0.03717, -0.00242, 0.0, -0.0, -0.00674, 0.0, 0.0, -0.00634, 0.04076, 0.00021, -0.00632, 0.04076, -0.00021, -0.00632, -0.04076, 0.00021, -0.00632, -0.04076, -0.00021, -0.00632, -0.0, -0.0, -0.00125, -0.0, 0.0, 0.00634, 0.0, -0.0, 0.00674, -0.0, 0.03588, 0.00459, -0.00022, 0.03717, 0.00242, 0.0, -0.0, 0.00125, 0.00022, -0.03717, 0.00242, 0.0, -0.03588, 0.00459, 0.0, -0.0, -0.00563, 0.0, -0.0, 0.00563, -0.0, 0.0, -0.00563, -0.0, 0.0, 0.00563], [0.0, 0.0, 0.00131, 0.0, 0.0, -0.0089, 0.0, 0.0, -0.01524, -0.00639, 0.02827, -0.0056, -0.00545, 0.0295, -0.00088, -0.0, 0.0, -0.00131, 0.00545, -0.0295, -0.00088, 0.00639, -0.02827, -0.0056, -0.00545, -0.0295, 0.00088, -0.00639, -0.02827, 0.0056, 0.00639, 0.02827, 0.0056, 0.00545, 0.0295, 0.00088, 0.0, 0.0, 0.01524, -0.0, -0.0, 0.0089, -0.04717, -0.00036, 0.00982, -0.04717, 0.00036, 0.00982, 0.04717, -0.00036, 0.00982, 0.04717, 0.00036, 0.00982, 0.00545, 0.0295, -0.00088, 0.00639, 0.02827, -0.0056, -0.00639, -0.02827, -0.0056, -0.00545, -0.0295, -0.00088, 0.0, -0.0, -0.01524, 0.0, 0.0, -0.0089, 0.04717, 0.00036, -0.00982, 0.04717, -0.00036, -0.00982, -0.04717, 0.00036, -0.00982, -0.04717, -0.00036, -0.00982, -0.0, -0.0, -0.00131, -0.0, 0.0, 0.0089, 0.0, -0.0, 0.01524, 0.00639, -0.02827, 0.0056, 0.00545, -0.0295, 0.00088, 0.0, -0.0, 0.00131, -0.00545, 0.0295, 0.00088, -0.00639, 0.02827, 0.0056, 0.0, 0.0, -0.0069, 0.0, 0.0, 0.0069, -0.0, -0.0, -0.0069, -0.0, -0.0, 0.0069]]), "Au":array([[-0.0, -0.0, -0.00861, -0.0, 0.0, 0.02405, -0.0, -0.0, 0.05013, 0.00044, 0.00445, 0.03298, 0.00065, 0.00481, 0.01726, -0.0, -0.0, 0.00861, -0.00065, -0.00481, 0.01726, -0.00044, -0.00445, 0.03298, 0.00065, -0.00481, -0.01726, 0.00044, -0.00445, -0.03298, -0.00044, 0.00445, -0.03298, -0.00065, 0.00481, -0.01726, -0.0, 0.0, -0.05013, -0.0, -0.0, -0.02405, 5e-05, -0.00515, -0.01173, -5e-05, -0.00515, 0.01173, 5e-05, 0.00515, 0.01173, -5e-05, 0.00515, -0.01173, 0.00065, -0.00481, -0.01726, 0.00044, -0.00445, -0.03298, -0.00044, 0.00445, -0.03298, -0.00065, 0.00481, -0.01726, -0.0, -0.0, -0.05013, -0.0, 0.0, -0.02405, 5e-05, -0.00515, -0.01173, -5e-05, -0.00515, 0.01173, 5e-05, 0.00515, 0.01173, -5e-05, 0.00515, -0.01173, -0.0, -0.0, -0.00861, -0.0, -0.0, 0.02405, -0.0, 0.0, 0.05013, 0.00044, 0.00445, 0.03298, 0.00065, 0.00481, 0.01726, -0.0, -0.0, 0.00861, -0.00065, -0.00481, 0.01726, -0.00044, -0.00445, 0.03298, -0.0, 0.0, 0.03931, -0.0, -0.0, -0.03931, -0.0, 0.0, -0.03931, -0.0, -0.0, 0.03931], [-0.0, 0.0, -0.00049, -0.0, -0.0, -0.00191, -0.0, -0.0, -0.0056, -0.00614, 0.0539, -0.00508, -0.00592, 0.05568, -0.00099, -0.0, -0.0, 0.00049, 0.00592, -0.05568, -0.00099, 0.00614, -0.0539, -0.00508, -0.00592, -0.05568, 0.00099, -0.00614, -0.0539, 0.00508, 0.00614, 0.0539, 0.00508, 0.00592, 0.05568, 0.00099, -0.0, 0.0, 0.0056, -0.0, 0.0, 0.00191, 0.00068, 0.00406, 0.00129, -0.00068, 0.00406, -0.00129, 0.00068, -0.00406, -0.00129, -0.00068, -0.00406, 0.00129, -0.00592, -0.05568, 0.00099, -0.00614, -0.0539, 0.00508, 0.00614, 0.0539, 0.00508, 0.00592, 0.05568, 0.00099, -0.0, -0.0, 0.0056, -0.0, -0.0, 0.00191, 0.00068, 0.00406, 0.00129, -0.00068, 0.00406, -0.00129, 0.00068, -0.00406, -0.00129, -0.00068, -0.00406, 0.00129, -0.0, -0.0, -0.00049, -0.0, 0.0, -0.00191, -0.0, 0.0, -0.0056, -0.00614, 0.0539, -0.00508, -0.00592, 0.05568, -0.00099, -0.0, -0.0, 0.00049, 0.00592, -0.05568, -0.00099, 0.00614, -0.0539, -0.00508, -0.0, 0.0, -0.00608, -0.0, -0.0, 0.00608, -0.0, 0.0, 0.00608, -0.0, -0.0, -0.00608]]), "B1u":array([[0.0, -0.0, 0.00255, -0.0, 0.0, -0.00033, -0.0, -0.0, 0.00103, -0.00576, 0.04349, 0.00157, -0.00601, 0.04443, 0.00205, 0.0, 0.0, 0.00255, 0.00601, -0.04443, 0.00205, 0.00576, -0.04349, 0.00157, 0.00601, 0.04443, 0.00205, 0.00576, 0.04349, 0.00157, -0.00576, -0.04349, 0.00157, -0.00601, -0.04443, 0.00205, 0.0, 0.0, 0.00103, 0.0, 0.0, -0.00033, -0.03013, -0.00023, -0.0003, -0.03013, 0.00023, -0.0003, 0.03013, -0.00023, -0.0003, 0.03013, 0.00023, -0.0003, 0.00601, 0.04443, 0.00205, 0.00576, 0.04349, 0.00157, -0.00576, -0.04349, 0.00157, -0.00601, -0.04443, 0.00205, 0.0, 0.0, 0.00103, 0.0, 0.0, -0.00033, -0.03013, -0.00023, -0.0003, -0.03013, 0.00023, -0.0003, 0.03013, -0.00023, -0.0003, 0.03013, 0.00023, -0.0003, 0.0, 0.0, 0.00255, 0.0, 0.0, -0.00033, 0.0, 0.0, 0.00103, -0.00576, 0.04349, 0.00157, -0.00601, 0.04443, 0.00205, 0.0, 0.0, 0.00255, 0.00601, -0.04443, 0.00205, 0.00576, -0.04349, 0.00157, 0.0, 0.0, 0.00152, 0.0, 0.0, 0.00152, 0.0, 0.0, 0.00152, 0.0, 0.0, 0.00152], [-0.0, 0.0, -0.00129, -0.0, -0.0, 0.00147, -0.0, -0.0, 0.00176, 0.00135, -0.02353, 0.00029, 0.0017, -0.02585, -0.00086, -0.0, -0.0, -0.00129, -0.0017, 0.02585, -0.00086, -0.00135, 0.02353, 0.00029, -0.0017, -0.02585, -0.00086, -0.00135, -0.02353, 0.00029, 0.00135, 0.02353, 0.00029, 0.0017, 0.02585, -0.00086, -0.0, -0.0, 0.00176, -0.0, -0.0, 0.00147, -0.07276, -0.00062, 0.00118, -0.07276, 0.00062, 0.00118, 0.07276, -0.00062, 0.00118, 0.07276, 0.00062, 0.00118, -0.0017, -0.02585, -0.00086, -0.00135, -0.02353, 0.00029, 0.00135, 0.02353, 0.00029, 0.0017, 0.02585, -0.00086, -0.0, -0.0, 0.00176, -0.0, -0.0, 0.00147, -0.07276, -0.00062, 0.00118, -0.07276, 0.00062, 0.00118, 0.07276, -0.00062, 0.00118, 0.07276, 0.00062, 0.00118, -0.0, -0.0, -0.00129, -0.0, -0.0, 0.00147, -0.0, -0.0, 0.00176, 0.00135, -0.02353, 0.00029, 0.0017, -0.02585, -0.00086, -0.0, -0.0, -0.00129, -0.0017, 0.02585, -0.00086, -0.00135, 0.02353, 0.00029, -0.0, -0.0, 0.0004, -0.0, -0.0, 0.0004, -0.0, -0.0, 0.0004, -0.0, -0.0, 0.0004]]), "B2u":array([[0.00029, -0.00015, -0.0, -0.0005, -0.00014, 0.0, 0.00114, -0.00119, 0.0, -0.0004, 0.00066, -0.04079, 0.00063, 0.00067, -0.04134, 0.00029, 0.00015, 0.0, 0.00063, 0.00067, 0.04134, -0.0004, 0.00066, 0.04079, 0.00063, -0.00067, 0.04134, -0.0004, -0.00066, 0.04079, -0.0004, -0.00066, -0.04079, 0.00063, -0.00067, -0.04134, 0.00114, 0.00119, -0.0, -0.0005, 0.00014, -0.0, -2e-05, 0.00049, 0.0415, -2e-05, -0.00049, 0.0415, -2e-05, -0.00049, -0.0415, -2e-05, 0.00049, -0.0415, 0.00063, -0.00067, 0.04134, -0.0004, -0.00066, 0.04079, -0.0004, -0.00066, -0.04079, 0.00063, -0.00067, -0.04134, 0.00114, 0.00119, 0.0, -0.0005, 0.00014, 0.0, -2e-05, 0.00049, 0.0415, -2e-05, -0.00049, 0.0415, -2e-05, -0.00049, -0.0415, -2e-05, 0.00049, -0.0415, 0.00029, -0.00015, 0.0, -0.0005, -0.00014, -0.0, 0.00114, -0.00119, -0.0, -0.0004, 0.00066, -0.04079, 0.00063, 0.00067, -0.04134, 0.00029, 0.00015, -0.0, 0.00063, 0.00067, 0.04134, -0.0004, 0.00066, 0.04079, -0.00063, 0.00064, 0.0, -0.00063, -0.00064, -0.0, -0.00063, -0.00064, -0.0, -0.00063, 0.00064, 0.0], [-0.01199, -0.01458, -0.0, 0.02217, 0.00024, -0.0, -0.00137, 0.00507, -0.0, -0.01074, 0.03256, -0.0008, -0.01236, 0.02601, -0.00017, -0.01199, 0.01458, 0.0, -0.01236, 0.02601, 0.00017, -0.01074, 0.03256, 0.0008, -0.01236, -0.02601, 0.00017, -0.01074, -0.03256, 0.0008, -0.01074, -0.03256, -0.0008, -0.01236, -0.02601, -0.00017, -0.00137, -0.00507, 0.0, 0.02217, -0.00024, 0.0, 0.02751, 0.0005, -0.00185, 0.02751, -0.0005, -0.00185, 0.02751, -0.0005, 0.00185, 0.02751, 0.0005, 0.00185, -0.01236, -0.02601, 0.00017, -0.01074, -0.03256, 0.0008, -0.01074, -0.03256, -0.0008, -0.01236, -0.02601, -0.00017, -0.00137, -0.00507, -0.0, 0.02217, -0.00024, -0.0, 0.02751, 0.0005, -0.00185, 0.02751, -0.0005, -0.00185, 0.02751, -0.0005, 0.00185, 0.02751, 0.0005, 0.00185, -0.01199, -0.01458, -0.0, 0.02217, 0.00024, 0.0, -0.00137, 0.00507, 0.0, -0.01074, 0.03256, -0.0008, -0.01236, 0.02601, -0.00017, -0.01199, 0.01458, 0.0, -0.01236, 0.02601, 0.00017, -0.01074, 0.03256, 0.0008, -0.0109, 0.02764, 0.0, -0.0109, -0.02764, -0.0, -0.0109, -0.02764, -0.0, -0.0109, 0.02764, 0.0]]), "B3u":array([[0.00062, -0.05092, 0.0, -0.00901, -0.00596, -0.0, -0.00713, -0.00875, -0.0, -0.00299, -0.02617, -9e-05, -0.00276, -0.04487, 0.00041, -0.00062, -0.05092, -0.0, -0.00276, -0.04487, -0.00041, -0.00299, -0.02617, 9e-05, 0.00276, -0.04487, 0.00041, 0.00299, -0.02617, -9e-05, 0.00299, -0.02617, 9e-05, 0.00276, -0.04487, -0.00041, 0.00713, -0.00875, 0.0, 0.00901, -0.00596, 0.0, 0.0047, -0.00824, 0.00061, -0.0047, -0.00824, -0.00061, -0.0047, -0.00824, 0.00061, 0.0047, -0.00824, -0.00061, 0.00276, -0.04487, 0.00041, 0.00299, -0.02617, -9e-05, 0.00299, -0.02617, 9e-05, 0.00276, -0.04487, -0.00041, 0.00713, -0.00875, 0.0, 0.00901, -0.00596, 0.0, 0.0047, -0.00824, 0.00061, -0.0047, -0.00824, -0.00061, -0.0047, -0.00824, 0.00061, 0.0047, -0.00824, -0.00061, 0.00062, -0.05092, 0.0, -0.00901, -0.00596, -0.0, -0.00713, -0.00875, -0.0, -0.00299, -0.02617, -9e-05, -0.00276, -0.04487, 0.00041, -0.00062, -0.05092, 0.0, -0.00276, -0.04487, -0.00041, -0.00299, -0.02617, 9e-05, -0.00487, -0.0173, -0.0, 0.00487, -0.0173, -0.0, 0.00487, -0.0173, 0.0, -0.00487, -0.0173, 0.0], [-0.00289, 0.01801, 0.0, 0.00922, -0.04147, -0.0, 0.00864, -0.03692, 0.0, 0.00442, -0.00934, 0.00238, 0.00408, 0.01049, 0.00204, 0.00289, 0.01801, -0.0, 0.00408, 0.01049, -0.00204, 0.00442, -0.00934, -0.00238, -0.00408, 0.01049, 0.00204, -0.00442, -0.00934, 0.00238, -0.00442, -0.00934, -0.00238, -0.00408, 0.01049, -0.00204, -0.00864, -0.03692, -0.0, -0.00922, -0.04147, -0.0, -0.00403, -0.04172, -0.00219, 0.00403, -0.04172, 0.00219, 0.00403, -0.04172, -0.00219, -0.00403, -0.04172, 0.00219, -0.00408, 0.01049, 0.00204, -0.00442, -0.00934, 0.00238, -0.00442, -0.00934, -0.00238, -0.00408, 0.01049, -0.00204, -0.00864, -0.03692, -0.0, -0.00922, -0.04147, -0.0, -0.00403, -0.04172, -0.00219, 0.00403, -0.04172, 0.00219, 0.00403, -0.04172, -0.00219, -0.00403, -0.04172, 0.00219, -0.00289, 0.01801, 0.0, 0.00922, -0.04147, 0.0, 0.00864, -0.03692, 0.0, 0.00442, -0.00934, 0.00238, 0.00408, 0.01049, 0.00204, 0.00289, 0.01801, 0.0, 0.00408, 0.01049, -0.00204, 0.00442, -0.00934, -0.00238, 0.0055, -0.02164, -0.0, -0.0055, -0.02164, -0.0, -0.0055, -0.02164, 0.0, 0.0055, -0.02164, 0.0]])}, database_path = 'blue box_scsd_20220719.pkl', maxdist = 1.66, mondrian_limits = [0, 0])
model_objs_dict['blue_box'] = scsd_model('blue_box', array([['0.7383', '3.4705', '-0.0', 'C'], ['-5.1135', '1.3886', '0.0', 'C'], ['-4.9421', '2.8846', '0.0', 'C'], ['-2.8229', '3.3026', '1.1591', 'C'], ['-1.4762', '3.4344', '1.1823', 'C'], ['-0.7383', '3.4705', '0.0', 'C'], ['-1.4762', '3.4344', '-1.1823', 'C'], ['-2.8229', '3.3026', '-1.1591', 'C'], ['1.4762', '3.4344', '1.1823', 'C'], ['2.8229', '3.3026', '1.1591', 'C'], ['2.8229', '3.3026', '-1.1591', 'C'], ['1.4762', '3.4344', '-1.1823', 'C'], ['4.9421', '2.8846', '-0.0', 'C'], ['5.1135', '1.3886', '-0.0', 'C'], ['5.1503', '0.6819', '-1.1896', 'C'], ['5.1503', '-0.6819', '-1.1896', 'C'], ['5.1503', '-0.6819', '1.1896', 'C'], ['5.1503', '0.6819', '1.1896', 'C'], ['-1.4762', '-3.4344', '-1.1823', 'C'], ['-2.8229', '-3.3026', '-1.1591', 'C'], ['-2.8229', '-3.3026', '1.1591', 'C'], ['-1.4762', '-3.4344', '1.1823', 'C'], ['-4.9421', '-2.8846', '0.0', 'C'], ['-5.1135', '-1.3886', '0.0', 'C'], ['-5.1503', '-0.6819', '1.1896', 'C'], ['-5.1503', '0.6819', '1.1896', 'C'], ['-5.1503', '0.6819', '-1.1896', 'C'], ['-5.1503', '-0.6819', '-1.1896', 'C'], ['-0.7383', '-3.4705', '0.0', 'C'], ['5.1135', '-1.3886', '-0.0', 'C'], ['4.9421', '-2.8846', '-0.0', 'C'], ['2.8229', '-3.3026', '-1.1591', 'C'], ['1.4762', '-3.4344', '-1.1823', 'C'], ['0.7383', '-3.4705', '-0.0', 'C'], ['1.4762', '-3.4344', '1.1823', 'C'], ['2.8229', '-3.3026', '1.1591', 'C'], ['-3.4819', '3.1975', '0.0', 'N'], ['3.4819', '3.1975', '-0.0', 'N'], ['-3.4819', '-3.1975', '0.0', 'N'], ['3.4819', '-3.1975', '-0.0', 'N']]), 'D2h' , pca = {"Ag":array([[9e-05, -0.03887, -0.0, -0.02065, 0.00045, -0.0, -0.00777, -0.00069, -0.0, -0.00252, -0.0238, 0.00209, -0.00085, -0.0354, 0.00229, -9e-05, -0.03887, -0.0, -0.00085, -0.0354, -0.00229, -0.00252, -0.0238, -0.00209, 0.00085, -0.0354, 0.00229, 0.00252, -0.0238, 0.00209, 0.00252, -0.0238, -0.00209, 0.00085, -0.0354, -0.00229, 0.00777, -0.00069, 0.0, 0.02065, 0.00045, 0.0, 0.02363, 0.00025, -0.00196, 0.02363, -0.00025, -0.00196, 0.02363, -0.00025, 0.00196, 0.02363, 0.00025, 0.00196, -0.00085, 0.0354, -0.00229, -0.00252, 0.0238, -0.00209, -0.00252, 0.0238, 0.00209, -0.00085, 0.0354, 0.00229, -0.00777, 0.00069, -0.0, -0.02065, -0.00045, -0.0, -0.02363, -0.00025, 0.00196, -0.02363, 0.00025, 0.00196, -0.02363, 0.00025, -0.00196, -0.02363, -0.00025, -0.00196, -9e-05, 0.03887, 0.0, 0.02065, -0.00045, 0.0, 0.00777, 0.00069, 0.0, 0.00252, 0.0238, -0.00209, 0.00085, 0.0354, -0.00229, 9e-05, 0.03887, 0.0, 0.00085, 0.0354, 0.00229, 0.00252, 0.0238, 0.00209, -0.00392, -0.01577, -0.0, 0.00392, -0.01577, -0.0, -0.00392, 0.01577, 0.0, 0.00392, 0.01577, 0.0], [-0.00978, -0.03392, -0.0, 0.03113, -0.0006, 0.0, 0.0069, -0.00432, 0.0, 0.00239, 0.00129, 0.0091, -0.00149, -0.02358, 0.00987, 0.00978, -0.03392, -0.0, -0.00149, -0.02358, -0.00987, 0.00239, 0.00129, -0.0091, 0.00149, -0.02358, 0.00987, -0.00239, 0.00129, 0.0091, -0.00239, 0.00129, -0.0091, 0.00149, -0.02358, -0.00987, -0.0069, -0.00432, -0.0, -0.03113, -0.0006, -0.0, -0.04327, 0.00055, 0.00196, -0.04327, -0.00055, 0.00196, -0.04327, -0.00055, -0.00196, -0.04327, 0.00055, -0.00196, -0.00149, 0.02358, -0.00987, 0.00239, -0.00129, -0.0091, 0.00239, -0.00129, 0.0091, -0.00149, 0.02358, 0.00987, 0.0069, 0.00432, 0.0, 0.03113, 0.0006, 0.0, 0.04327, -0.00055, -0.00196, 0.04327, 0.00055, -0.00196, 0.04327, 0.00055, 0.00196, 0.04327, -0.00055, 0.00196, 0.00978, 0.03392, 0.0, -0.03113, 0.0006, -0.0, -0.0069, 0.00432, -0.0, -0.00239, -0.00129, -0.0091, 0.00149, 0.02358, -0.00987, -0.00978, 0.03392, 0.0, 0.00149, 0.02358, 0.00987, -0.00239, -0.00129, 0.0091, -0.00603, 0.01755, 0.0, 0.00603, 0.01755, 0.0, -0.00603, -0.01755, -0.0, 0.00603, -0.01755, -0.0]]), "B1g":array([[-0.02227, -0.00716, -0.0, -0.01287, 0.01567, -0.0, -0.02188, 0.01648, -0.0, -0.02329, 0.02002, 0.00151, -0.02261, 0.01312, 0.00161, -0.02227, 0.00716, -0.0, -0.02261, 0.01312, -0.00161, -0.02329, 0.02002, -0.00151, -0.02261, -0.01312, -0.00161, -0.02329, -0.02002, -0.00151, -0.02329, -0.02002, 0.00151, -0.02261, -0.01312, 0.00161, -0.02188, -0.01648, 0.0, -0.01287, -0.01567, 0.0, -0.00637, -0.01538, 9e-05, 0.00637, -0.01538, -9e-05, 0.00637, -0.01538, 9e-05, -0.00637, -0.01538, -9e-05, 0.02261, 0.01312, 0.00161, 0.02329, 0.02002, 0.00151, 0.02329, 0.02002, -0.00151, 0.02261, 0.01312, -0.00161, 0.02188, 0.01648, -0.0, 0.01287, 0.01567, -0.0, 0.00637, 0.01538, -9e-05, -0.00637, 0.01538, 9e-05, -0.00637, 0.01538, -9e-05, 0.00637, 0.01538, 9e-05, 0.02227, 0.00716, 0.0, 0.01287, -0.01567, 0.0, 0.02188, -0.01648, 0.0, 0.02329, -0.02002, -0.00151, 0.02261, -0.01312, -0.00161, 0.02227, -0.00716, 0.0, 0.02261, -0.01312, 0.00161, 0.02329, -0.02002, 0.00151, -0.0235, 0.02133, -0.0, -0.0235, -0.02133, -0.0, 0.0235, 0.02133, 0.0, 0.0235, -0.02133, 0.0], [0.0014, -0.00183, -0.0, 0.0048, -0.01129, -0.0, 0.00157, -0.00298, -0.0, 0.00253, 0.00725, 0.0515, 0.00311, 0.00727, 0.05108, 0.0014, 0.00183, -0.0, 0.00311, 0.00727, -0.05108, 0.00253, 0.00725, -0.0515, 0.00311, -0.00727, -0.05108, 0.00253, -0.00725, -0.0515, 0.00253, -0.00725, 0.0515, 0.00311, -0.00727, 0.05108, 0.00157, 0.00298, 0.0, 0.0048, 0.01129, 0.0, 0.00247, 0.008, -0.00162, -0.00247, 0.008, 0.00162, -0.00247, 0.008, -0.00162, 0.00247, 0.008, 0.00162, -0.00311, 0.00727, 0.05108, -0.00253, 0.00725, 0.0515, -0.00253, 0.00725, -0.0515, -0.00311, 0.00727, -0.05108, -0.00157, -0.00298, -0.0, -0.0048, -0.01129, -0.0, -0.00247, -0.008, 0.00162, 0.00247, -0.008, -0.00162, 0.00247, -0.008, 0.00162, -0.00247, -0.008, -0.00162, -0.0014, 0.00183, 0.0, -0.0048, 0.01129, 0.0, -0.00157, 0.00298, 0.0, -0.00253, -0.00725, -0.0515, -0.00311, -0.00727, -0.05108, -0.0014, -0.00183, 0.0, -0.00311, -0.00727, 0.05108, -0.00253, -0.00725, 0.0515, 0.00305, 0.00686, -0.0, 0.00305, -0.00686, -0.0, -0.00305, 0.00686, 0.0, -0.00305, -0.00686, 0.0]]), "B2g":array([[0.0, -0.0, 0.00835, 0.0, 0.0, 0.00552, 0.0, -0.0, 0.01237, -0.00392, 0.04825, 0.0088, -0.0038, 0.05076, 0.00745, -0.0, -0.0, 0.00835, 0.0038, -0.05076, 0.00745, 0.00392, -0.04825, 0.0088, 0.0038, 0.05076, 0.00745, 0.00392, 0.04825, 0.0088, -0.00392, -0.04825, 0.0088, -0.0038, -0.05076, 0.00745, 0.0, -0.0, 0.01237, -0.0, 0.0, 0.00552, 0.00079, 0.00533, 0.00227, -0.00079, 0.00533, -0.00227, 0.00079, -0.00533, -0.00227, -0.00079, -0.00533, 0.00227, -0.0038, -0.05076, -0.00745, -0.00392, -0.04825, -0.0088, 0.00392, 0.04825, -0.0088, 0.0038, 0.05076, -0.00745, 0.0, 0.0, -0.01237, 0.0, -0.0, -0.00552, -0.00079, -0.00533, -0.00227, 0.00079, -0.00533, 0.00227, -0.00079, 0.00533, 0.00227, 0.00079, 0.00533, -0.00227, -0.0, 0.0, -0.00835, -0.0, -0.0, -0.00552, 0.0, 0.0, -0.01237, 0.00392, -0.04825, -0.0088, 0.0038, -0.05076, -0.00745, 0.0, 0.0, -0.00835, -0.0038, 0.05076, -0.00745, -0.00392, 0.04825, -0.0088, -0.0, 0.0, 0.01077, 0.0, 0.0, 0.01077, -0.0, -0.0, -0.01077, 0.0, -0.0, -0.01077], [-0.0, -0.0, 0.03128, -0.0, 0.0, -0.02747, 0.0, -0.0, -0.03753, -0.017, 0.00018, -0.00092, -0.0161, 0.00175, 0.02289, 0.0, -0.0, 0.03128, 0.0161, -0.00175, 0.02289, 0.017, -0.00018, -0.00092, 0.0161, 0.00175, 0.02289, 0.017, 0.00018, -0.00092, -0.017, -0.00018, -0.00092, -0.0161, -0.00175, 0.02289, 0.0, -0.0, -0.03753, 0.0, 0.0, -0.02747, -9e-05, -0.02071, -0.01533, 9e-05, -0.02071, 0.01533, -9e-05, 0.02071, 0.01533, 9e-05, 0.02071, -0.01533, -0.0161, -0.00175, -0.02289, -0.017, -0.00018, 0.00092, 0.017, 0.00018, 0.00092, 0.0161, 0.00175, -0.02289, 0.0, 0.0, 0.03753, -0.0, -0.0, 0.02747, 9e-05, 0.02071, 0.01533, -9e-05, 0.02071, -0.01533, 9e-05, -0.02071, -0.01533, -9e-05, -0.02071, 0.01533, 0.0, 0.0, -0.03128, 0.0, -0.0, 0.02747, 0.0, 0.0, 0.03753, 0.017, -0.00018, 0.00092, 0.0161, -0.00175, -0.02289, -0.0, 0.0, -0.03128, -0.0161, 0.00175, -0.02289, -0.017, 0.00018, 0.00092, -0.0, -0.0, -0.01206, 0.0, -0.0, -0.01206, -0.0, 0.0, 0.01206, 0.0, 0.0, 0.01206]]), "B3g":array([[0.0, 0.0, -0.00231, 0.0, 0.0, -0.00185, -0.0, 0.0, 0.00204, 0.00605, -0.04488, 0.00562, 0.00607, -0.04577, 0.0043, -0.0, 0.0, 0.00231, -0.00607, 0.04577, 0.0043, -0.00605, 0.04488, 0.00562, 0.00607, 0.04577, -0.0043, 0.00605, 0.04488, -0.00562, -0.00605, -0.04488, -0.00562, -0.00607, -0.04577, -0.0043, -0.0, 0.0, -0.00204, -0.0, 0.0, 0.00185, 0.02737, 0.00012, 0.00259, 0.02737, -0.00012, 0.00259, -0.02737, 0.00012, 0.00259, -0.02737, -0.00012, 0.00259, -0.00607, -0.04577, 0.0043, -0.00605, -0.04488, 0.00562, 0.00605, 0.04488, 0.00562, 0.00607, 0.04577, 0.0043, -0.0, -0.0, 0.00204, 0.0, -0.0, -0.00185, -0.02737, -0.00012, -0.00259, -0.02737, 0.00012, -0.00259, 0.02737, -0.00012, -0.00259, 0.02737, 0.00012, -0.00259, -0.0, -0.0, 0.00231, -0.0, -0.0, 0.00185, -0.0, -0.0, -0.00204, -0.00605, 0.04488, -0.00562, -0.00607, 0.04577, -0.0043, 0.0, -0.0, -0.00231, 0.00607, -0.04577, -0.0043, 0.00605, -0.04488, -0.00562, -0.0, -0.0, 0.0048, 0.0, -0.0, -0.0048, -0.0, 0.0, 0.0048, 0.0, 0.0, -0.0048], [0.0, -0.0, -0.00174, 0.0, 0.0, 0.00529, 0.0, 0.0, 0.00968, 0.00077, 0.02053, 0.00427, 0.00036, 0.02271, 0.00103, -0.0, 0.0, 0.00174, -0.00036, -0.02271, 0.00103, -0.00077, -0.02053, 0.00427, 0.00036, -0.02271, -0.00103, 0.00077, -0.02053, -0.00427, -0.00077, 0.02053, -0.00427, -0.00036, 0.02271, -0.00103, 0.0, 0.0, -0.00968, -0.0, 0.0, -0.00529, 0.06997, 0.00073, -0.00395, 0.06997, -0.00073, -0.00395, -0.06997, 0.00073, -0.00395, -0.06997, -0.00073, -0.00395, -0.00036, 0.02271, 0.00103, -0.00077, 0.02053, 0.00427, 0.00077, -0.02053, 0.00427, 0.00036, -0.02271, 0.00103, 0.0, -0.0, 0.00968, 0.0, -0.0, 0.00529, -0.06997, -0.00073, 0.00395, -0.06997, 0.00073, 0.00395, 0.06997, -0.00073, 0.00395, 0.06997, 0.00073, 0.00395, -0.0, -0.0, 0.00174, -0.0, -0.0, -0.00529, 0.0, -0.0, -0.00968, -0.00077, -0.02053, -0.00427, -0.00036, -0.02271, -0.00103, 0.0, -0.0, -0.00174, 0.00036, 0.02271, -0.00103, 0.00077, 0.02053, -0.00427, 0.0, -0.0, 0.00568, -0.0, -0.0, -0.00568, 0.0, 0.0, 0.00568, -0.0, 0.0, -0.00568]]), "Au":array([[0.0, 0.0, 0.00658, 0.0, -0.0, -0.01814, -0.0, 0.0, -0.03787, -0.01187, 0.0201, -0.02502, -0.01209, 0.02017, -0.01318, 0.0, 0.0, -0.00658, 0.01209, -0.02017, -0.01318, 0.01187, -0.0201, -0.02502, -0.01209, -0.02017, 0.01318, -0.01187, -0.0201, 0.02502, 0.01187, 0.0201, 0.02502, 0.01209, 0.02017, 0.01318, 0.0, -0.0, 0.03787, 0.0, 0.0, 0.01814, 0.00015, 0.01524, 0.00888, -0.00015, 0.01524, -0.00888, 0.00015, -0.01524, -0.00888, -0.00015, -0.01524, 0.00888, -0.01209, -0.02017, 0.01318, -0.01187, -0.0201, 0.02502, 0.01187, 0.0201, 0.02502, 0.01209, 0.02017, 0.01318, 0.0, 0.0, 0.03787, 0.0, -0.0, 0.01814, 0.00015, 0.01524, 0.00888, -0.00015, 0.01524, -0.00888, 0.00015, -0.01524, -0.00888, -0.00015, -0.01524, 0.00888, 0.0, 0.0, 0.00658, 0.0, 0.0, -0.01814, 0.0, -0.0, -0.03787, -0.01187, 0.0201, -0.02502, -0.01209, 0.02017, -0.01318, 0.0, 0.0, -0.00658, 0.01209, -0.02017, -0.01318, 0.01187, -0.0201, -0.02502, 0.0, -0.0, -0.02977, 0.0, 0.0, 0.02977, 0.0, -0.0, 0.02977, 0.0, 0.0, -0.02977], [-0.0, -0.0, 0.00214, 0.0, -0.0, -0.01347, 0.0, -0.0, -0.02654, -0.00051, -0.038, -0.0146, -0.0014, -0.04089, -0.00513, -0.0, -0.0, -0.00214, 0.0014, 0.04089, -0.00513, 0.00051, 0.038, -0.0146, -0.0014, 0.04089, 0.00513, -0.00051, 0.038, 0.0146, 0.00051, -0.038, 0.0146, 0.0014, -0.04089, 0.00513, -0.0, 0.0, 0.02654, -0.0, 0.0, 0.01347, 0.00083, 0.01113, 0.00584, -0.00083, 0.01113, -0.00584, 0.00083, -0.01113, -0.00584, -0.00083, -0.01113, 0.00584, -0.0014, 0.04089, 0.00513, -0.00051, 0.038, 0.0146, 0.00051, -0.038, 0.0146, 0.0014, -0.04089, 0.00513, -0.0, -0.0, 0.02654, -0.0, -0.0, 0.01347, 0.00083, 0.01113, 0.00584, -0.00083, 0.01113, -0.00584, 0.00083, -0.01113, -0.00584, -0.00083, -0.01113, 0.00584, -0.0, -0.0, 0.00214, -0.0, 0.0, -0.01347, -0.0, 0.0, -0.02654, -0.00051, -0.038, -0.0146, -0.0014, -0.04089, -0.00513, -0.0, -0.0, -0.00214, 0.0014, 0.04089, -0.00513, 0.00051, 0.038, -0.0146, -0.0, 0.0, -0.01877, -0.0, -0.0, 0.01877, -0.0, 0.0, 0.01877, -0.0, -0.0, -0.01877]]), "B1u":array([[-0.0, 0.0, 0.01082, -0.0, -0.0, -0.01099, -0.0, -0.0, -0.00655, 0.00026, -0.03359, 0.00483, 0.00019, -0.03459, 0.00985, -0.0, -0.0, 0.01082, -0.00019, 0.03459, 0.00985, -0.00026, 0.03359, 0.00483, -0.00019, -0.03459, 0.00985, -0.00026, -0.03359, 0.00483, 0.00026, 0.03359, 0.00483, 0.00019, 0.03459, 0.00985, -0.0, -0.0, -0.00655, -0.0, -0.0, -0.01099, -0.03809, -0.00012, -0.01223, -0.03809, 0.00012, -0.01223, 0.03809, -0.00012, -0.01223, 0.03809, 0.00012, -0.01223, -0.00019, -0.03459, 0.00985, -0.00026, -0.03359, 0.00483, 0.00026, 0.03359, 0.00483, 0.00019, 0.03459, 0.00985, -0.0, -0.0, -0.00655, -0.0, -0.0, -0.01099, -0.03809, -0.00012, -0.01223, -0.03809, 0.00012, -0.01223, 0.03809, -0.00012, -0.01223, 0.03809, 0.00012, -0.01223, -0.0, -0.0, 0.01082, -0.0, -0.0, -0.01099, -0.0, -0.0, -0.00655, 0.00026, -0.03359, 0.00483, 0.00019, -0.03459, 0.00985, -0.0, -0.0, 0.01082, -0.00019, 0.03459, 0.00985, -0.00026, 0.03359, 0.00483, -0.0, -0.0, 0.00181, -0.0, -0.0, 0.00181, -0.0, -0.0, 0.00181, -0.0, -0.0, 0.00181], [0.0, 0.0, -0.00698, -0.0, -0.0, 0.00482, 0.0, 0.0, 0.00498, 0.009, -0.03103, -0.00294, 0.0075, -0.03195, -0.00573, 0.0, 0.0, -0.00698, -0.0075, 0.03195, -0.00573, -0.009, 0.03103, -0.00294, -0.0075, -0.03195, -0.00573, -0.009, -0.03103, -0.00294, 0.009, 0.03103, -0.00294, 0.0075, 0.03195, -0.00573, 0.0, 0.0, 0.00498, 0.0, 0.0, 0.00482, 0.05009, 1e-05, 0.00747, 0.05009, -1e-05, 0.00747, -0.05009, 1e-05, 0.00747, -0.05009, -1e-05, 0.00747, -0.0075, -0.03195, -0.00573, -0.009, -0.03103, -0.00294, 0.009, 0.03103, -0.00294, 0.0075, 0.03195, -0.00573, 0.0, 0.0, 0.00498, 0.0, 0.0, 0.00482, 0.05009, 1e-05, 0.00747, 0.05009, -1e-05, 0.00747, -0.05009, 1e-05, 0.00747, -0.05009, -1e-05, 0.00747, 0.0, 0.0, -0.00698, 0.0, 0.0, 0.00482, 0.0, 0.0, 0.00498, 0.009, -0.03103, -0.00294, 0.0075, -0.03195, -0.00573, 0.0, 0.0, -0.00698, -0.0075, 0.03195, -0.00573, -0.009, 0.03103, -0.00294, 0.0, 0.0, -0.00044, 0.0, 0.0, -0.00044, 0.0, 0.0, -0.00044, 0.0, 0.0, -0.00044]]), "B2u":array([[-0.01187, -0.01366, -0.0, 0.02019, -0.00018, -0.0, -0.00085, 0.00364, 0.0, -0.0116, 0.03158, 0.00813, -0.01193, 0.02517, 0.00876, -0.01187, 0.01366, 0.0, -0.01193, 0.02517, -0.00876, -0.0116, 0.03158, -0.00813, -0.01193, -0.02517, -0.00876, -0.0116, -0.03158, -0.00813, -0.0116, -0.03158, 0.00813, -0.01193, -0.02517, 0.00876, -0.00085, -0.00364, -0.0, 0.02019, 0.00018, 0.0, 0.02563, 0.00098, -0.00067, 0.02563, -0.00098, -0.00067, 0.02563, -0.00098, 0.00067, 0.02563, 0.00098, 0.00067, -0.01193, -0.02517, -0.00876, -0.0116, -0.03158, -0.00813, -0.0116, -0.03158, 0.00813, -0.01193, -0.02517, 0.00876, -0.00085, -0.00364, -0.0, 0.02019, 0.00018, -0.0, 0.02563, 0.00098, -0.00067, 0.02563, -0.00098, -0.00067, 0.02563, -0.00098, 0.00067, 0.02563, 0.00098, 0.00067, -0.01187, -0.01366, -0.0, 0.02019, -0.00018, 0.0, -0.00085, 0.00364, 0.0, -0.0116, 0.03158, 0.00813, -0.01193, 0.02517, 0.00876, -0.01187, 0.01366, 0.0, -0.01193, 0.02517, -0.00876, -0.0116, 0.03158, -0.00813, -0.01168, 0.0266, 0.0, -0.01168, -0.0266, -0.0, -0.01168, -0.0266, -0.0, -0.01168, 0.0266, 0.0], [-0.00055, -0.00516, -0.0, -0.0035, 0.00235, 0.0, 0.00618, 0.00036, 0.0, 0.00112, 0.01064, -0.04028, 0.00044, 0.01211, -0.04436, -0.00055, 0.00516, 0.0, 0.00044, 0.01211, 0.04436, 0.00112, 0.01064, 0.04028, 0.00044, -0.01211, 0.04436, 0.00112, -0.01064, 0.04028, 0.00112, -0.01064, -0.04028, 0.00044, -0.01211, -0.04436, 0.00618, -0.00036, -0.0, -0.0035, -0.00235, -0.0, -0.00139, 0.00124, -0.02567, -0.00139, -0.00124, -0.02567, -0.00139, -0.00124, 0.02567, -0.00139, 0.00124, 0.02567, 0.00044, -0.01211, 0.04436, 0.00112, -0.01064, 0.04028, 0.00112, -0.01064, -0.04028, 0.00044, -0.01211, -0.04436, 0.00618, -0.00036, 0.0, -0.0035, -0.00235, 0.0, -0.00139, 0.00124, -0.02567, -0.00139, -0.00124, -0.02567, -0.00139, -0.00124, 0.02567, -0.00139, 0.00124, 0.02567, -0.00055, -0.00516, -0.0, -0.0035, 0.00235, -0.0, 0.00618, 0.00036, -0.0, 0.00112, 0.01064, -0.04028, 0.00044, 0.01211, -0.04436, -0.00055, 0.00516, 0.0, 0.00044, 0.01211, 0.04436, 0.00112, 0.01064, 0.04028, -0.00247, 0.00719, 0.0, -0.00247, -0.00719, -0.0, -0.00247, -0.00719, -0.0, -0.00247, 0.00719, 0.0]]), "B3u":array([[-0.00237, 0.04022, -0.0, 0.01229, -0.02795, 0.0, 0.01061, -0.02338, 0.0, 0.00524, 0.00518, -0.00801, 0.00473, 0.03123, -0.00964, 0.00237, 0.04022, -0.0, 0.00473, 0.03123, 0.00964, 0.00524, 0.00518, 0.00801, -0.00473, 0.03123, -0.00964, -0.00524, 0.00518, -0.00801, -0.00524, 0.00518, 0.00801, -0.00473, 0.03123, 0.00964, -0.01061, -0.02338, -0.0, -0.01229, -0.02795, -0.0, -0.00588, -0.0266, -0.00204, 0.00588, -0.0266, 0.00204, 0.00588, -0.0266, -0.00204, -0.00588, -0.0266, 0.00204, -0.00473, 0.03123, -0.00964, -0.00524, 0.00518, -0.00801, -0.00524, 0.00518, 0.00801, -0.00473, 0.03123, 0.00964, -0.01061, -0.02338, -0.0, -0.01229, -0.02795, -0.0, -0.00588, -0.0266, -0.00204, 0.00588, -0.0266, 0.00204, 0.00588, -0.0266, -0.00204, -0.00588, -0.0266, 0.00204, -0.00237, 0.04022, 0.0, 0.01229, -0.02795, 0.0, 0.01061, -0.02338, 0.0, 0.00524, 0.00518, -0.00801, 0.00473, 0.03123, -0.00964, 0.00237, 0.04022, 0.0, 0.00473, 0.03123, 0.00964, 0.00524, 0.00518, 0.00801, 0.00705, -0.00851, -0.0, -0.00705, -0.00851, -0.0, -0.00705, -0.00851, 0.0, 0.00705, -0.00851, 0.0], [0.00539, 0.00765, -0.0, -0.00124, -0.01066, 0.0, -0.00815, -0.01156, -0.0, -0.0123, 0.00978, 0.04465, -0.0109, 0.00667, 0.04156, -0.00539, 0.00765, -0.0, -0.0109, 0.00667, -0.04156, -0.0123, 0.00978, -0.04465, 0.0109, 0.00667, 0.04156, 0.0123, 0.00978, 0.04465, 0.0123, 0.00978, -0.04465, 0.0109, 0.00667, -0.04156, 0.00815, -0.01156, 0.0, 0.00124, -0.01066, 0.0, 0.00229, -0.01124, -0.00144, -0.00229, -0.01124, 0.00144, -0.00229, -0.01124, -0.00144, 0.00229, -0.01124, 0.00144, 0.0109, 0.00667, 0.04156, 0.0123, 0.00978, 0.04465, 0.0123, 0.00978, -0.04465, 0.0109, 0.00667, -0.04156, 0.00815, -0.01156, 0.0, 0.00124, -0.01066, 0.0, 0.00229, -0.01124, -0.00144, -0.00229, -0.01124, 0.00144, -0.00229, -0.01124, -0.00144, 0.00229, -0.01124, 0.00144, 0.00539, 0.00765, 0.0, -0.00124, -0.01066, -0.0, -0.00815, -0.01156, -0.0, -0.0123, 0.00978, 0.04465, -0.0109, 0.00667, 0.04156, -0.00539, 0.00765, -0.0, -0.0109, 0.00667, -0.04156, -0.0123, 0.00978, -0.04465, -0.01023, 0.00415, -0.0, 0.01023, 0.00415, 0.0, 0.01023, 0.00415, 0.0, -0.01023, 0.00415, -0.0]])}, database_path = 'blue box_scsd_20220719.pkl', maxdist = 1.66, mondrian_limits = [0, 0])
model_objs_dict['ppc'] = scsd_model('ppc', array([['4.4751', '-2.4579', '3.6033', 'C'], ['5.5547', '-2.8461', '4.3927', 'C'], ['5.2675', '-4.071', '4.9333', 'C'], ['3.9729', '-4.4219', '4.484', 'C'], ['3.178', '-5.5045', '4.8535', 'C'], ['1.8431', '-5.6516', '4.484', 'C'], ['0.8919', '-6.5972', '4.9333', 'C'], ['-0.3125', '-6.2335', '4.3927', 'C'], ['-0.109', '-5.1045', '3.6033', 'C'], ['-1.1238', '-4.3732', '2.8147', 'C'], ['-2.2446', '-3.8881', '3.4546', 'C'], ['-3.2251', '-3.1598', '2.8147', 'C'], ['-4.3659', '-2.6466', '3.6032', 'C'], ['-5.242', '-3.3874', '4.3924', 'C'], ['-6.1592', '-2.5262', '4.9329', 'C'], ['-5.8158', '-1.2297', '4.4838', 'C'], ['-6.3559', '-0.0001', '4.8533', 'C'], ['-5.8159', '1.2296', '4.4839', 'C'], ['-6.1593', '2.5261', '4.9331', 'C'], ['-5.2422', '3.3874', '4.3927', 'C'], ['-4.3661', '2.6468', '3.6033', 'C'], ['-3.2254', '3.16', '2.8148', 'C'], ['-2.2448', '3.8882', '3.4547', 'C'], ['-1.124', '4.3733', '2.8148', 'C'], ['-0.1091', '5.1046', '3.6033', 'C'], ['-0.3125', '6.2336', '4.3927', 'C'], ['0.892', '6.5972', '4.9331', 'C'], ['1.8431', '5.6515', '4.4839', 'C'], ['3.178', '5.5043', '4.8533', 'C'], ['3.9728', '4.4218', '4.4838', 'C'], ['5.2674', '4.0709', '4.9329', 'C'], ['5.5546', '2.846', '4.3924', 'C'], ['4.475', '2.4577', '3.6032', 'C'], ['4.349', '1.2131', '2.8147', 'C'], ['4.4895', '-0.0001', '3.4546', 'C'], ['4.3491', '-1.2134', '2.8147', 'C'], ['-1.0001', '-4.2032', '1.4317', 'C'], ['-2.049', '-3.5493', '0.7007', 'C'], ['-3.1397', '-2.9677', '1.4317', 'C'], ['-4.1682', '-2.3291', '0.6744', 'C'], ['-4.1682', '-2.3291', '-0.6744', 'C'], ['-3.1397', '-2.9677', '-1.4317', 'C'], ['-2.049', '-3.5493', '-0.7007', 'C'], ['-1.0001', '-4.2032', '-1.4317', 'C'], ['0.0672', '-4.7747', '-0.6744', 'C'], ['0.0672', '-4.7747', '0.6744', 'C'], ['-3.1401', '2.9681', '1.4317', 'C'], ['-4.1687', '2.3297', '0.6744', 'C'], ['-4.1687', '2.3297', '-0.6744', 'C'], ['-3.1401', '2.9681', '-1.4317', 'C'], ['-2.0494', '3.5496', '-0.7007', 'C'], ['-1.0004', '4.2035', '-1.4317', 'C'], ['0.0668', '4.7751', '-0.6744', 'C'], ['0.0668', '4.7751', '0.6744', 'C'], ['-1.0004', '4.2035', '1.4317', 'C'], ['-2.0494', '3.5496', '0.7007', 'C'], ['4.14', '1.2352', '1.4317', 'C'], ['4.1012', '2.4452', '0.6744', 'C'], ['4.1012', '2.4452', '-0.6744', 'C'], ['4.14', '1.2352', '-1.4317', 'C'], ['4.0983', '-0.0002', '-0.7007', 'C'], ['4.1401', '-1.2355', '-1.4317', 'C'], ['4.1014', '-2.4455', '-0.6744', 'C'], ['4.1014', '-2.4455', '0.6744', 'C'], ['4.1401', '-1.2355', '1.4317', 'C'], ['4.0983', '-0.0002', '0.7007', 'C'], ['4.4751', '-2.4579', '-3.6033', 'C'], ['5.5547', '-2.8461', '-4.3927', 'C'], ['5.2675', '-4.071', '-4.9333', 'C'], ['3.9729', '-4.4219', '-4.484', 'C'], ['3.178', '-5.5045', '-4.8535', 'C'], ['1.8431', '-5.6516', '-4.484', 'C'], ['0.8919', '-6.5972', '-4.9333', 'C'], ['-0.3125', '-6.2335', '-4.3927', 'C'], ['-0.109', '-5.1045', '-3.6033', 'C'], ['-1.1238', '-4.3732', '-2.8147', 'C'], ['-2.2446', '-3.8881', '-3.4546', 'C'], ['-3.2251', '-3.1598', '-2.8147', 'C'], ['-4.3659', '-2.6466', '-3.6032', 'C'], ['-5.242', '-3.3874', '-4.3924', 'C'], ['-6.1592', '-2.5262', '-4.9329', 'C'], ['-5.8158', '-1.2297', '-4.4838', 'C'], ['-6.3559', '-0.0001', '-4.8533', 'C'], ['-5.8159', '1.2296', '-4.4839', 'C'], ['-6.1593', '2.5261', '-4.9331', 'C'], ['-5.2422', '3.3874', '-4.3927', 'C'], ['-4.3661', '2.6468', '-3.6033', 'C'], ['-3.2254', '3.16', '-2.8148', 'C'], ['-2.2448', '3.8882', '-3.4547', 'C'], ['-1.124', '4.3733', '-2.8148', 'C'], ['-0.1091', '5.1046', '-3.6033', 'C'], ['1.8431', '5.6515', '-4.4839', 'C'], ['3.178', '5.5043', '-4.8533', 'C'], ['3.9728', '4.4218', '-4.4838', 'C'], ['5.2674', '4.0709', '-4.9329', 'C'], ['5.5546', '2.846', '-4.3924', 'C'], ['4.475', '2.4577', '-3.6032', 'C'], ['4.349', '1.2131', '-2.8147', 'C'], ['4.4895', '-0.0001', '-3.4546', 'C'], ['4.3491', '-1.2134', '-2.8147', 'C'], ['-0.3125', '6.2336', '-4.3927', 'C'], ['0.892', '6.5972', '-4.9331', 'C'], ['3.5224', '-3.3978', '3.6557', 'N'], ['1.1814', '-4.7494', '3.6557', 'N'], ['-4.7034', '-1.3516', '3.6557', 'N'], ['-4.7037', '1.3517', '3.6557', 'N'], ['1.1812', '4.7493', '3.6557', 'N'], ['3.5222', '3.3975', '3.6557', 'N'], ['3.5224', '-3.3978', '-3.6557', 'N'], ['1.1814', '-4.7494', '-3.6557', 'N'], ['-4.7034', '-1.3516', '-3.6557', 'N'], ['-4.7037', '1.3517', '-3.6557', 'N'], ['1.1812', '4.7493', '-3.6557', 'N'], ['3.5222', '3.3975', '-3.6557', 'N']]), 'D3h' , maxdist = 1.75, mondrian_limits = [0, 0])
model_objs_dict['teropyrene'] = scsd_model('teropyrene', array([['7.9732', '0.0', '0.0', 'C'], ['7.2777', '-1.2267', '0.0', 'C'], ['5.7733', '-1.2523', '0.0', 'C'], ['5.0412', '0.0', '0.0', 'C'], ['5.7733', '1.2523', '0.0', 'C'], ['7.2777', '1.2267', '0.0', 'C'], ['5.0512', '-2.4578', '0.0', 'C'], ['3.6606', '-2.4666', '0.0', 'C'], ['2.8947', '-1.2632', '0.0', 'C'], ['3.6175', '0.0', '0.0', 'C'], ['2.8947', '1.2632', '0.0', 'C'], ['3.6606', '2.4666', '0.0', 'C'], ['5.0512', '2.4578', '0.0', 'C'], ['1.4495', '-1.2608', '0.0', 'C'], ['0.7166', '0.0', '0.0', 'C'], ['1.4495', '1.2608', '0.0', 'C'], ['0.6928', '-2.4598', '-0.0', 'C'], ['-0.6928', '-2.4598', '-0.0', 'C'], ['-1.4495', '-1.2608', '-0.0', 'C'], ['-0.7166', '0.0', '-0.0', 'C'], ['-1.4495', '1.2608', '-0.0', 'C'], ['-0.6928', '2.4598', '0.0', 'C'], ['0.6928', '2.4598', '0.0', 'C'], ['-2.8947', '-1.2632', '-0.0', 'C'], ['-3.6175', '0.0', '-0.0', 'C'], ['-2.8947', '1.2632', '-0.0', 'C'], ['-3.6606', '-2.4666', '-0.0', 'C'], ['-5.0512', '-2.4578', '-0.0', 'C'], ['-5.7733', '-1.2523', '-0.0', 'C'], ['-5.0412', '0.0', '-0.0', 'C'], ['-5.7733', '1.2523', '-0.0', 'C'], ['-5.0512', '2.4578', '-0.0', 'C'], ['-3.6606', '2.4666', '-0.0', 'C'], ['-7.2777', '-1.2267', '-0.0', 'C'], ['-7.9732', '0.0', '-0.0', 'C'], ['-7.2777', '1.2267', '-0.0', 'C']]), 'D2h' , maxdist = 1.75, mondrian_limits = [0.0, 0.0])
model_objs_dict['amethyrin'] = scsd_model('amethyrin', array([['4.2337', '-1.2724', '-0.0', 'C'], ['4.8437', '-2.6047', '-0.0', 'C'], ['3.7884', '-3.5854', '-0.0', 'C'], ['2.5047', '-2.8764', '-0.0', 'C'], ['1.2031', '-3.3907', '-0.0', 'C'], ['0.7186', '-4.7647', '-0.0', 'C'], ['-0.7186', '-4.7647', '0.0', 'C'], ['-1.2031', '-3.3907', '0.0', 'C'], ['-2.5047', '-2.8764', '0.0', 'C'], ['-3.7884', '-3.5854', '0.0', 'C'], ['-4.8437', '-2.6047', '0.0', 'C'], ['-4.2337', '-1.2724', '0.0', 'C'], ['-4.872', '0.0', '0.0', 'C'], ['-4.2337', '1.2724', '0.0', 'C'], ['-4.8437', '2.6047', '0.0', 'C'], ['-3.7884', '3.5854', '0.0', 'C'], ['-2.5047', '2.8764', '0.0', 'C'], ['-1.2031', '3.3907', '0.0', 'C'], ['-0.7186', '4.7647', '0.0', 'C'], ['0.7186', '4.7647', '-0.0', 'C'], ['1.2031', '3.3907', '-0.0', 'C'], ['2.5047', '2.8764', '-0.0', 'C'], ['3.7884', '3.5854', '-0.0', 'C'], ['4.8437', '2.6047', '-0.0', 'C'], ['4.2337', '1.2724', '-0.0', 'C'], ['4.872', '-0.0', '-0.0', 'C'], ['2.815', '-1.4813', '-0.0', 'N'], ['-0.0', '-2.5334', '0.0', 'N'], ['-2.815', '-1.4813', '0.0', 'N'], ['-2.815', '1.4813', '0.0', 'N'], ['0.0', '2.5334', '0.0', 'N'], ['2.815', '1.4813', '-0.0', 'N']]), 'D2h' , maxdist = 1.75, mondrian_limits = [-1.0, -1.0])#TOZGIB relaxed.pdb dipyriamethyrin 20220719
model_objs_dict['dipyriamethyrin'] = scsd_model('dipyriamethyrin', array([['1.2888', '-4.0282', '-0.0', 'C'], ['2.5587', '-4.7293', '-0.0', 'C'], ['3.6302', '-3.7947', '-0.0', 'C'], ['3.0995', '-2.4411', '0.0', 'C'], ['3.8105', '-1.2243', '0.0', 'C'], ['5.2599', '-1.2226', '0.0', 'C'], ['5.9608', '0.0', '0.0', 'C'], ['5.2599', '1.2226', '0.0', 'C'], ['3.8105', '1.2243', '0.0', 'C'], ['3.0995', '2.4411', '0.0', 'C'], ['3.6302', '3.7947', '0.0', 'C'], ['2.5587', '4.7293', '0.0', 'C'], ['1.2888', '4.0282', '0.0', 'C'], ['0.0', '4.611', '0.0', 'C'], ['-1.2888', '4.0282', '0.0', 'C'], ['-2.5587', '4.7293', '0.0', 'C'], ['-3.6302', '3.7947', '0.0', 'C'], ['-3.0995', '2.4411', '-0.0', 'C'], ['-3.8105', '1.2243', '-0.0', 'C'], ['-3.8105', '-1.2243', '-0.0', 'C'], ['-3.0995', '-2.4411', '-0.0', 'C'], ['-3.6302', '-3.7947', '-0.0', 'C'], ['-2.5587', '-4.7293', '-0.0', 'C'], ['-1.2888', '-4.0282', '-0.0', 'C'], ['0.0', '-4.611', '-0.0', 'C'], ['-5.2599', '1.2226', '-0.0', 'C'], ['-5.9608', '0.0', '-0.0', 'C'], ['-5.2599', '-1.2226', '-0.0', 'C'], ['1.6191', '-2.5806', '-0.0', 'N'], ['3.2078', '0.0', '0.0', 'N'], ['1.6191', '2.5806', '0.0', 'N'], ['-1.6191', '2.5806', '0.0', 'N'], ['-3.2078', '0.0', '-0.0', 'N'], ['-1.6191', '-2.5806', '-0.0', 'N']]), 'D2h' , maxdist = 1.65, mondrian_limits = [0, 0])#JEZXOE relaxed.pdb tetrabenzopentacene 20220719 #TOZGIB relaxed.pdb dipyriamethyrin 20220719
model_objs_dict['siamesetwin'] = scsd_model('siamesetwin', array([['-1.1316', '2.8374', '-0.0', 'C'], ['-2.4786', '3.3417', '-0.0', 'C'], ['-3.7178', '2.7301', '-0.0', 'C'], ['1.1316', '2.8374', '-0.0', 'C'], ['2.4786', '3.3417', '-0.0', 'C'], ['3.7178', '2.7301', '-0.0', 'C'], ['4.9782', '3.4774', '-0.0', 'C'], ['5.3724', '1.2388', '-0.0', 'C'], ['-4.9782', '3.4774', '-0.0', 'C'], ['-0.0', '3.6805', '-0.0', 'C'], ['5.9955', '2.561', '-0.0', 'C'], ['-5.9955', '2.561', '-0.0', 'C'], ['-5.3724', '1.2388', '-0.0', 'C'], ['-6.0399', '-0.0', '0.0', 'C'], ['1.1316', '-2.8374', '0.0', 'C'], ['2.4786', '-3.3417', '0.0', 'C'], ['0.0', '-3.6805', '0.0', 'C'], ['-1.1316', '-2.8374', '0.0', 'C'], ['-2.4786', '-3.3417', '0.0', 'C'], ['3.7178', '-2.7301', '0.0', 'C'], ['4.9782', '-3.4774', '0.0', 'C'], ['5.3724', '-1.2388', '0.0', 'C'], ['6.0399', '0.0', '0.0', 'C'], ['5.9955', '-2.561', '0.0', 'C'], ['-3.7178', '-2.7301', '0.0', 'C'], ['-4.9782', '-3.4774', '0.0', 'C'], ['-5.9955', '-2.561', '0.0', 'C'], ['-5.3724', '-1.2388', '0.0', 'C'], ['-0.6746', '1.54', '-0.0', 'N'], ['-0.6746', '-1.54', '0.0', 'N'], ['0.6746', '-1.54', '0.0', 'N'], ['0.6746', '1.54', '-0.0', 'N'], ['4.0109', '1.3779', '-0.0', 'N'], ['-4.0109', '1.3779', '-0.0', 'N'], ['4.0109', '-1.3779', '0.0', 'N'], ['-4.0109', '-1.3779', '0.0', 'N']]), 'D2h' , maxdist = 1.7, mondrian_limits = [0.0, 0.0])
model_objs_dict['tetrabenzopentacene'] = scsd_model('tetrabenzopentacene', array([['2.5133', '-1.4313', '0.0', 'C'], ['3.8461', '-0.7196', '0.0', 'C'], ['3.8461', '0.7196', '-0.0', 'C'], ['5.0981', '1.4281', '-0.0', 'C'], ['5.1618', '2.9214', '-0.0', 'C'], ['6.3783', '3.594', '-0.0', 'C'], ['7.5831', '2.8901', '-0.0', 'C'], ['5.0981', '-1.4281', '0.0', 'C'], ['5.1618', '-2.9214', '0.0', 'C'], ['6.3783', '-3.594', '0.0', 'C'], ['7.5831', '-2.8901', '0.0', 'C'], ['7.5973', '-1.5006', '0.0', 'C'], ['6.3271', '-0.7064', '0.0', 'C'], ['6.3271', '0.7064', '-0.0', 'C'], ['7.5973', '1.5006', '-0.0', 'C'], ['-7.5831', '-2.8901', '0.0', 'C'], ['-7.5973', '-1.5006', '0.0', 'C'], ['-6.3271', '-0.7064', '0.0', 'C'], ['-5.0981', '-1.4281', '0.0', 'C'], ['-3.8461', '-0.7196', '0.0', 'C'], ['-2.5133', '-1.4313', '0.0', 'C'], ['-1.2634', '-0.7367', '0.0', 'C'], ['0.0', '-1.4298', '0.0', 'C'], ['-7.5831', '2.8901', '-0.0', 'C'], ['-7.5973', '1.5006', '-0.0', 'C'], ['-6.3271', '0.7064', '-0.0', 'C'], ['-5.0981', '1.4281', '-0.0', 'C'], ['-3.8461', '0.7196', '-0.0', 'C'], ['-1.2634', '0.7367', '-0.0', 'C'], ['-2.5133', '1.4313', '-0.0', 'C'], ['-6.3783', '-3.594', '0.0', 'C'], ['-5.1618', '-2.9214', '0.0', 'C'], ['-5.1618', '2.9214', '-0.0', 'C'], ['-6.3783', '3.594', '-0.0', 'C'], ['1.2634', '-0.7367', '0.0', 'C'], ['2.5133', '1.4313', '-0.0', 'C'], ['1.2634', '0.7367', '-0.0', 'C'], ['-0.0', '1.4298', '-0.0', 'C']]), 'D2h' , maxdist = 1.65, mondrian_limits = [-0.5, -0.5])
model_objs_dict['pentacene'] = scsd_model('pentacene', array([['2.5133', '-1.4313', '0.0', 'C'], ['3.8461', '-0.7196', '0.0', 'C'], ['3.8461', '0.7196', '-0.0', 'C'], ['5.0981', '1.4281', '-0.0', 'C'], ['5.0981', '-1.4281', '0.0', 'C'], ['6.3271', '-0.7064', '0.0', 'C'], ['6.3271', '0.7064', '-0.0', 'C'], ['-6.3271', '-0.7064', '0.0', 'C'], ['-5.0981', '-1.4281', '0.0', 'C'], ['-3.8461', '-0.7196', '0.0', 'C'], ['-2.5133', '-1.4313', '0.0', 'C'], ['-1.2634', '-0.7367', '0.0', 'C'], ['0.0', '-1.4298', '0.0', 'C'], ['-6.3271', '0.7064', '-0.0', 'C'], ['-5.0981', '1.4281', '-0.0', 'C'], ['-3.8461', '0.7196', '-0.0', 'C'], ['-1.2634', '0.7367', '-0.0', 'C'], ['-2.5133', '1.4313', '-0.0', 'C'], ['1.2634', '-0.7367', '0.0', 'C'], ['2.5133', '1.4313', '-0.0', 'C'], ['1.2634', '0.7367', '-0.0', 'C'], ['-0.0', '1.4298', '-0.0', 'C']]), 'D2h' , maxdist = 1.65, mondrian_limits = [-1, -1])#VAFRUP10.pdb blue box 20220719

model_objs_dict['carbamazepine'] = scsd_model('carbamazepine', array([[-1.2827, -0.0, 0.6603], [-2.0885, -0.0, 1.539], [-3.3991, -0.0, 1.1703], [-3.6161, 0.0, -0.1869], [-2.8212, 0.0, -1.2149], [-1.6037, 0.0, -0.672], [-0.6673, 0.0, -1.8154], [0.6673, 0.0, -1.8154], [1.6037, 0.0, -0.672], [2.8212, 0.0, -1.2149], [3.6161, 0.0, -0.1869], [3.3991, -0.0, 1.1703], [2.0885, -0.0, 1.539], [1.2827, -0.0, 0.6603], [-0.0, -0.0, 1.0392]]), 'C2v' , pca = {"A1":array([[0.01682, -0.0, -0.02682, 0.08356, -0.0, 0.00525, 0.13465, -0.0, 0.0406, 0.13102, -0.0, 0.05027, 0.06955, -0.0, 0.0196, 0.01828, 0.0, -0.01451, -0.01298, 0.0, -0.02407, 0.01298, 0.0, -0.02407, -0.01828, 0.0, -0.01451, -0.06955, -0.0, 0.0196, -0.13102, -0.0, 0.05027, -0.13465, -0.0, 0.0406, -0.08356, -0.0, 0.00525, -0.01682, 0.0, -0.02682, 0.0, 0.0, -0.10064], [-0.02736, 0.0, -0.04583, 0.05668, 0.0, 0.02609, 0.00939, -0.0, 0.06815, -0.10201, -0.0, 0.0824, -0.11245, -0.0, 0.00864, -0.04333, 0.0, -0.04198, -0.0118, 0.0, -0.03368, 0.0118, 0.0, -0.03368, 0.04333, 0.0, -0.04198, 0.11245, -0.0, 0.00864, 0.10201, -0.0, 0.0824, -0.00939, -0.0, 0.06815, -0.05668, -0.0, 0.02609, 0.02736, 0.0, -0.04583, 0.0, 0.0, -0.1276]]), "A2":array([[0.0, 0.03647, 0.0, 0.0, 0.1902, -0.0, 0.0, 0.145, -0.0, -0.0, -0.06214, 0.0, 0.0, -0.17717, -0.0, -0.0, -0.09865, -0.0, 0.0, -0.08115, -0.0, 0.0, 0.08115, -0.0, -0.0, 0.09865, -0.0, -0.0, 0.17717, 0.0, -0.0, 0.06214, -0.0, -0.0, -0.145, -0.0, -0.0, -0.1902, 0.0, -0.0, -0.03647, -0.0, 0.0, 0.0, -0.0], [0.0, -0.17422, -0.0, 0.0, -0.11156, 0.0, -0.0, 0.0679, 0.0, -0.0, 0.16534, -0.0, -0.0, 0.00984, -0.0, 0.0, -0.19609, 0.0, 0.0, -0.12815, 0.0, 0.0, 0.12815, 0.0, 0.0, 0.19609, 0.0, 0.0, -0.00984, 0.0, 0.0, -0.16534, 0.0, 0.0, -0.0679, 0.0, 0.0, 0.11156, -0.0, 0.0, 0.17422, 0.0, 0.0, -0.0, 0.0]]), "B1":array([[0.01453, 0.0, -0.02812, 0.14713, -0.0, -0.007, 0.1041, -0.0, -0.04601, -0.04752, 0.0, -0.05199, -0.10695, 0.0, -0.05327, -0.07491, 0.0, -0.02648, -0.0311, -0.0, -0.01524, -0.0311, -0.0, 0.01524, -0.07491, 0.0, 0.02648, -0.10695, 0.0, 0.05327, -0.04752, 0.0, 0.05199, 0.1041, 0.0, 0.04601, 0.14713, -0.0, 0.007, 0.01453, -0.0, 0.02812, -0.01055, 0.0, 0.0], [-0.0508, 0.0, 0.07053, -0.05005, -0.0, 0.02166, 0.06795, -0.0, -0.023, 0.10366, -0.0, -0.04162, 0.04487, 0.0, -0.05643, -0.08155, 0.0, 0.04346, -0.08705, -0.0, 0.02926, -0.08705, -0.0, -0.02926, -0.08155, 0.0, -0.04346, 0.04487, 0.0, 0.05643, 0.10366, -0.0, 0.04162, 0.06795, 0.0, 0.023, -0.05005, 0.0, -0.02166, -0.0508, 0.0, -0.07053, 0.10594, -0.0, 0.0]]), "B2":array([[-0.0, 0.09585, -0.0, 0.0, 0.01577, -0.0, 0.0, -0.11769, 0.0, 0.0, -0.17077, -0.0, 0.0, -0.09343, 0.0, 0.0, 0.03893, 0.0, 0.0, 0.10841, 0.0, 0.0, 0.10841, 0.0, 0.0, 0.03893, 0.0, 0.0, -0.09343, 0.0, 0.0, -0.17077, 0.0, 0.0, -0.11769, 0.0, 0.0, 0.01577, -0.0, 0.0, 0.09585, 0.0, -0.0, 0.24584, 0.0], [-0.0, 0.03398, -0.0, -0.0, -0.10319, -0.0, 0.0, -0.10944, -0.0, -0.0, 0.06422, -0.0, -0.0, 0.13203, -0.0, -0.0, 0.07061, 0.0, -0.0, -0.20177, -0.0, -0.0, -0.20177, -0.0, -0.0, 0.07061, 0.0, -0.0, 0.13203, -0.0, -0.0, 0.06422, -0.0, -0.0, -0.10944, -0.0, -0.0, -0.10319, -0.0, -0.0, 0.03398, -0.0, -0.0, 0.22712, -0.0]])}, database_path = 'carbamazepine_scsd_20220713.pkl', maxdist = 1.75)
model_objs_dict['dibenzazepine'] = scsd_model('dibenzazepine', array([[-1.2827, -0.0, 0.6603], [-2.0885, -0.0, 1.539], [-3.3991, -0.0, 1.1703], [-3.6161, 0.0, -0.1869], [-2.8212, 0.0, -1.2149], [-1.6037, 0.0, -0.672], [-0.6673, 0.0, -1.8154], [0.6673, 0.0, -1.8154], [1.6037, 0.0, -0.672], [2.8212, 0.0, -1.2149], [3.6161, 0.0, -0.1869], [3.3991, -0.0, 1.1703], [2.0885, -0.0, 1.539], [1.2827, -0.0, 0.6603], [-0.0, -0.0, 1.0392]]), 'C2v' , pca = {"A1":array([[0.01682, -0.0, -0.02682, 0.08356, -0.0, 0.00525, 0.13465, -0.0, 0.0406, 0.13102, -0.0, 0.05027, 0.06955, -0.0, 0.0196, 0.01828, 0.0, -0.01451, -0.01298, 0.0, -0.02407, 0.01298, 0.0, -0.02407, -0.01828, 0.0, -0.01451, -0.06955, -0.0, 0.0196, -0.13102, -0.0, 0.05027, -0.13465, -0.0, 0.0406, -0.08356, -0.0, 0.00525, -0.01682, 0.0, -0.02682, 0.0, 0.0, -0.10064], [-0.02736, 0.0, -0.04583, 0.05668, 0.0, 0.02609, 0.00939, -0.0, 0.06815, -0.10201, -0.0, 0.0824, -0.11245, -0.0, 0.00864, -0.04333, 0.0, -0.04198, -0.0118, 0.0, -0.03368, 0.0118, 0.0, -0.03368, 0.04333, 0.0, -0.04198, 0.11245, -0.0, 0.00864, 0.10201, -0.0, 0.0824, -0.00939, -0.0, 0.06815, -0.05668, -0.0, 0.02609, 0.02736, 0.0, -0.04583, 0.0, 0.0, -0.1276]]), "A2":array([[0.0, 0.03647, 0.0, 0.0, 0.1902, -0.0, 0.0, 0.145, -0.0, -0.0, -0.06214, 0.0, 0.0, -0.17717, -0.0, -0.0, -0.09865, -0.0, 0.0, -0.08115, -0.0, 0.0, 0.08115, -0.0, -0.0, 0.09865, -0.0, -0.0, 0.17717, 0.0, -0.0, 0.06214, -0.0, -0.0, -0.145, -0.0, -0.0, -0.1902, 0.0, -0.0, -0.03647, -0.0, 0.0, 0.0, -0.0], [0.0, -0.17422, -0.0, 0.0, -0.11156, 0.0, -0.0, 0.0679, 0.0, -0.0, 0.16534, -0.0, -0.0, 0.00984, -0.0, 0.0, -0.19609, 0.0, 0.0, -0.12815, 0.0, 0.0, 0.12815, 0.0, 0.0, 0.19609, 0.0, 0.0, -0.00984, 0.0, 0.0, -0.16534, 0.0, 0.0, -0.0679, 0.0, 0.0, 0.11156, -0.0, 0.0, 0.17422, 0.0, 0.0, -0.0, 0.0]]), "B1":array([[0.01453, 0.0, -0.02812, 0.14713, -0.0, -0.007, 0.1041, -0.0, -0.04601, -0.04752, 0.0, -0.05199, -0.10695, 0.0, -0.05327, -0.07491, 0.0, -0.02648, -0.0311, -0.0, -0.01524, -0.0311, -0.0, 0.01524, -0.07491, 0.0, 0.02648, -0.10695, 0.0, 0.05327, -0.04752, 0.0, 0.05199, 0.1041, 0.0, 0.04601, 0.14713, -0.0, 0.007, 0.01453, -0.0, 0.02812, -0.01055, 0.0, 0.0], [-0.0508, 0.0, 0.07053, -0.05005, -0.0, 0.02166, 0.06795, -0.0, -0.023, 0.10366, -0.0, -0.04162, 0.04487, 0.0, -0.05643, -0.08155, 0.0, 0.04346, -0.08705, -0.0, 0.02926, -0.08705, -0.0, -0.02926, -0.08155, 0.0, -0.04346, 0.04487, 0.0, 0.05643, 0.10366, -0.0, 0.04162, 0.06795, 0.0, 0.023, -0.05005, 0.0, -0.02166, -0.0508, 0.0, -0.07053, 0.10594, -0.0, 0.0]]), "B2":array([[-0.0, 0.09585, -0.0, 0.0, 0.01577, -0.0, 0.0, -0.11769, 0.0, 0.0, -0.17077, -0.0, 0.0, -0.09343, 0.0, 0.0, 0.03893, 0.0, 0.0, 0.10841, 0.0, 0.0, 0.10841, 0.0, 0.0, 0.03893, 0.0, 0.0, -0.09343, 0.0, 0.0, -0.17077, 0.0, 0.0, -0.11769, 0.0, 0.0, 0.01577, -0.0, 0.0, 0.09585, 0.0, -0.0, 0.24584, 0.0], [-0.0, 0.03398, -0.0, -0.0, -0.10319, -0.0, 0.0, -0.10944, -0.0, -0.0, 0.06422, -0.0, -0.0, 0.13203, -0.0, -0.0, 0.07061, 0.0, -0.0, -0.20177, -0.0, -0.0, -0.20177, -0.0, -0.0, 0.07061, 0.0, -0.0, 0.13203, -0.0, -0.0, 0.06422, -0.0, -0.0, -0.10944, -0.0, -0.0, -0.10319, -0.0, -0.0, 0.03398, -0.0, -0.0, 0.22712, -0.0]])}, database_path = 'carbamazepine_scsd_20220713.pkl', maxdist = 1.75)
model_objs_dict['xanthene'] = scsd_model('xanthene', array([['1.198', '-0.0', '0.6299', 'C'], ['2.342', '-0.0', '1.4048', 'C'], ['3.57', '-0.0', '0.7913', 'C'], ['3.661', '0.0', '-0.5981', 'C'], ['2.5085', '0.0', '-1.3482', 'C'], ['1.248', '0.0', '-0.7583', 'C'], ['0.0', '0.0', '-1.5722', 'C'], ['-1.198', '-0.0', '0.6299', 'C'], ['-2.342', '-0.0', '1.4048', 'C'], ['-3.57', '-0.0', '0.7913', 'C'], ['-3.661', '0.0', '-0.5981', 'C'], ['-2.5085', '0.0', '-1.3482', 'C'], ['-1.248', '0.0', '-0.7583', 'C'], ['-0.0', '-0.0', '1.3294', 'O']]), 'C2v' , pca = {"A1":array([[-0.01513, 0.0, -0.02023, -0.09259, -0.0, 0.01652, -0.13258, -0.0, 0.02379, -0.12326, -0.0, 0.02998, -0.06395, 0.0, -0.00291, -0.01185, 0.0, -0.02125, -0.0, -0.0, 0.01887, 0.01513, 0.0, -0.02023, 0.09259, -0.0, 0.01652, 0.13258, -0.0, 0.02379, 0.12326, -0.0, 0.02998, 0.06395, 0.0, -0.00291, 0.01185, 0.0, -0.02125, 0.0, 0.0, -0.0707], [0.00812, 0.0, 0.05484, 0.05358, 0.0, -0.00487, 0.00466, 0.0, -0.08036, -0.06011, 0.0, -0.07284, -0.07626, 0.0, -0.0276, -0.02283, -0.0, 0.04194, -0.0, -0.0, 0.12338, -0.00812, -0.0, 0.05484, -0.05358, 0.0, -0.00487, -0.00466, 0.0, -0.08036, 0.06011, 0.0, -0.07284, 0.07626, 0.0, -0.0276, 0.02283, -0.0, 0.04194, -0.0, -0.0, 0.0544]]), "A2":array([[-0.0, 0.03091, -0.0, -0.0, 0.14336, -0.0, -0.0, 0.07882, -0.0, -0.0, -0.08058, -0.0, -0.0, -0.12025, 0.0, -0.0, -0.04607, -0.0, 0.0, -0.0, -0.0, -0.0, -0.03091, 0.0, -0.0, -0.14336, 0.0, -0.0, -0.07882, -0.0, -0.0, 0.08058, 0.0, -0.0, 0.12025, -0.0, -0.0, 0.04607, 0.0, 0.0, 0.0, -0.0], [-0.0, 0.09045, 0.0, 0.0, 0.07454, -0.0, 0.0, -0.08497, 0.0, 0.0, -0.07659, 0.0, 0.0, 0.06694, 0.0, 0.0, 0.1065, 0.0, 0.0, -0.0, 0.0, 0.0, -0.09045, -0.0, 0.0, -0.07454, 0.0, 0.0, 0.08497, 0.0, 0.0, 0.07659, -0.0, 0.0, -0.06694, -0.0, 0.0, -0.1065, -0.0, 0.0, -0.0, 0.0]]), "B1":array([[0.03645, -0.0, 0.02791, 0.11125, -0.0, 0.0337, 0.06061, -0.0, 0.0164, -0.07898, -0.0, 0.02706, -0.10291, 0.0, 0.05115, -0.03621, -0.0, 0.03284, 0.00992, -0.0, 0.0, 0.03645, -0.0, -0.02791, 0.11125, 0.0, -0.0337, 0.06061, -0.0, -0.0164, -0.07898, 0.0, -0.02706, -0.10291, 0.0, -0.05115, -0.03621, 0.0, -0.03284, 0.00967, 0.0, 0.0], [-0.00016, 0.0, -0.09803, 0.0171, -0.0, -0.04291, 0.04036, -0.0, 0.09543, -0.01368, -0.0, 0.06443, 0.01146, -0.0, -0.0283, -0.03513, 0.0, -0.09109, -0.08631, -0.0, 0.0, -0.00016, -0.0, 0.09803, 0.0171, -0.0, 0.04291, 0.04036, -0.0, -0.09543, -0.01368, 0.0, -0.06443, 0.01146, 0.0, 0.0283, -0.03513, 0.0, 0.09109, 0.04645, 0.0, 0.0]]), "B2":array([[-0.0, 0.05807, 0.0, -0.0, -0.01193, 0.0, -0.0, -0.09381, -0.0, -0.0, -0.10727, -0.0, -0.0, -0.03699, 0.0, -0.0, 0.04976, 0.0, -0.0, 0.14536, -0.0, -0.0, 0.05807, 0.0, -0.0, -0.01193, 0.0, -0.0, -0.09381, -0.0, -0.0, -0.10727, -0.0, -0.0, -0.03699, 0.0, -0.0, 0.04976, 0.0, -0.0, 0.13898, -0.0], [0.0, 0.05104, 0.0, 0.0, -0.06321, -0.0, -0.0, -0.07852, -0.0, -0.0, 0.03314, -0.0, -0.0, 0.06605, -0.0, -0.0, 0.03602, -0.0, -0.0, -0.21653, -0.0, -0.0, 0.05104, 0.0, -0.0, -0.06321, -0.0, -0.0, -0.07852, -0.0, -0.0, 0.03314, -0.0, -0.0, 0.06605, 0.0, -0.0, 0.03602, 0.0, 0.0, 0.1275, -0.0]])}, database_path = 'xanthene_scsd_20220713.pkl', maxdist = 1.75)
model_objs_dict['tcnq'] = scsd_model('tcnq', array([['-2.8901', '0.0', '0.0', 'C'], ['-0.6888', '1.1899', '0.0', 'C'], ['-0.6888', '-1.1899', '0.0', 'C'], ['-3.6106', '-1.1815', '0.0', 'C'], ['-1.4208', '0.0', '0.0', 'C'], ['-3.6106', '1.1815', '0.0', 'C'], ['2.8901', '-0.0', '-0.0', 'C'], ['0.6888', '-1.1899', '-0.0', 'C'], ['0.6888', '1.1899', '-0.0', 'C'], ['3.6106', '1.1815', '-0.0', 'C'], ['1.4208', '-0.0', '-0.0', 'C'], ['3.6106', '-1.1815', '-0.0', 'C'], ['-4.1863', '-2.1917', '0.0', 'N'], ['-4.1863', '2.1917', '0.0', 'N'], ['4.1863', '2.1917', '-0.0', 'N'], ['4.1863', '-2.1917', '-0.0', 'N']]), 'D2h' , pca = {"Ag":array([[0.05077, -0.0, 0.0, 0.00017, 0.01413, 0.0, 0.00017, -0.01413, -0.0, 0.07483, -0.00698, 0.0, 0.0179, -0.0, 0.0, 0.07483, 0.00698, 0.0, -0.05077, 0.0, -0.0, -0.00017, -0.01413, -0.0, -0.00017, 0.01413, 0.0, -0.07483, 0.00698, -0.0, -0.0179, 0.0, -0.0, -0.07483, -0.00698, -0.0, 0.11734, -0.0273, 0.0, 0.11734, 0.0273, 0.0, -0.11734, 0.0273, -0.0, -0.11734, -0.0273, -0.0], [0.05944, 0.0, 0.0, -0.00506, 0.00206, -0.0, -0.00506, -0.00206, -0.0, 0.05598, 0.02949, 0.0, -0.00313, 0.0, -0.0, 0.05598, -0.02949, 0.0, -0.05944, 0.0, -0.0, 0.00506, -0.00206, 0.0, 0.00506, 0.00206, 0.0, -0.05598, -0.02949, -0.0, 0.00313, -0.0, 0.0, -0.05598, 0.02949, -0.0, -0.01331, 0.14396, 0.0, -0.01331, -0.14396, -0.0, 0.01331, -0.14396, -0.0, 0.01331, 0.14396, 0.0]]), "B1g":array([[0.0, -0.06032, 0.0, -0.03927, -0.02981, 0.0, 0.03927, -0.02981, -0.0, -0.0429, -0.03474, 0.0, 0.0, -0.05197, 0.0, 0.0429, -0.03474, 0.0, 0.0, 0.06032, -0.0, 0.03927, 0.02981, -0.0, -0.03927, 0.02981, 0.0, 0.0429, 0.03474, -0.0, 0.0, 0.05197, -0.0, -0.0429, 0.03474, -0.0, -0.1057, 0.00822, 0.0, 0.1057, 0.00822, 0.0, 0.1057, -0.00822, -0.0, -0.1057, -0.00822, -0.0], [0.0, 0.0259, 0.0, 0.08801, 0.03909, 0.0, -0.08801, 0.03909, -0.0, -0.00519, -0.0248, 0.0, 0.0, 0.0534, 0.0, 0.00519, -0.0248, 0.0, 0.0, -0.0259, -0.0, -0.08801, -0.03909, -0.0, 0.08801, -0.03909, 0.0, 0.00519, 0.0248, -0.0, 0.0, -0.0534, -0.0, -0.00519, 0.0248, -0.0, -0.05867, -0.06024, 0.0, 0.05867, -0.06024, 0.0, 0.05867, 0.06024, -0.0, -0.05867, 0.06024, -0.0]]), "B2g":array([[0.0, -0.0, -0.0, 0.0, -0.0, -0.15316, -0.0, 0.0, 0.15316, -0.0, -0.0, -0.02969, 0.0, -0.0, -0.0, -0.0, 0.0, 0.02969, 0.0, 0.0, 0.0, -0.0, 0.0, 0.15316, 0.0, -0.0, -0.15316, 0.0, 0.0, 0.02969, 0.0, 0.0, 0.0, 0.0, -0.0, -0.02969, 0.0, -0.0, -0.06715, 0.0, 0.0, 0.06715, -0.0, 0.0, 0.06715, -0.0, -0.0, -0.06715], [-0.0, 0.0, 0.0, -0.0, -0.0, -0.00552, 0.0, 0.0, 0.00552, 0.0, -0.0, -0.1608, 0.0, -0.0, -0.0, 0.0, 0.0, 0.1608, 0.0, 0.0, -0.0, 0.0, 0.0, 0.00552, -0.0, -0.0, -0.00552, -0.0, 0.0, 0.1608, 0.0, 0.0, 0.0, -0.0, -0.0, -0.1608, -0.0, -0.0, 0.08369, 0.0, 0.0, -0.08369, 0.0, 0.0, -0.08369, -0.0, -0.0, 0.08369]]), "B3g":array([[0.0, -0.0, 0.11458, -0.0, -0.0, 0.04475, 0.0, 0.0, 0.04475, 0.0, -0.0, 0.02437, 0.0, -0.0, 0.08311, 0.0, 0.0, 0.02437, 0.0, 0.0, -0.11458, 0.0, 0.0, -0.04475, -0.0, -0.0, -0.04475, -0.0, 0.0, -0.02437, 0.0, 0.0, -0.08311, -0.0, -0.0, -0.02437, 0.0, 0.0, -0.08204, 0.0, -0.0, -0.08204, -0.0, -0.0, 0.08204, -0.0, 0.0, 0.08204], [-0.0, 0.0, -0.02208, 0.0, -0.0, 0.07126, 0.0, 0.0, 0.07126, -0.0, -0.0, -0.07628, 0.0, -0.0, 0.09005, -0.0, 0.0, -0.07628, 0.0, -0.0, 0.02208, -0.0, 0.0, -0.07126, -0.0, -0.0, -0.07126, 0.0, 0.0, 0.07628, 0.0, 0.0, -0.09005, 0.0, -0.0, 0.07628, 0.0, 0.0, 0.0464, 0.0, -0.0, 0.0464, -0.0, -0.0, -0.0464, -0.0, 0.0, -0.0464]]), "Au":array([[-0.0, -0.0, -0.0, 0.0, 0.0, -0.00127, -0.0, 0.0, 0.00127, -0.0, 0.0, 0.0862, -0.0, -0.0, -0.0, -0.0, 0.0, -0.0862, -0.0, -0.0, 0.0, 0.0, 0.0, -0.00127, -0.0, 0.0, 0.00127, -0.0, -0.0, 0.0862, -0.0, -0.0, 0.0, -0.0, 0.0, -0.0862, -0.0, -0.0, 0.16253, 0.0, 0.0, -0.16253, -0.0, -0.0, 0.16253, -0.0, 0.0, -0.16253], [0.0, -0.0, -0.0, -0.0, 0.0, 0.19912, 0.0, -0.0, -0.19912, -0.0, 0.0, 0.03426, -0.0, 0.0, -0.0, -0.0, 0.0, -0.03426, -0.0, -0.0, -0.0, -0.0, -0.0, 0.19912, 0.0, 0.0, -0.19912, -0.0, -0.0, 0.03426, -0.0, 0.0, 0.0, -0.0, 0.0, -0.03426, -0.0, -0.0, -0.01662, -0.0, -0.0, 0.01662, -0.0, 0.0, -0.01662, 0.0, 0.0, 0.01662]]), "B1u":array([[-0.0, -0.0, 0.00022, -0.0, -0.0, 0.09419, -0.0, -0.0, 0.09419, -0.0, 0.0, -0.04296, -0.0, -0.0, 0.0614, -0.0, -0.0, -0.04296, -0.0, -0.0, 0.00022, -0.0, -0.0, 0.09419, -0.0, -0.0, 0.09419, -0.0, -0.0, -0.04296, -0.0, -0.0, 0.0614, -0.0, 0.0, -0.04296, -0.0, 0.0, -0.08204, -0.0, 0.0, -0.08204, -0.0, 0.0, -0.08204, 0.0, 0.0, -0.08204], [-0.0, 0.0, -0.14232, -0.0, -0.0, 0.05362, -0.0, -0.0, 0.05362, -0.0, -0.0, -0.03954, -0.0, 0.0, -0.02859, -0.0, 0.0, -0.03954, -0.0, 0.0, -0.14232, -0.0, -0.0, 0.05362, -0.0, -0.0, 0.05362, -0.0, 0.0, -0.03954, -0.0, 0.0, -0.02859, -0.0, -0.0, -0.03954, -0.0, 0.0, 0.07138, -0.0, 0.0, 0.07138, -0.0, 0.0, 0.07138, 0.0, 0.0, 0.07138]]), "B2u":array([[0.04097, -0.0, -0.0, 0.04492, 0.00156, -0.0, 0.04492, -0.00156, -0.0, 0.00547, 0.02611, 0.0, 0.0432, -0.0, -0.0, 0.00547, -0.02611, -0.0, 0.04097, -0.0, 0.0, 0.04492, 0.00156, 0.0, 0.04492, -0.00156, 0.0, 0.00547, 0.02611, 0.0, 0.0432, -0.0, 0.0, 0.00547, -0.02611, -0.0, -0.09248, 0.08486, 0.0, -0.09248, -0.08486, -0.0, -0.09248, 0.08486, 0.0, -0.09248, -0.08486, -0.0], [-0.03461, -0.0, -0.0, -0.04842, 0.00418, 0.0, -0.04842, -0.00418, 0.0, 0.03684, 0.05135, -0.0, -0.04412, 0.0, -0.0, 0.03684, -0.05135, -0.0, -0.03461, 0.0, 0.0, -0.04842, 0.00418, -0.0, -0.04842, -0.00418, -0.0, 0.03684, 0.05135, 0.0, -0.04412, 0.0, 0.0, 0.03684, -0.05135, 0.0, 0.05094, 0.08249, -0.0, 0.05094, -0.08249, -0.0, 0.05094, 0.08249, 0.0, 0.05094, -0.08249, 0.0]]), "B3u":array([[-0.0, -0.01478, 0.0, 0.00356, -0.0483, -0.0, -0.00356, -0.0483, 0.0, -0.04868, 0.02027, -0.0, -0.0, -0.04527, -0.0, 0.04868, 0.02027, 0.0, -0.0, -0.01478, -0.0, 0.00356, -0.0483, 0.0, -0.00356, -0.0483, -0.0, -0.04868, 0.02027, -0.0, -0.0, -0.04527, 0.0, 0.04868, 0.02027, 0.0, -0.10151, 0.05805, -0.0, 0.10151, 0.05805, 0.0, -0.10151, 0.05805, -0.0, 0.10151, 0.05805, 0.0], [0.0, 0.01179, -0.0, -0.00995, 0.07022, 0.0, 0.00995, 0.07022, -0.0, -0.02687, -0.0572, -0.0, 0.0, 0.05455, -0.0, 0.02687, -0.0572, 0.0, 0.0, 0.01179, -0.0, -0.00995, 0.07022, -0.0, 0.00995, 0.07022, 0.0, -0.02687, -0.0572, 0.0, 0.0, 0.05455, 0.0, 0.02687, -0.0572, -0.0, -0.07173, -0.04619, -0.0, 0.07173, -0.04619, 0.0, -0.07173, -0.04619, -0.0, 0.07173, -0.04619, 0.0]])}, database_path = 'tcnq_scsd_20220713.pkl', maxdist = 1.75)
model_objs_dict['ttf'] = scsd_model('ttf', array([['0.6572', '-3.1482', '-0.0', 'C'], ['-0.6572', '-3.1482', '-0.0', 'C'], ['0.0', '-0.6752', '-0.0', 'C'], ['-0.6572', '3.1482', '0.0', 'C'], ['0.6572', '3.1482', '0.0', 'C'], ['0.0', '0.6752', '0.0', 'C'], ['1.4776', '-1.6247', '-0.0', 'S'], ['-1.4776', '-1.6247', '-0.0', 'S'], ['-1.4776', '1.6247', '0.0', 'S'], ['1.4776', '1.6247', '0.0', 'S']]), 'D2h' , pca = {"Ag":array([[-0.0054, 0.23534, 0.0, 0.0054, 0.23534, 0.0, 0.0, 0.02078, 0.0, 0.0054, -0.23534, -0.0, -0.0054, -0.23534, -0.0, -0.0, -0.02078, -0.0, -0.00184, 0.00224, 0.0, 0.00184, 0.00224, 0.0, 0.00184, -0.00224, -0.0, -0.00184, -0.00224, -0.0], [0.02976, 0.00533, 0.0, -0.02976, 0.00533, -0.0, -0.0, -0.13963, -0.0, -0.02976, -0.00533, -0.0, 0.02976, -0.00533, 0.0, 0.0, 0.13963, 0.0, -0.1566, 0.03104, -0.0, 0.1566, 0.03104, 0.0, 0.1566, -0.03104, 0.0, -0.1566, -0.03104, -0.0]]), "B1g":array([[-0.03109, 0.14382, 0.0, -0.03109, -0.14382, 0.0, -0.09067, 0.0, 0.0, 0.03109, -0.14382, -0.0, 0.03109, 0.14382, -0.0, 0.09067, 0.0, -0.0, -0.0287, 0.05455, 0.0, -0.0287, -0.05455, 0.0, 0.0287, -0.05455, -0.0, 0.0287, 0.05455, -0.0], [0.06996, -0.05326, 0.0, 0.06996, 0.05326, 0.0, -0.16429, -0.0, 0.0, -0.06996, 0.05326, -0.0, -0.06996, -0.05326, -0.0, 0.16429, -0.0, -0.0, -0.08093, 0.00116, 0.0, -0.08093, -0.00116, 0.0, 0.08093, -0.00116, -0.0, 0.08093, 0.00116, -0.0]]), "B2g":array([[-0.0, 0.0, -0.07897, -0.0, 0.0, -0.07897, -0.0, -0.0, 0.06163, 0.0, -0.0, 0.07897, 0.0, -0.0, 0.07897, 0.0, -0.0, -0.06163, -0.0, -0.0, 0.14022, -0.0, -0.0, 0.14022, 0.0, 0.0, -0.14022, 0.0, 0.0, -0.14022], [0.0, 0.0, 0.00166, -0.0, -0.0, 0.00166, -0.0, -0.0, 0.34634, 0.0, 0.0, -0.00166, 0.0, 0.0, -0.00166, 0.0, -0.0, -0.34634, -0.0, 0.0, -0.07518, -0.0, 0.0, -0.07518, 0.0, -0.0, 0.07518, 0.0, -0.0, 0.07518]]), "B3g":array([[-0.0, 0.0, 0.17304, 0.0, 0.0, -0.17304, -0.0, -0.0, -0.0, -0.0, -0.0, -0.17304, -0.0, -0.0, 0.17304, 0.0, 0.0, 0.0, 0.0, 0.0, -0.07696, -0.0, 0.0, 0.07696, -0.0, -0.0, 0.07696, 0.0, -0.0, -0.07696]]), "Au":array([[-0.0, -0.0, 0.08527, -0.0, -0.0, -0.08527, 0.0, 0.0, -0.0, 0.0, -0.0, 0.08527, 0.0, -0.0, -0.08527, -0.0, 0.0, 0.0, -0.0, 0.0, 0.16473, -0.0, -0.0, -0.16473, -0.0, -0.0, 0.16473, -0.0, 0.0, -0.16473], [0.0, -0.0, 0.16473, -0.0, -0.0, -0.16473, -0.0, -0.0, 0.0, 0.0, -0.0, 0.16473, 0.0, -0.0, -0.16473, 0.0, -0.0, -0.0, -0.0, 0.0, -0.08527, -0.0, -0.0, 0.08527, -0.0, -0.0, -0.08527, -0.0, 0.0, 0.08527]]), "B1u":array([[-0.0, -0.0, -0.125, -0.0, -0.0, -0.125, -0.0, -0.0, 0.08418, 0.0, -0.0, -0.125, 0.0, -0.0, -0.125, -0.0, -0.0, 0.08418, -0.0, -0.0, 0.08291, -0.0, -0.0, 0.08291, -0.0, -0.0, 0.08291, -0.0, -0.0, 0.08291], [0.0, -0.0, -0.00076, -0.0, 0.0, -0.00076, 0.0, 0.0, -0.24849, -0.0, 0.0, -0.00076, 0.0, 0.0, -0.00076, 0.0, 0.0, -0.24849, 0.0, 0.0, 0.125, 0.0, 0.0, 0.125, 0.0, 0.0, 0.125, 0.0, 0.0, 0.125]]), "B2u":array([[0.08931, 0.05311, 0.0, 0.08931, -0.05311, -0.0, -0.10626, 0.0, 0.0, 0.08931, 0.05311, 0.0, 0.08931, -0.05311, -0.0, -0.10626, 0.0, 0.0, -0.03619, 0.10026, 0.0, -0.03619, -0.10026, -0.0, -0.03619, 0.10026, 0.0, -0.03619, -0.10026, -0.0], [0.06201, -0.17035, -0.0, 0.06201, 0.17035, 0.0, -0.03797, -0.0, -0.0, 0.06201, -0.17035, -0.0, 0.06201, 0.17035, 0.0, -0.03797, -0.0, -0.0, -0.04302, -0.00065, 0.0, -0.04302, 0.00065, -0.0, -0.04302, -0.00065, -0.0, -0.04302, 0.00065, 0.0]]), "B3u":array([[-0.01253, 0.12417, 0.0, 0.01253, 0.12417, 0.0, -0.0, -0.11304, -0.0, -0.01253, 0.12417, -0.0, 0.01253, 0.12417, -0.0, 0.0, -0.11304, 0.0, 0.01252, -0.06764, 0.0, -0.01252, -0.06764, -0.0, 0.01252, -0.06764, 0.0, -0.01252, -0.06764, -0.0], [0.18333, 0.00199, 0.0, -0.18333, 0.00199, -0.0, 0.0, -0.0756, 0.0, 0.18333, 0.00199, 0.0, -0.18333, 0.00199, -0.0, -0.0, -0.0756, -0.0, 0.01584, 0.03581, 0.0, -0.01584, 0.03581, -0.0, 0.01584, 0.03581, 0.0, -0.01584, 0.03581, -0.0]])}, database_path = 'ttf_scsd_20220713.pkl', maxdist = 2.4)
model_objs_dict['dhbq'] = scsd_model('dhbq', array([['-1.2028', '0.751', '-0.0', 'C'], ['-1.2028', '-0.751', '0.0', 'C'], ['-0.0', '1.4626', '-0.0', 'C'], ['0.0', '-1.4626', '0.0', 'C'], ['1.2028', '0.751', '-0.0', 'C'], ['1.2028', '-0.751', '0.0', 'C'], ['-2.354', '1.3321', '-0.0', 'O'], ['-2.354', '-1.3321', '0.0', 'O'], ['2.354', '1.3321', '-0.0', 'O'], ['2.354', '-1.3321', '0.0', 'O']]), 'D2h' , pca = {"Ag":array([[-0.08539, -0.00211, -0.0, -0.08539, 0.00211, -0.0, -0.0, -0.08519, -0.0, 0.0, 0.08519, 0.0, 0.08539, -0.00211, 0.0, 0.08539, 0.00211, 0.0, -0.02286, 0.15289, 0.0, -0.02286, -0.15289, -0.0, 0.02286, 0.15289, 0.0, 0.02286, -0.15289, -0.0], [0.09954, -0.0743, 0.0, 0.09954, 0.0743, 0.0, 0.0, 0.0142, 0.0, -0.0, -0.0142, -0.0, -0.09954, -0.0743, -0.0, -0.09954, 0.0743, -0.0, 0.0745, 0.06966, 0.0, 0.0745, -0.06966, 0.0, -0.0745, 0.06966, -0.0, -0.0745, -0.06966, -0.0]]), "B1g":array([[-0.01656, -0.03723, 0.0, 0.01656, -0.03723, -0.0, 0.12546, -0.0, -0.0, -0.12546, -0.0, 0.0, -0.01656, 0.03723, 0.0, 0.01656, 0.03723, -0.0, 0.11565, -0.08012, -0.0, -0.11565, -0.08012, -0.0, 0.11565, 0.08012, 0.0, -0.11565, 0.08012, 0.0], [-0.03528, 0.12945, -0.0, 0.03528, 0.12945, 0.0, 0.10284, -0.0, -0.0, -0.10284, -0.0, 0.0, -0.03528, -0.12945, -0.0, 0.03528, -0.12945, 0.0, -0.05698, -0.05459, -0.0, 0.05698, -0.05459, 0.0, -0.05698, 0.05459, -0.0, 0.05698, 0.05459, 0.0]]), "B2g":array([[0.0, -0.0, 0.02073, 0.0, 0.0, -0.02073, -0.0, 0.0, 0.20742, 0.0, 0.0, -0.20742, 0.0, -0.0, 0.02073, 0.0, 0.0, -0.02073, 0.0, -0.0, -0.12556, 0.0, 0.0, 0.12556, 0.0, -0.0, -0.12556, 0.0, 0.0, 0.12556], [0.0, 0.0, -0.16257, 0.0, -0.0, 0.16257, -0.0, 0.0, 0.08618, 0.0, 0.0, -0.08618, 0.0, 0.0, -0.16257, 0.0, -0.0, 0.16257, 0.0, 0.0, 0.04434, 0.0, -0.0, -0.04434, 0.0, 0.0, 0.04434, 0.0, -0.0, -0.04434]]), "B3g":array([[0.0, -0.0, -0.16546, 0.0, 0.0, -0.16546, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -0.0, 0.16546, 0.0, 0.0, 0.16546, 0.0, 0.0, 0.08454, 0.0, -0.0, 0.08454, 0.0, 0.0, -0.08454, 0.0, -0.0, -0.08454]]), "Au":array([[0.0, 0.0, 0.06086, 0.0, 0.0, -0.06086, 0.0, 0.0, -0.0, 0.0, 0.0, 0.0, 0.0, -0.0, -0.06086, 0.0, -0.0, 0.06086, 0.0, -0.0, 0.18914, 0.0, -0.0, -0.18914, 0.0, 0.0, -0.18914, 0.0, 0.0, 0.18914], [0.0, 0.0, 0.18914, -0.0, 0.0, -0.18914, -0.0, -0.0, -0.0, -0.0, -0.0, 0.0, -0.0, -0.0, -0.18914, -0.0, -0.0, 0.18914, -0.0, -0.0, -0.06086, -0.0, -0.0, 0.06086, -0.0, 0.0, 0.06086, -0.0, 0.0, -0.06086]]), "B1u":array([[0.0, -0.0, 0.03707, -0.0, -0.0, 0.03707, -0.0, -0.0, 0.17587, -0.0, -0.0, 0.17587, -0.0, -0.0, 0.03707, -0.0, -0.0, 0.03707, -0.0, -0.0, -0.125, -0.0, -0.0, -0.125, -0.0, -0.0, -0.125, -0.0, -0.0, -0.125], [0.0, 0.0, -0.125, -0.0, -0.0, -0.125, -0.0, -0.0, 0.13467, -0.0, -0.0, 0.13467, -0.0, -0.0, -0.125, -0.0, -0.0, -0.125, -0.0, -0.0, 0.05767, -0.0, -0.0, 0.05767, -0.0, -0.0, 0.05767, -0.0, -0.0, 0.05767]]), "B2u":array([[0.10097, -0.04061, -0.0, 0.10097, 0.04061, 0.0, -0.05717, -0.0, -0.0, -0.05717, -0.0, -0.0, 0.10097, 0.04061, 0.0, 0.10097, -0.04061, -0.0, -0.07239, -0.08625, -0.0, -0.07239, 0.08625, 0.0, -0.07239, 0.08625, 0.0, -0.07239, -0.08625, -0.0], [0.04794, -0.06361, -0.0, 0.04794, 0.06361, 0.0, -0.14769, 0.0, 0.0, -0.14769, 0.0, 0.0, 0.04794, 0.06361, 0.0, 0.04794, -0.06361, -0.0, 0.0259, 0.11328, 0.0, 0.0259, -0.11328, -0.0, 0.0259, -0.11328, -0.0, 0.0259, 0.11328, 0.0]]), "B3u":array([[0.05311, -0.09308, 0.0, -0.05311, -0.09308, -0.0, -0.0, -0.01141, -0.0, 0.0, -0.01141, 0.0, -0.05311, -0.09308, -0.0, 0.05311, -0.09308, 0.0, 0.06906, 0.09878, 0.0, -0.06906, 0.09878, -0.0, -0.06906, 0.09878, -0.0, 0.06906, 0.09878, 0.0], [0.02882, 0.07224, -0.0, -0.02882, 0.07224, -0.0, 0.0, -0.04158, 0.0, -0.0, -0.04158, -0.0, -0.02882, 0.07224, 0.0, 0.02882, 0.07224, -0.0, 0.14535, -0.05145, 0.0, -0.14535, -0.05145, -0.0, -0.14535, -0.05145, -0.0, 0.14535, -0.05145, 0.0]])}, database_path = 'dhbq_scsd_20220713.pkl', maxdist = 1.75)
model_objs_dict['triptycene'] = scsd_model('triptycene', array([['0.0', '-0.0', '-1.2995', 'C'], ['0.0', '-0.0', '1.2995', 'C'], ['-1.4017', '-0.0', '-0.7021', 'C'], ['-2.5974', '0.0', '-1.4048', 'C'], ['-3.8042', '0.0001', '-0.6953', 'C'], ['-3.8042', '0.0001', '0.6953', 'C'], ['-2.5974', '0.0', '1.4048', 'C'], ['-1.4017', '-0.0', '0.7021', 'C'], ['0.7009', '1.2139', '-0.7021', 'C'], ['1.2987', '2.2494', '-1.4048', 'C'], ['1.902', '3.2946', '-0.6953', 'C'], ['1.902', '3.2946', '0.6953', 'C'], ['1.2987', '2.2494', '1.4048', 'C'], ['0.7009', '1.2139', '0.7021', 'C'], ['0.7009', '-1.214', '-0.7021', 'C'], ['1.2987', '-2.2494', '-1.4048', 'C'], ['1.9021', '-3.2946', '-0.6953', 'C'], ['1.9021', '-3.2946', '0.6953', 'C'], ['1.2987', '-2.2494', '1.4048', 'C'], ['0.7009', '-1.214', '0.7021', 'C']]), 'D3h' , pca = {"A'1":array([[0.0, 0.0, -0.02974, 0.0, -0.0, 0.02974, -0.02584, 0.0, -0.00029, -0.05665, 0.0, 0.01184, -0.08046, -0.0, -0.00035, -0.08046, -0.0, 0.00035, -0.05665, 0.0, -0.01184, -0.02584, 0.0, 0.00029, 0.01292, 0.02238, -0.00029, 0.02833, 0.04906, 0.01184, 0.04023, 0.06968, -0.00035, 0.04023, 0.06968, 0.00035, 0.02833, 0.04906, -0.01184, 0.01292, 0.02238, 0.00029, 0.01292, -0.02238, -0.00029, 0.02833, -0.04906, 0.01184, 0.04023, -0.06968, -0.00035, 0.04023, -0.06968, 0.00035, 0.02833, -0.04906, -0.01184, 0.01292, -0.02238, 0.00029], [0.0, 0.0, -0.00031, 0.0, 0.0, 0.00031, 0.02131, 0.0, 0.01622, 0.00534, -0.0, 0.11068, 0.00555, -0.0, 0.02724, 0.00555, 0.0, -0.02724, 0.00534, 0.0, -0.11068, 0.02131, -0.0, -0.01622, -0.01066, -0.01846, 0.01622, -0.00267, -0.00463, 0.11068, -0.00277, -0.0048, 0.02724, -0.00277, -0.0048, -0.02724, -0.00267, -0.00463, -0.11068, -0.01066, -0.01846, -0.01622, -0.01066, 0.01846, 0.01622, -0.00267, 0.00463, 0.11068, -0.00277, 0.0048, 0.02724, -0.00277, 0.0048, -0.02724, -0.00267, 0.00463, -0.11068, -0.01066, 0.01846, -0.01622]]), "A'2":array([[0.0, 0.0, -0.0, 0.0, 0.0, -0.0, 0.0, 0.06698, 0.0, 0.0, 0.04777, 0.0, 0.0, -0.0573, 0.0, 0.0, -0.0573, -0.0, 0.0, 0.04777, -0.0, 0.0, 0.06698, -0.0, 0.058, -0.03349, 0.0, 0.04137, -0.02389, 0.0, -0.04962, 0.02865, -0.0, -0.04962, 0.02865, 0.0, 0.04137, -0.02389, -0.0, 0.058, -0.03349, -0.0, -0.058, -0.03349, -0.0, -0.04137, -0.02389, -0.0, 0.04962, 0.02865, -0.0, 0.04962, 0.02865, 0.0, -0.04137, -0.02389, 0.0, -0.058, -0.03349, 0.0], [-0.0, -0.0, 0.0, -0.0, -0.0, 0.0, -0.0, -0.07186, 0.0, 0.0, 0.07285, -0.0, 0.0, -0.02326, -0.0, 0.0, -0.02326, 0.0, 0.0, 0.07285, 0.0, -0.0, -0.07186, -0.0, -0.06223, 0.03593, 0.0, 0.06309, -0.03643, 0.0, -0.02014, 0.01163, 0.0, -0.02014, 0.01163, -0.0, 0.06309, -0.03643, -0.0, -0.06223, 0.03593, -0.0, 0.06223, 0.03593, -0.0, -0.06309, -0.03643, 0.0, 0.02014, 0.01163, -0.0, 0.02014, 0.01163, 0.0, -0.06309, -0.03643, -0.0, 0.06223, 0.03593, 0.0]]), "A''1":array([[0.0, -0.0, -0.0, 0.0, -0.0, -0.0, 0.0, 0.02544, 0.0, 0.0, 0.08831, 0.0, -0.0, 0.05418, 0.0, 0.0, -0.05418, 0.0, 0.0, -0.08831, 0.0, 0.0, -0.02544, 0.0, 0.02203, -0.01272, -0.0, 0.07648, -0.04416, -0.0, 0.04692, -0.02709, -0.0, -0.04692, 0.02709, -0.0, -0.07648, 0.04416, -0.0, -0.02203, 0.01272, -0.0, -0.02203, -0.01272, 0.0, -0.07648, -0.04416, 0.0, -0.04692, -0.02709, 0.0, 0.04692, 0.02709, -0.0, 0.07648, 0.04416, -0.0, 0.02203, 0.01272, -0.0], [0.0, 0.0, 0.0, -0.0, -0.0, -0.0, -0.0, 0.01048, -0.0, 0.0, 0.05832, -0.0, -0.0, -0.09998, 0.0, 0.0, 0.09998, 0.0, -0.0, -0.05832, 0.0, 0.0, -0.01048, -0.0, 0.00907, -0.00524, 0.0, 0.05051, -0.02916, -0.0, -0.08659, 0.04999, -0.0, 0.08659, -0.04999, -0.0, -0.05051, 0.02916, 0.0, -0.00907, 0.00524, 0.0, -0.00907, -0.00524, 0.0, -0.05051, -0.02916, 0.0, 0.08659, 0.04999, 0.0, -0.08659, -0.04999, -0.0, 0.05051, 0.02916, -0.0, 0.00907, 0.00524, -0.0]]), "A''2":array([[0.0, -0.0, -0.06509, -0.0, 0.0, -0.06509, -0.01278, -0.0, -0.04336, -0.04487, -0.0, 0.00654, -0.02045, -0.0, 0.05852, 0.02045, 0.0, 0.05852, 0.04487, 0.0, 0.00654, 0.01278, 0.0, -0.04336, 0.00639, 0.01107, -0.04336, 0.02243, 0.03886, 0.00654, 0.01023, 0.01771, 0.05852, -0.01023, -0.01771, 0.05852, -0.02243, -0.03886, 0.00654, -0.00639, -0.01107, -0.04336, 0.00639, -0.01107, -0.04336, 0.02243, -0.03886, 0.00654, 0.01023, -0.01771, 0.05852, -0.01023, 0.01771, 0.05852, -0.02243, 0.03886, 0.00654, -0.00639, 0.01107, -0.04336], [-0.0, 0.0, -0.10522, 0.0, -0.0, -0.10522, 0.01943, -0.0, -0.00628, 0.05767, 0.0, 0.02125, 0.0506, -0.0, 0.02011, -0.0506, -0.0, 0.02011, -0.05767, -0.0, 0.02125, -0.01943, -0.0, -0.00628, -0.00972, -0.01683, -0.00628, -0.02884, -0.04994, 0.02125, -0.0253, -0.04382, 0.02011, 0.0253, 0.04382, 0.02011, 0.02884, 0.04994, 0.02125, 0.00972, 0.01683, -0.00628, -0.00972, 0.01683, -0.00628, -0.02884, 0.04994, 0.02125, -0.0253, 0.04382, 0.02011, 0.0253, -0.04382, 0.02011, 0.02884, -0.04994, 0.02125, 0.00972, -0.01683, -0.00628]]), "E'1":array([[0.04053, -0.0, -0.0, 0.04053, -0.0, -0.0, 0.03808, -0.0, -0.001, 0.03694, -0.0, -0.00055, 0.03503, -0.0, 0.00045, 0.03503, -0.0, -0.00045, 0.03694, -0.0, 0.00055, 0.03808, -0.0, 0.001, 0.02467, 0.00774, 0.0005, -0.02466, 0.03556, 0.00027, -0.0753, 0.0637, -0.00022, -0.0753, 0.0637, 0.00022, -0.02466, 0.03556, -0.00027, 0.02467, 0.00774, -0.0005, 0.02467, -0.00774, 0.0005, -0.02466, -0.03556, 0.00027, -0.0753, -0.0637, -0.00022, -0.0753, -0.0637, 0.00022, -0.02466, -0.03556, -0.00027, 0.02467, -0.00774, -0.0005], [-0.05263, -0.0, 0.0, -0.05263, 0.0, 0.0, -0.01379, 0.0, 0.0027, 0.05175, -0.0, -0.0518, 0.06988, -0.0, -0.01079, 0.06988, -0.0, 0.01079, 0.05175, -0.0, 0.0518, -0.01379, 0.0, -0.0027, -0.04595, 0.01857, -0.00135, -0.02172, 0.04242, 0.0259, 0.04006, 0.01722, 0.0054, 0.04006, 0.01722, -0.0054, -0.02172, 0.04242, -0.0259, -0.04595, 0.01857, 0.00135, -0.04595, -0.01857, -0.00135, -0.02172, -0.04242, 0.0259, 0.04006, -0.01722, 0.0054, 0.04006, -0.01722, -0.0054, -0.02172, -0.04242, -0.0259, -0.04595, -0.01857, 0.00135]]), "E'2":array([[-0.02141, -0.03709, 0.0, -0.02141, -0.03709, 0.0, -0.01982, -0.01926, 0.00045, -0.01879, 0.04117, 0.00046, -0.018, 0.10088, -0.00027, -0.018, 0.10088, 0.00027, -0.01879, 0.04117, -0.00046, -0.01982, -0.01926, -0.00045, -0.01982, -0.03433, -0.00091, -0.01879, -0.03255, -0.00092, -0.018, -0.03117, 0.00054, -0.018, -0.03117, -0.00054, -0.01879, -0.03255, 0.00092, -0.01982, -0.03433, 0.00091, -0.00677, -0.0268, 0.00045, 0.04505, 0.00431, 0.00046, 0.09636, 0.03485, -0.00027, 0.09636, 0.03485, 0.00027, 0.04505, 0.00431, -0.00046, -0.00677, -0.0268, -0.00045], [0.0024, 0.00415, 0.0, 0.0024, 0.00415, -0.0, -0.00551, 0.04778, 0.00239, -0.02506, 0.07972, 0.01846, -0.01654, -0.04868, 0.0131, -0.01654, -0.04868, -0.0131, -0.02506, 0.07972, -0.01846, -0.00551, 0.04778, -0.00239, -0.00551, -0.00954, -0.00478, -0.02506, -0.04341, -0.03693, -0.01654, -0.02864, -0.02619, -0.01654, -0.02864, 0.02619, -0.02506, -0.04341, 0.03693, -0.00551, -0.00954, 0.00478, 0.04413, 0.01912, 0.00239, 0.08157, 0.01816, 0.01846, -0.03389, -0.03866, 0.0131, -0.03389, -0.03866, -0.0131, 0.08157, 0.01816, -0.01846, 0.04413, 0.01912, -0.00239]]), "E'3":array([[0.02004, -0.0347, -0.0, 0.02004, -0.0347, -0.0, 0.01957, -0.01674, -0.00058, 0.02099, 0.04397, -0.00026, 0.02166, 0.10367, 4e-05, 0.02166, 0.10367, -4e-05, 0.02099, 0.04397, 0.00026, 0.01957, -0.01674, 0.00058, 0.00471, -0.02532, -0.00058, -0.04857, 0.00381, -0.00026, -0.10061, 0.03308, 4e-05, -0.10061, 0.03308, -4e-05, -0.04857, 0.00381, 0.00026, 0.00471, -0.02532, 0.00058, 0.01957, -0.0339, 0.00116, 0.02099, -0.03635, 0.00053, 0.02166, -0.03752, -8e-05, 0.02166, -0.03752, 8e-05, 0.02099, -0.03635, -0.00053, 0.01957, -0.0339, -0.00116], [-0.00946, 0.01638, 0.0, -0.00946, 0.01638, -0.0, 0.0018, 0.05032, -0.00168, 0.01889, 0.07992, -0.00454, 0.02996, -0.05345, -0.00058, 0.02996, -0.05345, 0.00058, 0.01889, 0.07992, 0.00454, 0.0018, 0.05032, 0.00168, -0.04448, 0.02361, -0.00168, -0.07865, 0.0236, -0.00454, 0.0313, -0.05267, -0.00058, 0.0313, -0.05267, 0.00058, -0.07865, 0.0236, 0.00454, -0.04448, 0.02361, 0.00168, 0.0018, -0.00311, 0.00335, 0.01889, -0.03271, 0.00909, 0.02996, -0.05189, 0.00116, 0.02996, -0.05189, -0.00116, 0.01889, -0.03271, -0.00909, 0.0018, -0.00311, -0.00335]]), "E''1":array([[-0.0, 0.00507, 0.0, 0.0, -0.00507, -0.0, -0.0, -0.02809, 0.0, -0.0, -0.1257, 0.0, -0.0, -0.07951, 0.0, -0.0, 0.07951, 0.0, 0.0, 0.1257, 0.0, 0.0, 0.02809, 0.0, 0.01235, -0.00669, 0.01649, 0.05733, -0.02639, 0.02423, 0.03549, -0.01804, 0.03105, -0.03549, 0.01804, 0.03105, -0.05733, 0.02639, 0.02423, -0.01235, 0.00669, 0.01649, -0.01235, -0.00669, -0.01649, -0.05733, -0.02639, -0.02423, -0.03549, -0.01804, -0.03105, 0.03549, 0.01804, -0.03105, 0.05733, 0.02639, -0.02423, 0.01235, 0.00669, -0.01649], [0.0, -0.05619, 0.0, -0.0, 0.05619, 0.0, -0.0, -0.00744, -0.0, -0.0, -0.00466, -0.0, -0.0, -0.00205, 0.0, 0.0, 0.00205, 0.0, 0.0, 0.00466, -0.0, -0.0, 0.00744, -0.0, 0.00659, 0.00397, -0.09294, 0.03017, 0.04759, -0.03478, 0.01285, 0.0202, 0.04568, -0.01285, -0.0202, 0.04568, -0.03017, -0.04759, -0.03478, -0.00659, -0.00397, -0.09294, -0.00659, 0.00397, 0.09294, -0.03017, 0.04759, 0.03478, -0.01285, 0.0202, -0.04568, 0.01285, -0.0202, -0.04568, 0.03017, -0.04759, 0.03478, 0.00659, -0.00397, 0.09294]]), "E''2":array([[-0.00162, 0.00094, -0.0, 0.00162, -0.00094, 0.0, -0.00101, -0.01342, 0.01479, -0.00618, -0.06026, 0.02255, -0.00347, -0.03928, 0.03095, 0.00347, 0.03928, 0.03095, 0.00618, 0.06026, 0.02255, 0.00101, 0.01342, 0.01479, 0.02324, -0.01342, -0.0, 0.10438, -0.06026, -0.0, 0.06804, -0.03928, -0.0, -0.06804, 0.03928, -0.0, -0.10438, 0.06026, -0.0, -0.02324, 0.01342, -0.0, 0.01111, 0.00758, -0.01479, 0.0491, 0.03548, -0.02255, 0.03229, 0.02265, -0.03095, -0.03229, -0.02265, -0.03095, -0.0491, -0.03548, -0.02255, -0.01111, -0.00758, -0.01479], [0.05395, -0.03115, -0.0, -0.05395, 0.03115, -0.0, -0.0014, -0.00477, -0.09229, -0.04033, 0.00182, -0.03077, -0.02174, -0.00334, 0.04959, 0.02174, 0.00334, 0.04959, 0.04033, -0.00182, -0.03077, 0.0014, 0.00477, -0.09229, 0.00826, -0.00477, 0.0, -0.00315, 0.00182, 0.0, 0.00578, -0.00334, -0.0, -0.00578, 0.00334, -0.0, 0.00315, -0.00182, 0.0, -0.00826, 0.00477, 0.0, 0.00343, 0.0036, 0.09229, -0.02174, 0.03401, 0.03077, -0.00798, 0.02049, -0.04959, 0.00798, -0.02049, -0.04959, 0.02174, -0.03401, 0.03077, -0.00343, -0.0036, 0.09229]]), "E''3":array([[0.00442, 0.00255, -0.0, -0.00442, -0.00255, -0.0, -0.00116, -0.01204, -0.01822, -0.00185, -0.05543, -0.02284, -0.0012, -0.03665, -0.02871, 0.0012, 0.03665, -0.02871, 0.00185, 0.05543, -0.02284, 0.00116, 0.01204, -0.01822, -0.01101, 0.00502, 0.01822, -0.04893, 0.02611, 0.02284, -0.03234, 0.01729, 0.02871, 0.03234, -0.01729, 0.02871, 0.04893, -0.02611, 0.02284, 0.01101, -0.00502, 0.01822, -0.02086, -0.01204, 0.0, -0.096, -0.05543, 0.0, -0.06348, -0.03665, 0.0, 0.06348, 0.03665, -0.0, 0.096, 0.05543, -0.0, 0.02086, 0.01204, -0.0], [-0.05044, -0.02912, 0.0, 0.05044, 0.02912, 0.0, 0.00611, -0.00513, 0.09286, 0.06078, -0.00469, 0.0322, 0.02477, -0.00412, -0.04553, -0.02477, 0.00412, -0.04553, -0.06078, 0.00469, 0.0322, -0.00611, 0.00513, 0.09286, -0.00139, 0.00786, -0.09286, 0.02633, 0.05498, -0.0322, 0.00882, 0.02351, 0.04553, -0.00882, -0.02351, 0.04553, -0.02633, -0.05498, -0.0322, 0.00139, -0.00786, -0.09286, -0.00889, -0.00513, -0.0, -0.00812, -0.00469, -0.0, -0.00713, -0.00412, 0.0, 0.00713, 0.00412, -0.0, 0.00812, 0.00469, 0.0, 0.00889, 0.00513, 0.0]])}, database_path = 'triptycene_scsd_20220713.pkl', maxdist = 1.75)
model_objs_dict['cyclam'] = scsd_model('cyclam', array([['2.723', '-1.283', '0.0', 'C'], ['3.3475', '0.0', '-0.0', 'C'], ['2.723', '1.283', '-0.0', 'C'], ['0.7324', '2.7291', '-0.0', 'C'], ['-0.7324', '2.7291', '-0.0', 'C'], ['-2.723', '1.283', '-0.0', 'C'], ['-3.3475', '-0.0', '0.0', 'C'], ['-2.723', '-1.283', '0.0', 'C'], ['-0.7324', '-2.7291', '0.0', 'C'], ['0.7324', '-2.7291', '0.0', 'C'], ['1.3822', '1.5124', '-0.0', 'N'], ['-1.3822', '1.5124', '-0.0', 'N'], ['-1.3822', '-1.5124', '0.0', 'N'], ['1.3822', '-1.5124', '0.0', 'N']]), 'D2h' , pca = {"Ag":array([[-0.10683, 0.03531, -0.0, -0.1239, 0.0, -0.0, -0.10683, -0.03531, -0.0, -0.0271, -0.04845, -0.0, 0.0271, -0.04845, -0.0, 0.10683, -0.03531, 0.0, 0.1239, -0.0, 0.0, 0.10683, 0.03531, 0.0, 0.0271, 0.04845, 0.0, -0.0271, 0.04845, 0.0, 0.0064, 0.01586, 0.0, -0.0064, 0.01586, 0.0, -0.0064, -0.01586, -0.0, 0.0064, -0.01586, -0.0], [-0.04361, -0.02532, -0.0, 0.01272, -0.0, 0.0, -0.04361, 0.02532, -0.0, -0.00946, 0.09749, 0.0, 0.00946, 0.09749, 0.0, 0.04361, 0.02532, 0.0, -0.01272, 0.0, -0.0, 0.04361, -0.02532, 0.0, 0.00946, -0.09749, -0.0, -0.00946, -0.09749, -0.0, 0.05152, 0.07319, 0.0, -0.05152, 0.07319, 0.0, -0.05152, -0.07319, -0.0, 0.05152, -0.07319, -0.0]]), "B1g":array([[-0.03934, 0.03145, 0.0, 0.0, 0.05974, -0.0, 0.03934, 0.03145, -0.0, 0.06124, -0.00046, -0.0, 0.06124, 0.00046, -0.0, 0.03934, -0.03145, -0.0, 0.0, -0.05974, 0.0, -0.03934, -0.03145, 0.0, -0.06124, 0.00046, 0.0, -0.06124, -0.00046, 0.0, 0.05933, 0.0883, -0.0, 0.05933, -0.0883, -0.0, -0.05933, -0.0883, 0.0, -0.05933, 0.0883, 0.0], [-0.10708, 0.03232, -0.0, -0.0, 0.00449, -0.0, 0.10708, 0.03232, -0.0, -0.01703, -0.09032, 0.0, -0.01703, 0.09032, 0.0, 0.10708, -0.03232, 0.0, 0.0, -0.00449, 0.0, -0.10708, -0.03232, 0.0, 0.01703, 0.09032, -0.0, 0.01703, -0.09032, -0.0, -0.05317, -0.01367, 0.0, -0.05317, 0.01367, 0.0, 0.05317, 0.01367, -0.0, 0.05317, -0.01367, -0.0]]), "B2g":array([[0.0, -0.0, -0.14455, 0.0, 0.0, 0.0, 0.0, 0.0, 0.14455, 0.0, -0.0, -0.08132, -0.0, -0.0, -0.08132, -0.0, 0.0, 0.14455, 0.0, -0.0, -0.0, -0.0, -0.0, -0.14455, -0.0, 0.0, 0.08132, 0.0, 0.0, 0.08132, -0.0, 0.0, 0.02413, 0.0, 0.0, 0.02413, 0.0, -0.0, -0.02413, -0.0, -0.0, -0.02413], [0.0, -0.0, -0.05394, 0.0, -0.0, 0.0, 0.0, -0.0, 0.05394, -0.0, -0.0, 0.0536, 0.0, -0.0, 0.0536, -0.0, -0.0, 0.05394, 0.0, 0.0, -0.0, -0.0, 0.0, -0.05394, 0.0, 0.0, -0.0536, -0.0, 0.0, -0.0536, 0.0, -0.0, -0.14246, 0.0, -0.0, -0.14246, -0.0, 0.0, 0.14246, -0.0, 0.0, 0.14246]]), "B3g":array([[0.0, -0.0, -0.07133, -0.0, -0.0, 0.07655, -0.0, 0.0, -0.07133, 0.0, 0.0, -0.06051, 0.0, 0.0, 0.06051, 0.0, 0.0, 0.07133, -0.0, 0.0, -0.07655, 0.0, -0.0, 0.07133, -0.0, -0.0, 0.06051, -0.0, -0.0, -0.06051, 0.0, -0.0, 0.07989, 0.0, 0.0, -0.07989, -0.0, 0.0, -0.07989, -0.0, -0.0, 0.07989], [-0.0, -0.0, 0.00962, -0.0, -0.0, 0.10684, 0.0, 0.0, 0.00962, 0.0, -0.0, -0.08215, -0.0, -0.0, 0.08215, -0.0, 0.0, -0.00962, -0.0, 0.0, -0.10684, -0.0, -0.0, -0.00962, -0.0, 0.0, 0.08215, 0.0, 0.0, -0.08215, -0.0, 0.0, -0.10481, -0.0, -0.0, 0.10481, 0.0, -0.0, 0.10481, 0.0, 0.0, -0.10481]]), "Au":array([[0.0, -0.0, 0.1073, 0.0, 0.0, 0.0, 0.0, -0.0, -0.1073, 0.0, 0.0, -0.01955, -0.0, 0.0, 0.01955, 0.0, 0.0, 0.1073, 0.0, 0.0, -0.0, 0.0, 0.0, -0.1073, 0.0, 0.0, -0.01955, -0.0, 0.0, 0.01955, 0.0, -0.0, -0.12315, -0.0, 0.0, 0.12315, 0.0, -0.0, -0.12315, -0.0, -0.0, 0.12315], [0.0, 0.0, 0.00343, 0.0, -0.0, -0.0, 0.0, -0.0, -0.00343, 0.0, 0.0, -0.21022, -0.0, 0.0, 0.21022, 0.0, 0.0, 0.00343, 0.0, -0.0, 0.0, 0.0, 0.0, -0.00343, 0.0, 0.0, -0.21022, -0.0, 0.0, 0.21022, -0.0, -0.0, 0.03636, -0.0, 0.0, -0.03636, -0.0, 0.0, 0.03636, 0.0, -0.0, -0.03636]]), "B1u":array([[-0.0, -0.0, 0.07252, 0.0, -0.0, 0.10496, 0.0, -0.0, 0.07252, 0.0, 0.0, -0.07792, 0.0, 0.0, -0.07792, 0.0, -0.0, 0.07252, 0.0, -0.0, 0.10496, 0.0, -0.0, 0.07252, 0.0, 0.0, -0.07792, 0.0, 0.0, -0.07792, -0.0, 0.0, -0.04709, 0.0, 0.0, -0.04709, 0.0, 0.0, -0.04709, -0.0, 0.0, -0.04709], [0.0, 0.0, 0.02456, 0.0, 0.0, -0.07427, 0.0, 0.0, 0.02456, 0.0, 0.0, -0.08787, 0.0, 0.0, -0.08787, 0.0, 0.0, 0.02456, 0.0, 0.0, -0.07427, 0.0, 0.0, 0.02456, 0.0, 0.0, -0.08787, 0.0, 0.0, -0.08787, -0.0, -0.0, 0.10044, 0.0, -0.0, 0.10044, 0.0, -0.0, 0.10044, -0.0, -0.0, 0.10044]]), "B2u":array([[-0.00175, 0.00014, 0.0, -0.23783, 0.0, 0.0, -0.00175, -0.00014, 0.0, 0.07498, -0.00758, 0.0, 0.07498, 0.00758, -0.0, -0.00175, 0.00014, -0.0, -0.23783, 0.0, -0.0, -0.00175, -0.00014, -0.0, 0.07498, -0.00758, -0.0, 0.07498, 0.00758, 0.0, 0.04568, -0.02177, -0.0, 0.04568, 0.02177, 0.0, 0.04568, -0.02177, -0.0, 0.04568, 0.02177, 0.0], [-0.08887, -0.02745, 0.0, 0.03637, -0.0, 0.0, -0.08887, 0.02745, 0.0, 0.03573, 0.09237, 0.0, 0.03573, -0.09237, -0.0, -0.08887, -0.02745, -0.0, 0.03637, -0.0, -0.0, -0.08887, 0.02745, -0.0, 0.03573, 0.09237, 0.0, 0.03573, -0.09237, -0.0, 0.03494, -0.02749, -0.0, 0.03494, 0.02749, 0.0, 0.03494, -0.02749, -0.0, 0.03494, 0.02749, 0.0]]), "B3u":array([[-0.0349, -0.05974, -0.0, -0.0, -0.08278, 0.0, 0.0349, -0.05974, 0.0, 0.06392, 0.04955, 0.0, -0.06392, 0.04955, -0.0, -0.0349, -0.05974, -0.0, -0.0, -0.08278, -0.0, 0.0349, -0.05974, 0.0, 0.06392, 0.04955, 0.0, -0.06392, 0.04955, -0.0, 0.02927, 0.05156, 0.0, -0.02927, 0.05156, -0.0, 0.02927, 0.05156, 0.0, -0.02927, 0.05156, -0.0], [-0.05684, 0.09761, -0.0, -0.0, -0.10253, 0.0, 0.05684, 0.09761, 0.0, 0.02855, 0.00601, 0.0, -0.02855, 0.00601, 0.0, -0.05684, 0.09761, -0.0, 0.0, -0.10253, -0.0, 0.05684, 0.09761, 0.0, 0.02855, 0.00601, -0.0, -0.02855, 0.00601, -0.0, 0.00627, -0.0524, 0.0, -0.00627, -0.0524, 0.0, 0.00627, -0.0524, -0.0, -0.00627, -0.0524, -0.0]])}, database_path = 'cyclam_scsd_20220713.pkl', maxdist = 1.75)
model_objs_dict['cb6'] = scsd_model('cb6', array([['-3.448', '1.9907', '1.9407', 'C'], ['-3.9802', '3.1946', '0.0', 'C'], ['-4.7567', '1.8496', '0.0', 'C'], ['-3.448', '1.9907', '-1.9407', 'C'], ['-4.8951', '-0.0', '1.7551', 'C'], ['-4.8951', '-0.0', '-1.7551', 'C'], ['-3.448', '-1.9907', '1.9407', 'C'], ['-4.7567', '-1.8496', '0.0', 'C'], ['-3.9802', '-3.1946', '0.0', 'C'], ['-3.448', '-1.9907', '-1.9407', 'C'], ['-2.4476', '-4.2393', '1.7551', 'C'], ['-2.4476', '-4.2393', '-1.7551', 'C'], ['0.0', '-3.9814', '1.9407', 'C'], ['-0.7765', '-5.0443', '-0.0', 'C'], ['0.7765', '-5.0443', '-0.0', 'C'], ['0.0', '-3.9814', '-1.9407', 'C'], ['-2.4476', '4.2393', '-1.7551', 'C'], ['-2.4476', '4.2393', '1.7551', 'C'], ['3.448', '-1.9907', '-1.9407', 'C'], ['3.9802', '-3.1946', '-0.0', 'C'], ['4.7567', '-1.8496', '-0.0', 'C'], ['3.448', '-1.9907', '1.9407', 'C'], ['4.8951', '0.0', '-1.7551', 'C'], ['4.8951', '0.0', '1.7551', 'C'], ['3.448', '1.9907', '-1.9407', 'C'], ['4.7567', '1.8496', '-0.0', 'C'], ['3.9802', '3.1946', '-0.0', 'C'], ['3.448', '1.9907', '1.9407', 'C'], ['2.4476', '4.2393', '-1.7551', 'C'], ['2.4476', '4.2393', '1.7551', 'C'], ['-0.0', '3.9814', '-1.9407', 'C'], ['0.7765', '5.0443', '0.0', 'C'], ['-0.7765', '5.0443', '0.0', 'C'], ['-0.0', '3.9814', '1.9407', 'C'], ['2.4476', '-4.2393', '1.7551', 'C'], ['2.4476', '-4.2393', '-1.7551', 'C'], ['-3.2083', '3.1306', '1.2219', 'N'], ['-4.3153', '1.2131', '1.2219', 'N'], ['-4.3153', '1.2131', '-1.2219', 'N'], ['-3.2083', '3.1306', '-1.2219', 'N'], ['-4.3153', '-1.2131', '1.2219', 'N'], ['-3.2083', '-3.1306', '1.2219', 'N'], ['-3.2083', '-3.1306', '-1.2219', 'N'], ['-4.3153', '-1.2131', '-1.2219', 'N'], ['-1.107', '-4.3437', '1.2219', 'N'], ['1.107', '-4.3437', '1.2219', 'N'], ['1.107', '-4.3437', '-1.2219', 'N'], ['-1.107', '-4.3437', '-1.2219', 'N'], ['3.2083', '-3.1306', '-1.2219', 'N'], ['4.3153', '-1.2131', '-1.2219', 'N'], ['4.3153', '-1.2131', '1.2219', 'N'], ['3.2083', '-3.1306', '1.2219', 'N'], ['4.3153', '1.2131', '-1.2219', 'N'], ['3.2083', '3.1306', '-1.2219', 'N'], ['3.2083', '3.1306', '1.2219', 'N'], ['4.3153', '1.2131', '1.2219', 'N'], ['1.107', '4.3437', '-1.2219', 'N'], ['-1.107', '4.3437', '-1.2219', 'N'], ['-1.107', '4.3437', '1.2219', 'N'], ['1.107', '4.3437', '1.2219', 'N'], ['-2.9961', '1.7298', '3.0445', 'O'], ['-2.9961', '1.7298', '-3.0445', 'O'], ['-2.9961', '-1.7298', '3.0445', 'O'], ['-2.9961', '-1.7298', '-3.0445', 'O'], ['0.0', '-3.4596', '3.0445', 'O'], ['0.0', '-3.4596', '-3.0445', 'O'], ['2.9961', '-1.7298', '-3.0445', 'O'], ['2.9961', '-1.7298', '3.0445', 'O'], ['2.9961', '1.7298', '-3.0445', 'O'], ['2.9961', '1.7298', '3.0445', 'O'], ['-0.0', '3.4596', '-3.0445', 'O'], ['-0.0', '3.4596', '3.0445', 'O']]), 'D6h' , pca = {"A1g":array([[-0.01023, 0.00591, 0.00535, 0.01658, -0.00991, -0.0, 0.01688, -0.0094, -0.0, -0.01023, 0.00591, -0.00535, -0.00321, 0.0, -0.00191, -0.00321, 0.0, 0.00191, -0.01023, -0.00591, 0.00535, 0.01688, 0.0094, -0.0, 0.01658, 0.00991, -0.0, -0.01023, -0.00591, -0.00535, -0.0016, -0.00278, -0.00191, -0.0016, -0.00278, 0.00191, 0.0, -0.01181, 0.00535, 0.00029, 0.01932, 0.0, -0.00029, 0.01932, 0.0, -0.0, -0.01181, -0.00535, -0.0016, 0.00278, 0.00191, -0.0016, 0.00278, -0.00191, 0.01023, -0.00591, -0.00535, -0.01658, 0.00991, 0.0, -0.01688, 0.0094, 0.0, 0.01023, -0.00591, 0.00535, 0.00321, -0.0, 0.00191, 0.00321, -0.0, -0.00191, 0.01023, 0.00591, -0.00535, -0.01688, -0.0094, 0.0, -0.01658, -0.00991, 0.0, 0.01023, 0.00591, 0.00535, 0.0016, 0.00278, 0.00191, 0.0016, 0.00278, -0.00191, -0.0, 0.01181, -0.00535, -0.00029, -0.01932, -0.0, 0.00029, -0.01932, -0.0, 0.0, 0.01181, 0.00535, 0.0016, -0.00278, -0.00191, 0.0016, -0.00278, 0.00191, 0.00129, -0.0006, 0.00099, 0.00116, -0.00082, 0.00099, 0.00116, -0.00082, -0.00099, 0.00129, -0.0006, -0.00099, 0.00116, 0.00082, 0.00099, 0.00129, 0.0006, 0.00099, 0.00129, 0.0006, -0.00099, 0.00116, 0.00082, -0.00099, -0.00012, 0.00141, 0.00099, 0.00012, 0.00141, 0.00099, 0.00012, 0.00141, -0.00099, -0.00012, 0.00141, -0.00099, -0.00129, 0.0006, -0.00099, -0.00116, 0.00082, -0.00099, -0.00116, 0.00082, 0.00099, -0.00129, 0.0006, 0.00099, -0.00116, -0.00082, -0.00099, -0.00129, -0.0006, -0.00099, -0.00129, -0.0006, 0.00099, -0.00116, -0.00082, 0.00099, 0.00012, -0.00141, -0.00099, -0.00012, -0.00141, -0.00099, -0.00012, -0.00141, 0.00099, 0.00012, -0.00141, 0.00099, -0.02892, 0.0167, 0.0102, -0.02892, 0.0167, -0.0102, -0.02892, -0.0167, 0.0102, -0.02892, -0.0167, -0.0102, 0.0, -0.03339, 0.0102, 0.0, -0.03339, -0.0102, 0.02892, -0.0167, -0.0102, 0.02892, -0.0167, 0.0102, 0.02892, 0.0167, -0.0102, 0.02892, 0.0167, 0.0102, -0.0, 0.03339, -0.0102, -0.0, 0.03339, 0.0102], [0.00624, -0.0036, -0.00864, 0.02611, -0.01624, -0.0, 0.02712, -0.0145, -0.0, 0.00624, -0.0036, 0.00864, 0.00656, 0.0, -0.00346, 0.00656, 0.0, 0.00346, 0.00624, 0.0036, -0.00864, 0.02712, 0.0145, -0.0, 0.02611, 0.01624, -0.0, 0.00624, 0.0036, 0.00864, 0.00328, 0.00568, -0.00346, 0.00328, 0.00568, 0.00346, -0.0, 0.00721, -0.00864, 0.001, 0.03073, 0.0, -0.001, 0.03073, 0.0, -0.0, 0.00721, 0.00864, 0.00328, -0.00568, 0.00346, 0.00328, -0.00568, -0.00346, -0.00624, 0.0036, 0.00864, -0.02611, 0.01624, 0.0, -0.02712, 0.0145, 0.0, -0.00624, 0.0036, -0.00864, -0.00656, -0.0, 0.00346, -0.00656, -0.0, -0.00346, -0.00624, -0.0036, 0.00864, -0.02712, -0.0145, 0.0, -0.02611, -0.01624, 0.0, -0.00624, -0.0036, -0.00864, -0.00328, -0.00568, 0.00346, -0.00328, -0.00568, -0.00346, 0.0, -0.00721, 0.00864, -0.001, -0.03073, -0.0, 0.001, -0.03073, -0.0, 0.0, -0.00721, -0.00864, -0.00328, 0.00568, -0.00346, -0.00328, 0.00568, 0.00346, 0.00522, -0.00524, -0.00212, 0.00714, -0.0019, -0.00212, 0.00714, -0.0019, 0.00212, 0.00522, -0.00524, 0.00212, 0.00714, 0.0019, -0.00212, 0.00522, 0.00524, -0.00212, 0.00522, 0.00524, 0.00212, 0.00714, 0.0019, 0.00212, 0.00193, 0.00714, -0.00212, -0.00193, 0.00714, -0.00212, -0.00193, 0.00714, 0.00212, 0.00193, 0.00714, 0.00212, -0.00522, 0.00524, 0.00212, -0.00714, 0.0019, 0.00212, -0.00714, 0.0019, -0.00212, -0.00522, 0.00524, -0.00212, -0.00714, -0.0019, 0.00212, -0.00522, -0.00524, 0.00212, -0.00522, -0.00524, -0.00212, -0.00714, -0.0019, -0.00212, -0.00193, -0.00714, 0.00212, 0.00193, -0.00714, 0.00212, 0.00193, -0.00714, -0.00212, -0.00193, -0.00714, -0.00212, 0.00737, -0.00426, -0.01761, 0.00737, -0.00426, 0.01761, 0.00737, 0.00426, -0.01761, 0.00737, 0.00426, 0.01761, 0.0, 0.00851, -0.01761, 0.0, 0.00851, 0.01761, -0.00737, 0.00426, 0.01761, -0.00737, 0.00426, -0.01761, -0.00737, -0.00426, 0.01761, -0.00737, -0.00426, -0.01761, -0.0, -0.00851, 0.01761, -0.0, -0.00851, -0.01761]]), "A2g":array([[-0.00252, -0.00436, 0.0, 0.02174, 0.01183, 0.0, -0.00062, 0.02474, 0.0, -0.00252, -0.00436, -0.0, -0.0, 0.0007, 0.0, -0.0, 0.0007, -0.0, 0.00252, -0.00436, 0.0, 0.00062, 0.02474, 0.0, -0.02174, 0.01183, 0.0, 0.00252, -0.00436, -0.0, -0.00061, 0.00035, -0.0, -0.00061, 0.00035, 0.0, 0.00504, -0.0, 0.0, -0.02112, 0.01291, -0.0, -0.02112, -0.01291, -0.0, 0.00504, -0.0, -0.0, 0.00061, 0.00035, -0.0, 0.00061, 0.00035, 0.0, 0.00252, 0.00436, -0.0, -0.02174, -0.01183, -0.0, 0.00062, -0.02474, -0.0, 0.00252, 0.00436, 0.0, 0.0, -0.0007, -0.0, 0.0, -0.0007, 0.0, -0.00252, 0.00436, -0.0, -0.00062, -0.02474, -0.0, 0.02174, -0.01183, -0.0, -0.00252, 0.00436, 0.0, 0.00061, -0.00035, 0.0, 0.00061, -0.00035, -0.0, -0.00504, 0.0, -0.0, 0.02112, -0.01291, 0.0, 0.02112, 0.01291, 0.0, -0.00504, 0.0, 0.0, -0.00061, -0.00035, 0.0, -0.00061, -0.00035, -0.0, 0.00794, -0.00261, -0.00165, -0.00623, 0.00557, 0.00165, -0.00623, 0.00557, -0.00165, 0.00794, -0.00261, 0.00165, 0.00623, 0.00557, -0.00165, -0.00794, -0.00261, 0.00165, -0.00794, -0.00261, -0.00165, 0.00623, 0.00557, 0.00165, -0.00171, 0.00818, -0.00165, -0.00171, -0.00818, 0.00165, -0.00171, -0.00818, -0.00165, -0.00171, 0.00818, 0.00165, -0.00794, 0.00261, 0.00165, 0.00623, -0.00557, -0.00165, 0.00623, -0.00557, 0.00165, -0.00794, 0.00261, -0.00165, -0.00623, -0.00557, 0.00165, 0.00794, 0.00261, -0.00165, 0.00794, 0.00261, 0.00165, -0.00623, -0.00557, -0.00165, 0.00171, -0.00818, 0.00165, 0.00171, 0.00818, -0.00165, 0.00171, 0.00818, 0.00165, 0.00171, -0.00818, -0.00165, -0.01921, -0.03327, 0.0, -0.01921, -0.03327, -0.0, 0.01921, -0.03327, 0.0, 0.01921, -0.03327, -0.0, 0.03842, -0.0, 0.0, 0.03842, -0.0, -0.0, 0.01921, 0.03327, -0.0, 0.01921, 0.03327, 0.0, -0.01921, 0.03327, -0.0, -0.01921, 0.03327, 0.0, -0.03842, 0.0, -0.0, -0.03842, 0.0, 0.0], [-0.0008, -0.00138, 0.0, 0.00191, -0.01489, -0.0, -0.01385, -0.00579, -0.0, -0.0008, -0.00138, -0.0, 0.0, -0.00158, 0.0, 0.0, -0.00158, -0.0, 0.0008, -0.00138, 0.0, 0.01385, -0.00579, -0.0, -0.00191, -0.01489, -0.0, 0.0008, -0.00138, -0.0, 0.00137, -0.00079, 0.0, 0.00137, -0.00079, -0.0, 0.00159, -0.0, -0.0, 0.01194, 0.0091, 0.0, 0.01194, -0.0091, 0.0, 0.00159, -0.0, 0.0, -0.00137, -0.00079, -0.0, -0.00137, -0.00079, 0.0, 0.0008, 0.00138, -0.0, -0.00191, 0.01489, 0.0, 0.01385, 0.00579, 0.0, 0.0008, 0.00138, 0.0, -0.0, 0.00158, -0.0, -0.0, 0.00158, 0.0, -0.0008, 0.00138, -0.0, -0.01385, 0.00579, 0.0, 0.00191, 0.01489, 0.0, -0.0008, 0.00138, 0.0, -0.00137, 0.00079, -0.0, -0.00137, 0.00079, 0.0, -0.00159, 0.0, 0.0, -0.01194, -0.0091, -0.0, -0.01194, 0.0091, -0.0, -0.00159, 0.0, -0.0, 0.00137, 0.00079, 0.0, 0.00137, 0.00079, -0.0, 0.02126, -0.01446, -0.00425, -0.02315, 0.01118, 0.00425, -0.02315, 0.01118, -0.00425, 0.02126, -0.01446, 0.00425, 0.02315, 0.01118, -0.00425, -0.02126, -0.01446, 0.00425, -0.02126, -0.01446, -0.00425, 0.02315, 0.01118, 0.00425, 0.0019, 0.02564, -0.00425, 0.0019, -0.02564, 0.00425, 0.0019, -0.02564, -0.00425, 0.0019, 0.02564, 0.00425, -0.02126, 0.01446, 0.00425, 0.02315, -0.01118, -0.00425, 0.02315, -0.01118, 0.00425, -0.02126, 0.01446, -0.00425, -0.02315, -0.01118, 0.00425, 0.02126, 0.01446, -0.00425, 0.02126, 0.01446, 0.00425, -0.02315, -0.01118, -0.00425, -0.0019, -0.02564, 0.00425, -0.0019, 0.02564, -0.00425, -0.0019, 0.02564, 0.00425, -0.0019, -0.02564, -0.00425, 0.0039, 0.00675, -0.0, 0.0039, 0.00675, 0.0, -0.0039, 0.00675, 0.0, -0.0039, 0.00675, -0.0, -0.00779, 0.0, 0.0, -0.00779, 0.0, -0.0, -0.0039, -0.00675, 0.0, -0.0039, -0.00675, -0.0, 0.0039, -0.00675, -0.0, 0.0039, -0.00675, 0.0, 0.00779, -0.0, -0.0, 0.00779, -0.0, 0.0]]), "B1g":array([[-0.01272, 0.00734, 0.01019, -0.0, 0.0, 0.0015, -0.0, -0.0, 0.0015, 0.01272, -0.00734, 0.01019, 0.0, -0.00153, -0.0, 0.0, 0.00153, 0.0, 0.01272, 0.00734, -0.01019, -0.0, -0.0, -0.0015, -0.0, -0.0, -0.0015, -0.01272, -0.00734, -0.01019, -0.00133, 0.00077, -0.0, 0.00133, -0.00077, 0.0, -0.0, -0.01469, 0.01019, -0.0, -0.0, 0.0015, 0.0, -0.0, 0.0015, 0.0, 0.01469, 0.01019, -0.00133, -0.00077, 0.0, 0.00133, 0.00077, -0.0, 0.01272, -0.00734, -0.01019, 0.0, -0.0, -0.0015, 0.0, 0.0, -0.0015, -0.01272, 0.00734, -0.01019, -0.0, 0.00153, 0.0, -0.0, -0.00153, -0.0, -0.01272, -0.00734, 0.01019, 0.0, 0.0, 0.0015, 0.0, 0.0, 0.0015, 0.01272, 0.00734, 0.01019, 0.00133, -0.00077, 0.0, -0.00133, 0.00077, -0.0, 0.0, 0.01469, -0.01019, 0.0, 0.0, -0.0015, -0.0, 0.0, -0.0015, -0.0, -0.01469, -0.01019, 0.00133, 0.00077, -0.0, -0.00133, -0.00077, 0.0, -0.01009, 0.0053, 0.00815, -0.00963, 0.00609, 0.00815, 0.00963, -0.00609, 0.00815, 0.01009, -0.0053, 0.00815, 0.00963, 0.00609, -0.00815, 0.01009, 0.0053, -0.00815, -0.01009, -0.0053, -0.00815, -0.00963, -0.00609, -0.00815, 0.00046, -0.01139, 0.00815, -0.00046, -0.01139, 0.00815, 0.00046, 0.01139, 0.00815, -0.00046, 0.01139, 0.00815, 0.01009, -0.0053, -0.00815, 0.00963, -0.00609, -0.00815, -0.00963, 0.00609, -0.00815, -0.01009, 0.0053, -0.00815, -0.00963, -0.00609, 0.00815, -0.01009, -0.0053, 0.00815, 0.01009, 0.0053, 0.00815, 0.00963, 0.00609, 0.00815, -0.00046, 0.01139, -0.00815, 0.00046, 0.01139, -0.00815, -0.00046, -0.01139, -0.00815, 0.00046, -0.01139, -0.00815, -0.01909, 0.01102, 0.01387, 0.01909, -0.01102, 0.01387, 0.01909, 0.01102, -0.01387, -0.01909, -0.01102, -0.01387, -0.0, -0.02204, 0.01387, -0.0, 0.02204, 0.01387, 0.01909, -0.01102, -0.01387, -0.01909, 0.01102, -0.01387, -0.01909, -0.01102, 0.01387, 0.01909, 0.01102, 0.01387, 0.0, 0.02204, -0.01387, 0.0, -0.02204, -0.01387], [0.00782, -0.00451, 0.01417, 0.0, -0.0, 0.01804, 0.0, 0.0, 0.01804, -0.00782, 0.00451, 0.01417, 0.0, -0.01043, -0.0, 0.0, 0.01043, 0.0, -0.00782, -0.00451, -0.01417, 0.0, 0.0, -0.01804, 0.0, 0.0, -0.01804, 0.00782, 0.00451, -0.01417, -0.00903, 0.00521, -0.0, 0.00903, -0.00521, -0.0, 0.0, 0.00903, 0.01417, 0.0, 0.0, 0.01804, -0.0, 0.0, 0.01804, 0.0, -0.00903, 0.01417, -0.00903, -0.00521, 0.0, 0.00903, 0.00521, -0.0, -0.00782, 0.00451, -0.01417, -0.0, 0.0, -0.01804, -0.0, -0.0, -0.01804, 0.00782, -0.00451, -0.01417, -0.0, 0.01043, 0.0, -0.0, -0.01043, -0.0, 0.00782, 0.00451, 0.01417, -0.0, -0.0, 0.01804, -0.0, -0.0, 0.01804, -0.00782, -0.00451, 0.01417, 0.00903, -0.00521, 0.0, -0.00903, 0.00521, 0.0, -0.0, -0.00903, -0.01417, -0.0, -0.0, -0.01804, 0.0, -0.0, -0.01804, -0.0, 0.00903, -0.01417, 0.00903, 0.00521, -0.0, -0.00903, -0.00521, 0.0, 0.00299, -0.00056, 0.01529, 0.00198, -0.00231, 0.01529, -0.00198, 0.00231, 0.01529, -0.00299, 0.00056, 0.01529, -0.00198, -0.00231, -0.01529, -0.00299, -0.00056, -0.01529, 0.00299, 0.00056, -0.01529, 0.00198, 0.00231, -0.01529, -0.00101, 0.00287, 0.01529, 0.00101, 0.00287, 0.01529, -0.00101, -0.00287, 0.01529, 0.00101, -0.00287, 0.01529, -0.00299, 0.00056, -0.01529, -0.00198, 0.00231, -0.01529, 0.00198, -0.00231, -0.01529, 0.00299, -0.00056, -0.01529, 0.00198, 0.00231, 0.01529, 0.00299, 0.00056, 0.01529, -0.00299, -0.00056, 0.01529, -0.00198, -0.00231, 0.01529, 0.00101, -0.00287, -0.01529, -0.00101, -0.00287, -0.01529, 0.00101, 0.00287, -0.01529, -0.00101, 0.00287, -0.01529, 0.01387, -0.00801, 0.00832, -0.01387, 0.00801, 0.00832, -0.01387, -0.00801, -0.00832, 0.01387, 0.00801, -0.00832, -0.0, 0.01602, 0.00832, 0.0, -0.01602, 0.00832, -0.01387, 0.00801, -0.00832, 0.01387, -0.00801, -0.00832, 0.01387, 0.00801, 0.00832, -0.01387, -0.00801, 0.00832, 0.0, -0.01602, -0.00832, -0.0, 0.01602, -0.00832]]), "B2g":array([[-0.00646, -0.01118, 0.0, 0.0, -0.0, 0.00481, 0.0, 0.0, -0.00481, 0.00646, 0.01118, 0.0, -0.02144, -0.0, -0.00641, 0.02144, 0.0, -0.00641, -0.00646, 0.01118, -0.0, 0.0, -0.0, -0.00481, 0.0, 0.0, 0.00481, 0.00646, -0.01118, 0.0, 0.01072, 0.01856, 0.00641, -0.01072, -0.01856, 0.00641, 0.01291, 0.0, -0.0, 0.0, 0.0, 0.00481, -0.0, 0.0, -0.00481, -0.01291, 0.0, 0.0, -0.01072, 0.01856, 0.00641, 0.01072, -0.01856, 0.00641, 0.00646, 0.01118, -0.0, -0.0, 0.0, -0.00481, -0.0, -0.0, 0.00481, -0.00646, -0.01118, -0.0, 0.02144, 0.0, 0.00641, -0.02144, -0.0, 0.00641, 0.00646, -0.01118, 0.0, -0.0, 0.0, 0.00481, -0.0, -0.0, -0.00481, -0.00646, 0.01118, -0.0, -0.01072, -0.01856, -0.00641, 0.01072, 0.01856, -0.00641, -0.01291, -0.0, 0.0, -0.0, -0.0, -0.00481, 0.0, -0.0, 0.00481, 0.01291, -0.0, -0.0, 0.01072, -0.01856, -0.00641, -0.01072, 0.01856, -0.00641, 0.00833, -0.0143, -0.00027, -0.01655, 7e-05, 0.00027, 0.01655, -7e-05, 0.00027, -0.00833, 0.0143, -0.00027, -0.01655, -7e-05, 0.00027, 0.00833, 0.0143, -0.00027, -0.00833, -0.0143, -0.00027, 0.01655, 7e-05, 0.00027, 0.00822, 0.01436, -0.00027, 0.00822, -0.01436, 0.00027, -0.00822, 0.01436, 0.00027, -0.00822, -0.01436, -0.00027, -0.00833, 0.0143, 0.00027, 0.01655, -7e-05, -0.00027, -0.01655, 7e-05, -0.00027, 0.00833, -0.0143, 0.00027, 0.01655, 7e-05, -0.00027, -0.00833, -0.0143, 0.00027, 0.00833, 0.0143, 0.00027, -0.01655, -7e-05, -0.00027, -0.00822, -0.01436, 0.00027, -0.00822, 0.01436, -0.00027, 0.00822, -0.01436, -0.00027, 0.00822, 0.01436, 0.00027, -0.00996, -0.01725, -0.0, 0.00996, 0.01725, 0.0, -0.00996, 0.01725, -0.0, 0.00996, -0.01725, 0.0, 0.01991, -0.0, -0.0, -0.01991, -0.0, 0.0, 0.00996, 0.01725, 0.0, -0.00996, -0.01725, -0.0, 0.00996, -0.01725, 0.0, -0.00996, 0.01725, -0.0, -0.01991, 0.0, 0.0, 0.01991, 0.0, -0.0], [0.00112, 0.00194, 0.0, -0.0, -0.0, 0.00152, 0.0, 0.0, -0.00152, -0.00112, -0.00194, -0.0, 0.00503, 0.0, -0.02527, -0.00503, -0.0, -0.02527, 0.00112, -0.00194, 0.0, 0.0, -0.0, -0.00152, -0.0, -0.0, 0.00152, -0.00112, 0.00194, -0.0, -0.00252, -0.00436, 0.02527, 0.00252, 0.00436, 0.02527, -0.00224, -0.0, 0.0, 0.0, 0.0, 0.00152, -0.0, 0.0, -0.00152, 0.00224, -0.0, -0.0, 0.00252, -0.00436, 0.02527, -0.00252, 0.00436, 0.02527, -0.00112, -0.00194, -0.0, 0.0, 0.0, -0.00152, -0.0, -0.0, 0.00152, 0.00112, 0.00194, 0.0, -0.00503, -0.0, 0.02527, 0.00503, 0.0, 0.02527, -0.00112, 0.00194, -0.0, -0.0, 0.0, 0.00152, 0.0, 0.0, -0.00152, 0.00112, -0.00194, 0.0, 0.00252, 0.00436, -0.02527, -0.00252, -0.00436, -0.02527, 0.00224, 0.0, -0.0, -0.0, -0.0, -0.00152, 0.0, -0.0, 0.00152, -0.00224, 0.0, 0.0, -0.00252, 0.00436, -0.02527, 0.00252, -0.00436, -0.02527, -0.00701, 0.00778, 0.01464, 0.01024, -0.00217, -0.01464, -0.01024, 0.00217, -0.01464, 0.00701, -0.00778, 0.01464, 0.01024, 0.00217, -0.01464, -0.00701, -0.00778, 0.01464, 0.00701, 0.00778, 0.01464, -0.01024, -0.00217, -0.01464, -0.00324, -0.00996, 0.01464, -0.00324, 0.00996, -0.01464, 0.00324, -0.00996, -0.01464, 0.00324, 0.00996, 0.01464, 0.00701, -0.00778, -0.01464, -0.01024, 0.00217, 0.01464, 0.01024, -0.00217, 0.01464, -0.00701, 0.00778, -0.01464, -0.01024, -0.00217, 0.01464, 0.00701, 0.00778, -0.01464, -0.00701, -0.00778, -0.01464, 0.01024, 0.00217, 0.01464, 0.00324, 0.00996, -0.01464, 0.00324, -0.00996, 0.01464, -0.00324, 0.00996, 0.01464, -0.00324, -0.00996, -0.01464, -0.0079, -0.01369, -0.0, 0.0079, 0.01369, 0.0, -0.0079, 0.01369, -0.0, 0.0079, -0.01369, 0.0, 0.0158, -0.0, -0.0, -0.0158, -0.0, 0.0, 0.0079, 0.01369, 0.0, -0.0079, -0.01369, -0.0, 0.0079, -0.01369, 0.0, -0.0079, 0.01369, -0.0, -0.0158, 0.0, 0.0, 0.0158, 0.0, -0.0]]), "A1u":array([[-0.00103, -0.00179, 0.0, -0.0, -0.0, -2e-05, 0.0, 0.0, 2e-05, 0.00103, 0.00179, 0.0, 0.0, -5e-05, -0.0, 0.0, 5e-05, 0.0, 0.00103, -0.00179, -0.0, 0.0, -0.0, -2e-05, -0.0, 0.0, 2e-05, -0.00103, 0.00179, 0.0, 4e-05, -3e-05, -0.0, -4e-05, 3e-05, 0.0, 0.00206, -0.0, 0.0, 0.0, 0.0, -2e-05, -0.0, 0.0, 2e-05, -0.00206, 0.0, 0.0, 4e-05, 3e-05, 0.0, -4e-05, -3e-05, -0.0, -0.00103, -0.00179, -0.0, 0.0, 0.0, -2e-05, -0.0, -0.0, 2e-05, 0.00103, 0.00179, -0.0, -0.0, -5e-05, 0.0, -0.0, 5e-05, -0.0, 0.00103, -0.00179, 0.0, -0.0, 0.0, -2e-05, 0.0, -0.0, 2e-05, -0.00103, 0.00179, -0.0, 4e-05, -3e-05, 0.0, -4e-05, 3e-05, -0.0, 0.00206, 0.0, -0.0, -0.0, -0.0, -2e-05, 0.0, -0.0, 2e-05, -0.00206, 0.0, -0.0, 4e-05, 3e-05, -0.0, -4e-05, -3e-05, 0.0, -0.02102, 0.00945, 0.0118, 0.01869, -0.01348, -0.0118, -0.01869, 0.01348, -0.0118, 0.02102, -0.00945, 0.0118, -0.01869, -0.01348, 0.0118, 0.02102, 0.00945, -0.0118, -0.02102, -0.00945, -0.0118, 0.01869, 0.01348, 0.0118, 0.00232, -0.02293, 0.0118, 0.00232, 0.02293, -0.0118, -0.00232, -0.02293, -0.0118, -0.00232, 0.02293, 0.0118, -0.02102, 0.00945, 0.0118, 0.01869, -0.01348, -0.0118, -0.01869, 0.01348, -0.0118, 0.02102, -0.00945, 0.0118, -0.01869, -0.01348, 0.0118, 0.02102, 0.00945, -0.0118, -0.02102, -0.00945, -0.0118, 0.01869, 0.01348, 0.0118, 0.00232, -0.02293, 0.0118, 0.00232, 0.02293, -0.0118, -0.00232, -0.02293, -0.0118, -0.00232, 0.02293, 0.0118, -0.00199, -0.00345, 0.0, 0.00199, 0.00345, -0.0, 0.00199, -0.00345, 0.0, -0.00199, 0.00345, -0.0, 0.00398, 0.0, 0.0, -0.00398, -0.0, -0.0, -0.00199, -0.00345, -0.0, 0.00199, 0.00345, 0.0, 0.00199, -0.00345, -0.0, -0.00199, 0.00345, 0.0, 0.00398, -0.0, -0.0, -0.00398, 0.0, 0.0], [0.00938, 0.01624, -0.0, -0.0, 0.0, -0.00283, -0.0, 0.0, 0.00283, -0.00938, -0.01624, 0.0, 0.0, 0.01565, -0.0, 0.0, -0.01565, 0.0, -0.00938, 0.01624, -0.0, -0.0, 0.0, -0.00283, -0.0, -0.0, 0.00283, 0.00938, -0.01624, 0.0, -0.01355, 0.00782, -0.0, 0.01355, -0.00782, 0.0, -0.01876, 0.0, -0.0, -0.0, -0.0, -0.00283, 0.0, -0.0, 0.00283, 0.01876, 0.0, 0.0, -0.01355, -0.00782, 0.0, 0.01355, 0.00782, -0.0, 0.00938, 0.01624, 0.0, 0.0, -0.0, -0.00283, 0.0, -0.0, 0.00283, -0.00938, -0.01624, -0.0, -0.0, 0.01565, 0.0, -0.0, -0.01565, -0.0, -0.00938, 0.01624, 0.0, 0.0, -0.0, -0.00283, 0.0, 0.0, 0.00283, 0.00938, -0.01624, -0.0, -0.01355, 0.00782, 0.0, 0.01355, -0.00782, -0.0, -0.01876, -0.0, 0.0, 0.0, 0.0, -0.00283, -0.0, 0.0, 0.00283, 0.01876, -0.0, -0.0, -0.01355, -0.00782, -0.0, 0.01355, 0.00782, 0.0, 0.00336, 0.00976, 0.00451, 0.00677, 0.00779, -0.00451, -0.00677, -0.00779, -0.00451, -0.00336, -0.00976, 0.00451, -0.00677, 0.00779, 0.00451, -0.00336, 0.00976, -0.00451, 0.00336, -0.00976, -0.00451, 0.00677, -0.00779, 0.00451, -0.01013, -0.00197, 0.00451, -0.01013, 0.00197, -0.00451, 0.01013, -0.00197, -0.00451, 0.01013, 0.00197, 0.00451, 0.00336, 0.00976, 0.00451, 0.00677, 0.00779, -0.00451, -0.00677, -0.00779, -0.00451, -0.00336, -0.00976, 0.00451, -0.00677, 0.00779, 0.00451, -0.00336, 0.00976, -0.00451, 0.00336, -0.00976, -0.00451, 0.00677, -0.00779, 0.00451, -0.01013, -0.00197, 0.00451, -0.01013, 0.00197, -0.00451, 0.01013, -0.00197, -0.00451, 0.01013, 0.00197, 0.00451, 0.01386, 0.02401, -0.0, -0.01386, -0.02401, 0.0, -0.01386, 0.02401, -0.0, 0.01386, -0.02401, 0.0, -0.02772, 0.0, -0.0, 0.02772, 0.0, 0.0, 0.01386, 0.02401, 0.0, -0.01386, -0.02401, -0.0, -0.01386, 0.02401, 0.0, 0.01386, -0.02401, -0.0, -0.02772, -0.0, 0.0, 0.02772, -0.0, -0.0]]), "A2u":array([[0.01391, -0.00803, -0.00406, 0.0, -0.0, 0.00524, 0.0, -0.0, 0.00524, -0.01391, 0.00803, -0.00406, 0.00592, -0.0, 0.00476, -0.00592, -0.0, 0.00476, 0.01391, 0.00803, -0.00406, 0.0, 0.0, 0.00524, 0.0, 0.0, 0.00524, -0.01391, -0.00803, -0.00406, 0.00296, 0.00513, 0.00476, -0.00296, -0.00513, 0.00476, -0.0, 0.01606, -0.00406, -0.0, 0.0, 0.00524, 0.0, 0.0, 0.00524, -0.0, -0.01606, -0.00406, -0.00296, 0.00513, 0.00476, 0.00296, -0.00513, 0.00476, 0.01391, -0.00803, -0.00406, -0.0, 0.0, 0.00524, -0.0, 0.0, 0.00524, -0.01391, 0.00803, -0.00406, 0.00592, 0.0, 0.00476, -0.00592, 0.0, 0.00476, 0.01391, 0.00803, -0.00406, -0.0, -0.0, 0.00524, -0.0, -0.0, 0.00524, -0.01391, -0.00803, -0.00406, 0.00296, 0.00513, 0.00476, -0.00296, -0.00513, 0.00476, -0.0, 0.01606, -0.00406, 0.0, -0.0, 0.00524, -0.0, -0.0, 0.00524, 0.0, -0.01606, -0.00406, -0.00296, 0.00513, 0.00476, 0.00296, -0.00513, 0.00476, 0.00307, -0.00326, 0.00277, 0.00436, -0.00102, 0.00277, -0.00436, 0.00102, 0.00277, -0.00307, 0.00326, 0.00277, 0.00436, 0.00102, 0.00277, 0.00307, 0.00326, 0.00277, -0.00307, -0.00326, 0.00277, -0.00436, -0.00102, 0.00277, 0.00129, 0.00429, 0.00277, -0.00129, 0.00429, 0.00277, 0.00129, -0.00429, 0.00277, -0.00129, -0.00429, 0.00277, 0.00307, -0.00326, 0.00277, 0.00436, -0.00102, 0.00277, -0.00436, 0.00102, 0.00277, -0.00307, 0.00326, 0.00277, 0.00436, 0.00102, 0.00277, 0.00307, 0.00326, 0.00277, -0.00307, -0.00326, 0.00277, -0.00436, -0.00102, 0.00277, 0.00129, 0.00429, 0.00277, -0.00129, 0.00429, 0.00277, 0.00129, -0.00429, 0.00277, -0.00129, -0.00429, 0.00277, 0.02959, -0.01708, -0.01148, -0.02959, 0.01708, -0.01148, 0.02959, 0.01708, -0.01148, -0.02959, -0.01708, -0.01148, -0.0, 0.03416, -0.01148, -0.0, -0.03416, -0.01148, 0.02959, -0.01708, -0.01148, -0.02959, 0.01708, -0.01148, 0.02959, 0.01708, -0.01148, -0.02959, -0.01708, -0.01148, -0.0, 0.03416, -0.01148, 0.0, -0.03416, -0.01148], [0.00778, -0.00449, 0.00241, -0.0, 0.0, -0.00821, -0.0, 0.0, -0.00821, -0.00778, 0.00449, 0.00241, 0.02336, -0.0, -0.00337, -0.02336, -0.0, -0.00337, 0.00778, 0.00449, 0.00241, -0.0, -0.0, -0.00821, -0.0, -0.0, -0.00821, -0.00778, -0.00449, 0.00241, 0.01168, 0.02023, -0.00337, -0.01168, -0.02023, -0.00337, 0.0, 0.00898, 0.00241, -0.0, -0.0, -0.00821, 0.0, -0.0, -0.00821, 0.0, -0.00898, 0.00241, -0.01168, 0.02023, -0.00337, 0.01168, -0.02023, -0.00337, 0.00778, -0.00449, 0.00241, 0.0, -0.0, -0.00821, 0.0, -0.0, -0.00821, -0.00778, 0.00449, 0.00241, 0.02336, 0.0, -0.00337, -0.02336, 0.0, -0.00337, 0.00778, 0.00449, 0.00241, 0.0, 0.0, -0.00821, 0.0, 0.0, -0.00821, -0.00778, -0.00449, 0.00241, 0.01168, 0.02023, -0.00337, -0.01168, -0.02023, -0.00337, -0.0, 0.00898, 0.00241, 0.0, 0.0, -0.00821, -0.0, 0.0, -0.00821, 0.0, -0.00898, 0.00241, -0.01168, 0.02023, -0.00337, 0.01168, -0.02023, -0.00337, 0.00865, -0.01135, -0.00377, 0.01415, -0.00182, -0.00377, -0.01415, 0.00182, -0.00377, -0.00865, 0.01135, -0.00377, 0.01415, 0.00182, -0.00377, 0.00865, 0.01135, -0.00377, -0.00865, -0.01135, -0.00377, -0.01415, -0.00182, -0.00377, 0.0055, 0.01316, -0.00377, -0.0055, 0.01316, -0.00377, 0.0055, -0.01316, -0.00377, -0.0055, -0.01316, -0.00377, 0.00865, -0.01135, -0.00377, 0.01415, -0.00182, -0.00377, -0.01415, 0.00182, -0.00377, -0.00865, 0.01135, -0.00377, 0.01415, 0.00182, -0.00377, 0.00865, 0.01135, -0.00377, -0.00865, -0.01135, -0.00377, -0.01415, -0.00182, -0.00377, 0.0055, 0.01316, -0.00377, -0.0055, 0.01316, -0.00377, 0.0055, -0.01316, -0.00377, -0.0055, -0.01316, -0.00377, -0.00325, 0.00188, 0.01671, 0.00325, -0.00188, 0.01671, -0.00325, -0.00188, 0.01671, 0.00325, 0.00188, 0.01671, 0.0, -0.00375, 0.01671, 0.0, 0.00375, 0.01671, -0.00325, 0.00188, 0.01671, 0.00325, -0.00188, 0.01671, -0.00325, -0.00188, 0.01671, 0.00325, 0.00188, 0.01671, -0.0, -0.00375, 0.01671, -0.0, 0.00375, 0.01671]]), "B1u":array([[-0.00656, -0.01136, -0.0, 0.00843, -0.00465, -0.0, -0.00824, 0.00497, 0.0, -0.00656, -0.01136, 0.0, -0.02151, -0.0, -0.00545, -0.02151, -0.0, 0.00545, -0.00656, 0.01136, -0.0, -0.00824, -0.00497, 0.0, 0.00843, 0.00465, -0.0, -0.00656, 0.01136, 0.0, 0.01075, 0.01863, 0.00545, 0.01075, 0.01863, -0.00545, 0.01312, -0.0, -0.0, -0.00019, 0.00962, 0.0, -0.00019, -0.00962, -0.0, 0.01312, -0.0, 0.0, 0.01075, -0.01863, -0.00545, 0.01075, -0.01863, 0.00545, -0.00656, -0.01136, 0.0, 0.00843, -0.00465, -0.0, -0.00824, 0.00497, 0.0, -0.00656, -0.01136, -0.0, -0.02151, -0.0, -0.00545, -0.02151, -0.0, 0.00545, -0.00656, 0.01136, 0.0, -0.00824, -0.00497, 0.0, 0.00843, 0.00465, -0.0, -0.00656, 0.01136, -0.0, 0.01075, 0.01863, 0.00545, 0.01075, 0.01863, -0.00545, 0.01312, 0.0, 0.0, -0.00019, 0.00962, 0.0, -0.00019, -0.00962, -0.0, 0.01312, 0.0, -0.0, 0.01075, -0.01863, -0.00545, 0.01075, -0.01863, 0.00545, 0.00824, -0.01442, -9e-05, -0.01661, -7e-05, 9e-05, -0.01661, -7e-05, -9e-05, 0.00824, -0.01442, 9e-05, -0.01661, 7e-05, 9e-05, 0.00824, 0.01442, -9e-05, 0.00824, 0.01442, 9e-05, -0.01661, 7e-05, -9e-05, 0.00836, 0.01434, -9e-05, 0.00836, -0.01434, 9e-05, 0.00836, -0.01434, -9e-05, 0.00836, 0.01434, 9e-05, 0.00824, -0.01442, -9e-05, -0.01661, -7e-05, 9e-05, -0.01661, -7e-05, -9e-05, 0.00824, -0.01442, 9e-05, -0.01661, 7e-05, 9e-05, 0.00824, 0.01442, -9e-05, 0.00824, 0.01442, 9e-05, -0.01661, 7e-05, -9e-05, 0.00836, 0.01434, -9e-05, 0.00836, -0.01434, 9e-05, 0.00836, -0.01434, -9e-05, 0.00836, 0.01434, 9e-05, -0.01004, -0.01739, -0.0, -0.01004, -0.01739, -0.0, -0.01004, 0.01739, 0.0, -0.01004, 0.01739, -0.0, 0.02009, -0.0, 0.0, 0.02009, -0.0, -0.0, -0.01004, -0.01739, -0.0, -0.01004, -0.01739, 0.0, -0.01004, 0.01739, 0.0, -0.01004, 0.01739, 0.0, 0.02009, 0.0, -0.0, 0.02009, 0.0, -0.0], [0.0012, 0.00207, -0.0, 0.00359, 0.01566, -0.0, 0.01177, 0.01094, -0.0, 0.0012, 0.00207, 0.0, -0.01265, 0.0, 0.01216, -0.01265, 0.0, -0.01216, 0.0012, -0.00207, -0.0, 0.01177, -0.01094, -0.0, 0.00359, -0.01566, -0.0, 0.0012, -0.00207, 0.0, 0.00632, 0.01095, -0.01216, 0.00632, 0.01095, 0.01216, -0.0024, 0.0, -0.0, -0.01536, -0.00472, 0.0, -0.01536, 0.00472, 0.0, -0.0024, 0.0, 0.0, 0.00632, -0.01095, 0.01216, 0.00632, -0.01095, -0.01216, 0.0012, 0.00207, 0.0, 0.00359, 0.01566, 0.0, 0.01177, 0.01094, -0.0, 0.0012, 0.00207, -0.0, -0.01265, -0.0, 0.01216, -0.01265, -0.0, -0.01216, 0.0012, -0.00207, 0.0, 0.01177, -0.01094, -0.0, 0.00359, -0.01566, -0.0, 0.0012, -0.00207, -0.0, 0.00632, 0.01095, -0.01216, 0.00632, 0.01095, 0.01216, -0.0024, -0.0, 0.0, -0.01536, -0.00472, 0.0, -0.01536, 0.00472, 0.0, -0.0024, -0.0, -0.0, 0.00632, -0.01095, 0.01216, 0.00632, -0.01095, -0.01216, 0.00853, -0.00785, -0.00405, -0.01106, 0.00347, 0.00405, -0.01106, 0.00347, -0.00405, 0.00853, -0.00785, 0.00405, -0.01106, -0.00347, 0.00405, 0.00853, 0.00785, -0.00405, 0.00853, 0.00785, 0.00405, -0.01106, -0.00347, -0.00405, 0.00253, 0.01131, -0.00405, 0.00253, -0.01131, 0.00405, 0.00253, -0.01131, -0.00405, 0.00253, 0.01131, 0.00405, 0.00853, -0.00785, -0.00405, -0.01106, 0.00347, 0.00405, -0.01106, 0.00347, -0.00405, 0.00853, -0.00785, 0.00405, -0.01106, -0.00347, 0.00405, 0.00853, 0.00785, -0.00405, 0.00853, 0.00785, 0.00405, -0.01106, -0.00347, -0.00405, 0.00253, 0.01131, -0.00405, 0.00253, -0.01131, 0.00405, 0.00253, -0.01131, -0.00405, 0.00253, 0.01131, 0.00405, 0.01243, 0.02153, 0.0, 0.01243, 0.02153, 0.0, 0.01243, -0.02153, -0.0, 0.01243, -0.02153, 0.0, -0.02486, 0.0, -0.0, -0.02486, 0.0, 0.0, 0.01243, 0.02153, -0.0, 0.01243, 0.02153, -0.0, 0.01243, -0.02153, 0.0, 0.01243, -0.02153, -0.0, -0.02486, -0.0, 0.0, -0.02486, -0.0, 0.0]]), "B2u":array([[0.00379, -0.00219, -0.0013, -0.03103, 0.01829, 0.0, -0.03135, 0.01773, 0.0, 0.00379, -0.00219, 0.0013, 0.0, 0.00048, 0.0, 0.0, 0.00048, -0.0, -0.00379, -0.00219, 0.0013, 0.03135, 0.01773, 0.0, 0.03103, 0.01829, 0.0, -0.00379, -0.00219, -0.0013, 0.00042, -0.00024, 0.0, 0.00042, -0.00024, -0.0, 0.0, 0.00437, -0.0013, -0.00032, -0.03602, -0.0, 0.00032, -0.03602, -0.0, 0.0, 0.00437, 0.0013, -0.00042, -0.00024, -0.0, -0.00042, -0.00024, 0.0, 0.00379, -0.00219, -0.0013, -0.03103, 0.01829, -0.0, -0.03135, 0.01773, -0.0, 0.00379, -0.00219, 0.0013, -0.0, 0.00048, -0.0, -0.0, 0.00048, 0.0, -0.00379, -0.00219, 0.0013, 0.03135, 0.01773, 0.0, 0.03103, 0.01829, 0.0, -0.00379, -0.00219, -0.0013, 0.00042, -0.00024, -0.0, 0.00042, -0.00024, 0.0, 0.0, 0.00437, -0.0013, -0.00032, -0.03602, -0.0, 0.00032, -0.03602, -0.0, 0.0, 0.00437, 0.0013, -0.00042, -0.00024, 0.0, -0.00042, -0.00024, -0.0, -0.00837, 0.00488, 8e-05, -0.00841, 0.00481, 8e-05, -0.00841, 0.00481, -8e-05, -0.00837, 0.00488, -8e-05, 0.00841, 0.00481, -8e-05, 0.00837, 0.00488, -8e-05, 0.00837, 0.00488, 8e-05, 0.00841, 0.00481, 8e-05, -4e-05, -0.00968, 8e-05, 4e-05, -0.00968, 8e-05, 4e-05, -0.00968, -8e-05, -4e-05, -0.00968, -8e-05, -0.00837, 0.00488, 8e-05, -0.00841, 0.00481, 8e-05, -0.00841, 0.00481, -8e-05, -0.00837, 0.00488, -8e-05, 0.00841, 0.00481, -8e-05, 0.00837, 0.00488, -8e-05, 0.00837, 0.00488, 8e-05, 0.00841, 0.00481, 8e-05, -4e-05, -0.00968, 8e-05, 4e-05, -0.00968, 8e-05, 4e-05, -0.00968, -8e-05, -4e-05, -0.00968, -8e-05, 0.02467, -0.01425, -0.00277, 0.02467, -0.01425, 0.00277, -0.02467, -0.01425, 0.00277, -0.02467, -0.01425, -0.00277, 0.0, 0.02849, -0.00277, 0.0, 0.02849, 0.00277, 0.02467, -0.01425, -0.00277, 0.02467, -0.01425, 0.00277, -0.02467, -0.01425, 0.00277, -0.02467, -0.01425, -0.00277, 0.0, 0.02849, -0.00277, 0.0, 0.02849, 0.00277], [-0.01622, 0.00937, -0.00226, -0.00838, 0.00719, -0.0, -0.01042, 0.00366, -0.0, -0.01622, 0.00937, 0.00226, 0.0, 0.00144, 0.0, 0.0, 0.00144, -0.0, 0.01622, 0.00937, 0.00226, 0.01042, 0.00366, -0.0, 0.00838, 0.00719, -0.0, 0.01622, 0.00937, -0.00226, 0.00125, -0.00072, 0.0, 0.00125, -0.00072, -0.0, -0.0, -0.01873, -0.00226, -0.00204, -0.01085, -0.0, 0.00204, -0.01085, -0.0, -0.0, -0.01873, 0.00226, -0.00125, -0.00072, -0.0, -0.00125, -0.00072, 0.0, -0.01622, 0.00937, -0.00226, -0.00838, 0.00719, 0.0, -0.01042, 0.00366, 0.0, -0.01622, 0.00937, 0.00226, -0.0, 0.00144, -0.0, -0.0, 0.00144, 0.0, 0.01622, 0.00937, 0.00226, 0.01042, 0.00366, 0.0, 0.00838, 0.00719, 0.0, 0.01622, 0.00937, -0.00226, 0.00125, -0.00072, -0.0, 0.00125, -0.00072, 0.0, -0.0, -0.01873, -0.00226, -0.00204, -0.01085, -0.0, 0.00204, -0.01085, -0.0, -0.0, -0.01873, 0.00226, -0.00125, -0.00072, 0.0, -0.00125, -0.00072, -0.0, -0.01624, 0.00948, 0.00041, -0.01634, 0.00933, 0.00041, -0.01634, 0.00933, -0.00041, -0.01624, 0.00948, -0.00041, 0.01634, 0.00933, -0.00041, 0.01624, 0.00948, -0.00041, 0.01624, 0.00948, 0.00041, 0.01634, 0.00933, 0.00041, -9e-05, -0.01881, 0.00041, 9e-05, -0.01881, 0.00041, 9e-05, -0.01881, -0.00041, -9e-05, -0.01881, -0.00041, -0.01624, 0.00948, 0.00041, -0.01634, 0.00933, 0.00041, -0.01634, 0.00933, -0.00041, -0.01624, 0.00948, -0.00041, 0.01634, 0.00933, -0.00041, 0.01624, 0.00948, -0.00041, 0.01624, 0.00948, 0.00041, 0.01634, 0.00933, 0.00041, -9e-05, -0.01881, 0.00041, 9e-05, -0.01881, 0.00041, 9e-05, -0.01881, -0.00041, -9e-05, -0.01881, -0.00041, -0.02093, 0.01209, -0.00397, -0.02093, 0.01209, 0.00397, 0.02093, 0.01209, 0.00397, 0.02093, 0.01209, -0.00397, -0.0, -0.02417, -0.00397, 0.0, -0.02417, 0.00397, -0.02093, 0.01209, -0.00397, -0.02093, 0.01209, 0.00397, 0.02093, 0.01209, 0.00397, 0.02093, 0.01209, -0.00397, -0.0, -0.02417, -0.00397, -0.0, -0.02417, 0.00397]]), "E2g1":array([[0.01496, 0.00642, 0.00013, 0.00798, 0.00234, 0.0, 0.01737, -0.00301, 0.0, 0.01496, 0.00642, -0.00013, 0.02365, -0.0, 0.00281, 0.02365, -0.0, -0.00281, 0.01496, -0.00642, 0.00013, 0.01737, 0.00301, 0.0, 0.00798, -0.00234, 0.0, 0.01496, -0.00642, -0.00013, 0.00036, -0.01386, -0.00141, 0.00036, -0.01386, 0.00141, -0.0, -0.0195, -0.00026, -6e-05, -0.02229, -0.0, 6e-05, -0.02229, -0.0, -0.0, -0.0195, 0.00026, 0.00036, 0.01386, 0.00141, 0.00036, 0.01386, -0.00141, -0.01496, -0.00642, -0.00013, -0.00798, -0.00234, -0.0, -0.01737, 0.00301, -0.0, -0.01496, -0.00642, 0.00013, -0.02365, 0.0, -0.00281, -0.02365, 0.0, 0.00281, -0.01496, 0.00642, -0.00013, -0.01737, -0.00301, -0.0, -0.00798, 0.00234, -0.0, -0.01496, 0.00642, 0.00013, -0.00036, 0.01386, 0.00141, -0.00036, 0.01386, -0.00141, 0.0, 0.0195, 0.00026, 6e-05, 0.02229, 0.0, -6e-05, 0.02229, 0.0, 0.0, 0.0195, -0.00026, -0.00036, -0.01386, -0.00141, -0.00036, -0.01386, 0.00141, 0.00752, 0.00823, -0.0001, 0.02096, 0.00021, -6e-05, 0.02096, 0.00021, 6e-05, 0.00752, 0.00823, 0.0001, 0.02096, -0.00021, -6e-05, 0.00752, -0.00823, -0.0001, 0.00752, -0.00823, 0.0001, 0.02096, -0.00021, 6e-05, 0.00022, -0.02044, 0.00017, -0.00022, -0.02044, 0.00017, -0.00022, -0.02044, -0.00017, 0.00022, -0.02044, -0.00017, -0.00752, -0.00823, 0.0001, -0.02096, -0.00021, 6e-05, -0.02096, -0.00021, -6e-05, -0.00752, -0.00823, -0.0001, -0.02096, 0.00021, 6e-05, -0.00752, 0.00823, 0.0001, -0.00752, 0.00823, -0.0001, -0.02096, 0.00021, -6e-05, -0.00022, 0.02044, -0.00017, 0.00022, 0.02044, -0.00017, 0.00022, 0.02044, 0.00017, -0.00022, 0.02044, 0.00017, 0.01619, 0.00966, 0.00015, 0.01619, 0.00966, -0.00015, 0.01619, -0.00966, 0.00015, 0.01619, -0.00966, -0.00015, -0.0, -0.01837, -0.00029, 0.0, -0.01837, 0.00029, -0.01619, -0.00966, -0.00015, -0.01619, -0.00966, 0.00015, -0.01619, 0.00966, -0.00015, -0.01619, 0.00966, 0.00015, 0.0, 0.01837, 0.00029, -0.0, 0.01837, -0.00029], [0.00653, -0.00108, -0.0014, -0.02337, 0.01234, -0.0, -0.01905, 0.00908, -0.0, 0.00653, -0.00108, 0.0014, 0.00451, -0.0, 0.00307, 0.00451, -0.0, -0.00307, 0.00653, 0.00108, -0.0014, -0.01905, -0.00908, -0.0, -0.02337, -0.01234, -0.0, 0.00653, 0.00108, 0.0014, -0.00105, -0.002, -0.00154, -0.00105, -0.002, 0.00154, 0.0, -0.01238, 0.0028, 0.00066, 0.04745, 0.0, -0.00066, 0.04745, 0.0, 0.0, -0.01238, -0.0028, -0.00105, 0.002, 0.00154, -0.00105, 0.002, -0.00154, -0.00653, 0.00108, 0.0014, 0.02337, -0.01234, 0.0, 0.01905, -0.00908, 0.0, -0.00653, 0.00108, -0.0014, -0.00451, 0.0, -0.00307, -0.00451, 0.0, 0.00307, -0.00653, -0.00108, 0.0014, 0.01905, 0.00908, 0.0, 0.02337, 0.01234, 0.0, -0.00653, -0.00108, -0.0014, 0.00105, 0.002, 0.00154, 0.00105, 0.002, -0.00154, -0.0, 0.01238, -0.0028, -0.00066, -0.04745, -0.0, 0.00066, -0.04745, -0.0, -0.0, 0.01238, 0.0028, 0.00105, -0.002, -0.00154, 0.00105, -0.002, 0.00154, -0.00721, 0.00545, 0.0002, 0.0001, 0.00114, -0.0004, 0.0001, 0.00114, 0.0004, -0.00721, 0.00545, -0.0002, 0.0001, -0.00114, -0.0004, -0.00721, -0.00545, 0.0002, -0.00721, -0.00545, -0.0002, 0.0001, -0.00114, 0.0004, 8e-05, 0.00945, 0.0002, -8e-05, 0.00945, 0.0002, -8e-05, 0.00945, -0.0002, 8e-05, 0.00945, -0.0002, 0.00721, -0.00545, -0.0002, -0.0001, -0.00114, 0.0004, -0.0001, -0.00114, -0.0004, 0.00721, -0.00545, 0.0002, -0.0001, 0.00114, 0.0004, 0.00721, 0.00545, -0.0002, 0.00721, 0.00545, 0.0002, -0.0001, 0.00114, -0.0004, -8e-05, -0.00945, -0.0002, 8e-05, -0.00945, -0.0002, 8e-05, -0.00945, 0.0002, -8e-05, -0.00945, 0.0002, 0.0216, -0.00625, -0.00331, 0.0216, -0.00625, 0.00331, 0.0216, 0.00625, -0.00331, 0.0216, 0.00625, 0.00331, -0.0, -0.04366, 0.00663, 0.0, -0.04366, -0.00663, -0.0216, 0.00625, 0.00331, -0.0216, 0.00625, -0.00331, -0.0216, -0.00625, 0.00331, -0.0216, -0.00625, -0.00331, 0.0, 0.04366, -0.00663, -0.0, 0.04366, 0.00663]]), "E2g2":array([[-0.01563, 0.00902, -0.00052, -0.01957, 0.01174, 0.0, -0.01995, 0.01107, 0.0, -0.01563, 0.00902, 0.00052, -0.01119, 0.00635, -0.00125, -0.01119, 0.00635, 0.00125, 0.00164, 0.0152, 0.00026, 0.00281, 0.0077, 0.0, 0.01149, 0.01316, 0.0, 0.00164, 0.0152, -0.00026, 0.01119, 0.01939, 0.0025, 0.01119, 0.01939, -0.0025, 0.01235, 0.00902, 0.00026, 0.00565, 0.01653, 0.0, 0.00527, 0.00629, 0.0, 0.01235, 0.00902, -0.00026, -0.0111, 0.00652, 0.00125, -0.0111, 0.00652, -0.00125, 0.01563, -0.00902, 0.00052, 0.01957, -0.01174, -0.0, 0.01995, -0.01107, -0.0, 0.01563, -0.00902, -0.00052, 0.01119, -0.00635, 0.00125, 0.01119, -0.00635, -0.00125, -0.00164, -0.0152, -0.00026, -0.00281, -0.0077, -0.0, -0.01149, -0.01316, -0.0, -0.00164, -0.0152, 0.00026, -0.01119, -0.01939, -0.0025, -0.01119, -0.01939, 0.0025, -0.01235, -0.00902, -0.00026, -0.00565, -0.01653, -0.0, -0.00527, -0.00629, -0.0, -0.01235, -0.00902, 0.00026, 0.0111, -0.00652, -0.00125, 0.0111, -0.00652, 0.00125, -0.01711, 0.00999, 0.00015, -0.01721, 0.00982, 0.00015, -0.01721, 0.00982, -0.00015, -0.01711, 0.00999, -0.00015, -0.0029, 0.00988, -6e-05, 0.01, 0.01744, -9e-05, 0.01, 0.01744, 9e-05, -0.0029, 0.00988, 6e-05, 0.01011, 0.01738, -9e-05, 0.01001, 0.00243, -6e-05, 0.01001, 0.00243, 6e-05, 0.01011, 0.01738, 9e-05, 0.01711, -0.00999, -0.00015, 0.01721, -0.00982, -0.00015, 0.01721, -0.00982, 0.00015, 0.01711, -0.00999, 0.00015, 0.0029, -0.00988, 6e-05, -0.01, -0.01744, 9e-05, -0.01, -0.01744, -9e-05, 0.0029, -0.00988, -6e-05, -0.01011, -0.01738, 9e-05, -0.01001, -0.00243, 6e-05, -0.01001, -0.00243, -6e-05, -0.01011, -0.01738, -9e-05, -0.01317, 0.00761, -0.00093, -0.01317, 0.00761, 0.00093, -0.00115, 0.01721, 0.00046, -0.00115, 0.01721, -0.00046, 0.01548, 0.00761, 0.00046, 0.01548, 0.00761, -0.00046, 0.01317, -0.00761, 0.00093, 0.01317, -0.00761, -0.00093, 0.00115, -0.01721, -0.00046, 0.00115, -0.01721, 0.00046, -0.01548, -0.00761, -0.00046, -0.01548, -0.00761, 0.00046], [-0.01183, 0.00683, 0.00166, 0.03795, -0.02241, -0.0, 0.03838, -0.02166, -0.0, -0.01183, 0.00683, -0.00166, -0.00222, 0.00037, -0.00153, -0.00222, 0.00037, 0.00153, 0.00482, 0.00532, -0.00083, -0.02086, -0.01279, 0.0, -0.01645, -0.01074, 0.0, 0.00482, 0.00532, 0.00083, 0.00222, 0.00385, 0.00306, 0.00222, 0.00385, -0.00306, 0.00219, 0.00683, -0.00083, -0.00108, -0.01961, -0.0, -0.00065, -0.02446, -0.0, 0.00219, 0.00683, 0.00083, -0.00144, 0.00174, 0.00153, -0.00144, 0.00174, -0.00153, 0.01183, -0.00683, -0.00166, -0.03795, 0.02241, 0.0, -0.03838, 0.02166, 0.0, 0.01183, -0.00683, 0.00166, 0.00222, -0.00037, 0.00153, 0.00222, -0.00037, -0.00153, -0.00482, -0.00532, 0.00083, 0.02086, 0.01279, -0.0, 0.01645, 0.01074, -0.0, -0.00482, -0.00532, -0.00083, -0.00222, -0.00385, -0.00306, -0.00222, -0.00385, 0.00306, -0.00219, -0.00683, 0.00083, 0.00108, 0.01961, 0.0, 0.00065, 0.02446, 0.0, -0.00219, -0.00683, -0.00083, 0.00144, -0.00174, -0.00153, 0.00144, -0.00174, 0.00153, 0.00655, -0.00387, 0.00019, 0.00662, -0.00374, 0.00019, 0.00662, -0.00374, -0.00019, 0.00655, -0.00387, -0.00019, -0.00713, -0.00263, 0.00015, -0.0007, 0.001, -0.00035, -0.0007, 0.001, 0.00035, -0.00713, -0.00263, -0.00015, 0.00121, -0.00011, -0.00035, 0.00129, -0.0075, 0.00015, 0.00129, -0.0075, -0.00015, 0.00121, -0.00011, 0.00035, -0.00655, 0.00387, -0.00019, -0.00662, 0.00374, -0.00019, -0.00662, 0.00374, 0.00019, -0.00655, 0.00387, 0.00019, 0.00713, 0.00263, -0.00015, 0.0007, -0.001, 0.00035, 0.0007, -0.001, -0.00035, 0.00713, 0.00263, 0.00015, -0.00121, 0.00011, 0.00035, -0.00129, 0.0075, -0.00015, -0.00129, 0.0075, 0.00015, -0.00121, 0.00011, -0.00035, -0.03969, 0.02291, 0.00423, -0.03969, 0.02291, -0.00423, 0.01732, 0.01583, -0.00211, 0.01732, 0.01583, 0.00211, 0.00505, 0.02291, -0.00211, 0.00505, 0.02291, 0.00211, 0.03969, -0.02291, -0.00423, 0.03969, -0.02291, 0.00423, -0.01732, -0.01583, 0.00211, -0.01732, -0.01583, -0.00211, -0.00505, -0.02291, 0.00211, -0.00505, -0.02291, -0.00211]]), "E2g3":array([[-0.00193, 0.01547, -0.00016, -0.01076, 0.01282, 0.0, -0.00203, 0.00741, 0.0, -0.00193, 0.01547, 0.00016, 0.01118, 0.00642, 0.00127, 0.01118, 0.00642, -0.00127, 0.01629, 0.00941, 0.00031, 0.01851, 0.01032, -0.0, 0.01819, 0.01087, -0.0, 0.01629, 0.00941, -0.00031, 0.01115, 0.00647, 0.00127, 0.01115, 0.00647, -0.00127, 0.01243, -0.00941, -0.00016, 0.00541, -0.00546, -0.0, 0.00572, -0.01573, -0.0, 0.01243, -0.00941, 0.00016, -0.01118, 0.01937, 0.00253, -0.01118, 0.01937, -0.00253, 0.00193, -0.01547, 0.00016, 0.01076, -0.01282, -0.0, 0.00203, -0.00741, -0.0, 0.00193, -0.01547, -0.00016, -0.01118, -0.00642, -0.00127, -0.01118, -0.00642, 0.00127, -0.01629, -0.00941, -0.00031, -0.01851, -0.01032, 0.0, -0.01819, -0.01087, 0.0, -0.01629, -0.00941, 0.00031, -0.01115, -0.00647, -0.00127, -0.01115, -0.00647, 0.00127, -0.01243, 0.00941, 0.00016, -0.00541, 0.00546, 0.0, -0.00572, 0.01573, 0.0, -0.01243, 0.00941, -0.00016, 0.01118, -0.01937, -0.00253, 0.01118, -0.01937, 0.00253, -0.00983, 0.01745, 5e-05, 0.00304, 0.00995, 0.0001, 0.00304, 0.00995, -0.0001, -0.00983, 0.01745, -5e-05, 0.01698, 0.00973, -0.00015, 0.01692, 0.00984, -0.00015, 0.01692, 0.00984, 0.00015, 0.01698, 0.00973, 0.00015, 0.01014, -0.00234, 0.0001, 0.0102, -0.01723, 5e-05, 0.0102, -0.01723, -5e-05, 0.01014, -0.00234, -0.0001, 0.00983, -0.01745, -5e-05, -0.00304, -0.00995, -0.0001, -0.00304, -0.00995, 0.0001, 0.00983, -0.01745, 5e-05, -0.01698, -0.00973, 0.00015, -0.01692, -0.00984, 0.00015, -0.01692, -0.00984, -0.00015, -0.01698, -0.00973, -0.00015, -0.01014, 0.00234, -0.0001, -0.0102, 0.01723, -5e-05, -0.0102, 0.01723, 5e-05, -0.01014, 0.00234, 0.0001, -1e-05, 0.01806, -0.00017, -1e-05, 0.01806, 0.00017, 0.01566, 0.00904, 0.00034, 0.01566, 0.00904, -0.00034, 0.01564, -0.00904, -0.00017, 0.01564, -0.00904, 0.00017, 1e-05, -0.01806, 0.00017, 1e-05, -0.01806, -0.00017, -0.01566, -0.00904, -0.00034, -0.01566, -0.00904, 0.00034, -0.01564, 0.00904, 0.00017, -0.01564, 0.00904, -0.00017], [0.00371, -0.0048, -0.00186, -0.01529, 0.01126, 0.0, -0.02034, 0.01307, 0.0, 0.00371, -0.0048, 0.00186, -0.00185, 0.0, -0.00146, -0.00185, 0.0, 0.00146, -0.00973, -0.00562, 0.00373, 0.03774, 0.02068, -0.0, 0.03678, 0.02234, -0.0, -0.00973, -0.00562, -0.00373, -0.00092, -0.00161, -0.00146, -0.00092, -0.00161, 0.00146, -0.0023, 0.00562, -0.00186, 0.00115, -0.02415, -0.0, 0.00211, -0.01887, -0.0, -0.0023, 0.00562, 0.00186, 0.00185, -0.00321, -0.00292, 0.00185, -0.00321, 0.00292, -0.00371, 0.0048, 0.00186, 0.01529, -0.01126, -0.0, 0.02034, -0.01307, -0.0, -0.00371, 0.0048, -0.00186, 0.00185, -0.0, 0.00146, 0.00185, -0.0, -0.00146, 0.00973, 0.00562, -0.00373, -0.03774, -0.02068, 0.0, -0.03678, -0.02234, 0.0, 0.00973, 0.00562, 0.00373, 0.00092, 0.00161, 0.00146, 0.00092, 0.00161, -0.00146, 0.0023, -0.00562, 0.00186, -0.00115, 0.02415, 0.0, -0.00211, 0.01887, 0.0, 0.0023, -0.00562, -0.00186, -0.00185, 0.00321, 0.00292, -0.00185, 0.00321, -0.00292, -0.00108, -0.00048, -0.00044, -0.00812, 0.00347, 3e-05, -0.00812, 0.00347, -3e-05, -0.00108, -0.00048, 0.00044, 0.00824, 0.00465, 0.00041, 0.00815, 0.00482, 0.00041, 0.00815, 0.00482, -0.00041, 0.00824, 0.00465, -0.00041, -0.00105, -0.00877, 3e-05, -0.00095, -0.0007, -0.00044, -0.00095, -0.0007, 0.00044, -0.00105, -0.00877, -3e-05, 0.00108, 0.00048, 0.00044, 0.00812, -0.00347, -3e-05, 0.00812, -0.00347, 3e-05, 0.00108, 0.00048, -0.00044, -0.00824, -0.00465, -0.00041, -0.00815, -0.00482, -0.00041, -0.00815, -0.00482, 0.00041, -0.00824, -0.00465, 0.00041, 0.00105, 0.00877, -3e-05, 0.00095, 0.0007, 0.00044, 0.00095, 0.0007, -0.00044, 0.00105, 0.00877, 3e-05, 0.0146, -0.01461, -0.00442, 0.0146, -0.01461, 0.00442, -0.03456, -0.01995, 0.00884, -0.03456, -0.01995, -0.00884, -0.00535, 0.01995, -0.00442, -0.00535, 0.01995, 0.00442, -0.0146, 0.01461, 0.00442, -0.0146, 0.01461, -0.00442, 0.03456, 0.01995, -0.00884, 0.03456, 0.01995, 0.00884, 0.00535, -0.01995, 0.00442, 0.00535, -0.01995, -0.00442]]), "E1g1":array([[0.00686, -0.00725, -0.00657, 0.0, -0.0, -0.00043, 0.0, 0.0, -0.00173, -0.00686, 0.00725, -0.00657, -0.0, -0.00088, -0.0, 0.0, 0.00088, 0.0, -0.00686, -0.00725, 0.00657, 0.0, 0.0, 0.00173, 0.0, 0.0, 0.00043, 0.00686, 0.00725, 0.00657, -0.00231, -0.00488, 0.00218, 0.00231, 0.00488, 0.00218, 0.0, -0.01913, 0.01314, 0.0, 0.0, 0.00216, -0.0, 0.0, 0.00216, -0.0, 0.01913, 0.01314, -0.00231, 0.00488, -0.00218, 0.00231, -0.00488, -0.00218, -0.00686, 0.00725, 0.00657, -0.0, 0.0, 0.00043, -0.0, -0.0, 0.00173, 0.00686, -0.00725, 0.00657, 0.0, 0.00088, 0.0, -0.0, -0.00088, -0.0, 0.00686, 0.00725, -0.00657, -0.0, -0.0, -0.00173, -0.0, -0.0, -0.00043, -0.00686, -0.00725, -0.00657, 0.00231, 0.00488, -0.00218, -0.00231, -0.00488, -0.00218, -0.0, 0.01913, -0.01314, -0.0, -0.0, -0.00216, 0.0, -0.0, -0.00216, 0.0, -0.01913, -0.01314, 0.00231, -0.00488, 0.00218, -0.00231, 0.00488, 0.00218, -0.00573, 0.00032, 0.00149, 0.01295, -0.01057, -0.0105, -0.01295, 0.01057, -0.0105, 0.00573, -0.00032, 0.00149, -0.01295, -0.01057, 0.0105, 0.00573, 0.00032, -0.00149, -0.00573, -0.00032, -0.00149, 0.01295, 0.01057, 0.0105, 0.0001, -0.01138, 0.00902, -0.0001, -0.01138, 0.00902, 0.0001, 0.01138, 0.00902, -0.0001, 0.01138, 0.00902, 0.00573, -0.00032, -0.00149, -0.01295, 0.01057, 0.0105, 0.01295, -0.01057, 0.0105, -0.00573, 0.00032, -0.00149, 0.01295, 0.01057, -0.0105, -0.00573, -0.00032, 0.00149, 0.00573, 0.00032, 0.00149, -0.01295, -0.01057, -0.0105, -0.0001, 0.01138, -0.00902, 0.0001, 0.01138, -0.00902, -0.0001, -0.01138, -0.00902, 0.0001, -0.01138, -0.00902, 0.01195, -0.01201, -0.00989, -0.01195, 0.01201, -0.00989, -0.01195, -0.01201, 0.00989, 0.01195, 0.01201, 0.00989, 0.0, -0.03272, 0.01977, -0.0, 0.03272, 0.01977, -0.01195, 0.01201, 0.00989, 0.01195, -0.01201, 0.00989, 0.01195, 0.01201, -0.00989, -0.01195, -0.01201, -0.00989, -0.0, 0.03272, -0.01977, 0.0, -0.03272, -0.01977], [-0.00082, -0.00439, -0.00211, 0.0, 0.0, -0.00103, -0.0, 0.0, -0.00227, 0.00082, 0.00439, -0.00211, 0.0, -0.00418, -0.0, 0.0, 0.00418, 0.0, 0.00082, -0.00439, 0.00211, -0.0, -0.0, 0.00227, -0.0, -0.0, 0.00103, -0.00082, 0.00439, 0.00211, -0.00198, -0.00761, 0.00309, 0.00198, 0.00761, 0.00309, -0.0, -0.00297, 0.00422, 0.0, 0.0, 0.0033, -0.0, 0.0, 0.0033, -0.0, 0.00297, 0.00422, -0.00198, 0.00761, -0.00309, 0.00198, -0.00761, -0.00309, 0.00082, 0.00439, 0.00211, -0.0, -0.0, 0.00103, 0.0, -0.0, 0.00227, -0.00082, -0.00439, 0.00211, -0.0, 0.00418, 0.0, -0.0, -0.00418, -0.0, -0.00082, 0.00439, -0.00211, 0.0, 0.0, -0.00227, 0.0, 0.0, -0.00103, 0.00082, -0.00439, -0.00211, 0.00198, 0.00761, -0.00309, -0.00198, -0.00761, -0.00309, 0.0, 0.00297, -0.00422, -0.0, -0.0, -0.0033, 0.0, -0.0, -0.0033, 0.0, -0.00297, -0.00422, 0.00198, -0.00761, 0.00309, -0.00198, 0.00761, 0.00309, 0.01576, -0.01357, -0.0118, -0.02519, 0.012, 0.01352, 0.02519, -0.012, 0.01352, -0.01576, 0.01357, -0.0118, 0.02519, 0.012, -0.01352, -0.01576, -0.01357, 0.0118, 0.01576, 0.01357, 0.0118, -0.02519, -0.012, -0.01352, -0.00167, 0.00739, -0.00172, 0.00167, 0.00739, -0.00172, -0.00167, -0.00739, -0.00172, 0.00167, -0.00739, -0.00172, -0.01576, 0.01357, 0.0118, 0.02519, -0.012, -0.01352, -0.02519, 0.012, -0.01352, 0.01576, -0.01357, 0.0118, -0.02519, -0.012, 0.01352, 0.01576, 0.01357, -0.0118, -0.01576, -0.01357, -0.0118, 0.02519, 0.012, 0.01352, 0.00167, -0.00739, 0.00172, -0.00167, -0.00739, 0.00172, 0.00167, 0.00739, 0.00172, -0.00167, 0.00739, 0.00172, 0.0056, -0.00943, -0.00545, -0.0056, 0.00943, -0.00545, -0.0056, -0.00943, 0.00545, 0.0056, 0.00943, 0.00545, 0.0, -0.01913, 0.01091, 0.0, 0.01913, 0.01091, -0.0056, 0.00943, 0.00545, 0.0056, -0.00943, 0.00545, 0.0056, 0.00943, -0.00545, -0.0056, -0.00943, -0.00545, -0.0, 0.01913, -0.01091, -0.0, -0.01913, -0.01091]]), "E1g2":array([[-0.0168, 0.0097, 0.01394, 0.0, 0.0, 0.00197, -0.0, -0.0, 0.00197, 0.0168, -0.0097, 0.01394, -0.00564, 7e-05, 0.00201, 0.00564, -7e-05, 0.00201, -0.00952, -0.0029, 0.00697, -0.0, -0.0, 0.00049, -0.0, -0.0, 0.00148, 0.00952, 0.0029, 0.00697, -0.00012, 7e-05, 0.0, 0.00012, -7e-05, 0.0, -0.00225, 0.0097, -0.00697, -0.0, -0.0, -0.00148, -0.0, -0.0, -0.00049, 0.00225, -0.0097, -0.00697, 0.00288, -0.00485, 0.00201, -0.00288, 0.00485, 0.00201, 0.0168, -0.0097, -0.01394, -0.0, -0.0, -0.00197, 0.0, 0.0, -0.00197, -0.0168, 0.0097, -0.01394, 0.00564, -7e-05, -0.00201, -0.00564, 7e-05, -0.00201, 0.00952, 0.0029, -0.00697, 0.0, 0.0, -0.00049, 0.0, 0.0, -0.00148, -0.00952, -0.0029, -0.00697, 0.00012, -7e-05, -0.0, -0.00012, 7e-05, -0.0, 0.00225, -0.0097, 0.00697, 0.0, 0.0, 0.00148, 0.0, 0.0, 0.00049, -0.00225, 0.0097, 0.00697, -0.00288, 0.00485, -0.00201, 0.00288, -0.00485, -0.00201, -0.00915, 0.00576, 0.00809, -0.00956, 0.00504, 0.00809, 0.00956, -0.00504, 0.00809, 0.00915, -0.00576, 0.00809, 0.0023, 0.00397, -0.00131, -0.01374, -0.00577, 0.0094, 0.01374, 0.00577, 0.0094, -0.0023, -0.00397, -0.00131, -0.00187, 0.01478, -0.0094, -0.00229, -0.00398, 0.00131, 0.00229, 0.00398, 0.00131, 0.00187, -0.01478, -0.0094, 0.00915, -0.00576, -0.00809, 0.00956, -0.00504, -0.00809, -0.00956, 0.00504, -0.00809, -0.00915, 0.00576, -0.00809, -0.0023, -0.00397, 0.00131, 0.01374, 0.00577, -0.0094, -0.01374, -0.00577, -0.0094, 0.0023, 0.00397, 0.00131, 0.00187, -0.01478, 0.0094, 0.00229, 0.00398, -0.00131, -0.00229, -0.00398, -0.00131, -0.00187, 0.01478, 0.0094, -0.03091, 0.01785, 0.02143, 0.03091, -0.01785, 0.02143, -0.01713, -0.00602, 0.01071, 0.01713, 0.00602, 0.01071, -0.00335, 0.01785, -0.01071, 0.00335, -0.01785, -0.01071, 0.03091, -0.01785, -0.02143, -0.03091, 0.01785, -0.02143, 0.01713, 0.00602, -0.01071, -0.01713, -0.00602, -0.01071, 0.00335, -0.01785, 0.01071, -0.00335, 0.01785, 0.01071], [-0.00165, 0.00095, 0.00413, -0.0, 0.0, 0.0019, -0.0, -0.0, 0.0019, 0.00165, -0.00095, 0.00413, -0.00513, 0.0015, 0.00214, 0.00513, -0.0015, 0.00214, -0.00215, 0.00182, 0.00206, -0.0, -0.0, 0.00023, 0.0, 0.0, 0.00167, 0.00215, -0.00182, 0.00206, -0.00261, 0.0015, 0.0, 0.00261, -0.0015, -0.0, -0.00265, 0.00095, -0.00206, -0.0, 0.0, -0.00167, -0.0, -0.0, -0.00023, 0.00265, -0.00095, -0.00206, 0.00387, -0.00369, 0.00214, -0.00387, 0.00369, 0.00214, 0.00165, -0.00095, -0.00413, 0.0, -0.0, -0.0019, 0.0, 0.0, -0.0019, -0.00165, 0.00095, -0.00413, 0.00513, -0.0015, -0.00214, -0.00513, 0.0015, -0.00214, 0.00215, -0.00182, -0.00206, 0.0, 0.0, -0.00023, -0.0, -0.0, -0.00167, -0.00215, 0.00182, -0.00206, 0.00261, -0.0015, -0.0, -0.00261, 0.0015, 0.0, 0.00265, -0.00095, 0.00206, 0.0, -0.0, 0.00167, 0.0, 0.0, 0.00023, -0.00265, 0.00095, 0.00206, -0.00387, 0.00369, -0.00214, 0.00387, -0.00369, -0.00214, 0.00929, -0.00362, -0.00419, 0.00778, -0.00624, -0.00419, -0.00778, 0.00624, -0.00419, -0.00929, 0.00362, -0.00419, -0.01878, -0.00804, 0.01109, 0.02564, 0.01586, -0.01528, -0.02564, -0.01586, -0.01528, 0.01878, 0.00804, 0.01109, -0.00092, -0.03014, 0.01528, -0.00243, 0.02028, -0.01109, 0.00243, -0.02028, -0.01109, 0.00092, 0.03014, 0.01528, -0.00929, 0.00362, 0.00419, -0.00778, 0.00624, 0.00419, 0.00778, -0.00624, 0.00419, 0.00929, -0.00362, 0.00419, 0.01878, 0.00804, -0.01109, -0.02564, -0.01586, 0.01528, 0.02564, 0.01586, 0.01528, -0.01878, -0.00804, -0.01109, 0.00092, 0.03014, -0.01528, 0.00243, -0.02028, 0.01109, -0.00243, 0.02028, 0.01109, -0.00092, -0.03014, -0.01528, -0.01645, 0.0095, 0.01148, 0.01645, -0.0095, 0.01148, -0.00975, -0.00211, 0.00574, 0.00975, 0.00211, 0.00574, -0.00305, 0.0095, -0.00574, 0.00305, -0.0095, -0.00574, 0.01645, -0.0095, -0.01148, -0.01645, 0.0095, -0.01148, 0.00975, 0.00211, -0.00574, -0.00975, -0.00211, -0.00574, 0.00305, -0.0095, 0.00574, -0.00305, 0.0095, 0.00574]]), "E1g3":array([[-0.00957, 0.00312, 0.00711, 0.0, -0.0, 0.00188, 0.0, 0.0, 0.00061, 0.00957, -0.00312, 0.00711, -0.00575, -0.00036, 0.00212, 0.00575, 0.00036, 0.00212, -0.01706, -0.00985, 0.01421, 0.0, -0.0, 0.00249, 0.0, 0.0, 0.00249, 0.01706, 0.00985, 0.01421, -0.00318, -0.0048, 0.00212, 0.00318, 0.0048, 0.00212, -0.00208, -0.00985, 0.00711, 0.0, 0.0, 0.00061, -0.0, 0.0, 0.00188, 0.00208, 0.00985, 0.00711, 0.00062, 0.00036, 0.0, -0.00062, -0.00036, -0.0, 0.00957, -0.00312, -0.00711, -0.0, 0.0, -0.00188, -0.0, -0.0, -0.00061, -0.00957, 0.00312, -0.00711, 0.00575, 0.00036, -0.00212, -0.00575, -0.00036, -0.00212, 0.01706, 0.00985, -0.01421, -0.0, 0.0, -0.00249, -0.0, -0.0, -0.00249, -0.01706, -0.00985, -0.01421, 0.00318, 0.0048, -0.00212, -0.00318, -0.0048, -0.00212, 0.00208, 0.00985, -0.00711, -0.0, -0.0, -0.00061, 0.0, -0.0, -0.00188, -0.00208, -0.00985, -0.00711, -0.00062, -0.00036, -0.0, 0.00062, 0.00036, 0.0, -0.01282, 0.00511, 0.00874, 0.00149, -0.00379, -0.00068, -0.00149, 0.00379, -0.00068, 0.01282, -0.00511, 0.00874, -0.00934, -0.00476, 0.00806, -0.00879, -0.00571, 0.00806, 0.00879, 0.00571, 0.00806, 0.00934, 0.00476, 0.00806, -0.00253, 0.00319, -0.00068, -0.00199, -0.01366, 0.00874, 0.00199, 0.01366, 0.00874, 0.00253, -0.00319, -0.00068, 0.01282, -0.00511, -0.00874, -0.00149, 0.00379, 0.00068, 0.00149, -0.00379, 0.00068, -0.01282, 0.00511, -0.00874, 0.00934, 0.00476, -0.00806, 0.00879, 0.00571, -0.00806, -0.00879, -0.00571, -0.00806, -0.00934, -0.00476, -0.00806, 0.00253, -0.00319, 0.00068, 0.00199, 0.01366, -0.00874, -0.00199, -0.01366, -0.00874, -0.00253, 0.00319, 0.00068, -0.01783, 0.00598, 0.01079, 0.01783, -0.00598, 0.01079, -0.03193, -0.01844, 0.02158, 0.03193, 0.01844, 0.02158, -0.00373, -0.01844, 0.01079, 0.00373, 0.01844, 0.01079, 0.01783, -0.00598, -0.01079, -0.01783, 0.00598, -0.01079, 0.03193, 0.01844, -0.02158, -0.03193, -0.01844, -0.02158, 0.00373, 0.01844, -0.01079, -0.00373, -0.01844, -0.01079], [0.00106, 0.00069, -0.00149, 0.0, -0.0, -0.00124, 0.0, -0.0, -0.00085, -0.00106, -0.00069, -0.00149, 0.00651, 0.00189, -0.00227, -0.00651, -0.00189, -0.00227, 0.00099, 0.00057, -0.00298, 0.0, 0.0, -0.00209, 0.0, 0.0, -0.00209, -0.00099, -0.00057, -0.00298, 0.0049, 0.00469, -0.00227, -0.0049, -0.00469, -0.00227, 0.00113, 0.00057, -0.00149, -0.0, 0.0, -0.00085, -0.0, 0.0, -0.00124, -0.00113, -0.00057, -0.00149, -0.00328, -0.00189, -0.0, 0.00328, 0.00189, 0.0, -0.00106, -0.00069, 0.00149, -0.0, 0.0, 0.00124, -0.0, 0.0, 0.00085, 0.00106, 0.00069, 0.00149, -0.00651, -0.00189, 0.00227, 0.00651, 0.00189, 0.00227, -0.00099, -0.00057, 0.00298, -0.0, -0.0, 0.00209, -0.0, -0.0, 0.00209, 0.00099, 0.00057, 0.00298, -0.0049, -0.00469, 0.00227, 0.0049, 0.00469, 0.00227, -0.00113, -0.00057, 0.00149, 0.0, -0.0, 0.00085, 0.0, -0.0, 0.00124, 0.00113, 0.00057, 0.00149, 0.00328, 0.00189, 0.0, -0.00328, -0.00189, -0.0, -0.0265, 0.01547, 0.01579, 0.01824, -0.00915, -0.01117, -0.01824, 0.00915, -0.01117, 0.0265, -0.01547, 0.01579, -0.00841, -0.00607, 0.00462, -0.00946, -0.00425, 0.00462, 0.00946, 0.00425, 0.00462, 0.00841, 0.00607, 0.00462, 0.0012, 0.02037, -0.01117, 0.00015, -0.03069, 0.01579, -0.00015, 0.03069, 0.01579, -0.0012, -0.02037, -0.01117, 0.0265, -0.01547, -0.01579, -0.01824, 0.00915, 0.01117, 0.01824, -0.00915, 0.01117, -0.0265, 0.01547, -0.01579, 0.00841, 0.00607, -0.00462, 0.00946, 0.00425, -0.00462, -0.00946, -0.00425, -0.00462, -0.00841, -0.00607, -0.00462, -0.0012, -0.02037, 0.01117, -0.00015, 0.03069, -0.01579, 0.00015, -0.03069, -0.01579, 0.0012, 0.02037, 0.01117, 0.00879, -0.00312, -0.00487, -0.00879, 0.00312, -0.00487, 0.0159, 0.00918, -0.00974, -0.0159, -0.00918, -0.00974, 0.00169, 0.00918, -0.00487, -0.00169, -0.00918, -0.00487, -0.00879, 0.00312, 0.00487, 0.00879, -0.00312, 0.00487, -0.0159, -0.00918, 0.00974, 0.0159, 0.00918, 0.00974, -0.00169, -0.00918, 0.00487, 0.00169, 0.00918, 0.00487]]), "E1g4":array([[-0.01577, 0.00777, 0.01281, -0.0, 0.0, 0.0023, -0.0, 0.0, 0.00167, 0.01577, -0.00777, 0.01281, -0.00687, -0.0, 0.0024, 0.00687, -0.0, 0.0024, -0.01577, -0.00777, 0.01281, -0.0, -0.0, 0.00167, -0.0, -0.0, 0.0023, 0.01577, 0.00777, 0.01281, -0.00192, -0.00286, 0.0012, 0.00192, 0.00286, 0.0012, -0.00231, -0.0, 0.0, -0.0, -0.0, -0.00063, -0.0, -0.0, 0.00063, 0.00231, -0.0, -0.0, 0.00192, -0.00286, 0.0012, -0.00192, 0.00286, 0.0012, 0.01577, -0.00777, -0.01281, 0.0, -0.0, -0.0023, 0.0, -0.0, -0.00167, -0.01577, 0.00777, -0.01281, 0.00687, 0.0, -0.0024, -0.00687, 0.0, -0.0024, 0.01577, 0.00777, -0.01281, 0.0, 0.0, -0.00167, 0.0, 0.0, -0.0023, -0.01577, -0.00777, -0.01281, 0.00192, 0.00286, -0.0012, -0.00192, -0.00286, -0.0012, 0.00231, 0.0, -0.0, 0.0, 0.0, 0.00063, 0.0, 0.0, -0.00063, -0.00231, 0.0, 0.0, -0.00192, 0.00286, -0.0012, 0.00192, -0.00286, -0.0012, -0.01261, 0.00659, 0.00974, -0.00464, 0.00062, 0.0041, 0.00464, -0.00062, 0.0041, 0.01261, -0.00659, 0.00974, -0.00464, -0.00062, 0.0041, -0.01261, -0.00659, 0.00974, 0.01261, 0.00659, 0.00974, 0.00464, 0.00062, 0.0041, -0.00238, 0.00989, -0.00564, -0.00238, -0.00989, 0.00564, 0.00238, 0.00989, 0.00564, 0.00238, -0.00989, -0.00564, 0.01261, -0.00659, -0.00974, 0.00464, -0.00062, -0.0041, -0.00464, 0.00062, -0.0041, -0.01261, 0.00659, -0.00974, 0.00464, 0.00062, -0.0041, 0.01261, 0.00659, -0.00974, -0.01261, -0.00659, -0.00974, -0.00464, -0.00062, -0.0041, 0.00238, -0.00989, 0.00564, 0.00238, 0.00989, -0.00564, -0.00238, -0.00989, -0.00564, -0.00238, 0.00989, 0.00564, -0.02984, 0.01502, 0.01967, 0.02984, -0.01502, 0.01967, -0.02984, -0.01502, 0.01967, 0.02984, 0.01502, 0.01967, -0.00383, 0.0, 0.0, 0.00383, 0.0, 0.0, 0.02984, -0.01502, -0.01967, -0.02984, 0.01502, -0.01967, 0.02984, 0.01502, -0.01967, -0.02984, -0.01502, -0.01967, 0.00383, -0.0, -0.0, -0.00383, -0.0, -0.0], [-0.00124, -0.0002, 0.00289, -0.0, 0.0, 0.00176, -0.0, -0.0, 0.00126, 0.00124, 0.0002, 0.00289, -0.00636, 0.0, 0.0024, 0.00636, 0.0, 0.0024, -0.00124, 0.0002, 0.00289, -0.0, 0.0, 0.00126, -0.0, -0.0, 0.00176, 0.00124, -0.0002, 0.00289, -0.00421, -0.00124, 0.0012, 0.00421, 0.00124, 0.0012, -0.00159, 0.0, 0.0, -0.0, -0.0, -0.0005, 0.0, -0.0, 0.0005, 0.00159, 0.0, -0.0, 0.00421, -0.00124, 0.0012, -0.00421, 0.00124, 0.0012, 0.00124, 0.0002, -0.00289, 0.0, -0.0, -0.00176, 0.0, 0.0, -0.00126, -0.00124, -0.0002, -0.00289, 0.00636, -0.0, -0.0024, -0.00636, -0.0, -0.0024, 0.00124, -0.0002, -0.00289, 0.0, -0.0, -0.00126, 0.0, 0.0, -0.00176, -0.00124, 0.0002, -0.00289, 0.00421, 0.00124, -0.0012, -0.00421, -0.00124, -0.0012, 0.00159, -0.0, -0.0, 0.0, 0.0, 0.0005, -0.0, 0.0, -0.0005, -0.00159, -0.0, 0.0, -0.00421, 0.00124, -0.0012, 0.00421, -0.00124, -0.0012, 0.02176, -0.01199, -0.01251, -0.0055, 0.00159, 0.0035, 0.0055, -0.00159, 0.0035, -0.02176, 0.01199, -0.01251, -0.0055, -0.00159, 0.0035, 0.02176, 0.01199, -0.01251, -0.02176, -0.01199, -0.01251, 0.0055, 0.00159, 0.0035, -0.00087, -0.0304, 0.01601, -0.00087, 0.0304, -0.01601, 0.00087, -0.0304, -0.01601, 0.00087, 0.0304, 0.01601, -0.02176, 0.01199, 0.01251, 0.0055, -0.00159, -0.0035, -0.0055, 0.00159, -0.0035, 0.02176, -0.01199, 0.01251, 0.0055, 0.00159, -0.0035, -0.02176, -0.01199, 0.01251, 0.02176, 0.01199, 0.01251, -0.0055, -0.00159, -0.0035, 0.00087, 0.0304, -0.01601, 0.00087, -0.0304, 0.01601, -0.00087, 0.0304, 0.01601, -0.00087, -0.0304, -0.01601, -0.0145, 0.00726, 0.00915, 0.0145, -0.00726, 0.00915, -0.0145, -0.00726, 0.00915, 0.0145, 0.00726, 0.00915, -0.00193, 0.0, 0.0, 0.00193, 0.0, 0.0, 0.0145, -0.00726, -0.00915, -0.0145, 0.00726, -0.00915, 0.0145, 0.00726, -0.00915, -0.0145, -0.00726, -0.00915, 0.00193, -0.0, -0.0, -0.00193, -0.0, -0.0]]), "E1g5":array([[-0.00151, -0.00261, 0.0, 0.0, -0.0, 0.00083, 0.0, 0.0, -0.00083, 0.00151, 0.00261, -0.0, -0.00322, -0.00091, 0.00129, 0.00322, 0.00091, 0.00129, -0.01395, -0.0098, 0.01187, 0.0, -0.0, 0.00257, 0.0, 0.0, 0.00174, 0.01395, 0.0098, 0.01187, -0.00322, -0.00558, 0.00257, 0.00322, 0.00558, 0.00257, -0.00151, -0.01698, 0.01187, 0.0, 0.0, 0.00174, -0.0, 0.0, 0.00257, 0.00151, 0.01698, 0.01187, -0.00082, 0.00324, -0.00129, 0.00082, -0.00324, -0.00129, 0.00151, 0.00261, -0.0, -0.0, 0.0, -0.00083, -0.0, -0.0, 0.00083, -0.00151, -0.00261, 0.0, 0.00322, 0.00091, -0.00129, -0.00322, -0.00091, -0.00129, 0.01395, 0.0098, -0.01187, -0.0, 0.0, -0.00257, -0.0, -0.0, -0.00174, -0.01395, -0.0098, -0.01187, 0.00322, 0.00558, -0.00257, -0.00322, -0.00558, -0.00257, 0.00151, 0.01698, -0.01187, -0.0, -0.0, -0.00174, 0.0, -0.0, -0.00257, -0.00151, -0.01698, -0.01187, 0.00082, -0.00324, 0.00129, -0.00082, 0.00324, 0.00129, -0.01142, 0.00307, 0.00631, 0.00837, -0.00835, -0.00631, -0.00837, 0.00835, -0.00631, 0.01142, -0.00307, 0.00631, -0.0126, -0.00889, 0.01081, -0.00258, -0.0034, 0.0045, 0.00258, 0.0034, 0.0045, 0.0126, 0.00889, 0.01081, -0.00165, -0.00393, 0.0045, -0.0014, -0.01536, 0.01081, 0.0014, 0.01536, 0.01081, 0.00165, 0.00393, 0.0045, 0.01142, -0.00307, -0.00631, -0.00837, 0.00835, 0.00631, 0.00837, -0.00835, 0.00631, -0.01142, 0.00307, -0.00631, 0.0126, 0.00889, -0.01081, 0.00258, 0.0034, -0.0045, -0.00258, -0.0034, -0.0045, -0.0126, -0.00889, -0.01081, 0.00165, 0.00393, -0.0045, 0.0014, 0.01536, -0.01081, -0.0014, -0.01536, -0.01081, -0.00165, -0.00393, -0.0045, -0.00256, -0.00444, -0.0, 0.00256, 0.00444, -0.0, -0.02475, -0.01725, 0.01781, 0.02475, 0.01725, 0.01781, -0.00256, -0.03006, 0.01781, 0.00256, 0.03006, 0.01781, 0.00256, 0.00444, 0.0, -0.00256, -0.00444, 0.0, 0.02475, 0.01725, -0.01781, -0.02475, -0.01725, -0.01781, 0.00256, 0.03006, -0.01781, -0.00256, -0.03006, -0.01781], [-0.00141, -0.00245, 0.0, 0.0, 0.0, 0.00041, -0.0, -0.0, -0.00041, 0.00141, 0.00245, -0.0, -0.00438, -0.00359, 0.00157, 0.00438, 0.00359, 0.00157, -0.00091, -0.00216, 0.00312, -0.0, 0.0, 0.00279, -0.0, -0.0, 0.00238, 0.00091, 0.00216, 0.00312, -0.00438, -0.00758, 0.00313, 0.00438, 0.00758, 0.00313, -0.00141, -0.00186, 0.00312, -0.0, -0.0, 0.00238, 0.0, -0.0, 0.00279, 0.00141, 0.00186, 0.00312, 0.00092, 0.00558, -0.00157, -0.00092, -0.00558, -0.00157, 0.00141, 0.00245, -0.0, -0.0, -0.0, -0.00041, 0.0, 0.0, 0.00041, -0.00141, -0.00245, 0.0, 0.00438, 0.00359, -0.00157, -0.00438, -0.00359, -0.00157, 0.00091, 0.00216, -0.00312, 0.0, -0.0, -0.00279, 0.0, 0.0, -0.00238, -0.00091, -0.00216, -0.00312, 0.00438, 0.00758, -0.00313, -0.00438, -0.00758, -0.00313, 0.00141, 0.00186, -0.00312, 0.0, 0.0, -0.00238, -0.0, 0.0, -0.00279, -0.00141, -0.00186, -0.00312, -0.00092, -0.00558, 0.00157, 0.00092, 0.00558, 0.00157, 0.02361, -0.01605, -0.01494, -0.0257, 0.01242, 0.01494, 0.0257, -0.01242, 0.01494, -0.02361, 0.01605, -0.01494, 0.0187, 0.01098, -0.01003, -0.00475, -0.00534, 0.00491, 0.00475, 0.00534, 0.00491, -0.0187, -0.01098, -0.01003, -0.00225, -0.00679, 0.00491, 0.00016, 0.02168, -0.01003, -0.00016, -0.02168, -0.01003, 0.00225, 0.00679, 0.00491, -0.02361, 0.01605, 0.01494, 0.0257, -0.01242, -0.01494, -0.0257, 0.01242, -0.01494, 0.02361, -0.01605, 0.01494, -0.0187, -0.01098, 0.01003, 0.00475, 0.00534, -0.00491, -0.00475, -0.00534, -0.00491, 0.0187, 0.01098, 0.01003, 0.00225, 0.00679, -0.00491, -0.00016, -0.02168, 0.01003, 0.00016, 0.02168, 0.01003, -0.00225, -0.00679, -0.00491, -0.00205, -0.00354, -0.0, 0.00205, 0.00354, -0.0, -0.01331, -0.01004, 0.0088, 0.01331, 0.01004, 0.0088, -0.00205, -0.01655, 0.0088, 0.00205, 0.01655, 0.0088, 0.00205, 0.00354, 0.0, -0.00205, -0.00354, 0.0, 0.01331, 0.01004, -0.0088, -0.01331, -0.01004, -0.0088, 0.00205, 0.01655, -0.0088, -0.00205, -0.01655, -0.0088]]), "E1g6":array([[-0.01358, 0.00961, 0.01155, -0.0, 0.0, 0.00136, -0.0, 0.0, 0.00199, 0.01358, -0.00961, 0.01155, -0.00315, 0.00038, 0.0012, 0.00315, -0.00038, 0.0012, -0.00154, 0.00266, -0.0, -0.0, 0.0, -0.00063, -0.0, -0.0, 0.00063, 0.00154, -0.00266, 0.0, 0.00125, 0.00292, -0.0012, -0.00125, -0.00292, -0.0012, -0.00154, 0.01656, -0.01155, -0.0, -0.0, -0.00199, 0.0, -0.0, -0.00136, 0.00154, -0.01656, -0.01155, 0.00315, -0.00546, 0.00241, -0.00315, 0.00546, 0.00241, 0.01358, -0.00961, -0.01155, 0.0, -0.0, -0.00136, 0.0, -0.0, -0.00199, -0.01358, 0.00961, -0.01155, 0.00315, -0.00038, -0.0012, -0.00315, 0.00038, -0.0012, 0.00154, -0.00266, 0.0, 0.0, -0.0, 0.00063, 0.0, 0.0, -0.00063, -0.00154, 0.00266, -0.0, -0.00125, -0.00292, 0.0012, 0.00125, 0.00292, 0.0012, 0.00154, -0.01656, 0.01155, 0.0, 0.0, 0.00199, -0.0, 0.0, 0.00136, -0.00154, 0.01656, 0.01155, -0.00315, 0.00546, -0.00241, 0.00315, -0.00546, -0.00241, -0.00247, 0.00305, 0.00409, -0.01303, 0.00905, 0.01093, 0.01303, -0.00905, 0.01093, 0.00247, -0.00305, 0.00409, 0.00916, 0.00844, -0.00684, -0.01188, -0.00371, 0.00684, 0.01188, 0.00371, 0.00684, -0.00916, -0.00844, -0.00684, -0.00132, 0.01581, -0.01093, -0.00141, 0.00366, -0.00409, 0.00141, -0.00366, -0.00409, 0.00132, -0.01581, -0.01093, 0.00247, -0.00305, -0.00409, 0.01303, -0.00905, -0.01093, -0.01303, 0.00905, -0.01093, -0.00247, 0.00305, -0.00409, -0.00916, -0.00844, 0.00684, 0.01188, 0.00371, -0.00684, -0.01188, -0.00371, -0.00684, 0.00916, 0.00844, 0.00684, 0.00132, -0.01581, 0.01093, 0.00141, -0.00366, 0.00409, -0.00141, 0.00366, 0.00409, -0.00132, 0.01581, 0.01093, -0.02398, 0.01645, 0.01761, 0.02398, -0.01645, 0.01761, -0.00226, 0.00391, -0.0, 0.00226, -0.00391, -0.0, -0.00226, 0.02899, -0.01761, 0.00226, -0.02899, -0.01761, 0.02398, -0.01645, -0.01761, -0.02398, 0.01645, -0.01761, 0.00226, -0.00391, 0.0, -0.00226, 0.00391, 0.0, 0.00226, -0.02899, 0.01761, -0.00226, 0.02899, 0.01761], [-0.00091, 0.00334, 0.00394, -0.0, 0.0, 0.00181, -0.0, 0.0, 0.00271, 0.00091, -0.00334, 0.00394, -0.00364, 0.00296, 0.0016, 0.00364, -0.00296, 0.0016, -0.00244, 0.00423, -0.0, 0.0, 0.0, -0.00089, 0.0, 0.0, 0.00089, 0.00244, -0.00423, 0.0, -0.00074, 0.00463, -0.0016, 0.00074, -0.00463, -0.0016, -0.00244, 0.00246, -0.00394, 0.0, 0.0, -0.00271, 0.0, -0.0, -0.00181, 0.00244, -0.00246, -0.00394, 0.00364, -0.00631, 0.0032, -0.00364, 0.00631, 0.0032, 0.00091, -0.00334, -0.00394, 0.0, -0.0, -0.00181, 0.0, -0.0, -0.00271, -0.00091, 0.00334, -0.00394, 0.00364, -0.00296, -0.0016, -0.00364, 0.00296, -0.0016, 0.00244, -0.00423, 0.0, -0.0, -0.0, 0.00089, -0.0, -0.0, -0.00089, -0.00244, 0.00423, -0.0, 0.00074, -0.00463, 0.0016, -0.00074, 0.00463, 0.0016, 0.00244, -0.00246, 0.00394, -0.0, -0.0, 0.00271, -0.0, 0.0, 0.00181, -0.00244, 0.00246, 0.00394, -0.00364, 0.00631, -0.0032, 0.00364, -0.00631, -0.0032, -0.00463, 0.00638, 0.0053, 0.01848, -0.01058, -0.00963, -0.01848, 0.01058, -0.00963, 0.00463, -0.00638, 0.0053, -0.02633, -0.0114, 0.01493, 0.02304, 0.0171, -0.01493, -0.02304, -0.0171, -0.01493, 0.02633, 0.0114, 0.01493, -8e-05, -0.02129, 0.00963, -0.00321, 0.0072, -0.0053, 0.00321, -0.0072, -0.0053, 8e-05, 0.02129, 0.00963, 0.00463, -0.00638, -0.0053, -0.01848, 0.01058, 0.00963, 0.01848, -0.01058, 0.00963, -0.00463, 0.00638, -0.0053, 0.02633, 0.0114, -0.01493, -0.02304, -0.0171, 0.01493, 0.02304, 0.0171, 0.01493, -0.02633, -0.0114, -0.01493, 8e-05, 0.02129, -0.00963, 0.00321, -0.0072, 0.0053, -0.00321, 0.0072, 0.0053, -8e-05, -0.02129, -0.00963, -0.01307, 0.01088, 0.01023, 0.01307, -0.01088, 0.01023, -0.00289, 0.005, -0.0, 0.00289, -0.005, -0.0, -0.00289, 0.01676, -0.01023, 0.00289, -0.01676, -0.01023, 0.01307, -0.01088, -0.01023, -0.01307, 0.01088, -0.01023, 0.00289, -0.005, 0.0, -0.00289, 0.005, 0.0, 0.00289, -0.01676, 0.01023, -0.00289, 0.01676, 0.01023]]), "E1u1":array([[-0.01162, 0.00697, 0.00042, 0.0389, -0.02167, -0.0, 0.03776, -0.02109, -0.0, -0.01162, 0.00697, -0.00042, -0.001, 0.0, -0.00434, -0.001, 0.0, 0.00434, -0.01162, -0.00697, 0.00042, 0.03776, 0.02109, 0.0, 0.0389, 0.02167, 0.0, -0.01162, -0.00697, -0.00042, 0.00046, -0.00084, -0.00217, 0.00046, -0.00084, 0.00217, 0.00046, -0.0, -0.0, 0.0013, -0.00127, -0.0, 0.0013, 0.00127, 0.0, 0.00046, -0.0, 0.0, 0.00046, 0.00084, 0.00217, 0.00046, 0.00084, -0.00217, -0.01162, 0.00697, 0.00042, 0.0389, -0.02167, -0.0, 0.03776, -0.02109, -0.0, -0.01162, 0.00697, -0.00042, -0.001, -0.0, -0.00434, -0.001, -0.0, 0.00434, -0.01162, -0.00697, 0.00042, 0.03776, 0.02109, 0.0, 0.0389, 0.02167, 0.0, -0.01162, -0.00697, -0.00042, 0.00046, -0.00084, -0.00217, 0.00046, -0.00084, 0.00217, 0.00046, 0.0, 0.0, 0.0013, -0.00127, -0.0, 0.0013, 0.00127, 0.0, 0.00046, 0.0, -0.0, 0.00046, 0.00084, 0.00217, 0.00046, 0.00084, -0.00217, 0.00588, -0.00309, 0.00018, 0.00431, -0.00193, 0.00014, 0.00431, -0.00193, -0.00014, 0.00588, -0.00309, -0.00018, 0.00431, 0.00193, 0.00014, 0.00588, 0.00309, 0.00018, 0.00588, 0.00309, -0.00018, 0.00431, 0.00193, -0.00014, 0.00074, -0.00195, -4e-05, 0.00074, 0.00195, 4e-05, 0.00074, 0.00195, -4e-05, 0.00074, -0.00195, 4e-05, 0.00588, -0.00309, 0.00018, 0.00431, -0.00193, 0.00014, 0.00431, -0.00193, -0.00014, 0.00588, -0.00309, -0.00018, 0.00431, 0.00193, 0.00014, 0.00588, 0.00309, 0.00018, 0.00588, 0.00309, -0.00018, 0.00431, 0.00193, -0.00014, 0.00074, -0.00195, -4e-05, 0.00074, 0.00195, 4e-05, 0.00074, 0.00195, -4e-05, 0.00074, -0.00195, 4e-05, -0.03818, 0.0217, 0.00018, -0.03818, 0.0217, -0.00018, -0.03818, -0.0217, 0.00018, -0.03818, -0.0217, -0.00018, -0.00059, 0.0, -0.0, -0.00059, 0.0, -0.0, -0.03818, 0.0217, 0.00018, -0.03818, 0.0217, -0.00018, -0.03818, -0.0217, 0.00018, -0.03818, -0.0217, -0.00018, -0.00059, -0.0, 0.0, -0.00059, -0.0, 0.0], [-0.00073, 0.00369, 0.01881, -0.01652, 0.01261, -0.0, -0.01646, 0.01354, -0.0, -0.00073, 0.00369, -0.01881, -0.00033, -0.0, 0.00264, -0.00033, -0.0, -0.00264, -0.00073, -0.00369, 0.01881, -0.01646, -0.01354, -0.0, -0.01652, -0.01261, -0.0, -0.00073, -0.00369, -0.01881, 0.00496, -0.00305, 0.00132, 0.00496, -0.00305, -0.00132, 0.00566, 0.0, 0.0, 0.00615, -0.00041, 0.0, 0.00615, 0.00041, 0.0, 0.00566, 0.0, -0.0, 0.00496, 0.00305, -0.00132, 0.00496, 0.00305, 0.00132, -0.00073, 0.00369, 0.01881, -0.01652, 0.01261, 0.0, -0.01646, 0.01354, 0.0, -0.00073, 0.00369, -0.01881, -0.00033, 0.0, 0.00264, -0.00033, 0.0, -0.00264, -0.00073, -0.00369, 0.01881, -0.01646, -0.01354, 0.0, -0.01652, -0.01261, 0.0, -0.00073, -0.00369, -0.01881, 0.00496, -0.00305, 0.00132, 0.00496, -0.00305, -0.00132, 0.00566, -0.0, -0.0, 0.00615, -0.00041, -0.0, 0.00615, 0.00041, 0.0, 0.00566, 0.0, 0.0, 0.00496, 0.00305, -0.00132, 0.00496, 0.00305, 0.00132, 0.00684, 0.00146, 0.00352, 0.00402, 6e-05, 0.0042, 0.00402, 6e-05, -0.0042, 0.00684, 0.00146, -0.00352, 0.00402, -6e-05, 0.0042, 0.00684, -0.00146, 0.00352, 0.00684, -0.00146, -0.00352, 0.00402, -6e-05, -0.0042, 0.00674, -0.00174, 0.00068, 0.00674, 0.00174, -0.00068, 0.00674, 0.00174, 0.00068, 0.00674, -0.00174, -0.00068, 0.00684, 0.00146, 0.00352, 0.00402, 6e-05, 0.0042, 0.00402, 6e-05, -0.0042, 0.00684, 0.00146, -0.00352, 0.00402, -6e-05, 0.0042, 0.00684, -0.00146, 0.00352, 0.00684, -0.00146, -0.00352, 0.00402, -6e-05, -0.0042, 0.00674, -0.00174, 0.00068, 0.00674, 0.00174, -0.00068, 0.00674, 0.00174, 0.00068, 0.00674, -0.00174, -0.00068, -0.01422, 0.01183, 0.03905, -0.01422, 0.01183, -0.03905, -0.01422, -0.01183, 0.03905, -0.01422, -0.01183, -0.03905, 0.00628, -0.0, 0.0, 0.00628, -0.0, -0.0, -0.01422, 0.01183, 0.03905, -0.01422, 0.01183, -0.03905, -0.01422, -0.01183, 0.03905, -0.01422, -0.01183, -0.03905, 0.00628, 0.0, -0.0, 0.00628, 0.0, 0.0]]), "E1u2":array([[0.00037, 0.00064, -0.0, 0.00092, -0.00015, -0.0, -0.00059, 0.00072, 0.0, 0.00037, 0.00064, 0.0, -0.00016, 0.00055, -0.00228, -0.00016, 0.00055, 0.00228, -0.0117, -0.00633, 0.0008, 0.03802, 0.02227, -0.0, 0.03738, 0.02163, -0.0, -0.0117, -0.00633, -0.0008, -0.00016, -0.00028, -0.00456, -0.00016, -0.00028, 0.00456, 0.00037, -0.0133, 0.0008, 5e-05, 0.04319, 0.0, 0.00028, 0.04406, 0.0, 0.00037, -0.0133, -0.0008, 0.00056, 0.00014, -0.00228, 0.00056, 0.00014, 0.00228, 0.00037, 0.00064, 0.0, 0.00092, -0.00015, -0.0, -0.00059, 0.00072, 0.0, 0.00037, 0.00064, -0.0, -0.00016, 0.00055, -0.00228, -0.00016, 0.00055, 0.00228, -0.0117, -0.00633, 0.0008, 0.03802, 0.02227, -0.0, 0.03738, 0.02163, -0.0, -0.0117, -0.00633, -0.0008, -0.00016, -0.00028, -0.00456, -0.00016, -0.00028, 0.00456, 0.00037, -0.0133, 0.0008, 5e-05, 0.04319, 0.0, 0.00028, 0.04406, 0.0, 0.00037, -0.0133, -0.0008, 0.00056, 0.00014, -0.00228, 0.00056, 0.00014, 0.00228, 0.00139, -0.0003, 2e-05, -0.00096, 0.00105, -2e-05, -0.00096, 0.00105, 2e-05, 0.00139, -0.0003, -2e-05, 0.00576, 0.00329, 0.00011, 0.00434, 0.00304, 9e-05, 0.00434, 0.00304, -9e-05, 0.00576, 0.00329, -0.00011, 0.00046, 0.00528, 9e-05, -3e-05, 0.00663, 0.00011, -3e-05, 0.00663, -0.00011, 0.00046, 0.00528, -9e-05, 0.00139, -0.0003, 2e-05, -0.00096, 0.00105, -2e-05, -0.00096, 0.00105, 2e-05, 0.00139, -0.0003, -2e-05, 0.00576, 0.00329, 0.00011, 0.00434, 0.00304, 9e-05, 0.00434, 0.00304, -9e-05, 0.00576, 0.00329, -0.00011, 0.00046, 0.00528, 9e-05, -3e-05, 0.00663, 0.00011, -3e-05, 0.00663, -0.00011, 0.00046, 0.00528, -9e-05, 0.00041, 0.00071, 0.0, 0.00041, 0.00071, -0.0, -0.03908, -0.02209, 0.00108, -0.03908, -0.02209, -0.00108, 0.00041, -0.04489, 0.00108, 0.00041, -0.04489, -0.00108, 0.00041, 0.00071, -0.0, 0.00041, 0.00071, 0.0, -0.03908, -0.02209, 0.00108, -0.03908, -0.02209, -0.00108, 0.00041, -0.04489, 0.00108, 0.00041, -0.04489, -0.00108], [0.00334, 0.00578, 0.0, 0.00487, 0.00635, -0.0, 0.00306, 0.00739, 0.0, 0.00334, 0.00578, -0.0, 0.00061, 0.0062, 0.00139, 0.00061, 0.0062, -0.00139, -0.00525, 0.00083, 0.01808, -0.02057, -0.00847, 0.0, -0.0225, -0.00723, 0.0, -0.00525, 0.00083, -0.01808, 0.00061, 0.00106, 0.00278, 0.00061, 0.00106, -0.00278, 0.00334, -0.00413, 0.01808, 0.00498, -0.0231, -0.0, 0.00295, -0.02205, -0.0, 0.00334, -0.00413, -0.01808, 0.00507, 0.00363, 0.00139, 0.00507, 0.00363, -0.00139, 0.00334, 0.00578, -0.0, 0.00487, 0.00635, -0.0, 0.00306, 0.00739, 0.0, 0.00334, 0.00578, 0.0, 0.00061, 0.0062, 0.00139, 0.00061, 0.0062, -0.00139, -0.00525, 0.00083, 0.01808, -0.02057, -0.00847, 0.0, -0.0225, -0.00723, 0.0, -0.00525, 0.00083, -0.01808, 0.00061, 0.00106, 0.00278, 0.00061, 0.00106, -0.00278, 0.00334, -0.00413, 0.01808, 0.00498, -0.0231, -0.0, 0.00295, -0.02205, -0.0, 0.00334, -0.00413, -0.01808, 0.00507, 0.00363, 0.00139, 0.00507, 0.00363, -0.00139, 0.00532, 0.00479, -0.001, 0.00149, 0.007, 0.001, 0.00149, 0.007, -0.001, 0.00532, 0.00479, 0.001, 0.00231, 0.00633, 0.00206, 0.00132, 0.00363, 0.00305, 0.00132, 0.00363, -0.00305, 0.00231, 0.00633, -0.00206, 0.00248, 0.00295, 0.00305, 0.00433, 0.00516, 0.00206, 0.00433, 0.00516, -0.00206, 0.00248, 0.00295, -0.00305, 0.00532, 0.00479, -0.001, 0.00149, 0.007, 0.001, 0.00149, 0.007, -0.001, 0.00532, 0.00479, 0.001, 0.00231, 0.00633, 0.00206, 0.00132, 0.00363, 0.00305, 0.00132, 0.00363, -0.00305, 0.00231, 0.00633, -0.00206, 0.00248, 0.00295, 0.00305, 0.00433, 0.00516, 0.00206, 0.00433, 0.00516, -0.00206, 0.00248, 0.00295, -0.00305, 0.00291, 0.00503, 0.0, 0.00291, 0.00503, -0.0, -0.01717, -0.00656, 0.03717, -0.01717, -0.00656, -0.03717, 0.00291, -0.01815, 0.03717, 0.00291, -0.01815, -0.03717, 0.00291, 0.00503, -0.0, 0.00291, 0.00503, 0.0, -0.01717, -0.00656, 0.03717, -0.01717, -0.00656, -0.03717, 0.00291, -0.01815, 0.03717, 0.00291, -0.01815, -0.03717]]), "E1u3":array([[-0.0121, 0.00682, 0.0001, 0.03855, -0.02328, -0.0, 0.03983, -0.024, -0.0, -0.0121, 0.00682, -0.0001, -0.00039, -0.00071, -0.00233, -0.00039, -0.00071, 0.00233, 0.00014, -0.00025, -0.0, -0.00039, -0.00226, 0.0, 0.00215, -0.00079, -0.0, 0.00014, -0.00025, 0.0, 0.00081, -2e-05, 0.00233, 0.00081, -2e-05, -0.00233, 0.00014, 0.01389, -0.0001, 0.00087, -0.04649, 0.0, 0.00089, -0.04503, 0.0, 0.00014, 0.01389, 0.0001, -0.00039, 0.00067, 0.00466, -0.00039, 0.00067, -0.00466, -0.0121, 0.00682, 0.0001, 0.03855, -0.02328, -0.0, 0.03983, -0.024, -0.0, -0.0121, 0.00682, -0.0001, -0.00039, -0.00071, -0.00233, -0.00039, -0.00071, 0.00233, 0.00014, -0.00025, 0.0, -0.00039, -0.00226, 0.0, 0.00215, -0.00079, -0.0, 0.00014, -0.00025, -0.0, 0.00081, -2e-05, 0.00233, 0.00081, -2e-05, -0.00233, 0.00014, 0.01389, -0.0001, 0.00087, -0.04649, 0.0, 0.00089, -0.04503, 0.0, 0.00014, 0.01389, 0.0001, -0.00039, 0.00067, 0.00466, -0.00039, 0.00067, -0.00466, 0.00374, -0.00288, 6e-05, 0.00562, -0.00339, 1e-05, 0.00562, -0.00339, -1e-05, 0.00374, -0.00288, -6e-05, -0.00125, -0.0016, 5e-05, 0.00201, 0.00029, -5e-05, 0.00201, 0.00029, 5e-05, -0.00125, -0.0016, -5e-05, 0.00013, -0.00656, -1e-05, 0.00063, -0.00468, -6e-05, 0.00063, -0.00468, 6e-05, 0.00013, -0.00656, 1e-05, 0.00374, -0.00288, 6e-05, 0.00562, -0.00339, 1e-05, 0.00562, -0.00339, -1e-05, 0.00374, -0.00288, -6e-05, -0.00125, -0.0016, 5e-05, 0.00201, 0.00029, -5e-05, 0.00201, 0.00029, 5e-05, -0.00125, -0.0016, -5e-05, 0.00013, -0.00656, -1e-05, 0.00063, -0.00468, -6e-05, 0.00063, -0.00468, 6e-05, 0.00013, -0.00656, 1e-05, -0.03902, 0.02312, -0.00054, -0.03902, 0.02312, 0.00054, -0.00051, 0.00089, -0.0, -0.00051, 0.00089, 0.0, -0.00051, 0.04535, 0.00054, -0.00051, 0.04535, -0.00054, -0.03902, 0.02312, -0.00054, -0.03902, 0.02312, 0.00054, -0.00051, 0.00089, 0.0, -0.00051, 0.00089, -0.0, -0.00051, 0.04535, 0.00054, -0.00051, 0.04535, -0.00054], [-0.00464, -0.00161, 0.01689, -0.02065, 0.00754, 0.0, -0.01977, 0.00903, 0.0, -0.00464, -0.00161, -0.01689, 0.0011, -0.0063, 0.00171, 0.0011, -0.0063, -0.00171, 0.00371, -0.00643, -0.0, 0.00291, -0.00508, 0.0, 0.00295, -0.00506, 0.0, 0.00371, -0.00643, 0.0, 0.00491, -0.0041, -0.00171, 0.00491, -0.0041, 0.00171, 0.00371, 0.00322, -0.01689, 0.00206, 0.02164, -0.0, 0.0038, 0.02166, -0.0, 0.00371, 0.00322, 0.01689, 0.0011, -0.0019, -0.00341, 0.0011, -0.0019, 0.00341, -0.00464, -0.00161, 0.01689, -0.02065, 0.00754, 0.0, -0.01977, 0.00903, 0.0, -0.00464, -0.00161, -0.01689, 0.0011, -0.0063, 0.00171, 0.0011, -0.0063, -0.00171, 0.00371, -0.00643, 0.0, 0.00291, -0.00508, 0.0, 0.00295, -0.00506, 0.0, 0.00371, -0.00643, -0.0, 0.00491, -0.0041, -0.00171, 0.00491, -0.0041, 0.00171, 0.00371, 0.00322, -0.01689, 0.00206, 0.02164, -0.0, 0.0038, 0.02166, -0.0, 0.00371, 0.00322, 0.01689, 0.0011, -0.0019, -0.00341, 0.0011, -0.0019, 0.00341, 0.00132, -0.00403, 0.00295, 0.0001, -0.00452, 0.00246, 0.0001, -0.00452, -0.00246, 0.00132, -0.00403, -0.00295, 0.00406, -0.0054, 0.00048, 0.00265, -0.00621, -0.00048, 0.00265, -0.00621, 0.00048, 0.00406, -0.0054, -0.00048, 0.00387, -0.00235, -0.00246, 0.00284, -0.00316, -0.00295, 0.00284, -0.00316, 0.00295, 0.00387, -0.00235, 0.00246, 0.00132, -0.00403, 0.00295, 0.0001, -0.00452, 0.00246, 0.0001, -0.00452, -0.00246, 0.00132, -0.00403, -0.00295, 0.00406, -0.0054, 0.00048, 0.00265, -0.00621, -0.00048, 0.00265, -0.00621, 0.00048, 0.00406, -0.0054, -0.00048, 0.00387, -0.00235, -0.00246, 0.00284, -0.00316, -0.00295, 0.00284, -0.00316, 0.00295, 0.00387, -0.00235, 0.00246, -0.02033, 0.00598, 0.03529, -0.02033, 0.00598, -0.03529, 0.00499, -0.00864, 0.0, 0.00499, -0.00864, -0.0, 0.00499, 0.0206, -0.03529, 0.00499, 0.0206, 0.03529, -0.02033, 0.00598, 0.03529, -0.02033, 0.00598, -0.03529, 0.00499, -0.00864, -0.0, 0.00499, -0.00864, 0.0, 0.00499, 0.0206, -0.03529, 0.00499, 0.0206, 0.03529]]), "E1u4":array([[0.00681, -0.00356, -0.00013, -0.02052, 0.01317, -0.0, -0.02241, 0.01429, -0.0, 0.00681, -0.00356, 0.00013, -0.0, 0.00066, 0.0, -0.0, 0.00066, 0.0, -0.00681, -0.00356, 0.00013, 0.02241, 0.01429, -0.0, 0.02052, 0.01317, -0.0, -0.00681, -0.00356, -0.00013, -0.00048, -0.00018, -0.00398, -0.00048, -0.00018, 0.00398, -0.0, -0.01536, 0.00026, -2e-05, 0.05091, 0.0, 2e-05, 0.05091, 0.0, -0.0, -0.01536, -0.00026, 0.00048, -0.00018, -0.00398, 0.00048, -0.00018, 0.00398, 0.00681, -0.00356, -0.00013, -0.02052, 0.01317, -0.0, -0.02241, 0.01429, -0.0, 0.00681, -0.00356, 0.00013, 0.0, 0.00066, -0.0, 0.0, 0.00066, -0.0, -0.00681, -0.00356, 0.00013, 0.02241, 0.01429, 0.0, 0.02052, 0.01317, 0.0, -0.00681, -0.00356, -0.00013, -0.00048, -0.00018, -0.00398, -0.00048, -0.00018, 0.00398, -0.0, -0.01536, 0.00026, -2e-05, 0.05091, 0.0, 2e-05, 0.05091, 0.0, -0.0, -0.01536, -0.00026, 0.00048, -0.00018, -0.00398, 0.00048, -0.00018, 0.00398, -0.00123, 0.00152, -6e-05, -0.00377, 0.00258, 6e-05, -0.00377, 0.00258, -6e-05, -0.00123, 0.00152, 6e-05, 0.00377, 0.00258, -6e-05, 0.00123, 0.00152, 6e-05, 0.00123, 0.00152, -6e-05, 0.00377, 0.00258, 6e-05, 0.00035, 0.00638, -0.0, -0.00035, 0.00638, -0.0, -0.00035, 0.00638, 0.0, 0.00035, 0.00638, 0.0, -0.00123, 0.00152, -6e-05, -0.00377, 0.00258, 6e-05, -0.00377, 0.00258, -6e-05, -0.00123, 0.00152, 6e-05, 0.00377, 0.00258, -6e-05, 0.00123, 0.00152, 6e-05, 0.00123, 0.00152, -6e-05, 0.00377, 0.00258, 6e-05, 0.00035, 0.00638, -0.0, -0.00035, 0.00638, -0.0, -0.00035, 0.00638, 0.0, 0.00035, 0.00638, 0.0, 0.02188, -0.01308, 0.00013, 0.02188, -0.01308, -0.00013, -0.02188, -0.01308, -0.00013, -0.02188, -0.01308, 0.00013, -0.0, -0.05097, -0.00027, 0.0, -0.05097, 0.00027, 0.02188, -0.01308, 0.00013, 0.02188, -0.01308, -0.00013, -0.02188, -0.01308, -0.00013, -0.02188, -0.01308, 0.00013, 0.0, -0.05097, -0.00027, -0.0, -0.05097, 0.00027], [-0.00529, -0.00437, 0.00997, -0.01498, 0.00087, -0.0, -0.01338, 0.00143, -0.0, -0.00529, -0.00437, -0.00997, -0.0, -0.00726, -0.0, -0.0, -0.00726, 0.0, 0.00529, -0.00437, -0.00997, 0.01338, 0.00143, 0.0, 0.01498, 0.00087, 0.0, 0.00529, -0.00437, 0.00997, 0.00212, -0.0036, -0.0026, 0.00212, -0.0036, 0.0026, 0.0, 0.00479, -0.01993, -0.00128, 0.02571, 0.0, 0.00128, 0.02571, 0.0, 0.0, 0.00479, 0.01993, -0.00212, -0.0036, -0.0026, -0.00212, -0.0036, 0.0026, -0.00529, -0.00437, 0.00997, -0.01498, 0.00087, -0.0, -0.01338, 0.00143, -0.0, -0.00529, -0.00437, -0.00997, 0.0, -0.00726, 0.0, 0.0, -0.00726, -0.0, 0.00529, -0.00437, -0.00997, 0.01338, 0.00143, 0.0, 0.01498, 0.00087, 0.0, 0.00529, -0.00437, 0.00997, 0.00212, -0.0036, -0.0026, 0.00212, -0.0036, 0.0026, 0.0, 0.00479, -0.01993, -0.00128, 0.02571, 0.0, 0.00128, 0.02571, 0.0, -0.0, 0.00479, 0.01993, -0.00212, -0.0036, -0.0026, -0.00212, -0.0036, 0.0026, -0.00142, -0.00549, 0.00198, -0.00083, -0.00653, 0.00057, -0.00083, -0.00653, -0.00057, -0.00142, -0.00549, -0.00198, 0.00083, -0.00653, -0.00057, 0.00142, -0.00549, -0.00198, 0.00142, -0.00549, 0.00198, 0.00083, -0.00653, 0.00057, 0.0006, -0.00406, -0.00255, -0.0006, -0.00406, -0.00255, -0.0006, -0.00406, 0.00255, 0.0006, -0.00406, 0.00255, -0.00142, -0.00549, 0.00198, -0.00083, -0.00653, 0.00057, -0.00083, -0.00653, -0.00057, -0.00142, -0.00549, -0.00198, 0.00083, -0.00653, -0.00057, 0.00142, -0.00549, -0.00198, 0.00142, -0.00549, 0.00198, 0.00083, -0.00653, 0.00057, 0.0006, -0.00406, -0.00255, -0.0006, -0.00406, -0.00255, -0.0006, -0.00406, 0.00255, 0.0006, -0.00406, 0.00255, -0.01355, -0.0003, 0.02064, -0.01355, -0.0003, -0.02064, 0.01355, -0.0003, -0.02064, 0.01355, -0.0003, 0.02064, -0.0, 0.02316, -0.04127, -0.0, 0.02316, 0.04127, -0.01355, -0.0003, 0.02064, -0.01355, -0.0003, -0.02064, 0.01355, -0.0003, -0.02064, 0.01355, -0.0003, 0.02064, 0.0, 0.02316, -0.04127, 0.0, 0.02316, 0.04127]]), "E1u5":array([[-0.01307, 0.00754, 0.00022, 0.04269, -0.02465, -0.0, 0.04269, -0.02464, -0.0, -0.01307, 0.00754, -0.00022, -0.00077, -0.00043, -0.00369, -0.00077, -0.00043, 0.00369, -0.0064, -0.00401, 0.00011, 0.02102, 0.01054, 0.0, 0.02305, 0.01172, 0.0, -0.0064, -0.00401, -0.00011, 0.00074, -0.00043, -0.0, 0.00074, -0.00043, 0.0, 0.00027, 0.00754, -0.00011, 0.00138, -0.02582, -0.0, 0.00138, -0.02347, -0.0, 0.00027, 0.00754, 0.00011, -1e-05, 0.00088, 0.00369, -1e-05, 0.00088, -0.00369, -0.01307, 0.00754, 0.00022, 0.04269, -0.02465, 0.0, 0.04269, -0.02464, 0.0, -0.01307, 0.00754, -0.00022, -0.00077, -0.00043, -0.00369, -0.00077, -0.00043, 0.00369, -0.0064, -0.00401, 0.00011, 0.02102, 0.01054, 0.0, 0.02305, 0.01172, 0.0, -0.0064, -0.00401, -0.00011, 0.00074, -0.00043, 0.0, 0.00074, -0.00043, -0.0, 0.00027, 0.00754, -0.00011, 0.00138, -0.02582, -0.0, 0.00138, -0.02347, -0.0, 0.00027, 0.00754, 0.00011, -1e-05, 0.00088, 0.00369, -1e-05, 0.00088, -0.00369, 0.00522, -0.00324, 0.0001, 0.00542, -0.00291, 0.0001, 0.00542, -0.00291, -0.0001, 0.00522, -0.00324, -0.0001, 0.00162, 8e-05, 6e-05, 0.00434, 0.00187, 4e-05, 0.00434, 0.00187, -4e-05, 0.00162, 8e-05, -6e-05, 0.00055, -0.00469, -4e-05, 0.00074, -0.00145, -6e-05, 0.00074, -0.00145, 6e-05, 0.00055, -0.00469, 4e-05, 0.00522, -0.00324, 0.0001, 0.00542, -0.00291, 0.0001, 0.00542, -0.00291, -0.0001, 0.00522, -0.00324, -0.0001, 0.00162, 8e-05, 6e-05, 0.00434, 0.00187, 4e-05, 0.00434, 0.00187, -4e-05, 0.00162, 8e-05, -6e-05, 0.00055, -0.00469, -4e-05, 0.00074, -0.00145, -6e-05, 0.00074, -0.00145, 6e-05, 0.00055, -0.00469, 4e-05, -0.04234, 0.02445, -0.00037, -0.04234, 0.02445, 0.00037, -0.02158, -0.01151, -0.00019, -0.02158, -0.01151, 0.00019, -0.00082, 0.02445, 0.00019, -0.00082, 0.02445, -0.00019, -0.04234, 0.02445, -0.00037, -0.04234, 0.02445, 0.00037, -0.02158, -0.01151, -0.00019, -0.02158, -0.01151, 0.00019, -0.00082, 0.02445, 0.00019, -0.00082, 0.02445, -0.00019], [-0.00236, 0.00136, 0.01961, -0.02033, 0.01111, -0.0, -0.01979, 0.01205, -0.0, -0.00236, 0.00136, -0.01961, 0.0006, -0.00337, 0.00269, 0.0006, -0.00337, -0.00269, 0.00151, -0.00534, 0.0098, -0.00754, -0.00994, -0.0, -0.00795, -0.00956, -0.0, 0.00151, -0.00534, -0.0098, 0.00584, -0.00337, 0.0, 0.00584, -0.00337, -0.0, 0.00539, 0.00136, -0.0098, 0.0043, 0.01167, 0.0, 0.00484, 0.0115, 0.0, 0.00539, 0.00136, 0.0098, 0.00322, 0.00117, -0.00269, 0.00322, 0.00117, 0.00269, -0.00236, 0.00136, 0.01961, -0.02033, 0.01111, 0.0, -0.01979, 0.01205, 0.0, -0.00236, 0.00136, -0.01961, 0.0006, -0.00337, 0.00269, 0.0006, -0.00337, -0.00269, 0.00151, -0.00534, 0.0098, -0.00754, -0.00994, 0.0, -0.00795, -0.00956, 0.0, 0.00151, -0.00534, -0.0098, 0.00584, -0.00337, -0.0, 0.00584, -0.00337, 0.0, 0.00539, 0.00136, -0.0098, 0.0043, 0.01167, 0.0, 0.00484, 0.0115, 0.0, 0.00539, 0.00136, 0.0098, 0.00322, 0.00117, -0.00269, 0.00322, 0.00117, 0.00269, 0.00367, -0.00093, 0.00401, 0.00264, -0.00271, 0.00401, 0.00264, -0.00271, -0.00401, 0.00367, -0.00093, -0.00401, 0.00423, -0.00348, 0.00236, 0.00457, -0.00447, 0.00165, 0.00457, -0.00447, -0.00165, 0.00423, -0.00348, -0.00236, 0.00616, -0.00172, -0.00165, 0.00513, -0.00192, -0.00236, 0.00513, -0.00192, 0.00236, 0.00616, -0.00172, 0.00165, 0.00367, -0.00093, 0.00401, 0.00264, -0.00271, 0.00401, 0.00264, -0.00271, -0.00401, 0.00367, -0.00093, -0.00401, 0.00423, -0.00348, 0.00236, 0.00457, -0.00447, 0.00165, 0.00457, -0.00447, -0.00165, 0.00423, -0.00348, -0.00236, 0.00616, -0.00172, -0.00165, 0.00513, -0.00192, -0.00236, 0.00513, -0.00192, 0.00236, 0.00616, -0.00172, 0.00165, -0.01883, 0.01087, 0.04109, -0.01883, 0.01087, -0.04109, -0.00579, -0.01171, 0.02054, -0.00579, -0.01171, -0.02054, 0.00725, 0.01087, -0.02054, 0.00725, 0.01087, 0.02054, -0.01883, 0.01087, 0.04109, -0.01883, 0.01087, -0.04109, -0.00579, -0.01171, 0.02054, -0.00579, -0.01171, -0.02054, 0.00725, 0.01087, -0.02054, 0.00725, 0.01087, 0.02054]]), "E1u6":array([[0.00597, -0.00412, -0.00044, -0.02112, 0.01161, 0.0, -0.0197, 0.01093, 0.0, 0.00597, -0.00412, 0.00044, 0.00067, -0.00041, 0.00349, 0.00067, -0.00041, -0.00349, 0.01252, 0.00723, -0.00089, -0.04032, -0.02341, -0.0, -0.04043, -0.02321, -0.0, 0.01252, 0.00723, 0.00089, -2e-05, 0.00078, 0.00349, -2e-05, 0.00078, -0.00349, -0.00058, 0.00723, -0.00044, -0.00039, -0.02253, -0.0, -0.0005, -0.02409, -0.0, -0.00058, 0.00723, 0.00044, -0.00071, -0.00041, -0.0, -0.00071, -0.00041, 0.0, 0.00597, -0.00412, -0.00044, -0.02112, 0.01161, 0.0, -0.0197, 0.01093, 0.0, 0.00597, -0.00412, 0.00044, 0.00067, -0.00041, 0.00349, 0.00067, -0.00041, -0.00349, 0.01252, 0.00723, -0.00089, -0.04032, -0.02341, -0.0, -0.04043, -0.02321, -0.0, 0.01252, 0.00723, 0.00089, -2e-05, 0.00078, 0.00349, -2e-05, 0.00078, -0.00349, -0.00058, 0.00723, -0.00044, -0.00039, -0.02253, -0.0, -0.0005, -0.02409, -0.0, -0.00058, 0.00723, 0.00044, -0.00071, -0.00041, 0.0, -0.00071, -0.00041, -0.0, -0.00419, 0.00193, -0.00019, -0.00182, 0.00045, -4e-05, -0.00182, 0.00045, 4e-05, -0.00419, 0.00193, 0.00019, -0.00558, -0.00311, -0.00023, -0.00549, -0.00328, -0.00023, -0.00549, -0.00328, 0.00023, -0.00558, -0.00311, 0.00023, -0.00052, -0.0018, -4e-05, -0.00043, -0.00459, -0.00019, -0.00043, -0.00459, 0.00019, -0.00052, -0.0018, 4e-05, -0.00419, 0.00193, -0.00019, -0.00182, 0.00045, -4e-05, -0.00182, 0.00045, 4e-05, -0.00419, 0.00193, 0.00019, -0.00558, -0.00311, -0.00023, -0.00549, -0.00328, -0.00023, -0.00549, -0.00328, 0.00023, -0.00558, -0.00311, 0.00023, -0.00052, -0.0018, -4e-05, -0.00043, -0.00459, -0.00019, -0.00043, -0.00459, 0.00019, -0.00052, -0.0018, 4e-05, 0.02047, -0.01216, -0.00059, 0.02047, -0.01216, 0.00059, 0.04124, 0.02381, -0.00118, 0.04124, 0.02381, 0.00118, -0.0003, 0.02381, -0.00059, -0.0003, 0.02381, 0.00059, 0.02047, -0.01216, -0.00059, 0.02047, -0.01216, 0.00059, 0.04124, 0.02381, -0.00118, 0.04124, 0.02381, 0.00118, -0.0003, 0.02381, -0.00059, -0.0003, 0.02381, 0.00059], [0.00087, 0.00503, 0.01026, -0.0068, 0.01049, 0.0, -0.00771, 0.0119, 0.0, 0.00087, 0.00503, -0.01026, 0.0001, 0.00321, 0.00207, 0.0001, 0.00321, -0.00207, -0.00305, -0.00176, 0.02052, -0.0202, -0.01255, -0.0, -0.02097, -0.01122, -0.0, -0.00305, -0.00176, -0.02052, 0.00283, -0.00152, 0.00207, 0.00283, -0.00152, -0.00207, 0.0048, -0.00176, 0.01026, 0.00645, -0.01263, 0.0, 0.00568, -0.01114, -0.0, 0.0048, -0.00176, -0.01026, 0.00557, 0.00321, -0.0, 0.00557, 0.00321, 0.0, 0.00087, 0.00503, 0.01026, -0.0068, 0.01049, 0.0, -0.00771, 0.0119, 0.0, 0.00087, 0.00503, -0.01026, 0.0001, 0.00321, 0.00207, 0.0001, 0.00321, -0.00207, -0.00305, -0.00176, 0.02052, -0.0202, -0.01255, -0.0, -0.02097, -0.01122, -0.0, -0.00305, -0.00176, -0.02052, 0.00283, -0.00152, 0.00207, 0.00283, -0.00152, -0.00207, 0.0048, -0.00176, 0.01026, 0.00645, -0.01263, -0.0, 0.00568, -0.01114, 0.0, 0.0048, -0.00176, -0.01026, 0.00557, 0.00321, 0.0, 0.00557, 0.00321, -0.0, 0.00701, 0.00315, 0.00101, 0.00265, 0.00408, 0.00256, 0.00265, 0.00408, -0.00256, 0.00701, 0.00315, -0.00101, 0.00344, 0.00356, 0.00357, 0.0048, 0.0012, 0.00357, 0.0048, 0.0012, -0.00357, 0.00344, 0.00356, -0.00357, 0.00486, 0.00026, 0.00256, 0.00623, 0.0045, 0.00101, 0.00623, 0.0045, -0.00101, 0.00486, 0.00026, -0.00256, 0.00701, 0.00315, 0.00101, 0.00265, 0.00408, 0.00256, 0.00265, 0.00408, -0.00256, 0.00701, 0.00315, -0.00101, 0.00344, 0.00356, 0.00357, 0.0048, 0.0012, 0.00357, 0.0048, 0.0012, -0.00357, 0.00344, 0.00356, -0.00357, 0.00486, 0.00026, 0.00256, 0.00623, 0.0045, 0.00101, 0.00623, 0.0045, -0.00101, 0.00486, 0.00026, -0.00256, -0.00611, 0.00846, 0.0211, -0.00611, 0.00846, -0.0211, -0.01649, -0.00952, 0.04219, -0.01649, -0.00952, -0.04219, 0.00427, -0.00952, 0.0211, 0.00427, -0.00952, -0.0211, -0.00611, 0.00846, 0.0211, -0.00611, 0.00846, -0.0211, -0.01649, -0.00952, 0.04219, -0.01649, -0.00952, -0.04219, 0.00427, -0.00952, 0.0211, 0.00427, -0.00952, -0.0211]]), "E2u1":array([[-0.01014, 0.01314, -0.00033, 0.0, -0.0, -0.01154, 0.0, -0.0, -0.00685, 0.01014, -0.01314, -0.00033, 0.0, 0.00582, -0.0, 0.0, -0.00582, 0.0, 0.01014, 0.01314, 0.00033, 0.0, 0.0, 0.00685, 0.0, 0.0, 0.01154, -0.01014, -0.01314, 0.00033, 0.0109, 0.01306, 0.00909, -0.0109, -0.01306, 0.00909, 0.01261, 0.0, -0.0, 0.0, 0.0, 0.0047, -0.0, 0.0, -0.0047, -0.01261, 0.0, 0.0, 0.0109, -0.01306, -0.00909, -0.0109, 0.01306, -0.00909, -0.01014, 0.01314, -0.00033, -0.0, 0.0, -0.01154, -0.0, 0.0, -0.00685, 0.01014, -0.01314, -0.00033, -0.0, 0.00582, 0.0, -0.0, -0.00582, -0.0, 0.01014, 0.01314, 0.00033, -0.0, -0.0, 0.00685, -0.0, -0.0, 0.01154, -0.01014, -0.01314, 0.00033, 0.0109, 0.01306, 0.00909, -0.0109, -0.01306, 0.00909, 0.01261, -0.0, 0.0, -0.0, -0.0, 0.0047, 0.0, -0.0, -0.0047, -0.01261, -0.0, -0.0, 0.0109, -0.01306, -0.00909, -0.0109, 0.01306, -0.00909, -0.00975, 0.01028, -0.0049, -0.00333, 0.00655, -0.00181, 0.00333, -0.00655, -0.00181, 0.00975, -0.01028, -0.0049, 0.00333, 0.00655, 0.00181, 0.00975, 0.01028, 0.0049, -0.00975, -0.01028, 0.0049, -0.00333, -0.00655, 0.00181, 0.00804, 0.00742, 0.00309, 0.00804, -0.00742, -0.00309, -0.00804, 0.00742, -0.00309, -0.00804, -0.00742, 0.00309, -0.00975, 0.01028, -0.0049, -0.00333, 0.00655, -0.00181, 0.00333, -0.00655, -0.00181, 0.00975, -0.01028, -0.0049, 0.00333, 0.00655, 0.00181, 0.00975, 0.01028, 0.0049, -0.00975, -0.01028, 0.0049, -0.00333, -0.00655, 0.00181, 0.00804, 0.00742, 0.00309, 0.00804, -0.00742, -0.00309, -0.00804, 0.00742, -0.00309, -0.00804, -0.00742, 0.00309, -0.01604, 0.02055, 0.00424, 0.01604, -0.02055, 0.00424, 0.01604, 0.02055, -0.00424, -0.01604, -0.02055, -0.00424, 0.01954, 0.0, -0.0, -0.01954, -0.0, -0.0, -0.01604, 0.02055, 0.00424, 0.01604, -0.02055, 0.00424, 0.01604, 0.02055, -0.00424, -0.01604, -0.02055, -0.00424, 0.01954, -0.0, 0.0, -0.01954, -0.0, 0.0], [-0.01112, 0.00269, 0.01521, 0.0, -0.0, 0.00958, 0.0, -0.0, 0.0066, 0.01112, -0.00269, 0.01521, -0.0, -0.00548, -0.0, -0.0, 0.00548, 0.0, 0.01112, 0.00269, -0.01521, 0.0, -0.0, -0.0066, 0.0, 0.0, -0.00958, -0.01112, -0.00269, -0.01521, -0.00683, -0.00634, -0.00792, 0.00683, 0.00634, -0.00792, -0.00645, -0.0, -0.0, 0.0, 0.0, -0.00298, -0.0, 0.0, 0.00298, 0.00645, 0.0, 0.0, -0.00683, 0.00634, 0.00792, 0.00683, -0.00634, 0.00792, -0.01112, 0.00269, 0.01521, -0.0, 0.0, 0.00958, -0.0, 0.0, 0.0066, 0.01112, -0.00269, 0.01521, 0.0, -0.00548, 0.0, 0.0, 0.00548, -0.0, 0.01112, 0.00269, -0.01521, -0.0, 0.0, -0.0066, -0.0, -0.0, -0.00958, -0.01112, -0.00269, -0.01521, -0.00683, -0.00634, -0.00792, 0.00683, 0.00634, -0.00792, -0.00645, 0.0, 0.0, -0.0, -0.0, -0.00298, 0.0, -0.0, 0.00298, 0.00645, -0.0, -0.0, -0.00683, 0.00634, 0.00792, 0.00683, -0.00634, 0.00792, -0.00302, -0.00114, 0.01233, -0.01363, 0.00635, 0.01424, 0.01363, -0.00635, 0.01424, 0.00302, 0.00114, 0.01233, 0.01363, 0.00635, -0.01424, 0.00302, -0.00114, -0.01233, -0.00302, 0.00114, -0.01233, -0.01363, -0.00635, -0.01424, -0.00381, -0.01293, 0.00192, -0.00381, 0.01293, -0.00192, 0.00381, -0.01293, -0.00192, 0.00381, 0.01293, 0.00192, -0.00302, -0.00114, 0.01233, -0.01363, 0.00635, 0.01424, 0.01363, -0.00635, 0.01424, 0.00302, 0.00114, 0.01233, 0.01363, 0.00635, -0.01424, 0.00302, -0.00114, -0.01233, -0.00302, 0.00114, -0.01233, -0.01363, -0.00635, -0.01424, -0.00381, -0.01293, 0.00192, -0.00381, 0.01293, -0.00192, 0.00381, -0.01293, -0.00192, 0.00381, 0.01293, 0.00192, -0.01753, 0.00381, 0.01818, 0.01753, -0.00381, 0.01818, 0.01753, 0.00381, -0.01818, -0.01753, -0.00381, -0.01818, -0.01094, -0.0, -0.0, 0.01094, -0.0, 0.0, -0.01753, 0.00381, 0.01818, 0.01753, -0.00381, 0.01818, 0.01753, 0.00381, -0.01818, -0.01753, -0.00381, -0.01818, -0.01094, 0.0, 0.0, 0.01094, 0.0, -0.0]]), "E2u2":array([[0.00652, 0.01129, -0.0, -0.0, -0.0, -0.00494, 0.0, 0.0, 0.00494, -0.00652, -0.01129, 0.0, 0.0176, 0.0031, 0.00942, -0.0176, -0.0031, 0.00942, 0.01671, 0.00212, 0.00059, -0.0, -0.0, 0.01196, -0.0, 0.0, 0.00703, -0.01671, -0.00212, 0.00059, 0.00538, -0.0031, -0.0, -0.00538, 0.0031, -0.0, 0.00652, -0.01553, -0.00059, 0.0, -0.0, -0.00703, -0.0, -0.0, -0.01196, -0.00652, 0.01553, -0.00059, 0.00611, -0.01679, -0.00942, -0.00611, 0.01679, -0.00942, 0.00652, 0.01129, 0.0, 0.0, 0.0, -0.00494, -0.0, -0.0, 0.00494, -0.00652, -0.01129, -0.0, 0.0176, 0.0031, 0.00942, -0.0176, -0.0031, 0.00942, 0.01671, 0.00212, 0.00059, 0.0, 0.0, 0.01196, 0.0, -0.0, 0.00703, -0.01671, -0.00212, 0.00059, 0.00538, -0.0031, 0.0, -0.00538, 0.0031, -0.0, 0.00652, -0.01553, -0.00059, -0.0, 0.0, -0.00703, 0.0, 0.0, -0.01196, -0.00652, 0.01553, -0.00059, 0.00611, -0.01679, -0.00942, -0.00611, 0.01679, -0.00942, -0.00287, 0.01134, -0.00291, 0.01125, 0.00318, 0.00291, -0.01125, -0.00318, 0.00291, 0.00287, -0.01134, -0.00291, 0.01404, 0.00326, 0.00515, 0.00697, -0.0008, 0.00224, -0.00697, 0.0008, 0.00224, -0.01404, -0.00326, 0.00515, 0.00418, -0.00564, -0.00224, 0.0042, -0.01379, -0.00515, -0.0042, 0.01379, -0.00515, -0.00418, 0.00564, -0.00224, -0.00287, 0.01134, -0.00291, 0.01125, 0.00318, 0.00291, -0.01125, -0.00318, 0.00291, 0.00287, -0.01134, -0.00291, 0.01404, 0.00326, 0.00515, 0.00697, -0.0008, 0.00224, -0.00697, 0.0008, 0.00224, -0.01404, -0.00326, 0.00515, 0.00418, -0.00564, -0.00224, 0.0042, -0.01379, -0.00515, -0.0042, 0.01379, -0.00515, -0.00418, 0.00564, -0.00224, 0.01011, 0.01752, 0.0, -0.01011, -0.01752, -0.0, 0.02656, 0.00366, -0.0041, -0.02656, -0.00366, -0.0041, 0.01011, -0.02483, 0.0041, -0.01011, 0.02483, 0.0041, 0.01011, 0.01752, -0.0, -0.01011, -0.01752, 0.0, 0.02656, 0.00366, -0.0041, -0.02656, -0.00366, -0.0041, 0.01011, -0.02483, 0.0041, -0.01011, 0.02483, 0.0041], [-0.00326, -0.00565, 0.0, 0.0, -0.0, 0.00265, 0.0, 0.0, -0.00265, 0.00326, 0.00565, 0.0, -0.00792, -0.00264, -0.00818, 0.00792, 0.00264, -0.00818, 0.00806, 0.00842, -0.01508, 0.0, -0.0, -0.00989, 0.0, 0.0, -0.00724, -0.00806, -0.00842, -0.01508, -0.00458, 0.00264, -0.0, 0.00458, -0.00264, 0.0, -0.00326, -0.01119, 0.01508, 0.0, 0.0, 0.00724, -0.0, 0.0, 0.00989, 0.00326, 0.01119, 0.01508, -0.00167, 0.00818, 0.00818, 0.00167, -0.00818, 0.00818, -0.00326, -0.00565, -0.0, -0.0, 0.0, 0.00265, -0.0, -0.0, -0.00265, 0.00326, 0.00565, -0.0, -0.00792, -0.00264, -0.00818, 0.00792, 0.00264, -0.00818, 0.00806, 0.00842, -0.01508, -0.0, 0.0, -0.00989, -0.0, -0.0, -0.00724, -0.00806, -0.00842, -0.01508, -0.00458, 0.00264, 0.0, 0.00458, -0.00264, -0.0, -0.00326, -0.01119, 0.01508, -0.0, -0.0, 0.00724, 0.0, -0.0, 0.00989, 0.00326, 0.01119, 0.01508, -0.00167, 0.00818, 0.00818, 0.00167, -0.00818, 0.00818, 0.00828, -0.00918, -0.00157, -0.01209, 0.00258, 0.00157, 0.01209, -0.00258, 0.00157, -0.00828, 0.00918, -0.00157, 0.00084, 0.00345, -0.01243, 0.01169, 0.00818, -0.01401, -0.01169, -0.00818, -0.01401, -0.00084, -0.00345, -0.01243, -0.00124, -0.01421, 0.01401, -0.00257, -0.00245, 0.01243, 0.00257, 0.00245, 0.01243, 0.00124, 0.01421, 0.01401, 0.00828, -0.00918, -0.00157, -0.01209, 0.00258, 0.00157, 0.01209, -0.00258, 0.00157, -0.00828, 0.00918, -0.00157, 0.00084, 0.00345, -0.01243, 0.01169, 0.00818, -0.01401, -0.01169, -0.00818, -0.01401, -0.00084, -0.00345, -0.01243, -0.00124, -0.01421, 0.01401, -0.00257, -0.00245, 0.01243, 0.00257, 0.00245, 0.01243, 0.00124, 0.01421, 0.01401, -0.00523, -0.00906, -0.0, 0.00523, 0.00906, 0.0, 0.01251, 0.01327, -0.01871, -0.01251, -0.01327, -0.01871, -0.00523, -0.01747, 0.01871, 0.00523, 0.01747, 0.01871, -0.00523, -0.00906, 0.0, 0.00523, 0.00906, -0.0, 0.01251, 0.01327, -0.01871, -0.01251, -0.01327, -0.01871, -0.00523, -0.01747, 0.01871, 0.00523, 0.01747, 0.01871]]), "E2u3":array([[0.01698, -0.00244, 0.0003, -0.0, 0.0, 0.00678, -0.0, 0.0, 0.01176, -0.01698, 0.00244, 0.0003, 0.01737, -0.00311, 0.00934, -0.01737, 0.00311, 0.00934, 0.00638, -0.01105, -0.0, -0.0, -0.0, 0.00498, -0.0, -0.0, -0.00498, -0.00638, 0.01105, 0.0, -0.00599, -0.0166, -0.00934, 0.00599, 0.0166, -0.00934, -0.00638, -0.01593, -0.0003, -0.0, -0.0, -0.01176, -0.0, -0.0, -0.00678, 0.00638, 0.01593, -0.0003, -0.00539, -0.00311, -0.0, 0.00539, 0.00311, 0.0, 0.01698, -0.00244, 0.0003, 0.0, -0.0, 0.00678, 0.0, -0.0, 0.01176, -0.01698, 0.00244, 0.0003, 0.01737, -0.00311, 0.00934, -0.01737, 0.00311, 0.00934, 0.00638, -0.01105, 0.0, 0.0, 0.0, 0.00498, 0.0, 0.0, -0.00498, -0.00638, 0.01105, -0.0, -0.00599, -0.0166, -0.00934, 0.00599, 0.0166, -0.00934, -0.00638, -0.01593, -0.0003, 0.0, 0.0, -0.01176, 0.0, 0.0, -0.00678, 0.00638, 0.01593, -0.0003, -0.00539, -0.00311, 0.0, 0.00539, 0.00311, -0.0, 0.00743, 0.00061, 0.00185, 0.01409, -0.00349, 0.00494, -0.01409, 0.00349, 0.00494, -0.00743, -0.00061, 0.00185, 0.01091, -0.00325, 0.00309, -0.00264, -0.01107, -0.00309, 0.00264, 0.01107, -0.00309, -0.01091, 0.00325, 0.00309, -0.00402, -0.01395, -0.00494, -0.00424, -0.00613, -0.00185, 0.00424, 0.00613, -0.00185, 0.00402, 0.01395, -0.00494, 0.00743, 0.00061, 0.00185, 0.01409, -0.00349, 0.00494, -0.01409, 0.00349, 0.00494, -0.00743, -0.00061, 0.00185, 0.01091, -0.00325, 0.00309, -0.00264, -0.01107, -0.00309, 0.00264, 0.01107, -0.00309, -0.01091, 0.00325, 0.00309, -0.00402, -0.01395, -0.00494, -0.00424, -0.00613, -0.00185, 0.00424, 0.00613, -0.00185, 0.00402, 0.01395, -0.00494, 0.02712, -0.00428, -0.00444, -0.02712, 0.00428, -0.00444, 0.00985, -0.01706, 0.0, -0.00985, 0.01706, -0.0, -0.00985, -0.02563, 0.00444, 0.00985, 0.02563, 0.00444, 0.02712, -0.00428, -0.00444, -0.02712, 0.00428, -0.00444, 0.00985, -0.01706, -0.0, -0.00985, 0.01706, 0.0, -0.00985, -0.02563, 0.00444, 0.00985, 0.02563, 0.00444], [-0.00739, 0.00809, 0.01513, -0.0, -0.0, 0.00749, 0.0, -0.0, 0.01033, 0.00739, -0.00809, 0.01513, 0.00771, -0.00297, 0.00851, -0.00771, 0.00297, 0.00851, 0.00331, -0.00574, -0.0, 0.0, 0.0, 0.00284, 0.0, 0.0, -0.00284, -0.00331, 0.00574, 0.0, -0.00128, -0.00816, -0.00851, 0.00128, 0.00816, -0.00851, -0.00331, 0.01044, -0.01513, -0.0, -0.0, -0.01033, -0.0, 0.0, -0.00749, 0.00331, -0.01044, -0.01513, -0.00514, -0.00297, 0.0, 0.00514, 0.00297, 0.0, -0.00739, 0.00809, 0.01513, 0.0, 0.0, 0.00749, -0.0, 0.0, 0.01033, 0.00739, -0.00809, 0.01513, 0.00771, -0.00297, 0.00851, -0.00771, 0.00297, 0.00851, 0.00331, -0.00574, 0.0, -0.0, -0.0, 0.00284, -0.0, -0.0, -0.00284, -0.00331, 0.00574, -0.0, -0.00128, -0.00816, -0.00851, 0.00128, 0.00816, -0.00851, -0.00331, 0.01044, -0.01513, 0.0, 0.0, -0.01033, 0.0, -0.0, -0.00749, 0.00331, -0.01044, -0.01513, -0.00514, -0.00297, -0.0, 0.00514, 0.00297, -0.0, -0.0122, 0.0087, 0.01489, -0.00156, 0.00357, 0.01335, 0.00156, -0.00357, 0.01335, 0.0122, -0.0087, 0.01489, 0.01208, 0.00264, -0.00154, -0.00833, -0.00914, 0.00154, 0.00833, 0.00914, 0.00154, -0.01208, -0.00264, -0.00154, -0.00232, 0.00313, -0.01335, -0.00143, 0.01492, -0.01489, 0.00143, -0.01492, -0.01489, 0.00232, -0.00313, -0.01335, -0.0122, 0.0087, 0.01489, -0.00156, 0.00357, 0.01335, 0.00156, -0.00357, 0.01335, 0.0122, -0.0087, 0.01489, 0.01208, 0.00264, -0.00154, -0.00833, -0.00914, 0.00154, 0.00833, 0.00914, 0.00154, -0.01208, -0.00264, -0.00154, -0.00232, 0.00313, -0.01335, -0.00143, 0.01492, -0.01489, 0.00143, -0.01492, -0.01489, 0.00232, -0.00313, -0.01335, -0.0109, 0.01232, 0.01855, 0.0109, -0.01232, 0.01855, 0.00521, -0.00903, -0.0, -0.00521, 0.00903, 0.0, -0.00521, 0.0156, -0.01855, 0.00521, -0.0156, -0.01855, -0.0109, 0.01232, 0.01855, 0.0109, -0.01232, 0.01855, 0.00521, -0.00903, 0.0, -0.00521, 0.00903, -0.0, -0.00521, 0.0156, -0.01855, 0.00521, -0.0156, -0.01855]]), "E2u4":array([[0.01391, 0.00533, 0.0003, -0.0, 0.0, 0.00113, -0.0, 0.0, 0.01009, -0.01391, -0.00533, 0.0003, 0.02103, -0.0, 0.01124, -0.02103, -0.0, 0.01124, 0.01391, -0.00533, 0.0003, -0.0, -0.0, 0.01009, -0.0, -0.0, 0.00113, -0.01391, 0.00533, 0.0003, -0.00037, -0.01193, -0.00562, 0.00037, 0.01193, -0.00562, 0.0, -0.01875, -0.0006, 0.0, -0.0, -0.01122, -0.0, -0.0, -0.01122, -0.0, 0.01875, -0.0006, 0.00037, -0.01193, -0.00562, -0.00037, 0.01193, -0.00562, 0.01391, 0.00533, 0.0003, 0.0, -0.0, 0.00113, 0.0, -0.0, 0.01009, -0.01391, -0.00533, 0.0003, 0.02103, -0.0, 0.01124, -0.02103, 0.0, 0.01124, 0.01391, -0.00533, 0.0003, 0.0, 0.0, 0.01009, 0.0, 0.0, 0.00113, -0.01391, 0.00533, 0.0003, -0.00037, -0.01193, -0.00562, 0.00037, 0.01193, -0.00562, -0.0, -0.01875, -0.0006, -0.0, 0.0, -0.01122, 0.0, 0.0, -0.01122, 0.0, 0.01875, -0.0006, 0.00037, -0.01193, -0.00562, -0.00037, 0.01193, -0.00562, 0.00251, 0.00725, -0.00047, 0.0151, -0.00011, 0.00478, -0.0151, 0.00011, 0.00478, -0.00251, -0.00725, -0.00047, 0.0151, 0.00011, 0.00478, 0.00251, -0.00725, -0.00047, -0.00251, 0.00725, -0.00047, -0.0151, -0.00011, 0.00478, 7e-05, -0.01167, -0.00431, -7e-05, -0.01167, -0.00431, 7e-05, 0.01167, -0.00431, -7e-05, 0.01167, -0.00431, 0.00251, 0.00725, -0.00047, 0.0151, -0.00011, 0.00478, -0.0151, 0.00011, 0.00478, -0.00251, -0.00725, -0.00047, 0.0151, 0.00011, 0.00478, 0.00251, -0.00725, -0.00047, -0.00251, 0.00725, -0.00047, -0.0151, -0.00011, 0.00478, 7e-05, -0.01167, -0.00431, -7e-05, -0.01167, -0.00431, 7e-05, 0.01167, -0.00431, -7e-05, 0.01167, -0.00431, 0.02205, 0.00795, -0.00252, -0.02205, -0.00795, -0.00252, 0.02205, -0.00795, -0.00252, -0.02205, 0.00795, -0.00252, -0.0, -0.03024, 0.00504, 0.0, 0.03024, 0.00504, 0.02205, 0.00795, -0.00252, -0.02205, -0.00795, -0.00252, 0.02205, -0.00795, -0.00252, -0.02205, 0.00795, -0.00252, 0.0, -0.03024, 0.00504, 0.0, 0.03024, 0.00504], [-0.00259, 0.00832, 0.00891, -0.0, 0.0, 0.00296, -0.0, -0.0, 0.00763, 0.00259, -0.00832, 0.00891, 0.00886, -0.0, 0.01001, -0.00886, -0.0, 0.01001, -0.00259, -0.00832, 0.00891, -0.0, -0.0, 0.00763, -0.0, -0.0, 0.00296, 0.00259, 0.00832, 0.00891, 0.00207, -0.00631, -0.005, -0.00207, 0.00631, -0.005, 0.0, 0.01282, -0.01782, -0.0, -0.0, -0.01059, 0.0, -0.0, -0.01059, 0.0, -0.01282, -0.01782, -0.00207, -0.00631, -0.005, 0.00207, 0.00631, -0.005, -0.00259, 0.00832, 0.00891, 0.0, -0.0, 0.00296, 0.0, 0.0, 0.00763, 0.00259, -0.00832, 0.00891, 0.00886, 0.0, 0.01001, -0.00886, 0.0, 0.01001, -0.00259, -0.00832, 0.00891, 0.0, 0.0, 0.00763, 0.0, 0.0, 0.00296, 0.00259, 0.00832, 0.00891, 0.00207, -0.00631, -0.005, -0.00207, 0.00631, -0.005, -0.0, 0.01282, -0.01782, 0.0, 0.0, -0.01059, -0.0, 0.0, -0.01059, -0.0, -0.01282, -0.01782, -0.00207, -0.00631, -0.005, 0.00207, 0.00631, -0.005, -0.0117, 0.01024, 0.00942, 0.00617, 0.00072, 0.00684, -0.00617, -0.00072, 0.00684, 0.0117, -0.01024, 0.00942, 0.00617, -0.00072, 0.00684, -0.0117, -0.01024, 0.00942, 0.0117, 0.01024, 0.00942, -0.00617, 0.00072, 0.00684, -0.00069, 0.01028, -0.01626, 0.00069, 0.01028, -0.01626, -0.00069, -0.01028, -0.01626, 0.00069, -0.01028, -0.01626, -0.0117, 0.01024, 0.00942, 0.00617, 0.00072, 0.00684, -0.00617, -0.00072, 0.00684, 0.0117, -0.01024, 0.00942, 0.00617, -0.00072, 0.00684, -0.0117, -0.01024, 0.00942, 0.0117, 0.01024, 0.00942, -0.00617, 0.00072, 0.00684, -0.00069, 0.01028, -0.01626, 0.00069, 0.01028, -0.01626, -0.00069, -0.01028, -0.01626, 0.00069, -0.01028, -0.01626, -0.00389, 0.01288, 0.01111, 0.00389, -0.01288, 0.01111, -0.00389, -0.01288, 0.01111, 0.00389, 0.01288, 0.01111, 0.0, 0.01961, -0.02223, -0.0, -0.01961, -0.02223, -0.00389, 0.01288, 0.01111, 0.00389, -0.01288, 0.01111, -0.00389, -0.01288, 0.01111, 0.00389, 0.01288, 0.01111, -0.0, 0.01961, -0.02223, 0.0, -0.01961, -0.02223]]), "E2u5":array([[0.0155, -0.00895, 0.00023, -0.0, 0.0, 0.01042, -0.0, 0.0, 0.01042, -0.0155, 0.00895, 0.00023, 0.0096, -0.00508, 0.00521, -0.0096, 0.00508, 0.00521, -0.00238, -0.01377, -0.00011, -0.0, -0.0, -0.00111, -0.0, -0.0, -0.00931, 0.00238, 0.01377, -0.00011, -0.0096, -0.01663, -0.01042, 0.0096, 0.01663, -0.01042, -0.01073, -0.00895, -0.00011, -0.0, -0.0, -0.00931, 0.0, -0.0, -0.00111, 0.01073, 0.00895, -0.00011, -0.0092, 0.00578, 0.00521, 0.0092, -0.00578, 0.00521, 0.0155, -0.00895, 0.00023, 0.0, -0.0, 0.01042, 0.0, -0.0, 0.01042, -0.0155, 0.00895, 0.00023, 0.0096, -0.00508, 0.00521, -0.0096, 0.00508, 0.00521, -0.00238, -0.01377, -0.00011, 0.0, 0.0, -0.00111, 0.0, 0.0, -0.00931, 0.00238, 0.01377, -0.00011, -0.0096, -0.01663, -0.01042, 0.0096, 0.01663, -0.01042, -0.01073, -0.00895, -0.00011, 0.0, 0.0, -0.00931, -0.0, 0.0, -0.00111, 0.01073, 0.00895, -0.00011, -0.0092, 0.00578, 0.00521, 0.0092, -0.00578, 0.00521, 0.00999, -0.00567, 0.0037, 0.00991, -0.00581, 0.0037, -0.00991, 0.00581, 0.0037, -0.00999, 0.00567, 0.0037, 0.00397, -0.00572, 0.00087, -0.00702, -0.01197, -0.00456, 0.00702, 0.01197, -0.00456, -0.00397, 0.00572, 0.00087, -0.00685, -0.01206, -0.00456, -0.00693, 0.00058, 0.00087, 0.00693, -0.00058, 0.00087, 0.00685, 0.01206, -0.00456, 0.00999, -0.00567, 0.0037, 0.00991, -0.00581, 0.0037, -0.00991, 0.00581, 0.0037, -0.00999, 0.00567, 0.0037, 0.00397, -0.00572, 0.00087, -0.00702, -0.01197, -0.00456, 0.00702, 0.01197, -0.00456, -0.00397, 0.00572, 0.00087, -0.00685, -0.01206, -0.00456, -0.00693, 0.00058, 0.00087, 0.00693, -0.00058, 0.00087, 0.00685, 0.01206, -0.00456, 0.02461, -0.01421, -0.00507, -0.02461, 0.01421, -0.00507, -0.00401, -0.02147, 0.00254, 0.00401, 0.02147, 0.00254, -0.01659, -0.01421, 0.00254, 0.01659, 0.01421, 0.00254, 0.02461, -0.01421, -0.00507, -0.02461, 0.01421, -0.00507, -0.00401, -0.02147, 0.00254, 0.00401, 0.02147, 0.00254, -0.01659, -0.01421, 0.00254, 0.01659, 0.01421, 0.00254], [-0.0103, 0.00594, 0.01727, 0.0, -0.0, 0.00962, 0.0, -0.0, 0.00962, 0.0103, -0.00594, 0.01727, 0.00481, -0.00498, 0.00467, -0.00481, 0.00498, 0.00467, 0.00794, -0.00186, -0.00863, 0.0, -0.0, -0.00224, -0.0, -0.0, -0.00738, -0.00794, 0.00186, -0.00863, -0.00481, -0.00832, -0.00935, 0.00481, 0.00832, -0.00935, -0.00557, 0.00594, -0.00863, -0.0, -0.0, -0.00738, -0.0, -0.0, -0.00224, 0.00557, -0.00594, -0.00863, -0.00672, 0.00167, 0.00467, 0.00672, -0.00167, 0.00467, -0.0103, 0.00594, 0.01727, -0.0, 0.0, 0.00962, -0.0, 0.0, 0.00962, 0.0103, -0.00594, 0.01727, 0.00481, -0.00498, 0.00467, -0.00481, 0.00498, 0.00467, 0.00794, -0.00186, -0.00863, -0.0, 0.0, -0.00224, 0.0, 0.0, -0.00738, -0.00794, 0.00186, -0.00863, -0.00481, -0.00832, -0.00935, 0.00481, 0.00832, -0.00935, -0.00557, 0.00594, -0.00863, 0.0, 0.0, -0.00738, 0.0, 0.0, -0.00224, 0.00557, -0.00594, -0.00863, -0.00672, 0.00167, 0.00467, 0.00672, -0.00167, 0.00467, -0.00895, 0.00457, 0.01574, -0.00843, 0.00546, 0.01574, 0.00843, -0.00546, 0.01574, 0.00895, -0.00457, 0.01574, 0.01508, 0.0053, -0.00934, -0.00319, -0.00583, -0.00639, 0.00319, 0.00583, -0.00639, -0.01508, -0.0053, -0.00934, -0.00346, -0.00568, -0.00639, -0.00295, 0.01571, -0.00934, 0.00295, -0.01571, -0.00934, 0.00346, 0.00568, -0.00639, -0.00895, 0.00457, 0.01574, -0.00843, 0.00546, 0.01574, 0.00843, -0.00546, 0.01574, 0.00895, -0.00457, 0.01574, 0.01508, 0.0053, -0.00934, -0.00319, -0.00583, -0.00639, 0.00319, 0.00583, -0.00639, -0.01508, -0.0053, -0.00934, -0.00346, -0.00568, -0.00639, -0.00295, 0.01571, -0.00934, 0.00295, -0.01571, -0.00934, 0.00346, 0.00568, -0.00639, -0.01565, 0.00903, 0.02068, 0.01565, -0.00903, 0.02068, 0.01242, -0.00344, -0.01034, -0.01242, 0.00344, -0.01034, -0.00919, 0.00903, -0.01034, 0.00919, -0.00903, -0.01034, -0.01565, 0.00903, 0.02068, 0.01565, -0.00903, 0.02068, 0.01242, -0.00344, -0.01034, -0.01242, 0.00344, -0.01034, -0.00919, 0.00903, -0.01034, 0.00919, -0.00903, -0.01034]]), "E2u6":array([[-0.0021, 0.01409, -0.0003, 0.0, -0.0, -0.00957, -0.0, 0.0, -0.00127, 0.0021, -0.01409, -0.0003, 0.00993, 0.00518, 0.00535, -0.00993, -0.00518, 0.00535, 0.01535, 0.00886, 0.0006, 0.0, -0.0, 0.01085, 0.0, 0.0, 0.01085, -0.01535, -0.00886, 0.0006, 0.00945, 0.006, 0.00535, -0.00945, -0.006, 0.00535, 0.01116, -0.00886, -0.0003, -0.0, -0.0, -0.00127, -0.0, 0.0, -0.00957, -0.01116, 0.00886, -0.0003, 0.00993, -0.01719, -0.01069, -0.00993, 0.01719, -0.01069, -0.0021, 0.01409, -0.0003, -0.0, 0.0, -0.00957, 0.0, -0.0, -0.00127, 0.0021, -0.01409, -0.0003, 0.00993, 0.00518, 0.00535, -0.00993, -0.00518, 0.00535, 0.01535, 0.00886, 0.0006, -0.0, 0.0, 0.01085, -0.0, -0.0, 0.01085, -0.01535, -0.00886, 0.0006, 0.00945, 0.006, 0.00535, -0.00945, -0.006, 0.00535, 0.01116, -0.00886, -0.0003, 0.0, 0.0, -0.00127, 0.0, -0.0, -0.00957, -0.01116, 0.00886, -0.0003, 0.00993, -0.01719, -0.01069, -0.00993, 0.01719, -0.01069, -0.00716, 0.01239, -0.00464, 0.00454, 0.00557, 0.00051, -0.00454, -0.00557, 0.00051, 0.00716, -0.01239, -0.00464, 0.00977, 0.00558, 0.00413, 0.00971, 0.00567, 0.00413, -0.00971, -0.00567, 0.00413, -0.00977, -0.00558, 0.00413, 0.0071, 0.00115, 0.00051, 0.00715, -0.0124, -0.00464, -0.00715, 0.0124, -0.00464, -0.0071, -0.00115, 0.00051, -0.00716, 0.01239, -0.00464, 0.00454, 0.00557, 0.00051, -0.00454, -0.00557, 0.00051, 0.00716, -0.01239, -0.00464, 0.00977, 0.00558, 0.00413, 0.00971, 0.00567, 0.00413, -0.00971, -0.00567, 0.00413, -0.00977, -0.00558, 0.00413, 0.0071, 0.00115, 0.00051, 0.00715, -0.0124, -0.00464, -0.00715, 0.0124, -0.00464, -0.0071, -0.00115, 0.00051, -0.00348, 0.022, 0.00237, 0.00348, -0.022, 0.00237, 0.02427, 0.01401, -0.00473, -0.02427, -0.01401, -0.00473, 0.01732, -0.01401, 0.00237, -0.01732, 0.01401, 0.00237, -0.00348, 0.022, 0.00237, 0.00348, -0.022, 0.00237, 0.02427, 0.01401, -0.00473, -0.02427, -0.01401, -0.00473, 0.01732, -0.01401, 0.00237, -0.01732, 0.01401, 0.00237], [-0.00816, -0.0015, 0.00848, 0.0, -0.0, 0.00693, 0.0, -0.0, 0.00219, 0.00816, 0.0015, 0.00848, -0.0048, -0.00442, -0.00443, 0.0048, 0.00442, -0.00443, 0.01094, 0.00632, -0.01696, 0.0, -0.0, -0.00912, 0.0, 0.0, -0.00912, -0.01094, -0.00632, -0.01696, -0.00623, -0.00195, -0.00443, 0.00623, 0.00195, -0.00443, -0.00538, -0.00632, 0.00848, 0.0, 0.0, 0.00219, -0.0, 0.0, 0.00693, 0.00538, 0.00632, 0.00848, -0.0048, 0.00832, 0.00887, 0.0048, -0.00832, 0.00887, -0.00816, -0.0015, 0.00848, -0.0, 0.0, 0.00693, -0.0, 0.0, 0.00219, 0.00816, 0.0015, 0.00848, -0.0048, -0.00442, -0.00443, 0.0048, 0.00442, -0.00443, 0.01094, 0.00632, -0.01696, -0.0, 0.0, -0.00912, -0.0, -0.0, -0.00912, -0.01094, -0.00632, -0.01696, -0.00623, -0.00195, -0.00443, 0.00623, 0.00195, -0.00443, -0.00538, -0.00632, 0.00848, -0.0, -0.0, 0.00219, 0.0, -0.0, 0.00693, 0.00538, 0.00632, 0.00848, -0.0048, 0.00832, 0.00887, 0.0048, -0.00832, 0.00887, 0.00327, -0.00602, 0.00575, -0.01455, 0.00512, 0.00882, 0.01455, -0.00512, 0.00882, -0.00327, 0.00602, 0.00575, 0.00769, 0.0053, -0.01457, 0.00843, 0.00401, -0.01457, -0.00843, -0.00401, -0.01457, -0.00769, -0.0053, -0.01457, -0.00284, -0.01516, 0.00882, -0.00358, 0.00585, 0.00575, 0.00358, -0.00585, 0.00575, 0.00284, 0.01516, 0.00882, 0.00327, -0.00602, 0.00575, -0.01455, 0.00512, 0.00882, 0.01455, -0.00512, 0.00882, -0.00327, 0.00602, 0.00575, 0.00769, 0.0053, -0.01457, 0.00843, 0.00401, -0.01457, -0.00843, -0.00401, -0.01457, -0.00769, -0.0053, -0.01457, -0.00284, -0.01516, 0.00882, -0.00358, 0.00585, 0.00575, 0.00358, -0.00585, 0.00575, 0.00284, 0.01516, 0.00882, -0.01305, -0.00282, 0.01032, 0.01305, 0.00282, 0.01032, 0.01714, 0.0099, -0.02064, -0.01714, -0.0099, -0.02064, -0.00896, -0.0099, 0.01032, 0.00896, 0.0099, 0.01032, -0.01305, -0.00282, 0.01032, 0.01305, 0.00282, 0.01032, 0.01714, 0.0099, -0.02064, -0.01714, -0.0099, -0.02064, -0.00896, -0.0099, 0.01032, 0.00896, 0.0099, 0.01032]])}, database_path = 'cb6_scsd_20220713.pkl', maxdist = 1.75, mondrian_limits = (0, 0))
model_objs_dict['cb8'] = scsd_model('cb8', array([['-2.1137', '5.1029', '-1.9446', 'C'], ['-2.1137', '5.1029', '1.9446', 'C'], ['-1.803', '6.3518', '0.0', 'C'], ['-3.2164', '5.7663', '0.0', 'C'], ['-4.4936', '4.4936', '-1.8082', 'C'], ['-4.4936', '4.4936', '1.8082', 'C'], ['-5.1029', '2.1137', '-1.9446', 'C'], ['-5.1029', '2.1137', '1.9446', 'C'], ['-5.7663', '3.2164', '0.0', 'C'], ['-6.3518', '1.803', '0.0', 'C'], ['-6.3549', '0.0', '-1.8082', 'C'], ['-6.3549', '0.0', '1.8082', 'C'], ['-5.1029', '-2.1137', '-1.9446', 'C'], ['-5.1029', '-2.1137', '1.9446', 'C'], ['-6.3518', '-1.803', '0.0', 'C'], ['-5.7663', '-3.2164', '-0.0', 'C'], ['-4.4936', '-4.4936', '-1.8082', 'C'], ['-4.4936', '-4.4936', '1.8082', 'C'], ['-2.1137', '-5.1029', '-1.9446', 'C'], ['-2.1137', '-5.1029', '1.9446', 'C'], ['-3.2164', '-5.7663', '-0.0', 'C'], ['-1.803', '-6.3518', '-0.0', 'C'], ['-0.0', '-6.3549', '-1.8082', 'C'], ['-0.0', '-6.3549', '1.8082', 'C'], ['2.1137', '-5.1029', '1.9446', 'C'], ['2.1137', '-5.1029', '-1.9446', 'C'], ['1.803', '-6.3518', '-0.0', 'C'], ['3.2164', '-5.7663', '-0.0', 'C'], ['4.4936', '-4.4936', '1.8082', 'C'], ['4.4936', '-4.4936', '-1.8082', 'C'], ['5.1029', '-2.1137', '1.9446', 'C'], ['5.1029', '-2.1137', '-1.9446', 'C'], ['5.7663', '-3.2164', '-0.0', 'C'], ['6.3518', '-1.803', '-0.0', 'C'], ['6.3549', '-0.0', '1.8082', 'C'], ['6.3549', '-0.0', '-1.8082', 'C'], ['5.1029', '2.1137', '1.9446', 'C'], ['5.1029', '2.1137', '-1.9446', 'C'], ['6.3518', '1.803', '-0.0', 'C'], ['5.7663', '3.2164', '0.0', 'C'], ['4.4936', '4.4936', '1.8082', 'C'], ['4.4936', '4.4936', '-1.8082', 'C'], ['2.1137', '5.1029', '1.9446', 'C'], ['2.1137', '5.1029', '-1.9446', 'C'], ['3.2164', '5.7663', '0.0', 'C'], ['1.803', '6.3518', '0.0', 'C'], ['0.0', '6.3549', '1.8082', 'C'], ['0.0', '6.3549', '-1.8082', 'C'], ['1.2116', '-5.8184', '1.2152', 'N'], ['-1.2116', '5.8184', '1.2152', 'N'], ['-3.2576', '4.9709', '-1.2152', 'N'], ['-3.2576', '4.9709', '1.2152', 'N'], ['-4.9709', '3.2576', '-1.2152', 'N'], ['-4.9709', '3.2576', '1.2152', 'N'], ['-5.8184', '1.2116', '-1.2152', 'N'], ['-5.8184', '1.2116', '1.2152', 'N'], ['-5.8184', '-1.2116', '-1.2152', 'N'], ['-5.8184', '-1.2116', '1.2152', 'N'], ['-4.9709', '-3.2576', '-1.2152', 'N'], ['-4.9709', '-3.2576', '1.2152', 'N'], ['-3.2576', '-4.9709', '-1.2152', 'N'], ['-3.2576', '-4.9709', '1.2152', 'N'], ['-1.2116', '-5.8184', '-1.2152', 'N'], ['-1.2116', '-5.8184', '1.2152', 'N'], ['1.2116', '-5.8184', '-1.2152', 'N'], ['3.2576', '-4.9709', '1.2152', 'N'], ['3.2576', '-4.9709', '-1.2152', 'N'], ['4.9709', '-3.2576', '1.2152', 'N'], ['4.9709', '-3.2576', '-1.2152', 'N'], ['5.8184', '-1.2116', '1.2152', 'N'], ['5.8184', '-1.2116', '-1.2152', 'N'], ['5.8184', '1.2116', '1.2152', 'N'], ['5.8184', '1.2116', '-1.2152', 'N'], ['4.9709', '3.2576', '1.2152', 'N'], ['4.9709', '3.2576', '-1.2152', 'N'], ['3.2576', '4.9709', '1.2152', 'N'], ['3.2576', '4.9709', '-1.2152', 'N'], ['1.2116', '5.8184', '1.2152', 'N'], ['1.2116', '5.8184', '-1.2152', 'N'], ['-1.2116', '5.8184', '-1.2152', 'N'], ['-1.9162', '4.6262', '-3.068', 'O'], ['-1.9162', '4.6262', '3.068', 'O'], ['-4.6262', '1.9162', '-3.068', 'O'], ['-4.6262', '1.9162', '3.068', 'O'], ['-4.6262', '-1.9162', '-3.068', 'O'], ['-4.6262', '-1.9162', '3.068', 'O'], ['-1.9162', '-4.6262', '-3.068', 'O'], ['-1.9162', '-4.6262', '3.068', 'O'], ['1.9162', '-4.6262', '3.068', 'O'], ['1.9162', '-4.6262', '-3.068', 'O'], ['4.6262', '-1.9162', '3.068', 'O'], ['4.6262', '-1.9162', '-3.068', 'O'], ['4.6262', '1.9162', '3.068', 'O'], ['4.6262', '1.9162', '-3.068', 'O'], ['1.9162', '4.6262', '3.068', 'O'], ['1.9162', '4.6262', '-3.068', 'O']]), 'D8h' , pca = {"A1g":array([[0.00451, -0.01089, 0.00345, 0.00451, -0.01089, -0.00345, 0.00239, -0.00878, -0.0, 0.00452, -0.00789, -0.0, 0.00731, -0.00731, 0.00128, 0.00731, -0.00731, -0.00128, 0.01089, -0.00451, 0.00345, 0.01089, -0.00451, -0.00345, 0.00789, -0.00452, -0.0, 0.00878, -0.00239, -0.0, 0.01033, -0.0, 0.00128, 0.01033, -0.0, -0.00128, 0.01089, 0.00451, 0.00345, 0.01089, 0.00451, -0.00345, 0.00878, 0.00239, -0.0, 0.00789, 0.00452, -0.0, 0.00731, 0.00731, 0.00128, 0.00731, 0.00731, -0.00128, 0.00451, 0.01089, 0.00345, 0.00451, 0.01089, -0.00345, 0.00452, 0.00789, -0.0, 0.00239, 0.00878, -0.0, 0.0, 0.01033, 0.00128, 0.0, 0.01033, -0.00128, -0.00451, 0.01089, -0.00345, -0.00451, 0.01089, 0.00345, -0.00239, 0.00878, 0.0, -0.00452, 0.00789, 0.0, -0.00731, 0.00731, -0.00128, -0.00731, 0.00731, 0.00128, -0.01089, 0.00451, -0.00345, -0.01089, 0.00451, 0.00345, -0.00789, 0.00452, 0.0, -0.00878, 0.00239, 0.0, -0.01033, 0.0, -0.00128, -0.01033, 0.0, 0.00128, -0.01089, -0.00451, -0.00345, -0.01089, -0.00451, 0.00345, -0.00878, -0.00239, 0.0, -0.00789, -0.00452, 0.0, -0.00731, -0.00731, -0.00128, -0.00731, -0.00731, 0.00128, -0.00451, -0.01089, -0.00345, -0.00451, -0.01089, 0.00345, -0.00452, -0.00789, 0.0, -0.00239, -0.00878, 0.0, -0.0, -0.01033, -0.00128, -0.0, -0.01033, 0.00128, -0.00162, 0.00802, -0.00097, 0.00162, -0.00802, -0.00097, 0.00453, -0.00682, 0.00097, 0.00453, -0.00682, -0.00097, 0.00682, -0.00453, 0.00097, 0.00682, -0.00453, -0.00097, 0.00802, -0.00162, 0.00097, 0.00802, -0.00162, -0.00097, 0.00802, 0.00162, 0.00097, 0.00802, 0.00162, -0.00097, 0.00682, 0.00453, 0.00097, 0.00682, 0.00453, -0.00097, 0.00453, 0.00682, 0.00097, 0.00453, 0.00682, -0.00097, 0.00162, 0.00802, 0.00097, 0.00162, 0.00802, -0.00097, -0.00162, 0.00802, 0.00097, -0.00453, 0.00682, -0.00097, -0.00453, 0.00682, 0.00097, -0.00682, 0.00453, -0.00097, -0.00682, 0.00453, 0.00097, -0.00802, 0.00162, -0.00097, -0.00802, 0.00162, 0.00097, -0.00802, -0.00162, -0.00097, -0.00802, -0.00162, 0.00097, -0.00682, -0.00453, -0.00097, -0.00682, -0.00453, 0.00097, -0.00453, -0.00682, -0.00097, -0.00453, -0.00682, 0.00097, -0.00162, -0.00802, -0.00097, -0.00162, -0.00802, 0.00097, 0.00162, -0.00802, 0.00097, 0.00624, -0.01506, 0.00683, 0.00624, -0.01506, -0.00683, 0.01506, -0.00624, 0.00683, 0.01506, -0.00624, -0.00683, 0.01506, 0.00624, 0.00683, 0.01506, 0.00624, -0.00683, 0.00624, 0.01506, 0.00683, 0.00624, 0.01506, -0.00683, -0.00624, 0.01506, -0.00683, -0.00624, 0.01506, 0.00683, -0.01506, 0.00624, -0.00683, -0.01506, 0.00624, 0.00683, -0.01506, -0.00624, -0.00683, -0.01506, -0.00624, 0.00683, -0.00624, -0.01506, -0.00683, -0.00624, -0.01506, 0.00683], [0.00015, -0.00035, 0.00556, 0.00015, -0.00035, -0.00556, -0.00293, 0.01144, 0.0, -0.00602, 0.01016, 0.0, -0.00622, 0.00622, -0.00127, -0.00622, 0.00622, 0.00127, 0.00035, -0.00015, 0.00556, 0.00035, -0.00015, -0.00556, -0.01016, 0.00602, 0.0, -0.01144, 0.00293, 0.0, -0.0088, -0.0, -0.00127, -0.0088, -0.0, 0.00127, 0.00035, 0.00015, 0.00556, 0.00035, 0.00015, -0.00556, -0.01144, -0.00293, 0.0, -0.01016, -0.00602, 0.0, -0.00622, -0.00622, -0.00127, -0.00622, -0.00622, 0.00127, 0.00015, 0.00035, 0.00556, 0.00015, 0.00035, -0.00556, -0.00602, -0.01016, 0.0, -0.00293, -0.01144, 0.0, -0.0, -0.0088, -0.00127, 0.0, -0.0088, 0.00127, -0.00015, 0.00035, -0.00556, -0.00015, 0.00035, 0.00556, 0.00293, -0.01144, -0.0, 0.00602, -0.01016, -0.0, 0.00622, -0.00622, 0.00127, 0.00622, -0.00622, -0.00127, -0.00035, 0.00015, -0.00556, -0.00035, 0.00015, 0.00556, 0.01016, -0.00602, -0.0, 0.01144, -0.00293, -0.0, 0.0088, 0.0, 0.00127, 0.0088, 0.0, -0.00127, -0.00035, -0.00015, -0.00556, -0.00035, -0.00015, 0.00556, 0.01144, 0.00293, -0.0, 0.01016, 0.00602, -0.0, 0.00622, 0.00622, 0.00127, 0.00622, 0.00622, -0.00127, -0.00015, -0.00035, -0.00556, -0.00015, -0.00035, 0.00556, 0.00602, 0.01016, -0.0, 0.00293, 0.01144, -0.0, 0.0, 0.0088, 0.00127, -0.0, 0.0088, -0.00127, 0.00205, -0.00878, -4e-05, -0.00205, 0.00878, -4e-05, -0.00476, 0.00766, 4e-05, -0.00476, 0.00766, -4e-05, -0.00766, 0.00476, 4e-05, -0.00766, 0.00476, -4e-05, -0.00878, 0.00205, 4e-05, -0.00878, 0.00205, -4e-05, -0.00878, -0.00205, 4e-05, -0.00878, -0.00205, -4e-05, -0.00766, -0.00476, 4e-05, -0.00766, -0.00476, -4e-05, -0.00476, -0.00766, 4e-05, -0.00476, -0.00766, -4e-05, -0.00205, -0.00878, 4e-05, -0.00205, -0.00878, -4e-05, 0.00205, -0.00878, 4e-05, 0.00476, -0.00766, -4e-05, 0.00476, -0.00766, 4e-05, 0.00766, -0.00476, -4e-05, 0.00766, -0.00476, 4e-05, 0.00878, -0.00205, -4e-05, 0.00878, -0.00205, 4e-05, 0.00878, 0.00205, -4e-05, 0.00878, 0.00205, 4e-05, 0.00766, 0.00476, -4e-05, 0.00766, 0.00476, 4e-05, 0.00476, 0.00766, -4e-05, 0.00476, 0.00766, 4e-05, 0.00205, 0.00878, -4e-05, 0.00205, 0.00878, 4e-05, -0.00205, 0.00878, 4e-05, 0.00595, -0.01437, 0.01028, 0.00595, -0.01437, -0.01028, 0.01437, -0.00595, 0.01028, 0.01437, -0.00595, -0.01028, 0.01437, 0.00595, 0.01028, 0.01437, 0.00595, -0.01028, 0.00595, 0.01437, 0.01028, 0.00595, 0.01437, -0.01028, -0.00595, 0.01437, -0.01028, -0.00595, 0.01437, 0.01028, -0.01437, 0.00595, -0.01028, -0.01437, 0.00595, 0.01028, -0.01437, -0.00595, -0.01028, -0.01437, -0.00595, 0.01028, -0.00595, -0.01437, -0.01028, -0.00595, -0.01437, 0.01028]]), "A2g":array([[0.0005, 0.00021, -0.0, 0.0005, 0.00021, -0.0, 0.00678, -0.0085, 0.0, -0.00122, 0.01081, 0.0, 0.0009, 0.0009, 0.0, 0.0009, 0.0009, 0.0, 0.00021, 0.0005, -0.0, 0.00021, 0.0005, 0.0, 0.01081, -0.00122, 0.0, -0.0085, 0.00678, 0.0, -0.0, 0.00127, 0.0, 0.0, 0.00127, 0.0, -0.00021, 0.0005, 0.0, -0.00021, 0.0005, 0.0, 0.0085, 0.00678, 0.0, -0.01081, -0.00122, 0.0, -0.0009, 0.0009, 0.0, -0.0009, 0.0009, 0.0, -0.0005, 0.00021, 0.0, -0.0005, 0.00021, 0.0, 0.00122, 0.01081, 0.0, -0.00678, -0.0085, -0.0, -0.00127, 0.0, -0.0, -0.00127, 0.0, -0.0, -0.0005, -0.00021, 0.0, -0.0005, -0.00021, 0.0, -0.00678, 0.0085, -0.0, 0.00122, -0.01081, -0.0, -0.0009, -0.0009, -0.0, -0.0009, -0.0009, -0.0, -0.00021, -0.0005, 0.0, -0.00021, -0.0005, -0.0, -0.01081, 0.00122, -0.0, 0.0085, -0.00678, -0.0, 0.0, -0.00127, -0.0, -0.0, -0.00127, -0.0, 0.00021, -0.0005, -0.0, 0.00021, -0.0005, -0.0, -0.0085, -0.00678, -0.0, 0.01081, 0.00122, -0.0, 0.0009, -0.0009, -0.0, 0.0009, -0.0009, -0.0, 0.0005, -0.00021, -0.0, 0.0005, -0.00021, -0.0, -0.00122, -0.01081, -0.0, 0.00678, 0.0085, 0.0, 0.00127, -0.0, 0.0, 0.00127, -0.0, 0.0, -0.00422, 0.01875, -0.0025, 0.00422, -0.01875, -0.0025, -0.01027, 0.01625, -0.0025, -0.01027, 0.01625, 0.0025, 0.01625, -0.01027, 0.0025, 0.01625, -0.01027, -0.0025, -0.01875, 0.00422, -0.0025, -0.01875, 0.00422, 0.0025, 0.01875, 0.00422, 0.0025, 0.01875, 0.00422, -0.0025, -0.01625, -0.01027, -0.0025, -0.01625, -0.01027, 0.0025, 0.01027, 0.01625, 0.0025, 0.01027, 0.01625, -0.0025, -0.00422, -0.01875, -0.0025, -0.00422, -0.01875, 0.0025, -0.00422, 0.01875, 0.0025, 0.01027, -0.01625, 0.0025, 0.01027, -0.01625, -0.0025, -0.01625, 0.01027, -0.0025, -0.01625, 0.01027, 0.0025, 0.01875, -0.00422, 0.0025, 0.01875, -0.00422, -0.0025, -0.01875, -0.00422, -0.0025, -0.01875, -0.00422, 0.0025, 0.01625, 0.01027, 0.0025, 0.01625, 0.01027, -0.0025, -0.01027, -0.01625, -0.0025, -0.01027, -0.01625, 0.0025, 0.00422, 0.01875, 0.0025, 0.00422, 0.01875, -0.0025, 0.00422, -0.01875, 0.0025, -0.00785, -0.00325, -0.0, -0.00785, -0.00325, 0.0, -0.00325, -0.00785, -0.0, -0.00325, -0.00785, 0.0, 0.00325, -0.00785, -0.0, 0.00325, -0.00785, 0.0, 0.00785, -0.00325, -0.0, 0.00785, -0.00325, 0.0, 0.00785, 0.00325, 0.0, 0.00785, 0.00325, -0.0, 0.00325, 0.00785, 0.0, 0.00325, 0.00785, -0.0, -0.00325, 0.00785, 0.0, -0.00325, 0.00785, -0.0, -0.00785, 0.00325, 0.0, -0.00785, 0.00325, -0.0], [-0.00704, -0.00291, 0.0, -0.00704, -0.00291, 0.0, 0.00675, -0.00012, 0.0, 0.00469, 0.00486, 0.0, 0.01118, 0.01118, -0.0, 0.01118, 0.01118, -0.0, -0.00291, -0.00704, 0.0, -0.00291, -0.00704, -0.0, 0.00486, 0.00469, 0.0, -0.00012, 0.00675, 0.0, -0.0, 0.01581, -0.0, -0.0, 0.01581, -0.0, 0.00291, -0.00704, -0.0, 0.00291, -0.00704, -0.0, 0.00012, 0.00675, 0.0, -0.00486, 0.00469, 0.0, -0.01118, 0.01118, -0.0, -0.01118, 0.01118, -0.0, 0.00704, -0.00291, -0.0, 0.00704, -0.00291, -0.0, -0.00469, 0.00486, 0.0, -0.00675, -0.00012, 0.0, -0.01581, -0.0, 0.0, -0.01581, -0.0, 0.0, 0.00704, 0.00291, -0.0, 0.00704, 0.00291, -0.0, -0.00675, 0.00012, -0.0, -0.00469, -0.00486, -0.0, -0.01118, -0.01118, 0.0, -0.01118, -0.01118, 0.0, 0.00291, 0.00704, -0.0, 0.00291, 0.00704, 0.0, -0.00486, -0.00469, -0.0, 0.00012, -0.00675, -0.0, 0.0, -0.01581, 0.0, 0.0, -0.01581, 0.0, -0.00291, 0.00704, 0.0, -0.00291, 0.00704, 0.0, -0.00012, -0.00675, -0.0, 0.00486, -0.00469, -0.0, 0.01118, -0.01118, 0.0, 0.01118, -0.01118, 0.0, -0.00704, 0.00291, 0.0, -0.00704, 0.00291, 0.0, 0.00469, -0.00486, -0.0, 0.00675, 0.00012, -0.0, 0.01581, 0.0, -0.0, 0.01581, 0.0, -0.0, -0.0001, -0.00732, -0.00145, 0.0001, 0.00732, -0.00145, 0.00525, -0.00511, -0.00145, 0.00525, -0.00511, 0.00145, -0.00511, 0.00525, 0.00145, -0.00511, 0.00525, -0.00145, 0.00732, 0.0001, -0.00145, 0.00732, 0.0001, 0.00145, -0.00732, 0.0001, 0.00145, -0.00732, 0.0001, -0.00145, 0.00511, 0.00525, -0.00145, 0.00511, 0.00525, 0.00145, -0.00525, -0.00511, 0.00145, -0.00525, -0.00511, -0.00145, -0.0001, 0.00732, -0.00145, -0.0001, 0.00732, 0.00145, -0.0001, -0.00732, 0.00145, -0.00525, 0.00511, 0.00145, -0.00525, 0.00511, -0.00145, 0.00511, -0.00525, -0.00145, 0.00511, -0.00525, 0.00145, -0.00732, -0.0001, 0.00145, -0.00732, -0.0001, -0.00145, 0.00732, -0.0001, -0.00145, 0.00732, -0.0001, 0.00145, -0.00511, -0.00525, 0.00145, -0.00511, -0.00525, -0.00145, 0.00525, 0.00511, -0.00145, 0.00525, 0.00511, 0.00145, 0.0001, -0.00732, 0.00145, 0.0001, -0.00732, -0.00145, 0.0001, 0.00732, 0.00145, -0.02214, -0.00917, 0.0, -0.02214, -0.00917, 0.0, -0.00917, -0.02214, 0.0, -0.00917, -0.02214, 0.0, 0.00917, -0.02214, 0.0, 0.00917, -0.02214, 0.0, 0.02214, -0.00917, 0.0, 0.02214, -0.00917, 0.0, 0.02214, 0.00917, -0.0, 0.02214, 0.00917, -0.0, 0.00917, 0.02214, -0.0, 0.00917, 0.02214, -0.0, -0.00917, 0.02214, -0.0, -0.00917, 0.02214, -0.0, -0.02214, 0.00917, -0.0, -0.02214, 0.00917, -0.0]]), "B1g":array([[0.00807, 0.00334, 0.0, 0.00807, 0.00334, -0.0, -0.00447, 0.0066, 0.0, 0.00151, -0.00783, -0.0, 0.01168, -0.01168, -0.00389, 0.01168, -0.01168, 0.00389, -0.00334, -0.00807, 0.0, -0.00334, -0.00807, -0.0, 0.00783, -0.00151, -0.0, -0.0066, 0.00447, 0.0, -0.01652, 0.0, 0.00389, -0.01652, 0.0, -0.00389, -0.00334, 0.00807, 0.0, -0.00334, 0.00807, -0.0, -0.0066, -0.00447, 0.0, 0.00783, 0.00151, -0.0, 0.01168, 0.01168, -0.00389, 0.01168, 0.01168, 0.00389, 0.00807, -0.00334, 0.0, 0.00807, -0.00334, -0.0, 0.00151, 0.00783, 0.0, -0.00447, -0.0066, 0.0, -0.0, -0.01652, 0.00389, -0.0, -0.01652, -0.00389, -0.00807, -0.00334, -0.0, -0.00807, -0.00334, 0.0, 0.00447, -0.0066, -0.0, -0.00151, 0.00783, 0.0, -0.01168, 0.01168, 0.00389, -0.01168, 0.01168, -0.00389, 0.00334, 0.00807, -0.0, 0.00334, 0.00807, 0.0, -0.00783, 0.00151, 0.0, 0.0066, -0.00447, -0.0, 0.01652, -0.0, -0.00389, 0.01652, -0.0, 0.00389, 0.00334, -0.00807, -0.0, 0.00334, -0.00807, 0.0, 0.0066, 0.00447, -0.0, -0.00783, -0.00151, 0.0, -0.01168, -0.01168, 0.00389, -0.01168, -0.01168, -0.00389, -0.00807, 0.00334, -0.0, -0.00807, 0.00334, 0.0, -0.00151, -0.00783, -0.0, 0.00447, 0.0066, -0.0, 0.0, 0.01652, -0.00389, 0.0, 0.01652, 0.00389, -0.00012, -0.01262, 0.00017, 0.00012, 0.01262, 0.00017, 0.00901, -0.00884, 0.00017, 0.00901, -0.00884, -0.00017, 0.00884, -0.00901, 0.00017, 0.00884, -0.00901, -0.00017, -0.01262, -0.00012, -0.00017, -0.01262, -0.00012, 0.00017, -0.01262, 0.00012, -0.00017, -0.01262, 0.00012, 0.00017, 0.00884, 0.00901, 0.00017, 0.00884, 0.00901, -0.00017, 0.00901, 0.00884, 0.00017, 0.00901, 0.00884, -0.00017, 0.00012, -0.01262, -0.00017, 0.00012, -0.01262, 0.00017, -0.00012, -0.01262, -0.00017, -0.00901, 0.00884, -0.00017, -0.00901, 0.00884, 0.00017, -0.00884, 0.00901, -0.00017, -0.00884, 0.00901, 0.00017, 0.01262, 0.00012, 0.00017, 0.01262, 0.00012, -0.00017, 0.01262, -0.00012, 0.00017, 0.01262, -0.00012, -0.00017, -0.00884, -0.00901, -0.00017, -0.00884, -0.00901, 0.00017, -0.00901, -0.00884, -0.00017, -0.00901, -0.00884, 0.00017, -0.00012, 0.01262, 0.00017, -0.00012, 0.01262, -0.00017, 0.00012, 0.01262, -0.00017, 0.01263, 0.00523, 0.0, 0.01263, 0.00523, 0.0, -0.00523, -0.01263, 0.0, -0.00523, -0.01263, 0.0, -0.00523, 0.01263, 0.0, -0.00523, 0.01263, 0.0, 0.01263, -0.00523, 0.0, 0.01263, -0.00523, -0.0, -0.01263, -0.00523, -0.0, -0.01263, -0.00523, -0.0, 0.00523, 0.01263, -0.0, 0.00523, 0.01263, -0.0, 0.00523, -0.01263, -0.0, 0.00523, -0.01263, -0.0, -0.01263, 0.00523, -0.0, -0.01263, 0.00523, 0.0], [-0.0115, -0.00477, -0.0, -0.0115, -0.00477, -0.0, -0.01663, 0.00156, 0.0, -0.01066, -0.01287, 0.0, 0.00427, -0.00427, 0.00096, 0.00427, -0.00427, -0.00096, 0.00477, 0.0115, 0.0, 0.00477, 0.0115, -0.0, 0.01287, 0.01066, -0.0, -0.00156, 0.01663, 0.0, -0.00604, 0.0, -0.00096, -0.00604, -0.0, 0.00096, 0.00477, -0.0115, 0.0, 0.00477, -0.0115, -0.0, -0.00156, -0.01663, -0.0, 0.01287, -0.01066, -0.0, 0.00427, 0.00427, 0.00096, 0.00427, 0.00427, -0.00096, -0.0115, 0.00477, 0.0, -0.0115, 0.00477, -0.0, -0.01066, 0.01287, 0.0, -0.01663, -0.00156, 0.0, -0.0, -0.00604, -0.00096, -0.0, -0.00604, 0.00096, 0.0115, 0.00477, -0.0, 0.0115, 0.00477, 0.0, 0.01663, -0.00156, -0.0, 0.01066, 0.01287, -0.0, -0.00427, 0.00427, -0.00096, -0.00427, 0.00427, 0.00096, -0.00477, -0.0115, -0.0, -0.00477, -0.0115, 0.0, -0.01287, -0.01066, 0.0, 0.00156, -0.01663, -0.0, 0.00604, -0.0, 0.00096, 0.00604, 0.0, -0.00096, -0.00477, 0.0115, -0.0, -0.00477, 0.0115, 0.0, 0.00156, 0.01663, 0.0, -0.01287, 0.01066, 0.0, -0.00427, -0.00427, -0.00096, -0.00427, -0.00427, 0.00096, 0.0115, -0.00477, -0.0, 0.0115, -0.00477, 0.0, 0.01066, -0.01287, -0.0, 0.01663, 0.00156, -0.0, 0.0, 0.00604, 0.00096, 0.0, 0.00604, -0.00096, 0.01194, -0.00175, 0.00343, -0.01194, 0.00175, 0.00343, -0.00721, -0.00967, 0.00343, -0.00721, -0.00967, -0.00343, 0.00967, 0.00721, 0.00343, 0.00967, 0.00721, -0.00343, -0.00175, 0.01194, -0.00343, -0.00175, 0.01194, 0.00343, -0.00175, -0.01194, -0.00343, -0.00175, -0.01194, 0.00343, 0.00967, -0.00721, 0.00343, 0.00967, -0.00721, -0.00343, -0.00721, 0.00967, 0.00343, -0.00721, 0.00967, -0.00343, -0.01194, -0.00175, -0.00343, -0.01194, -0.00175, 0.00343, 0.01194, -0.00175, -0.00343, 0.00721, 0.00967, -0.00343, 0.00721, 0.00967, 0.00343, -0.00967, -0.00721, -0.00343, -0.00967, -0.00721, 0.00343, 0.00175, -0.01194, 0.00343, 0.00175, -0.01194, -0.00343, 0.00175, 0.01194, 0.00343, 0.00175, 0.01194, -0.00343, -0.00967, 0.00721, -0.00343, -0.00967, 0.00721, 0.00343, 0.00721, -0.00967, -0.00343, 0.00721, -0.00967, 0.00343, 0.01194, 0.00175, 0.00343, 0.01194, 0.00175, -0.00343, -0.01194, 0.00175, -0.00343, -0.00773, -0.0032, 0.0, -0.00773, -0.0032, -0.0, 0.0032, 0.00773, 0.0, 0.0032, 0.00773, -0.0, 0.0032, -0.00773, 0.0, 0.0032, -0.00773, -0.0, -0.00773, 0.0032, 0.0, -0.00773, 0.0032, -0.0, 0.00773, 0.0032, -0.0, 0.00773, 0.0032, 0.0, -0.0032, -0.00773, -0.0, -0.0032, -0.00773, 0.0, -0.0032, 0.00773, -0.0, -0.0032, 0.00773, 0.0, 0.00773, -0.0032, -0.0, 0.00773, -0.0032, 0.0]]), "B2g":array([[0.00412, -0.00996, 0.00544, 0.00412, -0.00996, -0.00544, -0.00025, 0.00128, 0.0, -0.00073, 0.00109, 0.0, -0.00076, -0.00076, -0.0, -0.00076, -0.00076, 0.0, -0.00996, 0.00412, -0.00544, -0.00996, 0.00412, 0.00544, 0.00109, -0.00073, -0.0, 0.00128, -0.00025, 0.0, 0.0, 0.00108, -0.0, 0.0, 0.00108, -0.0, 0.00996, 0.00412, 0.00544, 0.00996, 0.00412, -0.00544, -0.00128, -0.00025, 0.0, -0.00109, -0.00073, 0.0, 0.00076, -0.00076, -0.0, 0.00076, -0.00076, -0.0, -0.00412, -0.00996, -0.00544, -0.00412, -0.00996, 0.00544, 0.00073, 0.00109, 0.0, 0.00025, 0.00128, 0.0, -0.00108, 0.0, 0.0, -0.00108, 0.0, 0.0, -0.00412, 0.00996, -0.00544, -0.00412, 0.00996, 0.00544, 0.00025, -0.00128, -0.0, 0.00073, -0.00109, -0.0, 0.00076, 0.00076, 0.0, 0.00076, 0.00076, -0.0, 0.00996, -0.00412, 0.00544, 0.00996, -0.00412, -0.00544, -0.00109, 0.00073, 0.0, -0.00128, 0.00025, -0.0, -0.0, -0.00108, 0.0, -0.0, -0.00108, 0.0, -0.00996, -0.00412, -0.00544, -0.00996, -0.00412, 0.00544, 0.00128, 0.00025, -0.0, 0.00109, 0.00073, -0.0, -0.00076, 0.00076, 0.0, -0.00076, 0.00076, 0.0, 0.00412, 0.00996, 0.00544, 0.00412, 0.00996, -0.00544, -0.00073, -0.00109, -0.0, -0.00025, -0.00128, -0.0, 0.00108, -0.0, -0.0, 0.00108, -0.0, -0.0, -0.00145, 0.00015, -0.00092, 0.00145, -0.00015, -0.00092, -0.00092, -0.00113, 0.00092, -0.00092, -0.00113, -0.00092, -0.00113, -0.00092, -0.00092, -0.00113, -0.00092, 0.00092, -0.00015, 0.00145, -0.00092, -0.00015, 0.00145, 0.00092, 0.00015, 0.00145, 0.00092, 0.00015, 0.00145, -0.00092, 0.00113, -0.00092, 0.00092, 0.00113, -0.00092, -0.00092, 0.00092, -0.00113, -0.00092, 0.00092, -0.00113, 0.00092, -0.00145, -0.00015, -0.00092, -0.00145, -0.00015, 0.00092, -0.00145, 0.00015, 0.00092, 0.00092, 0.00113, -0.00092, 0.00092, 0.00113, 0.00092, 0.00113, 0.00092, 0.00092, 0.00113, 0.00092, -0.00092, 0.00015, -0.00145, 0.00092, 0.00015, -0.00145, -0.00092, -0.00015, -0.00145, -0.00092, -0.00015, -0.00145, 0.00092, -0.00113, 0.00092, -0.00092, -0.00113, 0.00092, 0.00092, -0.00092, 0.00113, 0.00092, -0.00092, 0.00113, -0.00092, 0.00145, 0.00015, 0.00092, 0.00145, 0.00015, -0.00092, 0.00145, -0.00015, 0.00092, 0.01002, -0.02419, 0.01236, 0.01002, -0.02419, -0.01236, -0.02419, 0.01002, -0.01236, -0.02419, 0.01002, 0.01236, 0.02419, 0.01002, 0.01236, 0.02419, 0.01002, -0.01236, -0.01002, -0.02419, -0.01236, -0.01002, -0.02419, 0.01236, -0.01002, 0.02419, -0.01236, -0.01002, 0.02419, 0.01236, 0.02419, -0.01002, 0.01236, 0.02419, -0.01002, -0.01236, -0.02419, -0.01002, -0.01236, -0.02419, -0.01002, 0.01236, 0.01002, 0.02419, 0.01236, 0.01002, 0.02419, -0.01236], [0.0033, -0.00798, -0.0035, 0.0033, -0.00798, 0.0035, 0.00572, -0.01446, -0.0, 0.00617, -0.01427, -0.0, -4e-05, -4e-05, -0.0, -4e-05, -4e-05, 0.0, -0.00798, 0.0033, 0.0035, -0.00798, 0.0033, -0.0035, -0.01427, 0.00617, 0.0, -0.01446, 0.00572, 0.0, 0.0, 5e-05, 0.0, 0.0, 5e-05, 0.0, 0.00798, 0.0033, -0.0035, 0.00798, 0.0033, 0.0035, 0.01446, 0.00572, 0.0, 0.01427, 0.00617, 0.0, 4e-05, -4e-05, 0.0, 4e-05, -4e-05, 0.0, -0.0033, -0.00798, 0.0035, -0.0033, -0.00798, -0.0035, -0.00617, -0.01427, -0.0, -0.00572, -0.01446, -0.0, -5e-05, 0.0, -0.0, -5e-05, 0.0, -0.0, -0.0033, 0.00798, 0.0035, -0.0033, 0.00798, -0.0035, -0.00572, 0.01446, 0.0, -0.00617, 0.01427, 0.0, 4e-05, 4e-05, 0.0, 4e-05, 4e-05, -0.0, 0.00798, -0.0033, -0.0035, 0.00798, -0.0033, 0.0035, 0.01427, -0.00617, -0.0, 0.01446, -0.00572, -0.0, -0.0, -5e-05, -0.0, -0.0, -5e-05, -0.0, -0.00798, -0.0033, 0.0035, -0.00798, -0.0033, -0.0035, -0.01446, -0.00572, -0.0, -0.01427, -0.00617, -0.0, -4e-05, 4e-05, -0.0, -4e-05, 4e-05, -0.0, 0.0033, 0.00798, -0.0035, 0.0033, 0.00798, 0.0035, 0.00617, 0.01427, 0.0, 0.00572, 0.01446, 0.0, 5e-05, -0.0, 0.0, 5e-05, -0.0, 0.0, -0.00646, 0.01358, -0.00048, 0.00646, -0.01358, -0.00048, 0.00504, -0.01417, 0.00048, 0.00504, -0.01417, -0.00048, -0.01417, 0.00504, -0.00048, -0.01417, 0.00504, 0.00048, -0.01358, 0.00646, -0.00048, -0.01358, 0.00646, 0.00048, 0.01358, 0.00646, 0.00048, 0.01358, 0.00646, -0.00048, 0.01417, 0.00504, 0.00048, 0.01417, 0.00504, -0.00048, -0.00504, -0.01417, -0.00048, -0.00504, -0.01417, 0.00048, -0.00646, -0.01358, -0.00048, -0.00646, -0.01358, 0.00048, -0.00646, 0.01358, 0.00048, -0.00504, 0.01417, -0.00048, -0.00504, 0.01417, 0.00048, 0.01417, -0.00504, 0.00048, 0.01417, -0.00504, -0.00048, 0.01358, -0.00646, 0.00048, 0.01358, -0.00646, -0.00048, -0.01358, -0.00646, -0.00048, -0.01358, -0.00646, 0.00048, -0.01417, -0.00504, -0.00048, -0.01417, -0.00504, 0.00048, 0.00504, 0.01417, 0.00048, 0.00504, 0.01417, -0.00048, 0.00646, 0.01358, 0.00048, 0.00646, 0.01358, -0.00048, 0.00646, -0.01358, 0.00048, 0.00017, -0.00042, -0.00724, 0.00017, -0.00042, 0.00724, -0.00042, 0.00017, 0.00724, -0.00042, 0.00017, -0.00724, 0.00042, 0.00017, -0.00724, 0.00042, 0.00017, 0.00724, -0.00017, -0.00042, 0.00724, -0.00017, -0.00042, -0.00724, -0.00017, 0.00042, 0.00724, -0.00017, 0.00042, -0.00724, 0.00042, -0.00017, -0.00724, 0.00042, -0.00017, 0.00724, -0.00042, -0.00017, 0.00724, -0.00042, -0.00017, -0.00724, 0.00017, 0.00042, -0.00724, 0.00017, 0.00042, 0.00724]]), "E2g1":array([[0.0026, 0.01308, 7e-05, 0.0026, 0.01308, -7e-05, -0.00163, 0.01431, 0.0, 0.00025, 0.00953, 0.0, 0.00471, 0.00471, 0.0, 0.00471, 0.00471, 0.0, 0.01308, 0.0026, -7e-05, 0.01308, 0.0026, 7e-05, 0.00953, 0.00025, -0.0, 0.01431, -0.00163, -0.0, 0.01757, -0.0, -0.00129, 0.01757, -0.0, 0.00129, 0.01308, -0.0026, -7e-05, 0.01308, -0.0026, 7e-05, 0.01431, 0.00163, -0.0, 0.00953, -0.00025, -0.0, 0.00471, -0.00471, -0.0, 0.00471, -0.00471, -0.0, 0.0026, -0.01308, 7e-05, 0.0026, -0.01308, -7e-05, 0.00025, -0.00953, -0.0, -0.00163, -0.01431, -0.0, -0.0, -0.01757, 0.00129, -0.0, -0.01757, -0.00129, -0.0026, -0.01308, -7e-05, -0.0026, -0.01308, 7e-05, 0.00163, -0.01431, -0.0, -0.00025, -0.00953, -0.0, -0.00471, -0.00471, -0.0, -0.00471, -0.00471, -0.0, -0.01308, -0.0026, 7e-05, -0.01308, -0.0026, -7e-05, -0.00953, -0.00025, 0.0, -0.01431, 0.00163, 0.0, -0.01757, 0.0, 0.00129, -0.01757, 0.0, -0.00129, -0.01308, 0.0026, 7e-05, -0.01308, 0.0026, -7e-05, -0.01431, -0.00163, 0.0, -0.00953, 0.00025, 0.0, -0.00471, 0.00471, 0.0, -0.00471, 0.00471, 0.0, -0.0026, 0.01308, -7e-05, -0.0026, 0.01308, 7e-05, -0.00025, 0.00953, 0.0, 0.00163, 0.01431, 0.0, 0.0, 0.01757, -0.00129, 0.0, 0.01757, 0.00129, 0.0, -0.01626, 4e-05, -0.0, 0.01626, 4e-05, 0.00284, 0.00944, -4e-05, 0.00284, 0.00944, 4e-05, 0.00944, 0.00284, 4e-05, 0.00944, 0.00284, -4e-05, 0.01626, -0.0, 4e-05, 0.01626, -0.0, -4e-05, 0.01626, 0.0, 4e-05, 0.01626, 0.0, -4e-05, 0.00944, -0.00284, 4e-05, 0.00944, -0.00284, -4e-05, 0.00284, -0.00944, -4e-05, 0.00284, -0.00944, 4e-05, -0.0, -0.01626, -4e-05, -0.0, -0.01626, 4e-05, 0.0, -0.01626, -4e-05, -0.00284, -0.00944, 4e-05, -0.00284, -0.00944, -4e-05, -0.00944, -0.00284, -4e-05, -0.00944, -0.00284, 4e-05, -0.01626, 0.0, -4e-05, -0.01626, 0.0, 4e-05, -0.01626, -0.0, -4e-05, -0.01626, -0.0, 4e-05, -0.00944, 0.00284, -4e-05, -0.00944, 0.00284, 4e-05, -0.00284, 0.00944, 4e-05, -0.00284, 0.00944, -4e-05, 0.0, 0.01626, 4e-05, 0.0, 0.01626, -4e-05, -0.0, 0.01626, -4e-05, 0.00422, 0.01355, 0.00013, 0.00422, 0.01355, -0.00013, 0.01355, 0.00422, -0.00013, 0.01355, 0.00422, 0.00013, 0.01355, -0.00422, -0.00013, 0.01355, -0.00422, 0.00013, 0.00422, -0.01355, 0.00013, 0.00422, -0.01355, -0.00013, -0.00422, -0.01355, -0.00013, -0.00422, -0.01355, 0.00013, -0.01355, -0.00422, 0.00013, -0.01355, -0.00422, -0.00013, -0.01355, 0.00422, 0.00013, -0.01355, 0.00422, -0.00013, -0.00422, 0.01355, -0.00013, -0.00422, 0.01355, 0.00013], [-0.00368, 0.00406, -0.00543, -0.00368, 0.00406, 0.00543, 0.00112, -0.00745, -0.0, 0.00182, -0.00601, -0.0, -0.00134, -0.00134, 0.0, -0.00134, -0.00134, 0.0, 0.00406, -0.00368, 0.00543, 0.00406, -0.00368, -0.00543, -0.00601, 0.00182, 0.0, -0.00745, 0.00112, 0.0, -0.00363, 0.0, 0.00019, -0.00363, 0.0, -0.00019, 0.00406, 0.00368, 0.00543, 0.00406, 0.00368, -0.00543, -0.00745, -0.00112, 0.0, -0.00601, -0.00182, 0.0, -0.00134, 0.00134, -0.0, -0.00134, 0.00134, -0.0, -0.00368, -0.00406, -0.00543, -0.00368, -0.00406, 0.00543, 0.00182, 0.00601, 0.0, 0.00112, 0.00745, 0.0, -0.0, 0.00363, -0.00019, -0.0, 0.00363, 0.00019, 0.00368, -0.00406, 0.00543, 0.00368, -0.00406, -0.00543, -0.00112, 0.00745, 0.0, -0.00182, 0.00601, 0.0, 0.00134, 0.00134, -0.0, 0.00134, 0.00134, -0.0, -0.00406, 0.00368, -0.00543, -0.00406, 0.00368, 0.00543, 0.00601, -0.00182, -0.0, 0.00745, -0.00112, -0.0, 0.00363, -0.0, -0.00019, 0.00363, -0.0, 0.00019, -0.00406, -0.00368, -0.00543, -0.00406, -0.00368, 0.00543, 0.00745, 0.00112, -0.0, 0.00601, 0.00182, -0.0, 0.00134, -0.00134, 0.0, 0.00134, -0.00134, 0.0, 0.00368, 0.00406, 0.00543, 0.00368, 0.00406, -0.00543, -0.00182, -0.00601, -0.0, -0.00112, -0.00745, -0.0, 0.0, -0.00363, 0.00019, 0.0, -0.00363, -0.00019, -0.00063, 0.00571, 0.00061, 0.00063, -0.00571, 0.00061, 0.00065, -0.00572, -0.00084, 0.00065, -0.00572, 0.00084, -0.00572, 0.00065, 0.00084, -0.00572, 0.00065, -0.00084, -0.00571, 0.00063, 0.00061, -0.00571, 0.00063, -0.00061, -0.00571, -0.00063, 0.00061, -0.00571, -0.00063, -0.00061, -0.00572, -0.00065, 0.00084, -0.00572, -0.00065, -0.00084, 0.00065, 0.00572, -0.00084, 0.00065, 0.00572, 0.00084, 0.00063, 0.00571, -0.00061, 0.00063, 0.00571, 0.00061, -0.00063, 0.00571, -0.00061, -0.00065, 0.00572, 0.00084, -0.00065, 0.00572, -0.00084, 0.00572, -0.00065, -0.00084, 0.00572, -0.00065, 0.00084, 0.00571, -0.00063, -0.00061, 0.00571, -0.00063, 0.00061, 0.00571, 0.00063, -0.00061, 0.00571, 0.00063, 0.00061, 0.00572, 0.00065, -0.00084, 0.00572, 0.00065, 0.00084, -0.00065, -0.00572, 0.00084, -0.00065, -0.00572, -0.00084, -0.00063, -0.00571, 0.00061, -0.00063, -0.00571, -0.00061, 0.00063, -0.00571, -0.00061, -0.01033, 0.01979, -0.01318, -0.01033, 0.01979, 0.01318, 0.01979, -0.01033, 0.01318, 0.01979, -0.01033, -0.01318, 0.01979, 0.01033, 0.01318, 0.01979, 0.01033, -0.01318, -0.01033, -0.01979, -0.01318, -0.01033, -0.01979, 0.01318, 0.01033, -0.01979, 0.01318, 0.01033, -0.01979, -0.01318, -0.01979, 0.01033, -0.01318, -0.01979, 0.01033, 0.01318, -0.01979, -0.01033, -0.01318, -0.01979, -0.01033, 0.01318, 0.01033, 0.01979, 0.01318, 0.01033, 0.01979, -0.01318]]), "E2g2":array([[-0.01062, 0.0071, 6e-05, -0.01062, 0.0071, -6e-05, -0.00673, 0.00631, 0.0, -0.00864, 0.0108, 0.0, -0.01191, 0.01191, 0.00116, -0.01191, 0.01191, -0.00116, -0.0071, 0.01062, 6e-05, -0.0071, 0.01062, -6e-05, -0.0108, 0.00864, 0.0, -0.00631, 0.00673, 0.0, -0.0, 0.00637, 0.0, -0.0, 0.00637, 0.0, 0.0071, 0.01062, -6e-05, 0.0071, 0.01062, 6e-05, 0.00631, 0.00673, -0.0, 0.0108, 0.00864, -0.0, 0.01191, 0.01191, -0.00116, 0.01191, 0.01191, 0.00116, 0.01062, 0.0071, -6e-05, 0.01062, 0.0071, 6e-05, 0.00864, 0.0108, -0.0, 0.00673, 0.00631, -0.0, 0.00637, -0.0, -0.0, 0.00637, -0.0, -0.0, 0.01062, -0.0071, -6e-05, 0.01062, -0.0071, 6e-05, 0.00673, -0.00631, -0.0, 0.00864, -0.0108, -0.0, 0.01191, -0.01191, -0.00116, 0.01191, -0.01191, 0.00116, 0.0071, -0.01062, -6e-05, 0.0071, -0.01062, 6e-05, 0.0108, -0.00864, -0.0, 0.00631, -0.00673, -0.0, 0.0, -0.00637, -0.0, 0.0, -0.00637, -0.0, -0.0071, -0.01062, 6e-05, -0.0071, -0.01062, -6e-05, -0.00631, -0.00673, 0.0, -0.0108, -0.00864, 0.0, -0.01191, -0.01191, 0.00116, -0.01191, -0.01191, -0.00116, -0.01062, -0.0071, 6e-05, -0.01062, -0.0071, -6e-05, -0.00864, -0.0108, 0.0, -0.00673, -0.00631, 0.0, -0.00637, 0.0, 0.0, -0.00637, 0.0, 0.0, 0.00836, -0.00449, 1e-05, -0.00836, 0.00449, 1e-05, -0.01099, 0.01102, -4e-05, -0.01099, 0.01102, 4e-05, -0.01102, 0.01099, -4e-05, -0.01102, 0.01099, 4e-05, -0.00449, 0.00836, -1e-05, -0.00449, 0.00836, 1e-05, 0.00449, 0.00836, 1e-05, 0.00449, 0.00836, -1e-05, 0.01102, 0.01099, 4e-05, 0.01102, 0.01099, -4e-05, 0.01099, 0.01102, 4e-05, 0.01099, 0.01102, -4e-05, 0.00836, 0.00449, 1e-05, 0.00836, 0.00449, -1e-05, 0.00836, -0.00449, -1e-05, 0.01099, -0.01102, 4e-05, 0.01099, -0.01102, -4e-05, 0.01102, -0.01099, 4e-05, 0.01102, -0.01099, -4e-05, 0.00449, -0.00836, 1e-05, 0.00449, -0.00836, -1e-05, -0.00449, -0.00836, -1e-05, -0.00449, -0.00836, 1e-05, -0.01102, -0.01099, -4e-05, -0.01102, -0.01099, 4e-05, -0.01099, -0.01102, -4e-05, -0.01099, -0.01102, 4e-05, -0.00836, -0.00449, -1e-05, -0.00836, -0.00449, 1e-05, -0.00836, 0.00449, -1e-05, -0.01204, 0.00635, 0.00014, -0.01204, 0.00635, -0.00014, -0.00635, 0.01204, 0.00014, -0.00635, 0.01204, -0.00014, 0.00635, 0.01204, -0.00014, 0.00635, 0.01204, 0.00014, 0.01204, 0.00635, -0.00014, 0.01204, 0.00635, 0.00014, 0.01204, -0.00635, -0.00014, 0.01204, -0.00635, 0.00014, 0.00635, -0.01204, -0.00014, 0.00635, -0.01204, 0.00014, -0.00635, -0.01204, 0.00014, -0.00635, -0.01204, -0.00014, -0.01204, -0.00635, 0.00014, -0.01204, -0.00635, -0.00014], [-2e-05, 0.00506, -0.00571, -2e-05, 0.00506, 0.00571, 0.00374, -0.00749, -0.0, 0.00428, -0.00589, -0.0, 0.00215, -0.00215, 0.00024, 0.00215, -0.00215, -0.00024, -0.00506, 2e-05, -0.00571, -0.00506, 2e-05, 0.00571, 0.00589, -0.00428, -0.0, 0.00749, -0.00374, -0.0, 0.0, -0.00165, -0.0, 0.0, -0.00165, -0.0, 0.00506, 2e-05, 0.00571, 0.00506, 2e-05, -0.00571, -0.00749, -0.00374, 0.0, -0.00589, -0.00428, 0.0, -0.00215, -0.00215, -0.00024, -0.00215, -0.00215, 0.00024, 2e-05, 0.00506, 0.00571, 2e-05, 0.00506, -0.00571, -0.00428, -0.00589, 0.0, -0.00374, -0.00749, 0.0, -0.00165, 0.0, 0.0, -0.00165, 0.0, -0.0, 2e-05, -0.00506, 0.00571, 2e-05, -0.00506, -0.00571, -0.00374, 0.00749, 0.0, -0.00428, 0.00589, 0.0, -0.00215, 0.00215, -0.00024, -0.00215, 0.00215, 0.00024, 0.00506, -2e-05, 0.00571, 0.00506, -2e-05, -0.00571, -0.00589, 0.00428, 0.0, -0.00749, 0.00374, 0.0, -0.0, 0.00165, 0.0, -0.0, 0.00165, 0.0, -0.00506, -2e-05, -0.00571, -0.00506, -2e-05, 0.00571, 0.00749, 0.00374, -0.0, 0.00589, 0.00428, -0.0, 0.00215, 0.00215, 0.00024, 0.00215, 0.00215, -0.00024, -2e-05, -0.00506, -0.00571, -2e-05, -0.00506, 0.00571, 0.00428, 0.00589, -0.0, 0.00374, 0.00749, -0.0, 0.00165, -0.0, -0.0, 0.00165, -0.0, 0.0, -0.00344, 0.00497, 0.00057, 0.00344, -0.00497, 0.00057, 0.00403, -0.00375, -0.00097, 0.00403, -0.00375, 0.00097, 0.00375, -0.00403, -0.00097, 0.00375, -0.00403, 0.00097, 0.00497, -0.00344, -0.00057, 0.00497, -0.00344, 0.00057, -0.00497, -0.00344, 0.00057, -0.00497, -0.00344, -0.00057, -0.00375, -0.00403, 0.00097, -0.00375, -0.00403, -0.00097, -0.00403, -0.00375, 0.00097, -0.00403, -0.00375, -0.00097, -0.00344, -0.00497, 0.00057, -0.00344, -0.00497, -0.00057, -0.00344, 0.00497, -0.00057, -0.00403, 0.00375, 0.00097, -0.00403, 0.00375, -0.00097, -0.00375, 0.00403, 0.00097, -0.00375, 0.00403, -0.00097, -0.00497, 0.00344, 0.00057, -0.00497, 0.00344, -0.00057, 0.00497, 0.00344, -0.00057, 0.00497, 0.00344, 0.00057, 0.00375, 0.00403, -0.00097, 0.00375, 0.00403, 0.00097, 0.00403, 0.00375, -0.00097, 0.00403, 0.00375, 0.00097, 0.00344, 0.00497, -0.00057, 0.00344, 0.00497, 0.00057, 0.00344, -0.00497, -0.00057, -0.00733, 0.02076, -0.01367, -0.00733, 0.02076, 0.01367, -0.02076, 0.00733, -0.01367, -0.02076, 0.00733, 0.01367, 0.02076, 0.00733, 0.01367, 0.02076, 0.00733, -0.01367, 0.00733, 0.02076, 0.01367, 0.00733, 0.02076, -0.01367, 0.00733, -0.02076, 0.01367, 0.00733, -0.02076, -0.01367, 0.02076, -0.00733, 0.01367, 0.02076, -0.00733, -0.01367, -0.02076, -0.00733, -0.01367, -0.02076, -0.00733, 0.01367, -0.00733, -0.02076, -0.01367, -0.00733, -0.02076, 0.01367]]), "E2g3":array([[-0.00581, 0.01403, 0.00011, -0.00581, 0.01403, -0.00011, -0.00589, 0.01434, 0.0, -0.00598, 0.01431, 0.0, -0.00533, 0.01174, 0.00085, -0.00533, 0.01174, -0.00085, 0.00387, 0.00935, 0.0, 0.00387, 0.00935, -0.0, -0.00119, 0.00636, 0.0, 0.00533, 0.00366, 0.0, 0.01207, 0.00454, -0.00085, 0.01207, 0.00454, 0.00085, 0.01403, 0.00581, -0.00011, 0.01403, 0.00581, 0.00011, 0.01434, 0.00589, 0.0, 0.01431, 0.00598, 0.0, 0.01174, 0.00533, -0.00085, 0.01174, 0.00533, 0.00085, 0.00935, -0.00387, 0.0, 0.00935, -0.00387, -0.0, 0.00636, 0.00119, 0.0, 0.00366, -0.00533, -0.0, 0.00454, -0.01207, 0.00085, 0.00454, -0.01207, -0.00085, 0.00581, -0.01403, -0.00011, 0.00581, -0.01403, 0.00011, 0.00589, -0.01434, -0.0, 0.00598, -0.01431, -0.0, 0.00533, -0.01174, -0.00085, 0.00533, -0.01174, 0.00085, -0.00387, -0.00935, -0.0, -0.00387, -0.00935, 0.0, 0.00119, -0.00636, -0.0, -0.00533, -0.00366, -0.0, -0.01207, -0.00454, 0.00085, -0.01207, -0.00454, -0.00085, -0.01403, -0.00581, 0.00011, -0.01403, -0.00581, -0.00011, -0.01434, -0.00589, -0.0, -0.01431, -0.00598, -0.0, -0.01174, -0.00533, 0.00085, -0.01174, -0.00533, -0.00085, -0.00935, 0.00387, -0.0, -0.00935, 0.00387, 0.0, -0.00636, -0.00119, -0.0, -0.00366, 0.00533, 0.0, -0.00454, 0.01207, -0.00085, -0.00454, 0.01207, 0.00085, 0.00595, -0.01433, 5e-05, -0.00595, 0.01433, 5e-05, -0.00593, 0.01434, -5e-05, -0.00593, 0.01434, 5e-05, -0.0014, 0.00979, -4e-05, -0.0014, 0.00979, 4e-05, 0.00792, 0.00593, 4e-05, 0.00792, 0.00593, -4e-05, 0.01433, 0.00595, 5e-05, 0.01433, 0.00595, -5e-05, 0.01434, 0.00593, 5e-05, 0.01434, 0.00593, -5e-05, 0.00979, 0.0014, 4e-05, 0.00979, 0.0014, -4e-05, 0.00593, -0.00792, -4e-05, 0.00593, -0.00792, 4e-05, 0.00595, -0.01433, -5e-05, 0.00593, -0.01434, 5e-05, 0.00593, -0.01434, -5e-05, 0.0014, -0.00979, 4e-05, 0.0014, -0.00979, -4e-05, -0.00792, -0.00593, -4e-05, -0.00792, -0.00593, 4e-05, -0.01433, -0.00595, -5e-05, -0.01433, -0.00595, 5e-05, -0.01434, -0.00593, -5e-05, -0.01434, -0.00593, 5e-05, -0.00979, -0.0014, -4e-05, -0.00979, -0.0014, 4e-05, -0.00593, 0.00792, 4e-05, -0.00593, 0.00792, -4e-05, -0.00595, 0.01433, -5e-05, -0.00573, 0.01383, 0.00019, -0.00573, 0.01383, -0.00019, 0.00476, 0.0115, -0.0, 0.00476, 0.0115, 0.0, 0.01383, 0.00573, -0.00019, 0.01383, 0.00573, 0.00019, 0.0115, -0.00476, -0.0, 0.0115, -0.00476, 0.0, 0.00573, -0.01383, -0.00019, 0.00573, -0.01383, 0.00019, -0.00476, -0.0115, 0.0, -0.00476, -0.0115, -0.0, -0.01383, -0.00573, 0.00019, -0.01383, -0.00573, -0.00019, -0.0115, 0.00476, 0.0, -0.0115, 0.00476, -0.0], [-0.00324, 0.00781, -0.00892, -0.00324, 0.00781, 0.00892, 0.00397, -0.0115, -0.0, 0.00532, -0.01094, -0.0, 0.00111, -0.00279, -0.00013, 0.00111, -0.00279, 0.00013, -0.00093, -0.00226, -0.0, -0.00093, -0.00226, -0.0, -0.00089, -0.00177, -0.0, -0.00062, -0.00189, -0.0, -0.00276, -0.00119, 0.00013, -0.00276, -0.00119, -0.00013, 0.00781, 0.00324, 0.00892, 0.00781, 0.00324, -0.00892, -0.0115, -0.00397, -0.0, -0.01094, -0.00532, -0.0, -0.00279, -0.00111, 0.00013, -0.00279, -0.00111, -0.00013, -0.00226, 0.00093, -0.0, -0.00226, 0.00093, 0.0, -0.00177, 0.00089, 0.0, -0.00189, 0.00062, -0.0, -0.00119, 0.00276, -0.00013, -0.00119, 0.00276, 0.00013, 0.00324, -0.00781, 0.00892, 0.00324, -0.00781, -0.00892, -0.00397, 0.0115, 0.0, -0.00532, 0.01094, 0.0, -0.00111, 0.00279, 0.00013, -0.00111, 0.00279, -0.00013, 0.00093, 0.00226, 0.0, 0.00093, 0.00226, 0.0, 0.00089, 0.00177, 0.0, 0.00062, 0.00189, 0.0, 0.00276, 0.00119, -0.00013, 0.00276, 0.00119, 0.00013, -0.00781, -0.00324, -0.00892, -0.00781, -0.00324, 0.00892, 0.0115, 0.00397, 0.0, 0.01094, 0.00532, 0.0, 0.00279, 0.00111, -0.00013, 0.00279, 0.00111, 0.00013, 0.00226, -0.00093, 0.0, 0.00226, -0.00093, -0.0, 0.00177, -0.00089, -0.0, 0.00189, -0.00062, 0.0, 0.00119, -0.00276, 0.00013, 0.00119, -0.00276, -0.00013, -0.00332, 0.00898, 0.00117, 0.00332, -0.00898, 0.00117, 0.004, -0.00869, -0.00117, 0.004, -0.00869, 0.00117, -0.00159, -0.00277, -6e-05, -0.00159, -0.00277, 6e-05, -0.00083, -0.00308, 6e-05, -0.00083, -0.00308, -6e-05, -0.00898, -0.00332, 0.00117, -0.00898, -0.00332, -0.00117, -0.00869, -0.004, 0.00117, -0.00869, -0.004, -0.00117, -0.00277, 0.00159, 6e-05, -0.00277, 0.00159, -6e-05, -0.00308, 0.00083, -6e-05, -0.00308, 0.00083, 6e-05, -0.00332, 0.00898, -0.00117, -0.004, 0.00869, 0.00117, -0.004, 0.00869, -0.00117, 0.00159, 0.00277, 6e-05, 0.00159, 0.00277, -6e-05, 0.00083, 0.00308, -6e-05, 0.00083, 0.00308, 6e-05, 0.00898, 0.00332, -0.00117, 0.00898, 0.00332, 0.00117, 0.00869, 0.004, -0.00117, 0.00869, 0.004, 0.00117, 0.00277, -0.00159, -6e-05, 0.00277, -0.00159, 6e-05, 0.00308, -0.00083, 6e-05, 0.00308, -0.00083, -6e-05, 0.00332, -0.00898, -0.00117, -0.01424, 0.03437, -0.02273, -0.01424, 0.03437, 0.02273, -0.001, -0.00241, -0.0, -0.001, -0.00241, 0.0, 0.03437, 0.01424, 0.02273, 0.03437, 0.01424, -0.02273, -0.00241, 0.001, -0.0, -0.00241, 0.001, 0.0, 0.01424, -0.03437, 0.02273, 0.01424, -0.03437, -0.02273, 0.001, 0.00241, 0.0, 0.001, 0.00241, -0.0, -0.03437, -0.01424, -0.02273, -0.03437, -0.01424, 0.02273, 0.00241, -0.001, 0.0, 0.00241, -0.001, -0.0]]), "E2g4":array([[0.00944, 0.00391, -0.0, 0.00944, 0.00391, 0.0, 0.00372, 0.00532, 0.0, 0.00639, -0.00113, -0.0, 0.01178, -0.00528, -0.00086, 0.01178, -0.00528, 0.00086, 0.01411, -0.00584, -6e-05, 0.01411, -0.00584, 6e-05, 0.01433, -0.00599, -0.0, 0.01437, -0.0059, -0.0, 0.01207, -0.0046, -0.00086, 0.01207, -0.0046, 0.00086, 0.00391, -0.00944, -0.0, 0.00391, -0.00944, 0.0, 0.00532, -0.00372, -0.0, -0.00113, -0.00639, -0.0, -0.00528, -0.01178, 0.00086, -0.00528, -0.01178, -0.00086, -0.00584, -0.01411, 6e-05, -0.00584, -0.01411, -6e-05, -0.00599, -0.01433, -0.0, -0.0059, -0.01437, -0.0, -0.0046, -0.01207, 0.00086, -0.0046, -0.01207, -0.00086, -0.00944, -0.00391, 0.0, -0.00944, -0.00391, -0.0, -0.00372, -0.00532, -0.0, -0.00639, 0.00113, 0.0, -0.01178, 0.00528, 0.00086, -0.01178, 0.00528, -0.00086, -0.01411, 0.00584, 6e-05, -0.01411, 0.00584, -6e-05, -0.01433, 0.00599, 0.0, -0.01437, 0.0059, 0.0, -0.01207, 0.0046, 0.00086, -0.01207, 0.0046, -0.00086, -0.00391, 0.00944, 0.0, -0.00391, 0.00944, -0.0, -0.00532, 0.00372, 0.0, 0.00113, 0.00639, 0.0, 0.00528, 0.01178, -0.00086, 0.00528, 0.01178, 0.00086, 0.00584, 0.01411, -6e-05, 0.00584, 0.01411, 6e-05, 0.00599, 0.01433, 0.0, 0.0059, 0.01437, 0.0, 0.0046, 0.01207, -0.00086, 0.0046, 0.01207, 0.00086, -0.00601, -0.00797, -1e-05, 0.00601, 0.00797, -1e-05, 0.00988, -0.00139, -1e-05, 0.00988, -0.00139, 1e-05, 0.01444, -0.00593, 4e-05, 0.01444, -0.00593, -4e-05, 0.0144, -0.00602, 4e-05, 0.0144, -0.00602, -4e-05, 0.00797, -0.00601, -1e-05, 0.00797, -0.00601, 1e-05, -0.00139, -0.00988, 1e-05, -0.00139, -0.00988, -1e-05, -0.00593, -0.01444, -4e-05, -0.00593, -0.01444, 4e-05, -0.00602, -0.0144, -4e-05, -0.00602, -0.0144, 4e-05, -0.00601, -0.00797, 1e-05, -0.00988, 0.00139, 1e-05, -0.00988, 0.00139, -1e-05, -0.01444, 0.00593, -4e-05, -0.01444, 0.00593, 4e-05, -0.0144, 0.00602, -4e-05, -0.0144, 0.00602, 4e-05, -0.00797, 0.00601, 1e-05, -0.00797, 0.00601, -1e-05, 0.00139, 0.00988, -1e-05, 0.00139, 0.00988, 1e-05, 0.00593, 0.01444, 4e-05, 0.00593, 0.01444, -4e-05, 0.00602, 0.0144, 4e-05, 0.00602, 0.0144, -4e-05, 0.00601, 0.00797, 1e-05, 0.01154, 0.00478, 0.0, 0.01154, 0.00478, 0.0, 0.01387, -0.00575, -0.0002, 0.01387, -0.00575, 0.0002, 0.00478, -0.01154, -0.0, 0.00478, -0.01154, -0.0, -0.00575, -0.01387, 0.0002, -0.00575, -0.01387, -0.0002, -0.01154, -0.00478, -0.0, -0.01154, -0.00478, -0.0, -0.01387, 0.00575, 0.0002, -0.01387, 0.00575, -0.0002, -0.00478, 0.01154, 0.0, -0.00478, 0.01154, 0.0, 0.00575, 0.01387, -0.0002, 0.00575, 0.01387, 0.0002], [-0.0039, -0.00161, 0.0, -0.0039, -0.00161, -0.0, -0.00181, -0.00097, 0.0, -0.00196, -0.00059, 0.0, -0.00351, 0.00073, -0.00019, -0.00351, 0.00073, 0.00019, 0.00767, -0.00318, 0.01013, 0.00767, -0.00318, -0.01013, -0.01089, 0.00592, 0.0, -0.01189, 0.00351, 0.0, -0.003, 0.00197, -0.00019, -0.003, 0.00197, 0.00019, -0.00161, 0.0039, 0.0, -0.00161, 0.0039, -0.0, -0.00097, 0.00181, 0.0, -0.00059, 0.00196, 0.0, 0.00073, 0.00351, 0.00019, 0.00073, 0.00351, -0.00019, -0.00318, -0.00767, -0.01013, -0.00318, -0.00767, 0.01013, 0.00592, 0.01089, 0.0, 0.00351, 0.01189, 0.0, 0.00197, 0.003, 0.00019, 0.00197, 0.003, -0.00019, 0.0039, 0.00161, -0.0, 0.0039, 0.00161, 0.0, 0.00181, 0.00097, -0.0, 0.00196, 0.00059, -0.0, 0.00351, -0.00073, 0.00019, 0.00351, -0.00073, -0.00019, -0.00767, 0.00318, -0.01013, -0.00767, 0.00318, 0.01013, 0.01089, -0.00592, -0.0, 0.01189, -0.00351, -0.0, 0.003, -0.00197, 0.00019, 0.003, -0.00197, -0.00019, 0.00161, -0.0039, -0.0, 0.00161, -0.0039, 0.0, 0.00097, -0.00181, -0.0, 0.00059, -0.00196, -0.0, -0.00073, -0.00351, -0.00019, -0.00073, -0.00351, 0.00019, 0.00318, 0.00767, 0.01013, 0.00318, 0.00767, -0.01013, -0.00592, -0.01089, -0.0, -0.00351, -0.01189, -0.0, -0.00197, -0.003, -0.00019, -0.00197, -0.003, 0.00019, 0.00232, 0.00072, 4e-05, -0.00232, -0.00072, 4e-05, -0.00215, -0.00113, 4e-05, -0.00215, -0.00113, -4e-05, -0.0081, 0.00385, 0.00139, -0.0081, 0.00385, -0.00139, -0.00844, 0.00301, 0.00139, -0.00844, 0.00301, -0.00139, -0.00072, 0.00232, 4e-05, -0.00072, 0.00232, -4e-05, -0.00113, 0.00215, -4e-05, -0.00113, 0.00215, 4e-05, 0.00385, 0.0081, -0.00139, 0.00385, 0.0081, 0.00139, 0.00301, 0.00844, -0.00139, 0.00301, 0.00844, 0.00139, 0.00232, 0.00072, -4e-05, 0.00215, 0.00113, -4e-05, 0.00215, 0.00113, 4e-05, 0.0081, -0.00385, -0.00139, 0.0081, -0.00385, 0.00139, 0.00844, -0.00301, -0.00139, 0.00844, -0.00301, 0.00139, 0.00072, -0.00232, -4e-05, 0.00072, -0.00232, 4e-05, 0.00113, -0.00215, 4e-05, 0.00113, -0.00215, -4e-05, -0.00385, -0.0081, 0.00139, -0.00385, -0.0081, -0.00139, -0.00301, -0.00844, 0.00139, -0.00301, -0.00844, -0.00139, -0.00232, -0.00072, -4e-05, -0.0027, -0.00112, -0.0, -0.0027, -0.00112, 0.0, 0.03507, -0.01453, 0.02274, 0.03507, -0.01453, -0.02274, -0.00112, 0.0027, -0.0, -0.00112, 0.0027, 0.0, -0.01453, -0.03507, -0.02274, -0.01453, -0.03507, 0.02274, 0.0027, 0.00112, 0.0, 0.0027, 0.00112, -0.0, -0.03507, 0.01453, -0.02274, -0.03507, 0.01453, 0.02274, 0.00112, -0.0027, 0.0, 0.00112, -0.0027, -0.0, 0.01453, 0.03507, 0.02274, 0.01453, 0.03507, -0.02274]]), "E1g1":array([[-0.00485, -0.00805, 0.0017, 0.00485, 0.00805, 0.0017, 0.0, 0.0, -0.00409, 0.0, 0.0, 0.00145, -0.00883, 0.00747, 0.00242, 0.00883, -0.00747, 0.00242, -0.01035, 0.00788, -0.00391, 0.01035, -0.00788, -0.00391, 0.0, 0.0, 0.0041, -0.0, 0.0, 0.00184, -0.0, 0.0021, 0.0, -0.0, -0.0021, -0.0, 0.01035, 0.00788, 0.00391, -0.01035, -0.00788, 0.00391, -0.0, 0.0, -0.00184, 0.0, 0.0, -0.0041, 0.00883, 0.00747, -0.00242, -0.00883, -0.00747, -0.00242, 0.00485, -0.00805, -0.0017, -0.00485, 0.00805, -0.0017, 0.0, 0.0, -0.00145, 0.0, -0.0, 0.00409, -0.0, -0.01747, 0.00348, 0.0, 0.01747, 0.00348, 0.00485, 0.00805, -0.0017, -0.00485, -0.00805, -0.0017, -0.0, -0.0, 0.00409, -0.0, 0.0, -0.00145, 0.00883, -0.00747, -0.00242, -0.00883, 0.00747, -0.00242, 0.01035, -0.00788, 0.00391, -0.01035, 0.00788, 0.00391, 0.0, -0.0, -0.0041, 0.0, -0.0, -0.00184, 0.0, -0.0021, -0.0, 0.0, 0.0021, 0.0, -0.01035, -0.00788, -0.00391, 0.01035, 0.00788, -0.00391, 0.0, -0.0, 0.00184, 0.0, -0.0, 0.0041, -0.00883, -0.00747, 0.00242, 0.00883, 0.00747, 0.00242, -0.00485, 0.00805, 0.0017, 0.00485, -0.00805, 0.0017, -0.0, 0.0, 0.00145, -0.0, 0.0, -0.00409, 0.0, 0.01747, -0.00348, -0.0, -0.01747, -0.00348, 0.00013, 0.01301, -0.00093, 0.00013, 0.01301, 0.00093, -0.00632, 0.0026, 0.00064, 0.00632, -0.0026, 0.00064, -0.00999, 0.00607, -0.00203, 0.00999, -0.00607, -0.00203, -0.00297, 0.00385, -0.00166, 0.00297, -0.00385, -0.00166, 0.00297, 0.00385, 0.00166, -0.00297, -0.00385, 0.00166, 0.00999, 0.00607, 0.00203, -0.00999, -0.00607, 0.00203, 0.00632, 0.0026, -0.00064, -0.00632, -0.0026, -0.00064, 0.00013, -0.01301, -0.00093, -0.00013, 0.01301, -0.00093, -0.00013, -0.01301, -0.00093, 0.00632, -0.0026, -0.00064, -0.00632, 0.0026, -0.00064, 0.00999, -0.00607, 0.00203, -0.00999, 0.00607, 0.00203, 0.00297, -0.00385, 0.00166, -0.00297, 0.00385, 0.00166, -0.00297, -0.00385, -0.00166, 0.00297, 0.00385, -0.00166, -0.00999, -0.00607, -0.00203, 0.00999, 0.00607, -0.00203, -0.00632, -0.0026, 0.00064, 0.00632, 0.0026, 0.00064, -0.00013, 0.01301, 0.00093, 0.00013, -0.01301, 0.00093, -0.00013, -0.01301, 0.00093, -0.00725, -0.01255, 0.00357, 0.00725, 0.01255, 0.00357, -0.01661, 0.01243, -0.00759, 0.01661, -0.01243, -0.00759, 0.01661, 0.01243, 0.00759, -0.01661, -0.01243, 0.00759, 0.00725, -0.01255, -0.00357, -0.00725, 0.01255, -0.00357, 0.00725, 0.01255, -0.00357, -0.00725, -0.01255, -0.00357, 0.01661, -0.01243, 0.00759, -0.01661, 0.01243, 0.00759, -0.01661, -0.01243, -0.00759, 0.01661, 0.01243, -0.00759, -0.00725, 0.01255, 0.00357, 0.00725, -0.01255, 0.00357], [0.00442, -0.01358, 0.00778, -0.00442, 0.01358, 0.00778, -0.0, 0.0, -0.00127, -0.0, -0.0, 6e-05, 0.0035, -0.00438, 0.0009, -0.0035, 0.00438, 0.0009, 0.00629, -0.005, 0.00481, -0.00629, 0.005, 0.00481, -0.0, -0.0, 0.0009, -0.0, 0.0, 0.00177, -0.0, 0.00125, 0.0, -0.0, -0.00125, 0.0, -0.00629, -0.005, -0.00481, 0.00629, 0.005, -0.00481, -0.0, -0.0, -0.00177, -0.0, 0.0, -0.0009, -0.0035, -0.00438, -0.0009, 0.0035, 0.00438, -0.0009, -0.00442, -0.01358, -0.00778, 0.00442, 0.01358, -0.00778, -0.0, -0.0, -6e-05, -0.0, -0.0, 0.00127, -0.0, -0.00928, -0.00029, 0.0, 0.00928, -0.00029, -0.00442, 0.01358, -0.00778, 0.00442, -0.01358, -0.00778, 0.0, -0.0, 0.00127, 0.0, -0.0, -6e-05, -0.0035, 0.00438, -0.0009, 0.0035, -0.00438, -0.0009, -0.00629, 0.005, -0.00481, 0.00629, -0.005, -0.00481, -0.0, 0.0, -0.0009, 0.0, -0.0, -0.00177, 0.0, -0.00125, -0.0, 0.0, 0.00125, -0.0, 0.00629, 0.005, 0.00481, -0.00629, -0.005, 0.00481, 0.0, 0.0, 0.00177, -0.0, -0.0, 0.0009, 0.0035, 0.00438, 0.0009, -0.0035, -0.00438, 0.0009, 0.00442, 0.01358, 0.00778, -0.00442, -0.01358, 0.00778, 0.0, -0.0, 6e-05, 0.0, 0.0, -0.00127, 0.0, 0.00928, 0.00029, -0.0, -0.00928, 0.00029, -0.00056, 0.00633, -0.00276, -0.00056, 0.00633, 0.00276, 0.00277, -0.00837, 0.0055, -0.00277, 0.00837, 0.0055, -0.00225, -0.00141, 3e-05, 0.00225, 0.00141, 3e-05, 0.00764, -0.00573, 0.00689, -0.00764, 0.00573, 0.00689, -0.00764, -0.00573, -0.00689, 0.00764, 0.00573, -0.00689, 0.00225, -0.00141, -3e-05, -0.00225, 0.00141, -3e-05, -0.00277, -0.00837, -0.0055, 0.00277, 0.00837, -0.0055, -0.00056, -0.00633, -0.00276, 0.00056, 0.00633, -0.00276, 0.00056, -0.00633, -0.00276, -0.00277, 0.00837, -0.0055, 0.00277, -0.00837, -0.0055, 0.00225, 0.00141, -3e-05, -0.00225, -0.00141, -3e-05, -0.00764, 0.00573, -0.00689, 0.00764, -0.00573, -0.00689, 0.00764, 0.00573, 0.00689, -0.00764, -0.00573, 0.00689, -0.00225, 0.00141, 3e-05, 0.00225, -0.00141, 3e-05, 0.00277, 0.00837, 0.0055, -0.00277, -0.00837, 0.0055, 0.00056, 0.00633, 0.00276, -0.00056, -0.00633, 0.00276, 0.00056, -0.00633, 0.00276, 0.00699, -0.02482, 0.01274, -0.00699, 0.02482, 0.01274, 0.01007, -0.00783, 0.00695, -0.01007, 0.00783, 0.00695, -0.01007, -0.00783, -0.00695, 0.01007, 0.00783, -0.00695, -0.00699, -0.02482, -0.01274, 0.00699, 0.02482, -0.01274, -0.00699, 0.02482, -0.01274, 0.00699, -0.02482, -0.01274, -0.01007, 0.00783, -0.00695, 0.01007, -0.00783, -0.00695, 0.01007, 0.00783, 0.00695, -0.01007, -0.00783, 0.00695, 0.00699, 0.02482, 0.01274, -0.00699, -0.02482, 0.01274]]), "E1g2":array([[0.00876, -0.00139, 0.00122, -0.00876, 0.00139, 0.00122, 0.0, -0.0, 0.00152, 0.0, -0.0, -0.00387, 0.01169, -0.01169, -0.004, -0.01169, 0.01169, -0.004, 0.00139, -0.00876, 0.00122, -0.00139, 0.00876, 0.00122, 0.0, -0.0, -0.00387, 0.0, 0.0, 0.00152, -0.0123, -0.00112, 0.00236, 0.0123, 0.00112, 0.00236, -0.01404, -0.00213, -0.00457, 0.01404, 0.00213, -0.00457, 0.0, 0.0, 0.00422, 0.0, 0.0, 0.0017, -0.00116, 0.00116, -0.0, 0.00116, -0.00116, 0.0, 0.00213, 0.01404, 0.00457, -0.00213, -0.01404, 0.00457, 0.0, 0.0, -0.0017, 0.0, -0.0, -0.00422, 0.00112, 0.0123, -0.00236, -0.00112, -0.0123, -0.00236, -0.00876, 0.00139, -0.00122, 0.00876, -0.00139, -0.00122, -0.0, 0.0, -0.00152, -0.0, 0.0, 0.00387, -0.01169, 0.01169, 0.004, 0.01169, -0.01169, 0.004, -0.00139, 0.00876, -0.00122, 0.00139, -0.00876, -0.00122, -0.0, 0.0, 0.00387, -0.0, -0.0, -0.00152, 0.0123, 0.00112, -0.00236, -0.0123, -0.00112, -0.00236, 0.01404, 0.00213, 0.00457, -0.01404, -0.00213, 0.00457, -0.0, -0.0, -0.00422, -0.0, -0.0, -0.0017, 0.00116, -0.00116, 0.0, -0.00116, 0.00116, -0.0, -0.00213, -0.01404, -0.00457, 0.00213, 0.01404, -0.00457, -0.0, -0.0, 0.0017, -0.0, 0.0, 0.00422, -0.00112, -0.0123, 0.00236, 0.00112, 0.0123, 0.00236, -0.00294, -0.00685, -0.00055, -0.00294, -0.00685, 0.00055, 0.00853, -0.00877, 0.00075, -0.00853, 0.00877, 0.00075, 0.00877, -0.00853, 0.00075, -0.00877, 0.00853, 0.00075, -0.00685, -0.00294, 0.00055, 0.00685, 0.00294, 0.00055, -0.0118, -0.00281, -0.002, 0.0118, 0.00281, -0.002, -0.006, 5e-05, -0.00233, 0.006, -5e-05, -0.00233, -5e-05, 0.006, 0.00233, 5e-05, -0.006, 0.00233, 0.00281, 0.0118, 0.002, -0.00281, -0.0118, 0.002, 0.00294, 0.00685, -0.00055, -0.00853, 0.00877, -0.00075, 0.00853, -0.00877, -0.00075, -0.00877, 0.00853, -0.00075, 0.00877, -0.00853, -0.00075, 0.00685, 0.00294, -0.00055, -0.00685, -0.00294, -0.00055, 0.0118, 0.00281, 0.002, -0.0118, -0.00281, 0.002, 0.006, -5e-05, 0.00233, -0.006, 5e-05, 0.00233, 5e-05, -0.006, -0.00233, -5e-05, 0.006, -0.00233, -0.00281, -0.0118, -0.002, 0.00281, 0.0118, -0.002, 0.00294, 0.00685, 0.00055, 0.01352, -0.0021, 0.00276, -0.01352, 0.0021, 0.00276, 0.0021, -0.01352, 0.00276, -0.0021, 0.01352, 0.00276, -0.02277, -0.00397, -0.00887, 0.02277, 0.00397, -0.00887, 0.00397, 0.02277, 0.00887, -0.00397, -0.02277, 0.00887, -0.01352, 0.0021, -0.00276, 0.01352, -0.0021, -0.00276, -0.0021, 0.01352, -0.00276, 0.0021, -0.01352, -0.00276, 0.02277, 0.00397, 0.00887, -0.02277, -0.00397, 0.00887, -0.00397, -0.02277, -0.00887, 0.00397, 0.02277, -0.00887], [-0.00821, 0.01087, -0.00664, 0.00821, -0.01087, -0.00664, -0.0, -0.0, -0.00071, -0.0, 0.0, 0.00174, -0.00869, 0.00869, 0.00081, 0.00869, -0.00869, 0.00081, -0.01087, 0.00821, -0.00664, 0.01087, -0.00821, -0.00664, -0.0, 0.0, 0.00174, 0.0, 0.0, -0.00071, -0.00187, 0.00091, -0.0016, 0.00187, -0.00091, -0.0016, -0.00649, -0.00175, -0.00574, 0.00649, 0.00175, -0.00574, 0.0, -0.0, -0.00073, -0.0, 0.0, -0.00173, 0.00105, -0.00105, 0.0, -0.00105, 0.00105, 0.0, 0.00175, 0.00649, 0.00574, -0.00175, -0.00649, 0.00574, 0.0, 0.0, 0.00173, 0.0, 0.0, 0.00073, -0.00091, 0.00187, 0.0016, 0.00091, -0.00187, 0.0016, 0.00821, -0.01087, 0.00664, -0.00821, 0.01087, 0.00664, 0.0, 0.0, 0.00071, 0.0, -0.0, -0.00174, 0.00869, -0.00869, -0.00081, -0.00869, 0.00869, -0.00081, 0.01087, -0.00821, 0.00664, -0.01087, 0.00821, 0.00664, 0.0, -0.0, -0.00174, -0.0, -0.0, 0.00071, 0.00187, -0.00091, 0.0016, -0.00187, 0.00091, 0.0016, 0.00649, 0.00175, 0.00574, -0.00649, -0.00175, 0.00574, -0.0, 0.0, 0.00073, 0.0, -0.0, 0.00173, -0.00105, 0.00105, -0.0, 0.00105, -0.00105, -0.0, -0.00175, -0.00649, -0.00574, 0.00175, 0.00649, -0.00574, -0.0, -0.0, -0.00173, -0.0, -0.0, -0.00073, 0.00091, -0.00187, -0.0016, -0.00091, 0.00187, -0.0016, 0.0038, -0.00452, 0.00434, 0.0038, -0.00452, -0.00434, -0.00533, 0.00649, -0.00261, 0.00533, -0.00649, -0.00261, -0.00649, 0.00533, -0.00261, 0.00649, -0.00533, -0.00261, -0.00452, 0.0038, -0.00434, 0.00452, -0.0038, -0.00434, 0.0021, 0.00312, -0.0002, -0.0021, -0.00312, -0.0002, -0.01046, -0.00252, -0.0074, 0.01046, 0.00252, -0.0074, 0.00252, 0.01046, 0.0074, -0.00252, -0.01046, 0.0074, -0.00312, -0.0021, 0.0002, 0.00312, 0.0021, 0.0002, -0.0038, 0.00452, 0.00434, 0.00533, -0.00649, 0.00261, -0.00533, 0.00649, 0.00261, 0.00649, -0.00533, 0.00261, -0.00649, 0.00533, 0.00261, 0.00452, -0.0038, 0.00434, -0.00452, 0.0038, 0.00434, -0.0021, -0.00312, 0.0002, 0.0021, 0.00312, 0.0002, 0.01046, 0.00252, 0.0074, -0.01046, -0.00252, 0.0074, -0.00252, -0.01046, -0.0074, 0.00252, 0.01046, -0.0074, 0.00312, 0.0021, -0.0002, -0.00312, -0.0021, -0.0002, -0.0038, 0.00452, -0.00434, -0.01311, 0.01945, -0.01086, 0.01311, -0.01945, -0.01086, -0.01945, 0.01311, -0.01086, 0.01945, -0.01311, -0.01086, -0.01202, -0.00266, -0.00806, 0.01202, 0.00266, -0.00806, 0.00266, 0.01202, 0.00806, -0.00266, -0.01202, 0.00806, 0.01311, -0.01945, 0.01086, -0.01311, 0.01945, 0.01086, 0.01945, -0.01311, 0.01086, -0.01945, 0.01311, 0.01086, 0.01202, 0.00266, 0.00806, -0.01202, -0.00266, 0.00806, -0.00266, -0.01202, -0.00806, 0.00266, 0.01202, -0.00806]]), "E1g3":array([[0.00772, -0.01071, 0.00413, -0.00772, 0.01071, 0.00413, -0.0, -0.0, -0.00168, -0.0, 0.0, -0.00393, 0.0074, -0.00887, -0.00269, -0.0074, 0.00887, -0.00269, -0.00766, -0.00474, -0.00165, 0.00766, 0.00474, -0.00165, -0.0, -0.0, -0.00161, 0.0, -0.0, 0.00346, -0.01584, 0.0, 0.00379, 0.01584, -0.0, 0.00379, -0.00766, 0.00474, -0.00165, 0.00766, -0.00474, -0.00165, 0.0, -0.0, 0.00346, -0.0, -0.0, -0.00161, 0.0074, 0.00887, -0.00269, -0.0074, -0.00887, -0.00269, 0.00772, 0.01071, 0.00413, -0.00772, -0.01071, 0.00413, -0.0, 0.0, -0.00393, 0.0, 0.0, -0.00168, 0.00133, 0.0, -0.0, -0.00133, 0.0, 0.0, -0.00772, 0.01071, -0.00413, 0.00772, -0.01071, -0.00413, 0.0, 0.0, 0.00168, 0.0, -0.0, 0.00393, -0.0074, 0.00887, 0.00269, 0.0074, -0.00887, 0.00269, 0.00766, 0.00474, 0.00165, -0.00766, -0.00474, 0.00165, -0.0, 0.0, 0.00161, -0.0, 0.0, -0.00346, 0.01584, -0.0, -0.00379, -0.01584, 0.0, -0.00379, 0.00766, -0.00474, 0.00165, -0.00766, 0.00474, 0.00165, -0.0, 0.0, -0.00346, -0.0, 0.0, 0.00161, -0.0074, -0.00887, 0.00269, 0.0074, 0.00887, 0.00269, -0.00772, -0.01071, -0.00413, 0.00772, 0.01071, -0.00413, 0.0, -0.0, 0.00393, -0.0, -0.0, 0.00168, -0.00133, -0.0, 0.0, 0.00133, -0.0, -0.0, -0.00388, 0.00366, -0.00175, -0.00388, 0.00366, 0.00175, 0.00586, -0.00991, 0.00194, -0.00586, 0.00991, 0.00194, 0.00201, -0.0059, -0.00089, -0.00201, 0.0059, -0.00089, -0.01174, -0.00017, -0.00057, 0.01174, 0.00017, -0.00057, -0.01174, 0.00017, -0.00057, 0.01174, -0.00017, -0.00057, 0.00201, 0.0059, -0.00089, -0.00201, -0.0059, -0.00089, 0.00586, 0.00991, 0.00194, -0.00586, -0.00991, 0.00194, 0.00388, 0.00366, 0.00175, -0.00388, -0.00366, 0.00175, 0.00388, -0.00366, -0.00175, -0.00586, 0.00991, -0.00194, 0.00586, -0.00991, -0.00194, -0.00201, 0.0059, 0.00089, 0.00201, -0.0059, 0.00089, 0.01174, 0.00017, 0.00057, -0.01174, -0.00017, 0.00057, 0.01174, -0.00017, 0.00057, -0.01174, 0.00017, 0.00057, -0.00201, -0.0059, 0.00089, 0.00201, 0.0059, 0.00089, -0.00586, -0.00991, -0.00194, 0.00586, 0.00991, -0.00194, -0.00388, -0.00366, -0.00175, 0.00388, 0.00366, -0.00175, 0.00388, -0.00366, 0.00175, 0.01205, -0.01727, 0.00798, -0.01205, 0.01727, 0.00798, -0.01237, -0.00709, -0.00332, 0.01237, 0.00709, -0.00332, -0.01237, 0.00709, -0.00332, 0.01237, -0.00709, -0.00332, 0.01205, 0.01727, 0.00798, -0.01205, -0.01727, 0.00798, -0.01205, 0.01727, -0.00798, 0.01205, -0.01727, -0.00798, 0.01237, 0.00709, 0.00332, -0.01237, -0.00709, 0.00332, 0.01237, -0.00709, 0.00332, -0.01237, 0.00709, 0.00332, -0.01205, -0.01727, -0.00798, 0.01205, 0.01727, -0.00798], [-0.00435, -5e-05, 0.00198, 0.00435, 5e-05, 0.00198, -0.0, -0.0, 0.00168, 0.0, -0.0, 0.00408, -0.00553, 0.00832, 0.00075, 0.00553, -0.00832, 0.00075, -0.01268, 0.00811, -0.00949, 0.01268, -0.00811, -0.00949, -0.0, -0.0, 0.00011, -0.0, 0.0, -0.00176, 7e-05, 0.0, -0.00331, -7e-05, 0.0, -0.00331, -0.01268, -0.00811, -0.00949, 0.01268, 0.00811, -0.00949, -0.0, -0.0, -0.00176, -0.0, -0.0, 0.00011, -0.00553, -0.00832, 0.00075, 0.00553, 0.00832, 0.00075, -0.00435, 5e-05, 0.00198, 0.00435, -5e-05, 0.00198, 0.0, -0.0, 0.00408, 0.0, 0.0, 0.00168, -0.00344, 0.0, -0.0, 0.00344, 0.0, -0.0, 0.00435, 5e-05, -0.00198, -0.00435, -5e-05, -0.00198, 0.0, 0.0, -0.00168, -0.0, 0.0, -0.00408, 0.00553, -0.00832, -0.00075, -0.00553, 0.00832, -0.00075, 0.01268, -0.00811, 0.00949, -0.01268, 0.00811, 0.00949, -0.0, 0.0, -0.00011, 0.0, -0.0, 0.00176, -7e-05, -0.0, 0.00331, 7e-05, -0.0, 0.00331, 0.01268, 0.00811, 0.00949, -0.01268, -0.00811, 0.00949, 0.0, 0.0, 0.00176, -0.0, 0.0, -0.00011, 0.00553, 0.00832, -0.00075, -0.00553, -0.00832, -0.00075, 0.00435, -5e-05, -0.00198, -0.00435, 5e-05, -0.00198, -0.0, 0.0, -0.00408, -0.0, -0.0, -0.00168, 0.00344, -0.0, 0.0, -0.00344, -0.0, 0.0, 0.00308, -0.00052, -0.00082, 0.00308, -0.00052, 0.00082, -0.00221, -0.00361, 0.00571, 0.00221, 0.00361, 0.00571, -0.01444, 0.00727, -0.00864, 0.01444, -0.00727, -0.00864, -0.00019, 0.0007, -0.00407, 0.00019, -0.0007, -0.00407, -0.00019, -0.0007, -0.00407, 0.00019, 0.0007, -0.00407, -0.01444, -0.00727, -0.00864, 0.01444, 0.00727, -0.00864, -0.00221, 0.00361, 0.00571, 0.00221, -0.00361, 0.00571, -0.00308, -0.00052, 0.00082, 0.00308, 0.00052, 0.00082, -0.00308, 0.00052, -0.00082, 0.00221, 0.00361, -0.00571, -0.00221, -0.00361, -0.00571, 0.01444, -0.00727, 0.00864, -0.01444, 0.00727, 0.00864, 0.00019, -0.0007, 0.00407, -0.00019, 0.0007, 0.00407, 0.00019, 0.0007, 0.00407, -0.00019, -0.0007, 0.00407, 0.01444, 0.00727, 0.00864, -0.01444, -0.00727, 0.00864, 0.00221, -0.00361, -0.00571, -0.00221, 0.00361, -0.00571, 0.00308, 0.00052, -0.00082, -0.00308, -0.00052, -0.00082, -0.00308, 0.00052, 0.00082, -0.00697, 0.00147, 0.00144, 0.00697, -0.00147, 0.00144, -0.02207, 0.01347, -0.01439, 0.02207, -0.01347, -0.01439, -0.02207, -0.01347, -0.01439, 0.02207, 0.01347, -0.01439, -0.00697, -0.00147, 0.00144, 0.00697, 0.00147, 0.00144, 0.00697, -0.00147, -0.00144, -0.00697, 0.00147, -0.00144, 0.02207, -0.01347, 0.01439, -0.02207, 0.01347, 0.01439, 0.02207, 0.01347, 0.01439, -0.02207, -0.01347, 0.01439, 0.00697, 0.00147, -0.00144, -0.00697, -0.00147, -0.00144]]), "E1g4":array([[-0.0023, 0.01315, -0.0039, 0.0023, -0.01315, -0.0039, -0.0, 0.0, 0.00395, -0.0, -0.0, 0.00193, 0.00093, 0.00093, 0.0, -0.00093, -0.00093, 0.0, 0.01315, -0.0023, 0.0039, -0.01315, 0.0023, 0.0039, 0.0, 0.0, -0.00193, -0.0, -0.0, -0.00395, 0.01222, -0.00137, -0.00295, -0.01222, 0.00137, -0.00295, -0.00248, -0.00971, -0.00216, 0.00248, 0.00971, -0.00216, 0.0, -0.0, -0.00144, -0.0, 0.0, 0.00426, -0.01226, -0.01226, 0.00357, 0.01226, 0.01226, 0.00357, -0.00971, -0.00248, -0.00216, 0.00971, 0.00248, -0.00216, 0.0, 0.0, 0.00426, -0.0, 0.0, -0.00144, -0.00137, 0.01222, -0.00295, 0.00137, -0.01222, -0.00295, 0.0023, -0.01315, 0.0039, -0.0023, 0.01315, 0.0039, 0.0, -0.0, -0.00395, 0.0, 0.0, -0.00193, -0.00093, -0.00093, -0.0, 0.00093, 0.00093, -0.0, -0.01315, 0.0023, -0.0039, 0.01315, -0.0023, -0.0039, -0.0, -0.0, 0.00193, 0.0, 0.0, 0.00395, -0.01222, 0.00137, 0.00295, 0.01222, -0.00137, 0.00295, 0.00248, 0.00971, 0.00216, -0.00248, -0.00971, 0.00216, -0.0, 0.0, 0.00144, 0.0, -0.0, -0.00426, 0.01226, 0.01226, -0.00357, -0.01226, -0.01226, -0.00357, 0.00971, 0.00248, 0.00216, -0.00971, -0.00248, 0.00216, -0.0, -0.0, -0.00426, 0.0, -0.0, 0.00144, 0.00137, -0.01222, 0.00295, -0.00137, 0.01222, 0.00295, 0.00268, -0.01135, 0.00165, 0.00268, -0.01135, -0.00165, 0.00014, 0.00536, -0.00178, -0.00014, -0.00536, -0.00178, 0.00536, 0.00014, 0.00178, -0.00536, -0.00014, 0.00178, 0.01135, -0.00268, 0.00165, -0.01135, 0.00268, 0.00165, 0.0065, -0.00298, -0.00084, -0.0065, 0.00298, -0.00084, -0.00924, -0.0092, -0.00102, 0.00924, 0.0092, -0.00102, -0.0092, -0.00924, -0.00102, 0.0092, 0.00924, -0.00102, -0.00298, 0.0065, -0.00084, 0.00298, -0.0065, -0.00084, -0.00268, 0.01135, 0.00165, -0.00014, -0.00536, 0.00178, 0.00014, 0.00536, 0.00178, -0.00536, -0.00014, -0.00178, 0.00536, 0.00014, -0.00178, -0.01135, 0.00268, -0.00165, 0.01135, -0.00268, -0.00165, -0.0065, 0.00298, 0.00084, 0.0065, -0.00298, 0.00084, 0.00924, 0.0092, 0.00102, -0.00924, -0.0092, 0.00102, 0.0092, 0.00924, 0.00102, -0.0092, -0.00924, 0.00102, 0.00298, -0.0065, 0.00084, -0.00298, 0.0065, 0.00084, -0.00268, 0.01135, -0.00165, -0.00355, 0.02078, -0.0078, 0.00355, -0.02078, -0.0078, 0.02078, -0.00355, 0.0078, -0.02078, 0.00355, 0.0078, -0.00434, -0.01495, -0.00358, 0.00434, 0.01495, -0.00358, -0.01495, -0.00434, -0.00358, 0.01495, 0.00434, -0.00358, 0.00355, -0.02078, 0.0078, -0.00355, 0.02078, 0.0078, -0.02078, 0.00355, -0.0078, 0.02078, -0.00355, -0.0078, 0.00434, 0.01495, 0.00358, -0.00434, -0.01495, 0.00358, 0.01495, 0.00434, 0.00358, -0.01495, -0.00434, 0.00358], [-0.00163, -0.00504, 0.0, 0.00163, 0.00504, 0.0, -0.0, 0.0, -0.00325, 0.0, -0.0, -0.00166, -0.00158, -0.00158, -0.0, 0.00158, 0.00158, 0.0, -0.00504, -0.00163, -0.0, 0.00504, 0.00163, -0.0, -0.0, -0.0, 0.00166, -0.0, 0.0, 0.00325, -0.00935, 0.00133, 0.00051, 0.00935, -0.00133, 0.00051, -0.01481, -0.00504, -0.00965, 0.01481, 0.00504, -0.00965, 0.0, 0.0, 0.00062, -0.0, -0.0, -0.00163, -0.0035, -0.0035, -0.00279, 0.0035, 0.0035, -0.00279, -0.00504, -0.01481, -0.00965, 0.00504, 0.01481, -0.00965, -0.0, -0.0, -0.00163, 0.0, 0.0, 0.00062, 0.00133, -0.00935, 0.00051, -0.00133, 0.00935, 0.00051, 0.00163, 0.00504, -0.0, -0.00163, -0.00504, -0.0, 0.0, -0.0, 0.00325, -0.0, 0.0, 0.00166, 0.00158, 0.00158, 0.0, -0.00158, -0.00158, -0.0, 0.00504, 0.00163, 0.0, -0.00504, -0.00163, 0.0, 0.0, 0.0, -0.00166, 0.0, -0.0, -0.00325, 0.00935, -0.00133, -0.00051, -0.00935, 0.00133, -0.00051, 0.01481, 0.00504, 0.00965, -0.01481, -0.00504, 0.00965, -0.0, -0.0, -0.00062, 0.0, 0.0, 0.00163, 0.0035, 0.0035, 0.00279, -0.0035, -0.0035, 0.00279, 0.00504, 0.01481, 0.00965, -0.00504, -0.01481, 0.00965, 0.0, 0.0, 0.00163, -0.0, -0.0, -0.00062, -0.00133, 0.00935, -0.00051, 0.00133, -0.00935, -0.00051, 0.00303, 0.00167, 0.00315, 0.00303, 0.00167, -0.00315, -0.00057, -0.00495, 0.00175, 0.00057, 0.00495, 0.00175, -0.00495, -0.00057, -0.00175, 0.00495, 0.00057, -0.00175, -0.00167, -0.00303, 0.00315, 0.00167, 0.00303, 0.00315, -0.01213, -0.00422, -0.00712, 0.01213, 0.00422, -0.00712, -0.00297, -0.00125, -0.00435, 0.00297, 0.00125, -0.00435, -0.00125, -0.00297, -0.00435, 0.00125, 0.00297, -0.00435, -0.00422, -0.01213, -0.00712, 0.00422, 0.01213, -0.00712, -0.00303, -0.00167, 0.00315, 0.00057, 0.00495, -0.00175, -0.00057, -0.00495, -0.00175, 0.00495, 0.00057, 0.00175, -0.00495, -0.00057, 0.00175, 0.00167, 0.00303, -0.00315, -0.00167, -0.00303, -0.00315, 0.01213, 0.00422, 0.00712, -0.01213, -0.00422, 0.00712, 0.00297, 0.00125, 0.00435, -0.00297, -0.00125, 0.00435, 0.00125, 0.00297, 0.00435, -0.00125, -0.00297, 0.00435, 0.00422, 0.01213, 0.00712, -0.00422, -0.01213, 0.00712, -0.00303, -0.00167, -0.00315, -0.00222, -0.00851, 0.00169, 0.00222, 0.00851, 0.00169, -0.00851, -0.00222, -0.00169, 0.00851, 0.00222, -0.00169, -0.02592, -0.00792, -0.01472, 0.02592, 0.00792, -0.01472, -0.00792, -0.02592, -0.01472, 0.00792, 0.02592, -0.01472, 0.00222, 0.00851, -0.00169, -0.00222, -0.00851, -0.00169, 0.00851, 0.00222, 0.00169, -0.00851, -0.00222, 0.00169, 0.02592, 0.00792, 0.01472, -0.02592, -0.00792, 0.01472, 0.00792, 0.02592, 0.01472, -0.00792, -0.02592, 0.01472]]), "E3g1":array([[-0.00498, 0.01203, -0.00377, 0.00498, -0.01203, -0.00377, -0.0, -0.0, 0.00318, 0.0, 0.0, 0.00318, -0.00341, 0.0053, 0.00182, 0.00341, -0.0053, 0.00182, 0.01218, 0.00096, 0.00366, -0.01218, -0.00096, 0.00366, 0.0, 0.0, -0.00021, 0.0, 0.0, -0.00405, 0.01554, -0.00067, -0.00347, -0.01554, 0.00067, -0.00347, 0.00331, -0.00799, -0.0, -0.00331, 0.00799, -0.0, 0.0, 0.0, -0.00307, 0.0, 0.0, 0.00307, -0.01051, -0.01146, 0.00347, 0.01051, 0.01146, 0.00347, -0.0093, -0.00793, -0.00366, 0.0093, 0.00793, -0.00366, 0.0, -0.0, 0.00405, 0.0, -0.0, 0.00021, -0.00133, 0.00616, -0.00182, 0.00133, -0.00616, -0.00182, 0.00498, -0.01203, 0.00377, -0.00498, 0.01203, 0.00377, 0.0, 0.0, -0.00318, -0.0, -0.0, -0.00318, 0.00341, -0.0053, -0.00182, -0.00341, 0.0053, -0.00182, -0.01218, -0.00096, -0.00366, 0.01218, 0.00096, -0.00366, -0.0, -0.0, 0.00021, -0.0, -0.0, 0.00405, -0.01554, 0.00067, 0.00347, 0.01554, -0.00067, 0.00347, -0.00331, 0.00799, 0.0, 0.00331, -0.00799, 0.0, -0.0, -0.0, 0.00307, -0.0, -0.0, -0.00307, 0.01051, 0.01146, -0.00347, -0.01051, -0.01146, -0.00347, 0.0093, 0.00793, 0.00366, -0.0093, -0.00793, 0.00366, -0.0, 0.0, -0.00405, -0.0, 0.0, -0.00021, 0.00133, -0.00616, 0.00182, -0.00133, 0.00616, 0.00182, 0.00348, -0.00783, 0.00168, 0.00348, -0.00783, -0.00168, -0.00307, 0.008, -0.00168, 0.00307, -0.008, -0.00168, 0.00179, 0.00332, 0.00148, -0.00179, -0.00332, 0.00148, 0.01306, -0.00169, 0.00166, -0.01306, 0.00169, 0.00166, 0.00987, -0.00181, -0.00023, -0.00987, 0.00181, -0.00023, -0.00569, -0.00826, 0.00023, 0.00569, 0.00826, 0.00023, -0.00804, -0.01043, -0.00166, 0.00804, 0.01043, -0.00166, -0.00361, 0.00108, -0.00148, 0.00361, -0.00108, -0.00148, -0.00348, 0.00783, 0.00168, 0.00307, -0.008, 0.00168, -0.00307, 0.008, 0.00168, -0.00179, -0.00332, -0.00148, 0.00179, 0.00332, -0.00148, -0.01306, 0.00169, -0.00166, 0.01306, -0.00169, -0.00166, -0.00987, 0.00181, 0.00023, 0.00987, -0.00181, 0.00023, 0.00569, 0.00826, -0.00023, -0.00569, -0.00826, -0.00023, 0.00804, 0.01043, 0.00166, -0.00804, -0.01043, 0.00166, 0.00361, -0.00108, 0.00148, -0.00361, 0.00108, 0.00148, -0.00348, 0.00783, -0.00168, -0.0078, 0.01884, -0.00775, 0.0078, -0.01884, -0.00775, 0.01952, 0.00112, 0.00674, -0.01952, -0.00112, 0.00674, 0.00506, -0.01221, 0.0, -0.00506, 0.01221, 0.0, -0.0146, -0.01301, -0.00674, 0.0146, 0.01301, -0.00674, 0.0078, -0.01884, 0.00775, -0.0078, 0.01884, 0.00775, -0.01952, -0.00112, -0.00674, 0.01952, 0.00112, -0.00674, -0.00506, 0.01221, -0.0, 0.00506, -0.01221, -0.0, 0.0146, 0.01301, 0.00674, -0.0146, -0.01301, 0.00674], [-0.00478, 0.01154, -0.00477, 0.00478, -0.01154, -0.00477, -0.0, 0.0, 0.00228, -0.0, 0.0, 0.00228, -0.00533, 0.00814, 0.00017, 0.00533, -0.00814, 0.00017, -0.00993, 0.00799, -0.00839, 0.00993, -0.00799, -0.00839, -0.0, 0.0, 1e-05, 0.0, 0.0, -0.00316, 0.00294, -0.00113, -0.00204, -0.00294, 0.00113, -0.00204, 0.00038, -0.00092, 0.0, -0.00038, 0.00092, -0.0, 0.0, 0.0, -0.0008, -0.0, -0.0, 0.0008, -0.00128, -0.00287, 0.00204, 0.00128, 0.00287, 0.00204, 0.00137, 0.01267, 0.00839, -0.00137, -0.01267, 0.00839, 0.0, 0.0, 0.00316, -0.0, -0.0, -1e-05, -0.00198, 0.00952, -0.00017, 0.00198, -0.00952, -0.00017, 0.00478, -0.01154, 0.00477, -0.00478, 0.01154, 0.00477, 0.0, -0.0, -0.00228, 0.0, -0.0, -0.00228, 0.00533, -0.00814, -0.00017, -0.00533, 0.00814, -0.00017, 0.00993, -0.00799, 0.00839, -0.00993, 0.00799, 0.00839, 0.0, -0.0, -1e-05, -0.0, -0.0, 0.00316, -0.00294, 0.00113, 0.00204, 0.00294, -0.00113, 0.00204, -0.00038, 0.00092, -0.0, 0.00038, -0.00092, 0.0, -0.0, -0.0, 0.0008, 0.0, 0.0, -0.0008, 0.00128, 0.00287, -0.00204, -0.00128, -0.00287, -0.00204, -0.00137, -0.01267, -0.00839, 0.00137, 0.01267, -0.00839, -0.0, -0.0, -0.00316, 0.0, 0.0, 1e-05, 0.00198, -0.00952, 0.00017, -0.00198, 0.00952, 0.00017, 0.00206, -0.00649, 0.00156, 0.00206, -0.00649, -0.00156, -0.00314, 0.00604, -0.00156, 0.00314, -0.00604, -0.00156, -0.00383, 0.00525, -0.0034, 0.00383, -0.00525, -0.0034, -0.00547, 0.00529, -0.00794, 0.00547, -0.00529, -0.00794, 0.00963, 0.00466, 0.00436, -0.00963, -0.00466, 0.00436, -0.01011, -0.00351, -0.00436, 0.01011, 0.00351, -0.00436, 0.00013, 0.00761, 0.00794, -0.00013, -0.00761, 0.00794, -0.00101, 0.00642, 0.0034, 0.00101, -0.00642, 0.0034, -0.00206, 0.00649, 0.00156, 0.00314, -0.00604, 0.00156, -0.00314, 0.00604, 0.00156, 0.00383, -0.00525, 0.0034, -0.00383, 0.00525, 0.0034, 0.00547, -0.00529, 0.00794, -0.00547, 0.00529, 0.00794, -0.00963, -0.00466, -0.00436, 0.00963, 0.00466, -0.00436, 0.01011, 0.00351, 0.00436, -0.01011, -0.00351, 0.00436, -0.00013, -0.00761, -0.00794, 0.00013, 0.00761, -0.00794, 0.00101, -0.00642, -0.0034, -0.00101, 0.00642, -0.0034, -0.00206, 0.00649, -0.00156, -0.00848, 0.02047, -0.00899, 0.00848, -0.02047, -0.00899, -0.01723, 0.0144, -0.0127, 0.01723, -0.0144, -0.0127, 0.00058, -0.0014, -0.0, -0.00058, 0.0014, -0.0, 0.002, 0.02237, 0.0127, -0.002, -0.02237, 0.0127, 0.00848, -0.02047, 0.00899, -0.00848, 0.02047, 0.00899, 0.01723, -0.0144, 0.0127, -0.01723, 0.0144, 0.0127, -0.00058, 0.0014, 0.0, 0.00058, -0.0014, 0.0, -0.002, -0.02237, -0.0127, 0.002, 0.02237, -0.0127]]), "E3g2":array([[-0.00176, -0.0113, 0.00305, 0.00176, 0.0113, 0.00305, 0.0, 0.0, -0.00463, 0.0, 0.0, -0.00036, -0.00573, 0.00356, 0.00145, 0.00573, -0.00356, 0.00145, -0.01315, 0.00545, -0.00429, 0.01315, -0.00545, -0.00429, 0.0, -0.0, 0.00304, 0.0, -0.0, 0.00304, -0.00657, 0.00154, 0.00145, 0.00657, -0.00154, 0.00145, 0.00675, 0.00924, 0.00305, -0.00675, -0.00924, 0.00305, -0.0, -0.0, -0.00036, -0.0, -0.0, -0.00463, 0.01179, 0.01071, -0.00335, -0.01179, -0.01071, -0.00335, 0.00782, -0.00324, 0.0, -0.00782, 0.00324, 0.0, 0.0, 0.0, -0.00304, 0.0, 0.0, 0.00304, 0.00077, -0.01591, 0.00335, -0.00077, 0.01591, 0.00335, 0.00176, 0.0113, -0.00305, -0.00176, -0.0113, -0.00305, -0.0, -0.0, 0.00463, -0.0, -0.0, 0.00036, 0.00573, -0.00356, -0.00145, -0.00573, 0.00356, -0.00145, 0.01315, -0.00545, 0.00429, -0.01315, 0.00545, 0.00429, -0.0, 0.0, -0.00304, -0.0, 0.0, -0.00304, 0.00657, -0.00154, -0.00145, -0.00657, 0.00154, -0.00145, -0.00675, -0.00924, -0.00305, 0.00675, 0.00924, -0.00305, 0.0, 0.0, 0.00036, 0.0, 0.0, 0.00463, -0.01179, -0.01071, 0.00335, 0.01179, 0.01071, 0.00335, -0.00782, 0.00324, -0.0, 0.00782, -0.00324, -0.0, -0.0, -0.0, 0.00304, -0.0, -0.0, -0.00304, -0.00077, 0.01591, -0.00335, 0.00077, -0.01591, -0.00335, -0.00126, 0.01319, -0.00141, -0.00126, 0.01319, 0.00141, -0.00375, -0.00138, 0.00135, 0.00375, 0.00138, 0.00135, -0.00815, 0.00302, -0.00194, 0.00815, -0.00302, -0.00194, -0.0079, 0.00363, -0.00194, 0.0079, -0.00363, -0.00194, -0.00168, 0.00362, 0.00135, 0.00168, -0.00362, 0.00135, 0.01022, 0.00843, 0.00141, -0.01022, -0.00843, 0.00141, 0.00859, 0.00661, 0.00029, -0.00859, -0.00661, 0.00029, 0.0014, -0.01075, -0.00029, -0.0014, 0.01075, -0.00029, 0.00126, -0.01319, -0.00141, 0.00375, 0.00138, -0.00135, -0.00375, -0.00138, -0.00135, 0.00815, -0.00302, 0.00194, -0.00815, 0.00302, 0.00194, 0.0079, -0.00363, 0.00194, -0.0079, 0.00363, 0.00194, 0.00168, -0.00362, -0.00135, -0.00168, 0.00362, -0.00135, -0.01022, -0.00843, -0.00141, 0.01022, 0.00843, -0.00141, -0.00859, -0.00661, -0.00029, 0.00859, 0.00661, -0.00029, -0.0014, 0.01075, 0.00029, 0.0014, -0.01075, 0.00029, 0.00126, -0.01319, 0.00141, -0.00263, -0.0179, 0.00581, 0.00263, 0.0179, 0.00581, -0.02076, 0.0086, -0.00856, 0.02076, -0.0086, -0.00856, 0.0108, 0.01451, 0.00581, -0.0108, -0.01451, 0.00581, 0.01167, -0.00483, 0.0, -0.01167, 0.00483, -0.0, 0.00263, 0.0179, -0.00581, -0.00263, -0.0179, -0.00581, 0.02076, -0.0086, 0.00856, -0.02076, 0.0086, 0.00856, -0.0108, -0.01451, -0.00581, 0.0108, 0.01451, -0.00581, -0.01167, 0.00483, -0.0, 0.01167, -0.00483, 0.0], [-0.00552, 0.00893, -0.00619, 0.00552, -0.00893, -0.00619, -0.0, -0.0, -8e-05, 0.0, 0.0, 0.00091, -0.00448, 0.00524, -0.00101, 0.00448, -0.00524, -0.00101, -0.01582, 0.00655, -0.00938, 0.01582, -0.00655, -0.00938, 0.0, 0.0, -0.00011, 0.0, -0.0, -0.00011, -0.00687, -0.00053, -0.00101, 0.00687, 0.00053, -0.00101, -0.01022, -0.00241, -0.00619, 0.01022, 0.00241, -0.00619, -0.0, 0.0, 0.00091, 0.0, 0.0, -8e-05, -0.00355, -0.00309, -0.00031, 0.00355, 0.00309, -0.00031, -0.0025, 0.00103, -0.0, 0.0025, -0.00103, 0.0, 0.0, 0.0, 0.00135, -0.0, 0.0, -0.00135, -0.00032, 0.00469, 0.00031, 0.00032, -0.00469, 0.00031, 0.00552, -0.00893, 0.00619, -0.00552, 0.00893, 0.00619, 0.0, 0.0, 8e-05, -0.0, -0.0, -0.00091, 0.00448, -0.00524, 0.00101, -0.00448, 0.00524, 0.00101, 0.01582, -0.00655, 0.00938, -0.01582, 0.00655, 0.00938, -0.0, -0.0, 0.00011, -0.0, 0.0, 0.00011, 0.00687, 0.00053, 0.00101, -0.00687, -0.00053, 0.00101, 0.01022, 0.00241, 0.00619, -0.01022, -0.00241, 0.00619, 0.0, -0.0, -0.00091, -0.0, -0.0, 8e-05, 0.00355, 0.00309, 0.00031, -0.00355, -0.00309, 0.00031, 0.0025, -0.00103, 0.0, -0.0025, 0.00103, -0.0, -0.0, -0.0, -0.00135, 0.0, -0.0, 0.00135, 0.00032, -0.00469, -0.00031, -0.00032, 0.00469, -0.00031, 0.0052, -0.0091, 0.00556, 0.0052, -0.0091, -0.00556, -0.00207, -0.00021, -0.00014, 0.00207, 0.00021, -0.00014, -0.00838, 0.00404, -0.00599, 0.00838, -0.00404, -0.00599, -0.00878, 0.00307, -0.00599, 0.00878, -0.00307, -0.00599, -0.00132, 0.00162, -0.00014, 0.00132, -0.00162, -0.00014, -0.01011, -0.00276, -0.00556, 0.01011, 0.00276, -0.00556, -0.00126, 0.00546, 0.0033, 0.00126, -0.00546, 0.0033, -0.00475, -0.00297, -0.0033, 0.00475, 0.00297, -0.0033, -0.0052, 0.0091, 0.00556, 0.00207, 0.00021, 0.00014, -0.00207, -0.00021, 0.00014, 0.00838, -0.00404, 0.00599, -0.00838, 0.00404, 0.00599, 0.00878, -0.00307, 0.00599, -0.00878, 0.00307, 0.00599, 0.00132, -0.00162, 0.00014, -0.00132, 0.00162, 0.00014, 0.01011, 0.00276, 0.00556, -0.01011, -0.00276, 0.00556, 0.00126, -0.00546, -0.0033, -0.00126, 0.00546, -0.0033, 0.00475, 0.00297, 0.0033, -0.00475, -0.00297, 0.0033, -0.0052, 0.0091, -0.00556, -0.00913, 0.01678, -0.00943, 0.00913, -0.01678, -0.00943, -0.02642, 0.01094, -0.01489, 0.02642, -0.01094, -0.01489, -0.01832, -0.00541, -0.00943, 0.01832, 0.00541, -0.00943, -0.00481, 0.00199, -0.0, 0.00481, -0.00199, 0.0, 0.00913, -0.01678, 0.00943, -0.00913, 0.01678, 0.00943, 0.02642, -0.01094, 0.01489, -0.02642, 0.01094, 0.01489, 0.01832, 0.00541, 0.00943, -0.01832, -0.00541, 0.00943, 0.00481, -0.00199, 0.0, -0.00481, 0.00199, -0.0]]), "E3g3":array([[-0.00755, -0.00313, -0.0, 0.00755, 0.00313, 0.0, 0.0, -0.0, -0.00307, -0.0, -0.0, 0.00307, -0.01146, 0.01114, 0.00327, 0.01146, -0.01114, 0.00327, -0.00731, 0.00952, -0.00322, 0.00731, -0.00952, -0.00322, 0.0, 0.0, 0.00437, -0.0, -0.0, 0.00018, 0.00597, 0.00177, -0.00126, -0.00597, -0.00177, -0.00126, 0.01259, 0.00522, 0.00439, -0.01259, -0.00522, 0.00439, 0.0, -0.0, -0.0033, 0.0, 0.0, -0.0033, 0.00548, 0.00297, -0.00126, -0.00548, -0.00297, -0.00126, 0.00156, -0.0119, -0.00322, -0.00156, 0.0119, -0.00322, -0.0, -0.0, 0.00018, 0.0, 0.0, 0.00437, -0.00022, -0.01598, 0.00327, 0.00022, 0.01598, 0.00327, 0.00755, 0.00313, 0.0, -0.00755, -0.00313, -0.0, -0.0, 0.0, 0.00307, 0.0, 0.0, -0.00307, 0.01146, -0.01114, -0.00327, -0.01146, 0.01114, -0.00327, 0.00731, -0.00952, 0.00322, -0.00731, 0.00952, 0.00322, -0.0, -0.0, -0.00437, 0.0, 0.0, -0.00018, -0.00597, -0.00177, 0.00126, 0.00597, 0.00177, 0.00126, -0.01259, -0.00522, -0.00439, 0.01259, 0.00522, -0.00439, -0.0, 0.0, 0.0033, -0.0, -0.0, 0.0033, -0.00548, -0.00297, 0.00126, 0.00548, 0.00297, 0.00126, -0.00156, 0.0119, 0.00322, 0.00156, -0.0119, 0.00322, 0.0, 0.0, -0.00018, -0.0, -0.0, -0.00437, 0.00022, 0.01598, -0.00327, -0.00022, -0.01598, -0.00327, 0.00161, 0.0104, -6e-05, 0.00161, 0.0104, 6e-05, -0.0085, 0.00622, -6e-05, 0.0085, -0.00622, -6e-05, -0.01072, 0.00831, -0.00174, 0.01072, -0.00831, -0.00174, 0.00169, 0.0037, -0.00136, -0.00169, -0.0037, -0.00136, 0.00806, 0.00363, 0.00213, -0.00806, -0.00363, 0.00213, 0.00827, 0.00313, 0.00213, -0.00827, -0.00313, 0.00213, 0.00381, -0.00142, -0.00136, -0.00381, 0.00142, -0.00136, -0.0017, -0.01345, -0.00174, 0.0017, 0.01345, -0.00174, -0.00161, -0.0104, -6e-05, 0.0085, -0.00622, 6e-05, -0.0085, 0.00622, 6e-05, 0.01072, -0.00831, 0.00174, -0.01072, 0.00831, 0.00174, -0.00169, -0.0037, 0.00136, 0.00169, 0.0037, 0.00136, -0.00806, -0.00363, -0.00213, 0.00806, 0.00363, -0.00213, -0.00827, -0.00313, -0.00213, 0.00827, 0.00313, -0.00213, -0.00381, 0.00142, 0.00136, 0.00381, -0.00142, 0.00136, 0.0017, 0.01345, 0.00174, -0.0017, -0.01345, 0.00174, -0.00161, -0.0104, 6e-05, -0.01167, -0.00484, -0.0, 0.01167, 0.00484, 0.0, -0.01176, 0.01468, -0.0066, 0.01176, -0.01468, -0.0066, 0.02064, 0.00855, 0.00837, -0.02064, -0.00855, 0.00837, 0.00206, -0.0187, -0.0066, -0.00206, 0.0187, -0.0066, 0.01167, 0.00484, 0.0, -0.01167, -0.00484, -0.0, 0.01176, -0.01468, 0.0066, -0.01176, 0.01468, 0.0066, -0.02064, -0.00855, -0.00837, 0.02064, 0.00855, -0.00837, -0.00206, 0.0187, 0.0066, 0.00206, -0.0187, 0.0066], [-0.00478, -0.00198, -0.0, 0.00478, 0.00198, -0.0, -0.0, 0.0, -0.00121, -0.0, 0.0, 0.00121, -0.00486, 0.00318, 0.0009, 0.00486, -0.00318, 0.0009, -0.00875, 0.00288, -0.00357, 0.00875, -0.00288, -0.00357, 0.0, -0.0, 0.00237, -0.0, -0.0, 0.00194, -0.00673, 0.00158, -0.00168, 0.00673, -0.00158, -0.00168, -0.01694, -0.00702, -0.01272, 0.01694, 0.00702, -0.01272, 0.0, -0.0, -0.002, 0.0, 0.0, -0.002, -0.00364, -0.00588, -0.00168, 0.00364, 0.00588, -0.00168, -0.00415, -0.00823, -0.00357, 0.00415, 0.00823, -0.00357, 0.0, 0.0, 0.00194, -0.0, 0.0, 0.00237, -0.00119, -0.00569, 0.0009, 0.00119, 0.00569, 0.0009, 0.00478, 0.00198, -0.0, -0.00478, -0.00198, 0.0, 0.0, -0.0, 0.00121, 0.0, -0.0, -0.00121, 0.00486, -0.00318, -0.0009, -0.00486, 0.00318, -0.0009, 0.00875, -0.00288, 0.00357, -0.00875, 0.00288, 0.00357, -0.0, 0.0, -0.00237, 0.0, 0.0, -0.00194, 0.00673, -0.00158, 0.00168, -0.00673, 0.00158, 0.00168, 0.01694, 0.00702, 0.01272, -0.01694, -0.00702, 0.01272, -0.0, 0.0, 0.002, -0.0, -0.0, 0.002, 0.00364, 0.00588, 0.00168, -0.00364, -0.00588, 0.00168, 0.00415, 0.00823, 0.00357, -0.00415, -0.00823, 0.00357, -0.0, -0.0, -0.00194, 0.0, -0.0, -0.00237, 0.00119, 0.00569, -0.0009, -0.00119, -0.00569, -0.0009, 0.00355, 0.00037, 0.00226, 0.00355, 0.00037, -0.00226, -0.00277, -0.00225, 0.00226, 0.00277, 0.00225, 0.00226, -0.00884, 0.00301, -0.00344, 0.00884, -0.00301, -0.00344, -0.00022, -0.00212, 0.00131, 0.00022, 0.00212, 0.00131, -0.0095, -0.00399, -0.00853, 0.0095, 0.00399, -0.00853, -0.00954, -0.00389, -0.00853, 0.00954, 0.00389, -0.00853, -0.00166, 0.00134, 0.00131, 0.00166, -0.00134, 0.00131, -0.00412, -0.00838, -0.00344, 0.00412, 0.00838, -0.00344, -0.00355, -0.00037, 0.00226, 0.00277, 0.00225, -0.00226, -0.00277, -0.00225, -0.00226, 0.00884, -0.00301, 0.00344, -0.00884, 0.00301, 0.00344, 0.00022, 0.00212, -0.00131, -0.00022, -0.00212, -0.00131, 0.0095, 0.00399, 0.00853, -0.0095, -0.00399, 0.00853, 0.00954, 0.00389, 0.00853, -0.00954, -0.00389, 0.00853, 0.00166, -0.00134, -0.00131, -0.00166, 0.00134, -0.00131, 0.00412, 0.00838, 0.00344, -0.00412, -0.00838, 0.00344, -0.00355, -0.00037, -0.00226, -0.00657, -0.00272, -0.0, 0.00657, 0.00272, -0.0, -0.0158, 0.00593, -0.00686, 0.0158, -0.00593, -0.00686, -0.02843, -0.01178, -0.01807, 0.02843, 0.01178, -0.01807, -0.00698, -0.01537, -0.00686, 0.00698, 0.01537, -0.00686, 0.00657, 0.00272, 0.0, -0.00657, -0.00272, 0.0, 0.0158, -0.00593, 0.00686, -0.0158, 0.00593, 0.00686, 0.02843, 0.01178, 0.01807, -0.02843, -0.01178, 0.01807, 0.00698, 0.01537, 0.00686, -0.00698, -0.01537, 0.00686]]), "E3g4":array([[0.00878, -0.00589, 0.0026, -0.00878, 0.00589, 0.0026, -0.0, -0.0, 0.00014, 0.0, 0.0, -0.00419, 0.00981, -0.01084, -0.00358, -0.00981, 0.01084, -0.00358, -0.00292, -0.00706, -0.0, 0.00292, 0.00706, -0.0, -0.0, -0.0, -0.00255, 0.0, -0.0, 0.00255, -0.0146, -0.00073, 0.00358, 0.0146, 0.00073, 0.00358, -0.01037, 0.00204, -0.0026, 0.01037, -0.00204, -0.0026, -0.0, -0.0, 0.00419, -0.0, 0.0, -0.00014, 0.00408, 0.00586, -0.00108, -0.00408, -0.00586, -0.00108, 0.00602, 0.01452, 0.00546, -0.00602, -0.01452, 0.00546, 0.0, 0.0, -0.00313, 0.0, 0.0, -0.00313, 0.00126, 0.00702, -0.00108, -0.00126, -0.00702, -0.00108, -0.00878, 0.00589, -0.0026, 0.00878, -0.00589, -0.0026, 0.0, 0.0, -0.00014, -0.0, -0.0, 0.00419, -0.00981, 0.01084, 0.00358, 0.00981, -0.01084, 0.00358, 0.00292, 0.00706, 0.0, -0.00292, -0.00706, 0.0, 0.0, 0.0, 0.00255, -0.0, 0.0, -0.00255, 0.0146, 0.00073, -0.00358, -0.0146, -0.00073, -0.00358, 0.01037, -0.00204, 0.0026, -0.01037, 0.00204, 0.0026, 0.0, 0.0, -0.00419, 0.0, -0.0, 0.00014, -0.00408, -0.00586, 0.00108, 0.00408, 0.00586, 0.00108, -0.00602, -0.01452, -0.00546, 0.00602, 0.01452, -0.00546, -0.0, -0.0, 0.00313, -0.0, -0.0, 0.00313, -0.00126, -0.00702, 0.00108, 0.00126, 0.00702, 0.00108, -0.00376, -0.00139, -0.00141, -0.00376, -0.00139, 0.00141, 0.0076, -0.00944, 0.00108, -0.0076, 0.00944, 0.00108, 0.00588, -0.00778, 8e-05, -0.00588, 0.00778, 8e-05, -0.00966, -0.00134, -8e-05, 0.00966, 0.00134, -8e-05, -0.01205, -0.0013, -0.00108, 0.01205, 0.0013, -0.00108, -0.00167, 0.00364, -0.00141, 0.00167, -0.00364, -0.00141, 0.00348, 0.00895, 0.00245, -0.00348, -0.00895, 0.00245, 0.00387, 0.00879, 0.00245, -0.00387, -0.00879, 0.00245, 0.00376, 0.00139, -0.00141, -0.0076, 0.00944, -0.00108, 0.0076, -0.00944, -0.00108, -0.00588, 0.00778, -8e-05, 0.00588, -0.00778, -8e-05, 0.00966, 0.00134, 8e-05, -0.00966, -0.00134, 8e-05, 0.01205, 0.0013, 0.00108, -0.01205, -0.0013, 0.00108, 0.00167, -0.00364, 0.00141, -0.00167, 0.00364, 0.00141, -0.00348, -0.00895, -0.00245, 0.00348, 0.00895, -0.00245, -0.00387, -0.00879, -0.00245, 0.00387, 0.00879, -0.00245, 0.00376, 0.00139, 0.00141, 0.01363, -0.00974, 0.00541, -0.01363, 0.00974, 0.00541, -0.00456, -0.011, 0.0, 0.00456, 0.011, 0.0, -0.01652, 0.00275, -0.00541, 0.01652, -0.00275, -0.00541, 0.00979, 0.02364, 0.01024, -0.00979, -0.02364, 0.01024, -0.01363, 0.00974, -0.00541, 0.01363, -0.00974, -0.00541, 0.00456, 0.011, -0.0, -0.00456, -0.011, -0.0, 0.01652, -0.00275, 0.00541, -0.01652, 0.00275, 0.00541, -0.00979, -0.02364, -0.01024, 0.00979, 0.02364, -0.01024], [0.00378, -0.01114, 0.0058, -0.00378, 0.01114, 0.0058, -0.0, 0.0, 1e-05, -0.0, -0.0, -0.00062, 0.00462, -0.00492, -0.00068, -0.00462, 0.00492, -0.00068, -0.00155, -0.00373, 0.0, 0.00155, 0.00373, -0.0, -0.0, 0.0, -0.00121, -0.0, -0.0, 0.00121, -0.00675, -0.00022, 0.00068, 0.00675, 0.00022, 0.00068, -0.01055, -0.0052, -0.0058, 0.01055, 0.0052, -0.0058, -0.0, 0.0, 0.00062, -0.0, 0.0, -1e-05, -0.00519, -0.00461, -0.00112, 0.00519, 0.00461, -0.00112, -0.00594, -0.01433, -0.00961, 0.00594, 0.01433, -0.00961, -0.0, -0.0, 1e-05, 0.0, -0.0, 1e-05, 0.00041, -0.00694, -0.00112, -0.00041, 0.00694, -0.00112, -0.00378, 0.01114, -0.0058, 0.00378, -0.01114, -0.0058, 0.0, -0.0, -1e-05, 0.0, 0.0, 0.00062, -0.00462, 0.00492, 0.00068, 0.00462, -0.00492, 0.00068, 0.00155, 0.00373, -0.0, -0.00155, -0.00373, 0.0, 0.0, -0.0, 0.00121, 0.0, 0.0, -0.00121, 0.00675, 0.00022, -0.00068, -0.00675, -0.00022, -0.00068, 0.01055, 0.0052, 0.0058, -0.01055, -0.0052, 0.0058, 0.0, -0.0, -0.00062, 0.0, -0.0, 1e-05, 0.00519, 0.00461, 0.00112, -0.00519, -0.00461, 0.00112, 0.00594, 0.01433, 0.00961, -0.00594, -0.01433, 0.00961, 0.0, 0.0, -1e-05, -0.0, 0.0, -1e-05, -0.00041, 0.00694, 0.00112, 0.00041, -0.00694, 0.00112, 0.00036, 0.0023, -0.00123, 0.00036, 0.0023, 0.00123, 0.00347, -0.01068, 0.00588, -0.00347, 0.01068, 0.00588, -0.00366, -0.00228, -0.00325, 0.00366, 0.00228, -0.00325, 0.00098, -0.0042, 0.00325, -0.00098, 0.0042, 0.00325, -0.01001, -0.0051, -0.00588, 0.01001, 0.0051, -0.00588, -0.00137, -0.00188, -0.00123, 0.00137, 0.00188, -0.00123, -0.00368, -0.00653, -0.00525, 0.00368, 0.00653, -0.00525, -0.00201, -0.00722, -0.00525, 0.00201, 0.00722, -0.00525, -0.00036, -0.0023, -0.00123, -0.00347, 0.01068, -0.00588, 0.00347, -0.01068, -0.00588, 0.00366, 0.00228, 0.00325, -0.00366, -0.00228, 0.00325, -0.00098, 0.0042, -0.00325, 0.00098, -0.0042, -0.00325, 0.01001, 0.0051, 0.00588, -0.01001, -0.0051, 0.00588, 0.00137, 0.00188, 0.00123, -0.00137, -0.00188, 0.00123, 0.00368, 0.00653, 0.00525, -0.00368, -0.00653, 0.00525, 0.00201, 0.00722, 0.00525, -0.00201, -0.00722, 0.00525, -0.00036, -0.0023, 0.00123, 0.00637, -0.01861, 0.00962, -0.00637, 0.01861, 0.00962, -0.00243, -0.00586, 0.0, 0.00243, 0.00586, -0.0, -0.01767, -0.00866, -0.00962, 0.01767, 0.00866, -0.00962, -0.01079, -0.02605, -0.01518, 0.01079, 0.02605, -0.01518, -0.00637, 0.01861, -0.00962, 0.00637, -0.01861, -0.00962, 0.00243, 0.00586, -0.0, -0.00243, -0.00586, 0.0, 0.01767, 0.00866, 0.00962, -0.01767, -0.00866, 0.00962, 0.01079, 0.02605, 0.01518, -0.01079, -0.02605, 0.01518]]), "A1u":array([[0.00305, 0.00126, -0.0, -0.00305, -0.00126, -0.0, -0.0, 0.0, -0.00157, 0.0, 0.0, 0.00157, 0.0062, 0.0062, 0.0, -0.0062, -0.0062, -0.0, 0.00126, 0.00305, 0.0, -0.00126, -0.00305, -0.0, 0.0, -0.0, -0.00157, 0.0, -0.0, 0.00157, -0.0, 0.00877, 0.0, -0.0, -0.00877, -0.0, -0.00126, 0.00305, 0.0, 0.00126, -0.00305, 0.0, 0.0, -0.0, -0.00157, -0.0, 0.0, 0.00157, -0.0062, 0.0062, 0.0, 0.0062, -0.0062, -0.0, -0.00305, 0.00126, 0.0, 0.00305, -0.00126, -0.0, 0.0, 0.0, -0.00157, 0.0, 0.0, 0.00157, -0.00877, -0.0, 0.0, 0.00877, -0.0, -0.0, 0.00305, 0.00126, 0.0, -0.00305, -0.00126, 0.0, 0.0, -0.0, -0.00157, -0.0, -0.0, 0.00157, 0.0062, 0.0062, -0.0, -0.0062, -0.0062, 0.0, 0.00126, 0.00305, -0.0, -0.00126, -0.00305, 0.0, -0.0, 0.0, -0.00157, -0.0, 0.0, 0.00157, 0.0, 0.00877, -0.0, 0.0, -0.00877, 0.0, -0.00126, 0.00305, -0.0, 0.00126, -0.00305, -0.0, -0.0, 0.0, -0.00157, 0.0, -0.0, 0.00157, -0.0062, 0.0062, -0.0, 0.0062, -0.0062, 0.0, -0.00305, 0.00126, -0.0, 0.00305, -0.00126, 0.0, -0.0, -0.0, -0.00157, -0.0, -0.0, 0.00157, -0.00877, 0.0, -0.0, 0.00877, 0.0, 0.0, -0.00439, 0.01466, -0.00793, 0.00439, -0.01466, -0.00793, 0.00727, -0.01347, 0.00793, -0.00727, 0.01347, 0.00793, -0.01347, 0.00727, -0.00793, 0.01347, -0.00727, -0.00793, 0.01466, -0.00439, 0.00793, -0.01466, 0.00439, 0.00793, -0.01466, -0.00439, -0.00793, 0.01466, 0.00439, -0.00793, 0.01347, 0.00727, 0.00793, -0.01347, -0.00727, 0.00793, -0.00727, -0.01347, -0.00793, 0.00727, 0.01347, -0.00793, 0.00439, 0.01466, 0.00793, -0.00439, -0.01466, 0.00793, 0.00439, -0.01466, -0.00793, 0.00727, -0.01347, 0.00793, -0.00727, 0.01347, 0.00793, -0.01347, 0.00727, -0.00793, 0.01347, -0.00727, -0.00793, 0.01466, -0.00439, 0.00793, -0.01466, 0.00439, 0.00793, -0.01466, -0.00439, -0.00793, 0.01466, 0.00439, -0.00793, 0.01347, 0.00727, 0.00793, -0.01347, -0.00727, 0.00793, -0.00727, -0.01347, -0.00793, 0.00727, 0.01347, -0.00793, 0.00439, 0.01466, 0.00793, -0.00439, -0.01466, 0.00793, -0.00439, 0.01466, -0.00793, 0.00618, 0.00256, 0.0, -0.00618, -0.00256, 0.0, 0.00256, 0.00618, 0.0, -0.00256, -0.00618, 0.0, -0.00256, 0.00618, 0.0, 0.00256, -0.00618, 0.0, -0.00618, 0.00256, 0.0, 0.00618, -0.00256, 0.0, 0.00618, 0.00256, -0.0, -0.00618, -0.00256, -0.0, 0.00256, 0.00618, -0.0, -0.00256, -0.00618, -0.0, -0.00256, 0.00618, -0.0, 0.00256, -0.00618, -0.0, -0.00618, 0.00256, -0.0, 0.00618, -0.00256, -0.0], [-0.00798, -0.0033, 0.0, 0.00798, 0.0033, 0.0, 0.0, -0.0, 0.00611, 0.0, -0.0, -0.00611, -0.00677, -0.00677, -0.0, 0.00677, 0.00677, -0.0, -0.0033, -0.00798, -0.0, 0.0033, 0.00798, 0.0, 0.0, 0.0, 0.00611, 0.0, -0.0, -0.00611, -0.0, -0.00958, 0.0, -0.0, 0.00958, 0.0, 0.0033, -0.00798, -0.0, -0.0033, 0.00798, 0.0, 0.0, 0.0, 0.00611, 0.0, -0.0, -0.00611, 0.00677, -0.00677, -0.0, -0.00677, 0.00677, 0.0, 0.00798, -0.0033, -0.0, -0.00798, 0.0033, 0.0, -0.0, 0.0, 0.00611, 0.0, 0.0, -0.00611, 0.00958, -0.0, 0.0, -0.00958, -0.0, 0.0, -0.00798, -0.0033, 0.0, 0.00798, 0.0033, -0.0, -0.0, 0.0, 0.00611, -0.0, 0.0, -0.00611, -0.00677, -0.00677, 0.0, 0.00677, 0.00677, 0.0, -0.0033, -0.00798, 0.0, 0.0033, 0.00798, -0.0, -0.0, -0.0, 0.00611, -0.0, 0.0, -0.00611, 0.0, -0.00958, -0.0, 0.0, 0.00958, -0.0, 0.0033, -0.00798, 0.0, -0.0033, 0.00798, -0.0, -0.0, -0.0, 0.00611, -0.0, 0.0, -0.00611, 0.00677, -0.00677, -0.0, -0.00677, 0.00677, -0.0, 0.00798, -0.0033, 0.0, -0.00798, 0.0033, -0.0, 0.0, -0.0, 0.00611, -0.0, -0.0, -0.00611, 0.00958, 0.0, -0.0, -0.00958, 0.0, -0.0, -0.01295, 0.00299, -0.00216, 0.01295, -0.00299, -0.00216, -0.00704, -0.01127, 0.00216, 0.00704, 0.01127, 0.00216, -0.01127, -0.00704, -0.00216, 0.01127, 0.00704, -0.00216, 0.00299, -0.01295, 0.00216, -0.00299, 0.01295, 0.00216, -0.00299, -0.01295, -0.00216, 0.00299, 0.01295, -0.00216, 0.01127, -0.00704, 0.00216, -0.01127, 0.00704, 0.00216, 0.00704, -0.01127, -0.00216, -0.00704, 0.01127, -0.00216, 0.01295, 0.00299, 0.00216, -0.01295, -0.00299, 0.00216, 0.01295, -0.00299, -0.00216, -0.00704, -0.01127, 0.00216, 0.00704, 0.01127, 0.00216, -0.01127, -0.00704, -0.00216, 0.01127, 0.00704, -0.00216, 0.00299, -0.01295, 0.00216, -0.00299, 0.01295, 0.00216, -0.00299, -0.01295, -0.00216, 0.00299, 0.01295, -0.00216, 0.01127, -0.00704, 0.00216, -0.01127, 0.00704, 0.00216, 0.00704, -0.01127, -0.00216, -0.00704, 0.01127, -0.00216, 0.01295, 0.00299, 0.00216, -0.01295, -0.00299, 0.00216, -0.01295, 0.00299, -0.00216, -0.01566, -0.00649, 0.0, 0.01566, 0.00649, -0.0, -0.00649, -0.01566, 0.0, 0.00649, 0.01566, -0.0, 0.00649, -0.01566, 0.0, -0.00649, 0.01566, 0.0, 0.01566, -0.00649, 0.0, -0.01566, 0.00649, 0.0, -0.01566, -0.00649, -0.0, 0.01566, 0.00649, 0.0, -0.00649, -0.01566, -0.0, 0.00649, 0.01566, 0.0, 0.00649, -0.01566, -0.0, -0.00649, 0.01566, -0.0, 0.01566, -0.00649, -0.0, -0.01566, 0.00649, -0.0]]), "A2u":array([[0.00431, -0.01041, 0.00341, -0.00431, 0.01041, 0.00341, -0.0, -0.0, -0.00485, 0.0, 0.0, -0.00485, 0.00591, -0.00591, -0.00429, -0.00591, 0.00591, -0.00429, 0.01041, -0.00431, 0.00341, -0.01041, 0.00431, 0.00341, 0.0, -0.0, -0.00485, 0.0, 0.0, -0.00485, 0.00835, 0.0, -0.00429, -0.00835, -0.0, -0.00429, 0.01041, 0.00431, 0.00341, -0.01041, -0.00431, 0.00341, 0.0, -0.0, -0.00485, 0.0, 0.0, -0.00485, 0.00591, 0.00591, -0.00429, -0.00591, -0.00591, -0.00429, 0.00431, 0.01041, 0.00341, -0.00431, -0.01041, 0.00341, 0.0, 0.0, -0.00485, -0.0, 0.0, -0.00485, 0.0, 0.00835, -0.00429, 0.0, -0.00835, -0.00429, 0.00431, -0.01041, 0.00341, -0.00431, 0.01041, 0.00341, 0.0, 0.0, -0.00485, -0.0, -0.0, -0.00485, 0.00591, -0.00591, -0.00429, -0.00591, 0.00591, -0.00429, 0.01041, -0.00431, 0.00341, -0.01041, 0.00431, 0.00341, -0.0, 0.0, -0.00485, -0.0, -0.0, -0.00485, 0.00835, 0.0, -0.00429, -0.00835, 0.0, -0.00429, 0.01041, 0.00431, 0.00341, -0.01041, -0.00431, 0.00341, -0.0, 0.0, -0.00485, -0.0, -0.0, -0.00485, 0.00591, 0.00591, -0.00429, -0.00591, -0.00591, -0.00429, 0.00431, 0.01041, 0.00341, -0.00431, -0.01041, 0.00341, -0.0, -0.0, -0.00485, 0.0, -0.0, -0.00485, -0.0, 0.00835, -0.00429, -0.0, -0.00835, -0.00429, 0.00018, -0.00471, -0.00091, -0.00018, 0.00471, -0.00091, 0.00321, -0.00346, -0.00091, -0.00321, 0.00346, -0.00091, 0.00346, -0.00321, -0.00091, -0.00346, 0.00321, -0.00091, 0.00471, -0.00018, -0.00091, -0.00471, 0.00018, -0.00091, 0.00471, 0.00018, -0.00091, -0.00471, -0.00018, -0.00091, 0.00346, 0.00321, -0.00091, -0.00346, -0.00321, -0.00091, 0.00321, 0.00346, -0.00091, -0.00321, -0.00346, -0.00091, 0.00018, 0.00471, -0.00091, -0.00018, -0.00471, -0.00091, -0.00018, 0.00471, -0.00091, 0.00321, -0.00346, -0.00091, -0.00321, 0.00346, -0.00091, 0.00346, -0.00321, -0.00091, -0.00346, 0.00321, -0.00091, 0.00471, -0.00018, -0.00091, -0.00471, 0.00018, -0.00091, 0.00471, 0.00018, -0.00091, -0.00471, -0.00018, -0.00091, 0.00346, 0.00321, -0.00091, -0.00346, -0.00321, -0.00091, 0.00321, 0.00346, -0.00091, -0.00321, -0.00346, -0.00091, 0.00018, 0.00471, -0.00091, -0.00018, -0.00471, -0.00091, 0.00018, -0.00471, -0.00091, 0.00837, -0.0202, 0.00755, -0.00837, 0.0202, 0.00755, 0.0202, -0.00837, 0.00755, -0.0202, 0.00837, 0.00755, 0.0202, 0.00837, 0.00755, -0.0202, -0.00837, 0.00755, 0.00837, 0.0202, 0.00755, -0.00837, -0.0202, 0.00755, 0.00837, -0.0202, 0.00755, -0.00837, 0.0202, 0.00755, 0.0202, -0.00837, 0.00755, -0.0202, 0.00837, 0.00755, 0.0202, 0.00837, 0.00755, -0.0202, -0.00837, 0.00755, 0.00837, 0.0202, 0.00755, -0.00837, -0.0202, 0.00755], [0.00251, -0.00606, -0.00366, -0.00251, 0.00606, -0.00366, -0.0, -0.0, -0.00389, 0.0, 0.0, -0.00389, 0.01697, -0.01697, -0.00237, -0.01697, 0.01697, -0.00237, 0.00606, -0.00251, -0.00366, -0.00606, 0.00251, -0.00366, -0.0, -0.0, -0.00389, 0.0, 0.0, -0.00389, 0.024, 0.0, -0.00237, -0.024, -0.0, -0.00237, 0.00606, 0.00251, -0.00366, -0.00606, -0.00251, -0.00366, 0.0, -0.0, -0.00389, -0.0, 0.0, -0.00389, 0.01697, 0.01697, -0.00237, -0.01697, -0.01697, -0.00237, 0.00251, 0.00606, -0.00366, -0.00251, -0.00606, -0.00366, 0.0, -0.0, -0.00389, -0.0, 0.0, -0.00389, -0.0, 0.024, -0.00237, -0.0, -0.024, -0.00237, 0.00251, -0.00606, -0.00366, -0.00251, 0.00606, -0.00366, 0.0, 0.0, -0.00389, -0.0, -0.0, -0.00389, 0.01697, -0.01697, -0.00237, -0.01697, 0.01697, -0.00237, 0.00606, -0.00251, -0.00366, -0.00606, 0.00251, -0.00366, 0.0, 0.0, -0.00389, -0.0, -0.0, -0.00389, 0.024, 0.0, -0.00237, -0.024, -0.0, -0.00237, 0.00606, 0.00251, -0.00366, -0.00606, -0.00251, -0.00366, -0.0, 0.0, -0.00389, 0.0, -0.0, -0.00389, 0.01697, 0.01697, -0.00237, -0.01697, -0.01697, -0.00237, 0.00251, 0.00606, -0.00366, -0.00251, -0.00606, -0.00366, -0.0, 0.0, -0.00389, 0.0, -0.0, -0.00389, -0.0, 0.024, -0.00237, 0.0, -0.024, -0.00237, -0.00362, -0.00172, 0.00404, 0.00362, 0.00172, 0.00404, 0.00378, 0.00134, 0.00404, -0.00378, -0.00134, 0.00404, -0.00134, -0.00378, 0.00404, 0.00134, 0.00378, 0.00404, 0.00172, 0.00362, 0.00404, -0.00172, -0.00362, 0.00404, 0.00172, -0.00362, 0.00404, -0.00172, 0.00362, 0.00404, -0.00134, 0.00378, 0.00404, 0.00134, -0.00378, 0.00404, 0.00378, -0.00134, 0.00404, -0.00378, 0.00134, 0.00404, -0.00362, 0.00172, 0.00404, 0.00362, -0.00172, 0.00404, 0.00362, 0.00172, 0.00404, 0.00378, 0.00134, 0.00404, -0.00378, -0.00134, 0.00404, -0.00134, -0.00378, 0.00404, 0.00134, 0.00378, 0.00404, 0.00172, 0.00362, 0.00404, -0.00172, -0.00362, 0.00404, 0.00172, -0.00362, 0.00404, -0.00172, 0.00362, 0.00404, -0.00134, 0.00378, 0.00404, 0.00134, -0.00378, 0.00404, 0.00378, -0.00134, 0.00404, -0.00378, 0.00134, 0.00404, -0.00362, 0.00172, 0.00404, 0.00362, -0.00172, 0.00404, -0.00362, -0.00172, 0.00404, -0.00547, 0.01321, 0.00184, 0.00547, -0.01321, 0.00184, -0.01321, 0.00547, 0.00184, 0.01321, -0.00547, 0.00184, -0.01321, -0.00547, 0.00184, 0.01321, 0.00547, 0.00184, -0.00547, -0.01321, 0.00184, 0.00547, 0.01321, 0.00184, -0.00547, 0.01321, 0.00184, 0.00547, -0.01321, 0.00184, -0.01321, 0.00547, 0.00184, 0.01321, -0.00547, 0.00184, -0.01321, -0.00547, 0.00184, 0.01321, 0.00547, 0.00184, -0.00547, -0.01321, 0.00184, 0.00547, 0.01321, 0.00184]]), "B1u":array([[0.00433, -0.01046, 0.00739, -0.00433, 0.01046, 0.00739, 0.0, 0.0, 0.00166, -0.0, 0.0, 0.00166, 0.00121, 0.00121, 0.0, -0.00121, -0.00121, -0.0, -0.01046, 0.00433, -0.00739, 0.01046, -0.00433, -0.00739, -0.0, 0.0, -0.00166, -0.0, -0.0, -0.00166, 0.0, -0.00171, 0.0, -0.0, 0.00171, -0.0, 0.01046, 0.00433, 0.00739, -0.01046, -0.00433, 0.00739, -0.0, 0.0, 0.00166, -0.0, -0.0, 0.00166, -0.00121, 0.00121, 0.0, 0.00121, -0.00121, 0.0, -0.00433, -0.01046, -0.00739, 0.00433, 0.01046, -0.00739, -0.0, -0.0, -0.00166, 0.0, -0.0, -0.00166, 0.00171, -0.0, 0.0, -0.00171, -0.0, 0.0, 0.00433, -0.01046, 0.00739, -0.00433, 0.01046, 0.00739, -0.0, -0.0, 0.00166, 0.0, -0.0, 0.00166, 0.00121, 0.00121, -0.0, -0.00121, -0.00121, 0.0, -0.01046, 0.00433, -0.00739, 0.01046, -0.00433, -0.00739, 0.0, -0.0, -0.00166, 0.0, 0.0, -0.00166, -0.0, -0.00171, -0.0, -0.0, 0.00171, 0.0, 0.01046, 0.00433, 0.00739, -0.01046, -0.00433, 0.00739, 0.0, -0.0, 0.00166, 0.0, 0.0, 0.00166, -0.00121, 0.00121, -0.0, 0.00121, -0.00121, -0.0, -0.00433, -0.01046, -0.00739, 0.00433, 0.01046, -0.00739, 0.0, 0.0, -0.00166, -0.0, 0.0, -0.00166, 0.00171, 0.0, -0.0, -0.00171, 0.0, -0.0, 0.00341, -0.0062, 0.00589, -0.00341, 0.0062, 0.00589, 0.00197, -0.00679, 0.00589, -0.00197, 0.00679, 0.00589, -0.00679, 0.00197, -0.00589, 0.00679, -0.00197, -0.00589, -0.0062, 0.00341, -0.00589, 0.0062, -0.00341, -0.00589, 0.0062, 0.00341, 0.00589, -0.0062, -0.00341, 0.00589, 0.00679, 0.00197, 0.00589, -0.00679, -0.00197, 0.00589, -0.00197, -0.00679, -0.00589, 0.00197, 0.00679, -0.00589, -0.00341, -0.0062, -0.00589, 0.00341, 0.0062, -0.00589, -0.00341, 0.0062, 0.00589, 0.00197, -0.00679, 0.00589, -0.00197, 0.00679, 0.00589, -0.00679, 0.00197, -0.00589, 0.00679, -0.00197, -0.00589, -0.0062, 0.00341, -0.00589, 0.0062, -0.00341, -0.00589, 0.0062, 0.00341, 0.00589, -0.0062, -0.00341, 0.00589, 0.00679, 0.00197, 0.00589, -0.00679, -0.00197, 0.00589, -0.00197, -0.00679, -0.00589, 0.00197, 0.00679, -0.00589, -0.00341, -0.0062, -0.00589, 0.00341, 0.0062, -0.00589, 0.00341, -0.0062, 0.00589, 0.00653, -0.01576, 0.01046, -0.00653, 0.01576, 0.01046, -0.01576, 0.00653, -0.01046, 0.01576, -0.00653, -0.01046, 0.01576, 0.00653, 0.01046, -0.01576, -0.00653, 0.01046, -0.00653, -0.01576, -0.01046, 0.00653, 0.01576, -0.01046, 0.00653, -0.01576, 0.01046, -0.00653, 0.01576, 0.01046, -0.01576, 0.00653, -0.01046, 0.01576, -0.00653, -0.01046, 0.01576, 0.00653, 0.01046, -0.01576, -0.00653, 0.01046, -0.00653, -0.01576, -0.01046, 0.00653, 0.01576, -0.01046], [0.00075, -0.0018, 0.00054, -0.00075, 0.0018, 0.00054, 0.0, -0.0, 0.00379, -0.0, -0.0, 0.00379, -0.00311, -0.00311, -0.0, 0.00311, 0.00311, 0.0, -0.0018, 0.00075, -0.00054, 0.0018, -0.00075, -0.00054, 0.0, 0.0, -0.00379, 0.0, -0.0, -0.00379, 0.0, 0.00439, -0.0, 0.0, -0.00439, 0.0, 0.0018, 0.00075, 0.00054, -0.0018, -0.00075, 0.00054, 0.0, 0.0, 0.00379, 0.0, -0.0, 0.00379, 0.00311, -0.00311, -0.0, -0.00311, 0.00311, 0.0, -0.00075, -0.0018, -0.00054, 0.00075, 0.0018, -0.00054, -0.0, 0.0, -0.00379, 0.0, 0.0, -0.00379, -0.00439, 0.0, -0.0, 0.00439, 0.0, -0.0, 0.00075, -0.0018, 0.00054, -0.00075, 0.0018, 0.00054, -0.0, 0.0, 0.00379, 0.0, 0.0, 0.00379, -0.00311, -0.00311, 0.0, 0.00311, 0.00311, -0.0, -0.0018, 0.00075, -0.00054, 0.0018, -0.00075, -0.00054, -0.0, -0.0, -0.00379, -0.0, 0.0, -0.00379, -0.0, 0.00439, 0.0, -0.0, -0.00439, -0.0, 0.0018, 0.00075, 0.00054, -0.0018, -0.00075, 0.00054, -0.0, -0.0, 0.00379, -0.0, 0.0, 0.00379, 0.00311, -0.00311, 0.0, -0.00311, 0.00311, -0.0, -0.00075, -0.0018, -0.00054, 0.00075, 0.0018, -0.00054, 0.0, -0.0, -0.00379, -0.0, -0.0, -0.00379, -0.00439, -0.0, 0.0, 0.00439, -0.0, 0.0, 0.00857, -0.00927, 0.00852, -0.00857, 0.00927, 0.00852, 0.0005, -0.01262, 0.00852, -0.0005, 0.01262, 0.00852, -0.01262, 0.0005, -0.00852, 0.01262, -0.0005, -0.00852, -0.00927, 0.00857, -0.00852, 0.00927, -0.00857, -0.00852, 0.00927, 0.00857, 0.00852, -0.00927, -0.00857, 0.00852, 0.01262, 0.0005, 0.00852, -0.01262, -0.0005, 0.00852, -0.0005, -0.01262, -0.00852, 0.0005, 0.01262, -0.00852, -0.00857, -0.00927, -0.00852, 0.00857, 0.00927, -0.00852, -0.00857, 0.00927, 0.00852, 0.0005, -0.01262, 0.00852, -0.0005, 0.01262, 0.00852, -0.01262, 0.0005, -0.00852, 0.01262, -0.0005, -0.00852, -0.00927, 0.00857, -0.00852, 0.00927, -0.00857, -0.00852, 0.00927, 0.00857, 0.00852, -0.00927, -0.00857, 0.00852, 0.01262, 0.0005, 0.00852, -0.01262, -0.0005, 0.00852, -0.0005, -0.01262, -0.00852, 0.0005, 0.01262, -0.00852, -0.00857, -0.00927, -0.00852, 0.00857, 0.00927, -0.00852, 0.00857, -0.00927, 0.00852, -0.00547, 0.0132, -0.00524, 0.00547, -0.0132, -0.00524, 0.0132, -0.00547, 0.00524, -0.0132, 0.00547, 0.00524, -0.0132, -0.00547, -0.00524, 0.0132, 0.00547, -0.00524, 0.00547, 0.0132, 0.00524, -0.00547, -0.0132, 0.00524, -0.00547, 0.0132, -0.00524, 0.00547, -0.0132, -0.00524, 0.0132, -0.00547, 0.00524, -0.0132, 0.00547, 0.00524, -0.0132, -0.00547, -0.00524, 0.0132, 0.00547, -0.00524, 0.00547, 0.0132, 0.00524, -0.00547, -0.0132, 0.00524]]), "B2u":array([[-0.00771, -0.00319, 0.0, 0.00771, 0.00319, 0.0, 0.0, 0.0, -0.00387, -0.0, 0.0, 0.00387, -0.01334, 0.01334, 0.00189, 0.01334, -0.01334, 0.00189, 0.00319, 0.00771, -0.0, -0.00319, -0.00771, 0.0, -0.0, 0.0, 0.00387, -0.0, -0.0, -0.00387, 0.01886, -0.0, -0.00189, -0.01886, -0.0, -0.00189, 0.00319, -0.00771, 0.0, -0.00319, 0.00771, 0.0, -0.0, 0.0, -0.00387, -0.0, -0.0, 0.00387, -0.01334, -0.01334, 0.00189, 0.01334, 0.01334, 0.00189, -0.00771, 0.00319, -0.0, 0.00771, -0.00319, -0.0, -0.0, -0.0, 0.00387, -0.0, -0.0, -0.00387, 0.0, 0.01886, -0.00189, -0.0, -0.01886, -0.00189, -0.00771, -0.00319, -0.0, 0.00771, 0.00319, -0.0, -0.0, -0.0, -0.00387, 0.0, -0.0, 0.00387, -0.01334, 0.01334, 0.00189, 0.01334, -0.01334, 0.00189, 0.00319, 0.00771, 0.0, -0.00319, -0.00771, -0.0, 0.0, -0.0, 0.00387, 0.0, 0.0, -0.00387, 0.01886, 0.0, -0.00189, -0.01886, -0.0, -0.00189, 0.00319, -0.00771, -0.0, -0.00319, 0.00771, -0.0, 0.0, -0.0, -0.00387, 0.0, 0.0, 0.00387, -0.01334, -0.01334, 0.00189, 0.01334, 0.01334, 0.00189, -0.00771, 0.00319, 0.0, 0.00771, -0.00319, 0.0, 0.0, 0.0, 0.00387, 0.0, 0.0, -0.00387, 0.0, 0.01886, -0.00189, -0.0, -0.01886, -0.00189, 0.00026, -0.01271, 0.00106, -0.00026, 0.01271, 0.00106, -0.0088, 0.00917, -0.00106, 0.0088, -0.00917, -0.00106, -0.00917, 0.0088, -0.00106, 0.00917, -0.0088, -0.00106, 0.01271, -0.00026, 0.00106, -0.01271, 0.00026, 0.00106, 0.01271, 0.00026, 0.00106, -0.01271, -0.00026, 0.00106, -0.00917, -0.0088, -0.00106, 0.00917, 0.0088, -0.00106, -0.0088, -0.00917, -0.00106, 0.0088, 0.00917, -0.00106, 0.00026, 0.01271, 0.00106, -0.00026, -0.01271, 0.00106, -0.00026, 0.01271, 0.00106, -0.0088, 0.00917, -0.00106, 0.0088, -0.00917, -0.00106, -0.00917, 0.0088, -0.00106, 0.00917, -0.0088, -0.00106, 0.01271, -0.00026, 0.00106, -0.01271, 0.00026, 0.00106, 0.01271, 0.00026, 0.00106, -0.01271, -0.00026, 0.00106, -0.00917, -0.0088, -0.00106, 0.00917, 0.0088, -0.00106, -0.0088, -0.00917, -0.00106, 0.0088, 0.00917, -0.00106, 0.00026, 0.01271, 0.00106, -0.00026, -0.01271, 0.00106, 0.00026, -0.01271, 0.00106, -0.01163, -0.00482, 0.0, 0.01163, 0.00482, -0.0, 0.00482, 0.01163, 0.0, -0.00482, -0.01163, -0.0, 0.00482, -0.01163, 0.0, -0.00482, 0.01163, -0.0, -0.01163, 0.00482, 0.0, 0.01163, -0.00482, -0.0, -0.01163, -0.00482, -0.0, 0.01163, 0.00482, 0.0, 0.00482, 0.01163, -0.0, -0.00482, -0.01163, 0.0, 0.00482, -0.01163, -0.0, -0.00482, 0.01163, 0.0, -0.01163, 0.00482, -0.0, 0.01163, -0.00482, 0.0], [0.00372, 0.00154, 0.0, -0.00372, -0.00154, 0.0, -0.0, -0.0, -0.00977, 0.0, -0.0, 0.00977, 0.00056, -0.00056, 0.03789, -0.00056, 0.00056, 0.03789, -0.00154, -0.00372, 0.0, 0.00154, 0.00372, -0.0, -0.0, -0.0, 0.00977, 0.0, 0.0, -0.00977, -0.00079, 0.0, -0.03789, 0.00079, -0.0, -0.03789, -0.00154, 0.00372, -0.0, 0.00154, -0.00372, -0.0, 0.0, -0.0, -0.00977, -0.0, 0.0, 0.00977, 0.00056, 0.00056, 0.03789, -0.00056, -0.00056, 0.03789, 0.00372, -0.00154, -0.0, -0.00372, 0.00154, -0.0, 0.0, -0.0, 0.00977, -0.0, 0.0, -0.00977, -0.0, -0.00079, -0.03789, -0.0, 0.00079, -0.03789, 0.00372, 0.00154, -0.0, -0.00372, -0.00154, -0.0, 0.0, 0.0, -0.00977, -0.0, 0.0, 0.00977, 0.00056, -0.00056, 0.03789, -0.00056, 0.00056, 0.03789, -0.00154, -0.00372, -0.0, 0.00154, 0.00372, 0.0, 0.0, 0.0, 0.00977, -0.0, -0.0, -0.00977, -0.00079, -0.0, -0.03789, 0.00079, -0.0, -0.03789, -0.00154, 0.00372, 0.0, 0.00154, -0.00372, 0.0, -0.0, 0.0, -0.00977, 0.0, -0.0, 0.00977, 0.00056, 0.00056, 0.03789, -0.00056, -0.00056, 0.03789, 0.00372, -0.00154, 0.0, -0.00372, 0.00154, 0.0, -0.0, 0.0, 0.00977, 0.0, -0.0, -0.00977, 0.0, -0.00079, -0.03789, 0.0, 0.00079, -0.03789, 0.00066, -0.00101, -0.00782, -0.00066, 0.00101, -0.00782, -0.00025, 0.00118, 0.00782, 0.00025, -0.00118, 0.00782, -0.00118, 0.00025, 0.00782, 0.00118, -0.00025, 0.00782, 0.00101, -0.00066, -0.00782, -0.00101, 0.00066, -0.00782, 0.00101, 0.00066, -0.00782, -0.00101, -0.00066, -0.00782, -0.00118, -0.00025, 0.00782, 0.00118, 0.00025, 0.00782, -0.00025, -0.00118, 0.00782, 0.00025, 0.00118, 0.00782, 0.00066, 0.00101, -0.00782, -0.00066, -0.00101, -0.00782, -0.00066, 0.00101, -0.00782, -0.00025, 0.00118, 0.00782, 0.00025, -0.00118, 0.00782, -0.00118, 0.00025, 0.00782, 0.00118, -0.00025, 0.00782, 0.00101, -0.00066, -0.00782, -0.00101, 0.00066, -0.00782, 0.00101, 0.00066, -0.00782, -0.00101, -0.00066, -0.00782, -0.00118, -0.00025, 0.00782, 0.00118, 0.00025, 0.00782, -0.00025, -0.00118, 0.00782, 0.00025, 0.00118, 0.00782, 0.00066, 0.00101, -0.00782, -0.00066, -0.00101, -0.00782, 0.00066, -0.00101, -0.00782, 0.00517, 0.00214, 0.0, -0.00517, -0.00214, 0.0, -0.00214, -0.00517, 0.0, 0.00214, 0.00517, -0.0, -0.00214, 0.00517, 0.0, 0.00214, -0.00517, 0.0, 0.00517, -0.00214, 0.0, -0.00517, 0.00214, -0.0, 0.00517, 0.00214, -0.0, -0.00517, -0.00214, -0.0, -0.00214, -0.00517, -0.0, 0.00214, 0.00517, 0.0, -0.00214, 0.00517, -0.0, 0.00214, -0.00517, -0.0, 0.00517, -0.00214, -0.0, -0.00517, 0.00214, 0.0]]), "E2u1":array([[0.00861, -0.0059, -0.00269, -0.00861, 0.0059, -0.00269, 0.0, 0.0, -0.00532, -0.0, 0.0, -0.00958, 0.00933, -0.00933, -0.00947, -0.00933, 0.00933, -0.00947, 0.0059, -0.00861, -0.00269, -0.0059, 0.00861, -0.00269, -0.0, 0.0, -0.00958, -0.0, -0.0, -0.00532, 0.0, -0.00501, 0.0, 0.0, 0.00501, 0.0, -0.0059, -0.00861, 0.00269, 0.0059, 0.00861, 0.00269, -0.0, 0.0, 0.00532, -0.0, -0.0, 0.00958, -0.00933, -0.00933, 0.00947, 0.00933, 0.00933, 0.00947, -0.00861, -0.0059, 0.00269, 0.00861, 0.0059, 0.00269, -0.0, -0.0, 0.00958, 0.0, -0.0, 0.00532, -0.00501, 0.0, -0.0, 0.00501, 0.0, -0.0, 0.00861, -0.0059, -0.00269, -0.00861, 0.0059, -0.00269, -0.0, -0.0, -0.00532, 0.0, -0.0, -0.00958, 0.00933, -0.00933, -0.00947, -0.00933, 0.00933, -0.00947, 0.0059, -0.00861, -0.00269, -0.0059, 0.00861, -0.00269, 0.0, -0.0, -0.00958, 0.0, 0.0, -0.00532, -0.0, -0.00501, -0.0, -0.0, 0.00501, -0.0, -0.0059, -0.00861, 0.00269, 0.0059, 0.00861, 0.00269, 0.0, -0.0, 0.00532, 0.0, 0.0, 0.00958, -0.00933, -0.00933, 0.00947, 0.00933, 0.00933, 0.00947, -0.00861, -0.0059, 0.00269, 0.00861, 0.0059, 0.00269, 0.0, 0.0, 0.00958, -0.0, 0.0, 0.00532, -0.00501, -0.0, 0.0, 0.00501, -0.0, 0.0, 0.00444, -0.0012, -0.00242, -0.00444, 0.0012, -0.00242, 0.0066, -0.00645, -0.00592, -0.0066, 0.00645, -0.00592, 0.00645, -0.0066, -0.00592, -0.00645, 0.0066, -0.00592, 0.0012, -0.00444, -0.00242, -0.0012, 0.00444, -0.00242, -0.0012, -0.00444, 0.00242, 0.0012, 0.00444, 0.00242, -0.00645, -0.0066, 0.00592, 0.00645, 0.0066, 0.00592, -0.0066, -0.00645, 0.00592, 0.0066, 0.00645, 0.00592, -0.00444, -0.0012, 0.00242, 0.00444, 0.0012, 0.00242, -0.00444, 0.0012, -0.00242, 0.0066, -0.00645, -0.00592, -0.0066, 0.00645, -0.00592, 0.00645, -0.0066, -0.00592, -0.00645, 0.0066, -0.00592, 0.0012, -0.00444, -0.00242, -0.0012, 0.00444, -0.00242, -0.0012, -0.00444, 0.00242, 0.0012, 0.00444, 0.00242, -0.00645, -0.0066, 0.00592, 0.00645, 0.0066, 0.00592, -0.0066, -0.00645, 0.00592, 0.0066, 0.00645, 0.00592, -0.00444, -0.0012, 0.00242, 0.00444, 0.0012, 0.00242, 0.00444, -0.0012, -0.00242, 0.01354, -0.00939, -0.00035, -0.01354, 0.00939, -0.00035, 0.00939, -0.01354, -0.00035, -0.00939, 0.01354, -0.00035, -0.00939, -0.01354, 0.00035, 0.00939, 0.01354, 0.00035, -0.01354, -0.00939, 0.00035, 0.01354, 0.00939, 0.00035, 0.01354, -0.00939, -0.00035, -0.01354, 0.00939, -0.00035, 0.00939, -0.01354, -0.00035, -0.00939, 0.01354, -0.00035, -0.00939, -0.01354, 0.00035, 0.00939, 0.01354, 0.00035, -0.01354, -0.00939, 0.00035, 0.01354, 0.00939, 0.00035], [0.00174, -0.00653, 0.01172, -0.00174, 0.00653, 0.01172, 0.0, 0.0, 0.00707, 0.0, -0.0, 0.00691, 0.00098, -0.00098, 0.00905, -0.00098, 0.00098, 0.00905, 0.00653, -0.00174, 0.01172, -0.00653, 0.00174, 0.01172, 0.0, 0.0, 0.00691, -0.0, -0.0, 0.00707, -0.0, 0.00495, -0.0, -0.0, -0.00495, -0.0, -0.00653, -0.00174, -0.01172, 0.00653, 0.00174, -0.01172, -0.0, 0.0, -0.00707, 0.0, 0.0, -0.00691, -0.00098, -0.00098, -0.00905, 0.00098, 0.00098, -0.00905, -0.00174, -0.00653, -0.01172, 0.00174, 0.00653, -0.01172, 0.0, 0.0, -0.00691, 0.0, -0.0, -0.00707, 0.00495, -0.0, 0.0, -0.00495, -0.0, 0.0, 0.00174, -0.00653, 0.01172, -0.00174, 0.00653, 0.01172, -0.0, -0.0, 0.00707, -0.0, 0.0, 0.00691, 0.00098, -0.00098, 0.00905, -0.00098, 0.00098, 0.00905, 0.00653, -0.00174, 0.01172, -0.00653, 0.00174, 0.01172, -0.0, -0.0, 0.00691, 0.0, 0.0, 0.00707, 0.0, 0.00495, 0.0, 0.0, -0.00495, 0.0, -0.00653, -0.00174, -0.01172, 0.00653, 0.00174, -0.01172, 0.0, -0.0, -0.00707, -0.0, -0.0, -0.00691, -0.00098, -0.00098, -0.00905, 0.00098, 0.00098, -0.00905, -0.00174, -0.00653, -0.01172, 0.00174, 0.00653, -0.01172, -0.0, -0.0, -0.00691, -0.0, 0.0, -0.00707, 0.00495, 0.0, -0.0, -0.00495, 0.0, -0.0, 0.00213, -0.00666, 0.01023, -0.00213, 0.00666, 0.01023, 0.00037, -0.00152, 0.01017, -0.00037, 0.00152, 0.01017, 0.00152, -0.00037, 0.01017, -0.00152, 0.00037, 0.01017, 0.00666, -0.00213, 0.01023, -0.00666, 0.00213, 0.01023, -0.00666, -0.00213, -0.01023, 0.00666, 0.00213, -0.01023, -0.00152, -0.00037, -0.01017, 0.00152, 0.00037, -0.01017, -0.00037, -0.00152, -0.01017, 0.00037, 0.00152, -0.01017, -0.00213, -0.00666, -0.01023, 0.00213, 0.00666, -0.01023, -0.00213, 0.00666, 0.01023, 0.00037, -0.00152, 0.01017, -0.00037, 0.00152, 0.01017, 0.00152, -0.00037, 0.01017, -0.00152, 0.00037, 0.01017, 0.00666, -0.00213, 0.01023, -0.00666, 0.00213, 0.01023, -0.00666, -0.00213, -0.01023, 0.00666, 0.00213, -0.01023, -0.00152, -0.00037, -0.01017, 0.00152, 0.00037, -0.01017, -0.00037, -0.00152, -0.01017, 0.00037, 0.00152, -0.01017, -0.00213, -0.00666, -0.01023, 0.00213, 0.00666, -0.01023, 0.00213, -0.00666, 0.01023, 0.00159, -0.0122, 0.01378, -0.00159, 0.0122, 0.01378, 0.0122, -0.00159, 0.01378, -0.0122, 0.00159, 0.01378, -0.0122, -0.00159, -0.01378, 0.0122, 0.00159, -0.01378, -0.00159, -0.0122, -0.01378, 0.00159, 0.0122, -0.01378, 0.00159, -0.0122, 0.01378, -0.00159, 0.0122, 0.01378, 0.0122, -0.00159, 0.01378, -0.0122, 0.00159, 0.01378, -0.0122, -0.00159, -0.01378, 0.0122, 0.00159, -0.01378, -0.00159, -0.0122, -0.01378, 0.00159, 0.0122, -0.01378]]), "E2u2":array([[0.0019, 0.01153, 0.00197, -0.0019, -0.01153, 0.00197, -0.0, -0.0, 0.0097, 0.0, 0.0, 0.00535, 0.00378, 0.00378, -0.0, -0.00378, -0.00378, 0.0, 0.01153, 0.0019, -0.00197, -0.01153, -0.0019, -0.00197, -0.0, -0.0, -0.00535, -0.0, 0.0, -0.0097, 0.01443, -0.0, -0.00992, -0.01443, -0.0, -0.00992, 0.01153, -0.0019, -0.00197, -0.01153, 0.0019, -0.00197, -0.0, -0.0, -0.0097, 0.0, 0.0, -0.00535, 0.00378, -0.00378, -0.0, -0.00378, 0.00378, 0.0, 0.0019, -0.01153, 0.00197, -0.0019, 0.01153, 0.00197, 0.0, 0.0, 0.00535, -0.0, -0.0, 0.0097, -0.0, -0.01443, 0.00992, 0.0, 0.01443, 0.00992, 0.0019, 0.01153, 0.00197, -0.0019, -0.01153, 0.00197, 0.0, 0.0, 0.0097, -0.0, -0.0, 0.00535, 0.00378, 0.00378, 0.0, -0.00378, -0.00378, -0.0, 0.01153, 0.0019, -0.00197, -0.01153, -0.0019, -0.00197, 0.0, 0.0, -0.00535, 0.0, -0.0, -0.0097, 0.01443, 0.0, -0.00992, -0.01443, 0.0, -0.00992, 0.01153, -0.0019, -0.00197, -0.01153, 0.0019, -0.00197, 0.0, 0.0, -0.0097, -0.0, -0.0, -0.00535, 0.00378, -0.00378, 0.0, -0.00378, 0.00378, -0.0, 0.0019, -0.01153, 0.00197, -0.0019, 0.01153, 0.00197, -0.0, -0.0, 0.00535, 0.0, 0.0, 0.0097, -0.0, -0.01443, 0.00992, 0.0, 0.01443, 0.00992, -0.00017, 0.00996, 0.00561, 0.00017, -0.00996, 0.00561, 0.00249, 0.00447, 0.00209, -0.00249, -0.00447, 0.00209, 0.00447, 0.00249, -0.00209, -0.00447, -0.00249, -0.00209, 0.00996, -0.00017, -0.00561, -0.00996, 0.00017, -0.00561, 0.00996, 0.00017, -0.00561, -0.00996, -0.00017, -0.00561, 0.00447, -0.00249, -0.00209, -0.00447, 0.00249, -0.00209, 0.00249, -0.00447, 0.00209, -0.00249, 0.00447, 0.00209, -0.00017, -0.00996, 0.00561, 0.00017, 0.00996, 0.00561, 0.00017, -0.00996, 0.00561, 0.00249, 0.00447, 0.00209, -0.00249, -0.00447, 0.00209, 0.00447, 0.00249, -0.00209, -0.00447, -0.00249, -0.00209, 0.00996, -0.00017, -0.00561, -0.00996, 0.00017, -0.00561, 0.00996, 0.00017, -0.00561, -0.00996, -0.00017, -0.00561, 0.00447, -0.00249, -0.00209, -0.00447, 0.00249, -0.00209, 0.00249, -0.00447, 0.00209, -0.00249, 0.00447, 0.00209, -0.00017, -0.00996, 0.00561, 0.00017, 0.00996, 0.00561, -0.00017, 0.00996, 0.00561, 0.00271, 0.01826, -0.00056, -0.00271, -0.01826, -0.00056, 0.01826, 0.00271, 0.00056, -0.01826, -0.00271, 0.00056, 0.01826, -0.00271, 0.00056, -0.01826, 0.00271, 0.00056, 0.00271, -0.01826, -0.00056, -0.00271, 0.01826, -0.00056, 0.00271, 0.01826, -0.00056, -0.00271, -0.01826, -0.00056, 0.01826, 0.00271, 0.00056, -0.01826, -0.00271, 0.00056, 0.01826, -0.00271, 0.00056, -0.01826, 0.00271, 0.00056, 0.00271, -0.01826, -0.00056, -0.00271, 0.01826, -0.00056], [-0.00365, 0.00574, -0.00963, 0.00365, -0.00574, -0.00963, -0.0, 0.0, -0.00844, -0.0, -0.0, -0.00591, -0.00409, -0.00409, -0.0, 0.00409, 0.00409, 0.0, 0.00574, -0.00365, 0.00963, -0.00574, 0.00365, 0.00963, -0.0, -0.0, 0.00591, -0.0, 0.0, 0.00844, -0.00018, 0.0, 0.01021, 0.00018, 0.0, 0.01021, 0.00574, 0.00365, 0.00963, -0.00574, -0.00365, 0.00963, -0.0, -0.0, 0.00844, -0.0, -0.0, 0.00591, -0.00409, 0.00409, 0.0, 0.00409, -0.00409, 0.0, -0.00365, -0.00574, -0.00963, 0.00365, 0.00574, -0.00963, -0.0, -0.0, -0.00591, -0.0, 0.0, -0.00844, -0.0, 0.00018, -0.01021, -0.0, -0.00018, -0.01021, -0.00365, 0.00574, -0.00963, 0.00365, -0.00574, -0.00963, 0.0, -0.0, -0.00844, 0.0, 0.0, -0.00591, -0.00409, -0.00409, -0.0, 0.00409, 0.00409, -0.0, 0.00574, -0.00365, 0.00963, -0.00574, 0.00365, 0.00963, 0.0, 0.0, 0.00591, 0.0, -0.0, 0.00844, -0.00018, -0.0, 0.01021, 0.00018, -0.0, 0.01021, 0.00574, 0.00365, 0.00963, -0.00574, -0.00365, 0.00963, 0.0, 0.0, 0.00844, 0.0, 0.0, 0.00591, -0.00409, 0.00409, -0.0, 0.00409, -0.00409, -0.0, -0.00365, -0.00574, -0.00963, 0.00365, 0.00574, -0.00963, 0.0, 0.0, -0.00591, 0.0, -0.0, -0.00844, 0.0, 0.00018, -0.01021, 0.0, -0.00018, -0.01021, 0.00024, -0.00026, -0.00939, -0.00024, 0.00026, -0.00939, -0.00439, 0.0084, -0.0102, 0.00439, -0.0084, -0.0102, 0.0084, -0.00439, 0.0102, -0.0084, 0.00439, 0.0102, -0.00026, 0.00024, 0.00939, 0.00026, -0.00024, 0.00939, -0.00026, -0.00024, 0.00939, 0.00026, 0.00024, 0.00939, 0.0084, 0.00439, 0.0102, -0.0084, -0.00439, 0.0102, -0.00439, -0.0084, -0.0102, 0.00439, 0.0084, -0.0102, 0.00024, 0.00026, -0.00939, -0.00024, -0.00026, -0.00939, -0.00024, 0.00026, -0.00939, -0.00439, 0.0084, -0.0102, 0.00439, -0.0084, -0.0102, 0.0084, -0.00439, 0.0102, -0.0084, 0.00439, 0.0102, -0.00026, 0.00024, 0.00939, 0.00026, -0.00024, 0.00939, -0.00026, -0.00024, 0.00939, 0.00026, 0.00024, 0.00939, 0.0084, 0.00439, 0.0102, -0.0084, -0.00439, 0.0102, -0.00439, -0.0084, -0.0102, 0.00439, 0.0084, -0.0102, 0.00024, 0.00026, -0.00939, -0.00024, -0.00026, -0.00939, 0.00024, -0.00026, -0.00939, -0.00629, 0.00788, -0.01245, 0.00629, -0.00788, -0.01245, 0.00788, -0.00629, 0.01245, -0.00788, 0.00629, 0.01245, 0.00788, 0.00629, 0.01245, -0.00788, -0.00629, 0.01245, -0.00629, -0.00788, -0.01245, 0.00629, 0.00788, -0.01245, -0.00629, 0.00788, -0.01245, 0.00629, -0.00788, -0.01245, 0.00788, -0.00629, 0.01245, -0.00788, 0.00629, 0.01245, 0.00788, 0.00629, 0.01245, -0.00788, -0.00629, 0.01245, -0.00629, -0.00788, -0.01245, 0.00629, 0.00788, -0.01245]]), "E2u3":array([[0.00764, 0.00317, -0.0, -0.00764, -0.00317, 0.0, 0.0, -0.0, 0.00302, -0.0, -0.0, -0.00302, 0.00947, -0.00426, -0.00675, -0.00947, 0.00426, -0.00675, 0.01192, -0.00494, -0.00341, -0.01192, 0.00494, -0.00341, 0.0, -0.0, -0.01026, 0.0, 0.0, -0.01026, 0.00971, -0.00368, -0.00675, -0.00971, 0.00368, -0.00675, 0.00317, -0.00764, -0.0, -0.00317, 0.00764, 0.0, 0.0, 0.0, -0.00302, 0.0, -0.0, 0.00302, -0.00426, -0.00947, 0.00675, 0.00426, 0.00947, 0.00675, -0.00494, -0.01192, 0.00341, 0.00494, 0.01192, 0.00341, 0.0, 0.0, 0.01026, -0.0, 0.0, 0.01026, -0.00368, -0.00971, 0.00675, 0.00368, 0.00971, 0.00675, 0.00764, 0.00317, 0.0, -0.00764, -0.00317, -0.0, -0.0, 0.0, 0.00302, 0.0, 0.0, -0.00302, 0.00947, -0.00426, -0.00675, -0.00947, 0.00426, -0.00675, 0.01192, -0.00494, -0.00341, -0.01192, 0.00494, -0.00341, -0.0, 0.0, -0.01026, -0.0, -0.0, -0.01026, 0.00971, -0.00368, -0.00675, -0.00971, 0.00368, -0.00675, 0.00317, -0.00764, 0.0, -0.00317, 0.00764, -0.0, -0.0, -0.0, -0.00302, -0.0, 0.0, 0.00302, -0.00426, -0.00947, 0.00675, 0.00426, 0.00947, 0.00675, -0.00494, -0.01192, 0.00341, 0.00494, 0.01192, 0.00341, -0.0, -0.0, 0.01026, 0.0, -0.0, 0.01026, -0.00368, -0.00971, 0.00675, 0.00368, 0.00971, 0.00675, 0.00327, 0.00582, 0.00243, -0.00327, -0.00582, 0.00243, 0.00643, -0.0018, -0.00243, -0.00643, 0.0018, -0.00243, 0.00771, -0.00298, -0.00553, -0.00771, 0.00298, -0.00553, 0.00755, -0.00334, -0.00553, -0.00755, 0.00334, -0.00553, 0.00582, -0.00327, -0.00243, -0.00582, 0.00327, -0.00243, -0.0018, -0.00643, 0.00243, 0.0018, 0.00643, 0.00243, -0.00298, -0.00771, 0.00553, 0.00298, 0.00771, 0.00553, -0.00334, -0.00755, 0.00553, 0.00334, 0.00755, 0.00553, -0.00327, -0.00582, 0.00243, 0.00643, -0.0018, -0.00243, -0.00643, 0.0018, -0.00243, 0.00771, -0.00298, -0.00553, -0.00771, 0.00298, -0.00553, 0.00755, -0.00334, -0.00553, -0.00755, 0.00334, -0.00553, 0.00582, -0.00327, -0.00243, -0.00582, 0.00327, -0.00243, -0.0018, -0.00643, 0.00243, 0.0018, 0.00643, 0.00243, -0.00298, -0.00771, 0.00553, 0.00298, 0.00771, 0.00553, -0.00334, -0.00755, 0.00553, 0.00334, 0.00755, 0.00553, 0.00327, 0.00582, 0.00243, 0.01193, 0.00494, -0.0, -0.01193, -0.00494, 0.0, 0.01842, -0.00763, -8e-05, -0.01842, 0.00763, -8e-05, 0.00494, -0.01193, 0.0, -0.00494, 0.01193, 0.0, -0.00763, -0.01842, 8e-05, 0.00763, 0.01842, 8e-05, 0.01193, 0.00494, -0.0, -0.01193, -0.00494, 0.0, 0.01842, -0.00763, -8e-05, -0.01842, 0.00763, -8e-05, 0.00494, -0.01193, -0.0, -0.00494, 0.01193, -0.0, -0.00763, -0.01842, 8e-05, 0.00763, 0.01842, 8e-05], [-0.00161, -0.00067, -0.0, 0.00161, 0.00067, 0.0, 0.0, 0.0, -0.00022, -0.0, 0.0, 0.00022, -0.00295, -0.00248, 0.00616, 0.00295, 0.00248, 0.00616, 0.01025, -0.00425, 0.01584, -0.01025, 0.00425, 0.01584, 0.0, 0.0, 0.00904, 0.0, -0.0, 0.00904, -0.00033, 0.00384, 0.00616, 0.00033, -0.00384, 0.00616, -0.00067, 0.00161, -0.0, 0.00067, -0.00161, -0.0, -0.0, 0.0, 0.00022, 0.0, -0.0, -0.00022, -0.00248, 0.00295, -0.00616, 0.00248, -0.00295, -0.00616, -0.00425, -0.01025, -0.01584, 0.00425, 0.01025, -0.01584, -0.0, 0.0, -0.00904, 0.0, 0.0, -0.00904, 0.00384, 0.00033, -0.00616, -0.00384, -0.00033, -0.00616, -0.00161, -0.00067, -0.0, 0.00161, 0.00067, -0.0, -0.0, -0.0, -0.00022, 0.0, -0.0, 0.00022, -0.00295, -0.00248, 0.00616, 0.00295, 0.00248, 0.00616, 0.01025, -0.00425, 0.01584, -0.01025, 0.00425, 0.01584, -0.0, -0.0, 0.00904, -0.0, 0.0, 0.00904, -0.00033, 0.00384, 0.00616, 0.00033, -0.00384, 0.00616, -0.00067, 0.00161, 0.0, 0.00067, -0.00161, 0.0, 0.0, -0.0, 0.00022, -0.0, 0.0, -0.00022, -0.00248, 0.00295, -0.00616, 0.00248, -0.00295, -0.00616, -0.00425, -0.01025, -0.01584, 0.00425, 0.01025, -0.01584, 0.0, -0.0, -0.00904, -0.0, -0.0, -0.00904, 0.00384, 0.00033, -0.00616, -0.00384, -0.00033, -0.00616, 0.00214, -0.00693, 0.00143, -0.00214, 0.00693, 0.00143, -0.00339, 0.00641, -0.00143, 0.00339, -0.00641, -0.00143, 0.00544, -0.00234, 0.01459, -0.00544, 0.00234, 0.01459, 0.00551, -0.00219, 0.01459, -0.00551, 0.00219, 0.01459, -0.00693, -0.00214, -0.00143, 0.00693, 0.00214, -0.00143, 0.00641, 0.00339, 0.00143, -0.00641, -0.00339, 0.00143, -0.00234, -0.00544, -0.01459, 0.00234, 0.00544, -0.01459, -0.00219, -0.00551, -0.01459, 0.00219, 0.00551, -0.01459, -0.00214, 0.00693, 0.00143, -0.00339, 0.00641, -0.00143, 0.00339, -0.00641, -0.00143, 0.00544, -0.00234, 0.01459, -0.00544, 0.00234, 0.01459, 0.00551, -0.00219, 0.01459, -0.00551, 0.00219, 0.01459, -0.00693, -0.00214, -0.00143, 0.00693, 0.00214, -0.00143, 0.00641, 0.00339, 0.00143, -0.00641, -0.00339, 0.00143, -0.00234, -0.00544, -0.01459, 0.00234, 0.00544, -0.01459, -0.00219, -0.00551, -0.01459, 0.00219, 0.00551, -0.01459, 0.00214, -0.00693, 0.00143, -0.00441, -0.00183, 0.0, 0.00441, 0.00183, 0.0, 0.01496, -0.0062, 0.01916, -0.01496, 0.0062, 0.01916, -0.00183, 0.00441, -0.0, 0.00183, -0.00441, -0.0, -0.0062, -0.01496, -0.01916, 0.0062, 0.01496, -0.01916, -0.00441, -0.00183, -0.0, 0.00441, 0.00183, -0.0, 0.01496, -0.0062, 0.01916, -0.01496, 0.0062, 0.01916, -0.00183, 0.00441, 0.0, 0.00183, -0.00441, 0.0, -0.0062, -0.01496, -0.01916, 0.0062, 0.01496, -0.01916]]), "E2u4":array([[-0.00496, 0.01196, 0.00301, 0.00496, -0.01196, 0.00301, -0.0, 0.0, 0.01041, 0.0, 0.0, 0.01041, -0.00418, 0.00912, 0.00666, 0.00418, -0.00912, 0.00666, 0.00311, 0.00752, -0.0, -0.00311, -0.00752, -0.0, 0.0, -0.0, 0.00293, 0.0, -0.0, -0.00293, 0.0094, 0.00349, -0.00666, -0.0094, -0.00349, -0.00666, 0.01196, 0.00496, -0.00301, -0.01196, -0.00496, -0.00301, -0.0, -0.0, -0.01041, -0.0, -0.0, -0.01041, 0.00912, 0.00418, -0.00666, -0.00912, -0.00418, -0.00666, 0.00752, -0.00311, -0.0, -0.00752, 0.00311, 0.0, 0.0, 0.0, -0.00293, 0.0, 0.0, 0.00293, 0.00349, -0.0094, 0.00666, -0.00349, 0.0094, 0.00666, -0.00496, 0.01196, 0.00301, 0.00496, -0.01196, 0.00301, 0.0, -0.0, 0.01041, -0.0, -0.0, 0.01041, -0.00418, 0.00912, 0.00666, 0.00418, -0.00912, 0.00666, 0.00311, 0.00752, 0.0, -0.00311, -0.00752, 0.0, -0.0, 0.0, 0.00293, -0.0, 0.0, -0.00293, 0.0094, 0.00349, -0.00666, -0.0094, -0.00349, -0.00666, 0.01196, 0.00496, -0.00301, -0.01196, -0.00496, -0.00301, 0.0, 0.0, -0.01041, 0.0, 0.0, -0.01041, 0.00912, 0.00418, -0.00666, -0.00912, -0.00418, -0.00666, 0.00752, -0.00311, 0.0, -0.00752, 0.00311, -0.0, -0.0, -0.0, -0.00293, -0.0, -0.0, 0.00293, 0.00349, -0.0094, 0.00666, -0.00349, 0.0094, 0.00666, -0.00319, 0.00751, 0.00553, 0.00319, -0.00751, 0.00553, -0.00305, 0.00756, 0.00553, 0.00305, -0.00756, 0.00553, -0.00177, 0.00623, 0.00242, 0.00177, -0.00623, 0.00242, 0.00566, 0.00315, -0.00242, -0.00566, -0.00315, -0.00242, 0.00751, 0.00319, -0.00553, -0.00751, -0.00319, -0.00553, 0.00756, 0.00305, -0.00553, -0.00756, -0.00305, -0.00553, 0.00623, 0.00177, -0.00242, -0.00623, -0.00177, -0.00242, 0.00315, -0.00566, 0.00242, -0.00315, 0.00566, 0.00242, 0.00319, -0.00751, 0.00553, -0.00305, 0.00756, 0.00553, 0.00305, -0.00756, 0.00553, -0.00177, 0.00623, 0.00242, 0.00177, -0.00623, 0.00242, 0.00566, 0.00315, -0.00242, -0.00566, -0.00315, -0.00242, 0.00751, 0.00319, -0.00553, -0.00751, -0.00319, -0.00553, 0.00756, 0.00305, -0.00553, -0.00756, -0.00305, -0.00553, 0.00623, 0.00177, -0.00242, -0.00623, -0.00177, -0.00242, 0.00315, -0.00566, 0.00242, -0.00315, 0.00566, 0.00242, -0.00319, 0.00751, 0.00553, -0.00808, 0.0195, -0.00039, 0.00808, -0.0195, -0.00039, 0.0048, 0.01159, -0.0, -0.0048, -0.01159, 0.0, 0.0195, 0.00808, 0.00039, -0.0195, -0.00808, 0.00039, 0.01159, -0.0048, -0.0, -0.01159, 0.0048, 0.0, -0.00808, 0.0195, -0.00039, 0.00808, -0.0195, -0.00039, 0.0048, 0.01159, 0.0, -0.0048, -0.01159, 0.0, 0.0195, 0.00808, 0.00039, -0.0195, -0.00808, 0.00039, 0.01159, -0.0048, 0.0, -0.01159, 0.0048, 0.0], [0.0036, -0.0087, 0.01561, -0.0036, 0.0087, 0.01561, 0.0, 0.0, 0.01192, -0.0, -0.0, 0.01192, 0.00347, 0.00206, 0.00777, -0.00347, -0.00206, 0.00777, 0.00048, 0.00115, -0.0, -0.00048, -0.00115, 0.0, 0.0, -0.0, 0.00196, 0.0, -0.0, -0.00196, -0.001, 0.00391, -0.00777, 0.001, -0.00391, -0.00777, -0.0087, -0.0036, -0.01561, 0.0087, 0.0036, -0.01561, -0.0, 0.0, -0.01192, 0.0, -0.0, -0.01192, 0.00206, -0.00347, -0.00777, -0.00206, 0.00347, -0.00777, 0.00115, -0.00048, -0.0, -0.00115, 0.00048, 0.0, 0.0, -0.0, -0.00196, 0.0, 0.0, 0.00196, 0.00391, 0.001, 0.00777, -0.00391, -0.001, 0.00777, 0.0036, -0.0087, 0.01561, -0.0036, 0.0087, 0.01561, -0.0, -0.0, 0.01192, 0.0, 0.0, 0.01192, 0.00347, 0.00206, 0.00777, -0.00347, -0.00206, 0.00777, 0.00048, 0.00115, 0.0, -0.00048, -0.00115, -0.0, -0.0, 0.0, 0.00196, -0.0, 0.0, -0.00196, -0.001, 0.00391, -0.00777, 0.001, -0.00391, -0.00777, -0.0087, -0.0036, -0.01561, 0.0087, 0.0036, -0.01561, 0.0, -0.0, -0.01192, -0.0, 0.0, -0.01192, 0.00206, -0.00347, -0.00777, -0.00206, 0.00347, -0.00777, 0.00115, -0.00048, 0.0, -0.00115, 0.00048, -0.0, -0.0, 0.0, -0.00196, -0.0, -0.0, 0.00196, 0.00391, 0.001, 0.00777, -0.00391, -0.001, 0.00777, 0.00227, -0.0066, 0.01512, -0.00227, 0.0066, 0.01512, 0.00306, -0.00627, 0.01512, -0.00306, 0.00627, 0.01512, -0.00445, 0.00238, 0.00053, 0.00445, -0.00238, 0.00053, 0.00483, -0.00147, -0.00053, -0.00483, 0.00147, -0.00053, -0.0066, -0.00227, -0.01512, 0.0066, 0.00227, -0.01512, -0.00627, -0.00306, -0.01512, 0.00627, 0.00306, -0.01512, 0.00238, 0.00445, -0.00053, -0.00238, -0.00445, -0.00053, -0.00147, -0.00483, 0.00053, 0.00147, 0.00483, 0.00053, -0.00227, 0.0066, 0.01512, 0.00306, -0.00627, 0.01512, -0.00306, 0.00627, 0.01512, -0.00445, 0.00238, 0.00053, 0.00445, -0.00238, 0.00053, 0.00483, -0.00147, -0.00053, -0.00483, 0.00147, -0.00053, -0.0066, -0.00227, -0.01512, 0.0066, 0.00227, -0.01512, -0.00627, -0.00306, -0.01512, 0.00627, 0.00306, -0.01512, 0.00238, 0.00445, -0.00053, -0.00238, -0.00445, -0.00053, -0.00147, -0.00483, 0.00053, 0.00147, 0.00483, 0.00053, 0.00227, -0.0066, 0.01512, 0.00579, -0.01398, 0.01948, -0.00579, 0.01398, 0.01948, 0.00148, 0.00356, -0.0, -0.00148, -0.00356, 0.0, -0.01398, -0.00579, -0.01948, 0.01398, 0.00579, -0.01948, 0.00356, -0.00148, 0.0, -0.00356, 0.00148, -0.0, 0.00579, -0.01398, 0.01948, -0.00579, 0.01398, 0.01948, 0.00148, 0.00356, 0.0, -0.00148, -0.00356, -0.0, -0.01398, -0.00579, -0.01948, 0.01398, 0.00579, -0.01948, 0.00356, -0.00148, -0.0, -0.00356, 0.00148, -0.0]]), "E1u1":array([[-0.00827, 0.00871, -0.00055, -0.00827, 0.00871, 0.00055, -0.004, 0.00754, 0.0, -0.00586, 0.01269, 0.0, -0.00785, 0.01155, 0.00252, -0.00785, 0.01155, -0.00252, 0.00774, 0.00724, 0.00034, 0.00774, 0.00724, -0.00034, -0.00239, 0.00183, -0.0, 0.00983, -0.00498, -0.0, 0.01924, -0.0, -0.00301, 0.01924, -0.0, 0.00301, 0.00774, -0.00724, 0.00034, 0.00774, -0.00724, -0.00034, 0.00983, 0.00498, -0.0, -0.00239, -0.00183, -0.0, -0.00785, -0.01155, 0.00252, -0.00785, -0.01155, -0.00252, -0.00827, -0.00871, -0.00055, -0.00827, -0.00871, 0.00055, -0.00586, -0.01269, -0.0, -0.004, -0.00754, -0.0, -0.00328, 0.0, 0.0, -0.00328, 0.0, 0.0, -0.00827, 0.00871, -0.00055, -0.00827, 0.00871, 0.00055, -0.004, 0.00754, 0.0, -0.00586, 0.01269, 0.0, -0.00785, 0.01155, 0.00252, -0.00785, 0.01155, -0.00252, 0.00774, 0.00724, 0.00034, 0.00774, 0.00724, -0.00034, -0.00239, 0.00183, 0.0, 0.00983, -0.00498, 0.0, 0.01924, -0.0, -0.00301, 0.01924, -0.0, 0.00301, 0.00774, -0.00724, 0.00034, 0.00774, -0.00724, -0.00034, 0.00983, 0.00498, 0.0, -0.00239, -0.00183, 0.0, -0.00785, -0.01155, 0.00252, -0.00785, -0.01155, -0.00252, -0.00827, -0.00871, -0.00055, -0.00827, -0.00871, 0.00055, -0.00586, -0.01269, -0.0, -0.004, -0.00754, -0.0, -0.00328, -0.0, -0.0, -0.00328, -0.0, -0.0, -0.00563, 0.00605, -0.00019, -0.00563, 0.00605, 0.00019, -0.00771, 0.01315, -1e-05, -0.00771, 0.01315, 1e-05, -0.00267, 0.00756, -2e-05, -0.00267, 0.00756, 2e-05, 0.01657, -0.00033, -0.0001, 0.01657, -0.00033, 0.0001, 0.01657, 0.00033, -0.0001, 0.01657, 0.00033, 0.0001, -0.00267, -0.00756, -2e-05, -0.00267, -0.00756, 2e-05, -0.00771, -0.01315, -1e-05, -0.00771, -0.01315, 1e-05, -0.00563, -0.00605, -0.00019, -0.00563, -0.00605, 0.00019, -0.00563, 0.00605, 0.00019, -0.00771, 0.01315, -1e-05, -0.00771, 0.01315, 1e-05, -0.00267, 0.00756, -2e-05, -0.00267, 0.00756, 2e-05, 0.01657, -0.00033, -0.0001, 0.01657, -0.00033, 0.0001, 0.01657, 0.00033, -0.0001, 0.01657, 0.00033, 0.0001, -0.00267, -0.00756, -2e-05, -0.00267, -0.00756, 2e-05, -0.00771, -0.01315, -1e-05, -0.00771, -0.01315, 1e-05, -0.00563, -0.00605, -0.00019, -0.00563, -0.00605, 0.00019, -0.00563, 0.00605, -0.00019, -0.00973, 0.00997, -0.00044, -0.00973, 0.00997, 0.00044, 0.01078, 0.01079, 0.0009, 0.01078, 0.01079, -0.0009, 0.01078, -0.01079, 0.0009, 0.01078, -0.01079, -0.0009, -0.00973, -0.00997, -0.00044, -0.00973, -0.00997, 0.00044, -0.00973, 0.00997, -0.00044, -0.00973, 0.00997, 0.00044, 0.01078, 0.01079, 0.0009, 0.01078, 0.01079, -0.0009, 0.01078, -0.01079, 0.0009, 0.01078, -0.01079, -0.0009, -0.00973, -0.00997, -0.00044, -0.00973, -0.00997, 0.00044], [0.00199, -0.00713, 0.00392, 0.00199, -0.00713, -0.00392, -0.00293, 0.00719, -0.0, -0.0075, 0.0065, 0.0, -0.00707, 0.0068, -0.00355, -0.00707, 0.0068, 0.00355, 0.0028, 0.00091, 0.00089, 0.0028, 0.00091, -0.00089, -0.00407, 0.00079, 0.0, 0.00291, 0.00627, 0.0, -0.0011, 0.0, -0.00061, -0.0011, 0.0, 0.00061, 0.0028, -0.00091, 0.00089, 0.0028, -0.00091, -0.00089, 0.00291, -0.00627, 0.0, -0.00407, -0.00079, 0.0, -0.00707, -0.0068, -0.00355, -0.00707, -0.0068, 0.00355, 0.00199, 0.00713, 0.00392, 0.00199, 0.00713, -0.00392, -0.0075, -0.0065, -0.0, -0.00293, -0.00719, -0.0, -0.00119, 0.0, 0.0, -0.00119, 0.0, 0.0, 0.00199, -0.00713, 0.00392, 0.00199, -0.00713, -0.00392, -0.00293, 0.00719, 0.0, -0.0075, 0.0065, 0.0, -0.00707, 0.0068, -0.00355, -0.00707, 0.0068, 0.00355, 0.0028, 0.00091, 0.00089, 0.0028, 0.00091, -0.00089, -0.00407, 0.00079, -0.0, 0.00291, 0.00627, -0.0, -0.0011, -0.0, -0.00061, -0.0011, -0.0, 0.00061, 0.0028, -0.00091, 0.00089, 0.0028, -0.00091, -0.00089, 0.00291, -0.00627, -0.0, -0.00407, -0.00079, -0.0, -0.00707, -0.0068, -0.00355, -0.00707, -0.0068, 0.00355, 0.00199, 0.00713, 0.00392, 0.00199, 0.00713, -0.00392, -0.0075, -0.0065, -0.0, -0.00293, -0.00719, 0.0, -0.00119, -0.0, -0.0, -0.00119, -0.0, -0.0, -0.00625, 0.00739, -0.00248, -0.00625, 0.00739, 0.00248, -0.00463, 0.00468, 0.00082, -0.00463, 0.00468, -0.00082, -0.00027, 0.00458, 0.00034, -0.00027, 0.00458, -0.00034, 0.0027, 0.00146, 0.0007, 0.0027, 0.00146, -0.0007, 0.0027, -0.00146, 0.0007, 0.0027, -0.00146, -0.0007, -0.00027, -0.00458, 0.00034, -0.00027, -0.00458, -0.00034, -0.00463, -0.00468, 0.00082, -0.00463, -0.00468, -0.00082, -0.00625, -0.00739, -0.00248, -0.00625, -0.00739, 0.00248, -0.00625, 0.00739, 0.00248, -0.00463, 0.00468, 0.00082, -0.00463, 0.00468, -0.00082, -0.00027, 0.00458, 0.00034, -0.00027, 0.00458, -0.00034, 0.0027, 0.00146, 0.0007, 0.0027, 0.00146, -0.0007, 0.0027, -0.00146, 0.0007, 0.0027, -0.00146, -0.0007, -0.00027, -0.00458, 0.00034, -0.00027, -0.00458, -0.00034, -0.00463, -0.00468, 0.00082, -0.00463, -0.00468, -0.00082, -0.00625, -0.00739, -0.00248, -0.00625, -0.00739, 0.00248, -0.00625, 0.00739, -0.00248, 0.01367, -0.03608, 0.01835, 0.01367, -0.03608, -0.01835, 0.00402, 0.00305, 0.00183, 0.00402, 0.00305, -0.00183, 0.00402, -0.00305, 0.00183, 0.00402, -0.00305, -0.00183, 0.01367, 0.03608, 0.01835, 0.01367, 0.03608, -0.01835, 0.01367, -0.03608, 0.01835, 0.01367, -0.03608, -0.01835, 0.00402, 0.00305, 0.00183, 0.00402, 0.00305, -0.00183, 0.00402, -0.00305, 0.00183, 0.00402, -0.00305, -0.00183, 0.01367, 0.03608, 0.01835, 0.01367, 0.03608, -0.01835]]), "E1u2":array([[-0.0003, 0.0123, -0.00063, -0.0003, 0.0123, 0.00063, -0.00669, 0.01241, 0.0, -0.00442, 0.00722, 0.0, 0.00281, 0.00281, 0.0, 0.00281, 0.00281, -0.0, 0.0123, -0.0003, 0.00063, 0.0123, -0.0003, -0.00063, 0.00722, -0.00442, -0.0, 0.01241, -0.00669, -0.0, 0.01408, -0.00345, -0.00248, 0.01408, -0.00345, 0.00248, 1e-05, -0.01089, -0.00022, 1e-05, -0.01089, 0.00022, 0.00183, 0.00088, 0.0, -0.01, -0.00505, 0.0, -0.01436, -0.01436, 0.00258, -0.01436, -0.01436, -0.00258, -0.01089, 1e-05, -0.00022, -0.01089, 1e-05, 0.00022, -0.00505, -0.01, 0.0, 0.00088, 0.00183, 0.0, -0.00345, 0.01408, -0.00248, -0.00345, 0.01408, 0.00248, -0.0003, 0.0123, -0.00063, -0.0003, 0.0123, 0.00063, -0.00669, 0.01241, 0.0, -0.00442, 0.00722, 0.0, 0.00281, 0.00281, -0.0, 0.00281, 0.00281, 0.0, 0.0123, -0.0003, 0.00063, 0.0123, -0.0003, -0.00063, 0.00722, -0.00442, -0.0, 0.01241, -0.00669, -0.0, 0.01408, -0.00345, -0.00248, 0.01408, -0.00345, 0.00248, 1e-05, -0.01089, -0.00022, 1e-05, -0.01089, 0.00022, 0.00183, 0.00088, -0.0, -0.01, -0.00505, -0.0, -0.01436, -0.01436, 0.00258, -0.01436, -0.01436, -0.00258, -0.01089, 1e-05, -0.00022, -0.01089, 1e-05, 0.00022, -0.00505, -0.01, -0.0, 0.00088, 0.00183, -0.0, -0.00345, 0.01408, -0.00248, -0.00345, 0.01408, 0.00248, -0.00472, 0.01608, -0.00028, -0.00472, 0.01608, 0.00028, -0.00076, 0.00738, -0.0002, -0.00076, 0.00738, 0.0002, 0.00738, -0.00076, 0.0002, 0.00738, -0.00076, -0.0002, 0.01608, -0.00472, 0.00028, 0.01608, -0.00472, -0.00028, 0.00782, -0.00347, -0.00016, 0.00782, -0.00347, 0.00016, -0.01151, -0.01104, -6e-05, -0.01151, -0.01104, 6e-05, -0.01104, -0.01151, -6e-05, -0.01104, -0.01151, 6e-05, -0.00347, 0.00782, -0.00016, -0.00347, 0.00782, 0.00016, -0.00472, 0.01608, 0.00028, -0.00076, 0.00738, -0.0002, -0.00076, 0.00738, 0.0002, 0.00738, -0.00076, 0.0002, 0.00738, -0.00076, -0.0002, 0.01608, -0.00472, 0.00028, 0.01608, -0.00472, -0.00028, 0.00782, -0.00347, -0.00016, 0.00782, -0.00347, 0.00016, -0.01151, -0.01104, -6e-05, -0.01151, -0.01104, 6e-05, -0.01104, -0.01151, -6e-05, -0.01104, -0.01151, 6e-05, -0.00347, 0.00782, -0.00016, -0.00347, 0.00782, 0.00016, -0.00472, 0.01608, -0.00028, 9e-05, 0.01499, -0.0014, 9e-05, 0.01499, 0.0014, 0.01499, 9e-05, 0.0014, 0.01499, 9e-05, -0.0014, 0.00166, -0.01483, 0.00069, 0.00166, -0.01483, -0.00069, -0.01483, 0.00166, 0.00069, -0.01483, 0.00166, -0.00069, 9e-05, 0.01499, -0.0014, 9e-05, 0.01499, 0.0014, 0.01499, 9e-05, 0.0014, 0.01499, 9e-05, -0.0014, 0.00166, -0.01483, 0.00069, 0.00166, -0.01483, -0.00069, -0.01483, 0.00166, 0.00069, -0.01483, 0.00166, -0.00069], [0.00591, -0.00786, 0.0037, 0.00591, -0.00786, -0.0037, 0.00195, 0.00046, -0.0, 0.0031, 0.0024, -0.0, -0.0023, -0.0023, -0.0, -0.0023, -0.0023, -0.0, -0.00786, 0.00591, -0.0037, -0.00786, 0.00591, 0.0037, 0.0024, 0.0031, 0.0, 0.00046, 0.00195, 0.0, 0.00521, -0.00406, 3e-05, 0.00521, -0.00406, -3e-05, -0.00078, -0.00144, -0.00925, -0.00078, -0.00144, 0.00925, 0.01285, 0.00012, -0.0, 0.00671, -0.00082, -0.0, -0.00015, -0.00015, 9e-05, -0.00015, -0.00015, -9e-05, -0.00144, -0.00078, -0.00925, -0.00144, -0.00078, 0.00925, -0.00082, 0.00671, -0.0, 0.00012, 0.01285, -0.0, -0.00406, 0.00521, 3e-05, -0.00406, 0.00521, -3e-05, 0.00591, -0.00786, 0.0037, 0.00591, -0.00786, -0.0037, 0.00195, 0.00046, 0.0, 0.0031, 0.0024, 0.0, -0.0023, -0.0023, 0.0, -0.0023, -0.0023, 0.0, -0.00786, 0.00591, -0.0037, -0.00786, 0.00591, 0.0037, 0.0024, 0.0031, -0.0, 0.00046, 0.00195, -0.0, 0.00521, -0.00406, 3e-05, 0.00521, -0.00406, -3e-05, -0.00078, -0.00144, -0.00925, -0.00078, -0.00144, 0.00925, 0.01285, 0.00012, 0.0, 0.00671, -0.00082, 0.0, -0.00015, -0.00015, 9e-05, -0.00015, -0.00015, -9e-05, -0.00144, -0.00078, -0.00925, -0.00144, -0.00078, 0.00925, -0.00082, 0.00671, 0.0, 0.00012, 0.01285, 0.0, -0.00406, 0.00521, 3e-05, -0.00406, 0.00521, -3e-05, 0.00133, 0.00321, 0.00147, 0.00133, 0.00321, -0.00147, -0.00099, 0.00044, 0.00215, -0.00099, 0.00044, -0.00215, 0.00044, -0.00099, -0.00215, 0.00044, -0.00099, 0.00215, 0.00321, 0.00133, -0.00147, 0.00321, 0.00133, 0.00147, 0.01163, 0.00388, -0.00327, 0.01163, 0.00388, 0.00327, 0.0034, 0.00154, -0.00457, 0.0034, 0.00154, 0.00457, 0.00154, 0.0034, -0.00457, 0.00154, 0.0034, 0.00457, 0.00388, 0.01163, -0.00327, 0.00388, 0.01163, 0.00327, 0.00133, 0.00321, -0.00147, -0.00099, 0.00044, 0.00215, -0.00099, 0.00044, -0.00215, 0.00044, -0.00099, -0.00215, 0.00044, -0.00099, 0.00215, 0.00321, 0.00133, -0.00147, 0.00321, 0.00133, 0.00147, 0.01163, 0.00388, -0.00327, 0.01163, 0.00388, 0.00327, 0.0034, 0.00154, -0.00457, 0.0034, 0.00154, 0.00457, 0.00154, 0.0034, -0.00457, 0.00154, 0.0034, 0.00457, 0.00388, 0.01163, -0.00327, 0.00388, 0.01163, 0.00327, 0.00133, 0.00321, 0.00147, 0.00919, -0.01058, 0.00637, 0.00919, -0.01058, -0.00637, -0.01058, 0.00919, -0.00637, -0.01058, 0.00919, 0.00637, -0.0199, -0.01108, -0.01947, -0.0199, -0.01108, 0.01947, -0.01108, -0.0199, -0.01947, -0.01108, -0.0199, 0.01947, 0.00919, -0.01058, 0.00637, 0.00919, -0.01058, -0.00637, -0.01058, 0.00919, -0.00637, -0.01058, 0.00919, 0.00637, -0.0199, -0.01108, -0.01947, -0.0199, -0.01108, 0.01947, -0.01108, -0.0199, -0.01947, -0.01108, -0.0199, 0.01947]]), "E1u3":array([[-0.00716, -0.00856, -0.00016, -0.00716, -0.00856, 0.00016, 0.00544, -0.01099, 0.0, 0.00012, 0.00224, 0.0, -0.01164, 0.00883, 0.00119, -0.01164, 0.00883, -0.00119, -0.00835, 0.00775, 0.00025, -0.00835, 0.00775, -0.00025, -0.01312, 0.00672, 0.0, -0.00869, 0.00336, 0.0, -0.0, 0.00343, 0.0, -0.0, 0.00343, -0.0, 0.00835, 0.00775, -0.00025, 0.00835, 0.00775, 0.00025, 0.00869, 0.00336, -0.0, 0.01312, 0.00672, -0.0, 0.01164, 0.00883, -0.00119, 0.01164, 0.00883, 0.00119, 0.00716, -0.00856, 0.00016, 0.00716, -0.00856, -0.00016, -0.00012, 0.00224, -0.0, -0.00544, -0.01099, -0.0, -0.0, -0.02036, 0.00368, -0.0, -0.02036, -0.00368, -0.00716, -0.00856, -0.00016, -0.00716, -0.00856, 0.00016, 0.00544, -0.01099, -0.0, 0.00012, 0.00224, -0.0, -0.01164, 0.00883, 0.00119, -0.01164, 0.00883, -0.00119, -0.00835, 0.00775, 0.00025, -0.00835, 0.00775, -0.00025, -0.01312, 0.00672, 0.0, -0.00869, 0.00336, 0.0, 0.0, 0.00343, -0.0, 0.0, 0.00343, 0.0, 0.00835, 0.00775, -0.00025, 0.00835, 0.00775, 0.00025, 0.00869, 0.00336, -0.0, 0.01312, 0.00672, -0.0, 0.01164, 0.00883, -0.00119, 0.01164, 0.00883, 0.00119, 0.00716, -0.00856, 0.00016, 0.00716, -0.00856, -0.00016, -0.00012, 0.00224, 0.0, -0.00544, -0.01099, 0.0, -0.0, -0.02036, 0.00368, -0.0, -0.02036, -0.00368, 0.00061, -0.01657, 0.00043, 0.00061, -0.01657, -0.00043, -0.00762, 0.00358, -0.00022, -0.00762, 0.00358, 0.00022, -0.0135, 0.00891, -0.00037, -0.0135, 0.00891, 0.00037, -0.00662, 0.00542, -0.00023, -0.00662, 0.00542, 0.00023, 0.00662, 0.00542, 0.00023, 0.00662, 0.00542, -0.00023, 0.0135, 0.00891, 0.00037, 0.0135, 0.00891, -0.00037, 0.00762, 0.00358, 0.00022, 0.00762, 0.00358, -0.00022, -0.00061, -0.01657, -0.00043, -0.00061, -0.01657, 0.00043, 0.00061, -0.01657, -0.00043, -0.00762, 0.00358, -0.00022, -0.00762, 0.00358, 0.00022, -0.0135, 0.00891, -0.00037, -0.0135, 0.00891, 0.00037, -0.00662, 0.00542, -0.00023, -0.00662, 0.00542, 0.00023, 0.00662, 0.00542, 0.00023, 0.00662, 0.00542, -0.00023, 0.0135, 0.00891, 0.00037, 0.0135, 0.00891, -0.00037, 0.00762, 0.00358, 0.00022, 0.00762, 0.00358, -0.00022, -0.00061, -0.01657, -0.00043, -0.00061, -0.01657, 0.00043, 0.00061, -0.01657, 0.00043, -0.01126, -0.01021, 0.00059, -0.01126, -0.01021, -0.00059, -0.00659, 0.00865, 0.00104, -0.00659, 0.00865, -0.00104, 0.00659, 0.00865, -0.00104, 0.00659, 0.00865, 0.00104, 0.01126, -0.01021, -0.00059, 0.01126, -0.01021, 0.00059, -0.01126, -0.01021, 0.00059, -0.01126, -0.01021, -0.00059, -0.00659, 0.00865, 0.00104, -0.00659, 0.00865, -0.00104, 0.00659, 0.00865, -0.00104, 0.00659, 0.00865, 0.00104, 0.01126, -0.01021, -0.00059, 0.01126, -0.01021, 0.00059], [0.00403, -0.0038, 0.00809, 0.00403, -0.0038, -0.00809, 0.00351, 0.01114, 0.0, 0.0019, 0.01045, 0.0, -0.00267, -0.0011, 0.00158, -0.00267, -0.0011, -0.00158, -0.00139, 0.0026, 0.00596, -0.00139, 0.0026, -0.00596, -0.00443, 0.00091, 0.0, -0.00565, 0.00098, -0.0, 0.0, -0.0019, -0.0, 0.0, -0.0019, 0.0, 0.00139, 0.0026, -0.00596, 0.00139, 0.0026, 0.00596, 0.00565, 0.00098, 0.0, 0.00443, 0.00091, -0.0, 0.00267, -0.0011, -0.00158, 0.00267, -0.0011, 0.00158, -0.00403, -0.0038, -0.00809, -0.00403, -0.0038, 0.00809, -0.0019, 0.01045, -0.0, -0.00351, 0.01114, -0.0, 0.0, 0.00637, -0.00087, 0.0, 0.00637, 0.00087, 0.00403, -0.0038, 0.00809, 0.00403, -0.0038, -0.00809, 0.00351, 0.01114, -0.0, 0.0019, 0.01045, -0.0, -0.00267, -0.0011, 0.00158, -0.00267, -0.0011, -0.00158, -0.00139, 0.0026, 0.00596, -0.00139, 0.0026, -0.00596, -0.00443, 0.00091, 0.0, -0.00565, 0.00098, 0.0, -0.0, -0.0019, 0.0, -0.0, -0.0019, -0.0, 0.00139, 0.0026, -0.00596, 0.00139, 0.0026, 0.00596, 0.00565, 0.00098, -0.0, 0.00443, 0.00091, -0.0, 0.00267, -0.0011, -0.00158, 0.00267, -0.0011, 0.00158, -0.00403, -0.0038, -0.00809, -0.00403, -0.0038, 0.00809, -0.0019, 0.01045, 0.0, -0.00351, 0.01114, 0.0, 0.0, 0.00637, -0.00087, 0.0, 0.00637, 0.00087, -0.00052, 0.00952, 0.00295, -0.00052, 0.00952, -0.00295, 0.00019, 0.0022, 0.00393, 0.00019, 0.0022, -0.00393, -0.00402, 0.00105, 0.00123, -0.00402, 0.00105, -0.00123, -0.00509, 0.00409, 0.00264, -0.00509, 0.00409, -0.00264, 0.00509, 0.00409, -0.00264, 0.00509, 0.00409, 0.00264, 0.00402, 0.00105, -0.00123, 0.00402, 0.00105, 0.00123, -0.00019, 0.0022, -0.00393, -0.00019, 0.0022, 0.00393, 0.00052, 0.00952, -0.00295, 0.00052, 0.00952, 0.00295, -0.00052, 0.00952, -0.00295, 0.00019, 0.0022, 0.00393, 0.00019, 0.0022, -0.00393, -0.00402, 0.00105, 0.00123, -0.00402, 0.00105, -0.00123, -0.00509, 0.00409, 0.00264, -0.00509, 0.00409, -0.00264, 0.00509, 0.00409, -0.00264, 0.00509, 0.00409, 0.00264, 0.00402, 0.00105, -0.00123, 0.00402, 0.00105, 0.00123, -0.00019, 0.0022, -0.00393, -0.00019, 0.0022, 0.00393, 0.00052, 0.00952, -0.00295, 0.00052, 0.00952, 0.00295, -0.00052, 0.00952, 0.00295, 0.01199, -0.02351, 0.01739, 0.01199, -0.02351, -0.01739, 0.00896, -0.00502, 0.01193, 0.00896, -0.00502, -0.01193, -0.00896, -0.00502, -0.01193, -0.00896, -0.00502, 0.01193, -0.01199, -0.02351, -0.01739, -0.01199, -0.02351, 0.01739, 0.01199, -0.02351, 0.01739, 0.01199, -0.02351, -0.01739, 0.00896, -0.00502, 0.01193, 0.00896, -0.00502, -0.01193, -0.00896, -0.00502, -0.01193, -0.00896, -0.00502, 0.01193, -0.01199, -0.02351, -0.01739, -0.01199, -0.02351, 0.01739]]), "E1u4":array([[-0.01127, 0.00062, 9e-05, -0.01127, 0.00062, -9e-05, 0.00124, -0.00152, 0.0, -0.00542, 0.01204, 0.0, -0.01455, 0.01455, 0.00299, -0.01455, 0.01455, -0.00299, -0.00062, 0.01127, 9e-05, -0.00062, 0.01127, -9e-05, -0.01204, 0.00542, 0.0, 0.00152, -0.00124, 0.0, 0.01459, 0.00152, -0.00234, 0.01459, 0.00152, 0.00234, 0.01195, 0.00089, -0.00013, 0.01195, 0.00089, 0.00013, 0.01428, 0.00626, -0.0, 0.0081, 0.00333, -0.0, 0.00203, -0.00203, -0.0, 0.00203, -0.00203, 0.0, -0.00089, -0.01195, 0.00013, -0.00089, -0.01195, -0.00013, -0.00333, -0.0081, 0.0, -0.00626, -0.01428, 0.0, -0.00152, -0.01459, 0.00234, -0.00152, -0.01459, -0.00234, -0.01127, 0.00062, 9e-05, -0.01127, 0.00062, -9e-05, 0.00124, -0.00152, -0.0, -0.00542, 0.01204, -0.0, -0.01455, 0.01455, 0.00299, -0.01455, 0.01455, -0.00299, -0.00062, 0.01127, 9e-05, -0.00062, 0.01127, -9e-05, -0.01204, 0.00542, -0.0, 0.00152, -0.00124, -0.0, 0.01459, 0.00152, -0.00234, 0.01459, 0.00152, 0.00234, 0.01195, 0.00089, -0.00013, 0.01195, 0.00089, 0.00013, 0.01428, 0.00626, -0.0, 0.0081, 0.00333, -0.0, 0.00203, -0.00203, 0.0, 0.00203, -0.00203, -0.0, -0.00089, -0.01195, 0.00013, -0.00089, -0.01195, -0.00013, -0.00333, -0.0081, 0.0, -0.00626, -0.01428, 0.0, -0.00152, -0.01459, 0.00234, -0.00152, -0.01459, -0.00234, -0.00391, -0.00713, 4e-05, -0.00391, -0.00713, -4e-05, -0.01195, 0.01227, -0.00033, -0.01195, 0.01227, 0.00033, -0.01227, 0.01195, -0.00033, -0.01227, 0.01195, 0.00033, 0.00713, 0.00391, 4e-05, 0.00713, 0.00391, -4e-05, 0.01715, 0.00403, 0.00023, 0.01715, 0.00403, -0.00023, 0.00748, 0.00021, -0.0001, 0.00748, 0.00021, 0.0001, -0.00021, -0.00748, 0.0001, -0.00021, -0.00748, -0.0001, -0.00403, -0.01715, -0.00023, -0.00403, -0.01715, 0.00023, -0.00391, -0.00713, -4e-05, -0.01195, 0.01227, -0.00033, -0.01195, 0.01227, 0.00033, -0.01227, 0.01195, -0.00033, -0.01227, 0.01195, 0.00033, 0.00713, 0.00391, 4e-05, 0.00713, 0.00391, -4e-05, 0.01715, 0.00403, 0.00023, 0.01715, 0.00403, -0.00023, 0.00748, 0.00021, -0.0001, 0.00748, 0.00021, 0.0001, -0.00021, -0.00748, 0.0001, -0.00021, -0.00748, -0.0001, -0.00403, -0.01715, -0.00023, -0.00403, -0.01715, 0.00023, -0.00391, -0.00713, 4e-05, -0.01463, -0.00137, 0.00074, -0.01463, -0.00137, -0.00074, 0.00137, 0.01463, 0.00074, 0.00137, 0.01463, -0.00074, 0.01275, -0.00235, -0.00051, 0.01275, -0.00235, 0.00051, 0.00235, -0.01275, 0.00051, 0.00235, -0.01275, -0.00051, -0.01463, -0.00137, 0.00074, -0.01463, -0.00137, -0.00074, 0.00137, 0.01463, 0.00074, 0.00137, 0.01463, -0.00074, 0.01275, -0.00235, -0.00051, 0.01275, -0.00235, 0.00051, 0.00235, -0.01275, 0.00051, 0.00235, -0.01275, -0.00051], [0.00098, -0.00841, 0.00707, 0.00098, -0.00841, -0.00707, -0.00114, 0.00227, 0.0, -0.0045, 0.01084, 0.0, -0.00709, 0.00709, -0.00072, -0.00709, 0.00709, 0.00072, 0.00841, -0.00098, 0.00707, 0.00841, -0.00098, -0.00707, -0.01084, 0.0045, 0.0, -0.00227, 0.00114, 0.0, -0.00121, 0.00305, 0.00361, -0.00121, 0.00305, -0.00361, -0.00503, -0.00047, 0.00093, -0.00503, -0.00047, -0.00093, -0.00045, 0.00033, 0.0, -0.00508, 0.00339, -0.0, -0.00148, 0.00148, 0.0, -0.00148, 0.00148, -0.0, 0.00047, 0.00503, -0.00093, 0.00047, 0.00503, 0.00093, -0.00339, 0.00508, 0.0, -0.00033, 0.00045, -0.0, -0.00305, 0.00121, -0.00361, -0.00305, 0.00121, 0.00361, 0.00098, -0.00841, 0.00707, 0.00098, -0.00841, -0.00707, -0.00114, 0.00227, -0.0, -0.0045, 0.01084, -0.0, -0.00709, 0.00709, -0.00072, -0.00709, 0.00709, 0.00072, 0.00841, -0.00098, 0.00707, 0.00841, -0.00098, -0.00707, -0.01084, 0.0045, -0.0, -0.00227, 0.00114, -0.0, -0.00121, 0.00305, 0.00361, -0.00121, 0.00305, -0.00361, -0.00503, -0.00047, 0.00093, -0.00503, -0.00047, -0.00093, -0.00045, 0.00033, -0.0, -0.00508, 0.00339, 0.0, -0.00148, 0.00148, -0.0, -0.00148, 0.00148, 0.0, 0.00047, 0.00503, -0.00093, 0.00047, 0.00503, 0.00093, -0.00339, 0.00508, -0.0, -0.00033, 0.00045, 0.0, -0.00305, 0.00121, -0.00361, -0.00305, 0.00121, 0.00361, -0.0052, 0.00225, 0.00196, -0.0052, 0.00225, -0.00196, -0.00445, 0.00165, 0.00385, -0.00445, 0.00165, -0.00385, -0.00165, 0.00445, 0.00385, -0.00165, 0.00445, -0.00385, -0.00225, 0.0052, 0.00196, -0.00225, 0.0052, -0.00196, 0.00298, 0.00441, 0.00278, 0.00298, 0.00441, -0.00278, -0.00203, 0.00195, -0.00132, -0.00203, 0.00195, 0.00132, -0.00195, 0.00203, 0.00132, -0.00195, 0.00203, -0.00132, -0.00441, -0.00298, -0.00278, -0.00441, -0.00298, 0.00278, -0.0052, 0.00225, -0.00196, -0.00445, 0.00165, 0.00385, -0.00445, 0.00165, -0.00385, -0.00165, 0.00445, 0.00385, -0.00165, 0.00445, -0.00385, -0.00225, 0.0052, 0.00196, -0.00225, 0.0052, -0.00196, 0.00298, 0.00441, 0.00278, 0.00298, 0.00441, -0.00278, -0.00203, 0.00195, -0.00132, -0.00203, 0.00195, 0.00132, -0.00195, 0.00203, 0.00132, -0.00195, 0.00203, -0.00132, -0.00441, -0.00298, -0.00278, -0.00441, -0.00298, 0.00278, -0.0052, 0.00225, 0.00196, 0.01013, -0.0314, 0.01934, 0.01013, -0.0314, -0.01934, 0.0314, -0.01013, 0.01934, 0.0314, -0.01013, -0.01934, -0.00951, -0.00895, -0.00098, -0.00951, -0.00895, 0.00098, 0.00895, 0.00951, 0.00098, 0.00895, 0.00951, -0.00098, 0.01013, -0.0314, 0.01934, 0.01013, -0.0314, -0.01934, 0.0314, -0.01013, 0.01934, 0.0314, -0.01013, -0.01934, -0.00951, -0.00895, -0.00098, -0.00951, -0.00895, 0.00098, 0.00895, 0.00951, 0.00098, 0.00895, 0.00951, -0.00098]]), "E3u1":array([[-0.00973, -0.00403, 0.0, -0.00973, -0.00403, 0.0, 0.0039, -0.00705, -0.0, -0.00222, 0.00774, 0.0, -0.01398, 0.0128, 0.00264, -0.01398, 0.0128, -0.00264, -0.00521, 0.0104, 0.00039, -0.00521, 0.0104, -0.00039, -0.01429, 0.00602, 0.0, -0.00384, 0.00147, 0.0, 0.00788, 0.00239, -0.00109, 0.00788, 0.00239, 0.00109, 0.01075, 0.00445, -0.00026, 0.01075, 0.00445, 0.00026, 0.01222, 0.00426, 0.0, 0.01165, 0.00563, 0.0, 0.00726, 0.00388, -0.00109, 0.00726, 0.00388, 0.00109, 0.00367, -0.01104, 0.00039, 0.00367, -0.01104, -0.00039, -0.00167, -0.00375, 0.0, -0.00585, -0.01436, -0.0, -0.00084, -0.01894, 0.00264, -0.00084, -0.01894, -0.00264, -0.00973, -0.00403, -0.0, -0.00973, -0.00403, -0.0, 0.0039, -0.00705, -0.0, -0.00222, 0.00774, 0.0, -0.01398, 0.0128, 0.00264, -0.01398, 0.0128, -0.00264, -0.00521, 0.0104, 0.00039, -0.00521, 0.0104, -0.00039, -0.01429, 0.00602, 0.0, -0.00384, 0.00147, -0.0, 0.00788, 0.00239, -0.00109, 0.00788, 0.00239, 0.00109, 0.01075, 0.00445, -0.00026, 0.01075, 0.00445, 0.00026, 0.01222, 0.00426, -0.0, 0.01165, 0.00563, -0.0, 0.00726, 0.00388, -0.00109, 0.00726, 0.00388, 0.00109, 0.00367, -0.01104, 0.00039, 0.00367, -0.01104, -0.00039, -0.00167, -0.00375, -0.0, -0.00585, -0.01436, -0.0, -0.00084, -0.01894, 0.00264, -0.00084, -0.01894, -0.00264, -0.00178, -0.0128, 0.00013, -0.00178, -0.0128, -0.00013, -0.01031, 0.00779, -0.00013, -0.01031, 0.00779, 0.00013, -0.0145, 0.01137, -0.00077, -0.0145, 0.01137, 0.00077, 0.0011, 0.0045, 3e-05, 0.0011, 0.0045, -3e-05, 0.01232, 0.00478, -8e-05, 0.01232, 0.00478, 8e-05, 0.01209, 0.00533, -8e-05, 0.01209, 0.00533, 8e-05, 0.00396, -0.0024, 3e-05, 0.00396, -0.0024, -3e-05, -0.00221, -0.01829, -0.00077, -0.00221, -0.01829, 0.00077, -0.00178, -0.0128, -0.00013, -0.01031, 0.00779, -0.00013, -0.01031, 0.00779, 0.00013, -0.0145, 0.01137, -0.00077, -0.0145, 0.01137, 0.00077, 0.0011, 0.0045, 3e-05, 0.0011, 0.0045, -3e-05, 0.01232, 0.00478, -8e-05, 0.01232, 0.00478, 8e-05, 0.01209, 0.00533, -8e-05, 0.01209, 0.00533, 8e-05, 0.00396, -0.0024, 3e-05, 0.00396, -0.0024, -3e-05, -0.00221, -0.01829, -0.00077, -0.00221, -0.01829, 0.00077, -0.00178, -0.0128, 0.00013, -0.01397, -0.00579, 0.0, -0.01397, -0.00579, -0.0, -0.00292, 0.01324, 0.00035, -0.00292, 0.01324, -0.00035, 0.00918, 0.0038, -0.00166, 0.00918, 0.0038, 0.00166, 0.0073, -0.01143, 0.00035, 0.0073, -0.01143, -0.00035, -0.01397, -0.00579, -0.0, -0.01397, -0.00579, 0.0, -0.00292, 0.01324, 0.00035, -0.00292, 0.01324, -0.00035, 0.00918, 0.0038, -0.00166, 0.00918, 0.0038, 0.00166, 0.0073, -0.01143, 0.00035, 0.0073, -0.01143, -0.00035], [-0.00474, -0.00196, -0.0, -0.00474, -0.00196, 0.0, -0.00518, -0.00429, -0.0, -0.0067, -0.00063, 0.0, -0.00103, 0.00432, -0.00257, -0.00103, 0.00432, 0.00257, 0.01075, -0.00354, 0.00643, 0.01075, -0.00354, -0.00643, -0.00741, 0.00219, 0.0, -0.00084, -0.00088, 0.0, -0.00189, -0.00132, 0.0016, -0.00189, -0.00132, -0.0016, -0.0017, -0.0007, 0.00788, -0.0017, -0.0007, -0.00788, -0.00563, -0.00294, 0.0, -0.00606, -0.00191, 0.0, -0.00227, -0.0004, 0.0016, -0.00227, -0.0004, -0.0016, 0.0051, 0.01011, 0.00643, 0.0051, 0.01011, -0.00643, -0.00122, 3e-05, 0.0, -0.00369, -0.00679, 0.0, 0.00233, -0.00379, -0.00257, 0.00233, -0.00379, 0.00257, -0.00474, -0.00196, 0.0, -0.00474, -0.00196, -0.0, -0.00518, -0.00429, -0.0, -0.0067, -0.00063, 0.0, -0.00103, 0.00432, -0.00257, -0.00103, 0.00432, 0.00257, 0.01075, -0.00354, 0.00643, 0.01075, -0.00354, -0.00643, -0.00741, 0.00219, -0.0, -0.00084, -0.00088, -0.0, -0.00189, -0.00132, 0.0016, -0.00189, -0.00132, -0.0016, -0.0017, -0.0007, 0.00788, -0.0017, -0.0007, -0.00788, -0.00563, -0.00294, -0.0, -0.00606, -0.00191, -0.0, -0.00227, -0.0004, 0.0016, -0.00227, -0.0004, -0.0016, 0.0051, 0.01011, 0.00643, 0.0051, 0.01011, -0.00643, -0.00122, 3e-05, -0.0, -0.00369, -0.00679, -0.0, 0.00233, -0.00379, -0.00257, 0.00233, -0.00379, 0.00257, -0.00412, -0.00251, -0.00139, -0.00412, -0.00251, 0.00139, -0.00469, -0.00113, 0.00139, -0.00469, -0.00113, -0.00139, -0.00727, 0.00366, 0.00292, -0.00727, 0.00366, -0.00292, -0.00025, -0.00014, 0.00535, -0.00025, -0.00014, -0.00535, -0.00481, -0.00284, 0.00289, -0.00481, -0.00284, -0.00289, -0.00541, -0.00139, 0.00289, -0.00541, -0.00139, -0.00289, -0.00027, -8e-05, 0.00535, -0.00027, -8e-05, -0.00535, -0.00255, -0.00773, 0.00292, -0.00255, -0.00773, -0.00292, -0.00412, -0.00251, 0.00139, -0.00469, -0.00113, 0.00139, -0.00469, -0.00113, -0.00139, -0.00727, 0.00366, 0.00292, -0.00727, 0.00366, -0.00292, -0.00025, -0.00014, 0.00535, -0.00025, -0.00014, -0.00535, -0.00481, -0.00284, 0.00289, -0.00481, -0.00284, -0.00289, -0.00541, -0.00139, 0.00289, -0.00541, -0.00139, -0.00289, -0.00027, -8e-05, 0.00535, -0.00027, -8e-05, -0.00535, -0.00255, -0.00773, 0.00292, -0.00255, -0.00773, -0.00292, -0.00412, -0.00251, -0.00139, -0.00977, -0.00404, -0.0, -0.00977, -0.00404, 0.0, 0.0288, -0.01074, 0.01761, 0.0288, -0.01074, -0.01761, 0.00939, 0.00389, 0.01649, 0.00939, 0.00389, -0.01649, 0.01277, 0.02796, 0.01761, 0.01277, 0.02796, -0.01761, -0.00977, -0.00404, 0.0, -0.00977, -0.00404, -0.0, 0.0288, -0.01074, 0.01761, 0.0288, -0.01074, -0.01761, 0.00939, 0.00389, 0.01649, 0.00939, 0.00389, -0.01649, 0.01277, 0.02796, 0.01761, 0.01277, 0.02796, -0.01761]]), "E3u2":array([[-0.00999, 0.00532, -0.00028, -0.00999, 0.00532, 0.00028, -0.0016, 0.00289, 0.0, -0.00596, 0.0134, 0.0, -0.01246, 0.01352, 0.00312, -0.01246, 0.01352, -0.00312, 0.0043, 0.01038, -0.0, 0.0043, 0.01038, -0.0, -0.00735, 0.00278, 0.0, 0.00716, -0.00323, -0.0, 0.01838, 0.00075, -0.00312, 0.01838, 0.00075, 0.00312, 0.01082, -0.0033, 0.00028, 0.01082, -0.0033, -0.00028, 0.01369, 0.00526, -0.0, 0.00317, 0.00091, -0.0, -0.00348, -0.00748, 0.0011, -0.00348, -0.00748, -0.0011, -0.00466, -0.01126, -0.00014, -0.00466, -0.01126, 0.00014, -0.00427, -0.01197, 0.0, -0.00545, -0.01148, 0.0, -0.00283, -0.00774, 0.0011, -0.00283, -0.00774, -0.0011, -0.00999, 0.00532, -0.00028, -0.00999, 0.00532, 0.00028, -0.0016, 0.00289, 0.0, -0.00596, 0.0134, 0.0, -0.01246, 0.01352, 0.00312, -0.01246, 0.01352, -0.00312, 0.0043, 0.01038, 0.0, 0.0043, 0.01038, 0.0, -0.00735, 0.00278, 0.0, 0.00716, -0.00323, -0.0, 0.01838, 0.00075, -0.00312, 0.01838, 0.00075, 0.00312, 0.01082, -0.0033, 0.00028, 0.01082, -0.0033, -0.00028, 0.01369, 0.00526, -0.0, 0.00317, 0.00091, -0.0, -0.00348, -0.00748, 0.0011, -0.00348, -0.00748, -0.0011, -0.00466, -0.01126, -0.00014, -0.00466, -0.01126, 0.00014, -0.00427, -0.01197, -0.0, -0.00545, -0.01148, -0.0, -0.00283, -0.00774, 0.0011, -0.00283, -0.00774, -0.0011, -0.00531, -0.00045, 3e-05, -0.00531, -0.00045, -3e-05, -0.01123, 0.01425, 0.00018, -0.01123, 0.01425, -0.00018, -0.00814, 0.0104, -0.00023, -0.00814, 0.0104, 0.00023, 0.01311, 0.0016, 0.00023, 0.01311, 0.0016, -0.00023, 0.01802, 0.00214, -0.00018, 0.01802, 0.00214, 0.00018, 0.00344, -0.00407, -3e-05, 0.00344, -0.00407, 3e-05, -0.00452, -0.01155, -0.00026, -0.00452, -0.01155, 0.00026, -0.00497, -0.01136, -0.00026, -0.00497, -0.01136, 0.00026, -0.00531, -0.00045, -3e-05, -0.01123, 0.01425, 0.00018, -0.01123, 0.01425, -0.00018, -0.00814, 0.0104, -0.00023, -0.00814, 0.0104, 0.00023, 0.01311, 0.0016, 0.00023, 0.01311, 0.0016, -0.00023, 0.01802, 0.00214, -0.00018, 0.01802, 0.00214, 0.00018, 0.00344, -0.00407, -3e-05, 0.00344, -0.00407, 3e-05, -0.00452, -0.01155, -0.00026, -0.00452, -0.01155, 0.00026, -0.00497, -0.01136, -0.00026, -0.00497, -0.01136, 0.00026, -0.00531, -0.00045, 3e-05, -0.01402, 0.00447, 9e-05, -0.01402, 0.00447, -9e-05, 0.00567, 0.01368, 0.0, 0.00567, 0.01368, -0.0, 0.01307, -0.00675, -9e-05, 0.01307, -0.00675, 9e-05, -0.00491, -0.01184, -0.00013, -0.00491, -0.01184, 0.00013, -0.01402, 0.00447, 9e-05, -0.01402, 0.00447, -9e-05, 0.00567, 0.01368, -0.0, 0.00567, 0.01368, 0.0, 0.01307, -0.00675, -9e-05, 0.01307, -0.00675, 9e-05, -0.00491, -0.01184, -0.00013, -0.00491, -0.01184, 0.00013], [0.00373, -0.00215, 0.00723, 0.00373, -0.00215, -0.00723, -8e-05, 0.00403, 0.0, 0.00047, 0.00715, 0.0, -0.00308, -0.00106, 0.00159, -0.00308, -0.00106, -0.00159, 0.00015, 0.00037, -0.0, 0.00015, 0.00037, -0.0, 0.00239, 0.00146, 0.0, -0.00066, 0.00272, -0.0, 0.00142, -0.00293, -0.00159, 0.00142, -0.00293, 0.00159, -0.00416, 0.00112, -0.00723, -0.00416, 0.00112, 0.00723, 0.00472, 0.00538, -0.0, 0.0029, 0.00279, -0.0, 0.00469, 0.00035, 0.00071, 0.00469, 0.00035, -0.00071, -0.00242, -0.00585, -0.01089, -0.00242, -0.00585, 0.01089, 0.00722, 0.01049, -0.0, 0.00232, 0.01252, -0.0, -0.00307, 0.00357, 0.00071, -0.00307, 0.00357, -0.00071, 0.00373, -0.00215, 0.00723, 0.00373, -0.00215, -0.00723, -8e-05, 0.00403, -0.0, 0.00047, 0.00715, -0.0, -0.00308, -0.00106, 0.00159, -0.00308, -0.00106, -0.00159, 0.00015, 0.00037, 0.0, 0.00015, 0.00037, 0.0, 0.00239, 0.00146, -0.0, -0.00066, 0.00272, 0.0, 0.00142, -0.00293, -0.00159, 0.00142, -0.00293, 0.00159, -0.00416, 0.00112, -0.00723, -0.00416, 0.00112, 0.00723, 0.00472, 0.00538, 0.0, 0.0029, 0.00279, 0.0, 0.00469, 0.00035, 0.00071, 0.00469, 0.00035, -0.00071, -0.00242, -0.00585, -0.01089, -0.00242, -0.00585, 0.01089, 0.00722, 0.01049, 0.0, 0.00232, 0.01252, 0.0, -0.00307, 0.00357, 0.00071, -0.00307, 0.00357, -0.00071, 0.00053, 0.00586, 0.00241, 0.00053, 0.00586, -0.00241, -0.0005, 0.00543, 0.00243, -0.0005, 0.00543, -0.00243, 0.00058, -0.00065, -0.00123, 0.00058, -0.00065, 0.00123, -0.00087, -4e-05, 0.00123, -0.00087, -4e-05, -0.00123, 0.0042, 0.00348, -0.00243, 0.0042, 0.00348, 0.00243, 0.00377, 0.00452, -0.00241, 0.00377, 0.00452, 0.00241, 0.0036, 0.00887, -0.00521, 0.0036, 0.00887, 0.00521, 0.00373, 0.00882, -0.00521, 0.00373, 0.00882, 0.00521, 0.00053, 0.00586, -0.00241, -0.0005, 0.00543, 0.00243, -0.0005, 0.00543, -0.00243, 0.00058, -0.00065, -0.00123, 0.00058, -0.00065, 0.00123, -0.00087, -4e-05, 0.00123, -0.00087, -4e-05, -0.00123, 0.0042, 0.00348, -0.00243, 0.0042, 0.00348, 0.00243, 0.00377, 0.00452, -0.00241, 0.00377, 0.00452, 0.00241, 0.0036, 0.00887, -0.00521, 0.0036, 0.00887, 0.00521, 0.00373, 0.00882, -0.00521, 0.00373, 0.00882, 0.00521, 0.00053, 0.00586, 0.00241, 0.00863, -0.01282, 0.01353, 0.00863, -0.01282, -0.01353, -4e-05, -0.0001, 0.0, -4e-05, -0.0001, 0.0, -0.01517, -0.00297, -0.01353, -0.01517, -0.00297, 0.01353, -0.01537, -0.0371, -0.02716, -0.01537, -0.0371, 0.02716, 0.00863, -0.01282, 0.01353, 0.00863, -0.01282, -0.01353, -4e-05, -0.0001, -0.0, -4e-05, -0.0001, -0.0, -0.01517, -0.00297, -0.01353, -0.01517, -0.00297, 0.01353, -0.01537, -0.0371, -0.02716, -0.01537, -0.0371, 0.02716]]), "E3u3":array([[-0.00491, 0.01184, -0.00084, -0.00491, 0.01184, 0.00084, -0.00427, 0.01155, 0.0, -0.00515, 0.01118, 0.0, -0.00273, 0.00773, 0.00218, -0.00273, 0.00773, -0.00218, 0.01063, 0.00452, 0.00057, 0.01063, 0.00452, -0.00057, 0.00288, -0.00063, 0.0, 0.01172, -0.00584, 0.0, 0.01813, -0.00228, -0.0026, 0.01813, -0.00228, 0.0026, 0.0038, -0.00918, -0.0, 0.0038, -0.00918, 0.0, 0.00694, 0.00232, 0.0, -0.00655, -0.00327, -0.0, -0.0112, -0.01444, 0.0026, -0.0112, -0.01444, -0.0026, -0.01071, -0.00432, -0.00057, -0.01071, -0.00432, 0.00057, -0.00415, -0.01242, -0.0, -0.00159, -0.00248, -0.0, -0.00354, 0.0074, -0.00218, -0.00354, 0.0074, 0.00218, -0.00491, 0.01184, -0.00084, -0.00491, 0.01184, 0.00084, -0.00427, 0.01155, -0.0, -0.00515, 0.01118, -0.0, -0.00273, 0.00773, 0.00218, -0.00273, 0.00773, -0.00218, 0.01063, 0.00452, 0.00057, 0.01063, 0.00452, -0.00057, 0.00288, -0.00063, -0.0, 0.01172, -0.00584, 0.0, 0.01813, -0.00228, -0.0026, 0.01813, -0.00228, 0.0026, 0.0038, -0.00918, 0.0, 0.0038, -0.00918, -0.0, 0.00694, 0.00232, 0.0, -0.00655, -0.00327, -0.0, -0.0112, -0.01444, 0.0026, -0.0112, -0.01444, -0.0026, -0.01071, -0.00432, -0.00057, -0.01071, -0.00432, 0.00057, -0.00415, -0.01242, -0.0, -0.00159, -0.00248, 0.0, -0.00354, 0.0074, -0.00218, -0.00354, 0.0074, 0.00218, -0.00533, 0.01153, 1e-05, -0.00533, 0.01153, -1e-05, -0.00439, 0.01192, 1e-05, -0.00439, 0.01192, -1e-05, 0.00235, 0.0037, 0.00018, 0.00235, 0.0037, -0.00018, 0.01692, -0.00273, 0.00025, 0.01692, -0.00273, -0.00025, 0.01285, -0.00214, -0.00027, 0.01285, -0.00214, 0.00027, -0.00757, -0.01059, 0.00027, -0.00757, -0.01059, -0.00027, -0.01003, -0.0139, -0.00025, -0.01003, -0.0139, 0.00025, -0.00428, 0.00096, -0.00018, -0.00428, 0.00096, 0.00018, -0.00533, 0.01153, -1e-05, -0.00439, 0.01192, 1e-05, -0.00439, 0.01192, -1e-05, 0.00235, 0.0037, 0.00018, 0.00235, 0.0037, -0.00018, 0.01692, -0.00273, 0.00025, 0.01692, -0.00273, -0.00025, 0.01285, -0.00214, -0.00027, 0.01285, -0.00214, 0.00027, -0.00757, -0.01059, 0.00027, -0.00757, -0.01059, -0.00027, -0.01003, -0.0139, -0.00025, -0.01003, -0.0139, 0.00025, -0.00428, 0.00096, -0.00018, -0.00428, 0.00096, 0.00018, -0.00533, 0.01153, 1e-05, -0.00595, 0.01436, -0.0011, -0.00595, 0.01436, 0.0011, 0.0132, 0.00557, 0.00091, 0.0132, 0.00557, -0.00091, 0.00612, -0.01477, 0.0, 0.00612, -0.01477, -0.0, -0.01328, -0.0054, -0.00091, -0.01328, -0.0054, 0.00091, -0.00595, 0.01436, -0.0011, -0.00595, 0.01436, 0.0011, 0.0132, 0.00557, 0.00091, 0.0132, 0.00557, -0.00091, 0.00612, -0.01477, -0.0, 0.00612, -0.01477, 0.0, -0.01328, -0.0054, -0.00091, -0.01328, -0.0054, 0.00091], [0.00298, -0.0072, 0.0046, 0.00298, -0.0072, -0.0046, -0.00149, 0.00962, -0.0, -0.00575, 0.00786, -0.0, -0.00638, 0.00561, -0.00239, -0.00638, 0.00561, 0.00239, -0.00525, 0.00389, -0.00346, -0.00525, 0.00389, 0.00346, -5e-05, 7e-05, -0.0, 0.00594, 0.00132, 0.0, 0.00431, -0.00108, 0.00072, 0.00431, -0.00108, -0.00072, 0.0011, -0.00267, 0.0, 0.0011, -0.00267, 0.0, 0.00331, -0.00533, 0.0, 0.00143, -0.00611, -0.0, -0.00228, -0.00381, -0.00072, -0.00228, -0.00381, 0.00072, 0.00096, 0.00646, 0.00346, 0.00096, 0.00646, -0.00346, -0.00514, -0.00326, -0.0, -2e-05, 9e-05, 0.0, 0.00054, 0.00848, 0.00239, 0.00054, 0.00848, -0.00239, 0.00298, -0.0072, 0.0046, 0.00298, -0.0072, -0.0046, -0.00149, 0.00962, 0.0, -0.00575, 0.00786, 0.0, -0.00638, 0.00561, -0.00239, -0.00638, 0.00561, 0.00239, -0.00525, 0.00389, -0.00346, -0.00525, 0.00389, 0.00346, -5e-05, 7e-05, 0.0, 0.00594, 0.00132, 0.0, 0.00431, -0.00108, 0.00072, 0.00431, -0.00108, -0.00072, 0.0011, -0.00267, -0.0, 0.0011, -0.00267, -0.0, 0.00331, -0.00533, -0.0, 0.00143, -0.00611, 0.0, -0.00228, -0.00381, -0.00072, -0.00228, -0.00381, 0.00072, 0.00096, 0.00646, 0.00346, 0.00096, 0.00646, -0.00346, -0.00514, -0.00326, -0.0, -2e-05, 9e-05, -0.0, 0.00054, 0.00848, 0.00239, 0.00054, 0.00848, -0.00239, -0.00054, 0.00826, -0.00209, -0.00054, 0.00826, 0.00209, -0.00546, 0.00622, -0.00209, -0.00546, 0.00622, 0.00209, 0.00478, 0.00179, -0.00068, 0.00478, 0.00179, 0.00068, 0.00599, -0.00188, 0.0001, 0.00599, -0.00188, -0.0001, 0.00534, -0.00185, 2e-05, 0.00534, -0.00185, -2e-05, -0.00246, -0.00508, -2e-05, -0.00246, -0.00508, 2e-05, -0.00291, -0.00557, -0.0001, -0.00291, -0.00557, 0.0001, -0.00464, -0.00211, 0.00068, -0.00464, -0.00211, -0.00068, -0.00054, 0.00826, 0.00209, -0.00546, 0.00622, -0.00209, -0.00546, 0.00622, 0.00209, 0.00478, 0.00179, -0.00068, 0.00478, 0.00179, 0.00068, 0.00599, -0.00188, 0.0001, 0.00599, -0.00188, -0.0001, 0.00534, -0.00185, 2e-05, 0.00534, -0.00185, -2e-05, -0.00246, -0.00508, -2e-05, -0.00246, -0.00508, 2e-05, -0.00291, -0.00557, -0.0001, -0.00291, -0.00557, 0.0001, -0.00464, -0.00211, 0.00068, -0.00464, -0.00211, -0.00068, -0.00054, 0.00826, -0.00209, 0.01369, -0.03304, 0.01705, 0.01369, -0.03304, -0.01705, -0.02094, 0.00923, -0.01406, -0.02094, 0.00923, 0.01406, 0.00377, -0.00911, 0.0, 0.00377, -0.00911, -0.0, 0.00828, 0.02133, 0.01406, 0.00828, 0.02133, -0.01406, 0.01369, -0.03304, 0.01705, 0.01369, -0.03304, -0.01705, -0.02094, 0.00923, -0.01406, -0.02094, 0.00923, 0.01406, 0.00377, -0.00911, -0.0, 0.00377, -0.00911, 0.0, 0.00828, 0.02133, 0.01406, 0.00828, 0.02133, -0.01406]]), "E3u4":array([[0.0035, 0.01089, -0.00015, 0.0035, 0.01089, 0.00015, -0.00518, 0.01314, 0.0, -0.00304, 0.00294, -0.0, 0.00819, -0.00281, -0.00018, 0.00819, -0.00281, 0.00018, 0.01089, -0.00451, 0.00017, 0.01089, -0.00451, -0.00017, 0.01168, -0.00457, -0.0, 0.01148, -0.00503, -0.0, 0.00777, -0.0038, -0.00018, 0.00777, -0.0038, 0.00018, -0.00522, -0.01018, -0.00015, -0.00522, -0.01018, 0.00015, -0.00423, 8e-05, -0.0, -0.01296, -0.00562, -0.0, -0.01433, -0.0125, 0.00296, -0.01433, -0.0125, -0.00296, -0.01055, 0.00437, -0.0, -0.01055, 0.00437, 0.0, -0.00282, -0.00686, -0.0, 0.00286, 0.00685, 0.0, -0.0013, 0.01897, -0.00296, -0.0013, 0.01897, 0.00296, 0.0035, 0.01089, -0.00015, 0.0035, 0.01089, 0.00015, -0.00518, 0.01314, 0.0, -0.00304, 0.00294, 0.0, 0.00819, -0.00281, -0.00018, 0.00819, -0.00281, 0.00018, 0.01089, -0.00451, 0.00017, 0.01089, -0.00451, -0.00017, 0.01168, -0.00457, 0.0, 0.01148, -0.00503, 0.0, 0.00777, -0.0038, -0.00018, 0.00777, -0.0038, 0.00018, -0.00522, -0.01018, -0.00015, -0.00522, -0.01018, 0.00015, -0.00423, 8e-05, 0.0, -0.01296, -0.00562, -0.0, -0.01433, -0.0125, 0.00296, -0.01433, -0.0125, -0.00296, -0.01055, 0.00437, 0.0, -0.01055, 0.00437, -0.0, -0.00282, -0.00686, -0.0, 0.00286, 0.00685, 0.0, -0.0013, 0.01897, -0.00296, -0.0013, 0.01897, 0.00296, -0.0023, 0.01716, -2e-05, -0.0023, 0.01716, 2e-05, 0.00393, 0.00263, -9e-05, 0.00393, 0.00263, 9e-05, 0.01213, -0.00532, 0.0006, 0.01213, -0.00532, -0.0006, 0.01234, -0.00481, 0.0006, 0.01234, -0.00481, -0.0006, 0.00092, -0.00464, -9e-05, 0.00092, -0.00464, 9e-05, -0.01376, -0.01051, -2e-05, -0.01376, -0.01051, 2e-05, -0.01025, -0.00875, -0.00038, -0.01025, -0.00875, 0.00038, -0.00106, 0.01344, 0.00038, -0.00106, 0.01344, -0.00038, -0.0023, 0.01716, 2e-05, 0.00393, 0.00263, -9e-05, 0.00393, 0.00263, 9e-05, 0.01213, -0.00532, 0.0006, 0.01213, -0.00532, -0.0006, 0.01234, -0.00481, 0.0006, 0.01234, -0.00481, -0.0006, 0.00092, -0.00464, -9e-05, 0.00092, -0.00464, 9e-05, -0.01376, -0.01051, -2e-05, -0.01376, -0.01051, 2e-05, -0.01025, -0.00875, -0.00038, -0.01025, -0.00875, 0.00038, -0.00106, 0.01344, 0.00038, -0.00106, 0.01344, -0.00038, -0.0023, 0.01716, -2e-05, 0.00678, 0.01222, 0.00026, 0.00678, 0.01222, -0.00026, 0.01173, -0.00486, 0.00078, 0.01173, -0.00486, -0.00078, -0.00385, -0.01343, 0.00026, -0.00385, -0.01343, -0.00026, -0.01442, 0.00597, 0.0, -0.01442, 0.00597, -0.0, 0.00678, 0.01222, 0.00026, 0.00678, 0.01222, -0.00026, 0.01173, -0.00486, 0.00078, 0.01173, -0.00486, -0.00078, -0.00385, -0.01343, 0.00026, -0.00385, -0.01343, -0.00026, -0.01442, 0.00597, -0.0, -0.01442, 0.00597, 0.0], [0.00591, -0.00241, 0.00078, 0.00591, -0.00241, -0.00078, 0.00116, 0.00498, 0.0, 0.00402, 0.00172, 0.0, -0.00325, -0.0019, -0.00383, -0.00325, -0.0019, 0.00383, -0.01326, 0.00549, -0.0122, -0.01326, 0.00549, 0.0122, 0.00588, -0.00461, 0.0, 0.00742, -0.0009, 0.0, -0.00096, 0.00364, -0.00383, -0.00096, 0.00364, 0.00383, 0.00589, -0.00247, 0.00078, 0.00589, -0.00247, -0.00078, 0.00162, -0.00406, 0.0, -0.0027, -0.00434, 0.0, -0.00058, -0.00614, -0.00449, -0.00058, -0.00614, 0.00449, 0.00041, -0.00017, 0.0, 0.00041, -0.00017, 0.0, 0.00179, -0.00628, -0.0, 0.00571, 0.00318, 0.0, 0.00393, 0.00475, 0.00449, 0.00393, 0.00475, -0.00449, 0.00591, -0.00241, 0.00078, 0.00591, -0.00241, -0.00078, 0.00116, 0.00498, -0.0, 0.00402, 0.00172, -0.0, -0.00325, -0.0019, -0.00383, -0.00325, -0.0019, 0.00383, -0.01326, 0.00549, -0.0122, -0.01326, 0.00549, 0.0122, 0.00588, -0.00461, -0.0, 0.00742, -0.0009, -0.0, -0.00096, 0.00364, -0.00383, -0.00096, 0.00364, 0.00383, 0.00589, -0.00247, 0.00078, 0.00589, -0.00247, -0.00078, 0.00162, -0.00406, -0.0, -0.0027, -0.00434, -0.0, -0.00058, -0.00614, -0.00449, -0.00058, -0.00614, 0.00449, 0.00041, -0.00017, -0.0, 0.00041, -0.00017, -0.0, 0.00179, -0.00628, -0.0, 0.00571, 0.00318, 0.0, 0.00393, 0.00475, 0.00449, 0.00393, 0.00475, -0.00449, 0.00453, 0.00612, -0.00138, 0.00453, 0.00612, 0.00138, -0.00029, 0.00307, -0.0023, -0.00029, 0.00307, 0.0023, 0.0065, -0.00136, -0.00614, 0.0065, -0.00136, 0.00614, 0.00556, -0.00363, -0.00614, 0.00556, -0.00363, 0.00614, -0.00238, -0.00196, -0.0023, -0.00238, -0.00196, 0.0023, -0.00113, -0.00753, -0.00138, -0.00113, -0.00753, 0.00138, -0.00105, -0.00601, -0.00365, -0.00105, -0.00601, 0.00365, 0.0035, 0.00499, 0.00365, 0.0035, 0.00499, -0.00365, 0.00453, 0.00612, 0.00138, -0.00029, 0.00307, -0.0023, -0.00029, 0.00307, 0.0023, 0.0065, -0.00136, -0.00614, 0.0065, -0.00136, 0.00614, 0.00556, -0.00363, -0.00614, 0.00556, -0.00363, 0.00614, -0.00238, -0.00196, -0.0023, -0.00238, -0.00196, 0.0023, -0.00113, -0.00753, -0.00138, -0.00113, -0.00753, 0.00138, -0.00105, -0.00601, -0.00365, -0.00105, -0.00601, 0.00365, 0.0035, 0.00499, 0.00365, 0.0035, 0.00499, -0.00365, 0.00453, 0.00612, -0.00138, 0.00847, -0.00741, 0.00203, 0.00847, -0.00741, -0.00203, -0.04144, 0.01716, -0.03263, -0.04144, 0.01716, 0.03263, 0.01123, -0.00075, 0.00203, 0.01123, -0.00075, -0.00203, -0.00403, 0.00167, -0.0, -0.00403, 0.00167, -0.0, 0.00847, -0.00741, 0.00203, 0.00847, -0.00741, -0.00203, -0.04144, 0.01716, -0.03263, -0.04144, 0.01716, 0.03263, 0.01123, -0.00075, 0.00203, 0.01123, -0.00075, -0.00203, -0.00403, 0.00167, 0.0, -0.00403, 0.00167, 0.0]])}, database_path = 'cb8_scsd_20220713.pkl', maxdist = 1.75, mondrian_limits = (0, 0))
model_objs_dict['porphyrin_pdb'] = scsd_model('porphyrin_pdb', array([[-1.09, 2.827, 0.0], [-0.6767, 4.203, 0.0], [0.6767, 4.203, 0.0], [1.09, 2.827, 0.0], [2.413, 2.413, 0.0], [2.827, 1.09, 0.0], [4.203, 0.6767, 0.0], [4.203, -0.6767, 0.0], [2.827, -1.09, 0.0], [2.413, -2.413, 0.0], [1.09, -2.827, 0.0], [0.6767, -4.203, 0.0], [-0.6767, -4.203, 0.0], [-1.09, -2.827, 0.0], [-2.413, -2.413, 0.0], [-2.827, -1.09, 0.0], [-4.203, -0.6767, 0.0], [-4.203, 0.6767, 0.0], [-2.827, 1.09, 0.0], [-2.413, 2.413, 0.0], [0.0, 1.998, 0.0], [1.998, 0.0, 0.0], [0.0, -1.998, 0.0], [-1.998, 0.0, 0.0]]), 'D4h' ,  database_path = 'porphyrin_pdb_scsd_20220713.pkl', maxdist = 1.75)
#pca = {"A1g":array([[0.01504, -0.04918, -0.0, 0.01464, -0.05999, -0.0, -0.01464, -0.05999, -0.0, -0.01504, -0.04918, -0.0, -0.02667, -0.02667, -0.0, -0.04918, -0.01504, -0.0, -0.05999, -0.01464, -0.0, -0.05999, 0.01464, 0.0, -0.04918, 0.01504, 0.0, -0.02667, 0.02667, 0.0, -0.01504, 0.04918, 0.0, -0.01464, 0.05999, 0.0, 0.01464, 0.05999, 0.0, 0.01504, 0.04918, 0.0, 0.02667, 0.02667, 0.0, 0.04918, 0.01504, 0.0, 0.05999, 0.01464, 0.0, 0.05999, -0.01464, -0.0, 0.04918, -0.01504, -0.0, 0.02667, -0.02667, -0.0, -0.0, -0.05398, -0.0, -0.05398, 0.0, 0.0, 0.0, 0.05398, 0.0, 0.05398, -0.0, -0.0], [0.05456, 0.00181, 0.0, -0.00806, 0.04729, 0.0, 0.00806, 0.04729, 0.0, -0.05456, 0.00181, 0.0, -0.04334, -0.04334, -0.0, 0.00181, -0.05456, -0.0, 0.04729, 0.00806, 0.0, 0.04729, -0.00806, -0.0, 0.00181, 0.05456, 0.0, -0.04334, 0.04334, 0.0, -0.05456, -0.00181, -0.0, 0.00806, -0.04729, -0.0, -0.00806, -0.04729, -0.0, 0.05456, -0.00181, -0.0, 0.04334, 0.04334, 0.0, -0.00181, 0.05456, 0.0, -0.04729, -0.00806, -0.0, -0.04729, 0.00806, 0.0, -0.00181, -0.05456, -0.0, 0.04334, -0.04334, -0.0, -0.0, -0.03953, -0.0, -0.03953, -0.0, 0.0, 0.0, 0.03953, 0.0, 0.03953, 0.0, -0.0]]), "A2g":array([[0.03025, 0.00906, 0.0, -0.06336, 0.00616, 0.0, -0.06336, -0.00616, 0.0, 0.03025, -0.00906, 0.0, 0.0389, -0.0389, -0.0, 0.00906, -0.03025, -0.0, 0.00616, 0.06336, 0.0, -0.00616, 0.06336, -0.0, -0.00906, -0.03025, 0.0, -0.0389, -0.0389, 0.0, -0.03025, -0.00906, -0.0, 0.06336, -0.00616, -0.0, 0.06336, 0.00616, -0.0, -0.03025, 0.00906, -0.0, -0.0389, 0.0389, 0.0, -0.00906, 0.03025, 0.0, -0.00616, -0.06336, -0.0, 0.00616, -0.06336, 0.0, 0.00906, 0.03025, -0.0, 0.0389, 0.0389, -0.0, 0.07299, 0.0, 0.0, 0.0, -0.07299, 0.0, -0.07299, -0.0, -0.0, -0.0, 0.07299, -0.0], [-0.01839, 0.05386, -0.0, -0.01025, 0.0646, -0.0, -0.01025, -0.0646, -0.0, -0.01839, -0.05386, -0.0, 0.01445, -0.01445, -0.0, 0.05386, 0.01839, -0.0, 0.0646, 0.01025, -0.0, -0.0646, 0.01025, 0.0, -0.05386, 0.01839, 0.0, -0.01445, -0.01445, 0.0, 0.01839, -0.05386, 0.0, 0.01025, -0.0646, 0.0, 0.01025, 0.0646, 0.0, 0.01839, 0.05386, 0.0, -0.01445, 0.01445, 0.0, -0.05386, -0.01839, 0.0, -0.0646, -0.01025, 0.0, 0.0646, -0.01025, -0.0, 0.05386, -0.01839, -0.0, 0.01445, 0.01445, -0.0, -0.04223, -0.0, -0.0, 0.0, 0.04223, -0.0, 0.04223, 0.0, 0.0, -0.0, -0.04223, 0.0]]), "B1g":array([[0.00695, 0.05388, 0.0, -0.00205, 0.06148, 0.0, 0.00205, 0.06148, 0.0, -0.00695, 0.05388, 0.0, -0.02272, 0.02272, 0.0, -0.05388, 0.00695, 0.0, -0.06148, -0.00205, 0.0, -0.06148, 0.00205, 0.0, -0.05388, -0.00695, 0.0, -0.02272, -0.02272, -0.0, -0.00695, -0.05388, -0.0, 0.00205, -0.06148, -0.0, -0.00205, -0.06148, -0.0, 0.00695, -0.05388, -0.0, 0.02272, -0.02272, -0.0, 0.05388, -0.00695, -0.0, 0.06148, 0.00205, -0.0, 0.06148, -0.00205, -0.0, 0.05388, 0.00695, -0.0, 0.02272, 0.02272, 0.0, 0.0, 0.0607, 0.0, -0.0607, 0.0, 0.0, -0.0, -0.0607, -0.0, 0.0607, -0.0, -0.0], [-0.03969, -0.00148, 0.0, -0.08155, 0.00599, 0.0, 0.08155, 0.00599, -0.0, 0.03969, -0.00148, -0.0, 0.03099, -0.03099, -0.0, 0.00148, -0.03969, -0.0, -0.00599, -0.08155, -0.0, -0.00599, 0.08155, 0.0, 0.00148, 0.03969, 0.0, 0.03099, 0.03099, 0.0, 0.03969, 0.00148, -0.0, 0.08155, -0.00599, -0.0, -0.08155, -0.00599, 0.0, -0.03969, 0.00148, 0.0, -0.03099, 0.03099, 0.0, -0.00148, 0.03969, 0.0, 0.00599, 0.08155, 0.0, 0.00599, -0.08155, -0.0, -0.00148, -0.03969, -0.0, -0.03099, -0.03099, -0.0, 0.0, 0.0173, 0.0, -0.0173, 0.0, 0.0, -0.0, -0.0173, -0.0, 0.0173, -0.0, -0.0]]), "B2g":array([[0.04521, -0.0279, -0.0, 0.01931, -0.01779, -0.0, 0.01931, 0.01779, 0.0, 0.04521, 0.0279, 0.0, 0.05217, 0.05217, 0.0, 0.0279, 0.04521, 0.0, 0.01779, 0.01931, 0.0, -0.01779, 0.01931, 0.0, -0.0279, 0.04521, 0.0, -0.05217, 0.05217, 0.0, -0.04521, 0.0279, 0.0, -0.01931, 0.01779, 0.0, -0.01931, -0.01779, -0.0, -0.04521, -0.0279, -0.0, -0.05217, -0.05217, -0.0, -0.0279, -0.04521, -0.0, -0.01779, -0.01931, -0.0, 0.01779, -0.01931, -0.0, 0.0279, -0.04521, -0.0, 0.05217, -0.05217, -0.0, 0.0622, -0.0, -0.0, 0.0, 0.0622, 0.0, -0.0622, 0.0, 0.0, -0.0, -0.0622, -0.0], [0.00988, 0.01698, 0.0, 0.09745, 0.01369, 0.0, 0.09745, -0.01369, -0.0, 0.00988, -0.01698, -0.0, -0.01858, -0.01858, -0.0, -0.01698, 0.00988, 0.0, -0.01369, 0.09745, 0.0, 0.01369, 0.09745, 0.0, 0.01698, 0.00988, 0.0, 0.01858, -0.01858, -0.0, -0.00988, -0.01698, -0.0, -0.09745, -0.01369, -0.0, -0.09745, 0.01369, 0.0, -0.00988, 0.01698, 0.0, 0.01858, 0.01858, 0.0, 0.01698, -0.00988, -0.0, 0.01369, -0.09745, -0.0, -0.01369, -0.09745, -0.0, -0.01698, -0.00988, -0.0, -0.01858, 0.01858, 0.0, -0.02063, -0.0, -0.0, -0.0, -0.02063, -0.0, 0.02063, 0.0, 0.0, 0.0, 0.02063, 0.0]]), "A1u":array([[0.0, -0.0, 0.06867, -0.0, 0.0, 0.10613, -0.0, -0.0, -0.10613, -0.0, 0.0, -0.06867, -0.0, -0.0, 0.0, -0.0, -0.0, 0.06867, -0.0, 0.0, 0.10613, 0.0, -0.0, -0.10613, -0.0, 0.0, -0.06867, -0.0, 0.0, 0.0, 0.0, 0.0, 0.06867, 0.0, -0.0, 0.10613, 0.0, 0.0, -0.10613, 0.0, -0.0, -0.06867, 0.0, 0.0, -0.0, 0.0, 0.0, 0.06867, 0.0, -0.0, 0.10613, -0.0, 0.0, -0.10613, 0.0, -0.0, -0.06867, 0.0, -0.0, -0.0, 0.0, 0.0, -0.0, 0.0, 0.0, -0.0, -0.0, -0.0, 0.0, -0.0, -0.0, 0.0], [0.0, -0.0, 0.1097, -0.0, -0.0, -0.07098, 0.0, -0.0, 0.07098, -0.0, 0.0, -0.1097, 0.0, -0.0, -0.0, 0.0, -0.0, 0.1097, -0.0, -0.0, -0.07098, -0.0, -0.0, 0.07098, -0.0, 0.0, -0.1097, -0.0, 0.0, -0.0, -0.0, -0.0, 0.1097, 0.0, 0.0, -0.07098, -0.0, 0.0, 0.07098, 0.0, -0.0, -0.1097, -0.0, 0.0, 0.0, -0.0, 0.0, 0.1097, 0.0, 0.0, -0.07098, 0.0, 0.0, 0.07098, 0.0, -0.0, -0.1097, 0.0, -0.0, 0.0, -0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -0.0, -0.0, -0.0, -0.0, -0.0]]), "A2u":array([[-0.0, -0.0, -0.02195, 0.0, 0.0, 0.09661, 0.0, 0.0, 0.09661, -0.0, 0.0, -0.02195, 0.0, 0.0, -0.04898, 0.0, -0.0, -0.02195, 0.0, -0.0, 0.09661, 0.0, 0.0, 0.09661, -0.0, 0.0, -0.02195, 0.0, -0.0, -0.04898, -0.0, 0.0, -0.02195, -0.0, -0.0, 0.09661, -0.0, -0.0, 0.09661, 0.0, -0.0, -0.02195, -0.0, -0.0, -0.04898, -0.0, 0.0, -0.02195, -0.0, -0.0, 0.09661, -0.0, -0.0, 0.09661, 0.0, -0.0, -0.02195, -0.0, 0.0, -0.04898, -0.0, 0.0, -0.10035, 0.0, 0.0, -0.10035, -0.0, -0.0, -0.10035, -0.0, 0.0, -0.10035], [-0.0, 0.0, -0.02748, -0.0, -0.0, 0.02772, 0.0, -0.0, 0.02772, 0.0, -0.0, -0.02748, -0.0, -0.0, -0.12869, -0.0, 0.0, -0.02748, 0.0, 0.0, 0.02772, -0.0, 0.0, 0.02772, 0.0, 0.0, -0.02748, -0.0, -0.0, -0.12869, -0.0, 0.0, -0.02748, 0.0, 0.0, 0.02772, -0.0, 0.0, 0.02772, -0.0, 0.0, -0.02748, 0.0, 0.0, -0.12869, 0.0, -0.0, -0.02748, -0.0, -0.0, 0.02772, 0.0, -0.0, 0.02772, -0.0, -0.0, -0.02748, 0.0, 0.0, -0.12869, 0.0, 0.0, 0.12821, 0.0, -0.0, 0.12821, 0.0, -0.0, 0.12821, -0.0, 0.0, 0.12821]]), "B1u":array([[0.0, -0.0, -0.06886, 0.0, 0.0, -0.04403, -0.0, 0.0, 0.04403, 0.0, -0.0, 0.06886, 0.0, 0.0, 0.13954, -0.0, 0.0, 0.06886, 0.0, -0.0, 0.04403, 0.0, 0.0, -0.04403, -0.0, 0.0, -0.06886, -0.0, -0.0, -0.13954, 0.0, 0.0, -0.06886, -0.0, -0.0, -0.04403, 0.0, -0.0, 0.04403, -0.0, -0.0, 0.06886, -0.0, -0.0, 0.13954, 0.0, -0.0, 0.06886, -0.0, 0.0, 0.04403, -0.0, -0.0, -0.04403, 0.0, -0.0, -0.06886, 0.0, -0.0, -0.13954, 0.0, 0.0, 0.0, 0.0, 0.0, -0.0, -0.0, -0.0, -0.0, -0.0, -0.0, -0.0], [0.0, -0.0, -0.06051, -0.0, -0.0, -0.08008, -0.0, -0.0, 0.08008, -0.0, -0.0, 0.06051, -0.0, -0.0, -0.11025, -0.0, 0.0, 0.06051, -0.0, -0.0, 0.08008, -0.0, 0.0, -0.08008, -0.0, 0.0, -0.06051, -0.0, 0.0, 0.11025, -0.0, -0.0, -0.06051, 0.0, 0.0, -0.08008, 0.0, 0.0, 0.08008, 0.0, -0.0, 0.06051, 0.0, 0.0, -0.11025, 0.0, -0.0, 0.06051, 0.0, 0.0, 0.08008, 0.0, -0.0, -0.08008, 0.0, -0.0, -0.06051, 0.0, -0.0, 0.11025, -0.0, -0.0, 0.0, -0.0, -0.0, 0.0, 0.0, 0.0, -0.0, 0.0, 0.0, 0.0]]), "B2u":array([[0.0, 0.0, -0.04087, 0.0, -0.0, -0.12177, -0.0, 0.0, -0.12177, -0.0, -0.0, -0.04087, 0.0, -0.0, -0.0, -0.0, -0.0, 0.04087, -0.0, -0.0, 0.12177, -0.0, -0.0, 0.12177, -0.0, 0.0, 0.04087, -0.0, -0.0, -0.0, 0.0, -0.0, -0.04087, -0.0, 0.0, -0.12177, 0.0, -0.0, -0.12177, 0.0, 0.0, -0.04087, -0.0, 0.0, 0.0, 0.0, 0.0, 0.04087, 0.0, -0.0, 0.12177, 0.0, -0.0, 0.12177, 0.0, -0.0, 0.04087, 0.0, 0.0, 0.0, -0.0, 0.0, 0.00081, 0.0, 0.0, -0.00081, 0.0, -0.0, 0.00081, -0.0, -0.0, -0.00081], [0.0, -0.0, 0.07237, -0.0, 0.0, -0.02376, 0.0, 0.0, -0.02376, -0.0, -0.0, 0.07237, -0.0, -0.0, 0.0, -0.0, -0.0, -0.07237, 0.0, 0.0, 0.02376, 0.0, 0.0, 0.02376, -0.0, 0.0, -0.07237, 0.0, 0.0, 0.0, -0.0, 0.0, 0.07237, 0.0, -0.0, -0.02376, -0.0, -0.0, -0.02376, 0.0, 0.0, 0.07237, 0.0, 0.0, -0.0, 0.0, 0.0, -0.07237, -0.0, 0.0, 0.02376, -0.0, -0.0, 0.02376, 0.0, -0.0, -0.07237, -0.0, -0.0, -0.0, 0.0, -0.0, 0.15716, -0.0, 0.0, -0.15716, 0.0, 0.0, 0.15716, 0.0, 0.0, -0.15716]]), "Egx":array([[-0.0, 0.0, 0.03221, -0.0, -0.0, -0.12309, 0.0, -0.0, -0.12309, -0.0, 0.0, 0.03221, -0.0, 0.0, 0.09434, -0.0, 0.0, 0.05216, -0.0, -0.0, 0.0352, -0.0, 0.0, -0.0352, -0.0, -0.0, -0.05216, -0.0, -0.0, -0.09434, -0.0, 0.0, -0.03221, 0.0, 0.0, 0.12309, -0.0, 0.0, 0.12309, 0.0, 0.0, -0.03221, 0.0, -0.0, -0.09434, 0.0, -0.0, -0.05216, 0.0, 0.0, -0.0352, 0.0, -0.0, 0.0352, 0.0, 0.0, 0.05216, 0.0, 0.0, 0.09434, 0.0, 0.0, 0.1181, 0.0, 0.0, 0.0, -0.0, 0.0, -0.1181, 0.0, -0.0, -0.0], [-0.0, 0.0, -0.07752, -0.0, -0.0, 0.01656, 0.0, -0.0, 0.01656, -0.0, 0.0, -0.07752, 0.0, 0.0, 0.05973, 0.0, 0.0, 0.08757, -0.0, 0.0, 0.07212, -0.0, -0.0, -0.07212, 0.0, -0.0, -0.08757, 0.0, 0.0, -0.05973, -0.0, 0.0, 0.07752, 0.0, 0.0, -0.01656, -0.0, 0.0, -0.01656, 0.0, 0.0, 0.07752, -0.0, -0.0, -0.05973, -0.0, -0.0, -0.08757, 0.0, -0.0, -0.07212, 0.0, 0.0, 0.07212, -0.0, 0.0, 0.08757, -0.0, -0.0, 0.05973, -0.0, 0.0, -0.13897, 0.0, 0.0, -0.0, 0.0, 0.0, 0.13897, 0.0, -0.0, 0.0]]), "Egy":array([[-0.0, 0.0, 0.04808, -0.0, 0.0, 0.03233, 0.0, 0.0, -0.03233, 0.0, -0.0, -0.04808, 0.0, -0.0, -0.08843, -0.0, -0.0, -0.03311, -0.0, -0.0, 0.11867, -0.0, -0.0, 0.11867, -0.0, 0.0, -0.03311, -0.0, 0.0, -0.08843, 0.0, -0.0, -0.04808, 0.0, -0.0, -0.03233, -0.0, -0.0, 0.03233, -0.0, 0.0, 0.04808, -0.0, 0.0, 0.08843, 0.0, 0.0, 0.03311, 0.0, 0.0, -0.11867, 0.0, 0.0, -0.11867, 0.0, -0.0, 0.03311, 0.0, -0.0, 0.08843, 0.0, -0.0, -0.0, -0.0, 0.0, -0.11764, -0.0, -0.0, 0.0, -0.0, -0.0, 0.11764], [0.0, -0.0, 0.09282, -0.0, 0.0, 0.08525, -0.0, 0.0, -0.08525, 0.0, 0.0, -0.09282, -0.0, -0.0, -0.05472, 0.0, -0.0, 0.07287, 0.0, 0.0, -0.01269, 0.0, -0.0, -0.01269, 0.0, -0.0, 0.07287, -0.0, -0.0, -0.05472, 0.0, 0.0, -0.09282, 0.0, -0.0, -0.08525, 0.0, -0.0, 0.08525, -0.0, -0.0, 0.09282, 0.0, 0.0, 0.05472, -0.0, 0.0, -0.07287, -0.0, -0.0, 0.01269, -0.0, 0.0, 0.01269, -0.0, 0.0, -0.07287, 0.0, 0.0, 0.05472, 0.0, -0.0, -0.0, -0.0, -0.0, 0.13838, -0.0, -0.0, 0.0, -0.0, 0.0, -0.13838]]), "Egx+y":array([[-0.0, -0.0, 0.05635, -0.0, -0.0, -0.05892, -0.0, -0.0, -0.10562, -0.0, -0.0, -0.01222, 0.0, -0.0, -0.0, 0.0, -0.0, 0.01222, -0.0, 0.0, 0.10562, -0.0, -0.0, 0.05892, 0.0, -0.0, -0.05635, 0.0, 0.0, -0.12448, 0.0, 0.0, -0.05635, 0.0, 0.0, 0.05892, 0.0, 0.0, 0.10562, 0.0, 0.0, 0.01222, -0.0, 0.0, 0.0, -0.0, 0.0, -0.01222, 0.0, -0.0, -0.10562, 0.0, 0.0, -0.05892, -0.0, 0.0, 0.05635, -0.0, -0.0, 0.12448, 0.0, -0.0, 0.08013, -0.0, -0.0, -0.08013, -0.0, 0.0, -0.08013, 0.0, 0.0, 0.08013], [0.0, -0.0, 0.01274, 0.0, -0.0, 0.07135, -0.0, 0.0, -0.04904, 0.0, 0.0, -0.12328, 0.0, -0.0, 0.0, -0.0, -0.0, 0.12328, 0.0, 0.0, 0.04904, -0.0, 0.0, -0.07135, -0.0, -0.0, -0.01274, 0.0, 0.0, -0.07988, -0.0, -0.0, -0.01274, -0.0, 0.0, -0.07135, 0.0, -0.0, 0.04904, -0.0, -0.0, 0.12328, -0.0, 0.0, -0.0, 0.0, 0.0, -0.12328, -0.0, -0.0, -0.04904, 0.0, -0.0, 0.07135, 0.0, 0.0, 0.01274, -0.0, -0.0, 0.07988, -0.0, 0.0, -0.10198, 0.0, 0.0, 0.10198, 0.0, -0.0, 0.10198, -0.0, -0.0, -0.10198]]), "Egx-y":array([[0.0, -0.0, -0.0119, 0.0, 0.0, -0.10845, 0.0, 0.0, -0.06172, 0.0, 0.0, 0.05826, -0.0, -0.0, 0.12849, 0.0, 0.0, 0.05826, 0.0, 0.0, -0.06172, -0.0, -0.0, -0.10845, -0.0, -0.0, -0.0119, -0.0, 0.0, -0.0, 0.0, 0.0, 0.0119, -0.0, -0.0, 0.10845, -0.0, -0.0, 0.06172, -0.0, -0.0, -0.05826, 0.0, 0.0, -0.12849, -0.0, -0.0, -0.05826, -0.0, -0.0, 0.06172, 0.0, 0.0, 0.10845, 0.0, 0.0, 0.0119, 0.0, -0.0, 0.0, -0.0, 0.0, 0.0831, 0.0, -0.0, 0.0831, 0.0, -0.0, -0.0831, -0.0, 0.0, -0.0831], [-0.0, 0.0, 0.12156, 0.0, 0.0, 0.04518, -0.0, -0.0, -0.06623, 0.0, -0.0, -0.0093, 0.0, 0.0, -0.08999, 0.0, 0.0, -0.0093, -0.0, -0.0, -0.06623, 0.0, -0.0, 0.04518, -0.0, 0.0, 0.12156, -0.0, -0.0, 0.0, 0.0, 0.0, -0.12156, -0.0, -0.0, -0.04518, 0.0, 0.0, 0.06623, -0.0, 0.0, 0.0093, -0.0, -0.0, 0.08999, -0.0, -0.0, 0.0093, 0.0, 0.0, 0.06623, -0.0, 0.0, -0.04518, 0.0, -0.0, -0.12156, 0.0, 0.0, -0.0, -0.0, 0.0, 0.10327, -0.0, 0.0, 0.10327, 0.0, -0.0, -0.10327, 0.0, -0.0, -0.10327]]), "Eux":array([[0.02841, 0.01541, 0.0, 0.07073, 0.01083, 0.0, 0.07073, -0.01083, -0.0, 0.02841, -0.01541, -0.0, 0.01069, -0.02721, -0.0, -0.03571, -0.03202, -0.0, -0.04714, -0.05665, -0.0, -0.04714, 0.05665, 0.0, -0.03571, 0.03202, 0.0, 0.01069, 0.02721, 0.0, 0.02841, 0.01541, 0.0, 0.07073, 0.01083, 0.0, 0.07073, -0.01083, -0.0, 0.02841, -0.01541, -0.0, 0.01069, -0.02721, -0.0, -0.03571, -0.03202, -0.0, -0.04714, -0.05665, -0.0, -0.04714, 0.05665, 0.0, -0.03571, 0.03202, 0.0, 0.01069, 0.02721, 0.0, 0.01343, 0.0, -0.0, -0.06737, 0.0, -0.0, 0.01343, 0.0, 0.0, -0.06737, 0.0, 0.0], [-0.03605, 0.04897, 0.0, 0.04469, 0.036, 0.0, 0.04469, -0.036, -0.0, -0.03605, -0.04897, -0.0, -0.04209, -0.05107, -0.0, 0.01831, -0.01512, -0.0, 0.03738, 0.01218, 0.0, 0.03738, -0.01218, -0.0, 0.01831, 0.01512, 0.0, -0.04209, 0.05107, 0.0, -0.03605, 0.04897, 0.0, 0.04469, 0.036, 0.0, 0.04469, -0.036, -0.0, -0.03605, -0.04897, -0.0, -0.04209, -0.05107, -0.0, 0.01831, -0.01512, -0.0, 0.03738, 0.01218, 0.0, 0.03738, -0.01218, -0.0, 0.01831, 0.01512, 0.0, -0.04209, 0.05107, 0.0, -0.07666, -0.0, -0.0, 0.03218, -0.0, 0.0, -0.07666, -0.0, 0.0, 0.03218, -0.0, -0.0]]), "Euy":array([[0.03162, -0.03641, -0.0, 0.05813, -0.0485, -0.0, -0.05813, -0.0485, 0.0, -0.03162, -0.03641, 0.0, -0.02655, 0.01253, 0.0, -0.01443, 0.02861, 0.0, -0.01097, 0.07157, 0.0, 0.01097, 0.07157, -0.0, 0.01443, 0.02861, -0.0, 0.02655, 0.01253, -0.0, 0.03162, -0.03641, -0.0, 0.05813, -0.0485, -0.0, -0.05813, -0.0485, 0.0, -0.03162, -0.03641, 0.0, -0.02655, 0.01253, 0.0, -0.01443, 0.02861, 0.0, -0.01097, 0.07157, 0.0, 0.01097, 0.07157, -0.0, 0.01443, 0.02861, -0.0, 0.02655, 0.01253, -0.0, -0.0, -0.06933, -0.0, 0.0, 0.01373, 0.0, -0.0, -0.06933, 0.0, 0.0, 0.01373, 0.0], [0.01666, 0.01862, 0.0, -0.00122, 0.04083, -0.0, 0.00122, 0.04083, -0.0, -0.01666, 0.01862, 0.0, -0.05104, -0.03834, 0.0, -0.04882, -0.03596, 0.0, -0.03688, 0.03836, 0.0, 0.03688, 0.03836, -0.0, 0.04882, -0.03596, -0.0, 0.05104, -0.03834, -0.0, 0.01666, 0.01862, -0.0, -0.00122, 0.04083, 0.0, 0.00122, 0.04083, 0.0, -0.01666, 0.01862, 0.0, -0.05104, -0.03834, 0.0, -0.04882, -0.03596, 0.0, -0.03688, 0.03836, 0.0, 0.03688, 0.03836, -0.0, 0.04882, -0.03596, -0.0, 0.05104, -0.03834, -0.0, 0.0, 0.02829, -0.0, 0.0, -0.07532, 0.0, 0.0, 0.02829, 0.0, 0.0, -0.07532, 0.0]]), "Eux+y":array([[0.04347, -0.01468, -0.0, 0.09139, -0.02635, -0.0, 0.00867, -0.04225, -0.0, -0.00228, -0.03636, -0.0, -0.01075, -0.01075, -0.0, -0.03636, -0.00228, 0.0, -0.04225, 0.00867, 0.0, -0.02635, 0.09139, 0.0, -0.01468, 0.04347, 0.0, 0.02776, 0.02776, 0.0, 0.04347, -0.01468, -0.0, 0.09139, -0.02635, -0.0, 0.00867, -0.04225, -0.0, -0.00228, -0.03636, -0.0, -0.01075, -0.01075, 0.0, -0.03636, -0.00228, 0.0, -0.04225, 0.00867, 0.0, -0.02635, 0.09139, 0.0, -0.01468, 0.04347, 0.0, 0.02776, 0.02776, 0.0, 0.01024, -0.04887, -0.0, -0.04887, 0.01024, 0.0, 0.01024, -0.04887, -0.0, -0.04887, 0.01024, 0.0], [-0.01351, 0.04802, 0.0, 0.02513, 0.05423, 0.0, 0.03, 0.00309, -0.0, -0.03712, -0.02123, 0.0, -0.06431, -0.06431, -0.0, -0.02123, -0.03712, -0.0, 0.00309, 0.03, 0.0, 0.05423, 0.02513, -0.0, 0.04802, -0.01351, -0.0, 0.00882, 0.00882, 0.0, -0.01351, 0.04802, 0.0, 0.02513, 0.05423, 0.0, 0.03, 0.00309, -0.0, -0.03712, -0.02123, 0.0, -0.06431, -0.06431, 0.0, -0.02123, -0.03712, -0.0, 0.00309, 0.03, 0.0, 0.05423, 0.02513, -0.0, 0.04802, -0.01351, -0.0, 0.00882, 0.00882, -0.0, -0.05274, 0.01961, 0.0, 0.01961, -0.05274, -0.0, -0.05274, 0.01961, 0.0, 0.01961, -0.05274, -0.0]]), "Eux-y":array([[0.00247, -0.03658, -0.0, -0.01102, -0.04171, -0.0, -0.0924, -0.02641, -0.0, -0.04283, -0.0155, -0.0, -0.02727, 0.02727, 0.0, 0.0155, 0.04283, 0.0, 0.02641, 0.0924, 0.0, 0.04171, 0.01102, 0.0, 0.03658, -0.00247, 0.0, 0.01117, -0.01117, 0.0, 0.00247, -0.03658, -0.0, -0.01102, -0.04171, -0.0, -0.0924, -0.02641, -0.0, -0.04283, -0.0155, -0.0, -0.02727, 0.02727, -0.0, 0.0155, 0.04283, 0.0, 0.02641, 0.0924, 0.0, 0.04171, 0.01102, 0.0, 0.03658, -0.00247, 0.0, 0.01117, -0.01117, -0.0, -0.00917, -0.04884, -0.0, 0.04884, 0.00917, 0.0, -0.00917, -0.04884, -0.0, 0.04884, 0.00917, 0.0], [0.03571, -0.02131, -0.0, -0.03716, 0.00075, -0.0, -0.02382, 0.05176, 0.0, 0.01475, 0.04646, 0.0, -0.00637, 0.00637, -0.0, -0.04646, -0.01475, -0.0, -0.05176, 0.02382, -0.0, -0.00075, 0.03716, 0.0, 0.02131, -0.03571, -0.0, 0.06359, -0.06359, -0.0, 0.03571, -0.02131, 0.0, -0.03716, 0.00075, -0.0, -0.02382, 0.05176, 0.0, 0.01475, 0.04646, 0.0, -0.00637, 0.00637, -0.0, -0.04646, -0.01475, -0.0, -0.05176, 0.02382, -0.0, -0.00075, 0.03716, 0.0, 0.02131, -0.03571, -0.0, 0.06359, -0.06359, 0.0, 0.05386, 0.0229, 0.0, -0.0229, -0.05386, -0.0, 0.05386, 0.0229, 0.0, -0.0229, -0.05386, -0.0]])},
model_objs_dict['bcr_pi_pdb'] = scsd_model('bcr_pi_pdb', array([[11.1301, -1.1288, 0.0], [12.3684, -0.5703, 0.0], [10.0689, -0.2872, 0.0], [8.7467, -0.5395, 0.0], [7.7507, 0.3787, 0.0], [6.4774, -0.0783, 0.0], [5.3393, 0.6398, 0.0], [4.1174, 0.0711, 0.0], [2.9342, 0.7264, 0.0], [1.8034, -0.0163, 0.0], [0.5287, 0.4179, 0.0], [-0.5287, -0.4179, 0.0], [-1.8034, 0.0163, 0.0], [-2.9342, -0.7264, 0.0], [-4.1174, -0.0711, 0.0], [-5.3393, -0.6398, 0.0], [-6.4774, 0.0783, 0.0], [-7.7507, -0.3787, 0.0], [-8.7467, 0.5395, 0.0], [-10.0689, 0.2872, 0.0], [-11.1301, 1.1288, 0.0], [-12.3684, 0.5703, 0.0]]), 'C2h' , pca = {"Ag":array([[-0.07838, 0.04298, 0.0, -0.09365, 0.06074, 0.0, -0.05905, 0.01515, -0.0, -0.05042, 0.02621, 0.0, -0.03195, -0.00117, 0.0, -0.03001, 0.01197, 0.0, -0.01863, -0.00782, 0.0, -0.01625, 0.00342, 0.0, -0.00221, -0.01661, 0.0, -0.00629, 0.00246, 0.0, 0.0017, -0.00865, 0.0, -0.0017, 0.00865, 0.0, 0.00629, -0.00246, 0.0, 0.00221, 0.01661, 0.0, 0.01625, -0.00342, 0.0, 0.01863, 0.00782, 0.0, 0.03001, -0.01197, 0.0, 0.03195, 0.00117, 0.0, 0.05042, -0.02621, 0.0, 0.05905, -0.01515, 0.0, 0.07838, -0.04298, 0.0, 0.09365, -0.06074, 0.0], [-0.03576, -0.08611, 0.0, -0.03991, -0.1103, 0.0, -0.03432, -0.05407, 0.0, -0.02335, -0.01982, -0.0, -0.01957, 0.01113, -0.0, -0.01824, 0.03199, -0.0, -0.01166, 0.01425, -0.0, -0.01023, 0.029, -0.0, -0.0045, 0.00708, -0.0, -0.00565, 0.02908, -0.0, 0.00056, -0.00915, -0.0, -0.00056, 0.00915, -0.0, 0.00565, -0.02908, -0.0, 0.0045, -0.00708, -0.0, 0.01023, -0.029, -0.0, 0.01166, -0.01425, -0.0, 0.01824, -0.03199, -0.0, 0.01957, -0.01113, -0.0, 0.02335, 0.01982, -0.0, 0.03432, 0.05407, -0.0, 0.03576, 0.08611, -0.0, 0.03991, 0.1103, -0.0]]), "Bg":array([[-0.0, -0.0, 0.02947, -0.0, -0.0, 0.31565, 0.0, -0.0, -0.02049, -0.0, -0.0, -0.0239, -0.0, -0.0, -0.05277, -0.0, -0.0, -0.00714, -0.0, -0.0, 0.00057, 0.0, -0.0, 0.01451, -0.0, 0.0, 0.01219, -0.0, -0.0, 0.01454, 0.0, -0.0, 0.00878, -0.0, 0.0, -0.00878, 0.0, 0.0, -0.01454, 0.0, -0.0, -0.01219, -0.0, 0.0, -0.01451, -0.0, 0.0, -0.00057, -0.0, 0.0, 0.00714, -0.0, 0.0, 0.05277, -0.0, 0.0, 0.0239, -0.0, 0.0, 0.02049, -0.0, 0.0, -0.02947, 0.0, 0.0, -0.31565], [-0.0, -0.0, -0.12986, 0.0, 0.0, -0.00883, 0.0, 0.0, -0.14889, -0.0, -0.0, -0.01369, -0.0, 0.0, -0.04551, -0.0, 0.0, 0.04596, -0.0, -0.0, 0.01215, -0.0, 0.0, 0.04894, -0.0, 0.0, 0.00029, -0.0, -0.0, 0.03622, 0.0, 0.0, -0.00965, -0.0, -0.0, 0.00965, 0.0, 0.0, -0.03622, 0.0, -0.0, -0.00029, 0.0, -0.0, -0.04894, -0.0, 0.0, -0.01215, -0.0, -0.0, -0.04596, -0.0, -0.0, 0.04551, -0.0, 0.0, 0.01369, -0.0, -0.0, 0.14889, -0.0, 0.0, 0.12986, -0.0, -0.0, 0.00883]]), "Au":array([[-0.0, -0.0, -0.08511, 0.0, -0.0, -0.1155, -0.0, -0.0, -0.06185, -0.0, 0.0, -0.03489, -0.0, -0.0, -0.01531, -0.0, -0.0, 0.0037, -0.0, 0.0, 0.01867, -0.0, 0.0, 0.03126, -0.0, 0.0, 0.04101, -0.0, -0.0, 0.04449, -0.0, 0.0, 0.04822, -0.0, -0.0, 0.04822, -0.0, -0.0, 0.04449, -0.0, 0.0, 0.04101, -0.0, -0.0, 0.03126, -0.0, 0.0, 0.01867, -0.0, -0.0, 0.0037, -0.0, -0.0, -0.01531, -0.0, -0.0, -0.03489, -0.0, 0.0, -0.06185, 0.0, -0.0, -0.08511, 0.0, 0.0, -0.1155], [0.0, -0.0, -0.01755, 0.0, 0.0, 0.12163, -0.0, 0.0, -0.00183, -0.0, 0.0, -0.00659, -0.0, 0.0, 0.00956, -0.0, 0.0, 0.02705, -0.0, 0.0, 0.04208, -0.0, 0.0, 0.05755, -0.0, 0.0, 0.06676, -0.0, -0.0, 0.0721, -0.0, 0.0, 0.07729, -0.0, -0.0, 0.07729, -0.0, -0.0, 0.0721, -0.0, -0.0, 0.06676, 0.0, -0.0, 0.05755, -0.0, 0.0, 0.04208, -0.0, 0.0, 0.02705, -0.0, 0.0, 0.00956, -0.0, 0.0, -0.00659, -0.0, 0.0, -0.00183, 0.0, 0.0, -0.01755, 0.0, 0.0, 0.12163]]), "Bu":array([[0.00701, -0.07708, -0.0, 0.00597, -0.09914, -0.0, 0.00461, -0.04819, -0.0, 0.00592, -0.03234, -0.0, 0.00382, -0.00521, -0.0, 0.00333, 0.00443, -0.0, 0.0027, 0.02446, -0.0, 0.00096, 0.03438, -0.0, -0.0007, 0.05476, -0.0, -0.00165, 0.05636, -0.0, -0.00144, 0.05889, -0.0, -0.00144, 0.05889, -0.0, -0.00165, 0.05636, -0.0, -0.0007, 0.05476, -0.0, 0.00096, 0.03438, -0.0, 0.0027, 0.02446, -0.0, 0.00333, 0.00443, -0.0, 0.00382, -0.00521, -0.0, 0.00592, -0.03234, -0.0, 0.00461, -0.04819, -0.0, 0.00701, -0.07708, -0.0, 0.00597, -0.09914, -0.0], [-0.00872, -0.03736, 0.0, -0.0443, -0.1069, -0.0, -0.01257, -0.01631, 0.0, -0.00425, 0.00179, 0.0, -0.00819, 0.02031, 0.0, -0.01082, 0.01977, 0.0, -0.00893, -0.0131, 0.0, -0.00636, -0.03136, 0.0, -0.00739, -0.06675, 0.0, -0.00323, -0.07247, 0.0, -0.00669, -0.07749, 0.0, -0.00669, -0.07749, 0.0, -0.00323, -0.07247, 0.0, -0.00739, -0.06675, 0.0, -0.00636, -0.03136, 0.0, -0.00893, -0.0131, 0.0, -0.01082, 0.01977, 0.0, -0.00819, 0.02031, 0.0, -0.00425, 0.00179, 0.0, -0.01257, -0.01631, 0.0, -0.00872, -0.03736, 0.0, -0.0443, -0.1069, 0.0]])}, database_path = 'bcr_pi_pdb_scsd_20220713.pkl', maxdist = 1.75)
model_objs_dict['calix4'] = scsd_model('calix4', array([[-0.0, -1.9528, -1.9094], [1.9528, 0.0, -1.9094], [0.0, 1.9528, -1.9094], [-1.9528, 0.0, -1.9094], [-0.0, -2.6293, -0.7292], [-1.2185, -2.9604, -0.1593], [-1.203, -3.6948, 1.0222], [-0.0, -4.042, 1.6127], [1.203, -3.6948, 1.0222], [1.2185, -2.9604, -0.1593], [2.5438, -2.5438, -0.6998], [2.9604, -1.2185, -0.1593], [2.6293, -0.0, -0.7292], [2.9604, 1.2185, -0.1593], [3.6948, 1.203, 1.0222], [4.042, -0.0, 1.6127], [3.6948, -1.203, 1.0222], [2.5438, 2.5438, -0.6998], [1.2185, 2.9604, -0.1593], [0.0, 2.6293, -0.7292], [-1.2185, 2.9604, -0.1593], [-1.203, 3.6948, 1.0222], [0.0, 4.042, 1.6127], [1.203, 3.6948, 1.0222], [-2.5438, 2.5438, -0.6998], [-2.9604, 1.2185, -0.1593], [-2.6293, 0.0, -0.7292], [-2.9604, -1.2185, -0.1593], [-3.6948, -1.203, 1.0222], [-4.042, 0.0, 1.6127], [-3.6948, 1.203, 1.0222], [-2.5438, -2.5438, -0.6998]]), 'C4v' , pca = {"A1":array([[0.0, 0.00557, 0.04999, -0.00557, -0.0, 0.04999, 0.0, -0.00557, 0.04999, 0.00557, -0.0, 0.04999, 0.0, 0.0036, 0.01102, 0.00032, 0.00314, -0.00836, 0.00035, 0.00219, -0.04678, -0.0, 0.00169, -0.06637, -0.00035, 0.00219, -0.04678, -0.00032, 0.00314, -0.00836, -0.00062, 0.00062, 0.00997, -0.00314, 0.00032, -0.00836, -0.0036, 0.0, 0.01102, -0.00314, -0.00032, -0.00836, -0.00219, -0.00035, -0.04678, -0.00169, 0.0, -0.06637, -0.00219, 0.00035, -0.04678, -0.00062, -0.00062, 0.00997, -0.00032, -0.00314, -0.00836, 0.0, -0.0036, 0.01102, 0.00032, -0.00314, -0.00836, 0.00035, -0.00219, -0.04678, 0.0, -0.00169, -0.06637, -0.00035, -0.00219, -0.04678, 0.00062, -0.00062, 0.00997, 0.00314, -0.00032, -0.00836, 0.0036, -0.0, 0.01102, 0.00314, 0.00032, -0.00836, 0.00219, 0.00035, -0.04678, 0.00169, 0.0, -0.06637, 0.00219, -0.00035, -0.04678, 0.00062, 0.00062, 0.00997], [-0.0, -0.04774, -0.01185, 0.04774, -0.0, -0.01185, 0.0, 0.04774, -0.01185, -0.04774, 0.0, -0.01185, -0.0, -0.01032, -0.00891, 0.00017, 0.0117, -0.00637, -0.0001, 0.05027, -0.00336, 0.0, 0.06913, -0.00283, 0.0001, 0.05027, -0.00336, -0.00017, 0.0117, -0.00637, -0.00114, 0.00114, -0.00268, -0.0117, 0.00017, -0.00637, 0.01032, -0.0, -0.00891, -0.0117, -0.00017, -0.00637, -0.05027, 0.0001, -0.00336, -0.06913, 0.0, -0.00283, -0.05027, -0.0001, -0.00336, -0.00114, -0.00114, -0.00268, -0.00017, -0.0117, -0.00637, -0.0, 0.01032, -0.00891, 0.00017, -0.0117, -0.00637, -0.0001, -0.05027, -0.00336, -0.0, -0.06913, -0.00283, 0.0001, -0.05027, -0.00336, 0.00114, -0.00114, -0.00268, 0.0117, -0.00017, -0.00637, -0.01032, 0.0, -0.00891, 0.0117, 0.00017, -0.00637, 0.05027, -0.0001, -0.00336, 0.06913, 0.0, -0.00283, 0.05027, 0.0001, -0.00336, 0.00114, 0.00114, -0.00268]]), "A2":array([[0.09714, -0.0, -0.0, 0.0, 0.09714, 0.0, -0.09714, -0.0, -0.0, -0.0, -0.09714, -0.0, 0.01709, 0.0, -0.0, -0.00212, -0.01079, -0.00177, -0.03776, 0.0068, -0.00757, -0.03335, 0.0, 0.0, -0.03776, -0.0068, 0.00757, -0.00212, 0.01079, 0.00177, 0.00868, 0.00868, -0.0, 0.01079, -0.00212, -0.00177, 0.0, 0.01709, -0.0, -0.01079, -0.00212, 0.00177, 0.0068, -0.03776, 0.00757, 0.0, -0.03335, -0.0, -0.0068, -0.03776, -0.00757, -0.00868, 0.00868, 0.0, 0.00212, 0.01079, -0.00177, -0.01709, -0.0, -0.0, 0.00212, -0.01079, 0.00177, 0.03776, 0.0068, 0.00757, 0.03335, -0.0, -0.0, 0.03776, -0.0068, -0.00757, -0.00868, -0.00868, -0.0, -0.01079, 0.00212, -0.00177, -0.0, -0.01709, -0.0, 0.01079, 0.00212, 0.00177, -0.0068, 0.03776, 0.00757, -0.0, 0.03335, 0.0, 0.0068, 0.03776, -0.00757, 0.00868, -0.00868, -0.0], [0.0388, -0.0, 0.0, -0.0, 0.0388, 0.0, -0.0388, 0.0, -0.0, 0.0, -0.0388, -0.0, -0.0025, 0.0, 0.0, 0.00764, -0.00167, -0.00027, 0.02396, 0.04309, -0.00996, 0.0753, -0.0, 0.0, 0.02396, -0.04309, 0.00996, 0.00764, 0.00167, 0.00027, -0.00848, -0.00848, -0.0, 0.00167, 0.00764, -0.00027, 0.0, -0.0025, 0.0, -0.00167, 0.00764, 0.00027, 0.04309, 0.02396, 0.00996, -0.0, 0.0753, -0.0, -0.04309, 0.02396, -0.00996, 0.00848, -0.00848, 0.0, -0.00764, 0.00167, -0.00027, 0.0025, -0.0, 0.0, -0.00764, -0.00167, 0.00027, -0.02396, 0.04309, 0.00996, -0.0753, 0.0, -0.0, -0.02396, -0.04309, -0.00996, 0.00848, 0.00848, 0.0, -0.00167, -0.00764, -0.00027, 0.0, 0.0025, 0.0, 0.00167, -0.00764, 0.00027, -0.04309, -0.02396, 0.00996, 0.0, -0.0753, 0.0, 0.04309, -0.02396, -0.00996, -0.00848, 0.00848, -0.0]]), "B1":array([[0.0, -0.02148, -0.01596, -0.02148, 0.0, 0.01596, 0.0, 0.02148, -0.01596, 0.02148, -0.0, 0.01596, 0.0, 0.00165, 0.00397, 6e-05, 0.01294, 0.01291, 8e-05, 0.03597, 0.03288, 0.0, 0.04739, 0.04339, -8e-05, 0.03597, 0.03288, -6e-05, 0.01294, 0.01291, 0.00054, 0.00054, 0.0, 0.01294, -6e-05, -0.01291, 0.00165, 0.0, -0.00397, 0.01294, 6e-05, -0.01291, 0.03597, 8e-05, -0.03288, 0.04739, -0.0, -0.04339, 0.03597, -8e-05, -0.03288, 0.00054, -0.00054, -0.0, -6e-05, -0.01294, 0.01291, -0.0, -0.00165, 0.00397, 6e-05, -0.01294, 0.01291, 8e-05, -0.03597, 0.03288, -0.0, -0.04739, 0.04339, -8e-05, -0.03597, 0.03288, -0.00054, -0.00054, -0.0, -0.01294, 6e-05, -0.01291, -0.00165, 0.0, -0.00397, -0.01294, -6e-05, -0.01291, -0.03597, -8e-05, -0.03288, -0.04739, 0.0, -0.04339, -0.03597, 8e-05, -0.03288, -0.00054, 0.00054, 0.0], [-0.0, -0.01712, 0.02719, -0.01712, 0.0, -0.02719, 0.0, 0.01712, 0.02719, 0.01712, 0.0, -0.02719, 0.0, 0.0027, 0.00243, 0.0, 0.01227, -0.0104, -1e-05, 0.03224, -0.03481, 0.0, 0.04234, -0.04689, 1e-05, 0.03224, -0.03481, -0.0, 0.01227, -0.0104, 0.00079, 0.00079, -0.0, 0.01227, -0.0, 0.0104, 0.0027, 0.0, -0.00243, 0.01227, 0.0, 0.0104, 0.03224, -1e-05, 0.03481, 0.04234, -0.0, 0.04689, 0.03224, 1e-05, 0.03481, 0.00079, -0.00079, -0.0, -0.0, -0.01227, -0.0104, -0.0, -0.0027, 0.00243, 0.0, -0.01227, -0.0104, -1e-05, -0.03224, -0.03481, -0.0, -0.04234, -0.04689, 1e-05, -0.03224, -0.03481, -0.00079, -0.00079, 0.0, -0.01227, 0.0, 0.0104, -0.0027, 0.0, -0.00243, -0.01227, -0.0, 0.0104, -0.03224, 1e-05, 0.03481, -0.04234, 0.0, 0.04689, -0.03224, -1e-05, 0.03481, -0.00079, 0.00079, 0.0]]), "B2":array([[0.04463, -0.0, 0.0, 0.0, -0.04463, -0.0, -0.04463, 0.0, -0.0, 0.0, 0.04463, -0.0, 0.02301, 0.0, 0.0, 0.01047, 0.01149, -0.01969, -0.01545, 0.0116, -0.01974, -0.02898, 0.0, 0.0, -0.01545, -0.0116, 0.01974, 0.01047, -0.01149, 0.01969, 0.0251, -0.0251, 0.03979, 0.01149, -0.01047, 0.01969, -0.0, -0.02301, -0.0, -0.01149, -0.01047, -0.01969, -0.0116, 0.01545, -0.01974, -0.0, 0.02898, 0.0, 0.0116, 0.01545, 0.01974, -0.0251, -0.0251, -0.03979, -0.01047, -0.01149, -0.01969, -0.02301, -0.0, 0.0, -0.01047, 0.01149, 0.01969, 0.01545, 0.0116, 0.01974, 0.02898, -0.0, 0.0, 0.01545, -0.0116, -0.01974, -0.0251, 0.0251, 0.03979, -0.01149, 0.01047, 0.01969, 0.0, 0.02301, 0.0, 0.01149, 0.01047, -0.01969, 0.0116, -0.01545, -0.01974, -0.0, -0.02898, -0.0, -0.0116, -0.01545, 0.01974, 0.0251, 0.0251, -0.03979], [-0.02503, -0.0, 0.0, 0.0, 0.02503, 0.0, 0.02503, 0.0, -0.0, 0.0, -0.02503, 0.0, -0.02193, 0.0, 0.0, -0.02121, -0.01625, -0.01978, -0.01869, -0.01874, -0.02077, -0.01786, -0.0, -0.0, -0.01869, 0.01874, 0.02077, -0.02121, 0.01625, 0.01978, -0.02422, 0.02422, 0.03509, -0.01625, 0.02121, 0.01978, -0.0, 0.02193, 0.0, 0.01625, 0.02121, -0.01978, 0.01874, 0.01869, -0.02077, -0.0, 0.01786, 0.0, -0.01874, 0.01869, 0.02077, 0.02422, 0.02422, -0.03509, 0.02121, 0.01625, -0.01978, 0.02193, -0.0, 0.0, 0.02121, -0.01625, 0.01978, 0.01869, -0.01874, 0.02077, 0.01786, -0.0, -0.0, 0.01869, 0.01874, -0.02077, 0.02422, -0.02422, 0.03509, 0.01625, -0.02121, 0.01978, 0.0, -0.02193, 0.0, -0.01625, -0.02121, -0.01978, -0.01874, -0.01869, -0.02077, 0.0, -0.01786, -0.0, 0.01874, -0.01869, 0.02077, -0.02422, -0.02422, -0.03509]]), "Ex":array([[-0.00247, 0.0, 0.0, -0.00809, -0.0, 0.09414, -0.00247, 0.0, -0.0, -0.00809, 0.0, -0.09414, -0.00164, -0.0, -0.0, -0.00154, -0.00046, -0.01393, -0.0011, -0.00056, -0.01363, -0.00072, -0.0, 0.0, -0.0011, 0.00056, 0.01363, -0.00154, 0.00046, 0.01393, -0.00176, 0.0002, 0.02952, -0.00046, 0.00014, -0.00379, -0.00293, -0.0, 0.02894, -0.00046, -0.00014, -0.00379, 0.00416, -0.00028, -0.06866, 0.00614, 0.0, -0.10147, 0.00416, 0.00028, -0.06866, -0.00176, -0.0002, 0.02952, -0.00154, -0.00046, 0.01393, -0.00164, -0.0, 0.0, -0.00154, 0.00046, -0.01393, -0.0011, 0.00056, -0.01363, -0.00072, -0.0, -0.0, -0.0011, -0.00056, 0.01363, -0.00176, 0.0002, -0.02952, -0.00046, 0.00014, 0.00379, -0.00293, 0.0, -0.02894, -0.00046, -0.00014, 0.00379, 0.00416, -0.00028, 0.06866, 0.00614, -0.0, 0.10147, 0.00416, 0.00028, 0.06866, -0.00176, -0.0002, -0.02952], [0.01114, 0.0, 0.0, 0.08858, -0.0, 0.00055, 0.01114, -0.0, 0.0, 0.08858, -0.0, -0.00055, 0.00521, 0.0, -0.0, 0.00434, 0.00426, 0.0047, 0.00434, 0.00408, 0.00588, 0.00475, 0.0, 0.0, 0.00434, -0.00408, -0.00588, 0.00434, -0.00426, -0.0047, 0.00285, 0.00113, -0.00154, -0.01603, 0.00074, -0.00579, 0.01983, -0.0, -0.00275, -0.01603, -0.00074, -0.00579, -0.08268, -0.00118, -0.01076, -0.11725, 0.0, -0.01121, -0.08268, 0.00118, -0.01076, 0.00285, -0.00113, -0.00154, 0.00434, 0.00426, -0.0047, 0.00521, 0.0, 0.0, 0.00434, -0.00426, 0.0047, 0.00434, -0.00408, 0.00588, 0.00475, -0.0, -0.0, 0.00434, 0.00408, -0.00588, 0.00285, 0.00113, 0.00154, -0.01603, 0.00074, 0.00579, 0.01983, -0.0, 0.00275, -0.01603, -0.00074, 0.00579, -0.08268, -0.00118, 0.01076, -0.11725, 0.0, 0.01121, -0.08268, 0.00118, 0.01076, 0.00285, -0.00113, 0.00154]]), "Ey":array([[0.0, -0.01936, 0.06035, -0.0, -0.00252, 0.0, -0.0, -0.01936, -0.06035, -0.0, -0.00252, -0.0, 0.0, -0.00117, 0.00378, 5e-05, 0.00804, -0.02447, 0.00014, 0.02558, -0.08143, -0.0, 0.03414, -0.11043, -0.00014, 0.02558, -0.08143, -5e-05, 0.00804, -0.02447, 0.00016, -0.00165, 0.00471, -4e-05, -0.0018, 0.00254, 0.0, -0.00179, -0.0, 4e-05, -0.0018, -0.00254, -0.00011, -0.0017, -0.0024, -0.0, -0.00155, 0.0, 0.00011, -0.0017, 0.0024, -0.00016, -0.00165, -0.00471, 5e-05, 0.00804, 0.02447, -0.0, -0.00117, -0.00378, -5e-05, 0.00804, 0.02447, -0.00014, 0.02558, 0.08143, 0.0, 0.03414, 0.11043, 0.00014, 0.02558, 0.08143, 0.00016, -0.00165, -0.00471, -4e-05, -0.0018, -0.00254, 0.0, -0.00179, 0.0, 4e-05, -0.0018, 0.00254, -0.00011, -0.0017, 0.0024, 0.0, -0.00155, 0.0, 0.00011, -0.0017, -0.0024, -0.00016, -0.00165, 0.00471], [-0.0, -0.0599, -0.05192, 0.0, -0.0016, -0.0, -0.0, -0.0599, 0.05192, -0.0, -0.0016, 0.0, -0.0, -0.00909, -0.02464, -7e-05, 0.01633, -0.01178, 6e-05, 0.06429, 0.01238, 0.0, 0.08804, 0.02546, -6e-05, 0.06429, 0.01238, 7e-05, 0.01633, -0.01178, 7e-05, -0.00101, -0.01928, 0.00204, -0.00195, -0.01002, -0.0, -0.00125, 0.0, -0.00204, -0.00195, 0.01002, -0.00237, -0.00307, 0.01054, 0.0, -0.00324, -0.0, 0.00237, -0.00307, -0.01054, -7e-05, -0.00101, 0.01928, -7e-05, 0.01633, 0.01178, 0.0, -0.00909, 0.02464, 7e-05, 0.01633, 0.01178, -6e-05, 0.06429, -0.01238, 0.0, 0.08804, -0.02546, 6e-05, 0.06429, -0.01238, 7e-05, -0.00101, 0.01928, 0.00204, -0.00195, 0.01002, -0.0, -0.00125, -0.0, -0.00204, -0.00195, -0.01002, -0.00237, -0.00307, -0.01054, -0.0, -0.00324, 0.0, 0.00237, -0.00307, 0.01054, -7e-05, -0.00101, -0.01928]]), "Ex+y":array([[0.00034, -0.00514, 0.04515, 0.00514, -0.00034, 0.04515, 0.00034, -0.00514, -0.04515, 0.00514, -0.00034, -0.04515, 0.00032, -4e-05, 0.00836, 0.00056, 0.00258, -0.01446, 0.00093, 0.00771, -0.05113, 0.00087, 0.01022, -0.06557, 0.00061, 0.00773, -0.04255, 0.00039, 0.00273, -0.00559, 0.00018, -0.00018, 0.01787, -0.00273, -0.00039, -0.00559, 4e-05, -0.00032, 0.00836, -0.00258, -0.00056, -0.01446, -0.00771, -0.00093, -0.05113, -0.01022, -0.00087, -0.06557, -0.00773, -0.00061, -0.04255, 0.00026, -0.00026, -0.0, 0.00056, 0.00258, 0.01446, 0.00032, -4e-05, -0.00836, 0.00039, 0.00273, 0.00559, 0.00061, 0.00773, 0.04255, 0.00087, 0.01022, 0.06557, 0.00093, 0.00771, 0.05113, 0.00018, -0.00018, -0.01787, -0.00273, -0.00039, 0.00559, 4e-05, -0.00032, -0.00836, -0.00258, -0.00056, 0.01446, -0.00771, -0.00093, 0.05113, -0.01022, -0.00087, 0.06557, -0.00773, -0.00061, 0.04255, 0.00026, -0.00026, -0.0], [0.00431, -0.03805, -0.03098, 0.03805, -0.00431, -0.03098, 0.00431, -0.03805, 0.03098, 0.03805, -0.00431, 0.03098, 0.00242, -0.0064, -0.01934, 0.00203, 0.01067, -0.0053, 0.00162, 0.04067, 0.00466, 0.00116, 0.05462, 0.00132, 0.0013, 0.03877, -0.01329, 0.00204, 0.00889, -0.0228, 0.00222, -0.00222, -0.03494, -0.00889, -0.00204, -0.0228, 0.0064, -0.00242, -0.01934, -0.01067, -0.00203, -0.0053, -0.04067, -0.00162, 0.00466, -0.05462, -0.00116, 0.00132, -0.03877, -0.0013, -0.01329, 0.00186, -0.00186, -0.0, 0.00203, 0.01067, 0.0053, 0.00242, -0.0064, 0.01934, 0.00204, 0.00889, 0.0228, 0.0013, 0.03877, 0.01329, 0.00116, 0.05462, -0.00132, 0.00162, 0.04067, -0.00466, 0.00222, -0.00222, 0.03494, -0.00889, -0.00204, 0.0228, 0.0064, -0.00242, 0.01934, -0.01067, -0.00203, 0.0053, -0.04067, -0.00162, -0.00466, -0.05462, -0.00116, -0.00132, -0.03877, -0.0013, 0.01329, 0.00186, -0.00186, -0.0]]), "Ex-y":array([[0.00175, 0.01258, -0.0327, 0.01258, 0.00175, 0.0327, 0.00175, 0.01258, 0.0327, 0.01258, 0.00175, -0.0327, 0.00112, 0.00053, -0.00152, 0.00093, -0.00547, 0.01325, 0.00044, -0.01695, 0.04476, 0.00024, -0.02265, 0.06152, 0.00046, -0.01696, 0.04635, 0.00089, -0.00542, 0.01496, 0.00088, 0.00088, 0.0, -0.00542, 0.00089, -0.01496, 0.00053, 0.00112, 0.00152, -0.00547, 0.00093, -0.01325, -0.01695, 0.00044, -0.04476, -0.02265, 0.00024, -0.06152, -0.01696, 0.00046, -0.04635, 0.00121, 0.00121, 0.00372, 0.00093, -0.00547, -0.01325, 0.00112, 0.00053, 0.00152, 0.00089, -0.00542, -0.01496, 0.00046, -0.01696, -0.04635, 0.00024, -0.02265, -0.06152, 0.00044, -0.01695, -0.04476, 0.00088, 0.00088, 0.0, -0.00542, 0.00089, 0.01496, 0.00053, 0.00112, -0.00152, -0.00547, 0.00093, 0.01325, -0.01695, 0.00044, 0.04476, -0.02265, 0.00024, 0.06152, -0.01696, 0.00046, 0.04635, 0.00121, 0.00121, -0.00372], [0.00194, 0.03909, 0.03065, 0.03909, 0.00194, -0.03065, 0.00194, 0.03909, -0.03065, 0.03909, 0.00194, 0.03065, 0.00173, 0.00779, 0.01368, 0.00275, -0.00581, 0.01114, 0.00465, -0.03563, -0.00471, 0.00513, -0.05303, -0.01905, 0.00482, -0.04072, -0.01704, 0.00298, -0.01007, -0.00053, 0.00167, 0.00167, 0.0, -0.01007, 0.00298, 0.00053, 0.00779, 0.00173, -0.01368, -0.00581, 0.00275, -0.01114, -0.03563, 0.00465, 0.00471, -0.05303, 0.00513, 0.01905, -0.04072, 0.00482, 0.01704, 0.00112, 0.00112, -0.02194, 0.00275, -0.00581, -0.01114, 0.00173, 0.00779, -0.01368, 0.00298, -0.01007, 0.00053, 0.00482, -0.04072, 0.01704, 0.00513, -0.05303, 0.01905, 0.00465, -0.03563, 0.00471, 0.00167, 0.00167, 0.0, -0.01007, 0.00298, -0.00053, 0.00779, 0.00173, 0.01368, -0.00581, 0.00275, 0.01114, -0.03563, 0.00465, -0.00471, -0.05303, 0.00513, -0.01905, -0.04072, 0.00482, -0.01704, 0.00112, 0.00112, 0.02194]])}, database_path = 'calix4_scsd_20220713.pkl', maxdist = 1.65, mondrian_limits = (-0.5, -0.5))
model_objs_dict['ctc'] = scsd_model('ctc', array([['-1.7998', '0.0', '1.8333', 'C'], ['0.8999', '1.5587', '1.8333', 'C'], ['0.8999', '-1.5587', '1.8333', 'C'], ['-1.4949', '-1.2565', '1.066', 'C'], ['1.8356', '0.6663', '1.066', 'C'], ['1.8356', '-0.6663', '1.066', 'C'], ['-0.3408', '1.9229', '1.066', 'C'], ['-1.4949', '1.2565', '1.066', 'C'], ['-0.3407', '-1.9229', '1.066', 'C'], ['-2.58', '-1.6581', '0.2433', 'C'], ['2.726', '1.4053', '0.2433', 'C'], ['2.726', '-1.4053', '0.2432', 'C'], ['-0.146', '3.0634', '0.2433', 'C'], ['-2.58', '1.6582', '0.2433', 'C'], ['-0.146', '-3.0634', '0.2432', 'C'], ['-2.4057', '-2.7011', '-0.6562', 'C'], ['3.5421', '0.7328', '-0.6562', 'C'], ['3.5421', '-0.7328', '-0.6563', 'C'], ['-1.1364', '3.4339', '-0.6563', 'C'], ['-2.4057', '2.7012', '-0.6563', 'C'], ['-1.1364', '-3.434', '-0.6563', 'C'], ['-3.2528', '-3.2364', '-1.5696', 'O'], ['-3.2528', '3.2364', '-1.5696', 'O'], ['4.4292', '-1.1988', '-1.5696', 'O'], ['-1.1764', '-4.4352', '-1.5696', 'O'], ['4.4292', '1.1988', '-1.5696', 'O'], ['-1.1764', '4.4352', '-1.5696', 'O']]), 'C3v' , pca = {"A1":array([[0.01628, 0.0, -0.04968, -0.00814, -0.0141, -0.04968, -0.00814, 0.0141, -0.04968, 0.0047, -0.00067, -0.02653, -0.00177, -0.0044, -0.02653, -0.00177, 0.0044, -0.02653, -0.00293, -0.00373, -0.02653, 0.0047, 0.00067, -0.02653, -0.00293, 0.00373, -0.02653, 0.00387, -0.00983, 0.0018, 0.00658, -0.00826, 0.0018, 0.00658, 0.00826, 0.0018, -0.01044, 0.00157, 0.0018, 0.00387, 0.00983, 0.0018, -0.01044, -0.00157, 0.0018, -0.00457, -0.01717, 0.03235, 0.01716, -0.00462, 0.03235, 0.01716, 0.00462, 0.03235, -0.01258, 0.01255, 0.03235, -0.00457, 0.01717, 0.03235, -0.01258, -0.01255, 0.03235, -0.00652, -0.03015, 0.06496, -0.00652, 0.03015, 0.06496, 0.02937, 0.00942, 0.06496, -0.02285, -0.02072, 0.06496, 0.02937, -0.00942, 0.06496, -0.02285, 0.02072, 0.06496], [0.00714, 0.0, -0.01682, -0.00357, -0.00618, -0.01682, -0.00357, 0.00618, -0.01682, 0.00951, 0.00535, 0.00513, -0.00939, -0.00556, 0.00513, -0.00939, 0.00556, 0.00513, -0.00012, -0.01091, 0.00513, 0.00951, -0.00535, 0.00513, -0.00012, 0.01091, 0.00513, 0.02012, 0.0132, 0.01533, -0.0215, -0.01083, 0.01533, -0.0215, 0.01083, 0.01533, 0.00137, -0.02403, 0.01533, 0.02012, -0.0132, 0.01533, 0.00137, 0.02403, 0.01533, 0.02337, 0.03072, 0.01997, -0.03829, -0.00488, 0.01997, -0.03829, 0.00488, 0.01997, 0.01492, -0.0356, 0.01997, 0.02337, -0.03072, 0.01997, 0.01492, 0.0356, 0.01997, 0.04, 0.04518, 0.01932, 0.04, -0.04518, 0.01932, -0.05913, 0.01205, 0.01932, 0.01913, 0.05724, 0.01932, -0.05913, -0.01205, 0.01932, 0.01913, -0.05724, 0.01932]]), "A2":array([[-0.0, 0.00102, 0.0, 0.00089, -0.00051, -0.0, -0.00089, -0.00051, -0.0, 0.0063, 0.0044, -0.00887, 0.00696, 0.00326, 0.00887, -0.00696, 0.00326, -0.00887, 0.00066, -0.00765, -0.00887, -0.0063, 0.0044, 0.00887, -0.00066, -0.00765, 0.00887, 0.01911, 0.0161, -0.02985, 0.0235, 0.0085, 0.02985, -0.0235, 0.0085, -0.02985, 0.00439, -0.0246, -0.02986, -0.01911, 0.0161, 0.02985, -0.00439, -0.0246, 0.02985, 0.02102, 0.00791, -0.02021, 0.01736, 0.01425, 0.02021, -0.01736, 0.01425, -0.0202, -0.00366, -0.02216, -0.02021, -0.02102, 0.00791, 0.02021, 0.00366, -0.02216, 0.0202, 0.03582, 0.02603, -0.04949, -0.03582, 0.02603, 0.0495, -0.04045, 0.01801, -0.04949, -0.00463, -0.04404, 0.04949, 0.04045, 0.01801, 0.0495, 0.00463, -0.04404, -0.0495], [-1e-05, 0.02321, 4e-05, 0.0201, -0.01161, -2e-05, -0.0201, -0.0116, -2e-05, 0.00845, 0.00099, 0.0035, 0.00508, 0.00682, -0.00349, -0.00508, 0.00683, 0.00349, -0.00337, -0.00781, 0.00348, -0.00845, 0.00099, -0.00348, 0.00337, -0.00781, -0.0035, 0.02259, -0.00771, 0.01344, 0.00462, 0.02341, -0.0134, -0.00462, 0.02342, 0.01343, -0.01796, -0.01571, 0.01339, -0.02259, -0.0077, -0.0134, 0.01798, -0.01571, -0.01345, 0.03445, -0.02119, 0.00867, -0.00113, 0.04041, -0.00864, 0.00112, 0.04042, 0.0087, -0.03555, -0.01923, 0.00865, -0.03444, -0.02118, -0.00867, 0.03558, -0.01924, -0.0087, 0.05093, -0.03948, 0.0227, -0.0509, -0.03945, -0.02269, 0.00872, 0.06383, 0.02274, 0.05966, -0.02437, -0.02274, -0.00872, 0.06382, -0.02265, -0.05961, -0.02435, 0.02265]]), "E1":array([[0.00563, 0.0, 0.05971, 0.01124, -0.00324, -0.02986, 0.01124, 0.00324, -0.02986, 0.00022, 0.00097, 0.0238, 0.00875, -0.00188, -0.00155, 0.00875, 0.00188, -0.00155, 0.00527, -0.00596, -0.02226, 0.00022, -0.00097, 0.0238, 0.00527, 0.00596, -0.02226, -0.00791, 0.00134, 0.03255, 0.01021, -0.00387, 0.02861, 0.01021, 0.00387, 0.02861, 0.00333, -0.01308, -0.06116, -0.00791, -0.00134, 0.03255, 0.00333, 0.01308, -0.06116, -0.01001, 0.00747, -0.00469, 0.01199, -0.00185, 0.05876, 0.01199, 0.00185, 0.05876, -0.00388, -0.01439, -0.05408, -0.01001, -0.00747, -0.00469, -0.00388, 0.01439, -0.05408, -0.0182, 0.00817, 0.00269, -0.0182, -0.00817, 0.00269, 0.01465, 0.00318, 0.09064, -0.00609, 0.02277, -0.09333, 0.01465, -0.00318, 0.09064, -0.00609, -0.02277, -0.09333], [0.04384, 0.0, -0.03897, 0.0183, 0.01474, 0.01949, 0.0183, -0.01474, 0.01949, 0.01623, 0.00356, -0.01847, 0.01338, 0.00836, 0.01415, 0.01338, -0.00836, 0.01415, 0.00448, 6e-05, 0.00432, 0.01623, -0.00356, -0.01847, 0.00448, -6e-05, 0.00432, 0.00059, 0.01616, -0.02836, 0.01078, 0.01697, 0.01227, 0.01078, -0.01697, 0.01227, -0.02302, -0.00923, 0.01608, 0.00059, -0.01616, -0.02836, -0.02302, 0.00923, 0.01608, -0.02662, 0.0261, -0.01705, 0.00947, 0.0092, 0.01128, 0.00947, -0.0092, 0.01128, -0.03915, -0.0228, 0.00577, -0.02662, -0.0261, -0.01705, -0.03915, 0.0228, 0.00577, -0.04361, 0.04091, -0.02767, -0.04361, -0.04091, -0.02767, 0.00893, -0.01774, 0.01162, -0.06814, 0.03391, 0.01605, 0.00893, 0.01774, 0.01162, -0.06814, -0.03391, 0.01605]]), "E2":array([[-0.00601, -0.01278, -0.02501, -0.00601, -0.01042, 0.05003, -0.00806, -0.0116, -0.02501, -0.00541, -0.00971, 0.00139, -0.00289, -0.0023, 0.01917, -0.00822, -0.00247, -0.02056, -0.00055, -0.00365, 0.01917, 0.00197, -0.00836, -0.02056, -0.0057, -0.00954, 0.00139, -0.00498, -0.00935, 0.03062, 0.00082, 0.0075, 0.02521, -0.0114, 0.00714, -0.05583, 0.00609, 0.00446, 0.02521, 0.01188, -0.00631, -0.05583, -0.00561, -0.00899, 0.03062, -0.00522, -0.00917, 0.0605, -0.00173, 0.01733, -0.0091, -0.00883, 0.01727, -0.0514, 0.01587, 0.00717, -0.0091, 0.01937, 0.00099, -0.0514, -0.00533, -0.00911, 0.0605, -0.00554, -0.00936, 0.09158, 0.0304, 0.00306, -0.08637, -0.01255, 0.02786, -0.08637, -0.00534, -0.00948, 0.09158, 0.00167, 0.02774, -0.00521, 0.02319, 0.01532, -0.00521], [-0.00121, -0.00245, 0.00074, -0.00121, -0.00209, -0.00147, -0.00152, -0.00227, 0.00074, 0.00384, 0.0061, 0.01622, -4e-05, -0.01144, -0.01315, 0.00725, -0.01172, -0.00307, -0.00989, -0.00576, -0.01315, -0.01378, 0.00042, -0.00307, 0.00336, 0.00638, 0.01622, 0.0112, 0.01963, 0.02377, 0.00357, -0.02139, -0.02403, 0.01903, -0.02128, 0.00025, -0.02032, -0.0076, -0.02403, -0.02795, 0.00584, 0.00025, 0.0114, 0.01952, 0.02377, 0.01993, 0.03441, 0.03016, 0.01557, -0.03143, -0.02204, 0.0242, -0.03148, -0.00812, -0.03501, -0.00224, -0.02204, -0.03936, 0.00522, -0.00812, 0.01984, 0.03446, 0.03016, 0.02936, 0.05025, 0.03499, -0.05543, 0.01188, -0.00292, 0.038, -0.04207, -0.00292, 0.02884, 0.05056, 0.03499, 0.0202, -0.04176, -0.03207, -0.04627, -0.00339, -0.03207]]), "E3":array([[0.00638, -0.00848, 0.01541, 0.00415, -0.00977, 0.01541, 0.00638, -0.01105, -0.03083, -0.00687, -0.00634, 0.01299, 0.00892, 0.00278, 0.01299, 0.00315, 0.00186, -0.01673, 0.00523, -0.01092, 0.00373, 0.00684, -0.00999, 0.00373, -0.00318, -0.0018, -0.01673, -0.02231, -0.00597, 0.03821, 0.01632, 0.01633, 0.03821, 0.00151, 0.0145, -0.02489, 0.00733, -0.01636, -0.01332, 0.01051, -0.01453, -0.01332, -0.01331, 0.00594, -0.02489, -0.03398, -0.00025, 0.03289, 0.0172, 0.0293, 0.03289, 0.00771, 0.02829, -0.00172, 0.01158, -0.02208, -0.03117, 0.01333, -0.02107, -0.03117, -0.02835, 0.00747, -0.00172, -0.05238, -0.0002, 0.06118, 0.01892, -0.02689, -0.05201, 0.00639, 0.04232, -0.00916, -0.03984, 0.01563, -0.00916, 0.02636, 0.04526, 0.06118, 0.01383, -0.02983, -0.05201], [0.02522, -0.00771, -0.02265, -0.00594, -0.02569, -0.02265, 0.02522, -0.04368, 0.04531, 0.01102, -0.00168, -0.00629, -0.00405, -0.01038, -0.00629, 0.0121, -0.01778, 0.02721, -0.00238, -0.01066, -0.02093, 0.01043, -0.00327, -0.02093, 0.00935, -0.01937, 0.02721, -0.00305, 0.02224, -0.02465, -0.01774, 0.01376, -0.02465, 0.01736, -0.00106, 0.04351, -0.01303, -0.00708, -0.01886, 0.01265, 0.00775, -0.01886, -0.00776, -0.01556, 0.04351, -0.01945, 0.02789, -0.00957, -0.01443, 0.03078, -0.00957, 0.005, 0.02249, 0.02648, -0.0119, 0.00402, -0.0169, 0.00247, 0.01232, -0.0169, -0.02197, 0.00692, 0.02648, -0.0336, 0.0522, -0.02763, 0.0037, 0.02358, -0.01408, 0.00983, 0.0402, 0.04171, -0.03973, 0.01158, 0.04171, -0.02841, 0.05519, -0.02763, -0.02227, 0.00859, -0.01408]])}, database_path = 'ctc_scsd_20220713.pkl', maxdist = 1.75, mondrian_limits = (0, 0))
model_objs_dict['perylenediimide'] = scsd_model('perylenediimide', array([['2.8466', '0.0', '0.0', 'C'], ['3.5555', '-1.2245', '0.0', 'C'], ['2.8669', '-2.4096', '0.0', 'C'], ['1.43', '0.0', '0.0', 'C'], ['5.0362', '-1.2366', '0.0', 'C'], ['-1.4726', '-2.424', '-0.0', 'C'], ['-2.8669', '-2.4096', '-0.0', 'C'], ['-3.5555', '-1.2245', '-0.0', 'C'], ['1.4726', '-2.424', '-0.0', 'C'], ['0.7355', '-1.2489', '-0.0', 'C'], ['-0.7355', '-1.2489', '-0.0', 'C'], ['-5.0362', '-1.2366', '-0.0', 'C'], ['1.4726', '2.424', '0.0', 'C'], ['2.8669', '2.4096', '0.0', 'C'], ['3.5555', '1.2245', '0.0', 'C'], ['-1.4726', '2.424', '0.0', 'C'], ['-0.7355', '1.2489', '0.0', 'C'], ['0.7355', '1.2489', '0.0', 'C'], ['5.0362', '1.2366', '0.0', 'C'], ['-2.8466', '0.0', '-0.0', 'C'], ['-3.5555', '1.2245', '-0.0', 'C'], ['-2.8669', '2.4096', '-0.0', 'C'], ['-1.43', '0.0', '-0.0', 'C'], ['-5.0362', '1.2366', '-0.0', 'C'], ['5.6853', '0.0', '0.0', 'N'], ['-5.6853', '0.0', '-0.0', 'N'], ['5.6753', '-2.2737', '0.0', 'O'], ['-5.6753', '-2.2737', '-0.0', 'O'], ['5.6753', '2.2737', '0.0', 'O'], ['-5.6753', '2.2737', '-0.0', 'O']]), 'D2h' , pca = {"Ag":array([[-0.02756, 0.0, -0.0, -0.04735, -0.00568, -0.0, -0.04793, -0.00448, -0.0, -0.00748, 0.0, -0.0, -0.07252, -0.00025, -0.0, 0.04147, 0.00169, 0.0, 0.04793, -0.00448, 0.0, 0.04735, -0.00568, 0.0, -0.04147, 0.00169, -0.0, -0.01081, 0.01276, 0.0, 0.01081, 0.01276, 0.0, 0.07252, -0.00025, 0.0, -0.04147, -0.00169, -0.0, -0.04793, 0.00448, -0.0, -0.04735, 0.00568, -0.0, 0.04147, -0.00169, 0.0, 0.01081, -0.01276, -0.0, -0.01081, -0.01276, -0.0, -0.07252, 0.00025, -0.0, 0.02756, -0.0, 0.0, 0.04735, 0.00568, 0.0, 0.04793, 0.00448, 0.0, 0.00748, -0.0, 0.0, 0.07252, 0.00025, 0.0, -0.0812, 0.0, -0.0, 0.0812, -0.0, 0.0, -0.07893, 0.00664, -0.0, 0.07893, 0.00664, 0.0, -0.07893, -0.00664, -0.0, 0.07893, -0.00664, 0.0], [-0.00181, -0.0, -0.0, -0.00034, 0.04236, 0.0, -1e-05, 0.08969, 0.0, -0.00481, 0.0, -0.0, 0.00169, 0.04557, 0.0, -0.00296, 0.06996, 0.0, 1e-05, 0.08969, 0.0, 0.00034, 0.04236, 0.0, 0.00296, 0.06996, 0.0, -0.00517, 0.00523, 0.0, 0.00517, 0.00523, 0.0, -0.00169, 0.04557, 0.0, 0.00296, -0.06996, -0.0, -1e-05, -0.08969, -0.0, -0.00034, -0.04236, -0.0, -0.00296, -0.06996, -0.0, 0.00517, -0.00523, -0.0, -0.00517, -0.00523, -0.0, 0.00169, -0.04557, -0.0, 0.00181, -0.0, 0.0, 0.00034, -0.04236, -0.0, 1e-05, -0.08969, -0.0, 0.00481, -0.0, 0.0, -0.00169, -0.04557, -0.0, -0.00221, 0.0, -0.0, 0.00221, -0.0, 0.0, 0.00082, 0.08638, 0.0, -0.00082, 0.08638, 0.0, 0.00082, -0.08638, -0.0, -0.00082, -0.08638, -0.0], [0.00394, -0.0, 0.0, 0.02252, 0.00833, 0.0, 0.06645, 0.00193, 0.0, 0.00076, -0.0, 0.0, -0.02388, 0.0065, -0.0, -0.09495, -0.0, -0.0, -0.06645, 0.00193, -0.0, -0.02252, 0.00833, -0.0, 0.09495, -0.0, 0.0, 0.01944, -0.02209, -0.0, -0.01944, -0.02209, -0.0, 0.02388, 0.0065, 0.0, 0.09495, 0.0, 0.0, 0.06645, -0.00193, 0.0, 0.02252, -0.00833, 0.0, -0.09495, 0.0, -0.0, -0.01944, 0.02209, 0.0, 0.01944, 0.02209, 0.0, -0.02388, -0.0065, -0.0, -0.00394, 0.0, -0.0, -0.02252, -0.00833, -0.0, -0.06645, -0.00193, -0.0, -0.00076, 0.0, -0.0, 0.02388, -0.0065, 0.0, -0.02302, 0.0, -0.0, 0.02302, -0.0, 0.0, -0.07844, -0.00919, -0.0, 0.07844, -0.00919, 0.0, -0.07844, 0.00919, -0.0, 0.07844, 0.00919, 0.0], [0.02176, 0.0, -0.0, 0.0138, -0.03436, 0.0, 0.00967, -0.0817, -0.0, 0.01301, -0.0, 0.0, -0.00487, 0.02242, 0.0, -0.00406, -0.0116, -0.0, -0.00967, -0.0817, -0.0, -0.0138, -0.03436, -0.0, 0.00406, -0.0116, -0.0, 0.01758, 0.01088, 0.0, -0.01758, 0.01088, -0.0, 0.00487, 0.02242, 0.0, 0.00406, 0.0116, 0.0, 0.00967, 0.0817, 0.0, 0.0138, 0.03436, 0.0, -0.00406, 0.0116, 0.0, -0.01758, -0.01088, -0.0, 0.01758, -0.01088, 0.0, -0.00487, -0.02242, -0.0, -0.02176, 0.0, -0.0, -0.0138, 0.03436, 0.0, -0.00967, 0.0817, 0.0, -0.01301, 0.0, -0.0, 0.00487, -0.02242, -0.0, -0.01802, 0.0, -0.0, 0.01802, -0.0, 0.0, 0.0076, 0.09995, 0.0, -0.0076, 0.09995, 0.0, 0.0076, -0.09995, -0.0, -0.0076, -0.09995, -0.0], [0.02354, 0.0, -0.0, -0.00135, -0.01779, 0.0, 0.0102, -0.00385, -0.0, 0.00811, -0.0, 0.0, -0.02524, -0.03482, -0.0, -0.02452, 0.05801, 0.0, -0.0102, -0.00385, -0.0, 0.00135, -0.01779, -0.0, 0.02452, 0.05801, 0.0, 0.00893, 0.01195, 0.0, -0.00893, 0.01195, 0.0, 0.02524, -0.03482, -0.0, 0.02452, -0.05801, -0.0, 0.0102, 0.00385, 0.0, -0.00135, 0.01779, 0.0, -0.02452, -0.05801, -0.0, -0.00893, -0.01195, -0.0, 0.00893, -0.01195, -0.0, -0.02524, 0.03482, 0.0, -0.02354, 0.0, -0.0, 0.00135, 0.01779, 0.0, -0.0102, 0.00385, -0.0, -0.00811, 0.0, -0.0, 0.02524, 0.03482, 0.0, -0.10831, 0.0, -0.0, 0.10831, -0.0, 0.0, 0.05821, -0.01789, 0.0, -0.05821, -0.01789, -0.0, 0.05821, 0.01789, 0.0, -0.05821, 0.01789, -0.0], [-0.02327, 0.0, 0.0, -0.00357, 0.01564, 0.0, 0.00545, 0.02451, -0.0, -0.01917, 0.0, -0.0, -0.03109, 0.03066, -0.0, -0.0178, -0.07603, -0.0, -0.00545, 0.02451, 0.0, 0.00357, 0.01564, 0.0, 0.0178, -0.07603, -0.0, -0.02656, -0.02805, -0.0, 0.02656, -0.02805, -0.0, 0.03109, 0.03066, 0.0, 0.0178, 0.07603, 0.0, 0.00545, -0.02451, -0.0, -0.00357, -0.01564, -0.0, -0.0178, 0.07603, 0.0, 0.02656, 0.02805, 0.0, -0.02656, 0.02805, 0.0, -0.03109, -0.03066, -0.0, 0.02327, -0.0, 0.0, 0.00357, -0.01564, -0.0, -0.00545, -0.02451, -0.0, 0.01917, -0.0, 0.0, 0.03109, -0.03066, 0.0, -0.05456, 0.0, -0.0, 0.05456, -0.0, 0.0, 0.04682, 0.01046, 0.0, -0.04682, 0.01046, -0.0, 0.04682, -0.01046, 0.0, -0.04682, -0.01046, -0.0], [-0.02247, -0.0, -0.0, 0.02447, -0.00474, 0.0, -0.03053, -0.02604, 0.0, 0.01518, -0.0, 0.0, -0.04941, 0.05718, 0.0, -0.02806, 0.03111, 0.0, 0.03053, -0.02604, 0.0, -0.02447, -0.00474, -0.0, 0.02806, 0.03111, 0.0, -0.02887, 0.0156, -0.0, 0.02887, 0.0156, 0.0, 0.04941, 0.05718, 0.0, 0.02806, -0.03111, -0.0, -0.03053, 0.02604, -0.0, 0.02447, 0.00474, 0.0, -0.02806, -0.03111, -0.0, 0.02887, -0.0156, 0.0, -0.02887, -0.0156, -0.0, -0.04941, -0.05718, -0.0, 0.02247, -0.0, 0.0, -0.02447, 0.00474, -0.0, 0.03053, 0.02604, 0.0, -0.01518, 0.0, -0.0, 0.04941, -0.05718, -0.0, 0.05214, -0.0, 0.0, -0.05214, 0.0, -0.0, 0.01734, -0.02789, -0.0, -0.01734, -0.02789, -0.0, 0.01734, 0.02789, 0.0, -0.01734, 0.02789, 0.0], [0.01733, 0.0, -0.0, 0.05526, 0.01372, 0.0, -0.01916, 0.00902, -0.0, 0.03559, -0.0, 0.0, -0.04349, -0.07092, -0.0, 0.00382, -0.0039, -0.0, 0.01916, 0.00902, 0.0, -0.05526, 0.01372, -0.0, -0.00382, -0.0039, -0.0, -0.02967, -0.03256, -0.0, 0.02967, -0.03256, -0.0, 0.04349, -0.07092, -0.0, -0.00382, 0.0039, 0.0, -0.01916, -0.00902, -0.0, 0.05526, -0.01372, 0.0, 0.00382, 0.0039, 0.0, 0.02967, 0.03256, 0.0, -0.02967, 0.03256, 0.0, -0.04349, 0.07092, 0.0, -0.01733, 0.0, -0.0, -0.05526, -0.01372, -0.0, 0.01916, -0.00902, 0.0, -0.03559, 0.0, -0.0, 0.04349, 0.07092, 0.0, 0.03332, -0.0, 0.0, -0.03332, 0.0, -0.0, -0.00165, 0.02749, 0.0, 0.00165, 0.02749, 0.0, -0.00165, -0.02749, -0.0, 0.00165, -0.02749, -0.0], [0.07934, 0.0, -0.0, 0.06917, 0.01991, -0.0, 0.00249, 0.00919, 0.0, 0.00559, 0.0, -0.0, -0.01305, 0.03969, 0.0, 0.04893, -0.01287, 0.0, -0.00249, 0.00919, 0.0, -0.06917, 0.01991, -0.0, -0.04893, -0.01287, -0.0, 0.04391, 0.00915, 0.0, -0.04391, 0.00915, -0.0, 0.01305, 0.03969, 0.0, -0.04893, 0.01287, -0.0, 0.00249, -0.00919, -0.0, 0.06917, -0.01991, 0.0, 0.04893, 0.01287, 0.0, -0.04391, -0.00915, -0.0, 0.04391, -0.00915, 0.0, -0.01305, -0.03969, -0.0, -0.07934, 0.0, -0.0, -0.06917, -0.01991, -0.0, -0.00249, -0.00919, -0.0, -0.00559, 0.0, -0.0, 0.01305, -0.03969, -0.0, -0.03813, 0.0, -0.0, 0.03813, -0.0, 0.0, -0.00881, -0.02497, -0.0, 0.00881, -0.02497, -0.0, -0.00881, 0.02497, 0.0, 0.00881, 0.02497, 0.0], [-0.02746, -0.0, -0.0, 0.01945, -0.01026, -0.0, 0.09088, -0.01379, 0.0, -0.05607, -0.0, -0.0, -0.01262, 0.00316, -0.0, 0.05788, 0.02206, 0.0, -0.09088, -0.01379, -0.0, -0.01945, -0.01026, -0.0, -0.05788, 0.02206, -0.0, -0.06988, 0.00719, -0.0, 0.06988, 0.00719, 0.0, 0.01262, 0.00316, 0.0, -0.05788, -0.02206, -0.0, 0.09088, 0.01379, 0.0, 0.01945, 0.01026, 0.0, 0.05788, -0.02206, 0.0, 0.06988, -0.00719, 0.0, -0.06988, -0.00719, -0.0, -0.01262, -0.00316, -0.0, 0.02746, -0.0, 0.0, -0.01945, 0.01026, -0.0, -0.09088, 0.01379, -0.0, 0.05607, -0.0, 0.0, 0.01262, -0.00316, 0.0, -0.00119, 0.0, -0.0, 0.00119, -0.0, 0.0, -0.00444, -0.00431, -0.0, 0.00444, -0.00431, 0.0, -0.00444, 0.00431, -0.0, 0.00444, 0.00431, 0.0], [-0.02642, -0.0, -0.0, -0.0358, 0.03972, 0.0, 0.03723, -0.01028, 0.0, -0.0188, -0.0, -0.0, -0.06315, -0.01434, 0.0, 0.01887, 0.00223, 0.0, -0.03723, -0.01028, -0.0, 0.0358, 0.03972, 0.0, -0.01887, 0.00223, -0.0, 0.07183, -0.0013, 0.0, -0.07183, -0.0013, -0.0, 0.06315, -0.01434, 0.0, -0.01887, -0.00223, -0.0, 0.03723, 0.01028, 0.0, -0.0358, -0.03972, -0.0, 0.01887, -0.00223, 0.0, -0.07183, 0.0013, -0.0, 0.07183, 0.0013, 0.0, -0.06315, 0.01434, -0.0, 0.02642, -0.0, 0.0, 0.0358, -0.03972, -0.0, -0.03723, 0.01028, -0.0, 0.0188, -0.0, 0.0, 0.06315, 0.01434, 0.0, 0.06207, -0.0, 0.0, -0.06207, 0.0, -0.0, 0.02839, 0.0028, 0.0, -0.02839, 0.0028, -0.0, 0.02839, -0.0028, 0.0, -0.02839, -0.0028, -0.0], [-0.00287, 0.0, -0.0, 0.01906, -0.08101, 0.0, -0.00162, 0.05095, -0.0, -0.05606, 0.0, -0.0, -0.02389, -0.01304, -0.0, -0.01019, -0.02106, -0.0, 0.00162, 0.05095, 0.0, -0.01906, -0.08101, -0.0, 0.01019, -0.02106, -0.0, 0.02078, 0.04746, 0.0, -0.02078, 0.04746, 0.0, 0.02389, -0.01304, 0.0, 0.01019, 0.02106, 0.0, -0.00162, -0.05095, -0.0, 0.01906, 0.08101, 0.0, -0.01019, 0.02106, 0.0, -0.02078, -0.04746, -0.0, 0.02078, -0.04746, -0.0, -0.02389, 0.01304, -0.0, 0.00287, -0.0, 0.0, -0.01906, 0.08101, 0.0, 0.00162, -0.05095, -0.0, 0.05606, -0.0, 0.0, 0.02389, 0.01304, 0.0, 0.03144, -0.0, 0.0, -0.03144, 0.0, -0.0, 0.00117, 0.00812, 0.0, -0.00117, 0.00812, 0.0, 0.00117, -0.00812, -0.0, -0.00117, -0.00812, -0.0], [-0.08182, -0.0, -0.0, 0.0271, 0.03859, 0.0, 0.0108, 0.00063, -0.0, 0.06563, 0.0, 0.0, 0.01118, -0.01801, -0.0, -0.00482, -0.02302, -0.0, -0.0108, 0.00063, -0.0, -0.0271, 0.03859, 0.0, 0.00482, -0.02302, -0.0, 0.00049, 0.08737, 0.0, -0.00049, 0.08737, 0.0, -0.01118, -0.01801, -0.0, 0.00482, 0.02302, 0.0, 0.0108, -0.00063, 0.0, 0.0271, -0.03859, -0.0, -0.00482, 0.02302, 0.0, -0.00049, -0.08737, -0.0, 0.00049, -0.08737, -0.0, 0.01118, 0.01801, 0.0, 0.08182, -0.0, 0.0, -0.0271, -0.03859, -0.0, -0.0108, -0.00063, -0.0, -0.06563, 0.0, -0.0, -0.01118, 0.01801, 0.0, -0.02352, 0.0, -0.0, 0.02352, -0.0, 0.0, -0.00122, 0.00372, 0.0, 0.00122, 0.00372, 0.0, -0.00122, -0.00372, -0.0, 0.00122, -0.00372, -0.0], [0.1304, 0.0, 0.0, -0.0511, 0.01593, 0.0, 0.0209, 0.007, -0.0, 0.0387, 0.0, 0.0, -0.01948, -0.00303, 0.0, -0.0134, -0.02106, -0.0, -0.0209, 0.007, -0.0, 0.0511, 0.01593, 0.0, 0.0134, -0.02106, -0.0, -0.04119, 0.05263, 0.0, 0.04119, 0.05263, 0.0, 0.01948, -0.00303, 0.0, 0.0134, 0.02106, 0.0, 0.0209, -0.007, 0.0, -0.0511, -0.01593, -0.0, -0.0134, 0.02106, 0.0, 0.04119, -0.05263, -0.0, -0.04119, -0.05263, -0.0, -0.01948, 0.00303, -0.0, -0.1304, 0.0, -0.0, 0.0511, -0.01593, 0.0, -0.0209, -0.007, -0.0, -0.0387, 0.0, -0.0, 0.01948, 0.00303, 0.0, 0.03064, -0.0, 0.0, -0.03064, 0.0, -0.0, 0.00067, 0.00048, 0.0, -0.00067, 0.00048, -0.0, 0.00067, -0.00048, 0.0, -0.00067, -0.00048, -0.0], [-0.02403, -0.0, 0.0, -0.0175, -0.0587, 0.0, 0.03516, 0.03, -0.0, 0.15249, -0.0, 0.0, -0.01025, 0.02269, 0.0, 0.02787, -0.00054, 0.0, -0.03516, 0.03, -0.0, 0.0175, -0.0587, -0.0, -0.02787, -0.00054, -0.0, 0.00966, -0.03328, -0.0, -0.00966, -0.03328, -0.0, 0.01025, 0.02269, 0.0, -0.02787, 0.00054, -0.0, 0.03516, -0.03, 0.0, -0.0175, 0.0587, 0.0, 0.02787, 0.00054, 0.0, -0.00966, 0.03328, 0.0, 0.00966, 0.03328, 0.0, -0.01025, -0.02269, -0.0, 0.02403, -0.0, 0.0, 0.0175, 0.0587, 0.0, -0.03516, -0.03, -0.0, -0.15249, 0.0, -0.0, 0.01025, -0.02269, -0.0, 0.0061, -0.0, 0.0, -0.0061, 0.0, -0.0, 0.00225, -0.00617, -0.0, -0.00225, -0.00617, -0.0, 0.00225, 0.00617, 0.0, -0.00225, 0.00617, 0.0]]), "B1g":array([[-0.0, 0.00675, 0.0, -0.02133, 0.01423, -0.0, -0.05212, 9e-05, -0.0, 0.0, 0.00705, 0.0, -0.02772, 0.03902, -0.0, -0.04044, 0.00062, -0.0, -0.05212, -9e-05, -0.0, -0.02133, -0.01423, -0.0, -0.04044, -0.00062, -0.0, -0.00959, 0.00254, 0.0, -0.00959, -0.00254, -0.0, -0.02772, -0.03902, -0.0, 0.04044, -0.00062, 0.0, 0.05212, 9e-05, 0.0, 0.02133, 0.01423, 0.0, 0.04044, 0.00062, 0.0, 0.00959, -0.00254, -0.0, 0.00959, 0.00254, 0.0, 0.02772, 0.03902, 0.0, -0.0, -0.00675, -0.0, 0.02133, -0.01423, 0.0, 0.05212, -9e-05, 0.0, -0.0, -0.00705, -0.0, 0.02772, -0.03902, 0.0, -0.0, 0.05347, 0.0, -0.0, -0.05347, -0.0, -0.05043, 0.05129, 0.0, -0.05043, -0.05129, -0.0, 0.05043, 0.05129, 0.0, 0.05043, -0.05129, -0.0], [0.0, 0.02237, -0.0, -0.01355, 0.01552, 0.0, -0.00227, 0.03739, -0.0, 0.0, 0.02304, 0.0, -0.03228, -0.01886, 0.0, 0.01048, -0.04525, -0.0, -0.00227, -0.03739, -0.0, -0.01355, -0.01552, -0.0, 0.01048, 0.04525, -0.0, 0.00307, 0.0174, 0.0, 0.00307, -0.0174, -0.0, -0.03228, 0.01886, -0.0, -0.01048, 0.04525, 0.0, 0.00227, 0.03739, 0.0, 0.01355, 0.01552, 0.0, -0.01048, -0.04525, 0.0, -0.00307, -0.0174, -0.0, -0.00307, 0.0174, 0.0, 0.03228, -0.01886, 0.0, 0.0, -0.02237, -0.0, 0.01355, -0.01552, -0.0, 0.00227, -0.03739, 0.0, 0.0, -0.02304, -0.0, 0.03228, 0.01886, -0.0, 0.0, -0.04607, 0.0, 0.0, 0.04607, -0.0, -0.06815, -0.05293, 0.0, -0.06815, 0.05293, -0.0, 0.06815, -0.05293, 0.0, 0.06815, 0.05293, -0.0], [0.0, 0.00913, -0.0, 0.01363, 0.01727, 0.0, 0.06262, 0.01017, 0.0, 0.0, -0.03401, 0.0, -0.01102, 0.03065, 0.0, 0.04964, 0.03375, 0.0, 0.06262, -0.01017, -0.0, 0.01363, -0.01727, -0.0, 0.04964, -0.03375, 0.0, 0.01237, -0.02477, -0.0, 0.01237, 0.02477, -0.0, -0.01102, -0.03065, -0.0, -0.04964, -0.03375, -0.0, -0.06262, 0.01017, 0.0, -0.01363, 0.01727, 0.0, -0.04964, 0.03375, -0.0, -0.01237, 0.02477, 0.0, -0.01237, -0.02477, 0.0, 0.01102, 0.03065, 0.0, 0.0, -0.00913, -0.0, -0.01363, -0.01727, -0.0, -0.06262, -0.01017, -0.0, 0.0, 0.03401, -0.0, 0.01102, -0.03065, -0.0, 0.0, -0.00287, 0.0, 0.0, 0.00287, -0.0, -0.04824, 0.0333, 0.0, -0.04824, -0.0333, -0.0, 0.04824, 0.0333, 0.0, 0.04824, -0.0333, -0.0], [-0.0, -0.00816, -0.0, -0.01625, -0.01901, 0.0, -0.02038, -0.03355, -0.0, 0.0, -0.04029, 0.0, -0.01861, -0.03534, -0.0, -0.02353, 0.07365, 0.0, -0.02038, 0.03355, -0.0, -0.01625, 0.01901, -0.0, -0.02353, -0.07365, -0.0, -0.0098, -0.03968, -0.0, -0.0098, 0.03968, 0.0, -0.01861, 0.03534, -0.0, 0.02353, -0.07365, -0.0, 0.02038, -0.03355, 0.0, 0.01625, -0.01901, 0.0, 0.02353, 0.07365, 0.0, 0.0098, 0.03968, 0.0, 0.0098, -0.03968, -0.0, 0.01861, -0.03534, 0.0, 0.0, 0.00816, 0.0, 0.01625, 0.01901, 0.0, 0.02038, 0.03355, 0.0, 0.0, 0.04029, -0.0, 0.01861, 0.03534, 0.0, 0.0, -0.02297, 0.0, 0.0, 0.02297, -0.0, -0.03201, -0.04082, 0.0, -0.03201, 0.04082, -0.0, 0.03201, -0.04082, 0.0, 0.03201, 0.04082, -0.0], [0.0, -0.03305, 0.0, 0.01465, -0.05185, -0.0, 0.01037, -0.07225, 0.0, 0.0, 0.01143, 0.0, 0.00268, -0.0272, 0.0, 0.0297, -0.03596, -0.0, 0.01037, 0.07225, -0.0, 0.01465, 0.05185, -0.0, 0.0297, 0.03596, -0.0, -0.00199, 0.02133, -0.0, -0.00199, -0.02133, 0.0, 0.00268, 0.0272, -0.0, -0.0297, 0.03596, 0.0, -0.01037, -0.07225, 0.0, -0.01465, -0.05185, 0.0, -0.0297, -0.03596, 0.0, 0.00199, -0.02133, 0.0, 0.00199, 0.02133, -0.0, -0.00268, -0.0272, 0.0, 0.0, 0.03305, -0.0, -0.01465, 0.05185, -0.0, -0.01037, 0.07225, 0.0, 0.0, -0.01143, -0.0, -0.00268, 0.0272, -0.0, 0.0, 0.0501, 0.0, 0.0, -0.0501, -0.0, -0.04234, 0.00936, 0.0, -0.04234, -0.00936, -0.0, 0.04234, 0.00936, 0.0, 0.04234, -0.00936, -0.0], [-0.0, 0.00112, -0.0, -0.0015, 0.07025, -0.0, 0.01126, -0.00091, 0.0, -0.0, 0.02105, 0.0, 0.01637, -0.02394, -0.0, 0.0123, -0.00902, 0.0, 0.01126, 0.00091, 0.0, -0.0015, -0.07025, 0.0, 0.0123, 0.00902, 0.0, -0.02411, -0.04744, -0.0, -0.02411, 0.04744, 0.0, 0.01637, 0.02394, 0.0, -0.0123, 0.00902, -0.0, -0.01126, -0.00091, -0.0, 0.0015, 0.07025, -0.0, -0.0123, -0.00902, -0.0, 0.02411, 0.04744, 0.0, 0.02411, -0.04744, -0.0, -0.01637, -0.02394, -0.0, -0.0, -0.00112, 0.0, 0.0015, -0.07025, 0.0, -0.01126, 0.00091, 0.0, -0.0, -0.02105, -0.0, -0.01637, 0.02394, 0.0, -0.0, 0.13805, -0.0, -0.0, -0.13805, 0.0, 0.00165, -0.04582, -0.0, 0.00165, 0.04582, 0.0, -0.00165, -0.04582, -0.0, -0.00165, 0.04582, 0.0], [-0.0, -0.00803, 0.0, 0.01032, 0.05629, -0.0, 0.01058, -0.02911, 0.0, 0.0, 0.06066, -0.0, -0.03891, -0.06806, 0.0, -0.00998, -0.00047, -0.0, 0.01058, 0.02911, -0.0, 0.01032, -0.05629, -0.0, -0.00998, 0.00047, 0.0, 0.00409, -0.01058, 0.0, 0.00409, 0.01058, -0.0, -0.03891, 0.06806, -0.0, 0.00998, 0.00047, 0.0, -0.01058, -0.02911, 0.0, -0.01032, 0.05629, 0.0, 0.00998, -0.00047, -0.0, -0.00409, 0.01058, -0.0, -0.00409, -0.01058, 0.0, 0.03891, -0.06806, 0.0, 0.0, 0.00803, -0.0, -0.01032, -0.05629, -0.0, -0.01058, 0.02911, -0.0, 0.0, -0.06066, 0.0, 0.03891, 0.06806, -0.0, 0.0, -0.04989, 0.0, 0.0, 0.04989, -0.0, 0.01023, 0.05607, 0.0, 0.01023, -0.05607, -0.0, -0.01023, 0.05607, 0.0, -0.01023, -0.05607, -0.0], [0.0, 0.00836, 0.0, -0.02582, 0.04146, -0.0, -0.00515, -0.07266, 0.0, 0.0, -0.04243, -0.0, -0.00696, 0.04864, 0.0, -0.00152, -0.03966, -0.0, -0.00515, 0.07266, -0.0, -0.02582, -0.04146, -0.0, -0.00152, 0.03966, 0.0, 0.04586, -0.02319, 0.0, 0.04586, 0.02319, -0.0, -0.00696, -0.04864, 0.0, 0.00152, 0.03966, 0.0, 0.00515, -0.07266, 0.0, 0.02582, 0.04146, 0.0, 0.00152, -0.03966, -0.0, -0.04586, 0.02319, -0.0, -0.04586, -0.02319, 0.0, 0.00696, 0.04864, -0.0, 0.0, -0.00836, -0.0, 0.02582, -0.04146, 0.0, 0.00515, 0.07266, -0.0, 0.0, 0.04243, 0.0, 0.00696, -0.04864, -0.0, 0.0, -0.03892, -0.0, 0.0, 0.03892, 0.0, 0.01284, -0.02384, 0.0, 0.01284, 0.02384, 0.0, -0.01284, -0.02384, -0.0, -0.01284, 0.02384, -0.0], [-0.0, -0.07129, 0.0, 0.0147, -0.02775, 0.0, 0.01988, 0.03271, -0.0, 0.0, -0.05749, -0.0, -0.02285, -0.01214, 0.0, -0.04047, -0.05375, 0.0, 0.01988, -0.03271, 0.0, 0.0147, 0.02775, 0.0, -0.04047, 0.05375, -0.0, 0.0028, -0.06557, -0.0, 0.0028, 0.06557, -0.0, -0.02285, 0.01214, 0.0, 0.04047, 0.05375, -0.0, -0.01988, 0.03271, -0.0, -0.0147, -0.02775, -0.0, 0.04047, -0.05375, 0.0, -0.0028, 0.06557, 0.0, -0.0028, -0.06557, 0.0, 0.02285, -0.01214, -0.0, 0.0, 0.07129, -0.0, -0.0147, 0.02775, -0.0, -0.01988, -0.03271, -0.0, 0.0, 0.05749, -0.0, 0.02285, 0.01214, 0.0, 0.0, 0.00144, -0.0, 0.0, -0.00144, 0.0, 0.00252, 0.01372, 0.0, 0.00252, -0.01372, 0.0, -0.00252, 0.01372, -0.0, -0.00252, -0.01372, -0.0], [0.0, -0.11217, 0.0, -0.0258, 0.01375, 0.0, -0.01988, 0.02432, 0.0, 0.0, 0.00123, 0.0, -0.00539, -0.01326, -0.0, 0.02182, 0.0252, -0.0, -0.01988, -0.02432, -0.0, -0.0258, -0.01375, -0.0, 0.02182, -0.0252, -0.0, 0.0802, 0.028, -0.0, 0.0802, -0.028, 0.0, -0.00539, 0.01326, -0.0, -0.02182, -0.0252, 0.0, 0.01988, 0.02432, 0.0, 0.0258, 0.01375, 0.0, -0.02182, 0.0252, 0.0, -0.0802, -0.028, 0.0, -0.0802, 0.028, -0.0, 0.00539, -0.01326, 0.0, 0.0, 0.11217, 0.0, 0.0258, -0.01375, 0.0, 0.01988, -0.02432, 0.0, 0.0, -0.00123, 0.0, 0.00539, 0.01326, 0.0, 0.0, 0.03007, -0.0, 0.0, -0.03007, 0.0, 0.001, -0.0045, -0.0, 0.001, 0.0045, -0.0, -0.001, -0.0045, 0.0, -0.001, 0.0045, 0.0], [-0.0, 0.00256, 0.0, -0.04903, 0.00344, -0.0, -0.03157, 0.01009, 0.0, 0.0, -0.01643, 0.0, 0.07536, -0.0272, 0.0, 0.03202, -0.01681, 0.0, -0.03157, -0.01009, 0.0, -0.04903, -0.00344, -0.0, 0.03202, 0.01681, -0.0, -0.00802, -0.03296, 0.0, -0.00802, 0.03296, 0.0, 0.07536, 0.0272, -0.0, -0.03202, 0.01681, -0.0, 0.03157, 0.01009, -0.0, 0.04903, 0.00344, 0.0, -0.03202, -0.01681, 0.0, 0.00802, 0.03296, -0.0, 0.00802, -0.03296, -0.0, -0.07536, -0.0272, 0.0, 0.0, -0.00256, -0.0, 0.04903, -0.00344, 0.0, 0.03157, -0.01009, -0.0, 0.0, 0.01643, -0.0, -0.07536, 0.0272, -0.0, 0.0, -0.035, 0.0, 0.0, 0.035, -0.0, -0.01999, 0.03311, 0.0, -0.01999, -0.03311, -0.0, 0.01999, 0.03311, 0.0, 0.01999, -0.03311, -0.0], [-0.0, 0.06442, -0.0, 0.00408, -0.02787, 0.0, -0.04859, 0.01191, -0.0, 0.0, -0.01309, 0.0, -0.05463, -0.01359, -0.0, 0.06154, -0.00709, -0.0, -0.04859, -0.01191, 0.0, 0.00408, 0.02787, 0.0, 0.06154, 0.00709, -0.0, 0.02162, -0.03176, -0.0, 0.02162, 0.03176, 0.0, -0.05463, 0.01359, 0.0, -0.06154, 0.00709, 0.0, 0.04859, 0.01191, -0.0, -0.00408, -0.02787, -0.0, -0.06154, -0.00709, 0.0, -0.02162, 0.03176, 0.0, -0.02162, -0.03176, -0.0, 0.05463, -0.01359, -0.0, 0.0, -0.06442, 0.0, -0.00408, 0.02787, 0.0, 0.04859, -0.01191, 0.0, 0.0, 0.01309, -0.0, 0.05463, 0.01359, 0.0, 0.0, 0.02519, -0.0, 0.0, -0.02519, 0.0, 0.03069, 0.00878, -0.0, 0.03069, -0.00878, 0.0, -0.03069, 0.00878, -0.0, -0.03069, -0.00878, 0.0], [-0.0, 0.08182, 0.0, -0.04272, -0.02647, 0.0, 0.04392, 0.0049, -0.0, -0.0, 0.01392, -0.0, 0.0104, -0.02539, 0.0, -0.04252, 0.00016, -0.0, 0.04392, -0.0049, -0.0, -0.04272, 0.02647, -0.0, -0.04252, -0.00016, -0.0, 0.05921, -0.00137, -0.0, 0.05921, 0.00137, -0.0, 0.0104, 0.02539, -0.0, 0.04252, -0.00016, 0.0, -0.04392, 0.0049, 0.0, 0.04272, -0.02647, 0.0, 0.04252, 0.00016, 0.0, -0.05921, 0.00137, 0.0, -0.05921, -0.00137, 0.0, -0.0104, -0.02539, 0.0, 0.0, -0.08182, -0.0, 0.04272, 0.02647, 0.0, -0.04392, -0.0049, 0.0, 0.0, -0.01392, -0.0, -0.0104, 0.02539, 0.0, 0.0, 0.03789, 0.0, 0.0, -0.03789, -0.0, -0.00217, 0.00852, -0.0, -0.00217, -0.00852, -0.0, 0.00217, 0.00852, 0.0, 0.00217, -0.00852, 0.0], [0.0, -0.03174, -0.0, -0.00252, -0.038, 0.0, -0.00011, -0.0025, 0.0, -0.0, 0.17225, 0.0, 0.00783, 0.04987, 0.0, 0.00556, 0.01022, -0.0, -0.00011, 0.0025, 0.0, -0.00252, 0.038, 0.0, 0.00556, -0.01022, -0.0, 0.01565, -0.07093, -0.0, 0.01565, 0.07093, 0.0, 0.00783, -0.04987, 0.0, -0.00556, -0.01022, 0.0, 0.00011, -0.0025, -0.0, 0.00252, -0.038, -0.0, -0.00556, 0.01022, 0.0, -0.01565, 0.07093, 0.0, -0.01565, -0.07093, -0.0, -0.00783, 0.04987, -0.0, -0.0, 0.03174, 0.0, 0.00252, 0.038, -0.0, 0.00011, 0.0025, -0.0, -0.0, -0.17225, 0.0, -0.00783, -0.04987, -0.0, -0.0, -0.02525, 0.0, -0.0, 0.02525, -0.0, -0.00567, -0.01579, 0.0, -0.00567, 0.01579, -0.0, 0.00567, -0.01579, 0.0, 0.00567, 0.01579, -0.0]]), "B2g":array([[-0.0, 0.0, -0.0, -0.0, -0.0, 0.03302, 0.0, -0.0, 0.04586, -0.0, 0.0, -0.0, 0.0, -0.0, 0.0667, -0.0, -0.0, 0.03282, 0.0, -0.0, 0.04586, -0.0, -0.0, 0.03302, -0.0, -0.0, 0.03282, 0.0, -0.0, 0.01746, -0.0, -0.0, 0.01746, -0.0, -0.0, 0.0667, 0.0, -0.0, -0.03282, -0.0, 0.0, -0.04586, -0.0, 0.0, -0.03302, 0.0, -0.0, -0.03282, -0.0, -0.0, -0.01746, 0.0, -0.0, -0.01746, 0.0, 0.0, -0.0667, -0.0, -0.0, -0.0, -0.0, 0.0, -0.03302, -0.0, 0.0, -0.04586, -0.0, -0.0, 0.0, -0.0, 0.0, -0.0667, -0.0, 0.0, 0.0, -0.0, -0.0, -0.0, -0.0, -0.0, 0.14638, -0.0, -0.0, 0.14638, -0.0, 0.0, -0.14638, -0.0, 0.0, -0.14638], [0.0, -0.0, -0.0, 0.0, -0.0, -0.03475, 0.0, -0.0, -0.09088, -0.0, 0.0, -0.0, 0.0, 0.0, 0.01328, 0.0, -0.0, -0.10701, -0.0, -0.0, -0.09088, -0.0, 0.0, -0.03475, -0.0, -0.0, -0.10701, -0.0, -0.0, -0.05455, -0.0, -0.0, -0.05455, -0.0, 0.0, 0.01328, -0.0, -0.0, 0.10701, 0.0, 0.0, 0.09088, -0.0, -0.0, 0.03475, 0.0, -0.0, 0.10701, 0.0, -0.0, 0.05455, 0.0, -0.0, 0.05455, 0.0, -0.0, -0.01328, -0.0, -0.0, -0.0, -0.0, -0.0, 0.03475, -0.0, 0.0, 0.09088, -0.0, -0.0, 0.0, -0.0, -0.0, -0.01328, -0.0, 0.0, 0.0, -0.0, -0.0, -0.0, -0.0, -0.0, 0.06076, -0.0, -0.0, 0.06076, -0.0, 0.0, -0.06076, -0.0, 0.0, -0.06076], [0.0, -0.0, 0.0, -0.0, -0.0, 0.06335, 0.0, -0.0, 0.09719, -0.0, -0.0, -0.0, -0.0, 0.0, 0.01353, -0.0, -0.0, -0.07678, -0.0, -0.0, 0.09719, -0.0, -0.0, 0.06335, -0.0, -0.0, -0.07678, 0.0, -0.0, -0.07582, 0.0, -0.0, -0.07582, 0.0, 0.0, 0.01353, 0.0, -0.0, 0.07678, 0.0, 0.0, -0.09719, -0.0, 0.0, -0.06335, 0.0, -0.0, 0.07678, -0.0, -0.0, 0.07582, -0.0, -0.0, 0.07582, -0.0, -0.0, -0.01353, -0.0, 0.0, 0.0, -0.0, 0.0, -0.06335, -0.0, 0.0, -0.09719, -0.0, 0.0, 0.0, 0.0, -0.0, -0.01353, -0.0, -0.0, 0.0, -0.0, 0.0, -0.0, -0.0, 0.0, -0.02464, -0.0, 0.0, -0.02464, -0.0, -0.0, 0.02464, -0.0, -0.0, 0.02464], [-0.0, -0.0, -0.0, 0.0, 0.0, -0.00281, -0.0, 0.0, -0.00703, -0.0, -0.0, -0.0, -0.0, 0.0, 0.1276, 0.0, -0.0, -0.0468, -0.0, 0.0, -0.00703, -0.0, -0.0, -0.00281, -0.0, -0.0, -0.0468, -0.0, -0.0, 0.07632, 0.0, -0.0, 0.07632, 0.0, 0.0, 0.1276, -0.0, -0.0, 0.0468, 0.0, -0.0, 0.00703, -0.0, 0.0, 0.00281, 0.0, -0.0, 0.0468, 0.0, -0.0, -0.07632, -0.0, -0.0, -0.07632, -0.0, -0.0, -0.1276, -0.0, 0.0, 0.0, -0.0, 0.0, 0.00281, 0.0, -0.0, 0.00703, -0.0, 0.0, 0.0, 0.0, -0.0, -0.1276, -0.0, -0.0, -0.0, -0.0, 0.0, 0.0, -0.0, 0.0, -0.05392, -0.0, 0.0, -0.05392, -0.0, -0.0, 0.05392, -0.0, -0.0, 0.05392], [-0.0, -0.0, -0.0, 0.0, 0.0, -0.04283, -0.0, -0.0, -0.01339, 0.0, 0.0, 0.0, 0.0, -0.0, 0.07806, 0.0, 0.0, 0.07434, 0.0, -0.0, -0.01339, 0.0, 0.0, -0.04283, -0.0, 0.0, 0.07434, -0.0, 0.0, -0.10593, 0.0, 0.0, -0.10593, -0.0, -0.0, 0.07806, -0.0, 0.0, -0.07434, -0.0, 0.0, 0.01339, 0.0, -0.0, 0.04283, 0.0, 0.0, -0.07434, 0.0, 0.0, 0.10593, -0.0, 0.0, 0.10593, 0.0, 0.0, -0.07806, 0.0, 0.0, -0.0, 0.0, -0.0, 0.04283, 0.0, 0.0, 0.01339, 0.0, -0.0, -0.0, -0.0, 0.0, -0.07806, 0.0, 0.0, -0.0, 0.0, -0.0, 0.0, 0.0, 0.0, -0.02575, 0.0, 0.0, -0.02575, 0.0, -0.0, 0.02575, 0.0, -0.0, 0.02575]]), "B3g":array([[0.0, 0.0, 0.01952, 0.0, 0.0, 0.0159, 0.0, -0.0, 0.05762, -0.0, -0.0, 0.01907, -0.0, -0.0, -0.05556, -0.0, 0.0, -0.06675, -0.0, -0.0, -0.05762, -0.0, 0.0, -0.0159, -0.0, -0.0, 0.06675, -0.0, 0.0, 0.01738, -0.0, 0.0, -0.01738, 0.0, -0.0, 0.05556, 0.0, -0.0, 0.06675, 0.0, 0.0, 0.05762, -0.0, -0.0, 0.0159, 0.0, -0.0, -0.06675, 0.0, -0.0, -0.01738, 0.0, -0.0, 0.01738, -0.0, 0.0, -0.05556, -0.0, 0.0, -0.01952, -0.0, -0.0, -0.0159, -0.0, 0.0, -0.05762, -0.0, 0.0, -0.01907, 0.0, 0.0, 0.05556, -0.0, 0.0, -0.0638, -0.0, -0.0, 0.0638, -0.0, -0.0, -0.11667, -0.0, -0.0, 0.11667, -0.0, 0.0, -0.11667, -0.0, 0.0, 0.11667], [0.0, 0.0, 0.01351, 0.0, -0.0, 0.03782, 0.0, -0.0, 0.08558, 0.0, -0.0, 0.00066, -0.0, -0.0, 0.05005, 0.0, -0.0, -0.0981, 0.0, -0.0, -0.08558, 0.0, -0.0, -0.03782, -0.0, 0.0, 0.0981, 0.0, 0.0, 0.01425, 0.0, 0.0, -0.01425, 0.0, -0.0, -0.05005, -0.0, 0.0, 0.0981, -0.0, 0.0, 0.08558, 0.0, 0.0, 0.03782, 0.0, 0.0, -0.0981, -0.0, -0.0, -0.01425, -0.0, -0.0, 0.01425, -0.0, 0.0, 0.05005, 0.0, 0.0, -0.01351, 0.0, 0.0, -0.03782, -0.0, 0.0, -0.08558, 0.0, 0.0, -0.00066, 0.0, 0.0, -0.05005, 0.0, -0.0, 0.02806, 0.0, 0.0, -0.02806, 0.0, 0.0, 0.07534, 0.0, 0.0, -0.07534, 0.0, -0.0, 0.07534, 0.0, -0.0, -0.07534], [-0.0, 0.0, 0.00944, 0.0, 0.0, 0.01052, 0.0, -0.0, 0.00434, 0.0, 0.0, -0.00242, 0.0, 0.0, 0.04026, 0.0, 0.0, -0.00071, 0.0, -0.0, -0.00434, 0.0, 0.0, -0.01052, 0.0, 0.0, 0.00071, -0.0, -0.0, -0.00255, -0.0, -0.0, 0.00255, -0.0, 0.0, -0.04026, -0.0, 0.0, 0.00071, -0.0, 0.0, 0.00434, 0.0, -0.0, 0.01052, -0.0, 0.0, -0.00071, 0.0, 0.0, 0.00255, 0.0, 0.0, -0.00255, 0.0, -0.0, 0.04026, 0.0, 0.0, -0.00944, 0.0, -0.0, -0.01052, -0.0, 0.0, -0.00434, 0.0, -0.0, 0.00242, -0.0, -0.0, -0.04026, 0.0, -0.0, 0.26666, 0.0, 0.0, -0.26666, 0.0, -0.0, -0.08789, 0.0, -0.0, 0.08789, 0.0, 0.0, -0.08789, 0.0, 0.0, 0.08789], [-0.0, 0.0, 0.12268, 0.0, -0.0, 0.09724, 0.0, 0.0, 0.00614, 0.0, -0.0, 0.09133, 0.0, 0.0, 0.04062, 0.0, -0.0, 0.06289, -0.0, 0.0, -0.00614, -0.0, -0.0, -0.09724, 0.0, 0.0, -0.06289, -0.0, 0.0, 0.01899, 0.0, 0.0, -0.01899, -0.0, 0.0, -0.04062, -0.0, 0.0, -0.06289, 0.0, -0.0, 0.00614, 0.0, 0.0, 0.09724, -0.0, 0.0, 0.06289, 0.0, -0.0, -0.01899, -0.0, -0.0, 0.01899, 0.0, -0.0, 0.04062, 0.0, 0.0, -0.12268, 0.0, 0.0, -0.09724, -0.0, -0.0, -0.00614, 0.0, 0.0, -0.09133, -0.0, -0.0, -0.04062, 0.0, 0.0, -0.02977, 0.0, -0.0, 0.02977, 0.0, 0.0, -0.01034, 0.0, 0.0, 0.01034, 0.0, -0.0, -0.01034, 0.0, -0.0, 0.01034], [-0.0, 0.0, 0.06089, -0.0, 0.0, -0.01122, -0.0, -0.0, -0.10036, -0.0, -0.0, 0.09576, 0.0, -0.0, -0.04106, 0.0, -0.0, -0.07807, 0.0, -0.0, 0.10036, -0.0, 0.0, 0.01122, 0.0, 0.0, 0.07807, 0.0, -0.0, 0.0476, 0.0, -0.0, -0.0476, -0.0, -0.0, 0.04106, -0.0, 0.0, 0.07807, -0.0, 0.0, -0.10036, 0.0, -0.0, -0.01122, -0.0, 0.0, -0.07807, -0.0, 0.0, -0.0476, -0.0, 0.0, 0.0476, 0.0, 0.0, -0.04106, 0.0, 0.0, -0.06089, 0.0, -0.0, 0.01122, 0.0, 0.0, 0.10036, 0.0, 0.0, -0.09576, -0.0, 0.0, 0.04106, 0.0, -0.0, 0.03186, 0.0, 0.0, -0.03186, 0.0, 0.0, 0.02443, 0.0, 0.0, -0.02443, 0.0, -0.0, 0.02443, 0.0, -0.0, -0.02443], [-0.0, -0.0, -0.02386, 0.0, 0.0, -0.02976, -0.0, -0.0, -0.04761, 0.0, 0.0, -0.02291, -0.0, -0.0, 0.13068, 0.0, -0.0, -0.03097, -0.0, -0.0, 0.04761, 0.0, 0.0, 0.02976, -0.0, 0.0, 0.03097, -0.0, -0.0, 0.04183, 0.0, -0.0, -0.04183, 0.0, -0.0, -0.13068, -0.0, 0.0, 0.03097, 0.0, 0.0, -0.04761, 0.0, -0.0, -0.02976, 0.0, 0.0, -0.03097, 0.0, 0.0, -0.04183, -0.0, 0.0, 0.04183, -0.0, 0.0, 0.13068, 0.0, -0.0, 0.02386, 0.0, -0.0, 0.02976, 0.0, 0.0, 0.04761, 0.0, -0.0, 0.02291, 0.0, 0.0, -0.13068, 0.0, 0.0, -0.06799, 0.0, -0.0, 0.06799, 0.0, -0.0, -0.05113, 0.0, -0.0, 0.05113, 0.0, 0.0, -0.05113, 0.0, 0.0, 0.05113], [0.0, 0.0, 0.04841, 0.0, -0.0, 0.09127, -0.0, -0.0, -0.06256, 0.0, 0.0, -0.18448, 0.0, 0.0, -0.01792, 0.0, -0.0, -0.03255, -0.0, 0.0, 0.06256, 0.0, -0.0, -0.09127, 0.0, 0.0, 0.03255, 0.0, -0.0, -0.02066, 0.0, -0.0, 0.02066, -0.0, 0.0, 0.01792, -0.0, 0.0, 0.03255, 0.0, -0.0, -0.06256, 0.0, 0.0, 0.09127, -0.0, 0.0, -0.03255, -0.0, 0.0, 0.02066, -0.0, 0.0, -0.02066, 0.0, -0.0, -0.01792, 0.0, 0.0, -0.04841, 0.0, 0.0, -0.09127, 0.0, -0.0, 0.06256, 0.0, -0.0, 0.18448, -0.0, -0.0, 0.01792, 0.0, 0.0, -0.00617, 0.0, -0.0, 0.00617, 0.0, 0.0, -0.00372, 0.0, 0.0, 0.00372, 0.0, -0.0, -0.00372, 0.0, -0.0, 0.00372], [-0.0, 0.0, 0.02927, 0.0, -0.0, -0.02318, -0.0, 0.0, 0.03021, 0.0, 0.0, -0.08003, 0.0, 0.0, -0.02659, 0.0, -0.0, 0.04103, 0.0, -0.0, -0.03021, -0.0, -0.0, 0.02318, 0.0, 0.0, -0.04103, 0.0, 0.0, 0.15068, -0.0, 0.0, -0.15068, -0.0, 0.0, 0.02659, -0.0, 0.0, -0.04103, -0.0, 0.0, 0.03021, 0.0, 0.0, -0.02318, -0.0, 0.0, 0.04103, -0.0, -0.0, -0.15068, 0.0, -0.0, 0.15068, 0.0, -0.0, -0.02659, 0.0, 0.0, -0.02927, 0.0, 0.0, 0.02318, 0.0, 0.0, -0.03021, 0.0, -0.0, 0.08003, -0.0, -0.0, 0.02659, 0.0, -0.0, 0.01944, 0.0, 0.0, -0.01944, 0.0, 0.0, 0.01399, 0.0, 0.0, -0.01399, 0.0, -0.0, 0.01399, 0.0, -0.0, -0.01399]]), "Au":array([[-0.0, 0.0, -0.0, -0.0, -0.0, -0.04399, 0.0, 0.0, -0.09184, 0.0, 0.0, 0.0, -0.0, 0.0, -0.04566, -0.0, -0.0, 0.0816, -0.0, -0.0, 0.09184, -0.0, -0.0, 0.04399, -0.0, -0.0, -0.0816, 0.0, -0.0, -0.02087, 0.0, -0.0, 0.02087, -0.0, -0.0, 0.04566, -0.0, -0.0, 0.0816, -0.0, 0.0, 0.09184, -0.0, -0.0, 0.04399, -0.0, -0.0, -0.0816, -0.0, -0.0, -0.02087, -0.0, -0.0, 0.02087, -0.0, 0.0, 0.04566, -0.0, 0.0, 0.0, -0.0, -0.0, -0.04399, -0.0, -0.0, -0.09184, -0.0, 0.0, -0.0, -0.0, -0.0, -0.04566, -0.0, 0.0, 0.0, -0.0, 0.0, -0.0, -0.0, -0.0, -0.08398, -0.0, 0.0, 0.08398, -0.0, -0.0, 0.08398, -0.0, 0.0, -0.08398], [0.0, 0.0, -0.0, -0.0, -0.0, -0.00869, -0.0, 0.0, -0.06149, 0.0, 0.0, 0.0, -0.0, -0.0, 0.04865, -0.0, -0.0, 0.08011, -0.0, -0.0, 0.06149, -0.0, -0.0, 0.00869, -0.0, -0.0, -0.08011, 0.0, -0.0, -0.01784, -0.0, -0.0, 0.01784, -0.0, 0.0, -0.04865, -0.0, -0.0, 0.08011, -0.0, 0.0, 0.06149, -0.0, -0.0, 0.00869, -0.0, -0.0, -0.08011, -0.0, -0.0, -0.01784, 0.0, -0.0, 0.01784, -0.0, -0.0, -0.04865, -0.0, 0.0, -0.0, -0.0, -0.0, -0.00869, -0.0, -0.0, -0.06149, -0.0, 0.0, -0.0, -0.0, 0.0, 0.04865, -0.0, -0.0, 0.0, -0.0, -0.0, -0.0, -0.0, -0.0, 0.12763, -0.0, 0.0, -0.12763, -0.0, -0.0, -0.12763, -0.0, 0.0, 0.12763], [0.0, 0.0, 0.0, -0.0, 0.0, -0.07965, 0.0, 0.0, -0.0818, 0.0, -0.0, 0.0, 0.0, 0.0, -0.02546, 0.0, 0.0, -0.10537, 0.0, -0.0, 0.0818, 0.0, 0.0, 0.07965, 0.0, 0.0, 0.10537, -0.0, 0.0, 0.02995, -0.0, 0.0, -0.02995, 0.0, -0.0, 0.02546, 0.0, 0.0, -0.10537, 0.0, 0.0, 0.0818, 0.0, 0.0, 0.07965, 0.0, 0.0, 0.10537, -0.0, 0.0, 0.02995, 0.0, 0.0, -0.02995, 0.0, 0.0, 0.02546, 0.0, 0.0, 0.0, 0.0, 0.0, -0.07965, 0.0, -0.0, -0.0818, 0.0, -0.0, -0.0, 0.0, -0.0, -0.02546, 0.0, 0.0, -0.0, 0.0, 0.0, 0.0, 0.0, -0.0, 0.0352, 0.0, 0.0, -0.0352, 0.0, -0.0, -0.0352, 0.0, 0.0, 0.0352], [-0.0, 0.0, 0.0, -0.0, 0.0, 0.0158, -0.0, 0.0, -0.04885, -0.0, 0.0, -0.0, -0.0, -0.0, 0.14558, -0.0, -0.0, -0.01445, -0.0, -0.0, 0.04885, -0.0, -0.0, -0.0158, -0.0, -0.0, 0.01445, -0.0, -0.0, 0.05366, -0.0, -0.0, -0.05366, -0.0, 0.0, -0.14558, -0.0, -0.0, -0.01445, -0.0, 0.0, 0.04885, -0.0, -0.0, -0.0158, -0.0, -0.0, 0.01445, -0.0, -0.0, 0.05366, 0.0, -0.0, -0.05366, -0.0, -0.0, -0.14558, -0.0, 0.0, 0.0, -0.0, -0.0, 0.0158, -0.0, -0.0, -0.04885, -0.0, 0.0, 0.0, -0.0, 0.0, 0.14558, -0.0, -0.0, 0.0, -0.0, -0.0, -0.0, -0.0, 0.0, -0.06139, -0.0, -0.0, 0.06139, -0.0, 0.0, 0.06139, -0.0, -0.0, -0.06139], [-0.0, -0.0, -0.0, 0.0, 0.0, -0.10224, 0.0, -0.0, 0.04783, 0.0, -0.0, 0.0, 0.0, 0.0, 0.06117, 0.0, 0.0, -0.01201, 0.0, 0.0, -0.04783, 0.0, 0.0, 0.10224, 0.0, 0.0, 0.01201, 0.0, 0.0, -0.11329, 0.0, 0.0, 0.11329, 0.0, -0.0, -0.06117, 0.0, 0.0, -0.01201, 0.0, -0.0, -0.04783, 0.0, 0.0, 0.10224, 0.0, 0.0, 0.01201, -0.0, 0.0, -0.11329, -0.0, 0.0, 0.11329, 0.0, 0.0, -0.06117, 0.0, -0.0, -0.0, 0.0, 0.0, -0.10224, 0.0, 0.0, 0.04783, 0.0, -0.0, -0.0, 0.0, -0.0, 0.06117, 0.0, 0.0, 0.0, 0.0, 0.0, -0.0, 0.0, -0.0, -0.01552, 0.0, 0.0, 0.01552, 0.0, -0.0, 0.01552, 0.0, 0.0, -0.01552], [-0.0, -0.0, -0.0, 0.0, -0.0, -0.09453, -0.0, -0.0, 0.06389, -0.0, -0.0, -0.0, -0.0, 0.0, 0.00024, -0.0, -0.0, 0.05371, -0.0, 0.0, -0.06389, -0.0, -0.0, 0.09453, -0.0, -0.0, -0.05371, 0.0, -0.0, 0.10598, 0.0, -0.0, -0.10598, -0.0, -0.0, -0.00024, -0.0, -0.0, 0.05371, -0.0, -0.0, -0.06389, -0.0, -0.0, 0.09453, -0.0, -0.0, -0.05371, -0.0, -0.0, 0.10598, -0.0, -0.0, -0.10598, -0.0, 0.0, -0.00024, -0.0, -0.0, 0.0, -0.0, -0.0, -0.09453, -0.0, 0.0, 0.06389, -0.0, -0.0, 0.0, -0.0, -0.0, 0.00024, -0.0, 0.0, -0.0, -0.0, 0.0, 0.0, -0.0, -0.0, 0.00536, -0.0, 0.0, -0.00536, -0.0, -0.0, -0.00536, -0.0, 0.0, 0.00536]]), "B1u":array([[-0.0, -0.0, 0.00303, 0.0, 0.0, -0.01829, -0.0, -0.0, 0.00816, -0.0, -0.0, 0.03865, -0.0, -0.0, -0.07375, -0.0, -0.0, 0.04837, -0.0, -0.0, 0.00816, -0.0, -0.0, -0.01829, -0.0, -0.0, 0.04837, 0.0, -0.0, 0.05352, -0.0, -0.0, 0.05352, -0.0, 0.0, -0.07375, -0.0, -0.0, 0.04837, -0.0, -0.0, 0.00816, -0.0, -0.0, -0.01829, -0.0, -0.0, 0.04837, -0.0, -0.0, 0.05352, 0.0, -0.0, 0.05352, -0.0, -0.0, -0.07375, -0.0, -0.0, 0.00303, -0.0, -0.0, -0.01829, -0.0, -0.0, 0.00816, -0.0, -0.0, 0.03865, -0.0, 0.0, -0.07375, -0.0, 0.0, -0.0968, -0.0, 0.0, -0.0968, -0.0, 0.0, -0.10114, -0.0, 0.0, -0.10114, -0.0, 0.0, -0.10114, -0.0, -0.0, -0.10114], [-0.0, -0.0, 0.04022, -0.0, 0.0, 0.03828, -0.0, -0.0, 0.05671, -0.0, -0.0, 0.05678, -0.0, -0.0, 0.02995, -0.0, -0.0, 0.08202, -0.0, -0.0, 0.05671, -0.0, -0.0, 0.03828, -0.0, -0.0, 0.08202, 0.0, -0.0, 0.07202, -0.0, -0.0, 0.07202, -0.0, -0.0, 0.02995, -0.0, -0.0, 0.08202, -0.0, -0.0, 0.05671, -0.0, -0.0, 0.03828, -0.0, -0.0, 0.08202, -0.0, -0.0, 0.07202, 0.0, -0.0, 0.07202, -0.0, -0.0, 0.02995, -0.0, -0.0, 0.04022, -0.0, -0.0, 0.03828, -0.0, -0.0, 0.05671, -0.0, -0.0, 0.05678, -0.0, -0.0, 0.02995, -0.0, 0.0, -0.02531, -0.0, 0.0, -0.02531, -0.0, -0.0, 0.07672, -0.0, 0.0, 0.07672, -0.0, -0.0, 0.07672, -0.0, -0.0, 0.07672], [-0.0, 0.0, -0.03678, 0.0, -0.0, -0.00679, -0.0, -0.0, 0.02519, -0.0, 0.0, -0.05829, -0.0, -0.0, -0.02838, -0.0, -0.0, -0.01834, -0.0, -0.0, 0.02519, -0.0, -0.0, -0.00679, -0.0, -0.0, -0.01834, 0.0, -0.0, -0.05783, 0.0, -0.0, -0.05783, -0.0, -0.0, -0.02838, -0.0, -0.0, -0.01834, -0.0, -0.0, 0.02519, -0.0, -0.0, -0.00679, -0.0, -0.0, -0.01834, -0.0, -0.0, -0.05783, -0.0, -0.0, -0.05783, -0.0, -0.0, -0.02838, -0.0, 0.0, -0.03678, -0.0, -0.0, -0.00679, -0.0, -0.0, 0.02519, -0.0, 0.0, -0.05829, -0.0, -0.0, -0.02838, -0.0, 0.0, -0.19899, -0.0, 0.0, -0.19899, -0.0, 0.0, 0.06813, -0.0, 0.0, 0.06813, -0.0, -0.0, 0.06813, -0.0, -0.0, 0.06813], [-0.0, 0.0, -0.02056, -0.0, -0.0, 0.03977, -0.0, -0.0, 0.11186, -0.0, 0.0, -0.07158, -0.0, -0.0, 0.01405, -0.0, -0.0, 0.0442, -0.0, -0.0, 0.11186, -0.0, -0.0, 0.03977, -0.0, -0.0, 0.0442, -0.0, -0.0, -0.06609, 0.0, -0.0, -0.06609, -0.0, 0.0, 0.01405, -0.0, -0.0, 0.0442, -0.0, -0.0, 0.11186, -0.0, -0.0, 0.03977, -0.0, -0.0, 0.0442, 0.0, -0.0, -0.06609, -0.0, -0.0, -0.06609, -0.0, -0.0, 0.01405, -0.0, 0.0, -0.02056, -0.0, -0.0, 0.03977, -0.0, -0.0, 0.11186, -0.0, 0.0, -0.07158, -0.0, 0.0, 0.01405, -0.0, -0.0, 0.03903, -0.0, -0.0, 0.03903, -0.0, 0.0, -0.05491, -0.0, 0.0, -0.05491, -0.0, 0.0, -0.05491, -0.0, -0.0, -0.05491], [0.0, -0.0, 0.10216, -0.0, -0.0, 0.09496, -0.0, -0.0, 0.01849, -0.0, -0.0, 0.06238, -0.0, -0.0, 0.02074, -0.0, -0.0, -0.08613, -0.0, -0.0, 0.01849, -0.0, -0.0, 0.09496, -0.0, -0.0, -0.08613, 0.0, -0.0, -0.00118, -0.0, -0.0, -0.00118, -0.0, -0.0, 0.02074, -0.0, -0.0, -0.08613, -0.0, -0.0, 0.01849, -0.0, -0.0, 0.09496, -0.0, -0.0, -0.08613, -0.0, -0.0, -0.00118, 0.0, -0.0, -0.00118, -0.0, -0.0, 0.02074, -0.0, -0.0, 0.10216, -0.0, -0.0, 0.09496, -0.0, -0.0, 0.01849, -0.0, -0.0, 0.06238, -0.0, -0.0, 0.02074, -0.0, 0.0, -0.05184, -0.0, 0.0, -0.05184, -0.0, 0.0, -0.03437, -0.0, 0.0, -0.03437, -0.0, -0.0, -0.03437, -0.0, -0.0, -0.03437], [-0.0, -0.0, -0.02568, 0.0, 0.0, 0.01639, 0.0, -0.0, -0.07032, -0.0, -0.0, 0.01773, 0.0, 0.0, 0.11262, 0.0, 0.0, 0.0606, 0.0, 0.0, -0.07032, 0.0, 0.0, 0.01639, 0.0, 0.0, 0.0606, 0.0, 0.0, -0.03551, -0.0, 0.0, -0.03551, 0.0, 0.0, 0.11262, 0.0, 0.0, 0.0606, 0.0, 0.0, -0.07032, 0.0, 0.0, 0.01639, 0.0, 0.0, 0.0606, -0.0, 0.0, -0.03551, 0.0, 0.0, -0.03551, 0.0, 0.0, 0.11262, 0.0, 0.0, -0.02568, 0.0, 0.0, 0.01639, 0.0, 0.0, -0.07032, 0.0, -0.0, 0.01773, 0.0, 0.0, 0.11262, 0.0, 0.0, -0.07165, 0.0, 0.0, -0.07165, 0.0, -0.0, -0.04327, 0.0, 0.0, -0.04327, 0.0, -0.0, -0.04327, 0.0, -0.0, -0.04327], [-0.0, -0.0, 0.10468, -0.0, -0.0, 0.0469, -0.0, 0.0, -0.05989, 0.0, 0.0, -0.01011, -0.0, -0.0, -0.07048, -0.0, -0.0, 0.06477, -0.0, -0.0, -0.05989, -0.0, -0.0, 0.0469, -0.0, -0.0, 0.06477, -0.0, -0.0, -0.06565, -0.0, -0.0, -0.06565, -0.0, 0.0, -0.07048, -0.0, -0.0, 0.06477, -0.0, -0.0, -0.05989, -0.0, -0.0, 0.0469, -0.0, -0.0, 0.06477, 0.0, -0.0, -0.06565, 0.0, -0.0, -0.06565, -0.0, -0.0, -0.07048, -0.0, -0.0, 0.10468, -0.0, -0.0, 0.0469, -0.0, -0.0, -0.05989, -0.0, 0.0, -0.01011, -0.0, 0.0, -0.07048, -0.0, -0.0, 0.02623, -0.0, -0.0, 0.02623, -0.0, -0.0, 0.0214, -0.0, -0.0, 0.0214, -0.0, 0.0, 0.0214, -0.0, 0.0, 0.0214], [0.0, 0.0, -0.16557, 0.0, 0.0, 0.06251, -0.0, -0.0, -0.00441, 0.0, -0.0, 0.15109, -0.0, -0.0, -0.04227, -0.0, -0.0, 0.00171, -0.0, -0.0, -0.00441, -0.0, -0.0, 0.06251, -0.0, -0.0, 0.00171, 0.0, -0.0, -0.03831, -0.0, -0.0, -0.03831, -0.0, -0.0, -0.04227, -0.0, -0.0, 0.00171, -0.0, -0.0, -0.00441, -0.0, -0.0, 0.06251, -0.0, -0.0, 0.00171, -0.0, -0.0, -0.03831, 0.0, -0.0, -0.03831, -0.0, -0.0, -0.04227, -0.0, 0.0, -0.16557, -0.0, -0.0, 0.06251, -0.0, -0.0, -0.00441, -0.0, -0.0, 0.15109, -0.0, -0.0, -0.04227, -0.0, -0.0, 0.02473, -0.0, -0.0, 0.02473, -0.0, -0.0, 0.01426, -0.0, 0.0, 0.01426, -0.0, -0.0, 0.01426, -0.0, -0.0, 0.01426]]), "B2u":array([[0.00452, 0.0, 0.0, 0.00758, -0.0366, -0.0, 0.0075, -0.06683, -0.0, 0.00727, -0.0, 0.0, 0.00961, -0.04625, -0.0, 0.00657, 0.05483, 0.0, 0.0075, 0.06683, 0.0, 0.00758, 0.0366, 0.0, 0.00657, -0.05483, -0.0, 0.00681, -0.01454, -0.0, 0.00681, 0.01454, 0.0, 0.00961, 0.04625, 0.0, 0.00657, 0.05483, 0.0, 0.0075, 0.06683, 0.0, 0.00758, 0.0366, 0.0, 0.00657, -0.05483, -0.0, 0.00681, -0.01454, -0.0, 0.00681, 0.01454, 0.0, 0.00961, 0.04625, 0.0, 0.00452, -0.0, -0.0, 0.00758, -0.0366, -0.0, 0.0075, -0.06683, -0.0, 0.00727, -0.0, -0.0, 0.00961, -0.04625, -0.0, 0.01312, -0.0, -0.0, 0.01312, -0.0, 0.0, 0.00826, -0.09331, -0.0, 0.00826, 0.09331, 0.0, 0.00826, 0.09331, 0.0, 0.00826, -0.09331, -0.0], [0.00829, 0.0, 0.0, -0.01106, -0.00643, 0.0, 0.01535, 0.01057, 0.0, -0.00203, 0.0, 0.0, -0.06545, -0.0072, 0.0, 0.00076, -0.01276, -0.0, 0.01535, -0.01057, -0.0, -0.01106, 0.00643, -0.0, 0.00076, 0.01276, 0.0, -0.0015, 0.00405, 0.0, -0.0015, -0.00405, -0.0, -0.06545, 0.0072, -0.0, 0.00076, -0.01276, -0.0, 0.01535, -0.01057, 0.0, -0.01106, 0.00643, 0.0, 0.00076, 0.01276, 0.0, -0.0015, 0.00405, 0.0, -0.0015, -0.00405, -0.0, -0.06545, 0.0072, 0.0, 0.00829, -0.0, -0.0, -0.01106, -0.00643, -0.0, 0.01535, 0.01057, -0.0, -0.00203, 0.0, -0.0, -0.06545, -0.0072, -0.0, -0.08489, 0.0, 0.0, -0.08489, 0.0, -0.0, -0.10296, -0.03103, 0.0, -0.10296, 0.03103, -0.0, -0.10296, 0.03103, 0.0, -0.10296, -0.03103, -0.0], [0.01071, 0.0, -0.0, 0.00313, 0.02971, -0.0, 0.00427, 0.06246, 0.0, -0.00492, 0.0, 0.0, 0.01885, -0.03181, -0.0, -0.01306, -0.06473, -0.0, 0.00427, -0.06246, -0.0, 0.00313, -0.02971, -0.0, -0.01306, 0.06473, 0.0, -0.0057, 0.02218, 0.0, -0.0057, -0.02218, -0.0, 0.01885, 0.03181, 0.0, -0.01306, -0.06473, -0.0, 0.00427, -0.06246, -0.0, 0.00313, -0.02971, -0.0, -0.01306, 0.06473, 0.0, -0.0057, 0.02218, 0.0, -0.0057, -0.02218, -0.0, 0.01885, 0.03181, 0.0, 0.01071, -0.0, 0.0, 0.00313, 0.02971, 0.0, 0.00427, 0.06246, 0.0, -0.00492, 0.0, -0.0, 0.01885, -0.03181, -0.0, 0.01812, -0.0, -0.0, 0.01812, -0.0, 0.0, 0.02042, -0.07777, -0.0, 0.02042, 0.07777, 0.0, 0.02042, 0.07777, 0.0, 0.02042, -0.07777, -0.0], [-0.02648, 0.0, -0.0, -0.00981, 0.01014, -0.0, -0.02731, -0.005, -0.0, 0.02863, -0.0, 0.0, 0.04495, 0.02452, -0.0, -0.05421, 0.00039, -0.0, -0.02731, 0.005, -0.0, -0.00981, -0.01014, -0.0, -0.05421, -0.00039, 0.0, 0.02484, -0.01509, -0.0, 0.02484, 0.01509, 0.0, 0.04495, -0.02452, 0.0, -0.05421, 0.00039, 0.0, -0.02731, 0.005, 0.0, -0.00981, -0.01014, -0.0, -0.05421, -0.00039, -0.0, 0.02484, -0.01509, -0.0, 0.02484, 0.01509, 0.0, 0.04495, -0.02452, -0.0, -0.02648, 0.0, -0.0, -0.00981, 0.01014, 0.0, -0.02731, -0.005, -0.0, 0.02863, -0.0, -0.0, 0.04495, 0.02452, 0.0, 0.10287, -0.0, -0.0, 0.10287, -0.0, 0.0, -0.07678, -0.00942, -0.0, -0.07678, 0.00942, 0.0, -0.07678, 0.00942, -0.0, -0.07678, -0.00942, 0.0], [0.06041, -0.0, -0.0, 0.02333, 0.01054, 0.0, 0.08846, 0.02528, 0.0, 0.02347, 0.0, -0.0, 0.02656, 0.03134, 0.0, 0.03637, 0.02328, 0.0, 0.08846, -0.02528, -0.0, 0.02333, -0.01054, -0.0, 0.03637, -0.02328, -0.0, 0.03788, -0.00444, -0.0, 0.03788, 0.00444, 0.0, 0.02656, -0.03134, -0.0, 0.03637, 0.02328, 0.0, 0.08846, -0.02528, -0.0, 0.02333, -0.01054, 0.0, 0.03637, -0.02328, -0.0, 0.03788, -0.00444, 0.0, 0.03788, 0.00444, -0.0, 0.02656, -0.03134, -0.0, 0.06041, -0.0, -0.0, 0.02333, 0.01054, -0.0, 0.08846, 0.02528, 0.0, 0.02347, -0.0, -0.0, 0.02656, 0.03134, 0.0, 0.01913, -0.0, 0.0, 0.01913, -0.0, -0.0, -0.01431, -0.00389, -0.0, -0.01431, 0.00389, 0.0, -0.01431, 0.00389, -0.0, -0.01431, -0.00389, 0.0], [0.05043, 0.0, 0.0, -0.04518, 0.03015, -0.0, 0.00822, 0.03432, -0.0, -0.04181, -0.0, -0.0, -0.00716, -0.01732, -0.0, -0.02709, 0.06492, 0.0, 0.00822, -0.03432, -0.0, -0.04518, -0.03015, -0.0, -0.02709, -0.06492, -0.0, -0.05165, -0.00534, -0.0, -0.05165, 0.00534, 0.0, -0.00716, 0.01732, 0.0, -0.02709, 0.06492, 0.0, 0.00822, -0.03432, -0.0, -0.04518, -0.03015, -0.0, -0.02709, -0.06492, -0.0, -0.05165, -0.00534, 0.0, -0.05165, 0.00534, -0.0, -0.00716, 0.01732, 0.0, 0.05043, -0.0, -0.0, -0.04518, 0.03015, 0.0, 0.00822, 0.03432, 0.0, -0.04181, 0.0, -0.0, -0.00716, -0.01732, -0.0, 0.02361, -0.0, 0.0, 0.02361, -0.0, -0.0, -0.00235, 0.00278, -0.0, -0.00235, -0.00278, 0.0, -0.00235, -0.00278, -0.0, -0.00235, 0.00278, 0.0], [0.03763, 0.0, 0.0, 0.03407, -0.06691, -0.0, -0.02844, 0.04097, -0.0, -0.01048, -0.0, 0.0, -0.04488, 0.0081, 0.0, 0.02711, -0.00086, -0.0, -0.02844, -0.04097, -0.0, 0.03407, 0.06691, 0.0, 0.02711, 0.00086, 0.0, -0.01489, -0.01542, 0.0, -0.01489, 0.01542, -0.0, -0.04488, -0.0081, -0.0, 0.02711, -0.00086, 0.0, -0.02844, -0.04097, -0.0, 0.03407, 0.06691, 0.0, 0.02711, 0.00086, -0.0, -0.01489, -0.01542, -0.0, -0.01489, 0.01542, 0.0, -0.04488, -0.0081, 0.0, 0.03763, -0.0, -0.0, 0.03407, -0.06691, -0.0, -0.02844, 0.04097, 0.0, -0.01048, 0.0, -0.0, -0.04488, 0.0081, -0.0, 0.08246, -0.0, 0.0, 0.08246, -0.0, -0.0, -0.00325, -0.00252, -0.0, -0.00325, 0.00252, 0.0, -0.00325, 0.00252, -0.0, -0.00325, -0.00252, 0.0], [-0.00843, 0.0, 0.0, -0.06814, 0.00839, 0.0, -0.01716, -0.0012, -0.0, 0.00145, 0.0, 0.0, -0.00566, 0.06258, 0.0, 0.04444, -0.01107, 0.0, -0.01716, 0.0012, -0.0, -0.06814, -0.00839, -0.0, 0.04444, 0.01107, -0.0, 0.00031, -0.04449, -0.0, 0.00031, 0.04449, 0.0, -0.00566, -0.06258, -0.0, 0.04444, -0.01107, -0.0, -0.01716, 0.0012, 0.0, -0.06814, -0.00839, 0.0, 0.04444, 0.01107, 0.0, 0.00031, -0.04449, -0.0, 0.00031, 0.04449, 0.0, -0.00566, -0.06258, 0.0, -0.00843, 0.0, -0.0, -0.06814, 0.00839, -0.0, -0.01716, -0.0012, -0.0, 0.00145, -0.0, -0.0, -0.00566, 0.06258, -0.0, -0.0022, 0.0, 0.0, -0.0022, 0.0, -0.0, 0.01479, -0.03637, 0.0, 0.01479, 0.03637, -0.0, 0.01479, 0.03637, 0.0, 0.01479, -0.03637, -0.0], [0.01631, -0.0, -0.0, 0.03362, 0.06071, 0.0, -0.02409, -0.01296, -0.0, -0.01708, 0.0, -0.0, 0.00502, -0.04333, 0.0, 0.04934, -0.01066, -0.0, -0.02409, 0.01296, -0.0, 0.03362, -0.06071, -0.0, 0.04934, 0.01066, 0.0, -0.00607, -0.06017, -0.0, -0.00607, 0.06017, 0.0, 0.00502, 0.04333, -0.0, 0.04934, -0.01066, 0.0, -0.02409, 0.01296, 0.0, 0.03362, -0.06071, -0.0, 0.04934, 0.01066, -0.0, -0.00607, -0.06017, -0.0, -0.00607, 0.06017, 0.0, 0.00502, 0.04333, 0.0, 0.01631, -0.0, -0.0, 0.03362, 0.06071, 0.0, -0.02409, -0.01296, -0.0, -0.01708, 0.0, 0.0, 0.00502, -0.04333, -0.0, 0.0131, -0.0, 0.0, 0.0131, -0.0, -0.0, -0.02162, 0.01314, 0.0, -0.02162, -0.01314, -0.0, -0.02162, -0.01314, 0.0, -0.02162, 0.01314, -0.0], [0.04493, -0.0, -0.0, -0.01996, 0.02797, 0.0, -0.03668, 0.01418, -0.0, 0.06023, -0.0, -0.0, -0.04808, -0.02206, 0.0, -0.01112, 0.02323, 0.0, -0.03668, -0.01418, -0.0, -0.01996, -0.02797, -0.0, -0.01112, -0.02323, -0.0, 0.08337, 0.00965, 0.0, 0.08337, -0.00965, -0.0, -0.04808, 0.02206, -0.0, -0.01112, 0.02323, 0.0, -0.03668, -0.01418, 0.0, -0.01996, -0.02797, 0.0, -0.01112, -0.02323, -0.0, 0.08337, 0.00965, 0.0, 0.08337, -0.00965, -0.0, -0.04808, 0.02206, 0.0, 0.04493, -0.0, -0.0, -0.01996, 0.02797, -0.0, -0.03668, 0.01418, -0.0, 0.06023, -0.0, 0.0, -0.04808, -0.02206, -0.0, 0.00516, -0.0, 0.0, 0.00516, -0.0, -0.0, 0.02271, 0.00354, 0.0, 0.02271, -0.00354, -0.0, 0.02271, -0.00354, 0.0, 0.02271, 0.00354, -0.0], [-0.00354, 0.0, -0.0, -0.05449, -0.0082, 0.0, 0.04766, -0.03484, -0.0, -0.0149, 0.0, 0.0, -0.02725, -0.04147, 0.0, 0.01332, -0.04283, 0.0, 0.04766, 0.03484, 0.0, -0.05449, 0.0082, 0.0, 0.01332, 0.04283, -0.0, 0.00598, 0.01122, 0.0, 0.00598, -0.01122, -0.0, -0.02725, 0.04147, 0.0, 0.01332, -0.04283, -0.0, 0.04766, 0.03484, -0.0, -0.05449, 0.0082, -0.0, 0.01332, 0.04283, 0.0, 0.00598, 0.01122, 0.0, 0.00598, -0.01122, -0.0, -0.02725, 0.04147, -0.0, -0.00354, 0.0, 0.0, -0.05449, -0.0082, 0.0, 0.04766, -0.03484, 0.0, -0.0149, 0.0, 0.0, -0.02725, -0.04147, 0.0, 0.07542, -0.0, -0.0, 0.07542, -0.0, 0.0, -0.00251, 0.0242, -0.0, -0.00251, -0.0242, 0.0, -0.00251, -0.0242, -0.0, -0.00251, 0.0242, 0.0], [0.03908, -0.0, -0.0, 0.02356, 0.01231, 0.0, 0.027, -0.0266, 0.0, 0.03724, -0.0, 0.0, -0.0371, 0.02287, 0.0, -0.06068, -0.02635, 0.0, 0.027, 0.0266, 0.0, 0.02356, -0.01231, 0.0, -0.06068, 0.02635, -0.0, -0.02241, -0.05146, -0.0, -0.02241, 0.05146, 0.0, -0.0371, -0.02287, 0.0, -0.06068, -0.02635, -0.0, 0.027, 0.0266, 0.0, 0.02356, -0.01231, -0.0, -0.06068, 0.02635, 0.0, -0.02241, -0.05146, -0.0, -0.02241, 0.05146, 0.0, -0.0371, -0.02287, -0.0, 0.03908, -0.0, 0.0, 0.02356, 0.01231, 0.0, 0.027, -0.0266, -0.0, 0.03724, -0.0, 0.0, -0.0371, 0.02287, 0.0, 0.00112, -0.0, -0.0, 0.00112, -0.0, 0.0, 0.02355, -0.00566, -0.0, 0.02355, 0.00566, 0.0, 0.02355, 0.00566, -0.0, 0.02355, -0.00566, 0.0], [0.14471, -0.0, -0.0, -0.01935, -0.02908, -0.0, -0.03074, -0.02847, -0.0, -0.03713, 0.0, -0.0, 0.04607, 0.0007, 0.0, -0.00776, -0.03399, 0.0, -0.03074, 0.02847, 0.0, -0.01935, 0.02908, 0.0, -0.00776, 0.03399, -0.0, 0.00885, 0.00758, -0.0, 0.00885, -0.00758, 0.0, 0.04607, -0.0007, 0.0, -0.00776, -0.03399, -0.0, -0.03074, 0.02847, -0.0, -0.01935, 0.02908, -0.0, -0.00776, 0.03399, 0.0, 0.00885, 0.00758, 0.0, 0.00885, -0.00758, -0.0, 0.04607, -0.0007, -0.0, 0.14471, -0.0, 0.0, -0.01935, -0.02908, 0.0, -0.03074, -0.02847, 0.0, -0.03713, 0.0, 0.0, 0.04607, 0.0007, 0.0, -0.03473, 0.0, -0.0, -0.03473, 0.0, 0.0, -0.01105, 0.00973, -0.0, -0.01105, -0.00973, 0.0, -0.01105, -0.00973, -0.0, -0.01105, 0.00973, 0.0], [0.0087, -0.0, -0.0, -0.02732, -0.02784, -0.0, -0.00156, 0.02099, 0.0, 0.15605, -0.0, -0.0, 0.03089, -0.03673, -0.0, 0.01791, -0.00329, 0.0, -0.00156, -0.02099, -0.0, -0.02732, 0.02784, -0.0, 0.01791, 0.00329, -0.0, -0.04504, -0.01513, -0.0, -0.04504, 0.01513, 0.0, 0.03089, 0.03673, 0.0, 0.01791, -0.00329, -0.0, -0.00156, -0.02099, -0.0, -0.02732, 0.02784, 0.0, 0.01791, 0.00329, 0.0, -0.04504, -0.01513, -0.0, -0.04504, 0.01513, 0.0, 0.03089, 0.03673, 0.0, 0.0087, -0.0, -0.0, -0.02732, -0.02784, -0.0, -0.00156, 0.02099, 0.0, 0.15605, -0.0, -0.0, 0.03089, -0.03673, -0.0, -0.02113, 0.0, -0.0, -0.02113, 0.0, 0.0, -0.00761, 0.01746, 0.0, -0.00761, -0.01746, -0.0, -0.00761, -0.01746, 0.0, -0.00761, 0.01746, -0.0]]), "B3u":array([[-0.0, 0.00638, -0.0, -0.02288, -0.00937, -0.0, -0.03756, -0.00029, -0.0, -0.0, 0.03583, -0.0, -0.02199, -0.03478, -0.0, 0.0384, 0.02592, 0.0, 0.03756, -0.00029, 0.0, 0.02288, -0.00937, 0.0, -0.0384, 0.02592, -0.0, -0.00532, 0.04325, -0.0, 0.00532, 0.04325, 0.0, 0.02199, -0.03478, 0.0, 0.0384, 0.02592, 0.0, 0.03756, -0.00029, 0.0, 0.02288, -0.00937, 0.0, -0.0384, 0.02592, -0.0, -0.00532, 0.04325, -0.0, 0.00532, 0.04325, 0.0, 0.02199, -0.03478, 0.0, -0.0, 0.00638, 0.0, -0.02288, -0.00937, -0.0, -0.03756, -0.00029, -0.0, -0.0, 0.03583, 0.0, -0.02199, -0.03478, -0.0, -0.0, -0.04612, 0.0, -0.0, -0.04612, -0.0, -0.03905, -0.04587, -0.0, 0.03905, -0.04587, 0.0, 0.03905, -0.04587, 0.0, -0.03905, -0.04587, -0.0], [0.0, -0.0171, 0.0, 0.02191, -0.00627, 0.0, 0.02797, -0.00256, 0.0, 0.0, -0.01806, 0.0, 0.02056, -0.04766, 0.0, -0.03898, 0.00344, -0.0, -0.02797, -0.00256, -0.0, -0.02191, -0.00627, -0.0, 0.03898, 0.00344, 0.0, -0.01034, -0.02004, -0.0, 0.01034, -0.02004, 0.0, -0.02056, -0.04766, -0.0, -0.03898, 0.00344, -0.0, -0.02797, -0.00256, -0.0, -0.02191, -0.00627, -0.0, 0.03898, 0.00344, 0.0, -0.01034, -0.02004, -0.0, 0.01034, -0.02004, 0.0, -0.02056, -0.04766, -0.0, 0.0, -0.0171, 0.0, 0.02191, -0.00627, 0.0, 0.02797, -0.00256, 0.0, 0.0, -0.01806, -0.0, 0.02056, -0.04766, 0.0, 0.0, -0.07569, 0.0, 0.0, -0.07569, -0.0, 0.03905, -0.05807, 0.0, -0.03905, -0.05807, -0.0, -0.03905, -0.05807, -0.0, 0.03905, -0.05807, 0.0], [-0.0, 0.00426, -0.0, 0.01945, 0.023, -0.0, 0.0309, 0.03243, -0.0, 0.0, -0.02371, -0.0, -0.03858, 0.00347, -0.0, -0.0486, 0.02581, -0.0, -0.0309, 0.03243, -0.0, -0.01945, 0.023, -0.0, 0.0486, 0.02581, 0.0, -0.00268, -0.01689, -0.0, 0.00268, -0.01689, 0.0, 0.03858, 0.00347, 0.0, -0.0486, 0.02581, -0.0, -0.0309, 0.03243, -0.0, -0.01945, 0.023, -0.0, 0.0486, 0.02581, 0.0, -0.00268, -0.01689, -0.0, 0.00268, -0.01689, 0.0, 0.03858, 0.00347, 0.0, -0.0, 0.00426, -0.0, 0.01945, 0.023, 0.0, 0.0309, 0.03243, 0.0, -0.0, -0.02371, -0.0, -0.03858, 0.00347, -0.0, -0.0, -0.02782, 0.0, -0.0, -0.02782, -0.0, -0.0815, 0.00865, -0.0, 0.0815, 0.00865, 0.0, 0.0815, 0.00865, 0.0, -0.0815, 0.00865, -0.0], [-0.0, -0.00181, -0.0, 0.00244, -0.01461, -0.0, -0.0107, -0.05622, -0.0, -0.0, -0.02305, -0.0, -0.00504, 0.00667, -0.0, -0.01068, -0.08022, -0.0, 0.0107, -0.05622, 0.0, -0.00244, -0.01461, -0.0, 0.01068, -0.08022, 0.0, 0.01207, -0.02399, 0.0, -0.01207, -0.02399, -0.0, 0.00504, 0.00667, -0.0, -0.01068, -0.08022, -0.0, 0.0107, -0.05622, 0.0, -0.00244, -0.01461, 0.0, 0.01068, -0.08022, 0.0, 0.01207, -0.02399, 0.0, -0.01207, -0.02399, -0.0, 0.00504, 0.00667, 0.0, -0.0, -0.00181, 0.0, 0.00244, -0.01461, 0.0, -0.0107, -0.05622, -0.0, -0.0, -0.02305, -0.0, -0.00504, 0.00667, 0.0, -0.0, 0.04035, -0.0, -0.0, 0.04035, 0.0, -0.05127, -0.05581, -0.0, 0.05127, -0.05581, 0.0, 0.05127, -0.05581, 0.0, -0.05127, -0.05581, -0.0], [-0.0, -0.01016, -0.0, 0.02991, -0.01059, 0.0, 0.01371, -0.06587, -0.0, -0.0, 0.04199, 0.0, -0.00711, -0.01518, -0.0, -0.0224, -0.02166, -0.0, -0.01371, -0.06587, -0.0, -0.02991, -0.01059, -0.0, 0.0224, -0.02166, -0.0, -0.02415, 0.04753, -0.0, 0.02415, 0.04753, 0.0, 0.00711, -0.01518, -0.0, -0.0224, -0.02166, 0.0, -0.01371, -0.06587, 0.0, -0.02991, -0.01059, 0.0, 0.0224, -0.02166, 0.0, -0.02415, 0.04753, -0.0, 0.02415, 0.04753, 0.0, 0.00711, -0.01518, 0.0, 0.0, -0.01016, -0.0, 0.02991, -0.01059, 0.0, 0.01371, -0.06587, 0.0, 0.0, 0.04199, 0.0, -0.00711, -0.01518, 0.0, 0.0, -0.0518, 0.0, 0.0, -0.0518, -0.0, -0.00822, 0.05672, -0.0, 0.00822, 0.05672, -0.0, 0.00822, 0.05672, 0.0, -0.00822, 0.05672, 0.0], [0.0, 0.04451, -0.0, -0.00612, 0.02422, -0.0, 0.0289, 0.03086, -0.0, -0.0, 0.02426, 0.0, -0.02845, -0.01858, -0.0, 0.05392, -0.06318, 0.0, -0.0289, 0.03086, -0.0, 0.00612, 0.02422, 0.0, -0.05392, -0.06318, -0.0, -0.01335, -0.03365, -0.0, 0.01335, -0.03365, 0.0, 0.02845, -0.01858, 0.0, 0.05392, -0.06318, 0.0, -0.0289, 0.03086, -0.0, 0.00612, 0.02422, -0.0, -0.05392, -0.06318, -0.0, -0.01335, -0.03365, -0.0, 0.01335, -0.03365, 0.0, 0.02845, -0.01858, 0.0, 0.0, 0.04451, -0.0, -0.00612, 0.02422, -0.0, 0.0289, 0.03086, 0.0, 0.0, 0.02426, -0.0, -0.02845, -0.01858, -0.0, 0.0, -0.057, -0.0, 0.0, -0.057, 0.0, 0.00576, 0.01756, 0.0, -0.00576, 0.01756, 0.0, -0.00576, 0.01756, -0.0, 0.00576, 0.01756, -0.0], [0.0, 0.00795, -0.0, 0.02861, 0.08419, 0.0, -0.03382, 0.01068, -0.0, 0.0, 0.04259, -0.0, 0.02152, 0.01896, -0.0, -0.01821, -0.01477, -0.0, 0.03382, 0.01068, 0.0, -0.02861, 0.08419, -0.0, 0.01821, -0.01477, -0.0, -0.04547, 0.01821, -0.0, 0.04547, 0.01821, 0.0, -0.02152, 0.01896, -0.0, -0.01821, -0.01477, 0.0, 0.03382, 0.01068, 0.0, -0.02861, 0.08419, -0.0, 0.01821, -0.01477, 0.0, -0.04547, 0.01821, -0.0, 0.04547, 0.01821, 0.0, -0.02152, 0.01896, 0.0, -0.0, 0.00795, 0.0, 0.02861, 0.08419, 0.0, -0.03382, 0.01068, -0.0, -0.0, 0.04259, 0.0, 0.02152, 0.01896, 0.0, -0.0, 0.01674, -0.0, -0.0, 0.01674, 0.0, 0.00503, -0.02517, -0.0, -0.00503, -0.02517, -0.0, -0.00503, -0.02517, 0.0, 0.00503, -0.02517, 0.0], [0.0, -0.00375, 0.0, -0.00975, 0.02723, -0.0, -0.04274, -0.01928, -0.0, -0.0, -0.00699, -0.0, -0.04245, 0.04417, 0.0, -0.02027, -0.00208, 0.0, 0.04274, -0.01928, 0.0, 0.00975, 0.02723, 0.0, 0.02027, -0.00208, 0.0, 0.06069, -0.00847, 0.0, -0.06069, -0.00847, -0.0, 0.04245, 0.04417, 0.0, -0.02027, -0.00208, -0.0, 0.04274, -0.01928, -0.0, 0.00975, 0.02723, -0.0, 0.02027, -0.00208, -0.0, 0.06069, -0.00847, 0.0, -0.06069, -0.00847, -0.0, 0.04245, 0.04417, 0.0, 0.0, -0.00375, -0.0, -0.00975, 0.02723, -0.0, -0.04274, -0.01928, -0.0, 0.0, -0.00699, 0.0, -0.04245, 0.04417, -0.0, 0.0, -0.07782, 0.0, 0.0, -0.07782, -0.0, 0.02881, -0.0003, 0.0, -0.02881, -0.0003, 0.0, -0.02881, -0.0003, -0.0, 0.02881, -0.0003, -0.0], [-0.0, 0.06099, -0.0, -0.01764, 0.01832, 0.0, 0.06153, -0.00713, 0.0, -0.0, 0.02952, -0.0, -0.03427, -0.00193, -0.0, -0.01826, 0.00035, -0.0, -0.06153, -0.00713, -0.0, 0.01764, 0.01832, -0.0, 0.01826, 0.00035, 0.0, 0.03186, 0.04852, 0.0, -0.03186, 0.04852, -0.0, 0.03427, -0.00193, 0.0, -0.01826, 0.00035, -0.0, -0.06153, -0.00713, -0.0, 0.01764, 0.01832, 0.0, 0.01826, 0.00035, 0.0, 0.03186, 0.04852, 0.0, -0.03186, 0.04852, -0.0, 0.03427, -0.00193, 0.0, 0.0, 0.06099, -0.0, -0.01764, 0.01832, 0.0, 0.06153, -0.00713, 0.0, 0.0, 0.02952, 0.0, -0.03427, -0.00193, -0.0, 0.0, 0.04893, -0.0, 0.0, 0.04893, 0.0, 0.02335, -0.02914, 0.0, -0.02335, -0.02914, 0.0, -0.02335, -0.02914, -0.0, 0.02335, -0.02914, -0.0], [0.0, -0.0881, 0.0, 0.01591, 0.01088, 0.0, 0.04342, -0.03004, -0.0, -0.0, 0.08097, 0.0, -0.0137, 0.03311, -0.0, 0.0386, 0.03178, 0.0, -0.04342, -0.03004, -0.0, -0.01591, 0.01088, -0.0, -0.0386, 0.03178, -0.0, 0.00102, -0.03623, 0.0, -0.00102, -0.03623, 0.0, 0.0137, 0.03311, -0.0, 0.0386, 0.03178, 0.0, -0.04342, -0.03004, 0.0, -0.01591, 0.01088, 0.0, -0.0386, 0.03178, -0.0, 0.00102, -0.03623, -0.0, -0.00102, -0.03623, -0.0, 0.0137, 0.03311, 0.0, 0.0, -0.0881, 0.0, 0.01591, 0.01088, 0.0, 0.04342, -0.03004, 0.0, 0.0, 0.08097, 0.0, -0.0137, 0.03311, 0.0, 0.0, 0.00551, 0.0, 0.0, 0.00551, -0.0, -0.00359, -0.02229, -0.0, 0.00359, -0.02229, -0.0, 0.00359, -0.02229, 0.0, -0.00359, -0.02229, 0.0], [-0.0, -0.07786, 0.0, 0.00774, 0.00144, 0.0, -0.0273, 0.01821, -0.0, 0.0, 0.00198, 0.0, -0.07372, -0.0516, -0.0, -0.01775, -0.01545, -0.0, 0.0273, 0.01821, -0.0, -0.00774, 0.00144, -0.0, 0.01775, -0.01545, -0.0, -0.01062, -0.00317, -0.0, 0.01062, -0.00317, -0.0, 0.07372, -0.0516, 0.0, -0.01775, -0.01545, 0.0, 0.0273, 0.01821, 0.0, -0.00774, 0.00144, -0.0, 0.01775, -0.01545, 0.0, -0.01062, -0.00317, 0.0, 0.01062, -0.00317, 0.0, 0.07372, -0.0516, 0.0, 0.0, -0.07786, 0.0, 0.00774, 0.00144, 0.0, -0.0273, 0.01821, 0.0, 0.0, 0.00198, 0.0, -0.07372, -0.0516, -0.0, 0.0, 0.05962, -0.0, 0.0, 0.05962, 0.0, 0.02706, 0.00955, -0.0, -0.02706, 0.00955, -0.0, -0.02706, 0.00955, 0.0, 0.02706, 0.00955, 0.0], [-0.0, -0.04877, 0.0, 0.03626, -0.05131, 0.0, 0.00887, 0.06697, -0.0, 0.0, 0.01934, -0.0, 0.00089, 0.05191, 0.0, -0.00414, -0.04288, 0.0, -0.00887, 0.06697, 0.0, -0.03626, -0.05131, -0.0, 0.00414, -0.04288, 0.0, 0.00703, 0.05203, 0.0, -0.00703, 0.05203, 0.0, -0.00089, 0.05191, 0.0, -0.00414, -0.04288, -0.0, -0.00887, 0.06697, -0.0, -0.03626, -0.05131, -0.0, 0.00414, -0.04288, -0.0, 0.00703, 0.05203, -0.0, -0.00703, 0.05203, -0.0, -0.00089, 0.05191, -0.0, -0.0, -0.04877, 0.0, 0.03626, -0.05131, 0.0, 0.00887, 0.06697, -0.0, -0.0, 0.01934, 0.0, 0.00089, 0.05191, -0.0, -0.0, -0.02384, -0.0, -0.0, -0.02384, 0.0, 0.00172, -0.01985, 0.0, -0.00172, -0.01985, 0.0, -0.00172, -0.01985, -0.0, 0.00172, -0.01985, -0.0], [-0.0, 0.00099, -0.0, 0.00679, 0.00274, -0.0, 0.01505, -0.02328, -0.0, -0.0, -0.10729, 0.0, -0.04092, 0.03763, -0.0, 0.03167, 0.01301, 0.0, -0.01505, -0.02328, -0.0, -0.00679, 0.00274, -0.0, -0.03167, 0.01301, -0.0, -0.06018, 0.02109, -0.0, 0.06018, 0.02109, 0.0, 0.04092, 0.03763, 0.0, 0.03167, 0.01301, 0.0, -0.01505, -0.02328, -0.0, -0.00679, 0.00274, 0.0, -0.03167, 0.01301, -0.0, -0.06018, 0.02109, -0.0, 0.06018, 0.02109, 0.0, 0.04092, 0.03763, 0.0, -0.0, 0.00099, -0.0, 0.00679, 0.00274, 0.0, 0.01505, -0.02328, 0.0, -0.0, -0.10729, -0.0, -0.04092, 0.03763, -0.0, -0.0, -0.00988, 0.0, -0.0, -0.00988, -0.0, 0.0182, -0.01667, 0.0, -0.0182, -0.01667, 0.0, -0.0182, -0.01667, -0.0, 0.0182, -0.01667, -0.0], [-0.0, 0.02349, -0.0, 0.09986, 0.01572, 0.0, -0.00677, -0.00492, 0.0, 0.0, -0.03771, 0.0, 0.00306, -0.02751, 0.0, 0.04605, 0.01241, 0.0, 0.00677, -0.00492, 0.0, -0.09986, 0.01572, -0.0, -0.04605, 0.01241, 0.0, 0.05436, 0.01018, 0.0, -0.05436, 0.01018, -0.0, -0.00306, -0.02751, 0.0, 0.04605, 0.01241, -0.0, 0.00677, -0.00492, -0.0, -0.09986, 0.01572, -0.0, -0.04605, 0.01241, -0.0, 0.05436, 0.01018, 0.0, -0.05436, 0.01018, -0.0, -0.00306, -0.02751, -0.0, 0.0, 0.02349, -0.0, 0.09986, 0.01572, 0.0, -0.00677, -0.00492, -0.0, 0.0, -0.03771, -0.0, 0.00306, -0.02751, -0.0, 0.0, 0.01899, 0.0, 0.0, 0.01899, -0.0, -0.00334, 0.00101, 0.0, 0.00334, 0.00101, 0.0, 0.00334, 0.00101, -0.0, -0.00334, 0.00101, -0.0]])}, database_path = 'perylenediimide_scsd_20220715.pkl', maxdist = 1.75)
model_objs_dict['naphthalenediimide'] = scsd_model('naphthalenediimide', array([['0.7087', '-0.0', '-0.0', 'C'], ['-1.3992', '-1.2378', '0.0', 'C'], ['-2.8841', '1.2388', '0.0', 'C'], ['1.3992', '-1.2378', '-0.0', 'C'], ['-2.8841', '-1.2388', '0.0', 'C'], ['-0.7069', '-2.4312', '0.0', 'C'], ['0.7069', '-2.4312', '-0.0', 'C'], ['-0.7087', '0.0', '0.0', 'C'], ['1.3992', '1.2378', '-0.0', 'C'], ['2.8841', '-1.2388', '-0.0', 'C'], ['-1.3992', '1.2378', '0.0', 'C'], ['2.8841', '1.2388', '-0.0', 'C'], ['0.7069', '2.4312', '-0.0', 'C'], ['-0.7069', '2.4312', '0.0', 'C'], ['-3.5007', '0.0', '0.0', 'N'], ['3.5007', '-0.0', '-0.0', 'N'], ['-3.551', '-2.2597', '0.0', 'O'], ['-3.551', '2.2597', '0.0', 'O'], ['3.551', '2.2597', '-0.0', 'O'], ['3.551', '-2.2597', '-0.0', 'O']]), 'D2h' , pca = {"Ag":array([[-0.00537, -0.0, -0.0, 0.02474, 0.02713, 0.0, 0.04286, -0.03892, 0.0, -0.02474, 0.02713, 0.0, 0.04286, 0.03892, 0.0, 0.00751, 0.04439, 0.0, -0.00751, 0.04439, 0.0, 0.00537, -0.0, 0.0, -0.02474, -0.02713, -0.0, -0.04286, 0.03892, -0.0, 0.02474, -0.02713, -0.0, -0.04286, -0.03892, -0.0, -0.00751, -0.04439, -0.0, 0.00751, -0.04439, -0.0, 0.05895, -0.0, 0.0, -0.05895, 0.0, -0.0, 0.05383, 0.04821, 0.0, 0.05383, -0.04821, 0.0, -0.05383, -0.04821, -0.0, -0.05383, 0.04821, -0.0], [0.00901, 0.0, -0.0, -0.01882, 0.0212, 0.0, -0.07176, -0.05302, -0.0, 0.01882, 0.0212, 0.0, -0.07176, 0.05302, -0.0, -0.02224, 0.02387, 0.0, 0.02224, 0.02387, 0.0, -0.00901, 0.0, -0.0, 0.01882, -0.0212, -0.0, 0.07176, 0.05302, 0.0, -0.01882, -0.0212, -0.0, 0.07176, -0.05302, 0.0, 0.02224, -0.02387, -0.0, -0.02224, -0.02387, -0.0, -0.0525, 0.0, -0.0, 0.0525, -0.0, 0.0, -0.02401, 0.05962, 0.0, -0.02401, -0.05962, -0.0, 0.02401, -0.05962, -0.0, 0.02401, 0.05962, 0.0]]), "B1g":array([[0.0, -0.05042, 0.0, 0.00176, 0.0511, -0.0, -0.01361, 0.01907, -0.0, 0.00176, -0.0511, 0.0, 0.01361, 0.01907, 0.0, -0.04176, 0.06006, 0.0, -0.04176, -0.06006, 0.0, -0.0, 0.05042, 0.0, -0.00176, -0.0511, 0.0, 0.01361, -0.01907, 0.0, -0.00176, 0.0511, -0.0, -0.01361, -0.01907, -0.0, 0.04176, -0.06006, -0.0, 0.04176, 0.06006, -0.0, -0.0, -0.02433, -0.0, -0.0, 0.02433, 0.0, 0.05859, -0.02655, -0.0, -0.05859, -0.02655, -0.0, -0.05859, 0.02655, 0.0, 0.05859, 0.02655, 0.0], [-0.0, 0.01132, 0.0, 0.03102, 0.00077, -0.0, -0.01661, 0.06617, 0.0, 0.03102, -0.00077, -0.0, 0.01661, 0.06617, -0.0, 0.07256, -0.00598, -0.0, 0.07256, 0.00598, -0.0, -0.0, -0.01132, -0.0, -0.03102, -0.00077, 0.0, 0.01661, -0.06617, -0.0, -0.03102, 0.00077, 0.0, -0.01661, -0.06617, 0.0, -0.07256, 0.00598, 0.0, -0.07256, -0.00598, 0.0, -0.0, 0.03706, -0.0, -0.0, -0.03706, 0.0, 0.05865, 0.0336, -0.0, -0.05865, 0.0336, 0.0, -0.05865, -0.0336, 0.0, 0.05865, -0.0336, -0.0]]), "B2g":array([[-0.0, 0.0, 0.0, 0.0, 0.0, -0.03033, 0.0, 0.0, -0.0281, 0.0, 0.0, -0.03033, -0.0, -0.0, 0.0281, -0.0, -0.0, -0.0917, 0.0, -0.0, -0.0917, -0.0, -0.0, -0.0, -0.0, -0.0, 0.03033, -0.0, -0.0, 0.0281, -0.0, -0.0, 0.03033, -0.0, -0.0, -0.0281, 0.0, 0.0, 0.0917, -0.0, 0.0, 0.0917, -0.0, -0.0, 0.0, -0.0, 0.0, -0.0, 0.0, 0.0, 0.09987, 0.0, -0.0, -0.09987, -0.0, -0.0, -0.09987, -0.0, 0.0, 0.09987], [0.0, -0.0, -0.0, -0.0, 0.0, -0.06898, -0.0, -0.0, 0.09458, 0.0, 0.0, -0.06898, -0.0, -0.0, -0.09458, 0.0, 0.0, 0.04211, 0.0, 0.0, 0.04211, -0.0, -0.0, -0.0, 0.0, -0.0, 0.06898, -0.0, -0.0, -0.09458, -0.0, -0.0, 0.06898, -0.0, -0.0, 0.09458, -0.0, -0.0, -0.04211, -0.0, -0.0, -0.04211, -0.0, 0.0, 0.0, -0.0, -0.0, -0.0, 0.0, 0.0, 0.04433, -0.0, -0.0, -0.04433, -0.0, -0.0, -0.04433, 0.0, 0.0, 0.04433]]), "B3g":array([[-0.0, 0.0, -0.00374, -0.0, 0.0, 0.00437, -0.0, -0.0, 0.01829, -0.0, 0.0, -0.00437, -0.0, 0.0, 0.01829, 0.0, 0.0, -0.00025, -0.0, 0.0, 0.00025, -0.0, -0.0, 0.00374, 0.0, -0.0, -0.00437, -0.0, 0.0, -0.01829, 0.0, -0.0, 0.00437, -0.0, -0.0, -0.01829, -0.0, -0.0, 0.00025, 0.0, -0.0, -0.00025, -0.0, -0.0, 0.20981, -0.0, 0.0, -0.20981, -0.0, 0.0, -0.12032, 0.0, 0.0, -0.12032, 0.0, -0.0, 0.12032, -0.0, -0.0, 0.12032], [0.0, -0.0, -0.03178, 0.0, 0.0, 0.08034, -0.0, 0.0, 0.01743, -0.0, 0.0, -0.08034, -0.0, -0.0, 0.01743, 0.0, 0.0, 0.07255, -0.0, 0.0, -0.07255, -0.0, -0.0, 0.03178, -0.0, -0.0, -0.08034, -0.0, -0.0, -0.01743, 0.0, -0.0, 0.08034, -0.0, 0.0, -0.01743, -0.0, -0.0, -0.07255, 0.0, -0.0, 0.07255, -0.0, 0.0, -0.05081, -0.0, -0.0, 0.05081, -0.0, -0.0, -0.03838, -0.0, 0.0, -0.03838, 0.0, 0.0, 0.03838, 0.0, -0.0, 0.03838]]), "Au":array([[-0.0, -0.0, 0.0, 0.0, 0.0, -0.01909, 0.0, 0.0, 0.06452, -0.0, -0.0, 0.01909, -0.0, -0.0, -0.06452, -0.0, -0.0, -0.01299, -0.0, -0.0, 0.01299, -0.0, -0.0, -0.0, -0.0, -0.0, -0.01909, -0.0, -0.0, 0.06452, -0.0, 0.0, 0.01909, -0.0, -0.0, -0.06452, -0.0, -0.0, -0.01299, -0.0, -0.0, 0.01299, -0.0, -0.0, 0.0, -0.0, -0.0, -0.0, -0.0, -0.0, -0.15341, -0.0, -0.0, 0.15341, -0.0, 0.0, -0.15341, -0.0, 0.0, 0.15341], [-0.0, -0.0, -0.0, -0.0, 0.0, 0.0811, -0.0, -0.0, -0.0264, -0.0, -0.0, -0.0811, -0.0, -0.0, 0.0264, -0.0, -0.0, 0.11183, -0.0, -0.0, -0.11183, -0.0, 0.0, 0.0, -0.0, -0.0, 0.0811, -0.0, -0.0, -0.0264, -0.0, 0.0, -0.0811, -0.0, -0.0, 0.0264, -0.0, -0.0, 0.11183, -0.0, -0.0, -0.11183, -0.0, -0.0, -0.0, -0.0, -0.0, 0.0, -0.0, -0.0, -0.03066, -0.0, -0.0, 0.03066, -0.0, 0.0, -0.03066, -0.0, 0.0, 0.03066]]), "B1u":array([[-0.0, -0.0, -0.02271, -0.0, -0.0, -0.03443, -0.0, 0.0, 0.02937, -0.0, -0.0, -0.03443, -0.0, -0.0, 0.02937, -0.0, -0.0, -0.07921, -0.0, -0.0, -0.07921, -0.0, 0.0, -0.02271, -0.0, -0.0, -0.03443, -0.0, -0.0, 0.02937, -0.0, -0.0, -0.03443, -0.0, -0.0, 0.02937, -0.0, -0.0, -0.07921, -0.0, -0.0, -0.07921, -0.0, -0.0, 0.12218, -0.0, -0.0, 0.12218, -0.0, -0.0, 0.03454, -0.0, -0.0, 0.03454, -0.0, -0.0, 0.03454, -0.0, -0.0, 0.03454], [-0.0, -0.0, 0.0897, -0.0, -0.0, 0.03601, -0.0, 0.0, -0.0082, -0.0, -0.0, 0.03601, -0.0, -0.0, -0.0082, -0.0, -0.0, -0.01003, -0.0, -0.0, -0.01003, -0.0, -0.0, 0.0897, -0.0, -0.0, 0.03601, -0.0, -0.0, -0.0082, -0.0, -0.0, 0.03601, -0.0, -0.0, -0.0082, -0.0, -0.0, -0.01003, -0.0, -0.0, -0.01003, -0.0, -0.0, 0.08828, -0.0, -0.0, 0.08828, -0.0, 0.0, -0.10677, -0.0, 0.0, -0.10677, -0.0, 0.0, -0.10677, -0.0, 0.0, -0.10677]]), "B2u":array([[0.06337, -0.0, 0.0, 0.00619, -0.04787, -0.0, -0.02398, 0.05727, 0.0, 0.00619, 0.04787, 0.0, -0.02398, -0.05727, -0.0, 0.00574, -0.00064, 0.0, 0.00574, 0.00064, -0.0, 0.06337, -0.0, -0.0, 0.00619, -0.04787, -0.0, -0.02398, 0.05727, 0.0, 0.00619, 0.04787, 0.0, -0.02398, -0.05727, -0.0, 0.00574, -0.00064, -0.0, 0.00574, 0.00064, 0.0, -0.06791, 0.0, -0.0, -0.06791, 0.0, 0.0, 0.01433, -0.06815, -0.0, 0.01433, 0.06815, 0.0, 0.01433, -0.06815, -0.0, 0.01433, 0.06815, 0.0], [-0.02087, 0.0, -0.0, -0.03296, 0.07356, 0.0, -0.04244, 0.00388, -0.0, -0.03296, -0.07356, -0.0, -0.04244, -0.00388, -0.0, 0.02217, -0.02482, -0.0, 0.02217, 0.02482, 0.0, -0.02087, 0.0, -0.0, -0.03296, 0.07356, 0.0, -0.04244, 0.00388, 0.0, -0.03296, -0.07356, -0.0, -0.04244, -0.00388, 0.0, 0.02217, -0.02482, -0.0, 0.02217, 0.02482, 0.0, -0.0143, 0.0, -0.0, -0.0143, 0.0, 0.0, 0.07083, -0.02206, -0.0, 0.07083, 0.02206, 0.0, 0.07083, -0.02206, -0.0, 0.07083, 0.02206, 0.0]]), "B3u":array([[0.0, 0.05074, -0.0, 0.0123, 0.03199, 0.0, -0.03805, -0.02799, -0.0, -0.0123, 0.03199, -0.0, 0.03805, -0.02799, 0.0, 0.00747, 0.04175, 0.0, -0.00747, 0.04175, -0.0, 0.0, 0.05074, -0.0, 0.0123, 0.03199, 0.0, -0.03805, -0.02799, -0.0, -0.0123, 0.03199, -0.0, 0.03805, -0.02799, 0.0, 0.00747, 0.04175, 0.0, -0.00747, 0.04175, -0.0, 0.0, -0.03867, 0.0, 0.0, -0.03867, -0.0, 0.05614, -0.05177, 0.0, -0.05614, -0.05177, -0.0, 0.05614, -0.05177, 0.0, -0.05614, -0.05177, -0.0], [-0.0, -0.06652, 0.0, -0.01383, -0.00929, -0.0, -0.05843, 0.05122, -0.0, 0.01383, -0.00929, 0.0, 0.05843, 0.05122, 0.0, -0.00237, 0.00855, -0.0, 0.00237, 0.00855, -0.0, 0.0, -0.06652, -0.0, -0.01383, -0.00929, -0.0, -0.05843, 0.05122, -0.0, 0.01383, -0.00929, 0.0, 0.05843, 0.05122, 0.0, -0.00237, 0.00855, 0.0, 0.00237, 0.00855, 0.0, 0.0, -0.1302, -0.0, 0.0, -0.1302, 0.0, 0.01758, 0.04788, 0.0, -0.01758, 0.04788, -0.0, 0.01758, 0.04788, 0.0, -0.01758, 0.04788, -0.0]])}, database_path = 'naphthalenediimide_scsd_20220713.pkl', maxdist = 1.75)

model_objs_dict['pyrene'] = scsd_model('pyrene', array([['3.5174', '-0.0', '-0.0', 'C'], ['2.8304', '-1.2084', '-0.0', 'C'], ['1.4258', '-1.2315', '-0.0', 'C'], ['0.6908', '-2.4582', '-0.0', 'C'], ['-0.6908', '-2.4582', '0.0', 'C'], ['-1.4258', '-1.2315', '0.0', 'C'], ['-2.8304', '-1.2084', '0.0', 'C'], ['-3.5174', '0.0', '0.0', 'C'], ['-2.8304', '1.2084', '0.0', 'C'], ['-1.4258', '1.2315', '0.0', 'C'], ['-0.6908', '2.4582', '0.0', 'C'], ['0.6908', '2.4582', '-0.0', 'C'], ['1.4258', '1.2315', '-0.0', 'C'], ['2.8304', '1.2084', '-0.0', 'C'], ['0.7172', '-0.0', '-0.0', 'C'], ['-0.7172', '0.0', '0.0', 'C']]), 'D2h' , pca = {"Ag":array([[-0.11844, 0.0, -0.0, -0.07414, 0.02594, -0.0, 0.02429, -0.07984, -0.0, -0.0194, 0.03895, 0.0, 0.0194, 0.03895, 0.0, -0.02429, -0.07984, -0.0, 0.07414, 0.02594, 0.0, 0.11844, -0.0, 0.0, 0.07414, -0.02594, 0.0, -0.02429, 0.07984, 0.0, 0.0194, -0.03895, -0.0, -0.0194, -0.03895, -0.0, 0.02429, 0.07984, 0.0, -0.07414, -0.02594, -0.0, -0.01133, 0.0, -0.0, 0.01133, -0.0, 0.0], [0.10282, -0.0, 0.0, 0.05298, 0.02042, 0.0, 0.06924, -0.08006, -0.0, -0.00088, 0.03666, 0.0, 0.00088, 0.03666, 0.0, -0.06924, -0.08006, -0.0, -0.05298, 0.02042, -0.0, -0.10282, 0.0, -0.0, -0.05298, -0.02042, -0.0, -0.06924, 0.08006, 0.0, 0.00088, -0.03666, -0.0, -0.00088, -0.03666, -0.0, 0.06924, 0.08006, 0.0, 0.05298, -0.02042, 0.0, 0.00551, -0.0, 0.0, -0.00551, 0.0, -0.0]]), "B1g":array([[-0.0, 0.01267, 0.0, -0.09415, 0.06457, 0.0, -0.02644, 0.02376, 0.0, -0.0287, -0.05557, -0.0, -0.0287, 0.05557, -0.0, -0.02644, -0.02376, 0.0, -0.09415, -0.06457, -0.0, 0.0, -0.01267, -0.0, 0.09415, -0.06457, -0.0, 0.02644, -0.02376, -0.0, 0.0287, 0.05557, 0.0, 0.0287, -0.05557, 0.0, 0.02644, 0.02376, -0.0, 0.09415, 0.06457, 0.0, 0.0, 0.04563, 0.0, 0.0, -0.04563, -0.0], [-0.0, 0.11528, 0.0, 0.041, 0.03306, 0.0, -0.01114, -0.0457, -0.0, -0.10288, 0.00272, 0.0, -0.10288, -0.00272, 0.0, -0.01114, 0.0457, -0.0, 0.041, -0.03306, -0.0, 0.0, -0.11528, -0.0, -0.041, -0.03306, -0.0, 0.01114, 0.0457, 0.0, 0.10288, -0.00272, -0.0, 0.10288, 0.00272, -0.0, 0.01114, -0.0457, 0.0, -0.041, 0.03306, 0.0, 0.0, -0.04454, 0.0, 0.0, 0.04454, -0.0]]), "B2g":array([[0.0, -0.0, 0.0, -0.0, -0.0, 0.13985, -0.0, -0.0, 0.02758, 0.0, 0.0, -0.08257, -0.0, 0.0, -0.08257, 0.0, -0.0, 0.02758, 0.0, 0.0, 0.13985, 0.0, 0.0, -0.0, 0.0, 0.0, -0.13985, 0.0, 0.0, -0.02758, -0.0, -0.0, 0.08257, 0.0, -0.0, 0.08257, -0.0, 0.0, -0.02758, 0.0, -0.0, -0.13985, -0.0, 0.0, -0.0, 0.0, -0.0, 0.0], [-0.0, 0.0, 0.0, 0.0, 0.0, -0.05635, -0.0, 0.0, 0.14747, 0.0, 0.0, -0.04618, -0.0, 0.0, -0.04618, 0.0, 0.0, 0.14747, 0.0, 0.0, -0.05635, 0.0, -0.0, -0.0, 0.0, -0.0, 0.05635, 0.0, -0.0, -0.14747, -0.0, -0.0, 0.04618, 0.0, -0.0, 0.04618, -0.0, -0.0, -0.14747, 0.0, -0.0, 0.05635, 0.0, 0.0, 0.0, -0.0, -0.0, -0.0]]), "B3g":array([[-0.0, -0.0, 0.10459, -0.0, 0.0, 0.00211, 0.0, 0.0, -0.07402, -0.0, -0.0, -0.106, 0.0, -0.0, 0.106, -0.0, 0.0, 0.07402, -0.0, 0.0, -0.00211, -0.0, 0.0, -0.10459, -0.0, -0.0, -0.00211, -0.0, -0.0, 0.07402, 0.0, 0.0, 0.106, -0.0, 0.0, -0.106, 0.0, -0.0, -0.07402, -0.0, -0.0, 0.00211, -0.0, 0.0, -0.03115, -0.0, -0.0, 0.03115], [0.0, 0.0, 0.1234, -0.0, -0.0, -0.08207, 0.0, 0.0, -0.02602, -0.0, 0.0, 0.07805, 0.0, 0.0, -0.07805, -0.0, 0.0, 0.02602, -0.0, -0.0, 0.08207, -0.0, 0.0, -0.1234, -0.0, 0.0, 0.08207, -0.0, -0.0, 0.02602, 0.0, -0.0, -0.07805, -0.0, -0.0, 0.07805, 0.0, -0.0, -0.02602, -0.0, 0.0, -0.08207, -0.0, 0.0, -0.00432, -0.0, -0.0, 0.00432]]), "Au":array([[-0.0, -0.0, -0.0, 0.0, -0.0, -0.1126, 0.0, -0.0, -0.07288, -0.0, -0.0, -0.06452, -0.0, -0.0, 0.06452, -0.0, 0.0, 0.07288, -0.0, 0.0, 0.1126, -0.0, -0.0, 0.0, -0.0, 0.0, -0.1126, -0.0, 0.0, -0.07288, 0.0, -0.0, -0.06452, 0.0, -0.0, 0.06452, -0.0, -0.0, 0.07288, -0.0, -0.0, 0.1126, -0.0, 0.0, 0.0, -0.0, 0.0, -0.0], [0.0, -0.0, 0.0, -0.0, 0.0, 0.09293, -0.0, -0.0, -0.03932, -0.0, -0.0, -0.11775, 0.0, -0.0, 0.11775, -0.0, 0.0, 0.03932, -0.0, -0.0, -0.09293, -0.0, 0.0, -0.0, -0.0, -0.0, 0.09293, -0.0, 0.0, -0.03932, 0.0, -0.0, -0.11775, -0.0, -0.0, 0.11775, -0.0, -0.0, 0.03932, -0.0, 0.0, -0.09293, -0.0, -0.0, 0.0, -0.0, -0.0, -0.0]]), "B1u":array([[-0.0, -0.0, 0.12542, -0.0, 0.0, 0.06229, 0.0, 0.0, -0.02637, -0.0, -0.0, -0.07516, -0.0, -0.0, -0.07516, -0.0, -0.0, -0.02637, -0.0, -0.0, 0.06229, -0.0, -0.0, 0.12542, -0.0, -0.0, 0.06229, -0.0, -0.0, -0.02637, -0.0, -0.0, -0.07516, -0.0, -0.0, -0.07516, -0.0, -0.0, -0.02637, -0.0, -0.0, 0.06229, -0.0, 0.0, -0.04694, -0.0, 0.0, -0.04694], [0.0, -0.0, -0.04214, 0.0, 0.0, -0.00407, -0.0, -0.0, 0.06114, 0.0, 0.0, -0.09986, 0.0, 0.0, -0.09986, 0.0, 0.0, 0.06114, 0.0, 0.0, -0.00407, 0.0, 0.0, -0.04214, 0.0, 0.0, -0.00407, 0.0, 0.0, 0.06114, 0.0, 0.0, -0.09986, 0.0, 0.0, -0.09986, 0.0, 0.0, 0.06114, 0.0, 0.0, -0.00407, 0.0, -0.0, 0.12771, 0.0, -0.0, 0.12771]]), "B2u":array([[-0.02389, 0.0, -0.0, -0.03555, 0.0609, 0.0, -0.01791, 0.04418, 0.0, 0.09121, 0.03319, 0.0, 0.09121, -0.03319, -0.0, -0.01791, -0.04418, -0.0, -0.03555, -0.0609, -0.0, -0.02389, 0.0, 0.0, -0.03555, 0.0609, 0.0, -0.01791, 0.04418, 0.0, 0.09121, 0.03319, 0.0, 0.09121, -0.03319, -0.0, -0.01791, -0.04418, -0.0, -0.03555, -0.0609, -0.0, -0.05161, 0.0, -0.0, -0.05161, 0.0, 0.0], [-0.15641, 0.0, 0.0, 0.0182, -0.10211, -0.0, 0.00238, -0.01163, -0.0, 0.05665, 0.01318, 0.0, 0.05665, -0.01318, -0.0, 0.00238, 0.01163, 0.0, 0.0182, 0.10211, 0.0, -0.15641, 0.0, 0.0, 0.0182, -0.10211, -0.0, 0.00238, -0.01163, -0.0, 0.05665, 0.01318, 0.0, 0.05665, -0.01318, -0.0, 0.00238, 0.01163, 0.0, 0.0182, 0.10211, 0.0, 0.00195, -0.0, -0.0, 0.00195, -0.0, 0.0]]), "B3u":array([[0.0, -0.10916, 0.0, -0.04894, -0.05281, -0.0, -0.03318, 0.03036, -0.0, -0.0241, 0.05035, -0.0, 0.0241, 0.05035, 0.0, 0.03318, 0.03036, 0.0, 0.04894, -0.05281, 0.0, -0.0, -0.10916, -0.0, -0.04894, -0.05281, -0.0, -0.03318, 0.03036, -0.0, -0.0241, 0.05035, -0.0, 0.0241, 0.05035, 0.0, 0.03318, 0.03036, 0.0, 0.04894, -0.05281, 0.0, -0.0, 0.05338, 0.0, -0.0, 0.05338, -0.0], [-0.0, -0.00223, 0.0, 0.09704, -0.06392, 0.0, 0.05221, 0.00568, 0.0, 0.00554, 0.07335, 0.0, -0.00554, 0.07335, -0.0, -0.05221, 0.00568, -0.0, -0.09704, -0.06392, -0.0, 0.0, -0.00223, -0.0, 0.09704, -0.06392, 0.0, 0.05221, 0.00568, 0.0, 0.00554, 0.07335, 0.0, -0.00554, 0.07335, -0.0, -0.05221, 0.00568, -0.0, -0.09704, -0.06392, -0.0, 0.0, -0.028, -0.0, 0.0, -0.028, 0.0]])}, database_path = 'pyrene_scsd_20220713.pkl', maxdist = 1.75, mondrian_limits = [-1.0, -1.0])
model_objs_dict['anthracene'] = scsd_model('anthracene', array([[3.6753, 0.7155, -0.0], [2.486, 1.4106, -0.0], [1.2267, 0.7226, -0.0], [0.0, 1.4233, 0.0], [3.6753, -0.7155, -0.0], [2.486, -1.4106, -0.0], [1.2267, -0.7226, -0.0], [-1.2267, 0.7226, 0.0], [-1.2267, -0.7226, 0.0], [-2.486, -1.4106, 0.0], [-3.6753, -0.7155, 0.0], [-3.6753, 0.7155, 0.0], [-2.486, 1.4106, 0.0], [-0.0, -1.4233, 0.0]]), 'D2h' , pca = {"Ag":array([[0.14039, -0.00473, 0.0, 0.06235, 0.00809, 0.0, 0.03925, 0.00225, 0.0, 0.0, 0.037, 0.0, 0.14039, 0.00473, 0.0, 0.06235, -0.00809, 0.0, 0.03925, -0.00225, 0.0, -0.03925, 0.00225, -0.0, -0.03925, -0.00225, -0.0, -0.06235, -0.00809, -0.0, -0.14039, 0.00473, -0.0, -0.14039, -0.00473, -0.0, -0.06235, 0.00809, -0.0, -0.0, -0.037, -0.0], [-0.00226, 0.09515, -0.0, 0.01854, 0.10645, 0.0, -0.01231, 0.02165, 0.0, -0.0, -0.04405, -0.0, -0.00226, -0.09515, -0.0, 0.01854, -0.10645, -0.0, -0.01231, -0.02165, -0.0, 0.01231, 0.02165, 0.0, 0.01231, -0.02165, -0.0, -0.01854, -0.10645, -0.0, 0.00226, -0.09515, -0.0, 0.00226, 0.09515, 0.0, -0.01854, 0.10645, 0.0, 0.0, 0.04405, 0.0]]), "B1g":array([[-0.04539, 0.0273, 0.0, -0.07726, -0.04971, 0.0, -0.01755, -0.09357, -0.0, 0.02266, 0.0, -0.0, 0.04539, 0.0273, -0.0, 0.07726, -0.04971, 0.0, 0.01755, -0.09357, -0.0, -0.01755, 0.09357, 0.0, 0.01755, 0.09357, 0.0, 0.07726, 0.04971, -0.0, 0.04539, -0.0273, -0.0, -0.04539, -0.0273, 0.0, -0.07726, 0.04971, -0.0, -0.02266, 0.0, 0.0], [-0.05217, -0.06258, 0.0, -0.057, -0.02123, -0.0, -0.03038, 0.05908, -0.0, -0.0992, -0.0, -0.0, 0.05217, -0.06258, -0.0, 0.057, -0.02123, 0.0, 0.03038, 0.05908, -0.0, -0.03038, -0.05908, 0.0, 0.03038, -0.05908, 0.0, 0.057, 0.02123, 0.0, 0.05217, 0.06258, -0.0, -0.05217, 0.06258, 0.0, -0.057, 0.02123, -0.0, 0.0992, -0.0, 0.0]]), "B2g":array([[-0.0, -0.0, -0.07947, 0.0, -0.0, -0.05653, 0.0, -0.0, 0.03659, -0.0, 0.0, 0.1548, -0.0, 0.0, 0.07947, 0.0, 0.0, 0.05653, 0.0, 0.0, -0.03659, 0.0, -0.0, 0.03659, 0.0, 0.0, -0.03659, 0.0, 0.0, 0.05653, 0.0, 0.0, 0.07947, 0.0, -0.0, -0.07947, 0.0, -0.0, -0.05653, 0.0, 0.0, -0.1548], [0.0, -0.0, -0.1246, -0.0, 0.0, 0.09384, -0.0, 0.0, -0.00244, 0.0, -0.0, -0.05825, -0.0, -0.0, 0.1246, -0.0, -0.0, -0.09384, -0.0, -0.0, 0.00244, -0.0, 0.0, -0.00244, -0.0, -0.0, 0.00244, -0.0, -0.0, -0.09384, 0.0, -0.0, 0.1246, 0.0, 0.0, -0.1246, -0.0, 0.0, 0.09384, -0.0, -0.0, 0.05825]]), "B3g":array([[-0.0, -0.0, -0.08797, -0.0, 0.0, 0.0989, 0.0, -0.0, 0.06313, -0.0, -0.0, -0.0, -0.0, 0.0, -0.08797, -0.0, -0.0, 0.0989, 0.0, 0.0, 0.06313, -0.0, -0.0, -0.06313, -0.0, 0.0, -0.06313, -0.0, -0.0, -0.0989, 0.0, 0.0, 0.08797, 0.0, -0.0, 0.08797, -0.0, 0.0, -0.0989, -0.0, -0.0, -0.0], [0.0, -0.0, 0.0093, 0.0, -0.0, -0.08874, -0.0, 0.0, 0.15196, 0.0, 0.0, 0.0, 0.0, -0.0, 0.0093, 0.0, 0.0, -0.08874, -0.0, -0.0, 0.15196, 0.0, 0.0, -0.15196, 0.0, -0.0, -0.15196, 0.0, 0.0, 0.08874, -0.0, -0.0, -0.0093, -0.0, 0.0, -0.0093, 0.0, -0.0, 0.08874, 0.0, 0.0, 0.0]]), "Au":array([[0.0, -0.0, 0.08181, 0.0, 0.0, 0.1333, 0.0, -0.0, 0.03489, -0.0, -0.0, 0.0, 0.0, 0.0, -0.08181, 0.0, 0.0, -0.1333, 0.0, -0.0, -0.03489, 0.0, 0.0, -0.03489, 0.0, 0.0, 0.03489, 0.0, -0.0, 0.1333, 0.0, -0.0, 0.08181, 0.0, 0.0, -0.08181, 0.0, -0.0, -0.1333, 0.0, -0.0, -0.0], [-0.0, 0.0, -0.13053, 0.0, -0.0, 0.06616, 0.0, -0.0, 0.05331, 0.0, -0.0, 0.0, 0.0, -0.0, 0.13053, 0.0, -0.0, -0.06616, 0.0, -0.0, -0.05331, 0.0, 0.0, -0.05331, 0.0, 0.0, 0.05331, 0.0, 0.0, 0.06616, 0.0, 0.0, -0.13053, 0.0, -0.0, 0.13053, 0.0, 0.0, -0.06616, -0.0, -0.0, -0.0]]), "B1u":array([[0.0, 0.0, -0.10849, -0.0, -0.0, -0.01651, -0.0, -0.0, 0.06547, -0.0, -0.0, 0.11907, -0.0, 0.0, -0.10849, -0.0, -0.0, -0.01651, -0.0, -0.0, 0.06547, -0.0, -0.0, 0.06547, -0.0, -0.0, 0.06547, -0.0, -0.0, -0.01651, -0.0, 0.0, -0.10849, -0.0, 0.0, -0.10849, -0.0, -0.0, -0.01651, -0.0, -0.0, 0.11907], [0.0, 0.0, -0.06124, 0.0, 0.0, 0.08826, 0.0, -0.0, 0.03674, 0.0, 0.0, -0.12752, 0.0, 0.0, -0.06124, 0.0, 0.0, 0.08826, 0.0, -0.0, 0.03674, 0.0, -0.0, 0.03674, 0.0, -0.0, 0.03674, 0.0, 0.0, 0.08826, 0.0, 0.0, -0.06124, 0.0, 0.0, -0.06124, 0.0, 0.0, 0.08826, 0.0, 0.0, -0.12752]]), "B2u":array([[-0.10003, -0.03939, -0.0, 0.04441, -0.06367, -0.0, 0.04096, -0.02528, -0.0, 0.02933, -0.0, -0.0, -0.10003, 0.03939, 0.0, 0.04441, 0.06367, 0.0, 0.04096, 0.02528, 0.0, 0.04096, 0.02528, 0.0, 0.04096, -0.02528, -0.0, 0.04441, -0.06367, -0.0, -0.10003, -0.03939, -0.0, -0.10003, 0.03939, 0.0, 0.04441, 0.06367, 0.0, 0.02933, -0.0, -0.0], [-0.06557, 0.0095, -0.0, -0.0188, 0.12737, 0.0, 0.03806, 0.0062, 0.0, 0.09261, 0.0, -0.0, -0.06557, -0.0095, -0.0, -0.0188, -0.12737, -0.0, 0.03806, -0.0062, -0.0, 0.03806, -0.0062, -0.0, 0.03806, 0.0062, 0.0, -0.0188, 0.12737, 0.0, -0.06557, 0.0095, 0.0, -0.06557, -0.0095, -0.0, -0.0188, -0.12737, -0.0, 0.09261, 0.0, -0.0]]), "B3u":array([[-0.03103, 0.09384, -0.0, -0.06052, 0.01358, -0.0, -0.01399, -0.06242, -0.0, 0.0, -0.09001, 0.0, 0.03103, 0.09384, 0.0, 0.06052, 0.01358, 0.0, 0.01399, -0.06242, 0.0, 0.01399, -0.06242, 0.0, -0.01399, -0.06242, -0.0, -0.06052, 0.01358, -0.0, -0.03103, 0.09384, -0.0, 0.03103, 0.09384, 0.0, 0.06052, 0.01358, 0.0, -0.0, -0.09001, -0.0], [0.08589, 0.01346, -0.0, 0.07153, 0.05, 0.0, 0.02738, -0.01001, 0.0, 0.0, -0.1069, 0.0, -0.08589, 0.01346, -0.0, -0.07153, 0.05, -0.0, -0.02738, -0.01001, -0.0, -0.02738, -0.01001, -0.0, 0.02738, -0.01001, 0.0, 0.07153, 0.05, 0.0, 0.08589, 0.01346, 0.0, -0.08589, 0.01346, -0.0, -0.07153, 0.05, -0.0, -0.0, -0.1069, -0.0]])}, database_path = 'anthracene_scsd_20220713.pkl', maxdist = 1.75)
model_objs_dict['naphthalene'] = scsd_model('naphthalene', array([['2.4265', '-0.7079', '0.0', 'C'], ['1.2391', '-1.4005', '-0.0', 'C'], ['0.0', '-0.71', '-0.0', 'C'], ['-1.2391', '-1.4005', '-0.0', 'C'], ['2.4265', '0.7079', '0.0', 'C'], ['0.0', '0.71', '0.0', 'C'], ['-1.2391', '1.4005', '0.0', 'C'], ['-2.4265', '0.7079', '-0.0', 'C'], ['-2.4265', '-0.7079', '-0.0', 'C'], ['1.2391', '1.4005', '0.0', 'C']]), 'D2h' , pca = {"Ag":array([[-0.11276, 0.00338, -0.0, -0.10561, -0.07205, -0.0, 0.0, 0.03176, 0.0, 0.10561, -0.07205, 0.0, -0.11276, -0.00338, -0.0, -0.0, -0.03176, -0.0, 0.10561, 0.07205, 0.0, 0.11276, -0.00338, 0.0, 0.11276, 0.00338, 0.0, -0.10561, 0.07205, -0.0], [0.09464, -0.03357, 0.0, 0.00529, -0.16121, -0.0, -0.0, -0.01718, -0.0, -0.00529, -0.16121, -0.0, 0.09464, 0.03357, 0.0, 0.0, 0.01718, 0.0, -0.00529, 0.16121, 0.0, -0.09464, 0.03357, -0.0, -0.09464, -0.03357, -0.0, 0.00529, 0.16121, 0.0]]), "B1g":array([[-0.05499, 0.06252, -0.0, -0.13171, 0.04141, -0.0, 0.05732, 0.0, 0.0, -0.13171, -0.04141, -0.0, 0.05499, 0.06252, -0.0, -0.05732, 0.0, -0.0, 0.13171, -0.04141, 0.0, 0.05499, -0.06252, 0.0, -0.05499, -0.06252, 0.0, 0.13171, 0.04141, 0.0], [-0.08733, -0.00654, 0.0, 0.02265, 0.08737, 0.0, -0.17546, 0.0, -0.0, 0.02265, -0.08737, -0.0, 0.08733, -0.00654, 0.0, 0.17546, 0.0, 0.0, -0.02265, -0.08737, -0.0, 0.08733, 0.00654, -0.0, -0.08733, 0.00654, -0.0, -0.02265, 0.08737, 0.0]]), "B2g":array([[0.0, 0.0, 0.1522, -0.0, 0.0, -0.05546, 0.0, -0.0, -0.08468, 0.0, 0.0, -0.05546, -0.0, -0.0, -0.1522, -0.0, 0.0, 0.08468, 0.0, -0.0, 0.05546, 0.0, -0.0, -0.1522, 0.0, 0.0, 0.1522, -0.0, -0.0, 0.05546], [0.0, 0.0, -0.03963, -0.0, 0.0, 0.08407, 0.0, 0.0, -0.2526, 0.0, 0.0, 0.08407, 0.0, 0.0, 0.03963, -0.0, -0.0, 0.2526, 0.0, -0.0, -0.08407, -0.0, 0.0, 0.03963, -0.0, -0.0, -0.03963, -0.0, -0.0, -0.08407]]), "B3g":array([[0.0, 0.0, -0.08451, 0.0, -0.0, 0.16549, -0.0, -0.0, -0.0, -0.0, -0.0, -0.16549, -0.0, -0.0, -0.08451, 0.0, 0.0, 0.0, -0.0, 0.0, -0.16549, 0.0, -0.0, 0.08451, 0.0, 0.0, 0.08451, 0.0, 0.0, 0.16549]]), "Au":array([[-0.0, 0.0, 0.10428, -0.0, -0.0, 0.14572, 0.0, -0.0, -0.0, -0.0, 0.0, -0.14572, -0.0, 0.0, -0.10428, -0.0, -0.0, 0.0, -0.0, 0.0, 0.14572, -0.0, -0.0, 0.10428, -0.0, -0.0, -0.10428, -0.0, -0.0, -0.14572], [-0.0, 0.0, 0.14572, -0.0, -0.0, -0.10428, -0.0, -0.0, 0.0, -0.0, 0.0, 0.10428, -0.0, 0.0, -0.14572, 0.0, -0.0, -0.0, -0.0, 0.0, -0.10428, -0.0, -0.0, 0.14572, -0.0, -0.0, -0.14572, -0.0, -0.0, 0.10428]]), "B1u":array([[-0.0, 0.0, 0.125, 0.0, 0.0, -0.06035, 0.0, 0.0, -0.1293, 0.0, 0.0, -0.06035, 0.0, -0.0, 0.125, 0.0, 0.0, -0.1293, 0.0, 0.0, -0.06035, 0.0, -0.0, 0.125, 0.0, -0.0, 0.125, 0.0, 0.0, -0.06035], [0.0, 0.0, 0.03389, 0.0, 0.0, -0.125, -0.0, 0.0, 0.18221, 0.0, 0.0, -0.125, 0.0, 0.0, 0.03389, -0.0, 0.0, 0.18221, 0.0, 0.0, -0.125, 0.0, 0.0, 0.03389, 0.0, -0.0, 0.03389, 0.0, 0.0, -0.125]]), "B2u":array([[0.02924, 0.04591, 0.0, -0.00674, 0.18095, 0.0, -0.04499, -0.0, -0.0, -0.00674, -0.18095, -0.0, 0.02924, -0.04591, -0.0, -0.04499, 0.0, 0.0, -0.00674, 0.18095, 0.0, 0.02924, 0.04591, 0.0, 0.02924, -0.04591, -0.0, -0.00674, -0.18095, -0.0], [-0.10665, 0.08719, 0.0, 0.04711, 0.01167, 0.0, 0.11908, 0.0, 0.0, 0.04711, -0.01167, -0.0, -0.10665, -0.08719, -0.0, 0.11908, -0.0, -0.0, 0.04711, 0.01167, 0.0, -0.10665, 0.08719, 0.0, -0.10665, -0.08719, -0.0, 0.04711, -0.01167, -0.0]]), "B3u":array([[-0.02911, -0.09928, -0.0, -0.08341, 0.02822, -0.0, 0.0, 0.14213, 0.0, 0.08341, 0.02822, 0.0, 0.02911, -0.09928, 0.0, -0.0, 0.14213, -0.0, -0.08341, 0.02822, -0.0, -0.02911, -0.09928, -0.0, 0.02911, -0.09928, 0.0, 0.08341, 0.02822, 0.0], [-0.13326, -0.0254, 0.0, 0.01443, 0.07731, 0.0, 0.0, -0.10383, 0.0, -0.01443, 0.07731, -0.0, 0.13326, -0.0254, 0.0, -0.0, -0.10383, -0.0, 0.01443, 0.07731, 0.0, -0.13326, -0.0254, -0.0, 0.13326, -0.0254, 0.0, -0.01443, 0.07731, -0.0]])}, maxdist = 1.75)
model_objs_dict['benzoquinone'] = scsd_model('benzoquinone', array([['0.696', '1.3181', '-0.0', 'C'], ['-0.696', '1.3181', '-0.0', 'C'], ['-1.4954', '-0.0', '0.0', 'C'], ['-0.696', '-1.3181', '0.0', 'C'], ['0.696', '-1.3181', '0.0', 'C'], ['1.4954', '0.0', '0.0', 'C'], ['2.7646', '0.0', '0.0', 'O'], ['-2.7646', '-0.0', '0.0', 'O']]), 'D2h' , pca = {"Ag":array([[0.06191, -0.07832, -0.0, -0.06191, -0.07832, -0.0, -0.1177, 0.0, -0.0, -0.06191, 0.07832, 0.0, 0.06191, 0.07832, 0.0, 0.1177, -0.0, 0.0, 0.20618, -0.0, 0.0, -0.20618, 0.0, -0.0], [0.08696, 0.17987, 0.0, -0.08696, 0.17987, 0.0, -0.01889, 0.0, -0.0, -0.08696, -0.17987, -0.0, 0.08696, -0.17987, -0.0, 0.01889, -0.0, 0.0, 0.07364, -0.0, 0.0, -0.07364, 0.0, -0.0]]), "B1g":array([[-0.08489, -0.00963, -0.0, -0.08489, 0.00963, 0.0, 0.0, 0.29687, 0.0, 0.08489, 0.00963, 0.0, 0.08489, -0.00963, -0.0, 0.0, -0.29687, -0.0, 0.0, 0.08448, -0.0, 0.0, -0.08448, 0.0], [-0.03057, -0.19867, 0.0, -0.03057, 0.19867, 0.0, 0.0, -0.0438, 0.0, 0.03057, 0.19867, -0.0, 0.03057, -0.19867, -0.0, 0.0, 0.0438, -0.0, 0.0, 0.04719, -0.0, 0.0, -0.04719, 0.0]]), "B3g":array([[-0.0, 0.0, 0.14144, 0.0, 0.0, -0.14144, 0.0, 0.0, -0.09469, -0.0, 0.0, -0.14144, -0.0, 0.0, 0.14144, 0.0, -0.0, 0.09469, 0.0, 0.0, -0.12244, 0.0, -0.0, 0.12244], [0.0, -0.0, 0.10426, -0.0, -0.0, -0.10426, -0.0, -0.0, 0.22322, -0.0, -0.0, -0.10426, 0.0, -0.0, 0.10426, -0.0, 0.0, -0.22322, -0.0, -0.0, 0.06825, -0.0, 0.0, -0.06825]]), "Au":array([[-0.0, -0.0, -0.25, -0.0, -0.0, 0.25, -0.0, 0.0, 0.0, -0.0, -0.0, -0.25, -0.0, -0.0, 0.25, -0.0, 0.0, -0.0, -0.0, 0.0, -0.0, -0.0, 0.0, 0.0]]), "B1u":array([[-0.0, -0.0, -0.125, -0.0, -0.0, -0.125, -0.0, -0.0, 0.02425, -0.0, -0.0, -0.125, -0.0, -0.0, -0.125, -0.0, -0.0, 0.02425, -0.0, -0.0, 0.22575, -0.0, -0.0, 0.22575], [-0.0, -0.0, -0.07181, -0.0, -0.0, -0.07181, -0.0, -0.0, 0.25, -0.0, -0.0, -0.07181, -0.0, -0.0, -0.07181, -0.0, -0.0, 0.25, -0.0, 0.0, -0.10638, -0.0, 0.0, -0.10638]]), "B2u":array([[-0.01259, -0.20831, -0.0, -0.01259, 0.20831, 0.0, -0.05537, 0.0, 0.0, -0.01259, -0.20831, -0.0, -0.01259, 0.20831, 0.0, -0.05537, 0.0, -0.0, 0.08055, -0.0, 0.0, 0.08055, -0.0, -0.0], [0.12462, -0.01362, -0.0, 0.12462, 0.01362, 0.0, -0.12904, 0.0, -0.0, 0.12462, -0.01362, 0.0, 0.12462, 0.01362, -0.0, -0.12904, 0.0, 0.0, -0.1202, 0.0, -0.0, -0.1202, 0.0, 0.0]]), "B3u":array([[-0.07952, -0.084, -0.0, 0.07952, -0.084, 0.0, 0.0, -0.05192, -0.0, -0.07952, -0.084, -0.0, 0.07952, -0.084, 0.0, 0.0, -0.05192, 0.0, 0.0, 0.21992, 0.0, 0.0, 0.21992, -0.0], [-0.12111, 0.09433, -0.0, 0.12111, 0.09433, 0.0, 0.0, -0.14007, -0.0, -0.12111, 0.09433, -0.0, 0.12111, 0.09433, 0.0, 0.0, -0.14007, 0.0, 0.0, -0.04859, -0.0, 0.0, -0.04859, 0.0]])}, database_path = 'benzoquinone_scsd_20220713.pkl', maxdist = 1.75)

#model_objs_dict['keggin'] = scsd_model('keggin', array([['-0.0', '-3.7005', '-3.7005', 'O'], ['-1.3091', '-1.3091', '3.1481', 'O'], ['-1.3091', '1.3091', '3.1481', 'O'], ['-3.7005', '3.7005', '0.0', 'O'], ['-1.3091', '3.1481', '-1.3091', 'O'], ['-1.3091', '1.3091', '-3.1481', 'O'], ['-3.7005', '0.0', '-3.7005', 'O'], ['-1.3091', '-1.3091', '-3.1481', 'O'], ['1.3091', '-3.1481', '-1.3091', 'O'], ['3.1481', '-1.3091', '-1.3091', 'O'], ['1.3091', '1.3091', '-3.1481', 'O'], ['-3.7005', '-0.0', '3.7005', 'O'], ['-3.7005', '-3.7005', '-0.0', 'O'], ['-1.3091', '-3.1481', '-1.3091', 'O'], ['-0.0', '-3.7005', '3.7005', 'O'], ['1.3091', '-3.1481', '1.3091', 'O'], ['3.7005', '-0.0', '3.7005', 'O'], ['3.1481', '-1.3091', '1.3091', 'O'], ['3.1481', '1.3091', '1.3091', 'O'], ['3.1481', '1.3091', '-1.3091', 'O'], ['0.0', '3.7005', '3.7005', 'O'], ['1.3091', '3.1481', '1.3091', 'O'], ['-3.1481', '1.3091', '1.3091', 'O'], ['0.0', '3.7005', '-3.7005', 'O'], ['1.3091', '3.1481', '-1.3091', 'O'], ['1.3091', '-1.3091', '-3.1481', 'O'], ['3.7005', '-3.7005', '-0.0', 'O'], ['3.7005', '0.0', '-3.7005', 'O'], ['3.7005', '3.7005', '0.0', 'O'], ['-3.1481', '1.3091', '-1.3091', 'O'], ['-3.1481', '-1.3091', '-1.3091', 'O'], ['-1.3091', '-3.1481', '1.3091', 'O'], ['1.3091', '-1.3091', '3.1481', 'O'], ['1.3091', '1.3091', '3.1481', 'O'], ['-1.3091', '3.1481', '1.3091', 'O'], ['-3.1481', '-1.3091', '1.3091', 'O'], ['2.5118', '-2.5118', '-0.0', 'W'], ['2.5118', '0.0', '-2.5118', 'W'], ['2.5118', '2.5118', '0.0', 'W'], ['-2.5118', '-0.0', '2.5118', 'W'], ['-2.5118', '2.5118', '0.0', 'W'], ['-2.5118', '0.0', '-2.5118', 'W'], ['-2.5118', '-2.5118', '-0.0', 'W'], ['-0.0', '-2.5118', '2.5118', 'W'], ['2.5118', '-0.0', '2.5118', 'W'], ['0.0', '2.5118', '2.5118', 'W'], ['0.0', '2.5118', '-2.5118', 'W'], ['-0.0', '-2.5118', '-2.5118', 'W']]), 'Oh', maxdist = 2.0)
model_objs_dict['keggin'] = scsd_model('keggin', array([['-0.0', '-3.7005', '-3.7005', 'O'], ['-1.3091', '-1.3091', '3.1481', 'O'], ['-1.3091', '1.3091', '3.1481', 'O'], ['-3.7005', '3.7005', '0.0', 'O'], ['-1.3091', '3.1481', '-1.3091', 'O'], ['-1.3091', '1.3091', '-3.1481', 'O'], ['-3.7005', '0.0', '-3.7005', 'O'], ['-1.3091', '-1.3091', '-3.1481', 'O'], ['1.3091', '-3.1481', '-1.3091', 'O'], ['3.1481', '-1.3091', '-1.3091', 'O'], ['1.3091', '1.3091', '-3.1481', 'O'], ['-3.7005', '-0.0', '3.7005', 'O'], ['-3.7005', '-3.7005', '-0.0', 'O'], ['-1.3091', '-3.1481', '-1.3091', 'O'], ['-0.0', '-3.7005', '3.7005', 'O'], ['1.3091', '-3.1481', '1.3091', 'O'], ['3.7005', '-0.0', '3.7005', 'O'], ['3.1481', '-1.3091', '1.3091', 'O'], ['3.1481', '1.3091', '1.3091', 'O'], ['3.1481', '1.3091', '-1.3091', 'O'], ['0.0', '3.7005', '3.7005', 'O'], ['1.3091', '3.1481', '1.3091', 'O'], ['-3.1481', '1.3091', '1.3091', 'O'], ['0.0', '3.7005', '-3.7005', 'O'], ['1.3091', '3.1481', '-1.3091', 'O'], ['1.3091', '-1.3091', '-3.1481', 'O'], ['3.7005', '-3.7005', '-0.0', 'O'], ['3.7005', '0.0', '-3.7005', 'O'], ['3.7005', '3.7005', '0.0', 'O'], ['-3.1481', '1.3091', '-1.3091', 'O'], ['-3.1481', '-1.3091', '-1.3091', 'O'], ['-1.3091', '-3.1481', '1.3091', 'O'], ['1.3091', '-1.3091', '3.1481', 'O'], ['1.3091', '1.3091', '3.1481', 'O'], ['-1.3091', '3.1481', '1.3091', 'O'], ['-3.1481', '-1.3091', '1.3091', 'O'], ['2.5118', '-2.5118', '-0.0', 'W'], ['2.5118', '0.0', '-2.5118', 'W'], ['2.5118', '2.5118', '0.0', 'W'], ['-2.5118', '-0.0', '2.5118', 'W'], ['-2.5118', '2.5118', '0.0', 'W'], ['-2.5118', '0.0', '-2.5118', 'W'], ['-2.5118', '-2.5118', '-0.0', 'W'], ['-0.0', '-2.5118', '2.5118', 'W'], ['2.5118', '-0.0', '2.5118', 'W'], ['0.0', '2.5118', '2.5118', 'W'], ['0.0', '2.5118', '-2.5118', 'W'], ['-0.0', '-2.5118', '-2.5118', 'W']]), 'Oh' , pca = {"A1g":array([[0.0, 0.02342, 0.02342, 0.0166, 0.0166, 0.00311, 0.0166, -0.0166, 0.00311, 0.02342, -0.02342, 0.0, 0.0166, 0.00311, 0.0166, 0.0166, -0.0166, -0.00311, 0.02342, -0.0, 0.02342, 0.0166, 0.0166, -0.00311, -0.0166, -0.00311, 0.0166, 0.00311, 0.0166, 0.0166, -0.0166, -0.0166, -0.00311, 0.02342, 0.0, -0.02342, 0.02342, 0.02342, -0.0, 0.0166, -0.00311, 0.0166, 0.0, 0.02342, -0.02342, -0.0166, -0.00311, -0.0166, -0.02342, 0.0, -0.02342, 0.00311, 0.0166, -0.0166, 0.00311, -0.0166, -0.0166, 0.00311, -0.0166, 0.0166, -0.0, -0.02342, -0.02342, -0.0166, 0.00311, -0.0166, -0.00311, -0.0166, -0.0166, -0.0, -0.02342, 0.02342, -0.0166, 0.00311, 0.0166, -0.0166, 0.0166, -0.00311, -0.02342, 0.02342, -0.0, -0.02342, -0.0, 0.02342, -0.02342, -0.02342, 0.0, -0.00311, -0.0166, 0.0166, -0.00311, 0.0166, 0.0166, 0.0166, -0.00311, -0.0166, -0.0166, 0.0166, 0.00311, -0.0166, -0.0166, 0.00311, 0.0166, 0.00311, -0.0166, -0.00311, 0.0166, -0.0166, -0.00201, 0.00201, 0.0, -0.00201, 0.0, 0.00201, -0.00201, -0.00201, 0.0, 0.00201, -0.0, -0.00201, 0.00201, -0.00201, -0.0, 0.00201, -0.0, 0.00201, 0.00201, 0.00201, -0.0, 0.0, 0.00201, -0.00201, -0.00201, 0.0, -0.00201, -0.0, -0.00201, -0.00201, -0.0, -0.00201, 0.00201, 0.0, 0.00201, 0.00201], [-0.0, 0.01536, 0.01536, -0.01091, -0.01091, -0.01037, -0.01091, 0.01091, -0.01037, 0.01536, -0.01536, -0.0, -0.01091, -0.01037, -0.01091, -0.01091, 0.01091, 0.01037, 0.01536, -0.0, 0.01536, -0.01091, -0.01091, 0.01037, 0.01091, 0.01037, -0.01091, -0.01037, -0.01091, -0.01091, 0.01091, 0.01091, 0.01037, 0.01536, 0.0, -0.01536, 0.01536, 0.01536, -0.0, -0.01091, 0.01037, -0.01091, 0.0, 0.01536, -0.01536, 0.01091, 0.01037, 0.01091, -0.01536, 0.0, -0.01536, -0.01037, -0.01091, 0.01091, -0.01037, 0.01091, 0.01091, -0.01037, 0.01091, -0.01091, -0.0, -0.01536, -0.01536, 0.01091, -0.01037, 0.01091, 0.01037, 0.01091, 0.01091, -0.0, -0.01536, 0.01536, 0.01091, -0.01037, -0.01091, 0.01091, -0.01091, 0.01037, -0.01536, 0.01536, 0.0, -0.01536, -0.0, 0.01536, -0.01536, -0.01536, 0.0, 0.01037, 0.01091, -0.01091, 0.01037, -0.01091, -0.01091, -0.01091, 0.01037, 0.01091, 0.01091, -0.01091, -0.01037, 0.01091, 0.01091, -0.01037, -0.01091, -0.01037, 0.01091, 0.01037, -0.01091, 0.01091, -0.01729, 0.01729, 0.0, -0.01729, 0.0, 0.01729, -0.01729, -0.01729, -0.0, 0.01729, -0.0, -0.01729, 0.01729, -0.01729, -0.0, 0.01729, -0.0, 0.01729, 0.01729, 0.01729, 0.0, 0.0, 0.01729, -0.01729, -0.01729, 0.0, -0.01729, -0.0, -0.01729, -0.01729, -0.0, -0.01729, 0.01729, 0.0, 0.01729, 0.01729]]), "A2g":array([[-0.0, -0.05719, 0.05719, -0.00076, 0.00076, -0.0, -0.00076, -0.00076, -0.0, -0.05719, -0.05719, 0.0, 0.00076, -0.0, -0.00076, -0.00076, -0.00076, 0.0, 0.05719, -0.0, -0.05719, -0.00076, 0.00076, 0.0, -0.00076, 0.0, -0.00076, 0.0, -0.00076, 0.00076, 0.00076, -0.00076, 0.0, 0.05719, -0.0, 0.05719, -0.05719, 0.05719, 0.0, 0.00076, 0.0, -0.00076, -0.0, -0.05719, -0.05719, -0.00076, 0.0, 0.00076, -0.05719, 0.0, 0.05719, 0.0, -0.00076, -0.00076, 0.0, 0.00076, -0.00076, 0.0, 0.00076, 0.00076, 0.0, 0.05719, -0.05719, -0.00076, -0.0, 0.00076, -0.0, 0.00076, -0.00076, 0.0, 0.05719, 0.05719, -0.00076, -0.0, -0.00076, 0.00076, 0.00076, 0.0, 0.05719, 0.05719, -0.0, -0.05719, 0.0, -0.05719, 0.05719, -0.05719, -0.0, -0.0, 0.00076, 0.00076, -0.0, -0.00076, 0.00076, 0.00076, 0.0, 0.00076, 0.00076, 0.00076, -0.0, 0.00076, -0.00076, -0.0, 0.00076, -0.0, 0.00076, -0.0, -0.00076, -0.00076, 0.00022, 0.00022, -0.0, -0.00022, -0.0, -0.00022, 0.00022, -0.00022, -0.0, 0.00022, 0.0, 0.00022, -0.00022, -0.00022, 0.0, 0.00022, 0.0, -0.00022, -0.00022, 0.00022, 0.0, -0.0, -0.00022, -0.00022, -0.00022, -0.0, 0.00022, -0.0, 0.00022, -0.00022, 0.0, 0.00022, 0.00022, 0.0, -0.00022, 0.00022], [-0.0, 0.00043, -0.00043, -0.01908, 0.01908, 0.0, -0.01908, -0.01908, 0.0, 0.00043, 0.00043, -0.0, 0.01908, -0.0, -0.01908, -0.01908, -0.01908, -0.0, -0.00043, -0.0, 0.00043, -0.01908, 0.01908, -0.0, -0.01908, 0.0, -0.01908, 0.0, -0.01908, 0.01908, 0.01908, -0.01908, -0.0, -0.00043, 0.0, -0.00043, 0.00043, -0.00043, -0.0, 0.01908, 0.0, -0.01908, 0.0, 0.00043, 0.00043, -0.01908, 0.0, 0.01908, 0.00043, 0.0, -0.00043, 0.0, -0.01908, -0.01908, 0.0, 0.01908, -0.01908, 0.0, 0.01908, 0.01908, -0.0, -0.00043, 0.00043, -0.01908, -0.0, 0.01908, -0.0, 0.01908, -0.01908, -0.0, -0.00043, -0.00043, -0.01908, -0.0, -0.01908, 0.01908, 0.01908, -0.0, -0.00043, -0.00043, 0.0, 0.00043, -0.0, 0.00043, -0.00043, 0.00043, 0.0, -0.0, 0.01908, 0.01908, -0.0, -0.01908, 0.01908, 0.01908, 0.0, 0.01908, 0.01908, 0.01908, 0.0, 0.01908, -0.01908, 0.0, 0.01908, -0.0, 0.01908, -0.0, -0.01908, -0.01908, -0.02034, -0.02034, -0.0, 0.02034, -0.0, 0.02034, -0.02034, 0.02034, 0.0, -0.02034, 0.0, -0.02034, 0.02034, 0.02034, 0.0, -0.02034, 0.0, 0.02034, 0.02034, -0.02034, -0.0, 0.0, 0.02034, 0.02034, 0.02034, -0.0, -0.02034, -0.0, -0.02034, 0.02034, -0.0, -0.02034, -0.02034, 0.0, 0.02034, -0.02034]]), "Ega":array([[-0.04192, 0.00871, 0.00871, -0.02382, -0.02382, -0.01833, 0.00183, -0.01162, 0.00735, -0.00871, 0.00872, -0.0, -0.02382, -0.01833, -0.02382, 0.01163, -0.00183, -0.00735, 0.00871, -0.04192, 0.00872, 0.01036, 0.01036, -0.00363, -0.00183, -0.00735, 0.01162, -0.01833, -0.02382, -0.02382, 0.02382, 0.02382, 0.01833, -0.00871, -0.0, 0.00872, 0.00871, 0.00872, -0.04192, 0.01036, -0.00363, 0.01036, -0.0, -0.00871, 0.00871, 0.02382, 0.01833, 0.02382, -0.00871, 0.04192, -0.00872, 0.00735, 0.00183, -0.01162, 0.00363, -0.01036, -0.01036, 0.00735, -0.01162, 0.00183, 0.04192, -0.00871, -0.00871, -0.01036, 0.00363, -0.01036, 0.01833, 0.02382, 0.02382, 0.0, 0.00871, -0.00871, -0.01163, 0.00735, 0.00183, -0.00183, 0.01162, -0.00735, 0.00871, -0.00872, 0.0, 0.00871, 0.0, -0.00872, -0.00871, -0.00872, 0.04192, -0.00735, -0.00183, 0.01162, -0.00363, 0.01036, 0.01036, 0.01163, -0.00735, -0.00183, -0.01163, 0.00183, 0.00735, -0.01036, -0.01036, 0.00363, 0.00183, 0.00735, -0.01162, -0.00735, 0.01162, -0.00183, -0.00057, 0.00057, 0.0, -0.00057, 0.0, 0.00057, 0.00057, 0.00057, 0.02864, 0.00057, -0.0, -0.00057, 0.00057, -0.00057, -0.0, -0.00057, -0.02864, -0.00057, -0.00057, -0.00057, -0.02864, 0.0, 0.00057, -0.00057, 0.00057, 0.02864, 0.00057, 0.02864, 0.00057, 0.00057, -0.0, -0.00057, 0.00057, -0.02864, -0.00057, -0.00057], [0.04158, 0.02379, 0.02379, -0.0109, -0.01088, -0.00461, -0.00776, -0.00707, 0.00365, -0.02379, 0.02377, 2e-05, -0.0109, -0.00461, -0.01088, 0.00706, 0.00778, -0.00361, 0.02373, 0.04153, 0.02384, 0.0116, 0.01159, 0.00271, 0.00776, -0.00365, 0.00707, -0.00458, -0.01092, -0.01092, 0.0109, 0.01088, 0.00461, -0.02379, 2e-05, 0.02377, 0.02373, 0.02384, 0.04153, 0.0116, 0.00271, 0.01159, -0.0, -0.02378, 0.02378, 0.0109, 0.00461, 0.01088, -0.02373, -0.04153, -0.02384, 0.00369, -0.00775, -0.00708, -0.00268, -0.0116, -0.0116, 0.00369, -0.00708, -0.00775, -0.04158, -0.02379, -0.02379, -0.0116, -0.00271, -0.01159, 0.00458, 0.01092, 0.01092, 0.0, 0.02378, -0.02378, -0.00706, 0.00361, -0.00778, 0.00776, 0.00707, -0.00365, 0.02379, -0.02377, -2e-05, 0.02379, -2e-05, -0.02377, -0.02373, -0.02384, -0.04153, -0.00369, 0.00775, 0.00708, 0.00268, 0.0116, 0.0116, 0.00706, -0.00361, 0.00778, -0.00706, -0.00778, 0.00361, -0.0116, -0.01159, -0.00271, -0.00776, 0.00365, -0.00707, -0.00369, 0.00708, 0.00775, 0.00858, -0.00856, -2e-05, 0.00858, -2e-05, -0.00856, -0.00853, -0.0086, -0.00318, -0.00858, 2e-05, 0.00856, -0.00858, 0.00856, 2e-05, 0.00853, 0.00318, 0.0086, 0.00853, 0.0086, 0.00318, -0.0, -0.00855, 0.00855, -0.00853, -0.00318, -0.0086, -0.00323, -0.00856, -0.00856, 0.0, 0.00855, -0.00855, 0.00323, 0.00856, 0.00856]]), "Egb":array([[0.04212, 0.00855, 0.00855, 0.01158, 0.002, 0.00722, 0.01037, -0.01037, 0.00391, 0.00854, -0.00855, 0.04213, 0.01158, 0.00722, 0.002, -0.02395, 0.02395, 0.01835, -0.00855, -0.0, -0.00854, 0.00201, 0.01158, -0.00722, -0.01037, -0.00391, 0.01037, 0.00391, 0.01037, 0.01037, -0.01158, -0.002, -0.00722, 0.00854, 0.04213, -0.00855, -0.00855, -0.00854, -0.0, 0.00201, -0.00722, 0.01158, -0.0, -0.00855, 0.00855, -0.01158, -0.00722, -0.002, 0.00855, 0.0, 0.00854, 0.00722, 0.01158, -0.00201, -0.01835, 0.02395, 0.02395, 0.00722, -0.00201, 0.01158, -0.04212, -0.00855, -0.00855, -0.00201, 0.00722, -0.01158, -0.00391, -0.01037, -0.01037, 0.0, 0.00855, -0.00855, 0.02395, -0.01835, -0.02395, -0.01037, 0.01037, -0.00391, -0.00854, 0.00855, -0.04213, -0.00854, -0.04213, 0.00855, 0.00855, 0.00854, 0.0, -0.00722, -0.01158, 0.00201, 0.01835, -0.02395, -0.02395, -0.02395, 0.01835, 0.02395, 0.02395, -0.02395, -0.01835, -0.00201, -0.01158, 0.00722, 0.01037, 0.00391, -0.01037, -0.00722, 0.00201, -0.01158, 0.00065, -0.00066, -0.02854, 0.00065, -0.02854, -0.00066, -0.00065, -0.00065, 0.0, -0.00065, 0.02854, 0.00066, -0.00065, 0.00066, 0.02854, 0.00065, -0.0, 0.00065, 0.00065, 0.00065, -0.0, 0.0, 0.00065, -0.00065, -0.00065, 0.0, -0.00065, -0.02853, 0.00065, 0.00065, -0.0, -0.00065, 0.00065, 0.02853, -0.00065, -0.00065], [-0.03566, 0.03086, 0.03086, 0.00175, -0.00818, 0.00063, 0.01196, -0.01194, -0.0033, 0.03089, -0.0309, -0.03564, 0.00175, 0.00063, -0.00818, -0.00549, 0.00551, -0.00201, -0.03086, 3e-05, -0.03091, -0.0082, 0.00174, -0.00064, -0.01196, 0.0033, 0.01194, -0.00326, 0.01194, 0.01194, -0.00175, 0.00818, -0.00063, 0.03089, -0.03564, -0.0309, -0.03086, -0.03091, 3e-05, -0.0082, -0.00064, 0.00174, -0.0, -0.03088, 0.03088, -0.00175, -0.00063, 0.00818, 0.03086, -3e-05, 0.03091, 0.00065, 0.00176, 0.00818, 0.00199, 0.00551, 0.00551, 0.00065, 0.00818, 0.00176, 0.03566, -0.03086, -0.03086, 0.0082, 0.00064, -0.00174, 0.00326, -0.01194, -0.01194, 0.0, 0.03088, -0.03088, 0.00549, 0.00201, -0.00551, -0.01196, 0.01194, 0.0033, -0.03089, 0.0309, 0.03564, -0.03089, 0.03564, 0.0309, 0.03086, 0.03091, -3e-05, -0.00065, -0.00176, -0.00818, -0.00199, -0.00551, -0.00551, -0.00549, -0.00201, 0.00551, 0.00549, -0.00551, 0.00201, 0.0082, -0.00174, 0.00064, 0.01196, -0.0033, -0.01194, -0.00065, -0.00818, -0.00176, -0.00867, 0.0087, 0.00035, -0.00867, 0.00035, 0.0087, 0.00869, 0.00868, 0.0, 0.00867, -0.00035, -0.0087, 0.00867, -0.0087, -0.00035, -0.00869, -0.0, -0.00868, -0.00869, -0.00868, -0.0, -0.0, -0.00869, 0.00869, 0.00869, 0.0, 0.00868, 0.00036, -0.00869, -0.00869, 0.0, 0.00869, -0.00869, -0.00036, 0.00869, 0.00869]]), "Egc":array([[0.0, -0.00843, -0.00843, 0.01058, 0.01058, 0.0039, 0.0116, -0.0019, 0.00735, -0.00843, 0.00844, 0.0, 0.0019, 0.00735, 0.0116, 0.0019, -0.0116, -0.00735, -0.00843, -0.0, -0.00844, -0.02409, -0.02409, 0.01859, 0.02409, 0.01859, -0.02409, 0.00735, 0.00191, 0.0116, -0.01058, -0.01058, -0.0039, 0.00842, -0.04181, -0.00844, 0.00842, 0.00844, 0.04181, 0.0116, -0.00735, 0.0019, -0.04181, 0.00843, -0.00843, -0.0019, -0.00735, -0.0116, 0.00843, 0.0, 0.00844, -0.01859, -0.02409, 0.02409, 0.00735, -0.0116, -0.00191, 0.0039, -0.01059, 0.01059, 0.0, 0.00843, 0.00843, -0.0116, 0.00735, -0.0019, -0.00735, -0.00191, -0.0116, 0.04181, -0.00843, 0.00843, -0.01058, 0.0039, 0.01058, -0.0116, 0.0019, -0.00735, 0.00843, -0.00844, -0.0, -0.00842, 0.04181, 0.00844, -0.00842, -0.00844, -0.04181, 0.01859, 0.02409, -0.02409, -0.00735, 0.0116, 0.00191, 0.01058, -0.0039, -0.01058, -0.0019, 0.0116, 0.00735, 0.02409, 0.02409, -0.01859, -0.02409, -0.01859, 0.02409, -0.0039, 0.01059, -0.01059, -0.0005, 0.0005, -0.0, 0.0005, 0.02856, -0.0005, 0.0005, 0.0005, -0.02856, -0.0005, -0.02856, 0.0005, 0.0005, -0.0005, 0.0, 0.0005, -0.0, 0.0005, -0.0005, -0.0005, 0.02856, -0.02856, -0.0005, 0.0005, -0.0005, 0.0, -0.0005, -0.0, -0.0005, -0.0005, 0.02856, 0.0005, -0.0005, 0.0, 0.0005, 0.0005], [-0.0, 0.02871, 0.02871, -0.01303, -0.01302, 0.00271, -0.00164, -0.00757, -0.00021, 0.02871, -0.02876, 1e-05, 0.0076, -0.00023, -0.00164, 0.0076, 0.00164, 0.00023, 0.02871, -1e-05, 0.02876, 0.00709, 0.00706, 0.0023, -0.00709, 0.0023, 0.00706, -0.00021, 0.00759, -0.00163, 0.01303, 0.01302, -0.00271, -0.02871, -0.03095, 0.02873, -0.02871, -0.02873, 0.03095, -0.00164, 0.00021, 0.00757, -0.03098, -0.02874, 0.02874, -0.0076, 0.00023, 0.00164, -0.02871, 1e-05, -0.02876, -0.00229, 0.00707, -0.00707, -0.00021, 0.00163, -0.00759, 0.00275, 0.01301, -0.01301, -0.0, -0.02871, -0.02871, 0.00164, -0.00021, -0.00757, 0.00021, -0.00759, 0.00163, 0.03098, 0.02874, -0.02874, 0.01303, 0.00271, -0.01302, 0.00164, 0.00757, 0.00021, -0.02871, 0.02876, -1e-05, 0.02871, 0.03095, -0.02873, 0.02871, 0.02873, -0.03095, 0.00229, -0.00707, 0.00707, 0.00021, -0.00163, 0.00759, -0.01303, -0.00271, 0.01302, -0.0076, -0.00164, -0.00023, -0.00709, -0.00706, -0.0023, 0.00709, -0.0023, -0.00706, -0.00275, -0.01301, 0.01301, -0.01004, 0.01007, -1e-05, 0.01007, 0.00804, -0.01006, 0.01007, 0.01006, -0.00804, -0.01007, -0.00804, 0.01006, 0.01004, -0.01007, 1e-05, 0.01004, -1e-05, 0.01007, -0.01007, -0.01006, 0.00804, -0.00804, -0.01004, 0.01004, -0.01004, 1e-05, -0.01007, -0.0, -0.01006, -0.01006, 0.00804, 0.01004, -0.01004, 0.0, 0.01006, 0.01006]]), "Egd":array([[0.0, -0.00824, -0.00824, 0.00187, 0.01158, 0.00722, -0.02396, 0.02396, -0.01825, 0.00823, -0.00825, -0.04242, 0.01052, 0.00381, 0.01052, 0.01052, -0.01052, -0.00381, 0.00823, 0.04242, 0.00825, 0.01157, 0.00187, -0.00722, -0.01157, -0.00722, 0.00187, 0.00722, 0.01158, 0.00187, -0.00187, -0.01158, -0.00722, -0.00824, -1e-05, 0.00824, -0.00824, -0.00824, 1e-05, -0.02396, 0.01825, -0.02396, 0.04242, 0.00824, -0.00824, -0.01052, -0.00381, -0.01052, -0.00823, -0.04242, -0.00825, 0.0038, 0.01052, -0.01052, 0.00722, -0.00187, -0.01158, -0.01825, 0.02396, -0.02396, 0.0, 0.00824, 0.00824, 0.02396, -0.01825, 0.02396, -0.00722, -0.01158, -0.00187, -0.04242, -0.00824, 0.00824, -0.00187, 0.00722, 0.01158, 0.02396, -0.02396, 0.01825, -0.00823, 0.00825, 0.04242, 0.00824, 1e-05, -0.00824, 0.00824, 0.00824, -1e-05, -0.0038, -0.01052, 0.01052, -0.00722, 0.00187, 0.01158, 0.00187, -0.00722, -0.01158, -0.01052, 0.01052, 0.00381, -0.01157, -0.00187, 0.00722, 0.01157, 0.00722, -0.00187, 0.01825, -0.02396, 0.02396, 0.00051, -0.00052, 0.0289, -0.00052, 0.0, 0.00052, -0.00052, -0.00052, -0.0, 0.00052, -0.0, -0.00052, -0.00051, 0.00052, -0.0289, -0.00051, 0.0289, -0.00052, 0.00052, 0.00052, 0.0, 0.0289, -0.00052, 0.00052, 0.00051, -0.0289, 0.00052, -0.0, -0.00052, -0.00052, -0.0289, 0.00052, -0.00052, 0.0, 0.00052, 0.00052], [0.0, 0.02517, 0.02517, 0.01067, -0.00578, -0.00097, 0.00731, -0.00732, -0.00142, -0.02511, 0.02517, -0.02975, -0.01221, 0.00342, -0.01224, -0.01221, 0.01224, -0.00342, -0.02511, 0.02975, -0.02517, -0.00574, 0.01066, 0.00096, 0.00574, 0.00096, 0.01066, -0.00103, -0.00574, 0.01067, -0.01067, 0.00578, 0.00097, 0.02519, 1e-05, -0.02509, 0.02519, 0.02509, -1e-05, 0.00731, 0.00142, 0.00732, 0.02971, -0.02517, 0.02517, 0.01221, -0.00342, 0.01224, 0.02511, -0.02975, 0.02517, 0.00339, -0.01221, 0.01221, -0.00103, -0.01067, 0.00574, -0.00147, -0.0073, 0.0073, -0.0, -0.02517, -0.02517, -0.00731, -0.00142, -0.00732, 0.00103, 0.00574, -0.01067, -0.02971, 0.02517, -0.02517, -0.01067, -0.00097, -0.00578, -0.00731, 0.00732, 0.00142, 0.02511, -0.02517, 0.02975, -0.02519, -1e-05, 0.02509, -0.02519, -0.02509, 1e-05, -0.00339, 0.01221, -0.01221, 0.00103, 0.01067, -0.00574, 0.01067, 0.00097, 0.00578, 0.01221, -0.01224, 0.00342, 0.00574, -0.01066, -0.00096, -0.00574, -0.00096, -0.01066, 0.00147, 0.0073, -0.0073, 0.01079, -0.01081, 0.00765, -0.01082, 0.0, 0.01079, -0.01082, -0.01079, -0.0, 0.01082, -0.0, -0.01079, -0.01079, 0.01081, -0.00765, -0.01079, 0.00765, -0.01081, 0.01082, 0.01079, 0.0, 0.00765, -0.01081, 0.01081, 0.01079, -0.00765, 0.01081, -0.0, -0.01081, -0.01081, -0.00765, 0.01081, -0.01081, 0.0, 0.01081, 0.01081]]), "T1gx":array([[-0.0, -0.00877, -0.00877, -4e-05, -0.00277, 0.00539, -4e-05, 0.00277, 0.00539, 0.06404, 0.05527, -0.0, -4e-05, 0.00539, -0.00277, -4e-05, 0.00277, -0.00539, 0.06404, 0.0, -0.05527, -4e-05, -0.00277, -0.00539, 4e-05, -0.00539, -0.00277, -0.01078, 0.00281, 0.00281, 4e-05, 0.00277, -0.00539, 0.06404, -0.0, 0.05527, 0.06404, -0.05527, -0.0, -4e-05, -0.00539, -0.00277, -0.0, -0.00877, 0.00877, 4e-05, -0.00539, 0.00277, -0.06404, -0.0, 0.05527, -0.01078, 0.00281, -0.00281, -0.01078, -0.00281, -0.00281, -0.01078, -0.00281, 0.00281, 0.0, 0.00877, 0.00877, 4e-05, 0.00539, 0.00277, 0.01078, -0.00281, -0.00281, 0.0, 0.00877, -0.00877, 4e-05, 0.00539, -0.00277, 4e-05, -0.00277, -0.00539, -0.06404, -0.05527, 0.0, -0.06404, 0.0, -0.05527, -0.06404, 0.05527, 0.0, 0.01078, -0.00281, 0.00281, 0.01078, 0.00281, 0.00281, -4e-05, -0.00539, 0.00277, 4e-05, -0.00277, 0.00539, 4e-05, 0.00277, 0.00539, -4e-05, 0.00539, 0.00277, 0.01078, 0.00281, -0.00281, -0.00729, -0.00425, -0.0, -0.00729, -0.0, -0.00425, -0.00729, 0.00425, -0.0, 0.00729, 0.0, 0.00425, 0.00729, 0.00425, 0.0, 0.00729, -0.0, -0.00425, 0.00729, -0.00425, 0.0, -0.0, -0.00304, 0.00304, -0.00729, 0.0, 0.00425, 0.0, 0.00304, 0.00304, 0.0, 0.00304, -0.00304, -0.0, -0.00304, -0.00304], [-0.0, 0.04712, 0.04712, -0.00752, 0.0044, -0.00704, -0.00752, -0.0044, -0.00704, -0.01555, 0.03157, -0.0, -0.00752, -0.00704, 0.0044, -0.00752, -0.0044, 0.00704, -0.01555, 0.0, -0.03157, -0.00752, 0.0044, 0.00704, 0.00752, 0.00704, 0.0044, 0.01407, 0.00312, 0.00312, 0.00752, -0.0044, 0.00704, -0.01555, 0.0, 0.03157, -0.01555, -0.03157, -0.0, -0.00752, 0.00704, 0.0044, 0.0, 0.04712, -0.04712, 0.00752, 0.00704, -0.0044, 0.01555, -0.0, 0.03157, 0.01407, 0.00312, -0.00312, 0.01407, -0.00312, -0.00312, 0.01407, -0.00312, 0.00312, -0.0, -0.04712, -0.04712, 0.00752, -0.00704, -0.0044, -0.01407, -0.00312, -0.00312, -0.0, -0.04712, 0.04712, 0.00752, -0.00704, 0.0044, 0.00752, 0.0044, 0.00704, 0.01555, -0.03157, 0.0, 0.01555, -0.0, -0.03157, 0.01555, 0.03157, 0.0, -0.01407, -0.00312, 0.00312, -0.01407, 0.00312, 0.00312, -0.00752, 0.00704, -0.0044, 0.00752, 0.0044, -0.00704, 0.00752, -0.0044, -0.00704, -0.00752, -0.00704, -0.0044, -0.01407, 0.00312, -0.00312, 0.01012, -0.0025, 0.0, 0.01012, -0.0, -0.0025, 0.01012, 0.0025, -0.0, -0.01012, 0.0, 0.0025, -0.01012, 0.0025, -0.0, -0.01012, -0.0, -0.0025, -0.01012, -0.0025, 0.0, 0.0, 0.01262, -0.01262, 0.01012, 0.0, 0.0025, -0.0, -0.01262, -0.01262, -0.0, -0.01262, 0.01262, 0.0, 0.01262, 0.01262]]), "T1gy":array([[-0.0, -0.0606, 0.0517, 0.00199, -0.00274, -0.00496, 0.00199, 0.00274, -0.00496, 0.0517, 0.0606, -0.0, 0.00075, 0.00991, 0.00075, 0.00199, 0.00274, 0.00496, 0.0089, -0.0, 0.0089, 0.00199, -0.00274, 0.00496, -0.00075, -0.00991, 0.00075, -0.00496, -0.00274, 0.00199, -0.00199, 0.00274, 0.00496, 0.0089, -0.0, -0.0089, 0.0517, -0.0606, -0.0, 0.00075, -0.00991, 0.00075, -0.0, -0.0606, -0.0517, -0.00075, -0.00991, -0.00075, -0.0089, 0.0, -0.0089, -0.00496, -0.00274, -0.00199, -0.00496, 0.00274, -0.00199, -0.00496, 0.00274, 0.00199, -0.0, 0.0606, -0.0517, -0.00075, 0.00991, -0.00075, 0.00496, 0.00274, -0.00199, 0.0, 0.0606, 0.0517, -0.00075, 0.00991, 0.00075, -0.00199, -0.00274, 0.00496, -0.0517, -0.0606, 0.0, -0.0089, 0.0, 0.0089, -0.0517, 0.0606, 0.0, 0.00496, 0.00274, 0.00199, 0.00496, -0.00274, 0.00199, 0.00075, -0.00991, -0.00075, -0.00199, -0.00274, -0.00496, -0.00199, 0.00274, -0.00496, 0.00075, 0.00991, -0.00075, 0.00496, -0.00274, -0.00199, -0.00779, -0.01186, 0.0, -0.00407, 0.0, 0.00407, -0.00779, 0.01186, 0.0, 0.00407, -0.0, -0.00407, 0.00779, 0.01186, -0.0, 0.00407, -0.0, 0.00407, 0.00779, -0.01186, -0.0, -0.0, -0.01186, -0.00779, -0.00407, 0.0, -0.00407, 0.0, 0.01186, -0.00779, 0.0, 0.01186, 0.00779, -0.0, -0.01186, 0.00779], [-0.0, -0.01427, -0.03341, 0.00166, -0.00432, -0.00625, 0.00166, 0.00432, -0.00625, -0.03341, 0.01427, 0.0, 0.00266, 0.0125, 0.00266, 0.00166, 0.00432, 0.00625, 0.04767, -0.0, 0.04767, 0.00166, -0.00432, 0.00625, -0.00266, -0.0125, 0.00266, -0.00625, -0.00432, 0.00166, -0.00166, 0.00432, 0.00625, 0.04767, -0.0, -0.04767, -0.03341, -0.01427, 0.0, 0.00266, -0.0125, 0.00266, -0.0, -0.01427, 0.03341, -0.00266, -0.0125, -0.00266, -0.04767, 0.0, -0.04767, -0.00625, -0.00432, -0.00166, -0.00625, 0.00432, -0.00166, -0.00625, 0.00432, 0.00166, 0.0, 0.01427, 0.03341, -0.00266, 0.0125, -0.00266, 0.00625, 0.00432, -0.00166, 0.0, 0.01427, -0.03341, -0.00266, 0.0125, 0.00266, -0.00166, -0.00432, 0.00625, 0.03341, -0.01427, -0.0, -0.04767, 0.0, 0.04767, 0.03341, 0.01427, -0.0, 0.00625, 0.00432, 0.00166, 0.00625, -0.00432, 0.00166, 0.00266, -0.0125, -0.00266, -0.00166, -0.00432, -0.00625, -0.00166, 0.00432, -0.00625, 0.00266, 0.0125, -0.00266, 0.00625, -0.00432, -0.00166, 0.00251, -0.01427, 0.0, -0.01678, 0.0, 0.01678, 0.00251, 0.01427, 0.0, 0.01678, -0.0, -0.01678, -0.00251, 0.01427, -0.0, 0.01678, -0.0, 0.01678, -0.00251, -0.01427, -0.0, 0.0, -0.01427, 0.00251, -0.01678, 0.0, -0.01678, 0.0, 0.01427, 0.00251, -0.0, 0.01427, -0.00251, -0.0, -0.01427, -0.00251]]), "T1gz":array([[-0.0, -0.06704, 0.07157, 0.00206, 0.00206, -0.00513, 0.00206, -0.00206, -0.00513, -0.00453, 0.00453, 0.0, -0.0028, 0.00256, 0.00075, 0.00206, -0.00206, 0.00513, -0.06704, 0.0, 0.07157, 0.00206, 0.00206, 0.00513, 0.0028, -0.00256, 0.00075, 0.00256, -0.0028, 0.00075, -0.00206, -0.00206, 0.00513, -0.06704, 0.0, -0.07157, -0.00453, -0.00453, -0.0, -0.0028, -0.00256, 0.00075, -0.0, -0.06704, -0.07157, 0.0028, -0.00256, -0.00075, 0.06704, -0.0, -0.07157, 0.00256, -0.0028, -0.00075, 0.00256, 0.0028, -0.00075, 0.00256, 0.0028, 0.00075, -0.0, 0.06704, -0.07157, 0.0028, 0.00256, -0.00075, -0.00256, 0.0028, -0.00075, 0.0, 0.06704, 0.07157, 0.0028, 0.00256, 0.00075, -0.00206, 0.00206, 0.00513, 0.00453, -0.00453, -0.0, 0.06704, -0.0, 0.07157, 0.00453, 0.00453, 0.0, -0.00256, 0.0028, 0.00075, -0.00256, -0.0028, 0.00075, -0.0028, -0.00256, -0.00075, -0.00206, 0.00206, -0.00513, -0.00206, -0.00206, -0.00513, -0.0028, 0.00256, -0.00075, -0.00256, -0.0028, -0.00075, 0.0036, -0.0036, 0.0, 0.00322, 0.0, 0.00682, 0.0036, 0.0036, 0.0, -0.00322, -0.0, -0.00682, -0.0036, 0.0036, -0.0, -0.00322, 0.0, 0.00682, -0.0036, -0.0036, -0.0, -0.0, -0.00322, -0.00682, 0.00322, -0.0, -0.00682, -0.0, 0.00322, -0.00682, 0.0, 0.00322, 0.00682, 0.0, -0.00322, 0.00682], [-0.0, -0.03502, -0.02707, 0.00015, 0.00015, -0.00142, 0.00015, -0.00015, -0.00142, 0.06209, -0.06209, -0.0, 0.00276, 0.00071, -0.00291, 0.00015, -0.00015, 0.00142, -0.03502, 0.0, -0.02707, 0.00015, 0.00015, 0.00142, -0.00276, -0.00071, -0.00291, 0.00071, 0.00276, -0.00291, -0.00015, -0.00015, 0.00142, -0.03502, -0.0, 0.02707, 0.06209, 0.06209, 0.0, 0.00276, -0.00071, -0.00291, -0.0, -0.03502, 0.02707, -0.00276, -0.00071, 0.00291, 0.03502, -0.0, 0.02707, 0.00071, 0.00276, 0.00291, 0.00071, -0.00276, 0.00291, 0.00071, -0.00276, -0.00291, 0.0, 0.03502, 0.02707, -0.00276, 0.00071, 0.00291, -0.00071, -0.00276, 0.00291, 0.0, 0.03502, -0.02707, -0.00276, 0.00071, -0.00291, -0.00015, 0.00015, 0.00142, -0.06209, 0.06209, 0.0, 0.03502, 0.0, -0.02707, -0.06209, -0.06209, -0.0, -0.00071, -0.00276, -0.00291, -0.00071, 0.00276, -0.00291, 0.00276, -0.00071, 0.00291, -0.00015, 0.00015, -0.00142, -0.00015, -0.00015, -0.00142, 0.00276, 0.00071, 0.00291, -0.00071, 0.00276, 0.00291, -0.0191, 0.0191, 0.0, 0.00779, 0.0, -0.01132, -0.0191, -0.0191, -0.0, -0.00779, -0.0, 0.01132, 0.0191, -0.0191, -0.0, -0.00779, 0.0, -0.01132, 0.0191, 0.0191, 0.0, -0.0, -0.00779, 0.01132, 0.00779, -0.0, 0.01132, 0.0, 0.00779, 0.01132, 0.0, 0.00779, -0.01132, -0.0, -0.00779, -0.01132]]), "T2gx1":array([[-0.0, 0.10244, -0.10244, 0.00044, 0.0121, 0.00331, -0.00044, 0.0121, -0.00331, -0.0, 0.0, -0.0121, -0.00044, -0.00331, -0.0121, 0.00044, -0.0121, -0.00331, 0.0, -0.0121, 0.0, -0.00044, -0.0121, 0.00331, -0.00044, -0.00331, 0.0121, 0.0, -0.00236, 0.00236, -0.00044, -0.0121, -0.00331, -0.0, 0.0121, -0.0, -0.0, -0.0, 0.0121, 0.00044, -0.00331, 0.0121, -0.0, -0.10244, -0.10244, 0.00044, 0.00331, 0.0121, -0.0, 0.0121, -0.0, 0.0, 0.00236, 0.00236, 0.0, 0.00236, -0.00236, 0.0, -0.00236, -0.00236, -0.0, -0.10244, 0.10244, -0.00044, 0.00331, -0.0121, -0.0, 0.00236, -0.00236, 0.0, 0.10244, 0.10244, 0.00044, -0.00331, -0.0121, 0.00044, -0.0121, 0.00331, 0.0, -0.0, 0.0121, 0.0, -0.0121, 0.0, 0.0, 0.0, -0.0121, -0.0, -0.00236, -0.00236, -0.0, -0.00236, 0.00236, -0.00044, 0.00331, 0.0121, -0.00044, 0.0121, 0.00331, 0.00044, 0.0121, -0.00331, 0.00044, 0.00331, -0.0121, -0.0, 0.00236, 0.00236, 0.0, -0.0, 0.00698, -0.0, -0.00698, -0.0, 0.0, 0.0, -0.00698, 0.0, 0.00698, 0.0, -0.0, 0.0, -0.00698, 0.0, -0.00698, 0.0, -0.0, -0.0, 0.00698, 0.0, 0.00713, 0.00713, -0.0, 0.00698, -0.0, 0.0, 0.00713, -0.00713, -0.0, -0.00713, -0.00713, -0.0, -0.00713, 0.00713], [0.0, 0.01125, -0.01125, -0.01146, -0.01288, -0.00438, 0.01146, -0.01288, 0.00438, -0.0, -0.0, 0.06913, 0.01146, 0.00438, 0.01288, -0.01146, 0.01288, 0.00438, -0.0, 0.06913, 0.0, 0.01146, 0.01288, -0.00438, 0.01146, 0.00438, -0.01288, -0.0, -0.00333, 0.00333, 0.01146, 0.01288, 0.00438, -0.0, -0.06913, 0.0, -0.0, 0.0, -0.06913, -0.01146, 0.00438, -0.01288, -0.0, -0.01125, -0.01125, -0.01146, -0.00438, -0.01288, 0.0, -0.06913, -0.0, -0.0, 0.00333, 0.00333, -0.0, 0.00333, -0.00333, -0.0, -0.00333, -0.00333, -0.0, -0.01125, 0.01125, 0.01146, -0.00438, 0.01288, 0.0, 0.00333, -0.00333, 0.0, 0.01125, 0.01125, -0.01146, 0.00438, 0.01288, -0.01146, 0.01288, -0.00438, 0.0, 0.0, -0.06913, 0.0, 0.06913, -0.0, 0.0, -0.0, 0.06913, 0.0, -0.00333, -0.00333, 0.0, -0.00333, 0.00333, 0.01146, -0.00438, -0.01288, 0.01146, -0.01288, -0.00438, -0.01146, -0.01288, 0.00438, -0.01146, -0.00438, 0.01288, 0.0, 0.00333, 0.00333, 0.0, -0.0, 0.00561, -0.0, -0.00561, 0.0, 0.0, 0.0, -0.00561, 0.0, 0.00561, -0.0, -0.0, 0.0, -0.00561, -0.0, -0.00561, -0.0, -0.0, -0.0, 0.00561, -0.0, -0.00286, -0.00286, 0.0, 0.00561, 0.0, 0.0, -0.00286, 0.00286, 0.0, 0.00286, 0.00286, -0.0, 0.00286, -0.00286]]), "T2gy1":array([[-0.0121, 0.0, -0.0, 0.01232, 0.00065, 0.00331, 0.01232, -0.00065, 0.00331, -0.0, 0.0, 0.0121, -0.00214, -0.0, 0.00214, -0.01232, 0.00065, 0.00331, 0.10211, -0.0, -0.10211, -0.01232, -0.00065, 0.00331, -0.00214, 0.0, -0.00214, -0.00331, -0.00065, -0.01232, -0.01232, -0.00065, -0.00331, -0.10211, 0.0, -0.10211, -0.0, -0.0, 0.0121, -0.00214, 0.0, 0.00214, 0.0121, -0.0, 0.0, 0.00214, 0.0, -0.00214, -0.10211, 0.0, 0.10211, 0.00331, 0.00065, -0.01232, 0.00331, -0.00065, -0.01232, -0.00331, 0.00065, -0.01232, 0.0121, -0.0, 0.0, 0.00214, -0.0, -0.00214, 0.00331, 0.00065, 0.01232, -0.0121, 0.0, -0.0, -0.00214, -0.0, -0.00214, -0.01232, 0.00065, -0.00331, 0.0, -0.0, -0.0121, 0.10211, -0.0, 0.10211, 0.0, 0.0, -0.0121, -0.00331, -0.00065, 0.01232, -0.00331, 0.00065, 0.01232, 0.00214, 0.0, 0.00214, 0.01232, -0.00065, -0.00331, 0.01232, 0.00065, -0.00331, 0.00214, -0.0, 0.00214, 0.00331, -0.00065, 0.01232, -0.0, 0.0, -0.00687, -0.00745, 0.0, -0.00745, 0.0, -0.0, -0.00687, 0.00745, -0.0, 0.00745, 0.0, -0.0, 0.00687, -0.00745, 0.0, 0.00745, -0.0, 0.0, 0.00687, 0.00687, 0.0, 0.0, 0.00745, -0.0, -0.00745, 0.00687, -0.0, 0.0, -0.00687, -0.0, -0.0, -0.00687, 0.0, -0.0], [0.06956, -0.0, 0.0, -0.01118, -0.01134, -0.0064, -0.01118, 0.01134, -0.0064, -0.0, 0.0, -0.06956, -0.00364, 0.0, 0.00364, 0.01118, -0.01134, -0.0064, 0.01123, -0.0, -0.01123, 0.01118, 0.01134, -0.0064, -0.00364, -0.0, -0.00364, 0.0064, 0.01134, 0.01118, 0.01118, 0.01134, 0.0064, -0.01123, 0.0, -0.01123, 0.0, 0.0, -0.06956, -0.00364, -0.0, 0.00364, -0.06956, 0.0, -0.0, 0.00364, -0.0, -0.00364, -0.01123, 0.0, 0.01123, -0.0064, -0.01134, 0.01118, -0.0064, 0.01134, 0.01118, 0.0064, -0.01134, 0.01118, -0.06956, 0.0, -0.0, 0.00364, 0.0, -0.00364, -0.0064, -0.01134, -0.01118, 0.06956, -0.0, 0.0, -0.00364, 0.0, -0.00364, 0.01118, -0.01134, 0.0064, 0.0, -0.0, 0.06956, 0.01123, -0.0, 0.01123, -0.0, -0.0, 0.06956, 0.0064, 0.01134, -0.01118, 0.0064, -0.01134, -0.01118, 0.00364, -0.0, 0.00364, -0.01118, 0.01134, 0.0064, -0.01118, -0.01134, 0.0064, 0.00364, 0.0, 0.00364, -0.0064, 0.01134, -0.01118, -0.0, 0.0, -0.00554, 0.00353, -0.0, 0.00353, 0.0, -0.0, -0.00554, -0.00353, 0.0, -0.00353, 0.0, -0.0, 0.00554, 0.00353, -0.0, -0.00353, -0.0, 0.0, 0.00554, 0.00554, 0.0, 0.0, -0.00353, 0.0, 0.00353, 0.00554, -0.0, 0.0, -0.00554, -0.0, -0.0, -0.00554, 0.0, -0.0]]), "T2gz1":array([[-0.01152, 0.0, -0.0, -0.00217, 0.00217, -0.0, 0.00217, 0.00217, -0.0, -0.10286, -0.10286, -0.0, 0.01253, 0.00316, 0.00069, 0.00217, 0.00217, 0.0, 0.0, 0.01152, 0.0, -0.00217, 0.00217, 0.0, -0.01253, -0.00316, 0.00069, -0.00316, -0.01253, -0.00069, 0.00217, -0.00217, 0.0, 0.0, 0.01152, 0.0, 0.10286, -0.10286, 0.0, -0.01253, 0.00316, -0.00069, -0.01152, 0.0, -0.0, -0.01253, -0.00316, -0.00069, -0.0, -0.01152, -0.0, -0.00316, -0.01253, 0.00069, 0.00316, -0.01253, -0.00069, 0.00316, -0.01253, 0.00069, 0.01152, -0.0, 0.0, 0.01253, -0.00316, 0.00069, 0.00316, 0.01253, 0.00069, 0.01152, -0.0, 0.0, 0.01253, -0.00316, -0.00069, -0.00217, -0.00217, 0.0, 0.10286, 0.10286, 0.0, -0.0, -0.01152, -0.0, -0.10286, 0.10286, -0.0, 0.00316, 0.01253, -0.00069, -0.00316, 0.01253, 0.00069, -0.01253, 0.00316, 0.00069, -0.00217, -0.00217, -0.0, 0.00217, -0.00217, -0.0, 0.01253, 0.00316, -0.00069, -0.00316, 0.01253, -0.00069, -0.00738, -0.00738, -0.0, 0.0, -0.00658, 0.0, 0.00738, -0.00738, 0.0, -0.0, 0.00658, -0.0, 0.00738, 0.00738, 0.0, -0.0, 0.00658, 0.0, -0.00738, 0.00738, -0.0, -0.00658, -0.0, -0.0, 0.0, -0.00658, -0.0, 0.00658, 0.0, -0.0, 0.00658, 0.0, 0.0, -0.00658, -0.0, 0.0], [0.07205, -0.0, 0.0, -0.00337, 0.00337, 0.0, 0.00337, 0.00337, 0.0, -0.01069, -0.01069, 0.0, -0.00956, -0.00619, -0.01064, 0.00337, 0.00337, -0.0, -0.0, -0.07205, 0.0, -0.00337, 0.00337, -0.0, 0.00956, 0.00619, -0.01064, 0.00619, 0.00956, 0.01064, 0.00337, -0.00337, -0.0, -0.0, -0.07205, 0.0, 0.01069, -0.01069, -0.0, 0.00956, -0.00619, 0.01064, 0.07205, -0.0, 0.0, 0.00956, 0.00619, 0.01064, 0.0, 0.07205, -0.0, 0.00619, 0.00956, -0.01064, -0.00619, 0.00956, 0.01064, -0.00619, 0.00956, -0.01064, -0.07205, 0.0, -0.0, -0.00956, 0.00619, -0.01064, -0.00619, -0.00956, -0.01064, -0.07205, 0.0, -0.0, -0.00956, 0.00619, 0.01064, -0.00337, -0.00337, -0.0, 0.01069, 0.01069, -0.0, 0.0, 0.07205, -0.0, -0.01069, 0.01069, 0.0, -0.00619, -0.00956, 0.01064, 0.00619, -0.00956, -0.01064, 0.00956, -0.00619, -0.01064, -0.00337, -0.00337, 0.0, 0.00337, -0.00337, 0.0, -0.00956, -0.00619, 0.01064, 0.00619, -0.00956, 0.01064, 0.00445, 0.00445, -0.0, -0.0, -0.00632, -0.0, -0.00445, 0.00445, 0.0, 0.0, 0.00632, 0.0, -0.00445, -0.00445, 0.0, 0.0, 0.00632, 0.0, 0.00445, -0.00445, -0.0, -0.00632, 0.0, -0.0, -0.0, -0.00632, -0.0, 0.00632, -0.0, 0.0, 0.00632, -0.0, 0.0, -0.00632, 0.0, -0.0]]), "T2gx2":array([[0.0, 0.01301, 0.01301, -0.00943, -0.01682, -0.00856, 0.00943, -0.01682, 0.00856, 0.0, -0.0, 0.03243, -0.00943, -0.00856, -0.01682, -0.00943, 0.01682, 0.00856, -0.0, -0.03243, -0.0, 0.00943, 0.01682, -0.00856, -0.00943, -0.00856, 0.01682, -0.01118, -0.01042, -0.01042, 0.00943, 0.01682, 0.00856, 0.0, 0.03243, 0.0, 0.0, -0.0, -0.03243, 0.00943, -0.00856, 0.01682, -0.0, -0.01301, 0.01301, 0.00943, 0.00856, 0.01682, 0.0, 0.03243, 0.0, 0.01118, 0.01042, -0.01042, -0.01118, 0.01042, 0.01042, 0.01118, -0.01042, 0.01042, -0.0, -0.01301, -0.01301, -0.00943, 0.00856, -0.01682, 0.01118, 0.01042, 0.01042, 0.0, 0.01301, -0.01301, 0.00943, -0.00856, -0.01682, -0.00943, 0.01682, -0.00856, -0.0, 0.0, -0.03243, -0.0, -0.03243, -0.0, -0.0, 0.0, 0.03243, -0.01118, -0.01042, 0.01042, 0.01118, -0.01042, -0.01042, -0.00943, 0.00856, 0.01682, 0.00943, -0.01682, -0.00856, -0.00943, -0.01682, 0.00856, 0.00943, 0.00856, -0.01682, -0.01118, 0.01042, -0.01042, 0.0, 0.0, -0.02194, 0.0, -0.02194, -0.0, 0.0, -0.0, 0.02194, -0.0, 0.02194, 0.0, -0.0, -0.0, 0.02194, -0.0, -0.02194, -0.0, -0.0, 0.0, -0.02194, 0.0, 0.00107, -0.00107, 0.0, 0.02194, 0.0, 0.0, 0.00107, 0.00107, -0.0, -0.00107, 0.00107, -0.0, -0.00107, -0.00107], [0.0, 0.04349, 0.04349, -0.00299, -0.01777, 0.00147, 0.00299, -0.01777, -0.00147, -0.0, 0.0, -0.02881, -0.00299, 0.00147, -0.01777, -0.00299, 0.01777, -0.00147, 0.0, 0.02881, 0.0, 0.00299, 0.01777, 0.00147, -0.00299, 0.00147, 0.01777, -0.00407, 0.00631, 0.00631, 0.00299, 0.01777, -0.00147, -0.0, -0.02881, -0.0, 0.0, -0.0, 0.02881, 0.00299, 0.00147, 0.01777, -0.0, -0.04349, 0.04349, 0.00299, -0.00147, 0.01777, -0.0, -0.02881, -0.0, 0.00407, -0.00631, 0.00631, -0.00407, -0.00631, -0.00631, 0.00407, 0.00631, -0.00631, -0.0, -0.04349, -0.04349, -0.00299, -0.00147, -0.01777, 0.00407, -0.00631, -0.00631, 0.0, 0.04349, -0.04349, 0.00299, 0.00147, -0.01777, -0.00299, 0.01777, 0.00147, 0.0, -0.0, 0.02881, 0.0, 0.02881, 0.0, -0.0, 0.0, -0.02881, -0.00407, 0.00631, -0.00631, 0.00407, 0.00631, 0.00631, -0.00299, -0.00147, 0.01777, 0.00299, -0.01777, 0.00147, -0.00299, -0.01777, -0.00147, 0.00299, -0.00147, -0.01777, -0.00407, -0.00631, 0.00631, 0.0, -0.0, 0.0071, 0.0, 0.0071, 0.0, 0.0, 0.0, -0.0071, -0.0, -0.0071, -0.0, -0.0, 0.0, -0.0071, -0.0, 0.0071, -0.0, -0.0, -0.0, 0.0071, -0.0, -0.01749, 0.01749, 0.0, -0.0071, 0.0, -0.0, -0.01749, -0.01749, 0.0, 0.01749, -0.01749, 0.0, 0.01749, 0.01749]]), "T2gy2":array([[-0.03222, 0.0, -0.0, -0.01704, -0.00955, -0.00857, -0.01704, 0.00955, -0.00857, -0.0, 0.0, -0.03222, -0.01044, -0.01112, -0.01044, 0.01704, -0.00955, -0.00857, 0.01286, -0.0, 0.01286, 0.01704, 0.00955, -0.00857, -0.01044, -0.01112, 0.01044, -0.00857, -0.00955, -0.01704, 0.01704, 0.00955, 0.00857, -0.01286, 0.0, 0.01286, -0.0, 0.0, -0.03222, -0.01044, 0.01112, -0.01044, 0.03222, -0.0, -0.0, 0.01044, 0.01112, 0.01044, -0.01286, 0.0, -0.01286, 0.00857, 0.00955, -0.01704, 0.00857, -0.00955, -0.01704, -0.00857, 0.00955, -0.01704, 0.03222, -0.0, 0.0, 0.01044, -0.01112, 0.01044, 0.00857, 0.00955, 0.01704, -0.03222, 0.0, 0.0, -0.01044, 0.01112, 0.01044, 0.01704, -0.00955, 0.00857, 0.0, -0.0, 0.03222, 0.01286, -0.0, -0.01286, 0.0, -0.0, 0.03222, -0.00857, -0.00955, 0.01704, -0.00857, 0.00955, 0.01704, 0.01044, -0.01112, -0.01044, -0.01704, 0.00955, 0.00857, -0.01704, -0.00955, 0.00857, 0.01044, 0.01112, -0.01044, 0.00857, -0.00955, 0.01704, 0.0, 0.0, 0.02205, -0.00069, 0.0, 0.00069, 0.0, 0.0, 0.02205, 0.00069, -0.0, -0.00069, -0.0, -0.0, -0.02205, -0.00069, 0.0, -0.00069, -0.0, -0.0, -0.02205, 0.02205, -0.0, 0.0, 0.00069, -0.0, 0.00069, 0.02205, 0.0, -0.0, -0.02205, 0.0, -0.0, -0.02205, -0.0, 0.0], [-0.03621, 0.0, -0.0, 0.01204, 0.00442, -0.00042, 0.01204, -0.00442, -0.00042, -0.0, 0.0, -0.03621, -0.00367, 0.00348, -0.00367, -0.01204, 0.00442, -0.00042, -0.05963, 0.0, -0.05963, -0.01204, -0.00442, -0.00042, -0.00367, 0.00348, 0.00367, -0.00042, 0.00442, 0.01204, -0.01204, -0.00442, 0.00042, 0.05963, -0.0, -0.05963, -0.0, 0.0, -0.03621, -0.00367, -0.00348, -0.00367, 0.03621, -0.0, -0.0, 0.00367, -0.00348, 0.00367, 0.05963, -0.0, 0.05963, 0.00042, -0.00442, 0.01204, 0.00042, 0.00442, 0.01204, -0.00042, -0.00442, 0.01204, 0.03621, -0.0, -0.0, 0.00367, 0.00348, 0.00367, 0.00042, -0.00442, -0.01204, -0.03621, 0.0, 0.0, -0.00367, -0.00348, 0.00367, -0.01204, 0.00442, 0.00042, 0.0, -0.0, 0.03621, -0.05963, 0.0, 0.05963, 0.0, -0.0, 0.03621, -0.00042, 0.00442, -0.01204, -0.00042, -0.00442, -0.01204, 0.00367, 0.00348, -0.00367, 0.01204, -0.00442, 0.00042, 0.01204, 0.00442, 0.00042, 0.00367, -0.00348, -0.00367, 0.00042, 0.00442, -0.01204, 0.0, -0.0, 0.0017, -0.01839, 0.0, 0.01839, 0.0, 0.0, 0.0017, 0.01839, -0.0, -0.01839, -0.0, 0.0, -0.0017, -0.01839, 0.0, -0.01839, -0.0, -0.0, -0.0017, 0.0017, -0.0, 0.0, 0.01839, -0.0, 0.01839, 0.0017, 0.0, 0.0, -0.0017, 0.0, -0.0, -0.0017, -0.0, -0.0]]), "T2gz2":array([[0.0323, -0.0, -0.0, 0.01025, 0.01025, 0.01127, -0.01025, 0.01025, -0.01127, 0.01322, -0.01322, -0.0, 0.01695, 0.00843, 0.0095, -0.01025, 0.01025, 0.01127, 0.0, 0.0323, -0.0, 0.01025, 0.01025, -0.01127, -0.01695, -0.00843, 0.0095, 0.00843, 0.01695, 0.0095, -0.01025, -0.01025, -0.01127, 0.0, 0.0323, 0.0, -0.01322, -0.01322, -0.0, -0.01695, 0.00843, -0.0095, 0.0323, -0.0, -0.0, -0.01695, -0.00843, -0.0095, -0.0, -0.0323, 0.0, 0.00843, 0.01695, -0.0095, -0.00843, 0.01695, 0.0095, -0.00843, 0.01695, -0.0095, -0.0323, 0.0, 0.0, 0.01695, -0.00843, 0.0095, -0.00843, -0.01695, -0.0095, -0.0323, 0.0, 0.0, 0.01695, -0.00843, -0.0095, 0.01025, -0.01025, 0.01127, -0.01322, 0.01322, 0.0, -0.0, -0.0323, -0.0, 0.01322, 0.01322, 0.0, -0.00843, -0.01695, 0.0095, 0.00843, -0.01695, -0.0095, -0.01695, 0.00843, 0.0095, 0.01025, -0.01025, -0.01127, -0.01025, -0.01025, 0.01127, 0.01695, 0.00843, -0.0095, 0.00843, -0.01695, 0.0095, 0.00083, -0.00083, -0.0, -0.0, -0.02205, -0.0, -0.00083, -0.00083, -0.0, 0.0, 0.02205, 0.0, -0.00083, 0.00083, 0.0, 0.0, 0.02205, 0.0, 0.00083, 0.00083, 0.0, 0.02205, -0.0, 0.0, -0.0, -0.02205, -0.0, -0.02205, 0.0, -0.0, -0.02205, 0.0, -0.0, 0.02205, -0.0, 0.0], [0.03415, -0.0, -0.0, 0.00186, 0.00186, 0.00063, -0.00186, 0.00186, -0.00063, -0.05517, 0.05517, 0.0, -0.01532, 0.00291, -0.00312, -0.00186, 0.00186, 0.00063, 0.0, 0.03415, -0.0, 0.00186, 0.00186, -0.00063, 0.01532, -0.00291, -0.00312, 0.00291, -0.01532, -0.00312, -0.00186, -0.00186, -0.00063, 0.0, 0.03415, -0.0, 0.05517, 0.05517, 0.0, 0.01532, 0.00291, 0.00312, 0.03415, 0.0, -0.0, 0.01532, -0.00291, 0.00312, -0.0, -0.03415, 0.0, 0.00291, -0.01532, 0.00312, -0.00291, -0.01532, -0.00312, -0.00291, -0.01532, 0.00312, -0.03415, 0.0, -0.0, -0.01532, -0.00291, -0.00312, -0.00291, 0.01532, 0.00312, -0.03415, -0.0, 0.0, -0.01532, -0.00291, 0.00312, 0.00186, -0.00186, 0.00063, 0.05517, -0.05517, -0.0, -0.0, -0.03415, 0.0, -0.05517, -0.05517, -0.0, -0.00291, 0.01532, -0.00312, 0.00291, 0.01532, 0.00312, 0.01532, 0.00291, -0.00312, 0.00186, -0.00186, -0.00063, -0.00186, -0.00186, 0.00063, -0.01532, 0.00291, 0.00312, 0.00291, 0.01532, -0.00312, 0.01842, -0.01842, -0.0, -0.0, -0.0043, -0.0, -0.01842, -0.01842, -0.0, 0.0, 0.0043, 0.0, -0.01842, 0.01842, 0.0, 0.0, 0.0043, 0.0, 0.01842, 0.01842, 0.0, 0.0043, 0.0, 0.0, -0.0, -0.0043, -0.0, -0.0043, -0.0, -0.0, -0.0043, -0.0, -0.0, 0.0043, 0.0, 0.0]]), "A1u":array([[0.0, -0.0, 0.0, 0.02946, -0.02946, -0.0, -0.02946, -0.02946, 0.0, 0.0, 0.0, 0.0, -0.02946, 0.0, 0.02946, 0.02946, 0.02946, 0.0, 0.0, 0.0, 0.0, -0.02946, 0.02946, -0.0, 0.02946, -0.0, 0.02946, 0.0, 0.02946, -0.02946, 0.02946, -0.02946, -0.0, 0.0, -0.0, -0.0, 0.0, -0.0, -0.0, 0.02946, -0.0, -0.02946, -0.0, -0.0, -0.0, -0.02946, -0.0, 0.02946, -0.0, -0.0, -0.0, -0.0, -0.02946, -0.02946, 0.0, -0.02946, 0.02946, -0.0, 0.02946, 0.02946, 0.0, 0.0, -0.0, 0.02946, 0.0, -0.02946, 0.0, 0.02946, -0.02946, -0.0, 0.0, 0.0, -0.02946, 0.0, -0.02946, -0.02946, -0.02946, 0.0, -0.0, -0.0, 0.0, -0.0, 0.0, 0.0, -0.0, 0.0, -0.0, -0.0, -0.02946, -0.02946, 0.0, -0.02946, 0.02946, -0.02946, -0.0, -0.02946, 0.02946, 0.02946, 0.0, -0.02946, 0.02946, -0.0, 0.02946, 0.0, 0.02946, -0.0, 0.02946, 0.02946, 0.0, -0.0, 0.0, -0.0, -0.0, 0.0, 0.0, 0.0, -0.0, 0.0, 0.0, -0.0, -0.0, 0.0, 0.0, 0.0, -0.0, 0.0, -0.0, -0.0, -0.0, -0.0, -0.0, 0.0, -0.0, 0.0, -0.0, 0.0, 0.0, 0.0, -0.0, 0.0, -0.0, 0.0, -0.0, -0.0], [0.00082, -0.06015, 0.05143, 0.00249, -0.00471, -0.00567, 0.0015, 0.00091, -0.00434, 0.05143, 0.06015, 0.00082, 0.00322, 0.01, -0.00159, 0.00249, 0.00471, 0.00567, 0.00872, -0.0, 0.00872, 0.0015, -0.00091, 0.00434, -0.00322, -0.01, -0.00159, -0.00434, -0.00091, 0.0015, -0.0015, 0.00091, 0.00434, 0.00872, -0.0, -0.00872, 0.05143, -0.06015, -0.00082, -0.00159, -0.01, 0.00322, -0.00082, -0.06015, -0.05143, 0.00159, -0.01, -0.00322, -0.00872, 0.0, -0.00872, -0.00567, -0.00471, -0.00249, -0.00434, 0.00091, -0.0015, -0.00567, 0.00471, 0.00249, 0.00082, 0.06015, -0.05143, -0.00322, 0.01, 0.00159, 0.00567, 0.00471, -0.00249, -0.00082, 0.06015, 0.05143, 0.00159, 0.01, 0.00322, -0.00249, -0.00471, 0.00567, -0.05143, -0.06015, 0.00082, -0.00872, 0.0, 0.00872, -0.05143, 0.06015, -0.00082, 0.00434, 0.00091, 0.0015, 0.00567, -0.00471, 0.00249, 0.00322, -0.01, 0.00159, -0.0015, -0.00091, -0.00434, -0.00249, 0.00471, -0.00567, -0.00159, 0.01, -0.00322, 0.00434, -0.00091, -0.0015, -0.00809, -0.01133, -0.0034, -0.00323, 0.0, 0.00323, -0.00809, 0.01133, 0.0034, 0.00323, -0.0, -0.00323, 0.00809, 0.01133, -0.0034, 0.00323, -0.0, 0.00323, 0.00809, -0.01133, 0.0034, 0.0034, -0.01133, -0.00809, -0.00323, 0.0, -0.00323, -0.0034, 0.01133, -0.00809, 0.0034, 0.01133, 0.00809, -0.0034, -0.01133, 0.00809]]), "A2u":array([[0.01695, -0.0, -0.0, 0.01465, 0.01465, -0.01848, -0.01465, 0.01465, 0.01848, -0.0, -0.0, -0.01695, 0.01465, -0.01848, 0.01465, 0.01465, -0.01465, 0.01848, -0.0, 0.01695, 0.0, -0.01465, -0.01465, -0.01848, -0.01465, 0.01848, 0.01465, -0.01848, 0.01465, 0.01465, 0.01465, 0.01465, -0.01848, -0.0, -0.01695, -0.0, -0.0, 0.0, 0.01695, -0.01465, -0.01848, -0.01465, -0.01695, -0.0, 0.0, 0.01465, -0.01848, 0.01465, 0.0, 0.01695, -0.0, 0.01848, -0.01465, 0.01465, -0.01848, -0.01465, -0.01465, 0.01848, 0.01465, -0.01465, 0.01695, 0.0, 0.0, -0.01465, -0.01848, -0.01465, -0.01848, 0.01465, 0.01465, -0.01695, 0.0, -0.0, 0.01465, 0.01848, -0.01465, -0.01465, 0.01465, 0.01848, 0.0, 0.0, -0.01695, 0.0, -0.01695, 0.0, 0.0, -0.0, 0.01695, 0.01848, -0.01465, 0.01465, -0.01848, -0.01465, -0.01465, 0.01465, 0.01848, -0.01465, 0.01465, -0.01465, 0.01848, -0.01465, -0.01465, -0.01848, -0.01465, 0.01848, 0.01465, 0.01848, 0.01465, -0.01465, 0.0, -0.0, 0.01085, 0.0, 0.01085, -0.0, 0.0, 0.0, -0.01085, -0.0, 0.01085, 0.0, -0.0, 0.0, 0.01085, -0.0, -0.01085, -0.0, -0.0, -0.0, -0.01085, 0.01085, 0.0, -0.0, 0.0, -0.01085, 0.0, -0.01085, -0.0, -0.0, 0.01085, -0.0, 0.0, -0.01085, 0.0, 0.0], [-0.04941, -0.0, -0.0, 0.00853, 0.00853, -0.00752, -0.00853, 0.00853, 0.00752, -0.0, -0.0, 0.04941, 0.00853, -0.00752, 0.00853, 0.00853, -0.00853, 0.00752, -0.0, -0.04941, 0.0, -0.00853, -0.00853, -0.00752, -0.00853, 0.00752, 0.00853, -0.00752, 0.00853, 0.00853, 0.00853, 0.00853, -0.00752, -0.0, 0.04941, -0.0, -0.0, 0.0, -0.04941, -0.00853, -0.00752, -0.00853, 0.04941, -0.0, 0.0, 0.00853, -0.00752, 0.00853, 0.0, -0.04941, -0.0, 0.00752, -0.00853, 0.00853, -0.00752, -0.00853, -0.00853, 0.00752, 0.00853, -0.00853, -0.04941, 0.0, 0.0, -0.00853, -0.00752, -0.00853, -0.00752, 0.00853, 0.00853, 0.04941, 0.0, -0.0, 0.00853, 0.00752, -0.00853, -0.00853, 0.00853, 0.00752, 0.0, 0.0, 0.04941, 0.0, 0.04941, 0.0, 0.0, -0.0, -0.04941, 0.00752, -0.00853, 0.00853, -0.00752, -0.00853, -0.00853, 0.00853, 0.00752, -0.00853, 0.00853, -0.00853, 0.00752, -0.00853, -0.00853, -0.00752, -0.00853, 0.00752, 0.00853, 0.00752, 0.00853, -0.00853, 0.0, -0.0, 0.00549, 0.0, 0.00549, -0.0, 0.0, 0.0, -0.00549, -0.0, 0.00549, 0.0, -0.0, 0.0, 0.00549, -0.0, -0.00549, -0.0, -0.0, -0.0, -0.00549, 0.00549, 0.0, -0.0, 0.0, -0.00549, 0.0, -0.00549, -0.0, -0.0, 0.00549, -0.0, 0.0, -0.00549, 0.0, 0.0]]), "Eu1":array([[0.02138, -0.00401, 0.02222, 0.0087, -0.00445, 0.02222, 0.0087, 0.00445, 0.02222, 0.02222, 0.00401, 0.02138, 0.00981, 0.00138, 0.00981, 0.0086, -0.00088, 0.01379, 0.01846, -0.0, 0.01846, 0.0086, 0.00088, 0.01379, 0.01237, -0.0, 0.01237, 0.02222, -0.00445, 0.0087, 0.0087, -0.00445, 0.02222, 0.01993, -0.0, 0.01993, 0.02222, -0.00401, 0.02138, 0.00981, -0.00138, 0.00981, 0.02138, 0.00401, 0.02222, 0.00981, 0.00138, 0.00981, 0.01846, -0.0, 0.01846, 0.01379, -0.00088, 0.0086, 0.01379, 0.00088, 0.0086, 0.02222, 0.00445, 0.0087, 0.02138, -0.00401, 0.02222, 0.00981, -0.00138, 0.00981, 0.02222, -0.00445, 0.0087, 0.02138, 0.00401, 0.02222, 0.01237, -0.0, 0.01237, 0.0087, 0.00445, 0.02222, 0.02222, 0.00401, 0.02138, 0.01993, -0.0, 0.01993, 0.02222, -0.00401, 0.02138, 0.01379, -0.00088, 0.0086, 0.01379, 0.00088, 0.0086, 0.01237, 0.0, 0.01237, 0.0086, -0.00088, 0.01379, 0.0086, 0.00088, 0.01379, 0.01237, 0.0, 0.01237, 0.02222, 0.00445, 0.0087, 0.01223, 0.00228, 0.00886, 0.00952, -0.0, 0.00952, 0.01223, -0.00228, 0.00886, 0.00952, -0.0, 0.00952, 0.01223, 0.00228, 0.00886, 0.01534, -0.0, 0.01534, 0.01223, -0.00228, 0.00886, 0.00886, 0.00228, 0.01223, 0.01534, -0.0, 0.01534, 0.00886, -0.00228, 0.01223, 0.00886, 0.00228, 0.01223, 0.00886, -0.00228, 0.01223], [0.0107, 0.01515, -0.01264, -0.00793, -0.00164, 0.03185, -0.00793, 0.00164, 0.03185, -0.01264, -0.01515, 0.0107, -0.00898, 0.01275, -0.00898, -0.00669, 0.00076, 0.01775, -0.0018, 0.0, -0.0018, -0.00669, -0.00076, 0.01775, -0.00577, 0.0, -0.00577, 0.03185, -0.00164, -0.00793, -0.00793, -0.00164, 0.03185, -0.03432, 0.0, -0.03432, -0.01264, 0.01515, 0.0107, -0.00898, -0.01275, -0.00898, 0.0107, -0.01515, -0.01264, -0.00898, 0.01275, -0.00898, -0.0018, 0.0, -0.0018, 0.01775, 0.00076, -0.00669, 0.01775, -0.00076, -0.00669, 0.03185, 0.00164, -0.00793, 0.0107, 0.01515, -0.01264, -0.00898, -0.01275, -0.00898, 0.03185, -0.00164, -0.00793, 0.0107, -0.01515, -0.01264, -0.00577, -0.0, -0.00577, -0.00793, 0.00164, 0.03185, -0.01264, -0.01515, 0.0107, -0.03432, 0.0, -0.03432, -0.01264, 0.01515, 0.0107, 0.01775, 0.00076, -0.00669, 0.01775, -0.00076, -0.00669, -0.00577, 0.0, -0.00577, -0.00669, 0.00076, 0.01775, -0.00669, -0.00076, 0.01775, -0.00577, -0.0, -0.00577, 0.03185, 0.00164, -0.00793, -0.00381, 0.01137, -0.01102, 0.00753, -0.0, 0.00753, -0.00381, -0.01137, -0.01102, 0.00753, -0.0, 0.00753, -0.00381, 0.01137, -0.01102, -0.01669, 0.0, -0.01669, -0.00381, -0.01137, -0.01102, -0.01102, 0.01137, -0.00381, -0.01669, 0.0, -0.01669, -0.01102, -0.01137, -0.00381, -0.01102, 0.01137, -0.00381, -0.01102, -0.01137, -0.00381]]), "Eu2":array([[0.0, 0.01726, 0.01726, -0.00366, 0.00923, 0.01827, -0.00154, 0.01059, 0.01831, 0.00251, 0.02334, 0.02033, -0.00366, 0.01827, 0.00923, 0.00366, 0.00923, 0.01827, -0.00251, 0.02033, 0.02334, 0.00154, 0.01059, 0.01831, -0.00154, 0.01831, 0.01059, 0.00305, 0.00823, 0.00823, -0.00366, 0.00923, 0.01827, 0.00251, 0.02033, 0.02334, -0.00251, 0.02334, 0.02033, 0.00154, 0.01831, 0.01059, 0.0, 0.02754, 0.02754, -0.00366, 0.01827, 0.00923, -0.00251, 0.02033, 0.02334, 0.0, 0.01022, 0.01022, -0.00305, 0.00823, 0.00823, 0.0, 0.01022, 0.01022, 0.0, 0.01726, 0.01726, 0.00154, 0.01831, 0.01059, 0.00305, 0.00823, 0.00823, 0.0, 0.02754, 0.02754, 0.00366, 0.01827, 0.00923, -0.00154, 0.01059, 0.01831, 0.00251, 0.02334, 0.02033, 0.00251, 0.02033, 0.02334, -0.00251, 0.02334, 0.02033, -0.0, 0.01022, 0.01022, -0.00305, 0.00823, 0.00823, 0.00366, 0.01827, 0.00923, 0.00366, 0.00923, 0.01827, 0.00154, 0.01059, 0.01831, -0.00154, 0.01831, 0.01059, -0.0, 0.01022, 0.01022, -0.00237, 0.00955, 0.00747, -0.00237, 0.00747, 0.00955, 0.00237, 0.00955, 0.00747, -0.00237, 0.00747, 0.00955, -0.00237, 0.00955, 0.00747, 0.00237, 0.00747, 0.00955, 0.00237, 0.00955, 0.00747, 0.0, 0.01346, 0.01346, 0.00237, 0.00747, 0.00955, 0.0, 0.01624, 0.01624, 0.0, 0.01346, 0.01346, 0.0, 0.01624, 0.01624], [-0.0, 0.00113, 0.00113, -0.00134, -0.00714, 0.03282, 0.0013, -0.00513, 0.01836, -0.01463, -0.01633, 0.01323, -0.00134, 0.03282, -0.00714, 0.00134, -0.00714, 0.03282, 0.01463, 0.01323, -0.01633, -0.0013, -0.00513, 0.01836, 0.0013, 0.01836, -0.00513, 0.01467, -0.00682, -0.00682, -0.00134, -0.00714, 0.03282, -0.01463, 0.01323, -0.01633, 0.01463, -0.01633, 0.01323, -0.0013, 0.01836, -0.00513, -0.0, -0.02736, -0.02736, -0.00134, 0.03282, -0.00714, 0.01463, 0.01323, -0.01633, 0.0, -0.00501, -0.00501, -0.01467, -0.00682, -0.00682, 0.0, -0.00501, -0.00501, 0.0, 0.00113, 0.00113, -0.0013, 0.01836, -0.00513, 0.01467, -0.00682, -0.00682, -0.0, -0.02736, -0.02736, 0.00134, 0.03282, -0.00714, 0.0013, -0.00513, 0.01836, -0.01463, -0.01633, 0.01323, -0.01463, 0.01323, -0.01633, 0.01463, -0.01633, 0.01323, -0.0, -0.00501, -0.00501, -0.01467, -0.00682, -0.00682, 0.00134, 0.03282, -0.00714, 0.00134, -0.00714, 0.03282, -0.0013, -0.00513, 0.01836, 0.0013, 0.01836, -0.00513, -0.0, -0.00501, -0.00501, 0.01169, -0.00442, -0.00981, 0.01169, -0.00981, -0.00442, -0.01169, -0.00442, -0.00981, 0.01169, -0.00981, -0.00442, 0.01169, -0.00442, -0.00981, -0.01169, -0.00981, -0.00442, -0.01169, -0.00442, -0.00981, 0.0, 0.00798, 0.00798, -0.01169, -0.00981, -0.00442, -0.0, -0.01577, -0.01577, 0.0, 0.00798, 0.00798, -0.0, -0.01577, -0.01577]]), "Eu3":array([[0.01753, 0.00066, -0.01974, 0.01015, -0.00132, -0.01301, 0.01015, 0.00132, -0.01301, 0.01974, 0.00066, -0.01753, 0.01142, -0.0, -0.01142, 0.00925, 0.00242, -0.01051, 0.04744, -0.0, -0.04744, 0.00925, -0.00242, -0.01051, 0.01021, -0.00237, -0.01021, 0.01301, 0.00132, -0.01015, 0.01015, -0.00132, -0.01301, 0.02306, 0.0, -0.02306, 0.01974, -0.00066, -0.01753, 0.01142, 0.0, -0.01142, 0.01753, -0.00066, -0.01974, 0.01142, 0.0, -0.01142, 0.04744, -0.0, -0.04744, 0.01051, -0.00242, -0.00925, 0.01051, 0.00242, -0.00925, 0.01301, -0.00132, -0.01015, 0.01753, 0.00066, -0.01974, 0.01142, -0.0, -0.01142, 0.01301, 0.00132, -0.01015, 0.01753, -0.00066, -0.01974, 0.01021, 0.00237, -0.01021, 0.01015, 0.00132, -0.01301, 0.01974, 0.00066, -0.01753, 0.02306, -0.0, -0.02306, 0.01974, -0.00066, -0.01753, 0.01051, -0.00242, -0.00925, 0.01051, 0.00242, -0.00925, 0.01021, 0.00237, -0.01021, 0.00925, 0.00242, -0.01051, 0.00925, -0.00242, -0.01051, 0.01021, -0.00237, -0.01021, 0.01301, -0.00132, -0.01015, 0.01525, -0.00605, -0.01077, 0.01515, -0.0, -0.01515, 0.01525, 0.00605, -0.01077, 0.01515, 0.0, -0.01515, 0.01525, -0.00605, -0.01077, 0.00714, -0.0, -0.00714, 0.01525, 0.00605, -0.01077, 0.01077, 0.00605, -0.01525, 0.00714, 0.0, -0.00714, 0.01077, -0.00605, -0.01525, 0.01077, 0.00605, -0.01525, 0.01077, -0.00605, -0.01525], [0.0256, -0.00428, -0.00912, -7e-05, -0.0011, -0.02045, -7e-05, 0.0011, -0.02045, 0.00912, -0.00428, -0.0256, 0.00048, -0.0, -0.00048, -0.00267, 0.00257, -0.03291, -0.0542, 0.0, 0.0542, -0.00267, -0.00257, -0.03291, 0.00081, 0.01103, -0.00081, 0.02045, 0.0011, 7e-05, -7e-05, -0.0011, -0.02045, 0.01069, -0.0, -0.01069, 0.00912, 0.00428, -0.0256, 0.00048, -0.0, -0.00048, 0.0256, 0.00428, -0.00912, 0.00048, 0.0, -0.00048, -0.0542, 0.0, 0.0542, 0.03291, -0.00257, 0.00267, 0.03291, 0.00257, 0.00267, 0.02045, -0.0011, 7e-05, 0.0256, -0.00428, -0.00912, 0.00048, 0.0, -0.00048, 0.02045, 0.0011, 7e-05, 0.0256, 0.00428, -0.00912, 0.00081, -0.01103, -0.00081, -7e-05, 0.0011, -0.02045, 0.00912, -0.00428, -0.0256, 0.01069, 0.0, -0.01069, 0.00912, 0.00428, -0.0256, 0.03291, -0.00257, 0.00267, 0.03291, 0.00257, 0.00267, 0.00081, -0.01103, -0.00081, -0.00267, 0.00257, -0.03291, -0.00267, -0.00257, -0.03291, 0.00081, 0.01103, -0.00081, 0.02045, -0.0011, 7e-05, 0.00277, 0.00736, 0.00587, -0.0041, -0.0, 0.0041, 0.00277, -0.00736, 0.00587, -0.0041, 0.0, 0.0041, 0.00277, 0.00736, 0.00587, 0.01078, -0.0, -0.01078, 0.00277, -0.00736, 0.00587, -0.00587, -0.00736, -0.00277, 0.01078, -0.0, -0.01078, -0.00587, 0.00736, -0.00277, -0.00587, -0.00736, -0.00277, -0.00587, 0.00736, -0.00277]]), "Eu4":array([[0.0, -0.06468, 0.06468, 0.00081, -0.00812, -0.00957, 0.00102, -0.00815, -0.01943, -0.01458, -0.0239, -0.00487, -0.00081, 0.00957, 0.00812, -0.00081, -0.00812, -0.00957, -0.01458, 0.00487, 0.0239, -0.00102, -0.00815, -0.01943, -0.00102, 0.01943, 0.00815, -0.0, -0.0088, 0.0088, 0.00081, -0.00812, -0.00957, 0.01458, 0.00487, 0.0239, 0.01458, -0.0239, -0.00487, 0.00102, 0.01943, 0.00815, -0.0, -0.00182, 0.00182, -0.00081, 0.00957, 0.00812, -0.01458, 0.00487, 0.0239, 0.0103, -0.00718, 0.00718, -0.0, -0.0088, 0.0088, -0.0103, -0.00718, 0.00718, -0.0, -0.06468, 0.06468, 0.00102, 0.01943, 0.00815, 0.0, -0.0088, 0.0088, -0.0, -0.00182, 0.00182, 0.00081, 0.00957, 0.00812, 0.00102, -0.00815, -0.01943, -0.01458, -0.0239, -0.00487, 0.01458, 0.00487, 0.0239, 0.01458, -0.0239, -0.00487, 0.0103, -0.00718, 0.00718, 0.0, -0.0088, 0.0088, 0.00081, 0.00957, 0.00812, -0.00081, -0.00812, -0.00957, -0.00102, -0.00815, -0.01943, -0.00102, 0.01943, 0.00815, -0.0103, -0.00718, 0.00718, 0.01208, -0.0075, 0.0135, -0.01208, -0.0135, 0.0075, -0.01208, -0.0075, 0.0135, -0.01208, -0.0135, 0.0075, 0.01208, -0.0075, 0.0135, 0.01208, -0.0135, 0.0075, -0.01208, -0.0075, 0.0135, -0.0, -0.01564, 0.01564, 0.01208, -0.0135, 0.0075, 0.0, 0.00228, -0.00228, -0.0, -0.01564, 0.01564, 0.0, 0.00228, -0.00228], [0.0, 0.01774, -0.01774, 0.00035, -0.0081, 0.01688, -0.00099, -0.00187, 0.02177, -0.01554, -0.03135, 0.02266, -0.00035, -0.01688, 0.0081, -0.00035, -0.0081, 0.01688, -0.01554, -0.02266, 0.03135, 0.00099, -0.00187, 0.02177, 0.00099, -0.02177, 0.00187, -0.0, -0.00808, 0.00808, 0.00035, -0.0081, 0.01688, 0.01554, -0.02266, 0.03135, 0.01554, -0.03135, 0.02266, -0.00099, -0.02177, 0.00187, -0.0, -0.02141, 0.02141, -0.00035, -0.01688, 0.0081, -0.01554, -0.02266, 0.03135, -0.01093, -0.00323, 0.00323, -0.0, -0.00808, 0.00808, 0.01093, -0.00323, 0.00323, 0.0, 0.01774, -0.01774, -0.00099, -0.02177, 0.00187, 0.0, -0.00808, 0.00808, -0.0, -0.02141, 0.02141, 0.00035, -0.01688, 0.0081, -0.00099, -0.00187, 0.02177, -0.01554, -0.03135, 0.02266, 0.01554, -0.02266, 0.03135, 0.01554, -0.03135, 0.02266, -0.01093, -0.00323, 0.00323, 0.0, -0.00808, 0.00808, 0.00035, -0.01688, 0.0081, -0.00035, -0.0081, 0.01688, 0.00099, -0.00187, 0.02177, 0.00099, -0.02177, 0.00187, 0.01093, -0.00323, 0.00323, -0.00224, -0.00912, 0.00352, 0.00224, -0.00352, 0.00912, 0.00224, -0.00912, 0.00352, 0.00224, -0.00352, 0.00912, -0.00224, -0.00912, 0.00352, -0.00224, -0.00352, 0.00912, 0.00224, -0.00912, 0.00352, -0.0, -0.00999, 0.00999, -0.00224, -0.00352, 0.00912, -0.0, -0.01041, 0.01041, -0.0, -0.00999, 0.00999, -0.0, -0.01041, 0.01041]]), "Eu5":array([[0.02007, 0.02513, -0.0081, 0.00849, 0.00849, -0.00723, 0.01146, 0.01146, 0.0, 0.02394, 0.02394, -0.0, 0.01099, 0.00828, 0.00087, 0.01146, 0.01146, -0.0, 0.02513, 0.02007, -0.0081, 0.00849, 0.00849, 0.00723, 0.01099, 0.00828, -0.00087, 0.00828, 0.01099, 0.00087, 0.00849, 0.00849, -0.00723, 0.02513, 0.02007, 0.0081, 0.01831, 0.01831, 0.0, 0.01476, 0.01196, -0.00127, 0.02007, 0.02513, 0.0081, 0.01099, 0.00828, 0.00087, 0.02513, 0.02007, -0.0081, 0.00828, 0.01099, -0.00087, 0.01196, 0.01476, -0.00127, 0.01196, 0.01476, 0.00127, 0.02007, 0.02513, -0.0081, 0.01476, 0.01196, -0.00127, 0.00828, 0.01099, 0.00087, 0.02007, 0.02513, 0.0081, 0.01476, 0.01196, 0.00127, 0.01146, 0.01146, -0.0, 0.02394, 0.02394, -0.0, 0.02513, 0.02007, 0.0081, 0.01831, 0.01831, -0.0, 0.00828, 0.01099, -0.00087, 0.01196, 0.01476, -0.00127, 0.01476, 0.01196, 0.00127, 0.01146, 0.01146, 0.0, 0.00849, 0.00849, 0.00723, 0.01099, 0.00828, -0.00087, 0.01196, 0.01476, 0.00127, 0.01043, 0.01043, 0.0, 0.01534, 0.0121, -0.00659, 0.01316, 0.01316, 0.0, 0.01534, 0.0121, -0.00659, 0.01043, 0.01043, -0.0, 0.01534, 0.0121, 0.00659, 0.01316, 0.01316, -0.0, 0.0121, 0.01534, -0.00659, 0.01534, 0.0121, 0.00659, 0.0121, 0.01534, 0.00659, 0.0121, 0.01534, -0.00659, 0.0121, 0.01534, 0.00659], [0.01684, -0.00672, 0.01653, -0.00637, -0.00637, 0.01366, -0.00123, -0.00123, 0.0, -0.02728, -0.02728, 0.0, -0.0044, 0.03674, -0.00244, -0.00123, -0.00123, 0.0, -0.00672, 0.01684, 0.01653, -0.00637, -0.00637, -0.01366, -0.0044, 0.03674, 0.00244, 0.03674, -0.0044, -0.00244, -0.00637, -0.00637, 0.01366, -0.00672, 0.01684, -0.01653, 0.00542, 0.00542, -0.0, -0.00157, 0.02259, 0.00142, 0.01684, -0.00672, -0.01653, -0.0044, 0.03674, -0.00244, -0.00672, 0.01684, 0.01653, 0.03674, -0.0044, 0.00244, 0.02259, -0.00157, 0.00142, 0.02259, -0.00157, -0.00142, 0.01684, -0.00672, 0.01653, -0.00157, 0.02259, 0.00142, 0.03674, -0.0044, -0.00244, 0.01684, -0.00672, -0.01653, -0.00157, 0.02259, -0.00142, -0.00123, -0.00123, -0.0, -0.02728, -0.02728, -0.0, -0.00672, 0.01684, -0.01653, 0.00542, 0.00542, 0.0, 0.03674, -0.0044, 0.00244, 0.02259, -0.00157, 0.00142, -0.00157, 0.02259, -0.00142, -0.00123, -0.00123, -0.0, -0.00637, -0.00637, -0.01366, -0.0044, 0.03674, 0.00244, 0.02259, -0.00157, -0.00142, 0.01191, 0.01191, -0.0, 0.00153, -0.0087, 0.0089, -0.00937, -0.00937, -0.0, 0.00153, -0.0087, 0.0089, 0.01191, 0.01191, 0.0, 0.00153, -0.0087, -0.0089, -0.00937, -0.00937, 0.0, -0.0087, 0.00153, 0.0089, 0.00153, -0.0087, -0.0089, -0.0087, 0.00153, -0.0089, -0.0087, 0.00153, 0.0089, -0.0087, 0.00153, -0.0089]]), "Eu6":array([[0.02483, 0.00281, -0.02204, 0.00455, -0.00455, 0.0, 0.00356, -0.00356, 0.01119, 0.01677, -0.01677, -0.0, 0.0078, -0.02179, 0.00063, 0.00356, -0.00356, -0.01119, -0.00281, -0.02483, 0.02204, 0.00455, -0.00455, -0.0, 0.0078, -0.02179, -0.00063, 0.02179, -0.0078, -0.00063, 0.00455, -0.00455, -0.0, -0.00281, -0.02483, -0.02204, 0.02506, -0.02506, 0.0, 0.00435, -0.02809, -0.0026, 0.02483, 0.00281, 0.02204, 0.0078, -0.02179, 0.00063, -0.00281, -0.02483, 0.02204, 0.02179, -0.0078, 0.00063, 0.02809, -0.00435, 0.0026, 0.02809, -0.00435, -0.0026, 0.02483, 0.00281, -0.02204, 0.00435, -0.02809, -0.0026, 0.02179, -0.0078, -0.00063, 0.02483, 0.00281, 0.02204, 0.00435, -0.02809, 0.0026, 0.00356, -0.00356, 0.01119, 0.01677, -0.01677, 0.0, -0.00281, -0.02483, -0.02204, 0.02506, -0.02506, 0.0, 0.02179, -0.0078, 0.00063, 0.02809, -0.00435, 0.0026, 0.00435, -0.02809, 0.0026, 0.00356, -0.00356, -0.01119, 0.00455, -0.00455, 0.0, 0.0078, -0.02179, -0.00063, 0.02809, -0.00435, -0.0026, 0.00489, -0.00489, 0.0, 0.00952, -0.0034, 0.00247, 0.01049, -0.01049, -0.0, 0.00952, -0.0034, 0.00247, 0.00489, -0.00489, -0.0, 0.00952, -0.0034, -0.00247, 0.01049, -0.01049, 0.0, 0.0034, -0.00952, -0.00247, 0.00952, -0.0034, -0.00247, 0.0034, -0.00952, 0.00247, 0.0034, -0.00952, -0.00247, 0.0034, -0.00952, 0.00247], [0.00585, -0.03092, 0.01725, 0.01065, -0.01065, -0.0, 0.01204, -0.01204, -0.00747, 0.01097, -0.01097, 0.0, 0.01223, 0.00299, 0.00072, 0.01204, -0.01204, 0.00747, 0.03092, -0.00585, -0.01725, 0.01065, -0.01065, 0.0, 0.01223, 0.00299, -0.00072, -0.00299, -0.01223, -0.00072, 0.01065, -0.01065, 0.0, 0.03092, -0.00585, 0.01725, 0.02899, -0.02899, 0.0, 0.0091, 0.01466, -0.00053, 0.00585, -0.03092, -0.01725, 0.01223, 0.00299, 0.00072, 0.03092, -0.00585, -0.01725, -0.00299, -0.01223, 0.00072, -0.01466, -0.0091, 0.00053, -0.01466, -0.0091, -0.00053, 0.00585, -0.03092, 0.01725, 0.0091, 0.01466, -0.00053, -0.00299, -0.01223, -0.00072, 0.00585, -0.03092, -0.01725, 0.0091, 0.01466, 0.00053, 0.01204, -0.01204, -0.00747, 0.01097, -0.01097, -0.0, 0.03092, -0.00585, 0.01725, 0.02899, -0.02899, -0.0, -0.00299, -0.01223, 0.00072, -0.01466, -0.0091, 0.00053, 0.0091, 0.01466, 0.00053, 0.01204, -0.01204, 0.00747, 0.01065, -0.01065, -0.0, 0.01223, 0.00299, -0.00072, -0.01466, -0.0091, -0.00053, 0.02347, -0.02347, 0.0, 0.00881, -0.01439, -0.00947, -0.00182, 0.00182, -0.0, 0.00881, -0.01439, -0.00947, 0.02347, -0.02347, -0.0, 0.00881, -0.01439, 0.00947, -0.00182, 0.00182, -0.0, 0.01439, -0.00881, 0.00947, 0.00881, -0.01439, 0.00947, 0.01439, -0.00881, -0.00947, 0.01439, -0.00881, 0.00947, 0.01439, -0.00881, -0.00947]]), "T1ux":array([[0.03241, -0.0, 0.0, 0.01587, -0.00122, -0.0003, 0.01587, 0.00122, -0.0003, 0.02871, 0.00068, 0.0, 0.01587, -0.0003, -0.00122, 0.01587, 0.00122, 0.0003, 0.02871, -0.0, -0.00068, 0.01587, -0.00122, 0.0003, 0.01587, -0.0003, 0.00122, 0.02092, -0.00129, -0.00129, 0.01587, -0.00122, -0.0003, 0.02871, -0.0, 0.00068, 0.02871, -0.00068, -0.0, 0.01587, 0.0003, -0.00122, 0.03241, -0.0, -0.0, 0.01587, -0.0003, -0.00122, 0.02871, -0.0, -0.00068, 0.02092, -0.00129, 0.00129, 0.02092, 0.00129, 0.00129, 0.02092, 0.00129, -0.00129, 0.03241, -0.0, -0.0, 0.01587, 0.0003, -0.00122, 0.02092, -0.00129, -0.00129, 0.03241, -0.0, -0.0, 0.01587, 0.0003, 0.00122, 0.01587, 0.00122, -0.0003, 0.02871, 0.00068, 0.0, 0.02871, -0.0, 0.00068, 0.02871, -0.00068, -0.0, 0.02092, -0.00129, 0.00129, 0.02092, 0.00129, 0.00129, 0.01587, 0.0003, 0.00122, 0.01587, 0.00122, 0.0003, 0.01587, -0.00122, 0.0003, 0.01587, -0.0003, 0.00122, 0.02092, 0.00129, -0.00129, 0.01889, -0.00396, -0.0, 0.01889, -0.0, -0.00396, 0.01889, 0.00396, 0.0, 0.01889, -0.0, -0.00396, 0.01889, -0.00396, 0.0, 0.01889, -0.0, 0.00396, 0.01889, 0.00396, 0.0, 0.0159, 0.0, -0.0, 0.01889, -0.0, 0.00396, 0.0159, 0.0, 0.0, 0.0159, -0.0, 0.0, 0.0159, -0.0, 0.0], [0.01704, 0.0, -0.0, -0.00796, -0.00155, 0.00996, -0.00796, 0.00155, 0.00996, -0.01937, -0.02421, -0.0, -0.00796, 0.00996, -0.00155, -0.00796, 0.00155, -0.00996, -0.01937, 0.0, 0.02421, -0.00796, -0.00155, -0.00996, -0.00796, 0.00996, 0.00155, 0.03715, -0.00166, -0.00166, -0.00796, -0.00155, 0.00996, -0.01937, 0.0, -0.02421, -0.01937, 0.02421, -0.0, -0.00796, -0.00996, -0.00155, 0.01704, -0.0, -0.0, -0.00796, 0.00996, -0.00155, -0.01937, 0.0, 0.02421, 0.03715, -0.00166, 0.00166, 0.03715, 0.00166, 0.00166, 0.03715, 0.00166, -0.00166, 0.01704, -0.0, -0.0, -0.00796, -0.00996, -0.00155, 0.03715, -0.00166, -0.00166, 0.01704, -0.0, 0.0, -0.00796, -0.00996, 0.00155, -0.00796, 0.00155, 0.00996, -0.01937, -0.02421, -0.0, -0.01937, 0.0, -0.02421, -0.01937, 0.02421, 0.0, 0.03715, -0.00166, 0.00166, 0.03715, 0.00166, 0.00166, -0.00796, -0.00996, 0.00155, -0.00796, 0.00155, -0.00996, -0.00796, -0.00155, -0.00996, -0.00796, 0.00996, 0.00155, 0.03715, 0.00166, -0.00166, -0.00246, 0.01437, -0.0, -0.00246, -0.0, 0.01437, -0.00246, -0.01437, -0.0, -0.00246, -0.0, 0.01437, -0.00246, 0.01437, -0.0, -0.00246, 0.0, -0.01437, -0.00246, -0.01437, -0.0, -0.01595, -0.0, 0.0, -0.00246, -0.0, -0.01437, -0.01595, 0.0, 0.0, -0.01595, 0.0, -0.0, -0.01595, -0.0, -0.0]]), "T1uy":array([[0.0, -0.03581, 0.02211, -0.00064, -0.01413, 0.00934, 0.00064, -0.01413, -0.00934, -0.02211, -0.03581, 0.0, 0.0006, 0.01729, 0.0006, 0.00064, -0.01413, 0.00934, -0.0, -0.00402, -0.0, -0.00064, -0.01413, -0.00934, 0.0006, 0.01729, -0.0006, 0.00934, -0.01413, -0.00064, -0.00064, -0.01413, 0.00934, -0.0, -0.00402, -0.0, 0.02211, -0.03581, -0.0, -0.0006, 0.01729, -0.0006, -0.0, -0.03581, -0.02211, 0.0006, 0.01729, 0.0006, -0.0, -0.00402, 0.0, 0.00934, -0.01413, 0.00064, -0.00934, -0.01413, -0.00064, -0.00934, -0.01413, 0.00064, 0.0, -0.03581, 0.02211, -0.0006, 0.01729, -0.0006, 0.00934, -0.01413, -0.00064, -0.0, -0.03581, -0.02211, -0.0006, 0.01729, 0.0006, 0.00064, -0.01413, -0.00934, -0.02211, -0.03581, -0.0, 0.0, -0.00402, 0.0, 0.02211, -0.03581, 0.0, 0.00934, -0.01413, 0.00064, -0.00934, -0.01413, -0.00064, -0.0006, 0.01729, 0.0006, 0.00064, -0.01413, 0.00934, -0.00064, -0.01413, -0.00934, 0.0006, 0.01729, -0.0006, -0.00934, -0.01413, 0.00064, 0.01582, -0.01282, -0.0, 0.0, -0.01866, -0.0, -0.01582, -0.01282, -0.0, -0.0, -0.01866, 0.0, 0.01582, -0.01282, -0.0, -0.0, -0.01866, -0.0, -0.01582, -0.01282, 0.0, 0.0, -0.01282, 0.01582, 0.0, -0.01866, 0.0, 0.0, -0.01282, -0.01582, 0.0, -0.01282, 0.01582, 0.0, -0.01282, -0.01582], [0.0, 0.01917, 0.00889, -0.00143, 0.00727, 0.00626, 0.00143, 0.00727, -0.00626, -0.00889, 0.01917, -0.0, -0.00348, 0.04535, -0.00348, 0.00143, 0.00727, 0.00626, 0.0, 0.04392, -0.0, -0.00143, 0.00727, -0.00626, -0.00348, 0.04535, 0.00348, 0.00626, 0.00727, -0.00143, -0.00143, 0.00727, 0.00626, 0.0, 0.04392, -0.0, 0.00889, 0.01917, -0.0, 0.00348, 0.04535, 0.00348, 0.0, 0.01917, -0.00889, -0.00348, 0.04535, -0.00348, -0.0, 0.04392, -0.0, 0.00626, 0.00727, 0.00143, -0.00626, 0.00727, -0.00143, -0.00626, 0.00727, 0.00143, 0.0, 0.01917, 0.00889, 0.00348, 0.04535, 0.00348, 0.00626, 0.00727, -0.00143, 0.0, 0.01917, -0.00889, 0.00348, 0.04535, -0.00348, 0.00143, 0.00727, -0.00626, -0.00889, 0.01917, -0.0, -0.0, 0.04392, 0.0, 0.00889, 0.01917, -0.0, 0.00626, 0.00727, 0.00143, -0.00626, 0.00727, -0.00143, 0.00348, 0.04535, -0.00348, 0.00143, 0.00727, 0.00626, -0.00143, 0.00727, -0.00626, -0.00348, 0.04535, 0.00348, -0.00626, 0.00727, 0.00143, 0.00605, 0.01383, -0.0, 0.0, 0.0036, -0.0, -0.00605, 0.01383, 0.0, 0.0, 0.0036, -0.0, 0.00605, 0.01383, -0.0, 0.0, 0.0036, 0.0, -0.00605, 0.01383, -0.0, 0.0, 0.01383, 0.00605, -0.0, 0.0036, -0.0, 0.0, 0.01383, -0.00605, 0.0, 0.01383, 0.00605, 0.0, 0.01383, -0.00605]]), "T1uz":array([[-0.0, -0.00815, 0.03626, -0.0017, -0.0017, 0.0255, -0.0017, 0.0017, 0.0255, -0.0, 0.0, 0.02715, -0.00281, 0.00409, 0.0122, 0.0017, -0.0017, 0.0255, -0.00815, -0.0, 0.03626, 0.0017, 0.0017, 0.0255, 0.00281, -0.00409, 0.0122, 0.00409, -0.00281, 0.0122, -0.0017, -0.0017, 0.0255, 0.00815, -0.0, 0.03626, 0.0, 0.0, 0.02715, -0.00281, -0.00409, 0.0122, 0.0, 0.00815, 0.03626, -0.00281, 0.00409, 0.0122, -0.00815, -0.0, 0.03626, -0.00409, 0.00281, 0.0122, -0.00409, -0.00281, 0.0122, 0.00409, 0.00281, 0.0122, 0.0, -0.00815, 0.03626, -0.00281, -0.00409, 0.0122, 0.00409, -0.00281, 0.0122, 0.0, 0.00815, 0.03626, 0.00281, 0.00409, 0.0122, -0.0017, 0.0017, 0.0255, -0.0, 0.0, 0.02715, 0.00815, -0.0, 0.03626, 0.0, -0.0, 0.02715, -0.00409, 0.00281, 0.0122, -0.00409, -0.00281, 0.0122, 0.00281, 0.00409, 0.0122, 0.0017, -0.0017, 0.0255, 0.0017, 0.0017, 0.0255, 0.00281, -0.00409, 0.0122, 0.00409, 0.00281, 0.0122, -0.0, -0.0, 0.01023, -0.00148, -0.0, 0.01712, 0.0, 0.0, 0.01023, -0.00148, -0.0, 0.01712, -0.0, -0.0, 0.01023, 0.00148, -0.0, 0.01712, -0.0, -0.0, 0.01023, -0.0, -0.00148, 0.01712, 0.00148, -0.0, 0.01712, 0.0, 0.00148, 0.01712, 0.0, -0.00148, 0.01712, 0.0, 0.00148, 0.01712], [-0.0, -0.02481, 0.02098, 0.00095, 0.00095, -0.03416, 0.00095, -0.00095, -0.03416, -0.0, 0.0, -0.02511, 0.00128, -0.00901, 0.00694, -0.00095, 0.00095, -0.03416, -0.02481, 0.0, 0.02098, -0.00095, -0.00095, -0.03416, -0.00128, 0.00901, 0.00694, -0.00901, 0.00128, 0.00694, 0.00095, 0.00095, -0.03416, 0.02481, -0.0, 0.02098, -0.0, 0.0, -0.02511, 0.00128, 0.00901, 0.00694, 0.0, 0.02481, 0.02098, 0.00128, -0.00901, 0.00694, -0.02481, 0.0, 0.02098, 0.00901, -0.00128, 0.00694, 0.00901, 0.00128, 0.00694, -0.00901, -0.00128, 0.00694, 0.0, -0.02481, 0.02098, 0.00128, 0.00901, 0.00694, -0.00901, 0.00128, 0.00694, 0.0, 0.02481, 0.02098, -0.00128, -0.00901, 0.00694, 0.00095, -0.00095, -0.03416, 0.0, -0.0, -0.02511, 0.02481, -0.0, 0.02098, -0.0, 0.0, -0.02511, 0.00901, -0.00128, 0.00694, 0.00901, 0.00128, 0.00694, -0.00128, -0.00901, 0.00694, -0.00095, 0.00095, -0.03416, -0.00095, -0.00095, -0.03416, -0.00128, 0.00901, 0.00694, -0.00901, -0.00128, 0.00694, -0.0, -0.0, 0.01328, -0.01566, 0.0, 0.00425, -0.0, 0.0, 0.01328, -0.01566, 0.0, 0.00425, 0.0, -0.0, 0.01328, 0.01566, -0.0, 0.00425, 0.0, -0.0, 0.01328, 0.0, -0.01566, 0.00425, 0.01566, -0.0, 0.00425, -0.0, 0.01566, 0.00425, 0.0, -0.01566, 0.00425, -0.0, 0.01566, 0.00425]]), "T2ux":array([[0.0, -0.0, 0.0, 0.00042, -0.00043, -0.0015, 0.00042, 0.00043, -0.0015, 0.07565, 0.08617, 0.0, -0.00042, 0.0015, 0.00043, 0.00042, 0.00043, 0.0015, -0.07565, 0.0, 0.08617, 0.00042, -0.00043, 0.0015, -0.00042, 0.0015, -0.00043, -0.0, 0.00141, -0.00141, 0.00042, -0.00043, -0.0015, -0.07565, 0.0, -0.08617, 0.07565, -0.08617, -0.0, -0.00042, -0.0015, 0.00043, 0.0, 0.0, -0.0, -0.00042, 0.0015, 0.00043, -0.07565, -0.0, 0.08617, -0.0, 0.00141, 0.00141, -0.0, -0.00141, 0.00141, -0.0, -0.00141, -0.00141, 0.0, 0.0, -0.0, -0.00042, -0.0015, 0.00043, 0.0, 0.00141, -0.00141, 0.0, -0.0, 0.0, -0.00042, -0.0015, -0.00043, 0.00042, 0.00043, -0.0015, 0.07565, 0.08617, 0.0, -0.07565, 0.0, -0.08617, 0.07565, -0.08617, -0.0, 0.0, 0.00141, 0.00141, 0.0, -0.00141, 0.00141, -0.00042, -0.0015, -0.00043, 0.00042, 0.00043, 0.0015, 0.00042, -0.00043, 0.0015, -0.00042, 0.0015, -0.00043, 0.0, -0.00141, -0.00141, -0.00427, 0.00281, 0.0, 0.00427, -0.0, -0.00281, -0.00427, -0.00281, 0.0, 0.00427, -0.0, -0.00281, -0.00427, 0.00281, 0.0, 0.00427, -0.0, 0.00281, -0.00427, -0.00281, -0.0, -0.0, 0.0, -0.0, 0.00427, -0.0, 0.00281, -0.0, -0.0, 0.0, -0.0, -0.0, 0.0, -0.0, 0.0, -0.0], [0.0, 0.0, 0.0, 0.00193, 0.004, -0.00262, 0.00193, -0.004, -0.00262, -0.04156, 0.033, 0.0, -0.00193, 0.00262, -0.004, 0.00193, -0.004, 0.00262, 0.04156, -0.0, 0.033, 0.00193, 0.004, 0.00262, -0.00193, 0.00262, 0.004, 0.0, 0.00178, -0.00178, 0.00193, 0.004, -0.00262, 0.04156, 0.0, -0.033, -0.04156, -0.033, -0.0, -0.00193, -0.00262, -0.004, -0.0, -0.0, 0.0, -0.00193, 0.00262, -0.004, 0.04156, -0.0, 0.033, 0.0, 0.00178, 0.00178, 0.0, -0.00178, 0.00178, 0.0, -0.00178, -0.00178, -0.0, -0.0, -0.0, -0.00193, -0.00262, -0.004, -0.0, 0.00178, -0.00178, -0.0, 0.0, -0.0, -0.00193, -0.00262, 0.004, 0.00193, -0.004, -0.00262, -0.04156, 0.033, 0.0, 0.04156, -0.0, -0.033, -0.04156, -0.033, -0.0, -0.0, 0.00178, 0.00178, -0.0, -0.00178, 0.00178, -0.00193, -0.00262, 0.004, 0.00193, -0.004, 0.00262, 0.00193, 0.004, 0.00262, -0.00193, 0.00262, 0.004, -0.0, -0.00178, -0.00178, -0.03798, 0.0453, 0.0, 0.03798, 0.0, -0.0453, -0.03798, -0.0453, -0.0, 0.03798, 0.0, -0.0453, -0.03798, 0.0453, 0.0, 0.03798, -0.0, 0.0453, -0.03798, -0.0453, -0.0, -0.0, -0.0, -0.0, 0.03798, -0.0, 0.0453, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -0.0, -0.0, -0.0]]), "T2uy":array([[-0.0, 0.0816, -0.08056, -0.00178, -0.00029, -0.00021, 0.00178, -0.00029, 0.00021, -0.08056, -0.0816, 0.0, -0.00222, 0.0, 0.00222, 0.00178, -0.00029, -0.00021, -0.0, 0.0, -0.0, -0.00178, -0.00029, 0.00021, -0.00222, -0.0, -0.00222, 0.00021, 0.00029, 0.00178, -0.00178, -0.00029, -0.00021, -0.0, 0.0, 0.0, 0.08056, -0.0816, -0.0, 0.00222, 0.0, -0.00222, 0.0, 0.0816, 0.08056, -0.00222, -0.0, 0.00222, 0.0, -0.0, 0.0, 0.00021, 0.00029, -0.00178, -0.00021, 0.00029, 0.00178, -0.00021, 0.00029, -0.00178, 0.0, 0.0816, -0.08056, 0.00222, -0.0, -0.00222, 0.00021, 0.00029, 0.00178, 0.0, 0.0816, 0.08056, 0.00222, -0.0, 0.00222, 0.00178, -0.00029, 0.00021, -0.08056, -0.0816, 0.0, 0.0, -0.0, -0.0, 0.08056, -0.0816, -0.0, 0.00021, 0.00029, -0.00178, -0.00021, 0.00029, 0.00178, 0.00222, 0.0, 0.00222, 0.00178, -0.00029, -0.00021, -0.00178, -0.00029, 0.00021, -0.00222, 0.0, -0.00222, -0.00021, 0.00029, -0.00178, -0.00348, 0.00075, -0.0, 0.0, 0.0, 0.0, 0.00348, 0.00075, -0.0, -0.0, -0.0, -0.0, -0.00348, 0.00075, 0.0, -0.0, 0.0, -0.0, 0.00348, 0.00075, 0.0, 0.0, -0.00075, 0.00348, 0.0, -0.0, 0.0, 0.0, -0.00075, -0.00348, -0.0, -0.00075, 0.00348, -0.0, -0.00075, -0.00348], [0.0, -0.05357, -0.05271, -0.00079, 0.00085, -0.00232, 0.00079, 0.00085, 0.00232, -0.05271, 0.05357, 0.0, -0.00159, 0.0, 0.00159, 0.00079, 0.00085, -0.00232, 0.0, -0.0, -0.0, -0.00079, 0.00085, 0.00232, -0.00159, -0.0, -0.00159, 0.00232, -0.00085, 0.00079, -0.00079, 0.00085, -0.00232, -0.0, 0.0, 0.0, 0.05271, 0.05357, -0.0, 0.00159, -0.0, -0.00159, -0.0, -0.05357, 0.05271, -0.00159, -0.0, 0.00159, -0.0, 0.0, 0.0, 0.00232, -0.00085, -0.00079, -0.00232, -0.00085, 0.00079, -0.00232, -0.00085, -0.00079, -0.0, -0.05357, -0.05271, 0.00159, 0.0, -0.00159, 0.00232, -0.00085, 0.00079, -0.0, -0.05357, 0.05271, 0.00159, 0.0, 0.00159, 0.00079, 0.00085, 0.00232, -0.05271, 0.05357, 0.0, 0.0, -0.0, -0.0, 0.05271, 0.05357, -0.0, 0.00232, -0.00085, -0.00079, -0.00232, -0.00085, 0.00079, 0.00159, -0.0, 0.00159, 0.00079, 0.00085, -0.00232, -0.00079, 0.00085, 0.00232, -0.00159, 0.0, -0.00159, -0.00232, -0.00085, -0.00079, -0.0259, 0.03359, 0.0, 0.0, 0.0, -0.0, 0.0259, 0.03359, -0.0, -0.0, -0.0, 0.0, -0.0259, 0.03359, 0.0, -0.0, 0.0, 0.0, 0.0259, 0.03359, -0.0, -0.0, -0.03359, 0.0259, 0.0, -0.0, -0.0, -0.0, -0.03359, -0.0259, -0.0, -0.03359, 0.0259, -0.0, -0.03359, -0.0259]]), "T2uz":array([[-0.0, 0.08867, -0.0771, -5e-05, 5e-05, 0.0, -5e-05, -5e-05, 0.0, 0.0, -0.0, 0.0, 0.00096, -0.00104, -0.00101, 5e-05, 5e-05, -0.0, -0.08867, 0.0, 0.0771, 5e-05, -5e-05, -0.0, -0.00096, 0.00104, -0.00101, 0.00104, -0.00096, 0.00101, -5e-05, 5e-05, -0.0, 0.08867, -0.0, 0.0771, 0.0, -0.0, 0.0, 0.00096, 0.00104, -0.00101, -0.0, -0.08867, -0.0771, 0.00096, -0.00104, -0.00101, -0.08867, 0.0, 0.0771, -0.00104, 0.00096, 0.00101, -0.00104, -0.00096, 0.00101, 0.00104, 0.00096, 0.00101, 0.0, 0.08867, -0.0771, 0.00096, 0.00104, -0.00101, 0.00104, -0.00096, 0.00101, -0.0, -0.08867, -0.0771, -0.00096, -0.00104, -0.00101, -5e-05, -5e-05, -0.0, -0.0, 0.0, -0.0, 0.08867, -0.0, 0.0771, -0.0, 0.0, -0.0, -0.00104, 0.00096, 0.00101, -0.00104, -0.00096, 0.00101, -0.00096, -0.00104, -0.00101, 5e-05, 5e-05, 0.0, 5e-05, -5e-05, 0.0, -0.00096, 0.00104, -0.00101, 0.00104, 0.00096, 0.00101, 0.0, 0.0, -0.0, 0.00285, -0.0, -0.00275, -0.0, 0.0, -0.0, 0.00285, 0.0, -0.00275, -0.0, -0.0, 0.0, -0.00285, 0.0, -0.00275, 0.0, -0.0, 0.0, -0.0, -0.00285, 0.00275, -0.00285, -0.0, -0.00275, 0.0, 0.00285, 0.00275, 0.0, -0.00285, 0.00275, 0.0, 0.00285, 0.00275], [-0.0, 0.03916, 0.04742, 0.00206, -0.00206, -0.0, 0.00206, 0.00206, -0.0, 0.0, -0.0, 0.0, 0.00083, -0.00634, 0.00442, -0.00206, -0.00206, 0.0, -0.03916, 0.0, -0.04742, -0.00206, 0.00206, -0.0, -0.00083, 0.00634, 0.00442, 0.00634, -0.00083, -0.00442, 0.00206, -0.00206, 0.0, 0.03916, 0.0, -0.04742, -0.0, -0.0, 0.0, 0.00083, 0.00634, 0.00442, -0.0, -0.03916, 0.04742, 0.00083, -0.00634, 0.00442, -0.03916, 0.0, -0.04742, -0.00634, 0.00083, -0.00442, -0.00634, -0.00083, -0.00442, 0.00634, 0.00083, -0.00442, 0.0, 0.03916, 0.04742, 0.00083, 0.00634, 0.00442, 0.00634, -0.00083, -0.00442, 0.0, -0.03916, 0.04742, -0.00083, -0.00634, 0.00442, 0.00206, 0.00206, 0.0, -0.0, 0.0, -0.0, 0.03916, -0.0, -0.04742, 0.0, 0.0, -0.0, -0.00634, 0.00083, -0.00442, -0.00634, -0.00083, -0.00442, -0.00083, -0.00634, 0.00442, -0.00206, -0.00206, -0.0, -0.00206, 0.00206, 0.0, -0.00083, 0.00634, 0.00442, 0.00634, 0.00083, -0.00442, 0.0, -0.0, 0.0, 0.0351, -0.0, -0.02822, 0.0, 0.0, 0.0, 0.0351, -0.0, -0.02822, -0.0, 0.0, -0.0, -0.0351, 0.0, -0.02822, -0.0, -0.0, -0.0, -0.0, -0.0351, 0.02822, -0.0351, 0.0, -0.02822, 0.0, 0.0351, 0.02822, -0.0, -0.0351, 0.02822, 0.0, 0.0351, 0.02822]])}, database_path = 'keggin_scsd_20220824.pkl', maxdist = 2.0)
model_objs_dict['betacarotene'] = scsd_model('betacarotene', array([['13.6587', '-1.3537', '0.0', 'C'], ['13.4511', '-2.8003', '0.0', 'C'], ['12.3497', '-3.3733', '0.0', 'C'], ['10.9864', '-2.6659', '0.0', 'C'], ['11.1301', '-1.1288', '0.0', 'C'], ['12.3684', '-0.5703', '0.0', 'C'], ['10.0689', '-0.2872', '0.0', 'C'], ['8.7467', '-0.5395', '0.0', 'C'], ['7.7507', '0.3787', '0.0', 'C'], ['6.4774', '-0.0783', '0.0', 'C'], ['5.3393', '0.6398', '0.0', 'C'], ['4.1174', '0.0711', '0.0', 'C'], ['2.9342', '0.7264', '0.0', 'C'], ['1.8034', '-0.0163', '0.0', 'C'], ['0.5287', '0.4179', '0.0', 'C'], ['-0.5287', '-0.4179', '0.0', 'C'], ['-1.8034', '0.0163', '0.0', 'C'], ['-2.9342', '-0.7264', '0.0', 'C'], ['-4.1174', '-0.0711', '0.0', 'C'], ['-5.3393', '-0.6398', '0.0', 'C'], ['-6.4774', '0.0783', '0.0', 'C'], ['-7.7507', '-0.3787', '0.0', 'C'], ['-8.7467', '0.5395', '0.0', 'C'], ['-10.0689', '0.2872', '0.0', 'C'], ['-11.1301', '1.1288', '0.0', 'C'], ['-12.3684', '0.5703', '0.0', 'C'], ['10.2664', '-3.1714', '1.2672', 'C'], ['10.2664', '-3.1714', '-1.2672', 'C'], ['12.6651', '0.92', '0.0', 'C'], ['8.0868', '1.8605', '0.0', 'C'], ['2.9206', '2.2447', '0.0', 'C'], ['-2.9206', '-2.2447', '0.0', 'C'], ['-8.0868', '-1.8605', '0.0', 'C'], ['-13.6587', '1.3537', '0.0', 'C'], ['-13.4511', '2.8003', '0.0', 'C'], ['-12.3497', '3.3733', '0.0', 'C'], ['-10.9864', '2.6659', '0.0', 'C'], ['-10.2664', '3.1714', '1.2672', 'C'], ['-10.2664', '3.1714', '-1.2672', 'C'], ['-12.6651', '-0.92', '0.0', 'C']]), 'C2h' , pca = {"Ag":array([[-0.03437, 0.02619, -0.0, -0.03896, 0.01846, -0.0, -0.0398, 0.00073, 0.0, -0.03187, -0.00661, -0.0, -0.02601, 0.00314, -0.0, -0.02668, 0.01768, -0.0, -0.01939, -0.0039, -0.0, -0.01634, 0.00403, -0.0, -0.0102, -0.00355, -0.0, -0.00951, 0.00516, -0.0, -0.00605, -0.00626, -0.0, -0.00569, 0.00218, -0.0, -0.00102, -0.00942, -0.0, -0.00233, 0.0035, -0.0, 0.00083, -0.00619, -0.0, -0.00083, 0.00619, -0.0, 0.00233, -0.0035, -0.0, 0.00102, 0.00942, -0.0, 0.00569, -0.00218, -0.0, 0.00605, 0.00626, -0.0, 0.00951, -0.00516, -0.0, 0.0102, 0.00355, -0.0, 0.01634, -0.00403, -0.0, 0.01939, 0.0039, -0.0, 0.02601, -0.00314, -0.0, 0.02668, -0.01768, -0.0, -0.03088, -0.01566, 0.01141, -0.03088, -0.01566, -0.01141, -0.02007, 0.02638, -0.0, -0.00484, -0.02096, -0.0, 0.00502, -0.03544, -0.0, -0.00502, 0.03544, -0.0, 0.00484, 0.02096, -0.0, 0.03437, -0.02619, -0.0, 0.03896, -0.01846, -0.0, 0.0398, -0.00073, -0.0, 0.03187, 0.00661, -0.0, 0.03088, 0.01566, 0.01141, 0.03088, 0.01566, -0.01141, 0.02007, -0.02638, -0.0], [-0.03678, -0.0197, -0.0, -0.02222, 0.01097, -0.0, -0.00639, 0.03484, -0.0, 0.0009, 0.03558, -0.0, -0.01422, 0.00421, 0.0, -0.0309, -0.02085, 0.0, -0.01032, 0.00145, 0.0, -0.00914, 0.00077, 0.0, -0.00515, -0.00083, 0.0, -0.00571, -0.00011, 0.0, -0.00285, -0.00142, 0.0, -0.00198, -0.00196, 0.0, -7e-05, -0.0034, 0.0, -0.00133, -0.00082, 0.0, -0.0003, -0.00027, 0.0, 0.0003, 0.00027, 0.0, 0.00133, 0.00082, 0.0, 7e-05, 0.0034, 0.0, 0.00198, 0.00196, 0.0, 0.00285, 0.00142, 0.0, 0.00571, 0.00011, 0.0, 0.00515, 0.00083, 0.0, 0.00914, -0.00077, 0.0, 0.01032, -0.00145, 0.0, 0.01422, -0.00421, 0.0, 0.0309, 0.02085, 0.0, 0.01206, 0.05178, -0.03132, 0.01206, 0.05178, 0.03132, -0.04564, -0.05263, 0.0, -0.00058, -0.00297, 0.0, 0.00428, -0.00772, 0.0, -0.00428, 0.00772, 0.0, 0.00058, 0.00297, 0.0, 0.03678, 0.0197, 0.0, 0.02222, -0.01097, 0.0, 0.00639, -0.03484, 0.0, -0.0009, -0.03558, 0.0, -0.01206, -0.05178, -0.03132, -0.01206, -0.05178, 0.03132, 0.04564, 0.05263, 0.0], [0.01182, 0.04775, 0.0, 0.01368, 0.05242, -0.0, 0.01836, 0.04197, 0.0, 0.01482, 0.02281, 0.0, 0.0116, 0.0184, -0.0, 0.01006, 0.03005, -0.0, 0.00992, 0.00128, -0.0, 0.00687, -0.00754, -0.0, 0.00485, -0.02269, -0.0, 0.00507, -0.0241, -0.0, 0.00295, -0.0186, -0.0, 0.00283, -0.02053, -0.0, 0.00134, -0.01496, -0.0, 0.00156, -0.01554, -0.0, -0.0005, 0.00041, -0.0, 0.0005, -0.00041, -0.0, -0.00156, 0.01554, -0.0, -0.00134, 0.01496, -0.0, -0.00283, 0.02053, -0.0, -0.00295, 0.0186, -0.0, -0.00507, 0.0241, -0.0, -0.00485, 0.02269, -0.0, -0.00687, 0.00754, -0.0, -0.00992, -0.00128, -0.0, -0.0116, -0.0184, -0.0, -0.01006, -0.03005, -0.0, 0.01462, 0.01522, 0.0059, 0.01462, 0.01522, -0.0059, 0.00664, 0.02629, -0.0, 0.0026, -0.03711, -0.0, -0.00043, -0.0088, -0.0, 0.00043, 0.0088, -0.0, -0.0026, 0.03711, -0.0, -0.01182, -0.04775, -0.0, -0.01368, -0.05242, -0.0, -0.01836, -0.04197, -0.0, -0.01482, -0.02281, -0.0, -0.01462, -0.01522, 0.0059, -0.01462, -0.01522, -0.0059, -0.00664, -0.02629, -0.0]]), "Bg":array([[0.0, -0.0, 0.05626, -0.0, -0.0, 0.02447, -0.0, -0.0, -0.01325, 0.0, 0.0, -0.02673, 0.0, -0.0, 0.00646, 0.0, -0.0, 0.04388, 0.0, -0.0, -0.00361, 0.0, -0.0, -0.00028, 0.0, -0.0, -0.01061, 0.0, -0.0, -0.00224, 0.0, -0.0, -0.00819, 0.0, -0.0, -0.00136, -0.0, 0.0, -0.0076, -0.0, -0.0, 0.00063, 0.0, -0.0, -0.00355, -0.0, 0.0, 0.00355, 0.0, 0.0, -0.00063, 0.0, -0.0, 0.0076, -0.0, 0.0, 0.00136, 0.0, 0.0, 0.00819, 0.0, 0.0, 0.00224, 0.0, 0.0, 0.01061, 0.0, 0.0, 0.00028, 0.0, 0.0, 0.00361, 0.0, 0.0, -0.00646, 0.0, 0.0, -0.04388, -0.01858, -0.02768, -0.05096, 0.01858, 0.02768, -0.05096, 0.0, -0.0, 0.07626, 0.0, -0.0, -0.03092, 0.0, -0.0, -0.0231, -0.0, 0.0, 0.0231, 0.0, 0.0, 0.03092, 0.0, 0.0, -0.05626, 0.0, 0.0, -0.02447, 0.0, 0.0, 0.01325, 0.0, -0.0, 0.02673, -0.01858, -0.02768, 0.05096, 0.01858, 0.02768, 0.05096, -0.0, 0.0, -0.07626], [-0.0, 0.0, 0.04881, -0.0, -0.0, 0.06634, 0.0, 0.0, 0.06463, -0.0, -0.0, 0.04562, 0.0, 0.0, 0.0287, 0.0, 0.0, 0.02935, 0.0, 0.0, 0.0125, 0.0, 0.0, -0.00323, 0.0, 0.0, -0.01163, 0.0, -0.0, -0.01804, 0.0, 0.0, -0.01449, 0.0, -0.0, -0.01452, -0.0, -0.0, -0.00691, -0.0, 0.0, -0.00793, -0.0, 0.0, -0.0009, 0.0, -0.0, 0.0009, 0.0, -0.0, 0.00793, 0.0, 0.0, 0.00691, -0.0, 0.0, 0.01452, 0.0, -0.0, 0.01449, 0.0, 0.0, 0.01804, 0.0, -0.0, 0.01163, 0.0, -0.0, 0.00323, 0.0, -0.0, -0.0125, 0.0, -0.0, -0.0287, 0.0, -0.0, -0.02935, 0.01248, 0.00104, 0.04566, -0.01248, -0.00104, 0.04566, -0.0, -0.0, 0.01086, 0.0, -0.0, -0.01066, -0.0, -0.0, 0.00252, 0.0, 0.0, -0.00252, 0.0, 0.0, 0.01066, 0.0, -0.0, -0.04881, 0.0, 0.0, -0.06634, 0.0, -0.0, -0.06463, 0.0, 0.0, -0.04562, 0.01248, 0.00104, -0.04566, -0.01248, -0.00104, -0.04566, 0.0, 0.0, -0.01086], [-0.0, -0.0, 0.01969, 0.0, 0.0, 0.03612, -0.0, 0.0, 0.01114, -0.0, 0.0, -0.00665, -0.0, -0.0, -0.0192, -0.0, -0.0, -0.01267, -0.0, -0.0, -0.03972, -0.0, 0.0, 0.0065, -0.0, -0.0, -0.00981, -0.0, -0.0, 0.02925, -0.0, -0.0, 0.03058, 0.0, -0.0, 0.03697, -0.0, -0.0, 0.03354, -0.0, -0.0, 0.01964, -0.0, -0.0, 0.01114, 0.0, 0.0, -0.01114, 0.0, 0.0, -0.01964, 0.0, 0.0, -0.03354, -0.0, 0.0, -0.03697, -0.0, 0.0, -0.03058, -0.0, 0.0, -0.02925, -0.0, 0.0, 0.00981, -0.0, -0.0, -0.0065, -0.0, 0.0, 0.03972, -0.0, 0.0, 0.0192, -0.0, 0.0, 0.01267, -0.00605, 0.00708, -0.00952, 0.00605, -0.00708, -0.00952, 0.0, -0.0, -0.03842, -0.0, -0.0, -0.07195, -0.0, 0.0, 0.04485, 0.0, -0.0, -0.04485, -0.0, 0.0, 0.07195, -0.0, 0.0, -0.01969, -0.0, -0.0, -0.03612, -0.0, -0.0, -0.01114, -0.0, -0.0, 0.00665, -0.00605, 0.00708, 0.00952, 0.00605, -0.00708, 0.00952, -0.0, 0.0, 0.03842]]), "Au":array([[0.0, 0.0, -0.00608, -0.0, 0.0, -0.03841, 0.0, 0.0, -0.05869, -0.0, -0.0, -0.05174, 0.0, 0.0, -0.01831, -0.0, 0.0, 0.00247, 0.0, 0.0, -0.00935, 0.0, 0.0, -0.00445, 0.0, -0.0, 0.00289, 0.0, -0.0, 0.00748, 0.0, -0.0, 0.01177, -0.0, -0.0, 0.01426, -0.0, -0.0, 0.01682, -0.0, -0.0, 0.01688, 0.0, -0.0, 0.01802, 0.0, -0.0, 0.01802, -0.0, -0.0, 0.01688, -0.0, -0.0, 0.01682, -0.0, -0.0, 0.01426, 0.0, 0.0, 0.01177, 0.0, -0.0, 0.00748, 0.0, -0.0, 0.00289, 0.0, 0.0, -0.00445, 0.0, -0.0, -0.00935, 0.0, 0.0, -0.01831, -0.0, 0.0, 0.00247, -0.0162, -0.01749, -0.06595, 0.0162, 0.01749, -0.06595, -0.0, -0.0, 0.03588, 0.0, -0.0, 0.00444, -0.0, -0.0, 0.01883, -0.0, -0.0, 0.01883, 0.0, -0.0, 0.00444, 0.0, 0.0, -0.00608, 0.0, 0.0, -0.03841, 0.0, 0.0, -0.05869, 0.0, 0.0, -0.05174, 0.0162, 0.01749, -0.06595, -0.0162, -0.01749, -0.06595, -0.0, 0.0, 0.03588], [0.0, -0.0, -0.0676, 0.0, -0.0, -0.04462, -0.0, -0.0, -0.01675, -0.0, 0.0, -0.00393, 0.0, 0.0, -0.02703, 0.0, -0.0, -0.05651, 0.0, 0.0, -0.0169, 0.0, -0.0, -0.00685, 0.0, 0.0, 0.00139, 0.0, -0.0, 0.00731, 0.0, 0.0, 0.01281, 0.0, 0.0, 0.01692, 0.0, -0.0, 0.02105, -0.0, -0.0, 0.02185, 0.0, 0.0, 0.02334, -0.0, 0.0, 0.02334, 0.0, -0.0, 0.02185, -0.0, -0.0, 0.02105, -0.0, 0.0, 0.01692, 0.0, -0.0, 0.01281, 0.0, -0.0, 0.00731, 0.0, -0.0, 0.00139, 0.0, -0.0, -0.00685, 0.0, 0.0, -0.0169, 0.0, -0.0, -0.02703, -0.0, -0.0, -0.05651, 0.012, 0.01851, 0.01487, -0.012, -0.01851, 0.01487, 0.0, 0.0, -0.08117, 0.0, 0.0, 0.00292, -0.0, 0.0, 0.02377, 0.0, 0.0, 0.02377, 0.0, 0.0, 0.00292, 0.0, -0.0, -0.0676, 0.0, -0.0, -0.04462, 0.0, 0.0, -0.01675, 0.0, 0.0, -0.00393, -0.012, -0.01851, 0.01487, 0.012, 0.01851, 0.01487, 0.0, 0.0, -0.08117], [0.0, -0.0, 0.03559, -0.0, 0.0, 0.04653, 0.0, 0.0, 0.02469, -0.0, 0.0, -0.00089, -0.0, 0.0, -0.01117, 0.0, 0.0, 0.00327, -0.0, 0.0, -0.03699, -0.0, -0.0, -0.01329, -0.0, -0.0, -0.03475, -0.0, -0.0, 0.00052, -0.0, -0.0, 0.00446, -0.0, -0.0, 0.02835, -0.0, 0.0, 0.02984, -0.0, 0.0, 0.04185, 0.0, -0.0, 0.04395, -0.0, -0.0, 0.04395, -0.0, -0.0, 0.04185, -0.0, 0.0, 0.02984, -0.0, -0.0, 0.02835, -0.0, 0.0, 0.00446, -0.0, -0.0, 0.00052, -0.0, 0.0, -0.03475, -0.0, -0.0, -0.01329, -0.0, -0.0, -0.03699, 0.0, 0.0, -0.01117, 0.0, -0.0, 0.00327, -0.0002, -0.00348, -0.00714, 0.0002, 0.00348, -0.00714, 0.0, -0.0, -0.01497, -0.0, -0.0, -0.09636, -0.0, 0.0, 0.01725, -0.0, 0.0, 0.01725, -0.0, 0.0, -0.09636, -0.0, -0.0, 0.03559, -0.0, -0.0, 0.04653, -0.0, 0.0, 0.02469, -0.0, 0.0, -0.00089, 0.0002, 0.00348, -0.00714, -0.0002, -0.00348, -0.00714, 0.0, -0.0, -0.01497]]), "Bu":array([[-0.00317, -0.03953, -0.0, -0.00703, -0.04692, -0.0, -0.00855, -0.04373, 0.0, -0.00688, -0.03161, -0.0, -0.00277, -0.02366, 0.0, -0.00141, -0.02763, 0.0, -0.00035, -0.0116, 0.0, 0.00013, -0.00564, 0.0, 0.00242, 0.00545, 0.0, 0.00205, 0.00909, 0.0, 0.00245, 0.01609, 0.0, 0.00208, 0.01946, 0.0, 0.00226, 0.02674, 0.0, 0.00199, 0.02706, 0.0, 0.0022, 0.02764, 0.0, 0.0022, 0.02764, 0.0, 0.00199, 0.02706, 0.0, 0.00226, 0.02674, 0.0, 0.00208, 0.01946, 0.0, 0.00245, 0.01609, 0.0, 0.00205, 0.00909, 0.0, 0.00242, 0.00545, 0.0, 0.00013, -0.00564, 0.0, -0.00035, -0.0116, 0.0, -0.00277, -0.02366, 0.0, -0.00141, -0.02763, 0.0, -0.0083, -0.02977, 0.00029, -0.0083, -0.02977, -0.00029, 0.00211, -0.02043, 0.0, 0.00515, 0.01256, 0.0, 0.00281, 0.03362, 0.0, 0.00281, 0.03362, 0.0, 0.00515, 0.01256, 0.0, -0.00317, -0.03953, 0.0, -0.00703, -0.04692, 0.0, -0.00855, -0.04373, 0.0, -0.00688, -0.03161, 0.0, -0.0083, -0.02977, -0.00029, -0.0083, -0.02977, 0.00029, 0.00211, -0.02043, 0.0], [-0.01811, -0.04126, 0.0, -0.00341, -0.00996, 0.0, 0.01127, 0.02106, -0.0, 0.01566, 0.02937, -0.0, -0.0012, -0.00461, 0.0, -0.01585, -0.036, 0.0, -0.0011, -0.0026, 0.0, 0.00081, -0.00022, 0.0, 0.00095, 0.00148, 0.0, 0.00119, 0.00182, 0.0, 0.00114, 0.00322, 0.0, 0.00125, 0.00415, 0.0, 0.00114, 0.00545, 0.0, 0.00116, 0.00582, 0.0, 0.00108, 0.00673, 0.0, 0.00108, 0.00673, 0.0, 0.00116, 0.00582, 0.0, 0.00114, 0.00545, 0.0, 0.00125, 0.00415, 0.0, 0.00114, 0.00322, 0.0, 0.00119, 0.00182, 0.0, 0.00095, 0.00148, 0.0, 0.00081, -0.00022, 0.0, -0.0011, -0.0026, 0.0, -0.0012, -0.00461, 0.0, -0.01585, -0.036, 0.0, 0.02666, 0.05188, -0.03383, 0.02666, 0.05188, 0.03383, -0.03068, -0.0685, 0.0, 0.00056, 0.00303, 0.0, 0.00094, 0.0066, 0.0, 0.00094, 0.0066, 0.0, 0.00056, 0.00303, 0.0, -0.01811, -0.04126, 0.0, -0.00341, -0.00996, 0.0, 0.01127, 0.02106, 0.0, 0.01566, 0.02937, 0.0, 0.02666, 0.05188, 0.03383, 0.02666, 0.05188, -0.03383, -0.03068, -0.0685, 0.0], [0.03935, 0.01184, -0.0, 0.055, 0.01342, 0.0, 0.04321, 0.0061, 0.0, 0.03352, 0.0001, 0.0, 0.02333, 0.00046, 0.0, 0.02681, 0.00579, 0.0, 0.01006, -0.00485, 0.0, 0.00744, -0.01173, 0.0, -0.00425, -0.01444, 0.0, -0.00279, -0.01734, 0.0, -0.00449, -0.0002, 0.0, -0.00576, 0.00564, 0.0, -0.00624, 0.02477, 0.0, -0.00831, 0.02541, 0.0, -0.00647, 0.02753, 0.0, -0.00647, 0.02753, 0.0, -0.00831, 0.02541, 0.0, -0.00624, 0.02477, 0.0, -0.00576, 0.00564, 0.0, -0.00449, -0.0002, 0.0, -0.00279, -0.01734, 0.0, -0.00425, -0.01444, 0.0, 0.00744, -0.01173, 0.0, 0.01006, -0.00485, 0.0, 0.02333, 0.00046, 0.0, 0.02681, 0.00579, 0.0, 0.03386, -0.00313, 0.00087, 0.03386, -0.00313, -0.00087, 0.01896, 0.00636, 0.0, -0.01712, -0.01478, 0.0, -0.00416, 0.04419, 0.0, -0.00416, 0.04419, 0.0, -0.01712, -0.01478, 0.0, 0.03935, 0.01184, 0.0, 0.055, 0.01342, 0.0, 0.04321, 0.0061, 0.0, 0.03352, 0.0001, 0.0, 0.03386, -0.00313, -0.00087, 0.03386, -0.00313, 0.00087, 0.01896, 0.00636, 0.0]])})#, database_path = 'betacarotene_scsd_20220413.pkl', maxdist = 1.65)

model_objs_dict['ptpop']                     = scsd_model('ptpop', model_ptpop, 'D4h', maxdist = 2.4)
model_objs_dict['tmbodipy']                  = scsd_model('tmbodipy', model_tmbodipy, 'C2v')
model_objs_dict['coronene']                  = scsd_model('coronene',model_coron,'D6h')
model_objs_dict['bcp']                       = scsd_model('bcp',model_bcp,'D3h')
model_objs_dict['dpm']                       = scsd_model('dpm',model_dpm,'C2v')
model_objs_dict['macac']                     = scsd_model('macac',model_macac,'D2h', maxdist = 2.2)
model_objs_dict['tetrabenzocoronene']        = scsd_model('tetrabenzocoronene',model_tbcor,'D2h')
model_objs_dict['azatriangulene']            = scsd_model('azatriangulene', model_azatrip,'D3h')
model_objs_dict['texaphyrin']                = scsd_model('texaphyrin',model_texaphyrin,'C2v')
model_objs_dict['tetrathienodibenzocoronene']= scsd_model('corenothiophene',model_corenothiophene,'D2h', maxdist = 2.0, mondrian_limits = [0,0])
model_objs_dict['HEM_pdb'] = scsd_model('HEM_pdb', array([['3.4229', '-0.0', '-0.0', 'C'], ['0.0', '3.4229', '0.0', 'C'], ['-3.4229', '0.0', '0.0', 'C'], ['0.0', '-3.4229', '-0.0', 'C'], ['2.7768', '1.2168', '-0.0', 'C'], ['3.4332', '2.4924', '-0.0', 'C'], ['2.4924', '3.4332', '0.0', 'C'], ['1.2168', '2.7768', '0.0', 'C'], ['2.6568', '4.954', '0.0', 'C'], ['4.954', '2.6568', '-0.0', 'C'], ['-1.2168', '2.7768', '0.0', 'C'], ['-2.4924', '3.4332', '0.0', 'C'], ['-3.4332', '2.4924', '0.0', 'C'], ['-2.7768', '1.2168', '0.0', 'C'], ['-2.6568', '4.954', '0.0', 'C'], ['-4.954', '2.6568', '0.0', 'C'], ['-2.7768', '-1.2168', '0.0', 'C'], ['-3.4332', '-2.4924', '0.0', 'C'], ['-2.4924', '-3.4332', '-0.0', 'C'], ['-1.2168', '-2.7768', '-0.0', 'C'], ['-4.954', '-2.6568', '0.0', 'C'], ['-2.6568', '-4.954', '-0.0', 'C'], ['1.2168', '-2.7768', '-0.0', 'C'], ['2.4924', '-3.4332', '-0.0', 'C'], ['3.4332', '-2.4924', '-0.0', 'C'], ['2.7768', '-1.2168', '-0.0', 'C'], ['2.6568', '-4.954', '-0.0', 'C'], ['4.954', '-2.6568', '-0.0', 'C'], ['0.0', '0.0', '0.0', 'FE'], ['1.4106', '1.4106', '-0.0', 'N'], ['-1.4106', '1.4106', '0.0', 'N'], ['-1.4106', '-1.4106', '0.0', 'N'], ['1.4106', '-1.4106', '-0.0', 'N']]), 'D4h' , pca = {"A1g":array([[0.03806, -0.0, 0.0, 0.0, 0.03806, 0.0, -0.03806, 0.0, 0.0, -0.0, -0.03806, -0.0, 0.0013, -0.00612, -0.0, -0.00497, -0.03708, -0.0, -0.03708, -0.00497, -0.0, -0.00612, 0.0013, 0.0, -0.05019, 0.01869, 0.0, 0.01869, -0.05019, -0.0, 0.00612, 0.0013, 0.0, 0.03708, -0.00497, -0.0, 0.00497, -0.03708, -0.0, -0.0013, -0.00612, -0.0, 0.05019, 0.01869, 0.0, -0.01869, -0.05019, -0.0, -0.0013, 0.00612, 0.0, 0.00497, 0.03708, 0.0, 0.03708, 0.00497, 0.0, 0.00612, -0.0013, -0.0, -0.01869, 0.05019, 0.0, 0.05019, -0.01869, -0.0, -0.00612, -0.0013, -0.0, -0.03708, 0.00497, 0.0, -0.00497, 0.03708, 0.0, 0.0013, 0.00612, 0.0, -0.05019, -0.01869, -0.0, 0.01869, 0.05019, 0.0, 0.0, 0.0, 0.0, -0.01237, -0.01237, -0.0, 0.01237, -0.01237, -0.0, 0.01237, 0.01237, 0.0, -0.01237, 0.01237, 0.0], [-0.02504, 0.0, -0.0, -0.0, -0.02504, -0.0, 0.02504, -0.0, -0.0, 0.0, 0.02504, 0.0, -0.0232, -0.011, -0.0, -0.03354, -0.00946, -0.0, -0.00946, -0.03354, -0.0, -0.011, -0.0232, -0.0, -0.01102, -0.04187, -0.0, -0.04187, -0.01102, -0.0, 0.011, -0.0232, -0.0, 0.00946, -0.03354, -0.0, 0.03354, -0.00946, -0.0, 0.0232, -0.011, -0.0, 0.01102, -0.04187, -0.0, 0.04187, -0.01102, -0.0, 0.0232, 0.011, 0.0, 0.03354, 0.00946, 0.0, 0.00946, 0.03354, 0.0, 0.011, 0.0232, 0.0, 0.04187, 0.01102, 0.0, 0.01102, 0.04187, 0.0, -0.011, 0.0232, 0.0, -0.00946, 0.03354, 0.0, -0.03354, 0.00946, 0.0, -0.0232, 0.011, 0.0, -0.01102, 0.04187, 0.0, -0.04187, 0.01102, 0.0, 0.0, 0.0, 0.0, -0.01224, -0.01224, -0.0, 0.01224, -0.01224, -0.0, 0.01224, 0.01224, 0.0, -0.01224, 0.01224, 0.0]]), "A2g":array([[-0.0, -0.00831, 0.0, 0.00831, -0.0, 0.0, 0.0, 0.00831, -0.0, -0.00831, 0.0, -0.0, -0.00624, 0.01658, -0.0, 0.03383, -0.02168, -0.0, 0.02168, -0.03383, -0.0, -0.01658, 0.00624, 0.0, -0.0315, 0.00891, 0.0, -0.00891, 0.0315, -0.0, -0.01658, -0.00624, 0.0, 0.02168, 0.03383, -0.0, 0.03383, 0.02168, -0.0, -0.00624, -0.01658, -0.0, -0.0315, -0.00891, 0.0, -0.00891, -0.0315, -0.0, 0.00624, -0.01658, 0.0, -0.03383, 0.02168, 0.0, -0.02168, 0.03383, 0.0, 0.01658, -0.00624, -0.0, 0.00891, -0.0315, 0.0, 0.0315, -0.00891, -0.0, 0.01658, 0.00624, -0.0, -0.02168, -0.03383, 0.0, -0.03383, -0.02168, 0.0, 0.00624, 0.01658, 0.0, 0.0315, 0.00891, -0.0, 0.00891, 0.0315, 0.0, -0.0, -0.0, -0.0, 0.04274, -0.04274, -0.0, 0.04274, 0.04274, -0.0, -0.04274, 0.04274, 0.0, -0.04274, -0.04274, 0.0], [0.0, -0.05013, -0.0, 0.05013, 0.0, 0.0, 0.0, 0.05013, 0.0, -0.05013, -0.0, -0.0, 0.00372, -0.00957, 0.0, -0.03981, 0.02568, 0.0, -0.02568, 0.03981, -0.0, 0.00957, -0.00372, 0.0, 0.00277, 0.01108, 0.0, -0.01108, -0.00277, -0.0, 0.00957, 0.00372, 0.0, -0.02568, -0.03981, -0.0, -0.03981, -0.02568, 0.0, 0.00372, 0.00957, 0.0, 0.00277, -0.01108, 0.0, -0.01108, 0.00277, -0.0, -0.00372, 0.00957, -0.0, 0.03981, -0.02568, -0.0, 0.02568, -0.03981, 0.0, -0.00957, 0.00372, -0.0, 0.01108, 0.00277, 0.0, -0.00277, -0.01108, -0.0, -0.00957, -0.00372, -0.0, 0.02568, 0.03981, 0.0, 0.03981, 0.02568, -0.0, -0.00372, -0.00957, -0.0, -0.00277, 0.01108, -0.0, 0.01108, -0.00277, 0.0, -0.0, -0.0, -0.0, 0.04366, -0.04366, 0.0, 0.04366, 0.04366, 0.0, -0.04366, 0.04366, -0.0, -0.04366, -0.04366, -0.0]]), "B1g":array([[-0.04032, 0.0, 0.0, 0.0, 0.04032, 0.0, 0.04032, -0.0, -0.0, -0.0, -0.04032, -0.0, -0.02803, 0.00358, 0.0, -0.0071, -0.01059, -0.0, 0.01059, 0.0071, -0.0, -0.00358, 0.02803, 0.0, 0.04624, 0.00046, -0.0, -0.00046, -0.04624, -0.0, 0.00358, 0.02803, 0.0, -0.01059, 0.0071, 0.0, 0.0071, -0.01059, -0.0, 0.02803, 0.00358, -0.0, -0.04624, 0.00046, 0.0, 0.00046, -0.04624, -0.0, 0.02803, -0.00358, -0.0, 0.0071, 0.01059, 0.0, -0.01059, -0.0071, 0.0, 0.00358, -0.02803, -0.0, 0.00046, 0.04624, 0.0, -0.04624, -0.00046, 0.0, -0.00358, -0.02803, -0.0, 0.01059, -0.0071, -0.0, -0.0071, 0.01059, 0.0, -0.02803, -0.00358, 0.0, 0.04624, -0.00046, -0.0, -0.00046, 0.04624, 0.0, 0.0, 0.0, 0.0, -0.02487, 0.02487, 0.0, 0.02487, 0.02487, -0.0, 0.02487, -0.02487, -0.0, -0.02487, -0.02487, 0.0], [-0.02644, -0.0, 0.0, 0.0, 0.02644, 0.0, 0.02644, -0.0, -0.0, -0.0, -0.02644, -0.0, -0.02176, 0.0092, 0.0, -0.0235, 0.0191, 0.0, -0.0191, 0.0235, 0.0, -0.0092, 0.02176, 0.0, -0.0333, 0.0331, 0.0, -0.0331, 0.0333, 0.0, 0.0092, 0.02176, 0.0, 0.0191, 0.0235, 0.0, 0.0235, 0.0191, -0.0, 0.02176, 0.0092, -0.0, 0.0333, 0.0331, -0.0, 0.0331, 0.0333, -0.0, 0.02176, -0.0092, -0.0, 0.0235, -0.0191, -0.0, 0.0191, -0.0235, -0.0, 0.0092, -0.02176, -0.0, 0.0331, -0.0333, -0.0, 0.0333, -0.0331, -0.0, -0.0092, -0.02176, -0.0, -0.0191, -0.0235, -0.0, -0.0235, -0.0191, 0.0, -0.02176, -0.0092, 0.0, -0.0333, -0.0331, 0.0, -0.0331, -0.0333, 0.0, -0.0, -0.0, -0.0, -0.01544, 0.01544, 0.0, 0.01544, 0.01544, -0.0, 0.01544, -0.01544, -0.0, -0.01544, -0.01544, 0.0]]), "B2g":array([[-0.0, 0.01813, 0.0, 0.01813, 0.0, -0.0, -0.0, -0.01813, -0.0, -0.01813, -0.0, 0.0, 0.01296, 0.02154, 0.0, 0.02487, 0.02507, 0.0, 0.02507, 0.02487, 0.0, 0.02154, 0.01296, 0.0, 0.03352, 0.02728, 0.0, 0.02728, 0.03352, 0.0, 0.02154, -0.01296, -0.0, 0.02507, -0.02487, -0.0, 0.02487, -0.02507, -0.0, 0.01296, -0.02154, -0.0, 0.03352, -0.02728, -0.0, 0.02728, -0.03352, -0.0, -0.01296, -0.02154, -0.0, -0.02487, -0.02507, -0.0, -0.02507, -0.02487, -0.0, -0.02154, -0.01296, -0.0, -0.02728, -0.03352, -0.0, -0.03352, -0.02728, -0.0, -0.02154, 0.01296, 0.0, -0.02507, 0.02487, 0.0, -0.02487, 0.02507, 0.0, -0.01296, 0.02154, 0.0, -0.03352, 0.02728, 0.0, -0.02728, 0.03352, 0.0, 0.0, 0.0, 0.0, 0.01735, 0.01735, 0.0, 0.01735, -0.01735, -0.0, -0.01735, -0.01735, -0.0, -0.01735, 0.01735, 0.0], [0.0, -0.00472, -0.0, -0.00472, 0.0, -0.0, -0.0, 0.00472, 0.0, 0.00472, -0.0, -0.0, 0.01619, 0.00342, 0.0, 0.0263, -0.00982, -0.0, -0.00982, 0.0263, 0.0, 0.00342, 0.01619, 0.0, -0.05563, 0.03219, 0.0, 0.03219, -0.05563, -0.0, 0.00342, -0.01619, -0.0, -0.00982, -0.0263, -0.0, 0.0263, 0.00982, 0.0, 0.01619, -0.00342, -0.0, -0.05563, -0.03219, -0.0, 0.03219, 0.05563, 0.0, -0.01619, -0.00342, -0.0, -0.0263, 0.00982, 0.0, 0.00982, -0.0263, -0.0, -0.00342, -0.01619, -0.0, -0.03219, 0.05563, 0.0, 0.05563, -0.03219, -0.0, -0.00342, 0.01619, 0.0, 0.00982, 0.0263, 0.0, -0.0263, -0.00982, -0.0, -0.01619, 0.00342, 0.0, 0.05563, 0.03219, 0.0, -0.03219, -0.05563, -0.0, -0.0, -0.0, -0.0, 0.01944, 0.01944, 0.0, 0.01944, -0.01944, -0.0, -0.01944, -0.01944, -0.0, -0.01944, 0.01944, 0.0]]), "A1u":array([[0.0, -0.0, -0.0, 0.0, 0.0, -0.0, 0.0, 0.0, 0.0, -0.0, -0.0, 0.0, -0.0, 0.0, 0.01296, -0.0, -0.0, 0.02509, -0.0, -0.0, -0.02509, 0.0, 0.0, -0.01296, -0.0, 0.0, -0.08695, 0.0, -0.0, 0.08695, -0.0, 0.0, 0.01296, -0.0, 0.0, 0.02509, -0.0, -0.0, -0.02509, 0.0, 0.0, -0.01296, -0.0, 0.0, 0.08695, 0.0, 0.0, -0.08695, 0.0, -0.0, 0.01296, 0.0, -0.0, 0.02509, 0.0, 0.0, -0.02509, -0.0, -0.0, -0.01296, -0.0, 0.0, 0.08695, 0.0, -0.0, -0.08695, 0.0, -0.0, 0.01296, 0.0, -0.0, 0.02509, 0.0, 0.0, -0.02509, -0.0, -0.0, -0.01296, 0.0, -0.0, 0.08695, -0.0, -0.0, -0.08695, 0.0, -0.0, 0.0, -0.0, 0.0, -0.0, -0.0, -0.0, 0.0, 0.0, -0.0, 0.0, 0.0, 0.0, -0.0], [-0.0, -0.0, 0.0, -0.0, 0.0, 0.0, -0.0, -0.0, -0.0, 0.0, -0.0, -0.0, -0.0, 0.0, 0.01385, 0.0, -0.0, 0.08466, 0.0, 0.0, -0.08466, -0.0, -0.0, -0.01385, -0.0, 0.0, 0.02649, -0.0, -0.0, -0.02649, 0.0, 0.0, 0.01385, 0.0, -0.0, 0.08466, 0.0, -0.0, -0.08466, -0.0, -0.0, -0.01385, -0.0, 0.0, -0.02649, -0.0, 0.0, 0.02649, 0.0, -0.0, 0.01385, -0.0, -0.0, 0.08466, -0.0, -0.0, -0.08466, 0.0, 0.0, -0.01385, 0.0, 0.0, -0.02649, 0.0, -0.0, 0.02649, -0.0, -0.0, 0.01385, -0.0, 0.0, 0.08466, -0.0, 0.0, -0.08466, 0.0, 0.0, -0.01385, 0.0, -0.0, -0.02649, 0.0, -0.0, 0.02649, 0.0, -0.0, 0.0, -0.0, 0.0, -0.0, -0.0, 0.0, -0.0, 0.0, -0.0, 0.0, 0.0, -0.0, 0.0]]), "A2u":array([[-0.0, 0.0, -0.02567, -0.0, 0.0, -0.02567, 0.0, 0.0, -0.02567, 0.0, 0.0, -0.02567, -0.0, 0.0, -0.01919, 0.0, 0.0, 0.01502, 0.0, -0.0, 0.01502, 0.0, 0.0, -0.01919, 0.0, 0.0, 0.04746, 0.0, 0.0, 0.04746, 0.0, -0.0, -0.01919, 0.0, 0.0, 0.01502, -0.0, 0.0, 0.01502, 0.0, -0.0, -0.01919, 0.0, -0.0, 0.04746, 0.0, 0.0, 0.04746, 0.0, -0.0, -0.01919, -0.0, -0.0, 0.01502, -0.0, 0.0, 0.01502, -0.0, -0.0, -0.01919, -0.0, -0.0, 0.04746, -0.0, -0.0, 0.04746, -0.0, 0.0, -0.01919, -0.0, -0.0, 0.01502, 0.0, -0.0, 0.01502, -0.0, 0.0, -0.01919, -0.0, 0.0, 0.04746, -0.0, -0.0, 0.04746, 0.0, 0.0, -0.07968, 0.0, 0.0, -0.04106, -0.0, 0.0, -0.04106, -0.0, -0.0, -0.04106, 0.0, 0.0, -0.04106], [-0.0, -0.0, -0.03942, -0.0, 0.0, -0.03942, 0.0, 0.0, -0.03942, -0.0, -0.0, -0.03942, 0.0, 0.0, -0.02787, 0.0, -0.0, -0.00027, 0.0, 0.0, -0.00027, 0.0, -0.0, -0.02787, 0.0, 0.0, 0.02634, -0.0, 0.0, 0.02634, 0.0, 0.0, -0.02787, -0.0, 0.0, -0.00027, 0.0, 0.0, -0.00027, -0.0, -0.0, -0.02787, -0.0, 0.0, 0.02634, -0.0, 0.0, 0.02634, -0.0, -0.0, -0.02787, -0.0, 0.0, -0.00027, -0.0, -0.0, -0.00027, -0.0, 0.0, -0.02787, 0.0, -0.0, 0.02634, -0.0, -0.0, 0.02634, -0.0, -0.0, -0.02787, 0.0, 0.0, -0.00027, -0.0, -0.0, -0.00027, 0.0, 0.0, -0.02787, 0.0, -0.0, 0.02634, 0.0, -0.0, 0.02634, -0.0, -0.0, 0.28979, -0.0, -0.0, -0.02918, 0.0, -0.0, -0.02918, 0.0, 0.0, -0.02918, -0.0, 0.0, -0.02918]]), "B1u":array([[-0.0, -0.0, -0.0, 0.0, -0.0, -0.0, -0.0, 0.0, 0.0, -0.0, 0.0, 0.0, 0.0, -0.0, 0.01401, -0.0, 0.0, 0.04246, -0.0, -0.0, 0.04246, -0.0, 0.0, 0.01401, 0.0, -0.0, 0.06801, -0.0, 0.0, 0.06801, 0.0, -0.0, -0.01401, 0.0, -0.0, -0.04246, 0.0, 0.0, -0.04246, -0.0, -0.0, -0.01401, -0.0, -0.0, -0.06801, -0.0, 0.0, -0.06801, -0.0, 0.0, 0.01401, 0.0, -0.0, 0.04246, 0.0, 0.0, 0.04246, 0.0, -0.0, 0.01401, 0.0, -0.0, 0.06801, -0.0, 0.0, 0.06801, -0.0, 0.0, -0.01401, -0.0, 0.0, -0.04246, -0.0, -0.0, -0.04246, 0.0, 0.0, -0.01401, 0.0, 0.0, -0.06801, 0.0, -0.0, -0.06801, -0.0, 0.0, -0.0, -0.0, -0.0, -0.00105, 0.0, -0.0, 0.00105, 0.0, 0.0, -0.00105, -0.0, 0.0, 0.00105], [-0.0, 0.0, -0.0, -0.0, -0.0, 0.0, 0.0, -0.0, -0.0, 0.0, 0.0, 0.0, -0.0, -0.0, 0.05523, -0.0, -0.0, 0.01777, -0.0, -0.0, 0.01777, -0.0, -0.0, 0.05523, -0.0, -0.0, -0.02201, -0.0, -0.0, -0.02201, 0.0, -0.0, -0.05523, 0.0, -0.0, -0.01777, -0.0, 0.0, -0.01777, -0.0, -0.0, -0.05523, 0.0, -0.0, 0.02201, -0.0, 0.0, 0.02201, 0.0, 0.0, 0.05523, 0.0, 0.0, 0.01777, 0.0, 0.0, 0.01777, 0.0, 0.0, 0.05523, 0.0, 0.0, -0.02201, 0.0, 0.0, -0.02201, -0.0, 0.0, -0.05523, -0.0, 0.0, -0.01777, 0.0, -0.0, -0.01777, 0.0, 0.0, -0.05523, -0.0, 0.0, 0.02201, 0.0, -0.0, 0.02201, 0.0, -0.0, 0.0, 0.0, 0.0, 0.05999, -0.0, 0.0, -0.05999, -0.0, -0.0, 0.05999, 0.0, -0.0, -0.05999]]), "B2u":array([[-0.0, -0.0, 0.05983, -0.0, -0.0, -0.05983, 0.0, -0.0, 0.05983, -0.0, 0.0, -0.05983, 0.0, -0.0, 0.02904, -0.0, -0.0, 0.01901, -0.0, -0.0, -0.01901, -0.0, 0.0, -0.02904, 0.0, 0.0, -0.04703, 0.0, -0.0, 0.04703, 0.0, 0.0, -0.02904, 0.0, 0.0, -0.01901, 0.0, -0.0, 0.01901, -0.0, -0.0, 0.02904, 0.0, 0.0, -0.04703, -0.0, -0.0, 0.04703, -0.0, 0.0, 0.02904, 0.0, -0.0, 0.01901, 0.0, 0.0, -0.01901, 0.0, -0.0, -0.02904, -0.0, 0.0, 0.04703, -0.0, -0.0, -0.04703, -0.0, -0.0, -0.02904, -0.0, -0.0, -0.01901, -0.0, -0.0, 0.01901, 0.0, 0.0, 0.02904, -0.0, -0.0, -0.04703, 0.0, 0.0, 0.04703, -0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -0.0, 0.0, -0.0, -0.0, -0.0, -0.0, 0.0, -0.0, 0.0], [0.0, 0.0, 0.07294, -0.0, 0.0, -0.07294, -0.0, -0.0, 0.07294, -0.0, -0.0, -0.07294, -0.0, 0.0, 0.02494, 0.0, -0.0, -0.00299, -0.0, 0.0, 0.00299, -0.0, -0.0, -0.02494, -0.0, -0.0, 0.0606, -0.0, 0.0, -0.0606, -0.0, -0.0, -0.02494, -0.0, 0.0, 0.00299, -0.0, -0.0, -0.00299, 0.0, -0.0, 0.02494, -0.0, 0.0, 0.0606, -0.0, 0.0, -0.0606, 0.0, -0.0, 0.02494, -0.0, -0.0, -0.00299, 0.0, -0.0, 0.00299, 0.0, 0.0, -0.02494, 0.0, 0.0, -0.0606, -0.0, 0.0, 0.0606, 0.0, 0.0, -0.02494, 0.0, -0.0, 0.00299, 0.0, -0.0, -0.00299, -0.0, 0.0, 0.02494, 0.0, -0.0, 0.0606, 0.0, 0.0, -0.0606, 0.0, 0.0, 0.0, -0.0, -0.0, -0.0, 0.0, -0.0, 0.0, 0.0, 0.0, 0.0, -0.0, 0.0, -0.0]]), "Egx":array([[-0.0, -0.0, -0.0, 0.0, -0.0, 0.07114, -0.0, 0.0, 0.0, -0.0, -0.0, -0.07114, 0.0, -0.0, 0.00582, -0.0, 0.0, -0.02697, 0.0, 0.0, 0.0041, 0.0, -0.0, 0.04379, 0.0, 0.0, -0.0079, -0.0, 0.0, -0.08597, -0.0, 0.0, 0.04379, -0.0, -0.0, 0.0041, 0.0, 0.0, -0.02697, -0.0, -0.0, 0.00582, -0.0, 0.0, -0.0079, -0.0, -0.0, -0.08597, -0.0, 0.0, -0.00582, 0.0, -0.0, 0.02697, -0.0, -0.0, -0.0041, -0.0, 0.0, -0.04379, 0.0, -0.0, 0.08597, -0.0, -0.0, 0.0079, 0.0, -0.0, -0.04379, 0.0, 0.0, -0.0041, -0.0, -0.0, 0.02697, 0.0, 0.0, -0.00582, 0.0, -0.0, 0.0079, 0.0, 0.0, 0.08597, -0.0, -0.0, -0.0, 0.0, 0.0, 0.03989, -0.0, 0.0, 0.03989, -0.0, -0.0, -0.03989, 0.0, -0.0, -0.03989], [-0.0, 0.0, 0.0, -0.0, 0.0, -0.04427, -0.0, 0.0, -0.0, 0.0, 0.0, 0.04427, 0.0, 0.0, -0.01394, 0.0, 0.0, -0.03147, 0.0, -0.0, 0.00552, 0.0, -0.0, -0.02147, 0.0, 0.0, 0.06387, -0.0, -0.0, -0.04674, -0.0, 0.0, -0.02147, -0.0, 0.0, 0.00552, -0.0, 0.0, -0.03147, -0.0, 0.0, -0.01394, 0.0, 0.0, 0.06387, -0.0, -0.0, -0.04674, -0.0, -0.0, 0.01394, -0.0, -0.0, 0.03147, -0.0, 0.0, -0.00552, -0.0, 0.0, 0.02147, 0.0, 0.0, 0.04674, -0.0, -0.0, -0.06387, 0.0, -0.0, 0.02147, 0.0, -0.0, -0.00552, 0.0, -0.0, 0.03147, 0.0, -0.0, 0.01394, -0.0, -0.0, -0.06387, 0.0, 0.0, 0.04674, 0.0, 0.0, 0.0, 0.0, 0.0, -0.04485, -0.0, 0.0, -0.04485, -0.0, -0.0, 0.04485, 0.0, -0.0, 0.04485]]), "Egy":array([[-0.0, -0.0, 0.07788, -0.0, -0.0, 0.0, 0.0, 0.0, -0.07788, 0.0, -0.0, -0.0, -0.0, 0.0, 0.04549, -0.0, 0.0, 0.00574, 0.0, -0.0, -0.0222, 0.0, -0.0, 0.00656, 0.0, -0.0, -0.07797, -0.0, -0.0, -0.0094, -0.0, 0.0, -0.00656, -0.0, 0.0, 0.0222, 0.0, 0.0, -0.00574, 0.0, 0.0, -0.04549, -0.0, 0.0, 0.07797, -0.0, 0.0, 0.0094, 0.0, -0.0, -0.04549, 0.0, -0.0, -0.00574, -0.0, 0.0, 0.0222, -0.0, 0.0, -0.00656, 0.0, 0.0, 0.0094, -0.0, 0.0, 0.07797, 0.0, -0.0, 0.00656, 0.0, -0.0, -0.0222, -0.0, -0.0, 0.00574, -0.0, -0.0, 0.04549, 0.0, -0.0, -0.07797, 0.0, -0.0, -0.0094, -0.0, -0.0, -0.0, 0.0, -0.0, 0.04368, 0.0, -0.0, -0.04368, -0.0, 0.0, -0.04368, -0.0, 0.0, 0.04368], [0.0, 0.0, -0.02256, -0.0, 0.0, 0.0, -0.0, -0.0, 0.02256, 0.0, 0.0, -0.0, 0.0, -0.0, -0.02219, 0.0, -0.0, 0.00132, -0.0, 0.0, -0.02342, 0.0, -0.0, -0.02998, -0.0, -0.0, -0.04983, 0.0, -0.0, 0.07476, -0.0, 0.0, 0.02998, 0.0, -0.0, 0.02342, 0.0, -0.0, -0.00132, -0.0, 0.0, 0.02219, -0.0, 0.0, 0.04983, -0.0, 0.0, -0.07476, -0.0, 0.0, 0.02219, -0.0, 0.0, -0.00132, 0.0, -0.0, 0.02342, -0.0, 0.0, 0.02998, -0.0, 0.0, -0.07476, 0.0, 0.0, 0.04983, 0.0, -0.0, -0.02998, -0.0, 0.0, -0.02342, -0.0, 0.0, 0.00132, 0.0, -0.0, -0.02219, 0.0, -0.0, -0.04983, 0.0, -0.0, 0.07476, 0.0, 0.0, 0.0, 0.0, -0.0, -0.03722, 0.0, 0.0, 0.03722, -0.0, 0.0, 0.03722, -0.0, -0.0, -0.03722]]), "Egx+y":array([[-0.0, -0.0, 0.0449, 0.0, 0.0, -0.0449, -0.0, 0.0, -0.0449, -0.0, -0.0, 0.0449, 0.0, -0.0, 0.02332, -0.0, -0.0, 0.01604, 0.0, -0.0, -0.01604, -0.0, 0.0, -0.02332, -0.0, 0.0, -0.04215, 0.0, -0.0, 0.04215, -0.0, -0.0, -0.02999, -0.0, 0.0, 0.01182, 0.0, 0.0, 0.01182, -0.0, 0.0, -0.02999, -0.0, 0.0, 0.05578, 0.0, -0.0, 0.05578, -0.0, 0.0, -0.02332, 0.0, 0.0, -0.01604, -0.0, 0.0, 0.01604, 0.0, -0.0, 0.02332, -0.0, 0.0, -0.04215, 0.0, -0.0, 0.04215, 0.0, 0.0, 0.02999, 0.0, -0.0, -0.01182, -0.0, -0.0, -0.01182, 0.0, -0.0, 0.02999, 0.0, -0.0, -0.05578, -0.0, 0.0, -0.05578, -0.0, -0.0, -0.0, -0.0, -0.0, 0.0, 0.0, -0.0, -0.05202, 0.0, 0.0, -0.0, -0.0, 0.0, 0.05202], [-0.0, 0.0, 0.04083, 0.0, 0.0, -0.04083, 0.0, -0.0, -0.04083, -0.0, -0.0, 0.04083, 0.0, 0.0, 0.00536, -0.0, 0.0, -0.01939, 0.0, -0.0, 0.01939, 0.0, 0.0, -0.00536, -0.0, -0.0, 0.08482, -0.0, 0.0, -0.08482, 0.0, 0.0, -0.03311, -0.0, 0.0, -0.02755, 0.0, -0.0, -0.02755, 0.0, 0.0, -0.03311, 0.0, -0.0, -0.0071, -0.0, -0.0, -0.0071, -0.0, -0.0, -0.00536, 0.0, -0.0, 0.01939, -0.0, 0.0, -0.01939, -0.0, -0.0, 0.00536, 0.0, -0.0, 0.08482, 0.0, 0.0, -0.08482, -0.0, -0.0, 0.03311, 0.0, -0.0, 0.02755, -0.0, 0.0, 0.02755, -0.0, -0.0, 0.03311, -0.0, 0.0, 0.0071, 0.0, 0.0, 0.0071, -0.0, -0.0, -0.0, 0.0, 0.0, -0.0, -0.0, 0.0, -0.06369, -0.0, -0.0, 0.0, 0.0, -0.0, 0.06369]]), "Egx-y":array([[-0.0, -0.0, 0.04425, 0.0, -0.0, 0.04425, -0.0, 0.0, -0.04425, -0.0, 0.0, -0.04425, -0.0, 0.0, 0.03071, 0.0, 0.0, -0.01162, 0.0, 0.0, -0.01162, -0.0, -0.0, 0.03071, -0.0, -0.0, -0.05256, -0.0, -0.0, -0.05256, -0.0, 0.0, 0.02267, -0.0, -0.0, 0.01901, 0.0, 0.0, -0.01901, 0.0, 0.0, -0.02267, 0.0, -0.0, 0.04512, 0.0, 0.0, -0.04512, 0.0, -0.0, -0.03071, -0.0, -0.0, 0.01162, -0.0, -0.0, 0.01162, 0.0, 0.0, -0.03071, 0.0, 0.0, 0.05256, 0.0, 0.0, 0.05256, 0.0, -0.0, -0.02267, 0.0, 0.0, -0.01901, -0.0, -0.0, 0.01901, -0.0, -0.0, 0.02267, -0.0, 0.0, -0.04512, -0.0, -0.0, 0.04512, -0.0, -0.0, -0.0, 0.0, -0.0, 0.04812, 0.0, -0.0, -0.0, -0.0, 0.0, -0.04812, -0.0, 0.0, 0.0], [0.0, 0.0, -0.01713, 0.0, 0.0, -0.01713, -0.0, -0.0, 0.01713, -0.0, -0.0, 0.01713, -0.0, 0.0, -0.02943, 0.0, -0.0, -0.01254, -0.0, 0.0, -0.01254, -0.0, -0.0, -0.02943, -0.0, -0.0, 0.03267, 0.0, 0.0, 0.03267, -0.0, -0.0, 0.00171, 0.0, -0.0, 0.03137, -0.0, 0.0, -0.03137, 0.0, -0.0, -0.00171, 0.0, -0.0, 0.09245, 0.0, -0.0, -0.09245, 0.0, -0.0, 0.02943, -0.0, 0.0, 0.01254, 0.0, -0.0, 0.01254, 0.0, 0.0, 0.02943, -0.0, -0.0, -0.03267, 0.0, 0.0, -0.03267, 0.0, 0.0, -0.00171, -0.0, 0.0, -0.03137, 0.0, -0.0, 0.03137, -0.0, 0.0, 0.00171, -0.0, 0.0, -0.09245, -0.0, 0.0, 0.09245, 0.0, 0.0, 0.0, -0.0, -0.0, -0.06539, -0.0, -0.0, 0.0, 0.0, 0.0, 0.06539, 0.0, 0.0, -0.0]]), "Eux":array([[0.0242, -0.0, 0.0, -0.00343, -0.0, -0.0, 0.0242, -0.0, -0.0, -0.00343, 0.0, -0.0, 0.02067, 0.00906, 0.0, 0.00879, 0.02109, 0.0, -0.01712, 0.02554, 0.0, -0.00295, 0.00742, 0.0, -0.05303, 0.03456, 0.0, 0.00317, 0.05555, 0.0, -0.00295, -0.00742, -0.0, -0.01712, -0.02554, -0.0, 0.00879, -0.02109, -0.0, 0.02067, -0.00906, -0.0, -0.05303, -0.03456, -0.0, 0.00317, -0.05555, -0.0, 0.02067, 0.00906, 0.0, 0.00879, 0.02109, 0.0, -0.01712, 0.02554, 0.0, -0.00295, 0.00742, 0.0, 0.00317, 0.05555, 0.0, -0.05303, 0.03456, 0.0, -0.00295, -0.00742, -0.0, -0.01712, -0.02554, -0.0, 0.00879, -0.02109, -0.0, 0.02067, -0.00906, -0.0, -0.05303, -0.03456, -0.0, 0.00317, -0.05555, -0.0, 0.01629, -0.0, -0.0, 0.02856, 0.00511, 0.0, 0.02856, -0.00511, -0.0, 0.02856, 0.00511, 0.0, 0.02856, -0.00511, -0.0], [-0.00426, -0.0, -0.0, -0.00578, -0.0, -0.0, -0.00426, 0.0, -0.0, -0.00578, -0.0, 0.0, -0.00406, 0.00025, 0.0, -0.01672, 0.01613, 0.0, 0.00845, -0.01363, -0.0, -0.00204, -0.0066, -0.0, 0.06419, -0.01917, -0.0, -0.02313, 0.09728, 0.0, -0.00204, 0.0066, 0.0, 0.00845, 0.01363, 0.0, -0.01672, -0.01613, -0.0, -0.00406, -0.00025, -0.0, 0.06419, 0.01917, 0.0, -0.02313, -0.09728, -0.0, -0.00406, 0.00025, -0.0, -0.01672, 0.01613, 0.0, 0.00845, -0.01363, -0.0, -0.00204, -0.0066, -0.0, -0.02313, 0.09728, 0.0, 0.06419, -0.01917, -0.0, -0.00204, 0.0066, 0.0, 0.00845, 0.01363, 0.0, -0.01672, -0.01613, -0.0, -0.00406, -0.00025, 0.0, 0.06419, 0.01917, 0.0, -0.02313, -0.09728, -0.0, -0.00232, 0.0, -0.0, -0.02679, -0.00456, -0.0, -0.02679, 0.00456, 0.0, -0.02679, -0.00456, -0.0, -0.02679, 0.00456, 0.0]]), "Euy":array([[-0.0, -0.00616, 0.0, 0.0, 0.02383, -0.0, 0.0, -0.00616, -0.0, 0.0, 0.02383, 0.0, 0.00787, -0.0028, -0.0, 0.02561, -0.01649, -0.0, 0.02201, 0.00959, -0.0, 0.00955, 0.02133, -0.0, 0.05991, 0.00329, -0.0, 0.03442, -0.04987, -0.0, -0.00955, 0.02133, 0.0, -0.02201, 0.00959, 0.0, -0.02561, -0.01649, 0.0, -0.00787, -0.0028, 0.0, -0.05991, 0.00329, 0.0, -0.03442, -0.04987, 0.0, 0.00787, -0.0028, -0.0, 0.02561, -0.01649, -0.0, 0.02201, 0.00959, -0.0, 0.00955, 0.02133, -0.0, 0.03442, -0.04987, -0.0, 0.05991, 0.00329, -0.0, -0.00955, 0.02133, 0.0, -0.02201, 0.00959, 0.0, -0.02561, -0.01649, 0.0, -0.00787, -0.0028, 0.0, -0.05991, 0.00329, 0.0, -0.03442, -0.04987, 0.0, 0.0, 0.01705, -0.0, 0.00648, 0.02306, -0.0, -0.00648, 0.02306, 0.0, 0.00648, 0.02306, -0.0, -0.00648, 0.02306, 0.0], [-0.0, 0.00451, -0.0, -0.0, -0.00369, 0.0, 0.0, 0.00451, -0.0, -0.0, -0.00369, -0.0, -0.01128, -0.00017, 0.0, -0.02059, 0.01086, 0.0, 0.01292, -0.02092, -0.0, -0.00104, -0.00636, 0.0, 0.08901, -0.02823, -0.0, -0.02531, 0.05969, 0.0, 0.00104, -0.00636, 0.0, -0.01292, -0.02092, 0.0, 0.02059, 0.01086, -0.0, 0.01128, -0.00017, -0.0, -0.08901, -0.02823, 0.0, 0.02531, 0.05969, -0.0, -0.01128, -0.00017, 0.0, -0.02059, 0.01086, 0.0, 0.01292, -0.02092, -0.0, -0.00104, -0.00636, -0.0, -0.02531, 0.05969, 0.0, 0.08901, -0.02823, -0.0, 0.00104, -0.00636, -0.0, -0.01292, -0.02092, 0.0, 0.02059, 0.01086, -0.0, 0.01128, -0.00017, -0.0, -0.08901, -0.02823, 0.0, 0.02531, 0.05969, -0.0, -0.0, -0.00452, 0.0, -0.01341, -0.01611, 0.0, 0.01341, -0.01611, -0.0, -0.01341, -0.01611, 0.0, 0.01341, -0.01611, -0.0]]), "Eux+y":array([[-0.01737, 0.0042, 0.0, 0.0042, -0.01737, -0.0, -0.01737, 0.0042, 0.0, 0.0042, -0.01737, -0.0, -0.02095, -0.00463, 0.0, -0.02589, -0.00336, 0.0, -0.00336, -0.02589, -0.0, -0.00463, -0.02095, -0.0, -0.00463, -0.02842, -0.0, -0.02842, -0.00463, 0.0, 0.00882, -0.00991, -0.0, 0.02789, 0.01171, -0.0, 0.01171, 0.02789, 0.0, -0.00991, 0.00882, 0.0, 0.0796, 0.0227, -0.0, 0.0227, 0.0796, 0.0, -0.02095, -0.00463, 0.0, -0.02589, -0.00336, 0.0, -0.00336, -0.02589, -0.0, -0.00463, -0.02095, -0.0, -0.02842, -0.00463, 0.0, -0.00463, -0.02842, -0.0, 0.00882, -0.00991, -0.0, 0.02789, 0.01171, -0.0, 0.01171, 0.02789, 0.0, -0.00991, 0.00882, 0.0, 0.0796, 0.0227, -0.0, 0.0227, 0.0796, 0.0, -0.0119, -0.0119, -0.0, -0.02198, -0.02198, 0.0, -0.01376, -0.01376, -0.0, -0.02198, -0.02198, -0.0, -0.01376, -0.01376, -0.0], [-0.00376, 0.00074, -0.0, 0.00074, -0.00376, 0.0, -0.00376, 0.00074, -0.0, 0.00074, -0.00376, -0.0, -0.01183, -0.00047, 0.0, -0.03182, 0.02114, 0.0, 0.02114, -0.03182, -0.0, -0.00047, -0.01183, -0.0, 0.13154, -0.04212, -0.0, -0.04212, 0.13154, 0.0, -0.0007, 0.00264, 0.0, -0.00507, -0.00204, 0.0, -0.00204, -0.00507, -0.0, 0.00264, -0.0007, -0.0, -0.02545, -0.00399, 0.0, -0.00399, -0.02545, -0.0, -0.01183, -0.00047, 0.0, -0.03182, 0.02114, 0.0, 0.02114, -0.03182, -0.0, -0.00047, -0.01183, -0.0, -0.04212, 0.13154, 0.0, 0.13154, -0.04212, -0.0, -0.0007, 0.00264, 0.0, -0.00507, -0.00204, 0.0, -0.00204, -0.00507, -0.0, 0.00264, -0.0007, -0.0, -0.02545, -0.00399, 0.0, -0.00399, -0.02545, -0.0, -0.00254, -0.00254, -0.0, -0.02296, -0.02296, -0.0, -0.00977, -0.00977, 0.0, -0.02296, -0.02296, 0.0, -0.00977, -0.00977, -0.0]]), "Eux-y":array([[0.01769, 0.00286, 0.0, -0.00286, -0.01769, -0.0, 0.01769, 0.00286, 0.0, -0.00286, -0.01769, -0.0, 0.00957, 0.00893, 0.0, -0.01215, 0.02803, 0.0, -0.02803, 0.01215, -0.0, -0.00893, -0.00957, -0.0, -0.07953, 0.02286, -0.0, -0.02286, 0.07953, 0.0, 0.00474, -0.02084, -0.0, 0.00358, -0.02491, -0.0, 0.02491, -0.00358, 0.0, 0.02084, -0.00474, 0.0, 0.00437, -0.02671, -0.0, 0.02671, -0.00437, 0.0, 0.00957, 0.00893, 0.0, -0.01215, 0.02803, 0.0, -0.02803, 0.01215, -0.0, -0.00893, -0.00957, -0.0, -0.02286, 0.07953, 0.0, -0.07953, 0.02286, -0.0, 0.00474, -0.02084, -0.0, 0.00358, -0.02491, -0.0, 0.02491, -0.00358, 0.0, 0.02084, -0.00474, 0.0, 0.00437, -0.02671, -0.0, 0.02671, -0.00437, 0.0, 0.01243, -0.01243, -0.0, 0.01515, -0.01515, 0.0, 0.02401, -0.02401, -0.0, 0.01515, -0.01515, 0.0, 0.02401, -0.02401, 0.0], [-0.0028, 0.00049, 0.0, -0.00049, 0.0028, -0.0, -0.0028, 0.00049, -0.0, -0.00049, 0.0028, 0.0, 0.00371, 0.00052, 0.0, -0.00102, 0.00308, 0.0, -0.00308, 0.00102, -0.0, -0.00052, -0.00371, -0.0, -0.02714, 0.00177, -0.0, -0.00177, 0.02714, 0.0, -0.00238, 0.01292, 0.0, 0.01923, 0.03074, 0.0, -0.03074, -0.01923, -0.0, -0.01292, 0.00238, -0.0, 0.13147, 0.04114, 0.0, -0.04114, -0.13147, -0.0, 0.00371, 0.00052, 0.0, -0.00102, 0.00308, -0.0, -0.00308, 0.00102, 0.0, -0.00052, -0.00371, -0.0, -0.00177, 0.02714, 0.0, -0.02714, 0.00177, -0.0, -0.00238, 0.01292, 0.0, 0.01923, 0.03074, 0.0, -0.03074, -0.01923, -0.0, -0.01292, 0.00238, -0.0, 0.13147, 0.04114, 0.0, -0.04114, -0.13147, -0.0, -0.00379, 0.00379, 0.0, -0.00757, 0.00757, -0.0, -0.02704, 0.02704, -0.0, -0.00757, 0.00757, 0.0, -0.02704, 0.02704, 0.0]])}, database_path = 'HEM_pdb_scsd_20221109.pkl', maxdist = 1.75)

model_objs_dict['CLA_pdb'] = scsd_model('CLA_pdb', array([['2.4275', '-2.4651', '0.0', 'C'], ['-2.4275', '-2.4651', '0.0', 'C'], ['-2.4275', '2.4651', '-0.0', 'C'], ['2.4275', '2.4651', '-0.0', 'C'], ['1.0646', '-2.939', '0.0', 'C'], ['-1.0646', '-2.939', '0.0', 'C'], ['-2.839', '-1.0585', '0.0', 'C'], ['-4.2374', '-0.7264', '0.0', 'C'], ['-4.2374', '0.7264', '-0.0', 'C'], ['-2.839', '1.0585', '-0.0', 'C'], ['-1.0646', '2.939', '-0.0', 'C'], ['1.0646', '2.939', '-0.0', 'C'], ['2.839', '1.0585', '-0.0', 'C'], ['4.2374', '0.7264', '-0.0', 'C'], ['4.2374', '-0.7264', '0.0', 'C'], ['2.839', '-1.0585', '0.0', 'C'], ['0.0', '0.0', '0.0', 'MG'], ['0.0', '-2.1436', '0.0', 'N'], ['-1.9928', '-0.0', '0.0', 'N'], ['-0.0', '2.1436', '-0.0', 'N'], ['1.9928', '0.0', '0.0', 'N']]), 'D2h' , pca = {"Ag":array([[-0.02049, -0.01573, -0.0, 0.02049, -0.01573, 0.0, 0.02049, 0.01573, 0.0, -0.02049, 0.01573, -0.0, -0.00878, -0.05294, -0.0, 0.00878, -0.05294, -0.0, 0.03831, -0.00924, 0.0, 0.0844, 0.00378, 0.0, 0.0844, -0.00378, 0.0, 0.03831, 0.00924, 0.0, 0.00878, 0.05294, 0.0, -0.00878, 0.05294, 0.0, -0.03831, 0.00924, -0.0, -0.0844, -0.00378, -0.0, -0.0844, 0.00378, -0.0, -0.03831, -0.00924, -0.0, -0.0, -0.0, -0.0, -0.0, -0.04774, -0.0, 0.04549, -0.0, 0.0, 0.0, 0.04774, 0.0, -0.04549, 0.0, -0.0], [-0.00974, 0.0299, -0.0, 0.00974, 0.0299, 0.0, 0.00974, -0.0299, -0.0, -0.00974, -0.0299, -0.0, 0.01951, 0.07059, 0.0, -0.01951, 0.07059, 0.0, 0.0264, -0.00982, 0.0, 0.05365, 0.02001, 0.0, 0.05365, -0.02001, 0.0, 0.0264, 0.00982, 0.0, -0.01951, -0.07059, -0.0, 0.01951, -0.07059, -0.0, -0.0264, 0.00982, -0.0, -0.05365, -0.02001, -0.0, -0.05365, 0.02001, -0.0, -0.0264, -0.00982, -0.0, -0.0, -0.0, -0.0, 0.0, 0.09119, 0.0, 0.02859, -0.0, 0.0, -0.0, -0.09119, -0.0, -0.02859, 0.0, -0.0]]), "B1g":array([[0.02221, -0.02796, -0.0, 0.02221, 0.02796, 0.0, -0.02221, 0.02796, 0.0, -0.02221, -0.02796, -0.0, 0.04679, -0.01707, 0.0, 0.04679, 0.01707, -0.0, 0.04268, 0.03951, -0.0, 0.04942, 0.00399, 0.0, -0.04942, 0.00399, 0.0, -0.04268, 0.03951, 0.0, -0.04679, 0.01707, -0.0, -0.04679, -0.01707, 0.0, -0.04268, -0.03951, 0.0, -0.04942, -0.00399, -0.0, 0.04942, -0.00399, -0.0, 0.04268, -0.03951, -0.0, 0.0, 0.0, 0.0, 0.02788, 0.0, -0.0, 0.0, 0.08559, 0.0, -0.02788, 0.0, 0.0, 0.0, -0.08559, -0.0], [-0.03985, 0.04227, 0.0, -0.03985, -0.04227, 0.0, 0.03985, -0.04227, -0.0, 0.03985, 0.04227, -0.0, 0.00441, 0.0063, 0.0, 0.00441, -0.0063, -0.0, 0.02429, -0.05893, -0.0, 0.03088, 0.01334, -0.0, -0.03088, 0.01334, -0.0, -0.02429, -0.05893, -0.0, -0.00441, -0.0063, -0.0, -0.00441, 0.0063, 0.0, -0.02429, 0.05893, 0.0, -0.03088, -0.01334, 0.0, 0.03088, -0.01334, 0.0, 0.02429, 0.05893, 0.0, 0.0, 0.0, 0.0, -0.10137, 0.0, -0.0, 0.0, 0.07229, -0.0, 0.10137, 0.0, 0.0, 0.0, -0.07229, 0.0]]), "B2g":array([[-0.0, 0.0, -0.02511, 0.0, -0.0, -0.02511, 0.0, 0.0, 0.02511, -0.0, 0.0, 0.02511, -0.0, -0.0, 0.07522, 0.0, -0.0, 0.07522, -0.0, 0.0, -0.03625, 0.0, 0.0, -0.08655, 0.0, -0.0, 0.08655, -0.0, -0.0, 0.03625, 0.0, 0.0, -0.07522, -0.0, 0.0, -0.07522, 0.0, -0.0, 0.03625, -0.0, -0.0, 0.08655, -0.0, 0.0, -0.08655, 0.0, 0.0, -0.03625, 0.0, 0.0, 0.0, 0.0, 0.0, -0.05375, 0.0, -0.0, -0.0, -0.0, -0.0, 0.05375, -0.0, 0.0, 0.0], [-0.0, -0.0, 0.04826, -0.0, -0.0, 0.04826, -0.0, 0.0, -0.04826, 0.0, 0.0, -0.04826, -0.0, 0.0, 0.01274, 0.0, 0.0, 0.01274, 0.0, 0.0, 0.03426, -0.0, -0.0, 0.04862, -0.0, 0.0, -0.04862, -0.0, -0.0, -0.03426, 0.0, -0.0, -0.01274, -0.0, -0.0, -0.01274, -0.0, -0.0, -0.03426, 0.0, 0.0, -0.04862, 0.0, -0.0, 0.04862, 0.0, 0.0, 0.03426, -0.0, -0.0, -0.0, 0.0, 0.0, -0.21224, 0.0, -0.0, 0.0, -0.0, -0.0, 0.21224, -0.0, 0.0, -0.0]]), "B3g":array([[-0.0, 0.0, -0.05615, 0.0, 0.0, 0.05615, -0.0, -0.0, 0.05615, -0.0, -0.0, -0.05615, -0.0, -0.0, -0.03437, 0.0, -0.0, 0.03437, 0.0, -0.0, 0.02896, 0.0, 0.0, -0.08272, 0.0, -0.0, -0.08272, 0.0, 0.0, 0.02896, 0.0, 0.0, 0.03437, -0.0, 0.0, -0.03437, -0.0, 0.0, -0.02896, -0.0, -0.0, 0.08272, -0.0, 0.0, 0.08272, -0.0, -0.0, -0.02896, -0.0, -0.0, -0.0, 0.0, 0.0, 0.0, -0.0, -0.0, 0.09559, -0.0, -0.0, -0.0, 0.0, 0.0, -0.09559], [0.0, -0.0, -0.06902, -0.0, -0.0, 0.06902, -0.0, 0.0, 0.06902, 0.0, 0.0, -0.06902, 0.0, -0.0, -0.07707, -0.0, -0.0, 0.07707, 0.0, 0.0, -0.03464, 0.0, 0.0, -0.00616, 0.0, -0.0, -0.00616, 0.0, -0.0, -0.03464, -0.0, 0.0, 0.07707, 0.0, 0.0, -0.07707, -0.0, -0.0, 0.03464, -0.0, -0.0, 0.00616, -0.0, 0.0, 0.00616, -0.0, 0.0, 0.03464, -0.0, -0.0, -0.0, -0.0, -0.0, -0.0, -0.0, 0.0, -0.12619, 0.0, 0.0, 0.0, 0.0, -0.0, 0.12619]]), "Au":array([[-0.0, -0.0, -0.10542, 0.0, 0.0, 0.10542, -0.0, 0.0, -0.10542, -0.0, -0.0, 0.10542, 0.0, 0.0, -0.05249, -0.0, 0.0, 0.05249, 0.0, -0.0, 0.0504, 0.0, -0.0, 0.0417, 0.0, -0.0, -0.0417, 0.0, -0.0, -0.0504, -0.0, 0.0, -0.05249, 0.0, 0.0, 0.05249, 0.0, 0.0, 0.0504, 0.0, 0.0, 0.0417, 0.0, 0.0, -0.0417, 0.0, 0.0, -0.0504, 0.0, 0.0, 0.0, 0.0, 0.0, -0.0, 0.0, 0.0, -0.0, -0.0, 0.0, 0.0, 0.0, 0.0, 0.0], [-0.0, -0.0, 0.05564, -0.0, -0.0, -0.05564, -0.0, -0.0, 0.05564, 0.0, 0.0, -0.05564, -0.0, -0.0, 0.02655, -0.0, -0.0, -0.02655, -0.0, 0.0, 0.03007, -0.0, -0.0, 0.13775, -0.0, -0.0, -0.13775, -0.0, 0.0, -0.03007, 0.0, -0.0, 0.02655, 0.0, -0.0, -0.02655, -0.0, -0.0, 0.03007, -0.0, 0.0, 0.13775, -0.0, 0.0, -0.13775, -0.0, -0.0, -0.03007, -0.0, -0.0, -0.0, 0.0, -0.0, -0.0, -0.0, -0.0, -0.0, 0.0, -0.0, 0.0, -0.0, -0.0, 0.0]]), "B1u":array([[-0.0, -0.0, 0.00375, -0.0, -0.0, 0.00375, 0.0, -0.0, 0.00375, 0.0, -0.0, 0.00375, 0.0, -0.0, 0.01475, 0.0, -0.0, 0.01475, -0.0, 0.0, 0.01364, -0.0, -0.0, 0.09286, -0.0, -0.0, 0.09286, -0.0, -0.0, 0.01364, -0.0, -0.0, 0.01475, -0.0, -0.0, 0.01475, -0.0, -0.0, 0.01364, -0.0, -0.0, 0.09286, -0.0, -0.0, 0.09286, -0.0, 0.0, 0.01364, 0.0, 0.0, -0.34933, 0.0, -0.0, -0.04077, -0.0, 0.0, -0.03458, 0.0, -0.0, -0.04077, -0.0, 0.0, -0.03458], [-0.0, 0.0, -0.0284, 0.0, -0.0, -0.0284, -0.0, -0.0, -0.0284, -0.0, -0.0, -0.0284, 0.0, -0.0, -0.06715, -0.0, -0.0, -0.06715, -0.0, 0.0, 0.0078, -0.0, -0.0, 0.0923, -0.0, -0.0, 0.0923, -0.0, -0.0, 0.0078, -0.0, -0.0, -0.06715, 0.0, -0.0, -0.06715, -0.0, -0.0, 0.0078, -0.0, -0.0, 0.0923, -0.0, -0.0, 0.0923, -0.0, 0.0, 0.0078, -0.0, -0.0, 0.09952, 0.0, -0.0, -0.02979, -0.0, 0.0, -0.02914, 0.0, -0.0, -0.02979, -0.0, 0.0, -0.02914]]), "B2u":array([[-0.05513, 0.03479, 0.0, -0.05513, -0.03479, -0.0, -0.05513, 0.03479, 0.0, -0.05513, -0.03479, -0.0, -0.02983, 0.03765, 0.0, -0.02983, -0.03765, -0.0, 0.01959, 0.00506, 0.0, 0.0582, 0.0114, 0.0, 0.0582, -0.0114, -0.0, 0.01959, -0.00506, -0.0, -0.02983, 0.03765, 0.0, -0.02983, -0.03765, -0.0, 0.01959, 0.00506, 0.0, 0.0582, 0.0114, 0.0, 0.0582, -0.0114, -0.0, 0.01959, -0.00506, -0.0, 0.0337, -0.0, -0.0, -0.04994, 0.0, 0.0, 0.04768, -0.0, 0.0, -0.04994, -0.0, -0.0, 0.04768, -0.0, -0.0], [0.00213, 0.0805, 0.0, 0.00213, -0.0805, -0.0, 0.00213, 0.0805, 0.0, 0.00213, -0.0805, -0.0, 0.04878, 0.00206, 0.0, 0.04878, -0.00206, -0.0, -0.0301, -0.05448, -0.0, 0.01834, 0.00911, 0.0, 0.01834, -0.00911, -0.0, -0.0301, 0.05448, 0.0, 0.04878, 0.00206, -0.0, 0.04878, -0.00206, 0.0, -0.0301, -0.05448, -0.0, 0.01834, 0.00911, 0.0, 0.01834, -0.00911, -0.0, -0.0301, 0.05448, 0.0, -0.03965, 0.0, 0.0, -0.0012, 0.0, 0.0, -0.05483, 0.0, 0.0, -0.0012, -0.0, -0.0, -0.05483, 0.0, -0.0]]), "B3u":array([[0.03043, -0.02667, 0.0, -0.03043, -0.02667, -0.0, 0.03043, -0.02667, 0.0, -0.03043, -0.02667, -0.0, 3e-05, 0.00267, 0.0, -3e-05, 0.00267, 0.0, -0.0538, -0.02579, -0.0, -0.06278, 0.06736, -0.0, 0.06278, 0.06736, 0.0, 0.0538, -0.02579, 0.0, 3e-05, 0.00267, -0.0, -3e-05, 0.00267, -0.0, -0.0538, -0.02579, -0.0, -0.06278, 0.06736, -0.0, 0.06278, 0.06736, 0.0, 0.0538, -0.02579, 0.0, -0.0, -0.007, 0.0, 0.0, 0.03772, 0.0, -0.0, -0.06903, 0.0, -0.0, 0.03772, -0.0, 0.0, -0.06903, -0.0], [0.00911, -0.02608, 0.0, -0.00911, -0.02608, -0.0, 0.00911, -0.02608, 0.0, -0.00911, -0.02608, 0.0, -0.00216, -0.03553, -0.0, 0.00216, -0.03553, 0.0, 0.06296, 0.04875, 0.0, -0.06867, -0.01342, -0.0, 0.06867, -0.01342, 0.0, -0.06296, 0.04875, -0.0, -0.00216, -0.03553, -0.0, 0.00216, -0.03553, 0.0, 0.06296, 0.04875, 0.0, -0.06867, -0.01342, -0.0, 0.06867, -0.01342, 0.0, -0.06296, 0.04875, -0.0, 0.0, -0.0149, 0.0, 0.0, 0.04951, 0.0, -0.0, 0.01738, 0.0, -0.0, 0.04951, -0.0, 0.0, 0.01738, -0.0]])}, database_path = 'CLA_scsd_20221108.pkl', maxdist = 2.1)
#[y.gen_assign_mat() for x,y in model_objs_dict.items()]
scsd_model_dict = model_objs_dict

#'bodipy'            : scsd_model('bodipy', model_bodipy, 'C2v', pca = pca_bodipy, database_path='bodipy_scsd_20210702_s.pkl'),
#'porphyrin'         : scsd_model('porphyrin', model_porphyrin, 'D4h', pca = pca_porphyrin, database_path='porphyrin_scsd_202203.pkl'),
#'carbamazepine'    : scsd_model('carbamazepine', model_carbamazepine, 'C2v', pca = pca_carbamazepine, database_path = 'carbemazepine_2_scsd_20210628_s.pkl'),
#'curcumin'             : scsd_model('curcumin', model_curcumin, 'C2v', pca = pca_curcumin, database_path = 'cur_scsd_20210625_s.pkl'),
#'salen'            : scsd_model('salen', model_salen, 'C2v', pca = pca_salen, database_path='salen_scsd_20210625_s.pkl'),
#'phthalocyanine'    : scsd_model('phthalocyanine', model_phthalocyanine_full, 'D4h', pca = pca_phthalocyanine_full, database_path='phthalocyanine_full_scsd_20210628_s.pkl'),
#'phthalocyanine_full': scsd_model('phthalocyanine', model_phthalocyanine_full, 'D4h', pca = pca_phthalocyanine_full, database_path='phthalocyanine_full_scsd_20210628_s.pkl'),
#'xanthene'             : scsd_model('xanthene', model_xanthene, 'C2v', pca = pca_xanthene, database_path='xanthene_scsd_20210628_s.pkl'),
#'tcnq'                 : scsd_model('tcnq', model_tcnq, 'D2h', pca = pca_tcnq, database_path='tcnq_scsd_20210708_s.pkl'),
#'ttf'               : scsd_model('ttf', model_ttf, 'D2h', pca = pca_ttf, database_path='ttf_scsd_20210628_s.pkl'),
#'pyrene'           : scsd_model('pyrene', model_pyrene, 'D2h', pca = pca_pyrene, database_path='pyrene_scsd_20210628_s.pkl'),
#'naphthalenediimide': scsd_model('naphthalenediimide', model_naphthalenediimide, 'D2h', pca = pca_naphthalenediimide, database_path='naphthalenediimide_scsd_20210628_s.pkl'),
#'dhbq'             : scsd_model('dhbq', model_dhbq, 'D2h', pca = pca_dhbq, database_path = 'dhbq_scsd_20210702_s.pkl'),
#'triptycene'       : scsd_model('triptycene', model_triptycene, 'D3h', pca = pca_triptycene, database_path='triptycene_scsd_20210923_s.pkl'),
#'pyrrole_modified_porphyrin':      scsd_model('pyrrole_modified_porphyrin', model_pyrrole_modified_porphyrin, 'C2v', pca = pca_pyrrole_modified_porphyrin, database_path = 'pyrrole_modified_porphyrin_scsd_20210701_s.pkl'),
#'chlorin'          : scsd_model('chlorin', model_pyrrole_modified_porphyrin, 'C2v', pca = pca_pyrrole_modified_porphyrin, database_path = 'pyrrole_modified_porphyrin_scsd_20210701_s.pkl') ,
#'corrole'          : scsd_model('corrole', model_corrole, 'C2v', pca = pca_corrole, database_path = 'corrole_scsd_20210701_s.pkl'),
#'corrphycene'      : scsd_model('corrphycene', model_corrphycene, 'C2v', pca = pca_corrphycene, database_path = 'corrphycene_scsd_20210701_s.pkl'),
#'norcorrole'       : scsd_model('norcorrole', model_norcorrole, 'D2h', pca = pca_norcorrole, database_path = 'norcorrole_scsd_20210701_s.pkl'),
#'porphycene'       : scsd_model('porphycene', model_porphycene, 'D2h', pca = pca_porphycene, database_path =  'porphycene_scsd_20220406.pkl'),
#'anthracene'       : scsd_model('anthracene', model_anthracene, 'D2h', pca = pca_anthracene, database_path = 'anthracene_scsd_20210716_s.pkl'),
#'benzoquinone'      : scsd_model('benzoquinone', model_benzoquinone, 'D2h', pca = pca_benzoquinone, database_path = 'benzoquinone_scsd_20210702_s.pkl'),
#'naphthalene'       : scsd_model('naphthalene', model_naphthalene, 'D2h', pca = pca_naphthalene, database_path = 'naphthalene_scsd_20210702_s.pkl'),
#'subporphyrin'      : scsd_model('subporphyrin', model_subporphyrin, 'D3h'),
#'cb6'               : scsd_model('cb6', model_cb6, 'D6h', pca = pca_cb6, database_path='cb6_scsd_20211116.pkl'),
#'cyclam'            : scsd_model('cyclam', model_cyclam, 'D2h', pca = pca_cyclam, database_path='cyclam_scsd_20220113.pkl'),
#'cb8'               : scsd_model('cb8', model_cb8, 'D8h', pca = pca_cb8, database_path='cb8_scsd_20220324.pkl')
